Package: ellmer
Title: Chat with Large Language Models
Version: 0.4.2
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-4757-117X")),
    person("Joe", "Cheng", role = "aut"),
    person("Aaron", "Jacobs", role = "aut"),
    person("Garrick", "Aden-Buie", , "garrick@posit.co", role = "aut",
           comment = c(ORCID = "0000-0002-7111-0077")),
    person("Barret", "Schloerke", , "barret@posit.co", role = "aut",
           comment = c(ORCID = "0000-0001-9986-114X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Chat with large language models from a range of providers
    including 'Claude' <https://claude.ai>, 'OpenAI'
    <https://chatgpt.com>, and more. Supports streaming, asynchronous
    calls, tool calling, and structured data extraction.
License: MIT + file LICENSE
URL: https://ellmer.tidyverse.org, https://github.com/tidyverse/ellmer
BugReports: https://github.com/tidyverse/ellmer/issues
Depends: R (>= 4.1)
Imports: cli, coro (>= 1.1.0), glue, httr2 (>= 1.2.3), jsonlite, later
        (>= 1.4.0), lifecycle, promises (>= 1.5.0), R6, rlang (>=
        1.3.0), S7 (>= 0.2.0), tibble, vctrs
Suggests: connectcreds, curl (>= 6.0.1), gargle, gitcreds, jose, knitr,
        magick, openssl, otel (>= 0.2.0), otelsdk (>= 0.2.0),
        paws.common, png, rmarkdown, shiny, shinychat (>= 0.3.0),
        testthat (>= 3.0.0), vcr (>= 2.0.0), withr
VignetteBuilder: knitr
Config/Needs/prices: cli, dplyr, jsonlite, jsonvalidate, stringr,
        tidyr, usethis
Config/Needs/website: tidyverse/tidytemplate, rmarkdown
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/testthat/start-first: chat, provider*
Encoding: UTF-8
Collate: 'utils-S7.R' 'types.R' 'ellmer-package.R' 'tools-def.R'
        'content.R' 'provider.R' 'as-json.R' 'batch-chat.R'
        'chat-structured.R' 'chat-tools-content.R' 'turns.R'
        'chat-tools.R' 'chat-utils.R' 'utils-coro.R' 'chat.R'
        'content-image.R' 'content-pdf.R' 'content-replay.R' 'httr2.R'
        'import-standalone-defer.R' 'import-standalone-obj-type.R'
        'import-standalone-purrr.R' 'import-standalone-types-check.R'
        'interpolate.R' 'live.R' 'otel.R' 'parallel-chat.R' 'params.R'
        'provider-any.R' 'provider-aws.R'
        'provider-openai-compatible.R' 'provider-azure.R'
        'provider-claude-files.R' 'provider-claude-tools.R'
        'provider-claude.R' 'provider-google.R' 'provider-cloudflare.R'
        'provider-databricks.R' 'provider-deepseek.R'
        'provider-github.R' 'provider-google-tools.R'
        'provider-google-upload.R' 'provider-groq.R'
        'provider-huggingface.R' 'provider-lmstudio.R'
        'provider-mistral.R' 'provider-ollama.R'
        'provider-openai-tools.R' 'provider-openai.R'
        'provider-openrouter.R' 'provider-perplexity.R'
        'provider-portkey.R' 'provider-posit.R' 'provider-snowflake.R'
        'provider-vllm.R' 'schema.R' 'stream-controller.R' 'tokens.R'
        'tools-built-in.R' 'tools-def-auto.R' 'utils-auth.R'
        'utils-callbacks.R' 'utils-cat.R' 'utils-merge.R'
        'utils-prettytime.R' 'utils.R' 'zzz.R'
NeedsCompilation: no
Packaged: 2026-07-13 13:22:49 UTC; nic
Author: Hadley Wickham [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4757-117X>),
  Joe Cheng [aut],
  Aaron Jacobs [aut],
  Garrick Aden-Buie [aut] (ORCID:
    <https://orcid.org/0000-0002-7111-0077>),
  Barret Schloerke [aut] (ORCID: <https://orcid.org/0000-0001-9986-114X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2026-07-13 23:00:09 UTC
Built: R 4.6.1; ; 2026-07-13 23:45:13 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",ellmer_dollars)
S3method("[",ellmer_prompt)
S3method("[[",ellmer_dollars)
S3method(Summary,ellmer_dollars)
S3method(format,ellmer_dollars)
S3method(print,Chat)
S3method(print,ellmer_dollars)
S3method(print,ellmer_output)
S3method(print,ellmer_prompt)
export(AssistantPartialTurn)
export(AssistantTurn)
export(Content)
export(ContentImage)
export(ContentImageInline)
export(ContentImageRemote)
export(ContentPDF)
export(ContentText)
export(ContentThinking)
export(ContentToolRequest)
export(ContentToolResult)
export(Provider)
export(SystemTurn)
export(Turn)
export(TypeArray)
export(TypeBasic)
export(TypeEnum)
export(TypeIgnore)
export(TypeJsonSchema)
export(TypeObject)
export(UserTurn)
export(batch_chat)
export(batch_chat_completed)
export(batch_chat_structured)
export(batch_chat_text)
export(chat)
export(chat_anthropic)
export(chat_aws_bedrock)
export(chat_azure_openai)
export(chat_claude)
export(chat_cloudflare)
export(chat_databricks)
export(chat_deepseek)
export(chat_github)
export(chat_google_gemini)
export(chat_google_vertex)
export(chat_groq)
export(chat_huggingface)
export(chat_lmstudio)
export(chat_mistral)
export(chat_ollama)
export(chat_openai)
export(chat_openai_compatible)
export(chat_openrouter)
export(chat_perplexity)
export(chat_portkey)
export(chat_snowflake)
export(chat_vllm)
export(claude_file_delete)
export(claude_file_download)
export(claude_file_get)
export(claude_file_list)
export(claude_file_upload)
export(claude_tool_web_fetch)
export(claude_tool_web_search)
export(content_image_file)
export(content_image_plot)
export(content_image_url)
export(content_pdf_file)
export(content_pdf_url)
export(contents_html)
export(contents_markdown)
export(contents_record)
export(contents_replay)
export(contents_text)
export(create_tool_def)
export(df_schema)
export(google_tool_web_fetch)
export(google_tool_web_search)
export(google_upload)
export(has_credentials)
export(interpolate)
export(interpolate_file)
export(interpolate_package)
export(live_browser)
export(live_console)
export(models_anthropic)
export(models_aws_bedrock)
export(models_claude)
export(models_github)
export(models_google_gemini)
export(models_google_vertex)
export(models_lmstudio)
export(models_mistral)
export(models_ollama)
export(models_openai)
export(models_portkey)
export(models_vllm)
export(openai_tool_web_search)
export(parallel_chat)
export(parallel_chat_structured)
export(parallel_chat_text)
export(params)
export(stream_controller)
export(token_usage)
export(tool)
export(tool_annotations)
export(tool_reject)
export(type_array)
export(type_boolean)
export(type_enum)
export(type_from_schema)
export(type_ignore)
export(type_integer)
export(type_number)
export(type_object)
export(type_string)
if (getRversion() < "4.3.0") importFrom("S7", "@")
import(S7)
import(httr2)
import(rlang)
importFrom(R6,R6Class)
importFrom(lifecycle,deprecated)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# ellmer <a href="https://ellmer.tidyverse.org"><img src="man/figures/logo.png" align="right" height="138" alt="ellmer website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/tidyverse/ellmer/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidyverse/ellmer/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

ellmer makes it easy to use large language models (LLM) from R. It
supports a wide variety of LLM providers and implements a rich set of
features including streaming outputs, tool/function calling, structured
data extraction, and more.

ellmer is one of a number of LLM-related packages created by Posit:

- Looking for something similar in python? Check out
  [chatlas](https://github.com/posit-dev/chatlas)!
- Want to evaluate your LLMs? Try
  [vitals](https://vitals.tidyverse.org).
- Need RAG? Take a look at [ragnar](https://ragnar.tidyverse.org).
- Want to make a beautiful LLM powered chatbot? Consider
  [shinychat](https://posit-dev.github.io/shinychat/).
- Working with MCP? Check out
  [mcptools](https://posit-dev.github.io/mcptools/).

## Installation

You can install ellmer from CRAN with:

``` r
install.packages("ellmer")
```

## Providers

ellmer supports a wide variety of model providers. Official providers
are actively maintained, with priority support for bug fixes and new
features. Community providers are contributed and maintained by the
community; contributions to improve them are especially welcome.

### Official providers

- Anthropic’s Claude: `chat_anthropic()`.
- AWS Bedrock: `chat_aws_bedrock()`.
- Azure OpenAI: `chat_azure_openai()`.
- Databricks: `chat_databricks()`.
- DeepSeek: `chat_deepseek()`.
- GitHub model marketplace: `chat_github()`.
- Google Gemini/Vertex AI: `chat_google_gemini()`,
  `chat_google_vertex()`.
- Ollama: `chat_ollama()`.
- OpenAI: `chat_openai()`.
- Posit AI: `chat_posit()`.
- Snowflake Cortex: `chat_snowflake()` and `chat_cortex_analyst()`.

### Community providers

- Cloudflare: `chat_cloudflare()`.
- Groq: `chat_groq()`.
- Hugging Face: `chat_huggingface()`.
- LM Studio: `chat_lmstudio()`.
- Mistral: `chat_mistral()`.
- OpenRouter: `chat_openrouter()`.
- perplexity.ai: `chat_perplexity()`.
- Portkey: `chat_portkey()`.
- VLLM: `chat_vllm()`.

### Provider/model choice

If you’re using ellmer inside an organisation, you may have internal
policies that limit you to models from big cloud providers,
e.g. `chat_azure_openai()`, `chat_aws_bedrock()`, `chat_databricks()`,
or `chat_snowflake()`.

If you’re using ellmer for your own exploration, you’ll have a lot more
freedom, so we have a few recommendations to help you get started:

- `chat_openai()` or `chat_anthropic()` are good places to start.
  `chat_openai()` defaults to **GPT-5.4**, but you can use
  `model = "gpt-5.4-nano"` for a cheaper, faster model.
  `chat_anthropic()` defaults to **Claude Sonnet 4.6**, which we have
  found to be particularly good at writing R code.

- `chat_google_gemini()` is a strong model with a free tier (with the
  downside that [your data is
  used](https://ai.google.dev/gemini-api/terms#unpaid-services) to
  improve the model), making it a great place to start if you don’t want
  to spend any money.

- `chat_ollama()`, which uses [Ollama](https://ollama.com), allows you
  to run models on your own computer. While the biggest models you can
  run locally aren’t as good as the state of the art hosted models, they
  don’t share your data and are effectively free.

### Authentication

Authentication works a little differently depending on the provider. A
few popular ones (including OpenAI and Anthropic) require you to obtain
an API key. We recommend you save it in an environment variable rather
than using it directly in your code, and if you deploy an app or report
that uses ellmer to another system, you’ll need to ensure that this
environment variable is available there, too.

ellmer also automatically detects many of the OAuth or IAM-based
credentials used by the big cloud providers (currently
`chat_azure_openai()`, `chat_aws_bedrock()`, `chat_databricks()`,
`chat_snowflake()`). That includes credentials for these platforms
managed by [Posit
Workbench](https://docs.posit.co/ide/server-pro/user/posit-workbench/managed-credentials/managed-credentials.html)
and [Posit
Connect](https://docs.posit.co/connect/user/oauth-integrations/#adding-oauth-integrations-to-deployed-content).

If you find cases where ellmer cannot detect credentials from one of
these cloud providers, feel free to open an issue; we’re happy to add
more auth mechanisms if needed.

## Using ellmer

You can work with ellmer in several different ways, depending on whether
you are working interactively or programmatically. They all start with
creating a new chat object:

``` r
library(ellmer)

chat <- chat_openai("Be terse", model = "gpt-4o-mini")
```

Chat objects are stateful [R6 objects](https://r6.r-lib.org): they
retain the context of the conversation, so each new query builds on the
previous ones. You call their methods with `$`.

### Interactive chat console

The most interactive and least programmatic way of using ellmer is to
chat directly in your R console or browser with `live_console(chat)` or
`live_browser()`:

``` r
live_console(chat)
#> ╔════════════════════════════════════════════════════════╗
#> ║  Entering chat console. Use """ for multi-line input.  ║
#> ║  Press Ctrl+C to quit.                                 ║
#> ╚════════════════════════════════════════════════════════╝
#> >>> Who were the original creators of R?
#> R was originally created by Ross Ihaka and Robert Gentleman at the University of
#> Auckland, New Zealand.
#>
#> >>> When was that?
#> R was initially released in 1995. Development began a few years prior to that,
#> in the early 1990s.
```

Keep in mind that the chat object retains state, so when you enter the
chat console, any previous interactions with that chat object are still
part of the conversation, and any interactions you have in the chat
console will persist after you exit back to the R prompt. This is true
regardless of which chat function you use.

### Interactive method call

The second most interactive way to chat is to call the `chat()` method:

``` r
chat$chat("What preceding languages most influenced R?")
#> R was primarily influenced by:
#> 
#> 1. **S** - The predecessor to R, which introduced many foundational concepts.
#> 2. **Scheme** - A dialect of Lisp that influenced R's functional programming 
#> aspects.
#> 3. **Fortran** - Influenced R's efficiency and mathematical capabilities.
#> 4. **C** - Impacted R's performance and low-level programming features.
#> 
#> These languages contributed to R's design and functionality in statistics and 
#> data analysis.
```

If you initialize the chat object in the global environment, the `chat`
method will stream the response to the console. When the entire response
is received, it’s also (invisibly) returned as a character vector. This
is useful when you want to see the response as it arrives, but you don’t
want to enter the chat console.

If you want to ask a question about an image, you can pass one or more
additional input arguments using `content_image_file()` and/or
`content_image_url()`:

``` r
chat$chat(
  content_image_url("https://www.r-project.org/Rlogo.png"),
  "Can you explain this logo?"
)
#> The logo features a stylized letter "R" within a circular shape. The design 
#> reflects the programming language R, which is widely used for statistical 
#> analysis and data visualization. The circular element suggests continuity and 
#> completeness, while the bold "R" emphasizes its identity. Overall, the logo 
#> conveys modernity and practicality, aligning with R's functionality in data 
#> science.
```

### Streaming vs capturing

In most circumstances, ellmer will stream the output to the console. You
can take control of this by setting the `echo` argument either when
creating the chat object or when calling `$chat()`. Set `echo = "none"`
to return a string instead:

``` r
my_function <- function() {
  chat <- chat_openai("Be terse", model = "gpt-4o-mini", echo = "none")
  chat$chat("What is 6 times 7?")
}
str(my_function())
#>  'ellmer_output' chr "6 times 7 is 42."
```

If needed, you can manually control this behaviour with the `echo`
argument. This is useful for programming with ellmer when the result is
either not intended for human consumption or when you want to process
the response before displaying it.

## Learning more

ellmer comes with a bunch of vignettes to help you learn more:

- Learn key vocabulary and see example use cases in
  `vignette("ellmer")`.
- Learn how to design your prompt in `vignette("prompt-design")`.
- Learn about tool/function calling in `vignette("tool-calling")`.
- Learn how to extract structured data in `vignette("structured-data")`.
- Learn about streaming and async APIs in `vignette("streaming-async")`.

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = ellmer:::eval_vignette()
)
vcr::setup_knitr()

## ----setup--------------------------------------------------------------------
library(ellmer)

## -----------------------------------------------------------------------------
chat <- chat_openai()
chat$chat_structured(
  "My name is Susan and I'm 13 years old",
  type = type_object(
    name = type_string(),
    age = type_number()
  )
)

## -----------------------------------------------------------------------------
chat <- chat_openai()
chat$chat_structured(
  content_image_url("https://www.r-project.org/Rlogo.png"),
  type = type_object(
    primary_shape = type_string(),
    primary_colour = type_string()
  )
)

## -----------------------------------------------------------------------------
prompts <- list(
  "I go by Alex. 42 years on this planet and counting.",
  "Pleased to meet you! I'm Jamal, age 27.",
  "They call me Li Wei. Nineteen years young.",
  "Fatima here. Just celebrated my 35th birthday last week.",
  "The name's Robert - 51 years old and proud of it.",
  "Kwame here - just hit the big 5-0 this year."
)
type_person <- type_object(
  name = type_string(),
  age = type_number()
)
chat <- chat_openai()
parallel_chat_structured(chat, prompts, type = type_person)

## -----------------------------------------------------------------------------
type_logical_vector <- type_array(type_boolean())
type_integer_vector <- type_array(type_integer())
type_double_vector <- type_array(type_number())
type_character_vector <- type_array(type_string())

## -----------------------------------------------------------------------------
list_of_integers <- type_array(type_integer_vector)

## -----------------------------------------------------------------------------
type_person2 <- type_object(
name = type_string(),
age = type_integer(),
hobbies = type_array(type_string())
)

## -----------------------------------------------------------------------------
type_person3 <- type_object(
  "A person",
  name = type_string("Name"),
  age = type_integer("Age, in years."),
  hobbies = type_array(
    type_string(),
    "List of hobbies. Should be exclusive and brief.",
  )
)

## -----------------------------------------------------------------------------
no_match <- list(
  "I like apples",
  "What time is it?",
  "This cheese is 3 years old",
  "My name is Hadley."
)
parallel_chat_structured(chat, no_match, type = type_person)

## -----------------------------------------------------------------------------
type_person <- type_object(
  name = type_string(required = FALSE),
  age = type_number(required = FALSE)
)
parallel_chat_structured(chat, no_match, type = type_person)

## -----------------------------------------------------------------------------
prompt <- r"(
* John Smith. Age: 30. Height: 180 cm. Weight: 80 kg.
* Jane Doe. Age: 25. Height: 5'5". Weight: 110 lb.
* Jose Rodriguez. Age: 40. Height: 190 cm. Weight: 90 kg.
* June Lee | Age: 35 | Height 175 cm | Weight: 70 kg
)"

## -----------------------------------------------------------------------------
type_people <- type_object(
  name = type_array(type_string()),
  age = type_array(type_integer()),
  height = type_array(type_number("in m")),
  weight = type_array(type_number("in kg"))
)

chat <- chat_openai()
chat$chat_structured(prompt, type = type_people)

## -----------------------------------------------------------------------------
type_people <- type_array(
  type_object(
    name = type_string(),
    age = type_integer(),
    height = type_number("in m"),
    weight = type_number("in kg")
  )
)

chat <- chat_openai()
chat$chat_structured(prompt, type = type_people)

## -----------------------------------------------------------------------------
text <- readLines(system.file(
  "examples/third-party-testing.txt",
  package = "ellmer"
))
# url <- "https://www.anthropic.com/news/third-party-testing"
# html <- rvest::read_html(url)
# text <- rvest::html_text2(rvest::html_element(html, "article"))

type_summary <- type_object(
  "Summary of the article.",
  author = type_string("Name of the article author"),
  topics = type_array(
    type_string(),
    'Array of topics, e.g. ["tech", "politics"]. Should be as specific as possible, and can overlap.'
  ),
  summary = type_string("Summary of the article. One or two paragraphs max"),
  coherence = type_integer(
    "Coherence of the article's key points, 0-100 (inclusive)"
  ),
  persuasion = type_number("Article's persuasion score, 0.0-1.0 (inclusive)")
)

chat <- chat_openai()
data <- chat$chat_structured(text, type = type_summary)
cat(data$summary)

str(data)

## -----------------------------------------------------------------------------
text <- "
  John works at Google in New York. He met with Sarah, the CEO of
  Acme Inc., last week in San Francisco.
"

type_named_entity <- type_object(
  name = type_string("The extracted entity name."),
  type = type_enum(c("person", "location", "organization"), "The entity type"),
  context = type_string("The context in which the entity appears in the text.")
)
type_named_entities <- type_array(type_named_entity)

chat <- chat_openai()
chat$chat_structured(text, type = type_named_entities)

## -----------------------------------------------------------------------------
text <- "
  The product was okay, but the customer service was terrible. I probably
  won't buy from them again.
"

type_sentiment <- type_object(
  "Extract the sentiment scores of a given text. Sentiment scores should sum to 1.",
  positive_score = type_number(
    "Positive sentiment score, ranging from 0.0 to 1.0."
  ),
  negative_score = type_number(
    "Negative sentiment score, ranging from 0.0 to 1.0."
  ),
  neutral_score = type_number(
    "Neutral sentiment score, ranging from 0.0 to 1.0."
  )
)

chat <- chat_openai()
str(chat$chat_structured(text, type = type_sentiment))

## -----------------------------------------------------------------------------
text <- "The new quantum computing breakthrough could revolutionize the tech industry."

type_score <- type_object(
  name = type_enum(
    c(
      "Politics",
      "Sports",
      "Technology",
      "Entertainment",
      "Business",
      "Other"
    ),
    "The category name",
  ),
  score = type_number(
    "The classification score for the category, ranging from 0.0 to 1.0."
  )
)
type_classification <- type_array(
  type_score,
  description = "Array of classification results. The scores should sum to 1."
)

chat <- chat_openai()
data <- chat$chat_structured(text, type = type_classification)
data

## -----------------------------------------------------------------------------
type_characteristics <- type_array(
  type_object(
    name = type_string(),
    value = type_string()
  ),
  description = "All characteristics"
)

text <- "
  The man is tall, with a beard and a scar on his left cheek. He has a deep voice and wears a black leather jacket.
"

chat <- chat_anthropic("Extract all characteristics of supplied character")
chat$chat_structured(text, type = type_characteristics)

## -----------------------------------------------------------------------------
type_asset <- type_object(
  assert_name = type_string(),
  owner = type_string(),
  location = type_string(),
  asset_value_low = type_integer(),
  asset_value_high = type_integer(),
  income_type = type_string(),
  income_low = type_integer(),
  income_high = type_integer(),
  tx_gt_1000 = type_boolean()
)
type_assets <- type_array(type_asset)

chat <- chat_openai()
image <- content_image_file("congressional-assets.png")
data <- chat$chat_structured(image, type = type_assets)
data

## -----------------------------------------------------------------------------
knitr::kable(token_usage())

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = ellmer:::eval_vignette()
)
vcr::setup_knitr()

## ----setup--------------------------------------------------------------------
library(ellmer)

## -----------------------------------------------------------------------------
question <- "
  How can I compute the mean and median of variables a, b, c, and so on,
  all the way up to z, grouped by age and sex.
"

## -----------------------------------------------------------------------------
chat <- chat_anthropic()
chat$chat(question)

## -----------------------------------------------------------------------------
chat <- chat_anthropic(
  system_prompt = "
  You are an expert R programmer who prefers the tidyverse.
"
)
chat$chat(question)

## -----------------------------------------------------------------------------
chat <- chat_anthropic(
  system_prompt = "
  You are an expert R programmer who prefers the tidyverse.
  Just give me the code. I don't want any explanation or sample data.
"
)
chat$chat(question)

## -----------------------------------------------------------------------------
chat <- chat_anthropic(
  system_prompt = "
  You are an expert R programmer who prefers data.table.
  Just give me the code. I don't want any explanation or sample data.
"
)
chat$chat(question)

chat <- chat_anthropic(
  system_prompt = "
  You are an expert R programmer who prefers base R.
  Just give me the code. I don't want any explanation or sample data.
"
)
chat$chat(question)

## -----------------------------------------------------------------------------
chat <- chat_anthropic(
  system_prompt = "
  You are an expert R programmer who prefers the tidyverse.
  Just give me the code. I don't want any explanation or sample data.

  Follow the tidyverse style guide:
  * Spread long function calls across multiple lines.
  * Where needed, always indent function calls with two spaces.
  * Only name arguments that are less commonly used.
  * Always use double quotes for strings.
  * Use the base pipe, `|>`, not the magrittr pipe `%>%`.
"
)
chat$chat(question)

## -----------------------------------------------------------------------------
chat <- chat_anthropic(
  system_prompt = "
  You are an expert R teacher.
  I am a new R user who wants to improve my programming skills.
  Help me understand the code you produce by explaining each function call with
  a brief comment. For more complicated calls, add documentation to each
  argument. Just give me the code. I don't want any explanation or sample data.
"
)
chat$chat(question)

## -----------------------------------------------------------------------------
chat <- chat_anthropic(
  system_prompt = "
  You are an expert R programmer.
  Just give me the code; no explanation in text.
  Use the `.by` argument rather than `group_by()`.
  dplyr 1.1.0 introduced per-operation grouping with the `.by` argument.
  e.g., instead of:

  transactions |>
    group_by(company, year) |>
    mutate(total = sum(revenue))

  write this:
  transactions |>
    mutate(
      total = sum(revenue),
      .by = c(company, year)
    )
"
)
chat$chat(question)

## -----------------------------------------------------------------------------
ingredients <- "
  ¾ cup (150g) dark brown sugar
  2 large eggs
  ¾ cup (165g) sour cream
  ½ cup (113g) unsalted butter, melted
  1 teaspoon vanilla extract
  ¾ teaspoon kosher salt
  ⅓ cup (80ml) neutral oil
  1½ cups (190g) all-purpose flour
  150g plus 1½ teaspoons sugar
"

## -----------------------------------------------------------------------------
instruct_json <- "
  You're an expert baker who also loves JSON. I am going to give you a list of
  ingredients and your job is to return nicely structured JSON. Just return the
  JSON and no other commentary.
"

chat <- chat_openai(instruct_json)
chat$chat(ingredients)

## -----------------------------------------------------------------------------
instruct_weight <- r"(
  Here are some examples of the sort of output I'm looking for:

  ¾ cup (150g) dark brown sugar
  {"name": "dark brown sugar", "quantity": 150, "unit": "g"}

  ⅓ cup (80ml) neutral oil
  {"name": "neutral oil", "quantity": 80, "unit": "ml"}

  2 t ground cinnamon
  {"name": "ground cinnamon", "quantity": 2, "unit": "teaspoon"}
)"

chat <- chat_openai(paste(instruct_json, instruct_weight))
chat$chat(ingredients)

## -----------------------------------------------------------------------------
instruct_weight <- r"(
  * If an ingredient has both weight and volume, extract only the weight:

  ¾ cup (150g) dark brown sugar
  [
    {"name": "dark brown sugar", "quantity": 150, "unit": "g"}
  ]

* If an ingredient only lists a volume, extract that.

  2 t ground cinnamon
  ⅓ cup (80ml) neutral oil
  [
    {"name": "ground cinnamon", "quantity": 2, "unit": "teaspoon"},
    {"name": "neutral oil", "quantity": 80, "unit": "ml"}
  ]
)"

## -----------------------------------------------------------------------------
instruct_unit <- r"(
* If the unit uses a fraction, convert it to a decimal.

  ⅓ cup sugar
  ½ teaspoon salt
  [
    {"name": "dark brown sugar", "quantity": 0.33, "unit": "cup"},
    {"name": "salt", "quantity": 0.5, "unit": "teaspoon"}
  ]

* Quantities are always numbers

  pinch of kosher salt
  [
    {"name": "kosher salt", "quantity": 1, "unit": "pinch"}
  ]

* Some ingredients don't have a unit.
  2 eggs
  1 lime
  1 apple
  [
    {"name": "egg", "quantity": 2},
    {"name": "lime", "quantity": 1},
    {"name", "apple", "quantity": 1}
  ]
)"

## -----------------------------------------------------------------------------
type_ingredient <- type_object(
  name = type_string("Ingredient name"),
  quantity = type_number(),
  unit = type_string("Unit of measurement")
)

type_ingredients <- type_array(type_ingredient)

chat <- chat_openai(c(instruct_json, instruct_weight))
chat$chat_structured(ingredients, type = type_ingredients)

## -----------------------------------------------------------------------------
instruct_weight_input <- r"(
  * If an ingredient has both weight and volume, extract only the weight:

    ¾ cup (150g) dark brown sugar
    [
      {"name": "dark brown sugar", "quantity": 150, "unit": "g", "input": "¾ cup (150g) dark brown sugar"}
    ]

  * If an ingredient only lists a volume, extract that.

    2 t ground cinnamon
    ⅓ cup (80ml) neutral oil
    [
      {"name": "ground cinnamon", "quantity": 2, "unit": "teaspoon", "input": "2 t ground cinnamon"},
      {"name": "neutral oil", "quantity": 80, "unit": "ml", "input": "⅓ cup (80ml) neutral oil"}
    ]
)"

## -----------------------------------------------------------------------------
recipe <- r"(
  In a large bowl, cream together one cup of softened unsalted butter and a
  quarter cup of white sugar until smooth. Beat in an egg and 1 teaspoon of
  vanilla extract. Gradually stir in 2 cups of all-purpose flour until the
  dough forms. Finally, fold in 1 cup of semisweet chocolate chips. Drop
  spoonfuls of dough onto an ungreased baking sheet and bake at 350°F (175°C)
  for 10-12 minutes, or until the edges are lightly browned. Let the cookies
  cool on the baking sheet for a few minutes before transferring to a wire
  rack to cool completely. Enjoy!
)"

## -----------------------------------------------------------------------------
chat <- chat_openai(c(instruct_json, instruct_weight_input))
chat$chat(recipe)

## -----------------------------------------------------------------------------
knitr::kable(token_usage())

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = ellmer:::eval_vignette()
)
vcr::setup_knitr()

## ----setup--------------------------------------------------------------------
library(ellmer)

## -----------------------------------------------------------------------------
chat <- chat_openai(model = "gpt-4o")
chat$chat("How long ago did Neil Armstrong touch down on the moon?")

## -----------------------------------------------------------------------------
#' Gets the current time in the given time zone.
#'
#' @param tz The time zone to get the current time in.
#' @return The current time in the given time zone.
get_current_time <- function(tz = "UTC") {
  format(Sys.time(), tz = tz, usetz = TRUE)
}

## -----------------------------------------------------------------------------
# Fake it for the vignette so we always get the same results
get_current_time <- function(tz = "UTC") {
  format(
    as.POSIXct("2025-06-25 11:53:23", tz = "America/Chicago"),
    tz = tz,
    usetz = TRUE
  )
}

## -----------------------------------------------------------------------------
get_current_time <- tool(
  get_current_time,
  name = "get_current_time",
  description = "Returns the current time.",
  arguments = list(
    tz = type_string(
      "Time zone to display the current time in. Defaults to `\"UTC\"`.",
      required = FALSE
    )
  )
)

## -----------------------------------------------------------------------------
get_current_time()

## -----------------------------------------------------------------------------
chat$register_tool(get_current_time)

## -----------------------------------------------------------------------------
chat$chat("How long ago did Neil Armstrong touch down on the moon?")

## -----------------------------------------------------------------------------
chat

## -----------------------------------------------------------------------------
get_weather <- tool(
  function(cities) {
    raining <- c(London = "heavy", Houston = "none", Chicago = "overcast")
    temperature <- c(London = "cool", Houston = "hot", Chicago = "warm")
    wind <- c(London = "strong", Houston = "weak", Chicago = "strong")

    data.frame(
      city = cities,
      raining = unname(raining[cities]),
      temperature = unname(temperature[cities]),
      wind = unname(wind[cities])
    )
  },
  name = "get_weather",
  description = "
    Report on weather conditions in multiple cities. For efficiency, request
    all weather updates using a single tool call
  ",
  arguments = list(
    cities = type_array(type_string(), "City names")
  )
)

## -----------------------------------------------------------------------------
chat <- chat_openai()
chat$register_tool(get_weather)
chat$chat("Give me a weather update for London and Chicago")

## -----------------------------------------------------------------------------
chat

## -----------------------------------------------------------------------------
# screenshot_website <- tool(
#   function(url) {
#     tmpf <- withr::local_tempfile(fileext = ".png")
#     webshot2::webshot(url, file = tmpf)
#     content_image_file(tmpf)
#   },
#   name = "screenshot_website",
#   description = "Take a screenshot of a website.",
#   arguments = list(
#     url = type_string("The URL of the website")
#   )
# )

## -----------------------------------------------------------------------------
# chat <- chat_openai()
# #> Using model = "gpt-4.1".
# chat$register_tool(screenshot_website)
# chat$chat("Describe the design aesthetic of https://tidyverse.org")
# #> https://tidyverse.org screenshot completed
# #> The design aesthetic of the Tidyverse website (https://tidyverse.org) is
# #> clean, modern, and minimalistic, with several distinct features:
# #>
# #> - **Color Palette**: The overall site uses a lot of white space with navy
# #>   and dark backgrounds for some elements, accentuated by the colorful
# #>   hexagonal logos for various R packages.
# #> - **Typography**: Simple, sans-serif fonts contribute to readability and
# #>   a contemporary look.
# #> - **Hexagonal Icons**: Prominent display of tidyverse package logos in
# #>   hexagonal shapes, emphasizing the modular, package-oriented
# #>   nature of the Tidyverse.
# #> - **Layout**: A balanced, spacious two-column layout. The left side
# #>   features graphic elements; the right side provides concise, text-based
# #>   information.
# #>
# #> Overall, the design communicates clarity, ease of use, and a focus on
# #> modern data science tools.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# stream <- chat$stream("What are some common uses of R?")
# coro::loop(for (chunk in stream) {
#   cat(toupper(chunk))
# })
# #>  R IS COMMONLY USED FOR:
# #>
# #>  1. **STATISTICAL ANALYSIS**: PERFORMING COMPLEX STATISTICAL TESTS AND ANALYSES.
# #>  2. **DATA VISUALIZATION**: CREATING GRAPHS, CHARTS, AND PLOTS USING PACKAGES LIKE  GGPLOT2.
# #>  3. **DATA MANIPULATION**: CLEANING AND TRANSFORMING DATA WITH PACKAGES LIKE DPLYR AND TIDYR.
# #>  4. **MACHINE LEARNING**: BUILDING PREDICTIVE MODELS WITH LIBRARIES LIKE CARET AND #>  RANDOMFOREST.
# #>  5. **BIOINFORMATICS**: ANALYZING BIOLOGICAL DATA AND GENOMIC STUDIES.
# #>  6. **ECONOMETRICS**: PERFORMING ECONOMIC DATA ANALYSIS AND MODELING.
# #>  7. **REPORTING**: GENERATING DYNAMIC REPORTS AND DASHBOARDS WITH R MARKDOWN.
# #>  8. **TIME SERIES ANALYSIS**: ANALYZING TEMPORAL DATA AND FORECASTING.
# #>
# #>  THESE USES MAKE R A POWERFUL TOOL FOR DATA SCIENTISTS, STATISTICIANS, AND RESEARCHERS.

## ----eval=FALSE---------------------------------------------------------------
# library(promises)
# 
# chat$chat_async("How's your day going?") %...>% print()
# #> I'm just a computer program, so I don't have feelings, but I'm here to help you with any questions you have.

## ----eval=FALSE---------------------------------------------------------------
# stream <- chat$stream_async("What are some common uses of R?")
# coro::async(function() {
#   for (chunk in await_each(stream)) {
#     cat(toupper(chunk))
#   }
# })()
# #>  R IS COMMONLY USED FOR:
# #>
# #>  1. **STATISTICAL ANALYSIS**: PERFORMING VARIOUS STATISTICAL TESTS AND MODELS.
# #>  2. **DATA VISUALIZATION**: CREATING PLOTS AND GRAPHS TO VISUALIZE DATA.
# #>  3. **DATA MANIPULATION**: CLEANING AND TRANSFORMING DATA WITH PACKAGES LIKE DPLYR.
# #>  4. **MACHINE LEARNING**: BUILDING PREDICTIVE MODELS AND ALGORITHMS.
# #>  5. **BIOINFORMATICS**: ANALYZING BIOLOGICAL DATA, ESPECIALLY IN GENOMICS.
# #>  6. **TIME SERIES ANALYSIS**: ANALYZING TEMPORAL DATA FOR TRENDS AND FORECASTS.
# #>  7. **REPORT GENERATION**: CREATING DYNAMIC REPORTS WITH R MARKDOWN.
# #>  8. **GEOSPATIAL ANALYSIS**: MAPPING AND ANALYZING GEOGRAPHIC DATA.

# Translated from
# https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/utils/_merge.py
merge_dicts <- function(left, right) {
  for (right_k in names(right)) {
    right_v <- right[[right_k]]
    left_v <- left[[right_k]]

    if (is.null(right_v)) {
      if (!has_name(left, right_k)) {
        left[right_k] <- list(NULL)
      }
    } else if (is.null(left_v)) {
      left[[right_k]] <- right_v
    } else if (identical(left_v, right_v)) {
      next
    } else if (is.character(left_v)) {
      left[[right_k]] <- paste0(left_v, right_v)
    } else if (is.integer(left_v)) {
      left[[right_k]] <- right_v
    } else if (is.list(left_v)) {
      if (!is.null(names(right_v))) {
        left[[right_k]] <- merge_dicts(left_v, right_v)
      } else {
        left[[right_k]] <- merge_lists(left_v, right_v)
      }
    } else if (!identical(class(left_v), class(right_v))) {
      stop(paste0(
        "additional_kwargs['",
        right_k,
        "'] already exists in this message, but with a different type."
      ))
    } else {
      stop(paste0(
        "Additional kwargs key ",
        right_k,
        " already exists in left dict and value has unsupported type ",
        class(left[[right_k]]),
        "."
      ))
    }
  }

  left
}

merge_lists <- function(left, right) {
  if (is.null(right)) {
    return(left)
  } else if (is.null(left)) {
    return(right)
  }

  for (e in right) {
    idx <- find_index(left, e)
    if (is.na(idx)) {
      left <- c(left, list(e))
    } else {
      # If a top-level "type" has been set for a chunk, it should no
      # longer be overridden by the "type" field in future chunks.
      if (!is.null(left[[idx]]$type) && !is.null(e$type)) {
        e$type <- NULL
      }
      left[[idx]] <- merge_dicts(left[[idx]], e)
    }
  }
  left
}

find_index <- function(left, e_right) {
  if (
    !is.list(e_right) ||
      !has_name(e_right, "index") ||
      !is.numeric(e_right$index)
  ) {
    return(NA)
  }

  matches_idx <- map_lgl(left, function(e_left) e_left$index == e_right$index)
  if (sum(matches_idx) != 1) {
    return(NA)
  }
  which(matches_idx)[[1]]
}

test_that("NULLS + values", {
  expect_equal(merge_dicts(list(), list(a = 1)), list(a = 1))
  expect_equal(merge_dicts(list(a = 1), list()), list(a = 1))
})

test_that("equal values are ok", {
  expect_equal(merge_dicts(list(a = 1), list(a = 1)), list(a = 1))
  expect_equal(merge_dicts(list(a = 1.5), list(a = 1.5)), list(a = 1.5))
  expect_equal(
    merge_dicts(list(a = list(b = 1)), list(a = list(c = 2))),
    list(a = list(b = 1, c = 2))
  )

  expect_equal(merge_dicts(list(a = TRUE), list(a = TRUE)), list(a = TRUE))
  expect_equal(merge_dicts(list(a = FALSE), list(a = FALSE)), list(a = FALSE))

  expect_equal(merge_dicts(list(a = "x"), list(a = "x")), list(a = "x"))
})

test_that("strings are concatenated", {
  expect_equal(merge_dicts(list(a = "a"), list(a = "b")), list(a = "ab"))
  expect_equal(
    merge_dicts(list(a = list(b = "a")), list(a = list(b = "b"))),
    list(a = list(b = "ab"))
  )
})

test_that("can merge dictionaries with different keys", {
  expect_equal(
    merge_dicts(list(a = 1, b = 2), list(a = 1)),
    list(a = 1, b = 2)
  )
  expect_equal(
    merge_dicts(list(a = 1, b = 2), list(c = NULL)),
    list(a = 1, b = 2, c = NULL)
  )
})

test_that("nulls don't overwrite existing values (#115)", {
  expect_equal(
    merge_dicts(list(a = 1, b = 2), list(a = NULL)),
    list(a = 1, b = 2)
  )
})

test_that("can merge lists", {
  expect_equal(
    merge_dicts(list(a = list(1, 2)), list()),
    list(a = list(1, 2))
  )
  expect_equal(
    merge_dicts(list(), list(a = list(1, 2))),
    list(a = list(1, 2))
  )
  expect_equal(
    merge_dicts(list(a = list(1, 2)), list(a = list(3))),
    list(a = list(1, 2, 3))
  )
})

test_that("respects index when merging lists", {
  expect_equal(
    merge_dicts(
      list(a = list(list(index = 0, b = "{"))),
      list(a = list(list(index = 0, b = "f")))
    ),
    list(a = list(list(index = 0, b = "{f")))
  )

  expect_equal(
    merge_dicts(
      list(a = list(list(index = 0, b = "a"))),
      list(a = list(list(index = 1, b = "b")))
    ),
    list(a = list(list(index = 0, b = "a"), list(index = 1, b = "b")))
  )
})

test_that("we can merge OpenAI's chunk format", {
  # Example from: https://platform.openai.com/docs/api-reference/chat-streaming/streaming
  messages <- list(
    '{"id":"chatcmpl-123","object":"chat.completion.chunk","created":1694268190,"model":"gpt-4o-mini", "system_fingerprint": "fp_44709d6fcb", "choices":[{"index":0,"delta":{"role":"assistant","content":""},"logprobs":null,"finish_reason":null}]}',
    '{"id":"chatcmpl-123","object":"chat.completion.chunk","created":1694268190,"model":"gpt-4o-mini", "system_fingerprint": "fp_44709d6fcb", "choices":[{"index":0,"delta":{"content":"Hello"},"logprobs":null,"finish_reason":null}]}',
    '{"id":"chatcmpl-123","object":"chat.completion.chunk","created":1694268190,"model":"gpt-4o-mini", "system_fingerprint": "fp_44709d6fcb", "choices":[{"index":0,"delta":{},"logprobs":null,"finish_reason":"stop"}]}'
  )
  chunks <- lapply(messages, jsonlite::parse_json)
  expect_equal(
    merge_dicts(merge_dicts(chunks[[1]], chunks[[2]]), chunks[[3]]),
    list(
      id = "chatcmpl-123",
      object = "chat.completion.chunk",
      created = 1694268190,
      model = "gpt-4o-mini",
      system_fingerprint = "fp_44709d6fcb",
      choices = list(list(
        index = 0,
        delta = list(
          role = "assistant",
          content = "Hello"
        ),
        logprobs = NULL,
        finish_reason = "stop"
      ))
    )
  )
})

#' Callback Manager
#'
#' A simple class to manage a collection of callback functions that can be
#' invoked sequentially with a single call to `$invoke()` with an object or data
#' to pass to the callback functions. Callbacks must take at least one argument
#' are invoked in reverse order of their registration.
#'
#' @noRd
CallbackManager <- R6Class(
  "CallbackManager",

  public = list(
    initialize = function(args = NULL) {
      private$args <- args
    },

    #' @description Add a callback function.
    #' @param callback A function to be called.
    #' @return A function that can be called to remove the callback.
    add = function(callback, call = caller_env()) {
      check_function2(callback, args = private$args, call = call)

      id <- private$next_id()
      private$callbacks[[id]] <- callback

      fn_remove <- function() {
        private$callbacks[[id]] <- NULL
        invisible()
      }
      invisible(fn_remove)
    },

    #' @description
    #' Invoke all registered callbacks with the provided arguments. Callbacks
    #' are invoked in reverse order of registration (last-in first-evaluated).
    #'
    #' @param ... Arguments to pass to the callbacks.
    #' @returns Nothing, callbacks are invoked for side effects).
    invoke = function(...) {
      if (length(private$callbacks) == 0) {
        return(invisible(NULL))
      }

      # Invoke callbacks in reverse insertion order
      for (id in rev(as.integer(names(private$callbacks)))) {
        res <- private$callbacks[[as.character(id)]](...)
        if (promises::is.promise(res)) {
          cli::cli_abort(c(
            "Can't use async callbacks with `$chat()` or `$stream()`.",
            i = "Async callbacks are supported, but you must use `$chat_async()` or `$stream_async()`."
          ))
        }
      }

      invisible(NULL)
    },

    #' @description
    #' Invoke all registered callbacks asynchronously with the provided
    #' arguments. As with `$invoke()`, callbacks are invoked in reverse order of
    #' registration (last-in first-evaluated).
    #'
    #' @param ... Arguments to pass to the callbacks.
    #' @returns Nothing, callbacks are invoked for side effects).
    invoke_async = async_method(function(self, private, ...) {
      if (length(private$callbacks) == 0) {
        return(invisible(NULL))
      }

      # Invoke callbacks in reverse insertion order
      for (id in rev(as.integer(names(private$callbacks)))) {
        coro::await(exec(private$callbacks[[as.character(id)]], ...))
      }

      invisible(NULL)
    }),

    #' @description Get the number of registered callbacks.
    #' @return Integer count of callbacks.
    count = function() {
      length(private$callbacks)
    },

    #' @description Clear all registered callbacks.
    clear = function() {
      private$callbacks <- list()
      invisible(NULL)
    },

    #' @describeIn Get callback list
    get_callbacks = function() {
      private$callbacks
    }
  ),

  private = list(
    callbacks = list(),
    args = NULL,

    id = 1L,
    next_id = function() {
      id <- private$id
      private$id <- private$id + 1L
      as.character(id)
    }
  )
)

# From https://github.com/r-lib/httr2/blob/da2724ae/R/utils.R#L179-L235
check_function2 <- function(
  x,
  ...,
  args = NULL,
  allow_null = FALSE,
  arg = caller_arg(x),
  call = caller_env()
) {
  check_function(
    x = x,
    allow_null = allow_null,
    arg = arg,
    call = call
  )

  if (!is.null(x)) {
    .check_function_args(
      f = x,
      expected_args = args,
      arg = arg,
      call = call
    )
  }
}

.check_function_args <- function(f, expected_args, arg, call) {
  if (is_null(expected_args)) {
    return(invisible(NULL))
  }

  actual_args <- fn_fmls_names(f) %||% character()
  missing_args <- setdiff(expected_args, actual_args)
  if (is_empty(missing_args)) {
    return(invisible(NULL))
  }

  n_expected_args <- length(expected_args)
  n_actual_args <- length(actual_args)

  if (n_actual_args == 0) {
    arg_info <- "instead it has no arguments"
  } else {
    arg_info <- paste0("it currently has {.arg {actual_args}}")
  }

  cli::cli_abort(
    paste0(
      "{.arg {arg}} must have the {cli::qty(n_expected_args)}argument{?s} {.arg {expected_args}}; ",
      arg_info,
      "."
    ),
    call = call,
    arg = arg
  )
}

test_that("CallbackManager catches argument mismatches", {
  callbacks <- CallbackManager$new(args = "data")

  expect_snapshot(error = TRUE, {
    callbacks$add("foo")
    callbacks$add(function(foo) NULL)
    callbacks$add(function(x, y) x + y)
  })

  expect_silent(callbacks$add(function(data) data))
  expect_silent(callbacks$invoke(data = data))
  expect_silent(callbacks$invoke(1))

  # Callbacks with invalid args throw standard R error
  expect_snapshot(error = TRUE, {
    callbacks$invoke()
    callbacks$invoke(1, 2)
  })
})

test_that("CallbackManager invokes callbacks in LIFO order", {
  callbacks <- CallbackManager$new()
  res1 <- NULL
  res2 <- NULL

  cb1 <- callbacks$add(function(value) {
    res1 <<- list(value = value, time = Sys.time())
  })

  cb2 <- callbacks$add(function(...) {
    res <- list(time = Sys.time())
    res["value"] <- list(...)
    res2 <<- res
  })

  expect_equal(callbacks$count(), 2)

  # Callbacks don't return a value
  expect_null(callbacks$invoke(list(x = 1, y = 2)))

  # Callbacks receive expected arguments
  expect_equal(res1$value, list(x = 1, y = 2))
  expect_equal(res2$value, list(x = 1, y = 2))
  # Callbacks are invoked in reverse order
  expect_true(res1$time > res2$time)
})

test_that("$add() returns a function to remove the callback", {
  callbacks <- CallbackManager$new()
  res1 <- NULL
  res2 <- NULL

  cb1 <- callbacks$add(function() {
    res1 <<- Sys.time()
  })
  cb2 <- callbacks$add(function() {
    res2 <<- Sys.time()
  })

  expect_equal(callbacks$count(), 2)
  callbacks$invoke()

  # Unregistering a callback
  res1_first <- res1
  res2_first <- res2
  cb1()
  expect_equal(callbacks$count(), 1)
  callbacks$invoke()
  expect_equal(res1, res1_first) # first callback result hasn't changed
  expect_true(res2 > res2_first) # second callback was evaluated

  # Unregistering callbacks are idempotent
  cb_list <- callbacks$get_callbacks()
  expect_null(cb_list[["1"]])
  cb1()
  # Callback list hasn't changed
  expect_equal(callbacks$get_callbacks(), cb_list)
})

test_that("$clear() clears all callbacks", {
  callbacks <- CallbackManager$new()
  res <- c()
  callbacks$add(function() res <<- c(res, 1))
  callbacks$add(function() res <<- c(res, 2))
  expect_equal(callbacks$count(), 2)

  callbacks$clear()
  expect_equal(callbacks$count(), 0)
  expect_equal(callbacks$get_callbacks(), list())

  # Invoking without registered callbacks means nothing happens
  expect_null(callbacks$invoke())
  expect_equal(res, c()) # nothing happened
})

#' OpenAI web search tool
#'
#' @description
#' Enables OpenAI models to search the web for up-to-date information. The search
#' behavior varies by model: non-reasoning models perform simple searches, while
#' reasoning models can perform agentic, iterative searches.
#'
#' Learn more at <https://platform.openai.com/docs/guides/tools-web-search>
#'
#' @param allowed_domains Character vector. Restrict searches to specific domains
#'   (e.g., `c("nytimes.com", "bbc.com")`). Maximum 20 domains. URLs will be
#'   automatically cleaned (http/https prefixes removed).
#' @param user_location List with optional elements: `country` (2-letter ISO code),
#'   `city`, `region`, and `timezone` (IANA timezone) to localize search results.
#' @param external_web_access Logical. Whether to allow live internet access
#'   (`TRUE`, default) or use only cached/indexed results (`FALSE`).
#'
#' @family built-in tools
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_openai()
#' chat$register_tool(openai_tool_web_search())
#' chat$chat("Very briefly summarise the top 3 news stories of the day")
#' chat$chat("Of those stories, which one do you think was the most interesting?")
#' }
openai_tool_web_search <- function(
  allowed_domains = NULL,
  user_location = NULL,
  external_web_access = TRUE
) {
  check_character(allowed_domains, allow_null = TRUE)
  check_bool(external_web_access)

  # Strip http/https from domains
  if (!is.null(allowed_domains)) {
    allowed_domains <- sub("^https?://", "", allowed_domains)
  }

  json <- compact(list(
    type = "web_search",
    filters = if (!is.null(allowed_domains)) {
      list(allowed_domains = allowed_domains)
    },
    user_location = user_location,
    external_web_access = external_web_access
  ))
  ToolBuiltIn(
    name = "web_search",
    description = "Search the web for up-to-date information.",
    annotations = tool_annotations(
      title = "Web search",
      read_only_hint = TRUE,
      open_world_hint = TRUE
    ),
    json = json
  )
}

#' Create a stream controller
#'
#' @description
#' Creates a controller that can cancel an in-progress stream. Pass it to
#' [Chat]'s `$stream()` or `$stream_async()` via the `controller` argument,
#' then call `$cancel()` from anywhere (e.g. a Shiny observer) to stop the
#' stream after the next chunk arrives.
#'
#' The same controller can be reused across multiple streams. Call
#' `$reset()` to clear the cancelled state, or pass it directly to a new
#' `$stream()` call — it will be reset automatically.
#'
#' @section Async cancellation in Shiny:
#'
#' In a Shiny app, use an [ExtendedTask][shiny::ExtendedTask] for
#' non-blocking chat and a `stream_controller()` to wire up a cancel
#' button:
#'
#' ```r
#' controller <- stream_controller()
#'
#' chat_task <- ExtendedTask$new(function(user_query, controller = NULL) {
#'   chat <- chat_openai(model = "gpt-4.1-nano")
#'   stream <- chat$stream_async(user_query, controller = controller)
#'   shinychat::markdown_stream("response", stream)
#' })
#'
#' observeEvent(input$ask, {
#'   controller <<- stream_controller()
#'   chat_task$invoke(input$query, controller = controller)
#' })
#'
#' observeEvent(input$cancel, {
#'   controller$cancel()
#' })
#' ```
#'
#' @return An `ellmer_stream_controller` object with the following
#'   elements:
#'
#'   * `$cancel(reason = "cancelled")`: Cancel the stream. The `reason`
#'     string is stored on the controller and used as the
#'     [AssistantPartialTurn][Turn]'s `reason` property.
#'   * `$reset()`: Clear the cancelled state and reason.
#'   * `$cancelled`: A logical flag indicating whether the controller
#'     has been cancelled.
#'   * `$reason`: The cancellation reason string, or `NULL` if not
#'     cancelled.
#'
#' @examplesIf rlang::is_interactive()
#' chat <- chat_openai(model = "gpt-5.4-nano")
#'
#' ctrl <- stream_controller()
#' stream <- chat$stream("Write a short story.", controller = ctrl)
#'
#' i <- 0
#' coro::loop(for (chunk in stream) {
#'   i <- i + 1
#'   if (i > 10) ctrl$cancel()
#' })
#'
#' chat
#'
#' @export
stream_controller <- function() {
  StreamController$new()
}

StreamController <- R6::R6Class(
  "ellmer_stream_controller",
  public = list(
    cancel = function(reason = "cancelled") {
      check_string(reason)
      private$.reason <- reason
      private$.cancelled <- TRUE
    },

    reset = function() {
      private$.cancelled <- FALSE
      private$.reason <- NULL
    },

    print = function(...) {
      status <- if (private$.cancelled) {
        sprintf('cancelled="%s"', private$.reason)
      } else {
        "active"
      }
      cat("<ellmer_stream_controller ", status, ">\n", sep = "")
      invisible(self)
    }
  ),
  active = list(
    cancelled = function(value) {
      if (!missing(value)) {
        cli::cli_abort(
          "Use the {.code $cancel()} method to cancel the stream or {.code $reset()} to reset the stream controller."
        )
      }
      private$.cancelled
    },
    reason = function(value) {
      if (!missing(value)) {
        cli::cli_abort(
          "Use the {.code $cancel()} method to cancel the stream or {.code $reset()} to reset the stream controller."
        )
      }
      private$.reason
    }
  ),
  private = list(
    .cancelled = FALSE,
    .reason = NULL
  )
)

as_controller <- function(controller, reset = TRUE, call = caller_env()) {
  if (is.null(controller)) {
    return(stream_controller())
  }

  if (!inherits(controller, "ellmer_stream_controller")) {
    cli::cli_abort(
      "{.arg controller} must be an {.cls ellmer_stream_controller} object created by {.fn stream_controller}.",
      call = call
    )
  }

  if (reset && controller$cancelled) {
    controller$reset()
  }

  controller
}

cat_line <- function(..., prefix = "") {
  lines <- paste0(..., collapse = "\n")
  sink <- cat_word_wrap(prefix = prefix)
  sink(lines)
  if (!endsWith(lines, "\n")) {
    sink("\n")
  }
  sink(coro::exhausted())
}

# Utility functions for performing word wrapping on streaming text. Our strategy
# for dealing with streaming is to buffer the most recent word, assuming that
# it's incomplete, and render it only when the next word break (space or
# newline) is encountered.

# Split a string into contiguous runs of non-spaces, " ", and newline.
#
# split_spaces("one  two    three") => c("one", "  ", "two", "    ", "three")
split_spaces <- function(text) {
  if (length(text) != 1 || !is.character(text)) {
    stop("Input must be a character vector of length 1")
  }

  pattern <- "([^ \\n]+| +|\\n)"
  result <- regmatches(text, gregexpr(pattern, text, perl = TRUE))[[1]]
  result
}

# A sink is a function that takes a string and returns `invisible()`. The
# `prepare_sink` function takes a raw function that can take a string as its
# first argument, and returns whatever, and returns a wrapper sink function. It
# also primes generators.
prepare_sink <- function(raw_sink_func, ...) {
  if (inherits(raw_sink_func, "coro_generator_instance")) {
    # Generators need to be primed before they can be used
    raw_sink_func(...)
  }

  function(str) {
    raw_sink_func(str, ...)
    invisible()
  }
}


# Sink decorators ---------------------------------------------------------
#
# A sink decorator is a sink that wraps another sink, modifying or coalescing or
# redacting or otherwise messing with the input before passing it to the target
# sink.
#
# We currently have two sink decorators:
# - sink_wordwrap: wraps a sink to perform word wrapping
# - sink_prefix: wraps a sink to prefix each line with a given string
#
# Creating a sink decorator is a two-step process:
# 1. Define a generator that takes a sink and some other arguments, and calls
#    `yield()` to get input and `sink()` to pass output to the target sink.
# 2. Define a wrapper function that takes the target sink and any other
#    arguments, and returns a sink that wraps the target sink with the generator
#    from step 1. It's important that the wrapper function calls `force()` on
#    all of its arguments, as the generator is not likely to trigger evaluation
#    until relatively late. It also needs to call `prepare_sink` to ensure that
#    the generator is primed before it's used.

sink_wordwrap <- function(sink, width) {
  force(sink)
  force(width)
  prepare_sink(sink_wordwrap_gen(sink, width))
}

sink_wordwrap_gen <- NULL
on_load(
  sink_wordwrap_gen <- coro::generator(function(sink, width) {
    buffer <- ""
    pos_cursor <- 1

    exhausted <- FALSE
    while (!exhausted) {
      input <- coro::yield()
      if (coro::is_exhausted(input)) {
        exhausted <- TRUE
        input <- ""
      }

      input <- paste0(buffer, input)
      buffer <- ""

      # Normalize whitespace, make our lives easier
      input <- gsub("\r\n", "\n", input)
      input <- gsub("\r", "\n", input)
      input <- gsub("\t", "    ", input)

      tokens <- split_spaces(input)
      for (i in seq_along(tokens)) {
        token <- tokens[[i]]
        is_last_token <- i == length(tokens)

        token_width <- nchar(token)
        is_space <- startsWith(token, " ")
        # Uncomment to render whitespace
        # token <- gsub(" ", "\u00B7", token)

        if (token == "\n") {
          # Always flush newlines immediately
          sink("\n")
          pos_cursor <- 1
        } else if (is_last_token && !exhausted) {
          # A trailing non-newline might not be complete; buffer it until we have
          # more text to process
          buffer <- token
        } else {
          # A regular token. See if it fits on the current line.
          soft_wrap <- pos_cursor + token_width > width

          if (soft_wrap && (pos_cursor != 1 || is_space)) {
            # It doesn't fit; wrap to the next line
            sink("\n")
            pos_cursor <- 1

            if (is_space) {
              # soft-wrapping due to spaces; skip rendering the spaces
              next
            }
          }

          # Render the token and update the cursor
          sink(token)
          pos_cursor <- pos_cursor + token_width
        }
      }
    }
  })
)

# Wrap a sink to prefix each line with a given string.
sink_prefix <- function(sink, prefix = "") {
  force(sink)
  force(prefix)
  prepare_sink(sink_prefix_gen(sink, prefix))
}

sink_prefix_gen <- NULL
on_load(
  sink_prefix_gen <- coro::generator(function(sink, prefix = "") {
    # When at line start, the next input will be prefixed with `prefix`. (But if
    # the next input never comes, we won't print the prefix.)
    at_line_start <- TRUE

    # In case the prefix itself contains characters that have special meaning when
    # used as a gsub replacement
    prefix_escaped <- gsub(perl = TRUE, "(\\\\|\\$)", "\\\\\\1", prefix)

    repeat {
      input <- coro::yield(invisible())
      if (coro::is_exhausted(input)) {
        break
      }
      if (nchar(input) == 0) {
        next
      }

      # We have non-empty input. If we're at the start of the line, it's time to
      # print a prefix.
      if (at_line_start) {
        sink(prefix)
        at_line_start <- FALSE
      }

      # Add the prefix after each newline in the input except the last one.
      transformed <- gsub(
        perl = TRUE,
        "\n(?!$)", # Matches \n that are NOT immediately followed by EOF
        paste0("\n", prefix_escaped),
        input
      )
      sink(transformed)

      # If the input ends with a newline, let the _next_ input be prefixed.
      if (endsWith(input, "\n")) {
        at_line_start <- TRUE
      }
    }
  })
)

# cat_word_wrap() is a function that returns a function that can be used to
# print text to the console with word wrapping. It uses the COLUMNS environment
# variable to work well from the terminal, and getWidth("options") to work well
# in RStudio.
#
# Each call to cat_word_wrap() (i.e. the creation of a catter) assumes that the
# current line of the given connection is empty. Each catter will assume it has
# sole possession of the connection (i.e. it ignores other code writing directly
# to its underlying connection).
#
# The catter buffers the last word of a line, assuming that it's incomplete. To
# ensure that the buffer gets flushed, cat a "\n" character.
cat_word_wrap <- function(
  con = stdout(),
  prefix = "",
  width = cli::console_width()
) {
  sink <- prepare_sink(cat, file = con, sep = "")

  if (nchar(prefix) > 0) {
    sink <- sink_prefix(sink, prefix)
  }

  sink <- sink_wordwrap(sink, width - nchar(prefix))

  sink
}

emitter <- function(echo, prefix) {
  if (echo == "output") {
    cat_word_wrap()
  } else if (echo == "all") {
    cat_word_wrap(prefix = "< ")
  } else {
    function(...) invisible()
  }
}

# Execute a function that sinks text to the console and capture the output.
capture_sink <- function(prefix, func, width = 64) {
  con <- file(open = "w+b")
  on.exit(close(con))
  sink <- cat_word_wrap(con, prefix = prefix, width = width)
  func(sink)
  sink(coro::exhausted())
  len <- seek(con, 0)
  readChar(con, len)
}

sink_test_equal <- function(prefix, expected, func, width = 64) {
  value <- capture_sink(prefix, func, width = width)
  rlang::inject(expect_equal(value, {{ expected }}))
}

test_that("wrap + prefix interactions", {
  sink_test_equal(
    prefix = ">>>",
    expected = ">>>\n",
    function(sink) {
      sink("\n")
    },
    width = 12
  )

  sink_test_equal(
    prefix = "> ",
    expected = "> a\n> b\n",
    function(sink) {
      sink("a\n")
      sink("b\n")
    },
    width = 12
  )

  sink_test_equal(
    prefix = "> ",
    expected = "> this_is_a_very_long_string_that_nevertheless_should_not_wrap\n",
    function(sink) {
      sink("this_is_a_very_long_string_that_nevertheless_should_not_wrap\n")
    },
    width = 12
  )

  # Just barely fits (11 chars wide)
  sink_test_equal(
    prefix = "12",
    expected = "12345678 AB\n",
    function(sink) {
      sink("345678 AB\n")
    },
    width = 12
  )

  # Doesn't fit, but can't wrap
  sink_test_equal(
    prefix = "12",
    expected = "123456789ABCD\n",
    function(sink) {
      sink("3456789ABCD\n")
    },
    width = 12
  )

  # Doesn't fit, but wraps
  sink_test_equal(
    prefix = "> ",
    expected = "> 345678 \n> ABCD\n",
    function(sink) {
      sink("345678 ABCD\n")
    },
    width = 12
  )
})

test_that("more realistic wrap examples", {
  for (prefix in c("?> ", "")) {
    for (width in c(70, 10000)) {
      rlang::inject({
        expect_snapshot(cat(capture_sink(
          prefix = !!prefix,
          width = !!width,
          function(sink) {
            sink(
              "It was the best of times, it was the worst of times, it was the age "
            )
            sink(
              "of wisdom, it was the age of foolishness, it was the epoch of belief"
            )
            sink(
              ", it was the epoch of incredulity, it was the season of Light, it wa"
            )
            sink(
              "s the season of Darkness, it was the spring of hope, it was the wint"
            )
            sink("er of despair.")
          }
        )))
      })
    }
  }

  expect_snapshot(cat(capture_sink(prefix = "Dickens> ", function(sink) {
    sink("It was the best of times,\nit was the worst of times,\n")
    sink("it was the age ")
    sink("of wisdom, it was the age of ")
    sink("foolishness,\nit was the epoch of belief")
    sink(", it was the epoch of incredulity, ")
    sink(paste(rep_len(" ", 200), collapse = ""))
    sink("it was the season of Light, it wa")
    sink("s the season of Darkness,\n\n\nit was the spring of hope,")
    sink(" it was the wint")
    sink("er of despair.")
  })))
})

#' @include turns.R
NULL

is_tool_request <- function(x) S7_inherits(x, ContentToolRequest)
is_tool_result <- function(x) S7_inherits(x, ContentToolResult)

match_tools <- function(turn, tools) {
  if (is.null(turn)) {
    return(NULL)
  }

  turn@contents <- map(turn@contents, function(content) {
    if (is_tool_request(content)) {
      content@tool <- content@tool %||% tools[[content@name]]
      return(content)
    }

    if (is_tool_result(content)) {
      content@request@tool <-
        content@request@tool %||% tools[[content@request@name]]
      return(content)
    }

    content
  })

  turn
}

on_load({
  invoke_tools <- coro::generator(function(
    turn,
    echo = "none",
    on_tool_request = function(request) invisible(),
    on_tool_result = function(result) invisible(),
    yield_request = FALSE,
    otel_span = NULL
  ) {
    tool_requests <- extract_tool_requests(turn)

    for (request in tool_requests) {
      maybe_echo_tool(request, echo = echo)
      if (yield_request) {
        yield(request)
      }

      rejected <- maybe_on_tool_request(request, on_tool_request)
      if (!is.null(rejected)) {
        maybe_echo_tool(rejected, echo = echo)
        on_tool_result(rejected)
        yield(rejected)
        next
      }

      result <- invoke_tool(request, otel_span = otel_span)

      if (promises::is.promise(result@value)) {
        cli::cli_abort(
          c(
            "Can't use async tools with `$chat()` or `$stream()`.",
            i = "Async tools are supported, but you must use `$chat_async()` or `$stream_async()`."
          ),
          class = "tool_async_error"
        )
      }

      maybe_echo_tool(result, echo = echo)
      on_tool_result(result)
      yield(result)
    }
  })

  # invoke_tools_async is intentionally *not* an _async_ generator, instead it
  # is a generator that returns promises. This lets the caller decide if the
  # tasks should be run in parallel or sequentially.
  invoke_tools_async <- coro::generator(function(
    turn,
    tools,
    echo = "none",
    on_tool_request = function(request) invisible(),
    on_tool_result = function(result) invisible(),
    yield_request = FALSE,
    otel_span = NULL
  ) {
    tool_requests <- extract_tool_requests(turn)

    invoke_tool_async_wrapper <- coro::async(function(request) {
      maybe_echo_tool(request, echo = echo)

      rejected <- coro::await(
        maybe_on_tool_request_async(request, on_tool_request)
      )
      if (!is.null(rejected)) {
        maybe_echo_tool(rejected, echo = echo)
        on_tool_result(rejected)
        return(rejected)
      }

      result <- coro::await(invoke_tool_async(request, otel_span = otel_span))

      maybe_echo_tool(result, echo = echo)
      on_tool_result(result)
      result
    })

    for (request in tool_requests) {
      if (yield_request) {
        yield(request)
      }
      yield(invoke_tool_async_wrapper(request))
    }
  })
})

gen_async_promise_all <- function(generator) {
  promises::promise_all(.list = coro::collect(generator))
}

extract_tool_requests <- function(turn) {
  if (is.null(turn)) {
    return(NULL)
  }

  turn@contents[map_lgl(turn@contents, is_tool_request)]
}

turn_has_tool_request <- function(turn) {
  if (is.null(turn)) {
    return(FALSE)
  }
  stopifnot(S7_inherits(turn, Turn))

  some(turn@contents, is_tool_request)
}

new_tool_result <- function(request, result = NULL, error = NULL) {
  check_exclusive(result, error)

  if (!is.null(error)) {
    ContentToolResult(error = error, request = request)
  } else if (is_tool_result(result)) {
    set_props(result, request = request)
  } else {
    ContentToolResult(value = result, request = request)
  }
}

# Also need to handle edge cases: https://platform.openai.com/docs/guides/function-calling/edge-cases
invoke_tool <- function(request, otel_span = NULL) {
  if (is.null(request@tool)) {
    return(new_tool_result(request, error = "Unknown tool"))
  }

  args <- tool_request_args(request)
  if (is_tool_result(args)) {
    # Failed to convert the arguments
    return(args)
  }

  tool_span <- local_tool_otel_span(request, parent = otel_span)

  tryCatch(
    {
      result <- do.call(request@tool, args)
      new_tool_result(request, result)
    },
    error = function(e) {
      record_tool_otel_span_error(tool_span, e)
      new_tool_result(request, error = e)
    }
  )
}

on_load(
  invoke_tool_async <- coro::async(function(request, otel_span = NULL) {
    if (is.null(request@tool)) {
      return(new_tool_result(request, error = "Unknown tool"))
    }

    args <- tool_request_args(request)
    if (is_tool_result(args)) {
      # Failed to convert the arguments
      return(args)
    }

    tool_span <- local_tool_otel_span(request, parent = otel_span)

    tryCatch(
      {
        result <- await(do.call(request@tool, args))
        new_tool_result(request, result)
      },
      error = function(e) {
        record_tool_otel_span_error(tool_span, e)
        new_tool_result(request, error = e)
      }
    )
  })
)

tool_request_args <- function(request) {
  tool <- request@tool
  args <- request@arguments

  if (is.null(tool) || !isTRUE(tool@convert)) {
    return(args)
  }

  extra_args <- setdiff(names(args), names(tool@arguments@properties))
  if (length(extra_args) > 0) {
    e <- catch_cnd(cli::cli_abort("Unused argument{?s}: {extra_args}"))
    return(new_tool_result(request, error = e))
  }

  args <- convert_from_type(args, tool@arguments)
  args[!map_lgl(args, is.null)]
}

maybe_on_tool_request <- function(
  request,
  on_tool_request = function(request) invisible()
) {
  tryCatch(
    {
      on_tool_request(request)
      NULL
    },
    ellmer_tool_reject = function(e) {
      ContentToolResult(error = e$message, request = request)
    }
  )
}

on_load(
  maybe_on_tool_request_async <- coro::async(
    function(request, on_tool_request = function(request) invisible()) {
      tryCatch(
        {
          coro::await(on_tool_request(request))
          NULL
        },
        ellmer_tool_reject = function(e) {
          ContentToolResult(error = e$message, request = request)
        }
      )
    }
  )
)

tool_results_as_turn <- function(results) {
  if (length(results) == 0) {
    return(NULL)
  }
  is_tool_result <- map_lgl(results, is_tool_result)
  if (!any(is_tool_result)) {
    return(NULL)
  }
  UserTurn(contents = results[is_tool_result])
}

turn_split_tool_results <- function(turn) {
  is_result <- map_lgl(turn@contents, is_tool_result)
  list(
    tool_results = turn@contents[is_result],
    contents = turn@contents[!is_result]
  )
}

turn_get_tool_errors <- function(turn = NULL) {
  if (is.null(turn)) {
    return(NULL)
  }
  stopifnot(S7_inherits(turn, Turn))

  if (length(turn@contents) == 0) {
    return(NULL)
  }

  is_result <- map_lgl(turn@contents, S7_inherits, ContentToolResult)
  if (!any(is_result)) {
    return(NULL)
  }

  is_error <- map_lgl(turn@contents[is_result], tool_errored)

  res <- turn@contents[is_result][is_error]
  if (length(res)) res else NULL
}

warn_tool_errors <- function(tool_errors) {
  # tool_errors is a list of errors returned from turn_get_tool_errors()
  if (length(tool_errors) == 0) {
    return()
  }

  errs <- map_chr(
    tool_errors[seq_len(min(3, length(tool_errors)))],
    function(result) {
      name <- result@request@name %||% "unknown_tool"
      id <- result@request@id
      error <- tool_error_string(result)
      cli::format_inline("[{.field {name}} ({id})]: {cli_escape(error)}")
    }
  )

  cli::cli_warn(
    c(
      "Failed to evaluate {length(tool_errors)} tool call{?s}.",
      set_names(errs, "x"),
      "i" = if (length(errs) < length(tool_errors)) {
        cli::format_inline(
          "{cli::symbol$ellipsis} and {length(tool_errors) - length(errs)} more."
        )
      }
    ),
    class = "ellmer_tool_failure"
  )
}

maybe_echo_tool <- function(x, echo = "output") {
  if (!identical(echo, "output")) {
    return(invisible(x))
  }

  if (is_tool_request(x)) {
    cli::cli_text(
      cli::col_blue(cli::symbol$circle),
      " [{cli::col_blue('tool call')}] ",
      cli_escape(format(x, show = "call_short"))
    )
    return(invisible(x))
  }

  if (!is_tool_result(x)) {
    # neither tool result or request
    return(invisible(x))
  }

  # ContentToolResult ----
  if (tool_errored(x)) {
    icon <- cli::col_red(cli::symbol$stop)
    header <- cli::col_red("Error: ")
    value <- tool_error_string(x)
  } else {
    icon <- cli::col_green(cli::symbol$record)
    header <- ""
    value <- tool_string_preview(x)
  }

  value <- cli::style_italic(value)

  if (grepl("\n", value)) {
    lines <- strsplit(value, "\n")[[1]]
    lines <- c(
      lines[seq_len(min(5, length(lines)))],
      if (length(lines) > 5) cli::symbol$ellipsis
    )
    lines <- cli::style_italic(lines)
    cli::cli_text("{icon} #> {header}{lines[1]}")
    for (line in lines[-1]) {
      cli::cli_text("\u00a0\u00a0#> {line}")
    }
  } else {
    max_width <- cli::console_width() - 7
    if (nchar(value) > max_width) {
      value <- substring(value, 1, max_width)
      value <- paste0(value, cli::symbol$ellipsis)
    }
    value <- cli::style_italic(value)
    cli::cli_text("{icon} #> {header}{value}")
  }

  invisible(x)
}

.onLoad <- function(libname, pkgname) {
  run_on_load()
  S7::methods_register()
  otel_cache_tracer()
}

# Work around S7 bug
rm(format)
rm(print)

# enable usage of <S7_object>@name in package code
#' @rawNamespace if (getRversion() < "4.3.0") importFrom("S7", "@")
NULL

#' @include provider.R
#' @include content.R
#' @include turns.R
#' @include tools-def.R
NULL

#' Chat with a Google Gemini or Vertex AI model
#'
#' @description
#' Google's AI offering is broken up into two parts: Gemini and Vertex AI.
#' Most enterprises are likely to use Vertex AI, and individuals are likely
#' to use Gemini.
#'
#' Use [google_upload()] to upload files (PDFs, images, video, audio, etc.)
#'
#' ## Authentication
#' These functions try a number of authentication strategies, in this order:
#'
#' * An API key set in the `GOOGLE_API_KEY` env var, or,
#'   for `chat_google_gemini()` only, `GEMINI_API_KEY`.
#' * Google's default application credentials, if the \pkg{gargle} package
#'   is installed.
#' * Viewer-based credentials on Posit Connect, if the \pkg{connectcreds}
#'   package.
#' * `r lifecycle::badge("experimental")`. An browser-based OAuth flow, if
#'   you're in an interactive session. This currently uses an unverified
#'   OAuth app (so you will get a scary warning); we plan to verify in the
#'   near future.
#'
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials A function that returns a list of authentication headers
#'   or `NULL`, the default, to use ambient credentials. See above for details.
#' @param model `r param_model("gemini-2.5-flash", "google_gemini")`
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @family chatbots
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_google_gemini()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_google_gemini <- function(
  system_prompt = NULL,
  base_url = "https://generativelanguage.googleapis.com/v1beta/",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  api_headers = character(),
  echo = NULL
) {
  model <- set_default(model, "gemini-2.5-flash")
  echo <- check_echo(echo)

  credentials <- as_credentials(
    "chat_google_gemini",
    default_google_credentials(variant = "gemini"),
    credentials = credentials,
    api_key = api_key
  )

  provider <- ProviderGoogleGemini(
    name = "Google/Gemini",
    base_url = base_url,
    model = model,
    params = params %||% params(),
    extra_args = api_args,
    extra_headers = api_headers,
    credentials = credentials
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

chat_google_gemini_test <- function(
  ...,
  model = "gemini-2.5-flash",
  params = NULL,
  echo = "none"
) {
  params <- params %||% params()
  params$temperature <- params$temperature %||% 0
  params$seed <- 1014

  chat_google_gemini(..., model = model, params = params, echo = echo)
}

#' @export
#' @rdname chat_google_gemini
#' @param location Location, e.g. `us-east1`, `me-central1`, `africa-south1` or
#'   `global`.
#' @param project_id Project ID.
chat_google_vertex <- function(
  location,
  project_id,
  system_prompt = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  api_headers = character(),
  echo = NULL
) {
  check_string(location)
  check_string(project_id)

  model <- set_default(model, "gemini-2.5-flash")
  echo <- check_echo(echo)
  credentials <- default_google_credentials(variant = "vertex")

  provider <- ProviderGoogleGemini(
    name = "Google/Vertex",
    base_url = vertex_url(location, project_id),
    model = model,
    params = params %||% params(),
    extra_args = api_args,
    extra_headers = api_headers,
    credentials = credentials
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

# https://cloud.google.com/vertex-ai/docs/reference/rest/v1/projects.locations.endpoints/generateContent
vertex_url <- function(location, project_id) {
  paste_c(
    c("https://", google_location(location), "aiplatform.googleapis.com"),
    "/v1",
    c("/projects/", project_id),
    c("/locations/", location),
    "/publishers/google/"
  )
}

ProviderGoogleGemini <- new_class(
  "ProviderGoogleGemini",
  parent = Provider,
  properties = list(
    model = prop_string()
  )
)

# Base request -----------------------------------------------------------------

method(base_request, ProviderGoogleGemini) <- function(provider) {
  req <- request(provider@base_url)
  req <- ellmer_req_credentials(req, provider@credentials(), "x-goog-api-key")
  req <- ellmer_req_robustify(req)
  req <- ellmer_req_user_agent(req)
  req <- req_error(req, body = function(resp) {
    json <- resp_body_json(resp, check_type = FALSE)
    json$error$message
  })
  req
}

# Chat -------------------------------------------------------------------------

method(chat_request, ProviderGoogleGemini) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  req <- base_request(provider)

  # Can't use chat_path() because it varies based on stream
  req <- req_url_path_append(req, "models")
  if (stream) {
    # https://ai.google.dev/api/generate-content#method:-models.streamgeneratecontent
    req <- req_url_path_append(
      req,
      paste0(provider@model, ":", "streamGenerateContent")
    )
    req <- req_url_query(req, alt = "sse")
  } else {
    # https://ai.google.dev/api/generate-content#method:-models.generatecontent
    req <- req_url_path_append(
      req,
      paste0(provider@model, ":", "generateContent")
    )
  }

  body <- chat_body(
    provider = provider,
    stream = stream,
    turns = turns,
    tools = tools,
    type = type
  )
  body <- modify_list(body, provider@extra_args)

  req <- req_body_json(req, body)
  req <- req_headers(req, !!!provider@extra_headers)
  req
}

method(chat_body, ProviderGoogleGemini) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  if (length(turns) >= 1 && is_system_turn(turns[[1]])) {
    system <- list(parts = list(text = turns[[1]]@text))
  } else {
    system <- list(parts = list(text = ""))
  }

  generation_config <- chat_params(provider, provider@params)
  if (!is.null(type)) {
    generation_config$response_mime_type <- "application/json"
    generation_config$response_schema <- as_json(provider, type)
  }

  if (has_name(generation_config, "thinkingBudget")) {
    generation_config$thinkingConfig <- list(
      thinkingBudget = generation_config$thinkingBudget,
      includeThoughts = TRUE
    )
    generation_config$thinkingBudget <- NULL
  }

  contents <- as_json(provider, turns)

  # https://ai.google.dev/api/caching#Tool
  if (length(tools) > 0) {
    is_builtin <- map_lgl(tools, \(tool) S7_inherits(tool, ToolBuiltIn))
    funs <- as_json(provider, unname(tools))

    tools <- c(
      compact(list(functionDeclarations = funs[!is_builtin])),
      unlist(funs[is_builtin], recursive = FALSE)
    )
  } else {
    tools <- NULL
  }

  compact(list(
    contents = contents,
    tools = tools,
    systemInstruction = system,
    generationConfig = generation_config
  ))
}

method(chat_params, ProviderGoogleGemini) <- function(provider, params) {
  standardise_params(
    params,
    c(
      temperature = "temperature",
      topP = "top_p",
      topK = "top_k",
      frequencyPenalty = "frequency_penalty",
      presencePenalty = "presence_penalty",
      seed = "seed",
      maxOutputTokens = "max_tokens",
      responseLogprobs = "log_probs",
      stopSequences = "stop_sequences",
      thinkingBudget = "reasoning_tokens"
    )
  )
}

# Gemini -> ellmer --------------------------------------------------------------

method(stream_parse, ProviderGoogleGemini) <- function(provider, event) {
  if (is.null(event)) {
    NULL
  } else {
    jsonlite::parse_json(event$data)
  }
}
method(stream_content, ProviderGoogleGemini) <- function(provider, event) {
  parts <- event$candidates[[1]]$content$parts
  if (is.null(parts) || length(parts) == 0) {
    return(NULL)
  }

  part <- parts[[1]]
  if (isTRUE(part$thought) && !is.null(part$text)) {
    ContentThinking(part$text)
  } else if (!is.null(part$text)) {
    ContentText(part$text)
  }
}
method(stream_merge_chunks, ProviderGoogleGemini) <- function(
  provider,
  result,
  chunk
) {
  if (is.null(result)) {
    chunk
  } else {
    merge_gemini_chunks(result, chunk)
  }
}

method(value_tokens, ProviderGoogleGemini) <- function(provider, json) {
  # https://ai.google.dev/api/generate-content#UsageMetadata
  usage <- json$usageMetadata

  # Total token count for the generation request (prompt + response candidates).
  # Not documented, but appears to include thinking and tool use, i.e.
  # usage$promptTokenCount + usage$candidatesTokenCount +
  #  usage$toolUsePromptTokenCount + usage$thoughtsTokenCount ==
  #  usage$totalTokenCount
  total <- usage$totalTokenCount %||% 0

  # Number of tokens in the prompt. When cachedContent is set, this is
  # still the total effective prompt size meaning this includes the number
  # of tokens in the cached content.
  input <- usage$promptTokenCount %||% 0

  cached <- usage$cachedContentTokenCount %||% 0

  tokens(
    input = input - cached,
    output = total - input,
    cached_input = cached
  )
}

method(value_turn, ProviderGoogleGemini) <- function(
  provider,
  result,
  has_type = FALSE
) {
  message <- result$candidates[[1]]$content

  contents <- lapply(message$parts, function(content) {
    if (isTRUE(content$thought) && has_name(content, "text")) {
      ContentThinking(content$text)
    } else if (has_name(content, "text")) {
      if (has_type) {
        ContentJson(string = content$text)
      } else {
        ContentText(content$text)
      }
    } else if (has_name(content, "functionCall")) {
      extra <- compact(list(
        thoughtSignature = content$thoughtSignature
      ))
      ContentToolRequest(
        content$functionCall$name,
        content$functionCall$name,
        content$functionCall$args,
        extra = extra
      )
    } else if (has_name(content, "inlineData")) {
      ContentImageInline(
        type = content$inlineData$mimeType,
        data = content$inlineData$data
      )
    } else {
      cli::cli_abort(
        "Unknown content type with names {.str {names(content)}}.",
        .internal = TRUE
      )
    }
  })
  contents <- compact(contents)
  tokens <- value_tokens(provider, result)
  cost <- get_token_cost(provider, tokens)
  AssistantTurn(contents, json = result, tokens = unlist(tokens), cost = cost)
}

# ellmer -> Gemini --------------------------------------------------------------

# https://ai.google.dev/api/caching#Content
method(as_json, list(ProviderGoogleGemini, Turn)) <- function(
  provider,
  x,
  ...
) {
  if (is_system_turn(x)) {
    # System messages go in the top-level API parameter
  } else if (is_user_turn(x)) {
    x <- turn_contents_expand(x)
    list(
      role = x@role,
      parts = as_json(provider, x@contents, ...)
    )
  } else if (is_assistant_turn(x)) {
    list(role = "model", parts = as_json(provider, x@contents, ...))
  } else {
    cli::cli_abort("Unknown role {x@role}", .internal = TRUE)
  }
}


method(as_json, list(ProviderGoogleGemini, ToolDef)) <- function(
  provider,
  x,
  ...
) {
  compact(list(
    name = x@name,
    description = x@description,
    parameters = as_json(provider, x@arguments, ...)
  ))
}

method(as_json, list(ProviderGoogleGemini, ContentText)) <- function(
  provider,
  x,
  ...
) {
  if (identical(x@text, "")) {
    # Gemini tool call requests can include a Content with empty text,
    # but it doesn't like it if you send this back
    NULL
  } else {
    list(text = x@text)
  }
}

method(as_json, list(ProviderGoogleGemini, ContentThinking)) <- function(
  provider,
  x,
  ...
) {
  # https://ai.google.dev/gemini-api/docs/thinking
  list(thought = TRUE, text = x@thinking)
}

method(as_json, list(ProviderGoogleGemini, ContentPDF)) <- function(
  provider,
  x,
  ...
) {
  list(
    inlineData = list(
      mimeType = x@type,
      data = x@data
    )
  )
}

# https://ai.google.dev/api/caching#FileData
method(as_json, list(ProviderGoogleGemini, ContentUploaded)) <- function(
  provider,
  x,
  ...
) {
  list(
    fileData = list(
      mimeType = x@mime_type,
      fileUri = x@uri
    )
  )
}

# https://ai.google.dev/api/caching#FileData
method(as_json, list(ProviderGoogleGemini, ContentImageRemote)) <- function(
  provider,
  x,
  ...
) {
  cli::cli_abort("Gemini doesn't support remote images")
}

# https://ai.google.dev/api/caching#Blob
method(as_json, list(ProviderGoogleGemini, ContentImageInline)) <- function(
  provider,
  x,
  ...
) {
  list(
    inlineData = list(
      mimeType = x@type,
      data = x@data
    )
  )
}

# https://ai.google.dev/api/caching#FunctionCall
method(as_json, list(ProviderGoogleGemini, ContentToolRequest)) <- function(
  provider,
  x,
  ...
) {
  compact(list(
    functionCall = list(
      name = x@id,
      args = x@arguments
    ),
    thoughtSignature = x@extra$thoughtSignature
  ))
}

# https://ai.google.dev/api/caching#FunctionResponse
method(as_json, list(ProviderGoogleGemini, ContentToolResult)) <- function(
  provider,
  x,
  ...
) {
  list(
    functionResponse = list(
      name = x@request@id,
      response = list(value = tool_string(x))
    )
  )
}

method(as_json, list(ProviderGoogleGemini, TypeObject)) <- function(
  provider,
  x,
  ...
) {
  if (x@additional_properties) {
    cli::cli_abort("{.arg .additional_properties} not supported for Gemini.")
  }

  if (length(x@properties) == 0) {
    return(list())
  }

  required <- map_lgl(x@properties, function(prop) prop@required)

  compact(list(
    type = "object",
    description = x@description,
    properties = as_json(provider, x@properties, ...),
    required = as.list(names2(x@properties)[required])
  ))
}

# Gemini-specific merge logic --------------------------------------------------

merge_last <- function() {
  function(left, right, path = NULL) {
    right
  }
}

merge_identical <- function() {
  function(left, right, path = NULL) {
    if (!identical(left, right)) {
      stop(
        "Expected identical values, but got ",
        deparse(left),
        " and ",
        deparse(right)
      )
    }
    left
  }
}

merge_any_or_empty <- function() {
  function(left, right, path = NULL) {
    if (!is.null(left) && nzchar(left)) {
      left
    } else if (!is.null(right) && nzchar(right)) {
      right
    } else {
      ""
    }
  }
}

merge_optional <- function(merge_func) {
  function(left, right, path = NULL) {
    if (is.null(left) && is.null(right)) {
      NULL
    } else {
      merge_func(left, right, path)
    }
  }
}

merge_objects <- function(...) {
  spec <- list(...)
  function(left, right, path = NULL) {
    if (is.null(left)) {
      return(right)
    } else if (is.null(right)) {
      return(left)
    }

    # cat(paste(collapse = "", path), "\n")
    stopifnot(is.list(left), is.list(right), all(nzchar(names(spec))))
    mapply(
      names(spec),
      spec,
      FUN = function(key, value) {
        value(left[[key]], right[[key]], c(path, ".", key))
      },
      USE.NAMES = TRUE,
      SIMPLIFY = FALSE
    )
  }
}

merge_candidate_lists <- function(...) {
  merge_unindexed <- merge_objects(...)
  merge_indexed <- merge_objects(index = merge_identical(), ...)

  function(left, right, path = NULL) {
    if (length(left) == 1 && length(right) == 1) {
      list(merge_unindexed(left[[1]], right[[1]], c(path, "[]")))
    } else {
      # left and right are lists of objects with [["index"]]
      # We need to find the elements that have matching indices and merge them
      left_indices <- vapply(left, `[[`, integer(1), "index")
      right_indices <- vapply(right, `[[`, integer(1), "index")
      # I know this seems weird, but according to Google's Go SDK, we should
      # only retain indices on the right that *already* appear on the left.
      # Citations:
      # https://github.com/google/generative-ai-go/blob/3d14f4039eaef321b15bcbf70839389d7f000233/genai/client_test.go#L655
      # https://github.com/google/generative-ai-go/blob/3d14f4039eaef321b15bcbf70839389d7f000233/genai/client.go#L396
      lapply(left_indices, function(index) {
        left_item <- left[[which(left_indices == index)]]
        right_item <- right[[which(right_indices == index)]]
        if (is.null(right_item)) {
          left_item
        } else {
          merge_indexed(left_item, right_item, c(path, "[", index, "]"))
        }
      })
    }
  }
}

merge_append <- function() {
  function(left, right, path = NULL) {
    c(left, right)
  }
}

merge_parts <- function() {
  function(left, right, path = NULL) {
    joined <- c(left, right)

    # Identify text parts
    is_text <- map_lgl(joined, ~ is.list(.x) && identical(names(.x), "text"))

    # Create groups for contiguous sections
    groups <- cumsum(c(TRUE, diff(is_text) != 0))

    # Split into groups and process each
    split_parts <- split(joined, groups)
    merged_split_parts <- map2(
      split_parts,
      split(is_text, groups),
      function(parts, is_text_group) {
        if (!is_text_group[[1]]) {
          # Non-text group: return parts unchanged
          return(parts)
        } else {
          # Text group: merge text values
          text_values <- map_chr(parts, ~ .x[["text"]])
          list(list(text = paste0(text_values, collapse = "")))
        }
      }
    )
    unlist(merged_split_parts, recursive = FALSE, use.names = FALSE)
  }
}

# Put it all together...
# https://ai.google.dev/api/generate-content#v1beta.GenerateContentResponse
merge_gemini_chunks <- merge_objects(
  candidates = merge_candidate_lists(
    content = merge_objects(
      role = merge_any_or_empty(),
      parts = merge_parts()
    ),
    finishReason = merge_last(),
    safetyRatings = merge_last(),
    citationMetadata = merge_optional(
      merge_objects(citationSources = merge_append())
    ),
    tokenCount = merge_last()
  ),
  promptFeedback = merge_last(),
  usageMetadata = merge_last()
)

default_google_credentials <- function(
  error_call = caller_env(),
  variant = c("gemini", "vertex")
) {
  variant <- arg_match(variant)

  api_key <- Sys.getenv("GOOGLE_API_KEY")
  if (variant == "gemini" && api_key == "") {
    api_key <- Sys.getenv("GEMINI_API_KEY")
  }
  if (nzchar(api_key)) {
    return(\() api_key)
  }

  gemini_scope <- switch(
    variant,
    gemini = "https://www.googleapis.com/auth/generative-language.retriever",
    # https://github.com/googleapis/python-genai/blob/cc9e470326e0c1b84ec3ce9891c9f96f6c74688e/google/genai/_api_client.py#L184
    vertex = "https://www.googleapis.com/auth/cloud-platform"
  )

  # Detect viewer-based credentials from Posit Connect.
  if (has_connect_viewer_token(scope = gemini_scope)) {
    return(function() {
      token <- connectcreds::connect_viewer_token(scope = gemini_scope)
      list(Authorization = paste("Bearer", token$access_token))
    })
  }

  if (is_testing()) {
    testthat::skip_if_not_installed("gargle")
  }

  check_installed("gargle", "for Google authentication")
  gargle::with_cred_funs(
    funs = list(
      # We don't want to use *all* of gargle's default credential functions --
      # in particular, we don't want to try and authenticate using the bundled
      # OAuth client -- so winnow down the list.
      credentials_app_default = gargle::credentials_app_default
    ),
    {
      token <- gargle::token_fetch(scopes = gemini_scope)
    },
    action = "replace"
  )

  if (is.null(token) && is_testing()) {
    testthat::skip("no Google credentials available")
  }

  if (is.null(token) && is_interactive()) {
    return(function() {
      function(req) {
        req_oauth_auth_code(
          req,
          client = gemini_client(),
          auth_url = "https://accounts.google.com/o/oauth2/auth",
          scope = "https://www.googleapis.com/auth/generative-language.retriever"
        )
      }
    })
  }

  if (is.null(token)) {
    cli::cli_abort(
      c(
        "No Google credentials are available.",
        "i" = "Try suppling an API key or configuring Google's application default credentials."
      ),
      call = error_call
    )
  }

  # gargle emits an httr-style token, which we awkwardly shim into something
  # httr2 can work with.

  if (!token$can_refresh()) {
    # TODO: Not really sure what to do in this case when the token expires.
    return(function() {
      list(Authorization = paste("Bearer", token$credentials$access_token))
    })
  }

  # gargle tokens don't track the expiry time, so we do it ourselves (with a
  # grace period).
  expiry <- Sys.time() + token$credentials$expires_in - 5
  return(function() {
    if (expiry < Sys.time()) {
      token$refresh()
    }
    list(Authorization = paste("Bearer", token$credentials$access_token))
  })
}

google_oauth_reset <- function() {
  httr2::oauth_cache_clear(gemini_client())
}

gemini_client <- function() {
  httr2::oauth_client(
    id = "148439353047-kit3pok9u920mhmqbc3c0pdr50bvb7pt.apps.googleusercontent.com",
    secret = httr2::obfuscated(
      "o2yDPr_4BNgZvhLT9kIZS6jAYp43sAzAjMrmW60FUC-N4btRmTwOQ1650vS2pDRSvbKK"
    ),
    token_url = "https://oauth2.googleapis.com/token",
    name = "gemini-r-client"
  )
}

# Pricing ----------------------------------------------------------------------

# Models -----------------------------------------------------------------------

#' @export
#' @rdname chat_google_gemini
models_google_gemini <- function(
  base_url = "https://generativelanguage.googleapis.com/v1beta/",
  api_key = NULL,
  credentials = NULL
) {
  check_string(base_url)

  credentials <- as_credentials(
    "models_google_gemini",
    default_google_credentials(variant = "gemini"),
    credentials = credentials,
    api_key = api_key
  )

  models_google(base_url, credentials = credentials, variant = "gemini")
}

#' @rdname chat_google_gemini
#' @export
models_google_vertex <- function(location, project_id, credentials = NULL) {
  check_string(location)
  check_string(project_id)

  credentials <- credentials %||% default_google_credentials(variant = "vertex")
  check_credentials(credentials)

  base_url <- paste_c(
    c("https://", google_location(location), "aiplatform.googleapis.com"),
    "/v1beta1",
    "/publishers/google/"
  )

  models_google(base_url, project_id = project_id, variant = "vertex")
}

models_google <- function(
  base_url = "https://generativelanguage.googleapis.com/v1beta/",
  credentials,
  project_id = NULL,
  variant = c("gemini", "vertex")
) {
  variant <- arg_match(variant)

  provider <- ProviderGoogleGemini(
    name = "Google/Gemini",
    model = "",
    base_url = base_url,
    # https://cloud.google.com/docs/authentication/troubleshoot-adc#user-creds-client-based
    credentials = credentials
  )

  req <- base_request(provider)
  if (variant == "vertex") {
    req <- req_headers(req, `x-goog-user-project` = project_id)
  }
  req <- req_headers(req, !!!provider@extra_headers)
  req <- req_url_path_append(req, "/models")
  resp <- req_perform(req)

  json <- resp_body_json(resp)

  if (variant == "vertex") {
    name <- map_chr(json$publisherModels, "[[", "name")
    name <- gsub("^publishers/google/models/", "", name)
    # this is the closest to "generateContent" in "supportedGenerationMethods" for Gemini
    # https://cloud.google.com/vertex-ai/docs/reference/rest/v1beta1/publishers.models
    can_generate <- json$publisherModels |>
      map_lgl(\(x) "openGenerationAiStudio" %in% names(x$supportedActions))
  } else {
    name <- map_chr(json$models, "[[", "name")
    name <- gsub("^models/", "", name)
    display_name <- map_chr(json$models, "[[", "displayName")

    methods <- map(json$models, \(x) unlist(x$supportedGenerationMethods))
    can_generate <- map_lgl(methods, \(x) "generateContent" %in% x)
  }

  df <- data.frame(id = name)
  df <- cbind(df, match_prices(provider@name, df$id))
  df <- df[can_generate, ]
  unrowname(df[order(df$id), ])
}

# for location "global", there is no location in the final base URL
# https://github.com/googleapis/python-genai/blob/cc9e470326e0c1b84ec3ce9891c9f96f6c74688e/google/genai/_api_client.py#L646-L654
google_location <- function(location) {
  if (location == "global") "" else paste0(location, "-")
}

test_that("can upload a file an ask questions about it", {
  vcr::local_cassette("google-upload")
  upload <- google_upload(test_path("apples.pdf"))

  chat <- chat_google_gemini_test()
  response <- chat$chat("What's the title of this document?", upload)
  expect_match(response, "Apples are tasty")
  expect_match(chat$chat("What apple is not tasty?"), "red delicious")
})

#' @include provider-openai-compatible.R
#' @include content.R
NULL

#' Chat with a model hosted on Snowflake
#'
#' @description
#' The Snowflake provider allows you to interact with LLM models available
#' through the [Cortex LLM REST API](https://docs.snowflake.com/en/user-guide/snowflake-cortex/cortex-llm-rest-api).
#'
#' ## Authentication
#'
#' `chat_snowflake()` picks up the following ambient Snowflake credentials:
#'
#' - A static OAuth token defined via the `SNOWFLAKE_TOKEN` environment
#'   variable.
#' - Key-pair authentication credentials defined via the `SNOWFLAKE_USER` and
#'   `SNOWFLAKE_PRIVATE_KEY` (which can be a PEM-encoded private key or a path
#'   to one) environment variables.
#' - Posit Workbench-managed Snowflake credentials for the corresponding
#'   `account`.
#' - Viewer-based credentials on Posit Connect. Requires the \pkg{connectcreds}
#'   package.
#'
#' ## Known limitations
#' Note that Snowflake-hosted models do not support images.
#'
#' @inheritParams chat_openai
#' @param account A Snowflake [account identifier](https://docs.snowflake.com/en/user-guide/admin-account-identifier),
#'   e.g. `"testorg-test_account"`. Defaults to the value of the
#'   `SNOWFLAKE_ACCOUNT` environment variable.
#' @param credentials A list of authentication headers to pass into
#'   [`httr2::req_headers()`], a function that returns them when called, or
#'   `NULL`, the default, to use ambient credentials.
#' @param model `r param_model("claude-3-7-sonnet")`
#' @inherit chat_openai return
#' @examplesIf has_credentials("snowflake")
#' chat <- chat_snowflake()
#' chat$chat("Tell me a joke in the form of a SQL query.")
#' @export
chat_snowflake <- function(
  system_prompt = NULL,
  account = snowflake_account(),
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  echo = c("none", "output", "all"),
  api_headers = character()
) {
  check_string(account, allow_empty = FALSE)
  model <- set_default(model, "claude-3-7-sonnet")
  params <- params %||% params()
  echo <- check_echo(echo)

  credentials <- credentials %||% default_snowflake_credentials(account)
  check_credentials(credentials)

  provider <- ProviderSnowflakeCortex(
    name = "Snowflake/Cortex",
    base_url = snowflake_url(account),
    account = account,
    credentials = credentials,
    model = model,
    params = params,
    extra_args = api_args,
    extra_headers = api_headers
  )

  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderSnowflakeCortex <- new_class(
  "ProviderSnowflakeCortex",
  parent = ProviderOpenAICompatible,
  properties = list(
    account = prop_string()
  )
)

method(base_request, ProviderSnowflakeCortex) <- function(provider) {
  req <- request(provider@base_url)
  req <- ellmer_req_credentials(req, provider@credentials())
  req <- ellmer_req_robustify(req)
  # Snowflake uses the User Agent header to identify "parter applications", so
  # identify requests as coming from "r_ellmer" (unless an explicit partner
  # application is set via the ambient SF_PARTNER environment variable).
  req <- ellmer_req_user_agent(req, Sys.getenv("SF_PARTNER"))

  # Snowflake-specific error response handling:
  req <- req_error(req, body = function(resp) resp_body_json(resp)$message)

  req
}

method(chat_path, ProviderSnowflakeCortex) <- function(provider) {
  "/api/v2/cortex/inference:complete"
}

# See: https://docs.snowflake.com/en/user-guide/snowflake-cortex/cortex-llm-rest-api#api-reference
method(chat_body, ProviderSnowflakeCortex) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  messages <- as_json(provider, turns)
  tools <- as_json(provider, unname(tools))

  if (!is.null(type)) {
    # Note: Snowflake uses a slightly different format than OpenAI.
    response_format <- list(type = "json", schema = as_json(provider, type))
  } else {
    response_format <- NULL
  }

  params <- chat_params(provider, provider@params)
  compact(
    list2(
      messages = messages,
      model = provider@model,
      !!!params,
      stream = stream,
      tools = tools,
      response_format = response_format
    )
  )
}

method(as_json, list(ProviderSnowflakeCortex, TypeObject)) <- function(
  provider,
  x,
  ...
) {
  # Unlike OpenAI, Snowflake does not support the "additionalProperties" field.
  names <- names2(x@properties)
  required <- map_lgl(x@properties, function(prop) prop@required)
  properties <- as_json(provider, x@properties, ...)
  names(properties) <- names
  list(
    type = "object",
    description = x@description %||% "",
    properties = properties,
    required = as.list(names[required])
  )
}

# See: https://docs.snowflake.com/en/user-guide/snowflake-cortex/cortex-llm-rest-api#optional-json-arguments
method(chat_params, ProviderSnowflakeCortex) <- function(provider, params) {
  standardise_params(
    params,
    c(
      temperature = "temperature",
      top_p = "top_p",
      max_tokens = "max_tokens"
    )
  )
}

# Snowflake -> ellmer --------------------------------------------------------

method(stream_content, ProviderSnowflakeCortex) <- function(provider, event) {
  if (length(event$choices) == 0) {
    return(NULL)
  }
  delta <- event$choices[[1]]$delta
  if (is.null(delta) || !identical(delta$type, "text")) {
    return(NULL)
  }
  text <- delta[["content"]] %||% delta[["text"]]
  if (is.null(text) || !nzchar(text)) {
    return(NULL)
  }
  ContentText(text)
}

method(stream_merge_chunks, ProviderSnowflakeCortex) <- function(
  provider,
  result,
  chunk
) {
  delta <- chunk$choices[[1]]$delta

  if (is.null(result)) {
    if (delta$type == "tool_use") {
      content_list <- list(
        list(
          type = "tool_use",
          tool_use_id = delta$tool_use_id,
          name = delta$name,
          input = delta$input %||% ""
        )
      )
    } else {
      content_list <- list(
        list(
          type = "text",
          text = delta$text %||% ""
        )
      )
    }
    # Non-streaming responses use "message" instead of "delta".
    chunk$choices[[1]]$message <- list(content_list = content_list)
    chunk$choices[[1]]$delta <- NULL
    return(chunk)
  }

  content_list <- result$choices[[1]]$message$content_list

  if (delta$type == "text") {
    text_idx <- NULL
    for (i in rev(seq_along(content_list))) {
      if (identical(content_list[[i]]$type, "text")) {
        text_idx <- i
        break
      }
    }
    if (is.null(text_idx)) {
      content_list[[length(content_list) + 1L]] <- list(
        type = "text",
        text = delta$text %||% ""
      )
    } else {
      paste(content_list[[text_idx]]$text) <- delta$text %||% ""
    }
  } else if (delta$type == "tool_use") {
    if (!is.null(delta$tool_use_id)) {
      content_list[[length(content_list) + 1L]] <- list(
        type = "tool_use",
        tool_use_id = delta$tool_use_id,
        name = delta$name,
        input = delta$input %||% ""
      )
    } else if (!is.null(delta$input)) {
      for (i in rev(seq_along(content_list))) {
        if (identical(content_list[[i]]$type, "tool_use")) {
          paste(content_list[[i]]$input) <- delta$input
          break
        }
      }
    }
  } else {
    cli::cli_abort(
      "Unsupported content type {.str {delta$type}}.",
      .internal = TRUE
    )
  }

  result$choices[[1]]$message$content_list <- content_list
  result$usage <- chunk$usage
  result
}

method(value_tokens, ProviderSnowflakeCortex) <- function(provider, json) {
  usage <- json$usage
  tokens(
    input = usage$prompt_tokens %||% 0,
    output = usage$completion_tokens %||% 0
  )
}

method(value_turn, ProviderSnowflakeCortex) <- function(
  provider,
  result,
  has_type = FALSE
) {
  raw_content <- result$choices[[1]]$message$content_list
  contents <- compact(
    lapply(raw_content, function(content) {
      if (identical(content$type, "text")) {
        if (!nzchar(content$text %||% "")) {
          return(NULL)
        }
        if (has_type) {
          ContentJson(string = content$text)
        } else {
          ContentText(content$text)
        }
      } else if (identical(content$type, "tool_use")) {
        # Streaming produces flat format; non-streaming has nested tool_use
        if (!is.null(content[["tool_use"]])) {
          content <- content[["tool_use"]]
        }
        input <- content[["input"]] %||% ""
        id <- content[["tool_use_id"]]
        name <- content[["name"]]
        if (is_string(input)) {
          input <- jsonlite::parse_json(input)
        }
        ContentToolRequest(id, name, input %||% list())
      } else {
        cli::cli_abort(
          "Unknown content type {.str {content$type}}.",
          .internal = TRUE
        )
      }
    })
  )
  tokens <- value_tokens(provider, result)
  cost <- get_token_cost(provider, tokens)
  AssistantTurn(contents, json = result, tokens = unlist(tokens), cost = cost)
}

# ellmer -> Snowflake --------------------------------------------------------

method(as_json, list(ProviderSnowflakeCortex, Turn)) <- function(
  provider,
  x,
  ...
) {
  # Attempting to omit the `content` field and use `content_list` instead
  # yields:
  #
  # > messages[0].content cannot be empty string (was '')
  #
  # So we emulate what Snowflake do and put the text content in both.
  if (S7_inherits(x@contents[[1]], ContentText)) {
    content <- x@contents[[1]]@text
    if (nchar(content) == 0) {
      content <- "<empty>"
    }
  } else if (S7_inherits(x@contents[[1]], ContentJson)) {
    # Match the existing as_json() implementation.
    content <- "<structured data/>"
  } else if (S7_inherits(x@contents[[1]], ContentToolResult)) {
    # Completely undocumented, but: it seems like the model is expecting the
    # tool result in textual format here, too -- otherwise it gets confused,
    # like it can't see the output.
    content <- tool_string(x@contents[[1]])
  } else if (S7_inherits(x@contents[[1]], ContentToolRequest)) {
    # Tool-only response (no preceding text).
    content <- "<empty>"
  } else {
    cli::cli_abort("Unsupported content type: {.cls {class(x@contents[[1]])}}.")
  }
  x <- turn_contents_expand(x)
  list(
    role = x@role,
    content_list = as_json(provider, x@contents, ...)
  )
}

# See: https://docs.snowflake.com/en/user-guide/snowflake-cortex/cortex-llm-rest-api#tools-configuration
method(as_json, list(ProviderSnowflakeCortex, ToolDef)) <- function(
  provider,
  x,
  ...
) {
  list(
    tool_spec = compact(
      list(
        type = "generic",
        name = x@name,
        description = x@description,
        input_schema = as_json(provider, x@arguments, ...)
      )
    )
  )
}

# See: https://docs.snowflake.com/en/user-guide/snowflake-cortex/cortex-llm-rest-api#tools-configuration
method(as_json, list(ProviderSnowflakeCortex, ContentToolRequest)) <- function(
  provider,
  x,
  ...
) {
  input <- x@arguments
  if (length(input) == 0) {
    # Snowflake requires an empty object, rather than an empty array.
    input <- set_names(list())
  }
  list(
    type = "tool_use",
    tool_use = list(
      tool_use_id = x@id,
      name = x@name,
      input = input
    )
  )
}

# See: https://docs.snowflake.com/en/user-guide/snowflake-cortex/cortex-llm-rest-api#tool-results
method(as_json, list(ProviderSnowflakeCortex, ContentToolResult)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "tool_results",
    tool_results = compact(
      list(
        tool_use_id = x@request@id,
        name = x@request@name,
        content = list(
          list(type = "text", text = tool_string(x))
        ),
        # TODO: Is this the correct format?
        status = if (tool_errored(x)) "error"
      )
    )
  )
}

# Utilities ------------------------------------------------------------------

snowflake_account <- function() {
  key_get("SNOWFLAKE_ACCOUNT")
}

snowflake_url <- function(account) {
  paste0("https://", account, ".snowflakecomputing.com")
}

default_snowflake_credentials <- function(account = snowflake_account()) {
  # Detect viewer-based credentials from Posit Connect.
  url <- snowflake_url(account)
  if (is_installed("connectcreds") && connectcreds::has_viewer_token(url)) {
    return(function() {
      token <- connectcreds::connect_viewer_token(url)
      list(
        Authorization = paste("Bearer", token$access_token),
        `X-Snowflake-Authorization-Token-Type` = "OAUTH"
      )
    })
  }

  token <- Sys.getenv("SNOWFLAKE_TOKEN")
  if (nchar(token) != 0) {
    return(function() {
      list(
        Authorization = paste("Bearer", token),
        # See: https://docs.snowflake.com/en/developer-guide/snowflake-rest-api/authentication#using-oauth
        `X-Snowflake-Authorization-Token-Type` = "OAUTH"
      )
    })
  }

  # Support for Snowflake key-pair authentication.
  # See: https://docs.snowflake.com/en/developer-guide/snowflake-rest-api/authentication#generate-a-jwt-token
  user <- Sys.getenv("SNOWFLAKE_USER")
  private_key <- Sys.getenv("SNOWFLAKE_PRIVATE_KEY")
  if (nchar(user) != 0 && nchar(private_key) != 0) {
    check_installed("jose", "for key-pair authentication")
    key <- openssl::read_key(private_key)
    return(function() {
      token <- snowflake_keypair_token(account, user, key)
      list(
        Authorization = paste("Bearer", token),
        `X-Snowflake-Authorization-Token-Type` = "KEYPAIR_JWT"
      )
    })
  }

  # Check for Workbench-managed credentials.
  sf_home <- Sys.getenv("SNOWFLAKE_HOME")
  if (grepl("posit-workbench", sf_home, fixed = TRUE)) {
    token <- workbench_snowflake_token(account, sf_home)
    if (!is.null(token)) {
      return(function() {
        # Ensure we get an up-to-date token.
        token <- workbench_snowflake_token(account, sf_home)
        list(
          Authorization = paste("Bearer", token),
          `X-Snowflake-Authorization-Token-Type` = "OAUTH"
        )
      })
    }
  }

  if (is_testing()) {
    testthat::skip("no Snowflake credentials available")
  }

  cli::cli_abort("No Snowflake credentials are available.")
}

snowflake_keypair_token <- function(
  account,
  user,
  key,
  cache = snowflake_keypair_cache(account, key),
  lifetime = 600L,
  reauth = FALSE
) {
  # Producing a signed JWT is a fairly expensive operation (in the order of
  # ~10ms), but adding a cache speeds this up approximately 500x.
  creds <- cache$get()
  if (reauth || is.null(creds) || creds$expiry < Sys.time()) {
    cache$clear()
    expiry <- Sys.time() + lifetime
    # We can't use openssl::fingerprint() here because it uses a different
    # algorithm.
    fp <- openssl::base64_encode(
      openssl::sha256(openssl::write_der(key$pubkey))
    )
    if (grepl(".+\\.privatelink$", account)) {
      # account identifier is everything up to the first period
      account <- gsub("^([^.]*).+", "\\1", account)
    }
    sub <- toupper(paste0(account, ".", user))
    iss <- paste0(sub, ".SHA256:", fp)
    # Note: Snowflake employs a malformed issuer claim, so we have to inject it
    # manually after jose's validation phase.
    claim <- jwt_claim("dummy", sub, exp = as.integer(expiry))
    claim$iss <- iss
    creds <- list(expiry = expiry, token = jwt_encode_sig(claim, key))
    cache$set(creds)
  }
  creds$token
}

snowflake_keypair_cache <- function(account, key) {
  credentials_cache(key = hash(c("sf", account, openssl::fingerprint(key))))
}

# Credential handling ----------------------------------------------------------

snowflake_credentials_exist <- function(...) {
  tryCatch(
    is_list(default_snowflake_credentials(...)),
    error = function(e) FALSE
  )
}

# Reads Posit Workbench-managed Snowflake credentials from a
# $SNOWFLAKE_HOME/connections.toml file, as used by the Snowflake Connector for
# Python implementation. The file will look as follows:
#
# [workbench]
# account = "account-id"
# token = "token"
# authenticator = "oauth"
workbench_snowflake_token <- function(account, sf_home) {
  cfg <- readLines(file.path(sf_home, "connections.toml"))
  # We don't attempt a full parse of the TOML syntax, instead relying on the
  # fact that this file will always contain only one section.
  if (!any(grepl(account, cfg, fixed = TRUE))) {
    # The configuration doesn't actually apply to this account.
    return(NULL)
  }
  line <- grepl("token = ", cfg, fixed = TRUE)
  token <- gsub("token = ", "", cfg[line])
  if (nchar(token) == 0) {
    return(NULL)
  }
  # Drop enclosing quotes.
  gsub("\"", "", token)
}

#' Create metadata for a tool
#'
#' @description
#' In order to use a function as a tool in a chat, you need to craft the right
#' call to [tool()]. This function helps you do that for documented functions by
#' extracting the function's R documentation and using an LLM to generate the
#' `tool()` call. It's meant to be used interactively while writing your
#' code, not as part of your final code.
#'
#' If the function has package documentation, that will be used. Otherwise, if
#' the source code of the function can be automatically detected, then the
#' comments immediately preceding the function are used (especially helpful if
#' those are roxygen2 comments). If neither are available, then just the function
#' signature is used.
#'
#' Note that this function is inherently imperfect. It can't handle all possible
#' R functions, because not all parameters are suitable for use in a tool call
#' (for example, because they're not serializable to simple JSON objects). The
#' documentation might not specify the expected shape of arguments to the level
#' of detail that would allow an exact JSON schema to be generated. Please be
#' sure to review the generated code before using it!
#'
#' @param topic A symbol or string literal naming the function to create
#'   metadata for. Can also be an expression of the form `pkg::fun`.
#' @param chat A `Chat` object used to generate the output. If `NULL`
#'   (the default) uses [chat_openai()].
#' @param echo Emit the registration code to the console. Defaults to `TRUE` in
#'   interactive sessions.
#' @param verbose If `TRUE`, print the input we send to the LLM, which may be
#'   useful for debugging unexpectedly poor results.
#'
#' @return A `register_tool` call that you can copy and paste into your code.
#'   Returned invisibly if `echo` is `TRUE`.
#'
#' @examples
#' \dontrun{
#'   # These are all equivalent
#'   create_tool_def(rnorm)
#'   create_tool_def(stats::rnorm)
#'   create_tool_def("rnorm")
#'   create_tool_def("rnorm", chat = chat_azure_openai())
#' }
#'
#' @export
create_tool_def <- function(
  topic,
  chat = NULL,
  echo = interactive(),
  verbose = FALSE
) {
  expr <- enexpr(topic)

  if (is.null(chat)) {
    chat <- chat_openai()
  } else if (is_chat(chat)) {
    chat <- chat$clone()
  } else {
    stop_input_type(chat, "a <Chat> object", allow_null = TRUE)
  }
  check_echo(echo)
  check_bool(verbose)

  tool_prompt <- read_file(system.file("tool_prompt.md", package = "ellmer"))
  chat$set_system_prompt(tool_prompt)

  pkg <- NULL
  fun <- format(expr)

  # Ensure `expr` is a string literal, a symbol, or an expression of the form
  # `pkg::fun` or `pkg:::fun`.
  if (is_call(expr)) {
    if (
      !identical(expr[[1]], quote(`::`)) ||
        !is_symbol(expr[[2]]) ||
        !is_symbol(expr[[3]])
    ) {
      cli::cli_abort(
        "Expected a symbol or a string literal, or an expression of the form `pkg::fun` or `pkg:::fun`."
      )
    }
    pkg <- as.character(expr[[2]])
    fun <- as.character(expr[[3]])
  } else if (is_symbol(expr)) {
    fun <- as.character(expr)
  } else if (!is_string(expr)) {
    cli::cli_abort(
      "Expected a symbol or a string literal, or an expression of the form `pkg::fun` or `pkg:::fun`."
    )
  }

  help_text <- get_help_text(fun, pkg)
  if (is.null(help_text)) {
    # Package help documentation wasn't found; try comments instead
    help_text <- paste(extract_comments_and_signature(topic), collapse = "\n")
  }

  topic_str <- format(expr)

  payload <- paste_c(
    c("<name>", topic_str, "</name>\n\n"),
    c("<documentation>\n", help_text, "</documentation>\n\n")
  )

  if (isTRUE(verbose)) {
    cli::cli_rule(cli::style_bold("Prompt"))
    message(payload, "\n\n")
    cli::cli_rule(cli::style_bold("Response"))
  }

  chat$chat(payload, echo = echo)
}

help_to_text <- function(help_files) {
  file_contents <- NULL
  fake_pager <- function(files, header, title, delete.file) {
    if (delete.file) {
      on.exit(unlink(files))
    }

    for (file in files) {
      file_contents <<- c(
        file_contents,
        readLines(files, warn = FALSE),
        "\n"
      )
    }
  }

  op <- options(pager = fake_pager)
  on.exit(options(op), add = TRUE)
  rd_opts <- tools::Rd2txt_options(underline_titles = FALSE)
  on.exit(tools::Rd2txt_options(rd_opts), add = TRUE)

  print(help_files)

  paste(file_contents, collapse = "\n")
}

get_help_text <- function(topic, package = NULL) {
  # The extra parens around topic and package are to ensure that they're
  # evaluated; see the last example on ?utils::help.
  help_files <- utils::help((topic), package = (package), help_type = "text")
  if (length(help_files) == 0) {
    return(NULL)
  } else {
    help_to_text(help_files)
  }
}

# If the function source cannot be found, at least provide the function
# signature
get_signature <- function(func) {
  args <- utils::capture.output(args(func))
  paste0(paste0(args[-length(args)], collapse = "\n"), " ...")
}

is_file_accessible <- function(filename) {
  !is.null(filename) && file.access(filename, 4) == 0
}

find_comments_start <- function(lines, start_line) {
  while (start_line > 1 && grepl("^\\s*#", lines[start_line - 1])) {
    start_line <- start_line - 1
  }
  start_line
}

extract_comments <- function(filename, start_line) {
  lines <- readLines(filename)
  comments_start <- find_comments_start(lines, start_line)
  if (comments_start == start_line) {
    return(NULL)
  }
  comments_lines <- lines[comments_start:(start_line - 1)]
  paste(comments_lines, collapse = "\n")
}

# Given a function, extract the comments (if any) and function signature. Relies
# on srcref information, so functions entered at the repl will be missing
# documentation.
extract_comments_and_signature <- function(func) {
  src <- attr(func, "srcref")
  sig <- get_signature(func)

  if (is.null(src)) {
    return(sig)
  }

  filename <- attr(src, "srcfile")$filename
  if (!is_file_accessible(filename)) {
    return(sig)
  }

  start_line <- src[1]
  comments <- extract_comments(filename, start_line)
  if (is.null(comments)) {
    return(sig)
  }

  paste0(comments, "\n", sig)
}

test_that("help topic extraction works", {
  print_help <- get_help_text("print")
  expect_type(print_help, "character")
  expect_gt(nchar(print_help), 50)

  expect_identical(get_help_text("print", "base"), print_help)
})

test_that("roxygen2 comment extraction works", {
  sys.source(test_path("tools-def.R"), environment(), keep.source = TRUE)
  aliased_function <- has_roxygen_comments

  expect_snapshot(extract_comments_and_signature(has_roxygen_comments))
  expect_snapshot(extract_comments_and_signature(aliased_function))
  expect_snapshot(extract_comments_and_signature(indented_comments))
  expect_snapshot(extract_comments_and_signature(no_srcfile))
})

test_that("basic signature extraction works", {
  sys.source(test_path("tools-def.R"), environment(), keep.source = TRUE)
  expect_snapshot(extract_comments_and_signature(no_roxygen_comments))
})

test_that("checks its inputs", {
  expect_snapshot(error = TRUE, {
    create_tool_def(print, chat = 1)
  })
})

MockedChat <- R6::R6Class(
  "MockedChat",
  inherit = Chat,
  public = list(
    i = 0,
    saved_chats = character(),

    initialize = function(saved_chats) {
      self$saved_chats <- saved_chats
    },

    chat = function(...) {
      self$i <- self$i + 1
      self$saved_chats[self$i]
    }
  )
)

mocked_chat <- function(chats) {
  MockedChat$new(saved_chats = chats)
}

test_that("as_chat() takes string or chat", {
  chat <- chat_openai_test()
  expect_equal(as_chat(chat), chat)

  expect_equal(as_chat("openai/gpt-4.1-nano"), chat("openai/gpt-4.1-nano"))

  chat <- 1
  expect_snapshot(as_chat(chat), error = TRUE)
})

is_chat <- function(x) {
  inherits(x, "Chat")
}

as_chat <- function(x, error_arg = caller_arg(x), error_call = caller_env()) {
  if (is_chat(x)) {
    x
  } else if (is_string(x)) {
    chat(x)
  } else {
    stop_input_type(
      x,
      c("a string", "a <Chat> object"),
      arg = error_arg,
      call = error_call
    )
  }
}

#' @keywords internal
#' @importFrom R6 R6Class
"_PACKAGE"

the <- new_environment()
the$credentials_cache <- new_environment()

silence_r_cmd_check_note <- function() {
  later::later()
}

## usethis namespace: start
#' @import httr2
#' @import rlang
#' @import S7
#' @importFrom lifecycle deprecated
## usethis namespace: end
NULL

#' Chat with a model hosted on the GitHub model marketplace
#'
#' @description
#' GitHub Models hosts a number of open source and OpenAI models. To access the
#' GitHub model marketplace, you will need to apply for and be accepted into the
#' beta access program. See <https://github.com/marketplace/models> for details.
#'
#' This function is a lightweight wrapper around [chat_openai_compatible()] with
#' the defaults tweaked for the GitHub Models marketplace.
#'
#' GitHub also supports the Azure AI Inference SDK, which you can use by setting
#' `base_url` to `"https://models.inference.ai.azure.com/"`. This endpoint was
#' used in \pkg{ellmer} v0.3.0 and earlier.
#'
#' @family chatbots
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("GITHUB_PAT")`
#' @param model `r param_model("gpt-4o")`
#' @param params Common model parameters, usually created by [params()].
#' @export
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_github()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_github <- function(
  system_prompt = NULL,
  base_url = "https://models.github.ai/inference/",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
) {
  check_installed("gitcreds")

  model <- set_default(model, "gpt-4.1")
  echo <- check_echo(echo)

  credentials <- as_credentials(
    "chat_github",
    function() github_key(),
    credentials = credentials,
    api_key = api_key
  )

  # https://docs.github.com/en/rest/models/inference?apiVersion=2022-11-28
  params <- params %||% params()

  chat_openai_compatible(
    name = "gitHub",
    system_prompt = system_prompt,
    base_url = base_url,
    credentials = credentials,
    model = model,
    params = params,
    api_args = api_args,
    echo = echo,
    api_headers = api_headers
  )
}

github_key <- function() {
  withCallingHandlers(
    gitcreds::gitcreds_get()$password,
    error = function(cnd) {
      cli::cli_abort(
        "Failed to find git credentials or GITHUB_PAT env var",
        parent = cnd
      )
    }
  )
}

#' @rdname chat_github
#' @export
models_github <- function(
  base_url = "https://models.github.ai/",
  api_key = NULL,
  credentials = NULL
) {
  credentials <- as_credentials(
    "models_github",
    function() github_key(),
    credentials = credentials,
    api_key = api_key
  )

  provider <- ProviderOpenAI(
    name = "github",
    model = "",
    credentials = credentials,
    base_url = base_url
  )

  req <- base_request(provider)
  req <- req_url_path_append(req, "/catalog/models")
  resp <- req_perform(req)

  json <- resp_body_json(resp)

  if (grepl("models.inference.ai.azure.com", base_url, fixed = TRUE)) {
    # Support listing models from the older Azure endpoint (ellmer <= 0.3.0)
    id <- map_chr(json, "[[", "name")
    publisher <- map_chr(json, "[[", "publisher")
    license <- map_chr(json, "[[", "license")
    task <- map_chr(json, "[[", "task")

    res <- data.frame(
      id = id,
      publisher = publisher,
      license = license,
      task = task
    )
    return(res)
  }

  id <- map_chr(json, "[[", "id")
  publisher <- map_chr(json, "[[", "publisher")
  registry <- map_chr(json, "[[", "registry")
  rate_limit_tier <- map_chr(json, "[[", "rate_limit_tier")
  version <- map_chr(json, "[[", "version")
  capabilities <- map_chr(
    map(json, "[[", "capabilities"),
    paste,
    collapse = ", "
  )

  data.frame(
    id = id,
    publisher = publisher,
    registry = registry,
    rate_limit_tier = rate_limit_tier,
    version = version,
    capabilities = capabilities
  )
}

#' @include provider-google.R
NULL

#' Chat with a model hosted on CloudFlare
#'
#' @description
#' [Cloudflare](https://www.cloudflare.com/developer-platform/products/workers-ai/)
#' Workers AI hosts a variety of open-source AI models. To use the Cloudflare
#' API, you must have an Account ID and an Access Token, which you can obtain
#' [by following these instructions](https://developers.cloudflare.com/workers-ai/get-started/rest-api/).
#'
#' Built on top of [chat_openai_compatible()].
#'
#' ## Known limitations
#' * Tool calling does not appear to work.
#' * Images don't appear to work.
#'
#' @family chatbots
#' @param model `r param_model("meta-llama/Llama-3.3-70b-instruct-fp8-fast")`
#' @param account The Cloudflare account ID. Taken from the
#'   `CLOUDFLARE_ACCOUNT_ID` env var, if defined.
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("CLOUDFLARE_API_KEY")`
#' @export
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_cloudflare()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_cloudflare <- function(
  account = cloudflare_account(),
  system_prompt = NULL,
  params = NULL,
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
) {
  # List at https://developers.cloudflare.com/workers-ai/models/
  # `@cf` appears to be part of the model name
  model <- set_default(model, "@cf/meta/llama-3.3-70b-instruct-fp8-fast")
  echo <- check_echo(echo)
  params <- params %||% params()

  credentials <- as_credentials(
    "chat_cloudflare",
    function() cloudflare_key(),
    credentials = credentials,
    api_key = api_key
  )

  # https://developers.cloudflare.com/workers-ai/configuration/open-ai-compatibility/
  cloudflare_api <- "https://api.cloudflare.com/client/v4/accounts/"
  base_url <- paste0(cloudflare_api, cloudflare_account(), "/ai/v1/")

  provider <- ProviderCloudflare(
    name = "Cloudflare",
    base_url = base_url,
    model = model,
    params = params,
    credentials = credentials,
    extra_args = api_args,
    extra_headers = api_headers
  )

  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderCloudflare <- new_class(
  "ProviderCloudflare",
  parent = ProviderOpenAICompatible
)

method(base_request_error, ProviderCloudflare) <- function(provider, req) {
  req_error(req, body = function(resp) {
    if (resp_content_type(resp) == "application/json") {
      resp_body_json(resp)$errors[[1]]$message
    } else if (resp_content_type(resp) == "text/plain") {
      resp_body_string(resp)
    }
  })
}


# Docs look like Gemini tool defs
# https://developers.cloudflare.com/workers-ai/features/function-calling/traditional/
method(as_json, list(ProviderCloudflare, ToolDef)) <-
  method(as_json, list(ProviderGoogleGemini, ToolDef))

method(as_json, list(ProviderCloudflare, TypeObject)) <-
  method(as_json, list(ProviderGoogleGemini, TypeObject))


chat_cloudflare_test <- function(..., model = NULL, echo = "none") {
  model <- model %||% "@cf/meta/llama-3.3-70b-instruct-fp8-fast"
  chat_cloudflare(model = model, ..., echo = echo)
}

cloudflare_key <- function() {
  key_get("CLOUDFLARE_API_KEY")
}

cloudflare_account <- function() {
  key_get("CLOUDFLARE_ACCOUNT_ID")
}

#' Claude web search tool
#'
#' @description
#' Enables Claude to search the web for up-to-date information. Your organization
#' administrator must enable web search in the Anthropic Console before using
#' this tool, as it costs extra ($10 per 1,000 tokens at time of writing).
#'
#' Learn more in <https://docs.claude.com/en/docs/agents-and-tools/tool-use/web-search-tool>.
#'
#' @param max_uses Integer. Maximum number of searches allowed per request.
#' @param allowed_domains Character vector. Restrict searches to specific domains
#'   (e.g., `c("nytimes.com", "bbc.com")`). Cannot be used with `blocked_domains`.
#' @param blocked_domains Character vector. Exclude specific domains from searches.
#'   Cannot be used with `allowed_domains`.
#' @param user_location List with optional elements: `country` (2-letter code),
#'   `city`, `region`, and `timezone` (IANA timezone) to localize search results.
#'
#' @family built-in tools
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_claude()
#' chat$register_tool(claude_tool_web_search())
#' chat$chat("What was in the news today?")
#' chat$chat("What's the biggest news in the economy?")
#' }
claude_tool_web_search <- function(
  max_uses = NULL,
  allowed_domains = NULL,
  blocked_domains = NULL,
  user_location = NULL
) {
  check_exclusive(allowed_domains, blocked_domains, .require = FALSE)

  check_number_whole(max_uses, allow_null = TRUE, min = 1)
  check_character(allowed_domains, allow_null = TRUE)
  check_character(blocked_domains, allow_null = TRUE)

  json <- compact(list(
    name = "web_search",
    type = "web_search_20250305",
    max_uses = max_uses,
    allowed_domains = allowed_domains,
    blocked_domains = blocked_domains,
    user_location = user_location
  ))
  ToolBuiltIn(
    name = "web_search",
    description = "Search the web for up-to-date information.",
    annotations = tool_annotations(
      title = "Web search",
      read_only_hint = TRUE,
      open_world_hint = TRUE
    ),
    json = json
  )
}

#' Claude web fetch tool
#'
#' @description
#' Enables Claude to fetch and analyze content from web URLs. Claude can only
#' fetch URLs that appear in the conversation context (user messages or
#' previous tool results). For security reasons, Claude cannot dynamically
#' construct URLs to fetch.
#'
#' Requires the `web-fetch-2025-09-10` beta header.
#' Learn more in <https://docs.claude.com/en/docs/agents-and-tools/tool-use/web-fetch-tool>.
#'
#' @param max_uses Integer. Maximum number of fetches allowed per request.
#' @param allowed_domains Character vector. Restrict fetches to specific domains.
#'   Cannot be used with `blocked_domains`.
#' @param blocked_domains Character vector. Exclude specific domains from fetches.
#'   Cannot be used with `allowed_domains`.
#' @param citations Logical. Whether to include citations in the response. Default is `TRUE`.
#' @param max_content_tokens Integer. Maximum number of tokens to fetch from each URL.
#'
#' @family built-in tools
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_claude(beta_headers = "web-fetch-2025-09-10")
#' chat$register_tool(claude_tool_web_fetch())
#' chat$chat("What are the latest package releases on https://tidyverse.org/blog")
#' }
claude_tool_web_fetch <- function(
  max_uses = NULL,
  allowed_domains = NULL,
  blocked_domains = NULL,
  citations = FALSE,
  max_content_tokens = NULL
) {
  check_exclusive(allowed_domains, blocked_domains, .require = FALSE)

  check_character(allowed_domains, allow_null = TRUE)
  check_character(blocked_domains, allow_null = TRUE)
  check_bool(citations)

  json <- compact(list(
    name = "web_fetch",
    type = "web_fetch_20250910",
    max_uses = max_uses,
    allowed_domains = allowed_domains,
    blocked_domains = blocked_domains,
    citations = list(enabled = citations),
    max_content_tokens = max_content_tokens
  ))
  ToolBuiltIn(
    name = "web_fetch",
    description = "Fetch and analyze content from a web URL.",
    annotations = tool_annotations(
      title = "Web fetch",
      read_only_hint = TRUE,
      open_world_hint = TRUE
    ),
    json = json
  )
}

test_aws_bedrock_provider <- function(
  cache_point = "5m",
  model = "anthropic.claude-3-5-haiku-20241022-v1:0"
) {
  cache_point <- as_bedrock_cache_point(cache_point, model)

  ProviderAWSBedrock(
    name = "ProviderAWSBedrock",
    base_url = "https://bedrock-runtime.us-east-1.amazonaws.com",
    model = model,
    profile = NULL,
    region = "us-east-1",
    cache = list(),
    cache_point = cache_point,
    params = list(),
    extra_args = list(),
    extra_headers = character()
  )
}

#' @include provider-openai-compatible.R
NULL

#' Chat with a model hosted on DeepSeek
#'
#' @description
#' Sign up at <https://platform.deepseek.com>.
#'
#' Built on top of [chat_openai_compatible()].
#'
#' ## Known limitations
#'
#' * Structured data extraction is not supported.
#' * Images are not supported.
#'
#' @export
#' @family chatbots
#' @inheritParams chat_openai
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("DEEPSEEK_API_KEY")`
#' @param base_url The base URL to the endpoint; the default uses DeepSeek.
#' @param model `r param_model("deepseek-chat")`
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_deepseek()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_deepseek <- function(
  system_prompt = NULL,
  base_url = "https://api.deepseek.com",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
) {
  model <- set_default(model, "deepseek-chat")
  echo <- check_echo(echo)

  credentials <- as_credentials(
    "chat_deepseek",
    function() deepseek_key(),
    credentials = credentials,
    api_key = api_key
  )

  params <- params %||% params()

  provider <- ProviderDeepSeek(
    name = "DeepSeek",
    base_url = base_url,
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderDeepSeek <- new_class(
  "ProviderDeepSeek",
  parent = ProviderOpenAICompatible
)

method(chat_params, ProviderDeepSeek) <- function(provider, params) {
  # https://platform.deepseek.com/api-docs/api/create-chat-completion
  standardise_params(
    params,
    c(
      frequency_penalty = "frequency_penalty",
      max_tokens = "max_tokens",
      presence_penalty = "presence_penalty",
      stop = "stop_sequences",
      temperature = "temperature",
      top_p = "top_p",
      logprobs = "log_probs",
      top_logprobs = "top_k"
    )
  )
}

method(as_json, list(ProviderDeepSeek, ContentText)) <- function(
  provider,
  x,
  ...
) {
  x@text
}

method(as_json, list(ProviderDeepSeek, Turn)) <- function(provider, x, ...) {
  if (is_user_turn(x)) {
    x <- turn_contents_expand(x)
    data <- turn_split_tool_results(x)

    # Text and tool results go in separate messages
    texts <- keep(data$contents, S7_inherits, ContentText)
    texts_out <- lapply(texts, function(text) {
      list(role = "user", content = as_json(provider, text, ...))
    })

    tools_out <- lapply(data$tool_results, function(tool) {
      list(
        role = "tool",
        content = tool_string(tool),
        tool_call_id = tool@request@id
      )
    })

    c(tools_out, texts_out)
  } else if (is_assistant_turn(x)) {
    # Tool requests come out of content and go into own argument
    text <- detect(x@contents, S7_inherits, ContentText)
    tools <- keep(x@contents, is_tool_request)

    list(compact(list(
      role = "assistant",
      content = as_json(provider, text, ...),
      tool_calls = as_json(provider, tools, ...)
    )))
  } else {
    as_json(super(provider, ProviderOpenAICompatible), x, ...)
  }
}

deepseek_key <- function() key_get("DEEPSEEK_API_KEY")

#' Chat with a model hosted on Mistral's La Platforme
#'
#' @description
#' Get your API key from <https://console.mistral.ai/api-keys>.
#'
#' Built on top of [chat_openai_compatible()].
#'
#' ## Known limitations
#'
#' * Tool calling is unstable.
#' * Images require a model that supports images.
#'
#' @export
#' @family chatbots
#' @param model `r param_model("mistral-large-latest")`
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("MISTRAL_API_KEY")`
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_mistral()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_mistral <- function(
  system_prompt = NULL,
  params = NULL,
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
) {
  params <- params %||% params()
  model <- set_default(model, "mistral-large-latest")
  echo <- check_echo(echo)

  credentials <- as_credentials(
    "chat_mistral",
    function() mistral_key(),
    credentials = credentials,
    api_key = api_key
  )

  provider <- ProviderMistral(
    name = "Mistral",
    base_url = mistral_base_url,
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

mistral_base_url <- "https://api.mistral.ai/v1/"
ProviderMistral <- new_class(
  "ProviderMistral",
  parent = ProviderOpenAICompatible
)

chat_mistral_test <- function(
  system_prompt = NULL,
  model = "mistral-large-latest",
  params = NULL,
  ...,
  echo = "none"
) {
  params <- params %||% params()
  params <- modify_list(list(seed = 1014, temperature = 0), params)

  chat_mistral(
    system_prompt = system_prompt,
    model = model,
    params = params,
    ...,
    echo = echo
  )
}

method(base_request, ProviderMistral) <- function(provider) {
  req <- base_request(super(provider, ProviderOpenAICompatible))
  req <- ellmer_req_robustify(req, after = function(resp) {
    as.numeric(resp_header(resp, "ratelimitbysize-reset", NA))
  })
  req <- req_error(req, body = function(resp) {
    if (resp_content_type(resp) == "application/json") {
      resp_body_json(resp)$message
    }
  })
  req <- req_throttle(req, capacity = 1, fill_time_s = 1)
  req
}

method(chat_body, ProviderMistral) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  body <- chat_body(
    super(provider, ProviderOpenAICompatible),
    stream = stream,
    turns = turns,
    tools = tools,
    type = type
  )

  # Mistral doensn't support stream options
  body$stream_options <- NULL

  body
}

method(chat_params, ProviderMistral) <- function(provider, params) {
  standardise_params(
    params,
    c(
      temperature = "temperature",
      top_p = "top_p",
      frequency_penalty = "frequency_penalty",
      presence_penalty = "presence_penalty",
      random_seed = "seed",
      max_tokens = "max_tokens",
      logprobs = "log_probs",
      stop = "stop_sequences"
    )
  )
}

mistral_key <- function() {
  key_get("MISTRAL_API_KEY")
}

# Models -----------------------------------------------------------------------

#' @export
#' @rdname chat_mistral
models_mistral <- function(api_key = mistral_key()) {
  provider <- ProviderMistral(
    name = "Mistral",
    model = "",
    base_url = mistral_base_url,
    credentials = function() api_key
  )

  req <- base_request(provider)
  req <- req_url_path_append(req, "/models")
  resp <- req_perform(req)

  json <- resp_body_json(resp)

  id <- map_chr(json$data, `[[`, "id")
  display_name <- map_chr(json$data, `[[`, "name")
  created_at <- as.POSIXct(map_int(json$data, `[[`, "created"))

  df <- data.frame(
    id = id,
    name = display_name,
    created_at = created_at
  )
  df <- cbind(df, match_prices("Mistral", df$id))
  df
}

#' @include tools-def.R
#' @include as-json.R

ToolBuiltIn <- new_class(
  "ToolBuiltIn",
  properties = list(
    name = prop_string(),
    description = prop_string(default = ""),
    annotations = class_list,
    json = class_any
  )
)

method(as_json, list(Provider, ToolBuiltIn)) <- function(provider, x, ...) {
  x@json
}

# Web search ---------------------------------------------------------------

ContentToolRequestSearch <- new_class(
  "ContentToolRequestSearch",
  parent = Content,
  properties = list(
    query = prop_string(),
    json = class_list
  )
)
method(format, ContentToolRequestSearch) <- function(x, ...) {
  cli::format_inline("[{.strong web search request}]: {.str {x@query}}")
}
method(as_json, list(Provider, ContentToolRequestSearch)) <- function(
  provider,
  x,
  ...
) {
  x@json
}

ContentToolResponseSearch <- new_class(
  "ContentToolResponseSearch",
  parent = Content,
  properties = list(
    urls = class_character,
    json = class_list
  )
)
method(format, ContentToolResponseSearch) <- function(x, ...) {
  paste0(
    cli::format_inline("[{.strong web search results}]:\n"),
    paste0("* ", x@urls, "\n", collapse = "")
  )
}

method(as_json, list(Provider, ContentToolResponseSearch)) <- function(
  provider,
  x,
  ...
) {
  x@json
}

# Web fetch -------------------------------------------------------------------

ContentToolRequestFetch <- new_class(
  "ContentToolRequestFetch",
  parent = Content,
  properties = list(
    url = prop_string(),
    json = class_list
  )
)
method(format, ContentToolRequestFetch) <- function(x, ...) {
  cli::format_inline("[{.strong web fetch request}]: {.url {x@url}}")
}
method(as_json, list(Provider, ContentToolRequestFetch)) <- function(
  provider,
  x,
  ...
) {
  x@json
}

ContentToolResponseFetch <- new_class(
  "ContentToolResponseFetch",
  parent = Content,
  properties = list(
    url = prop_string(),
    json = class_list
  )
)
method(format, ContentToolResponseFetch) <- function(x, ...) {
  cli::format_inline("[{.strong web fetch result}]: {.url {x@url}}")
}
method(as_json, list(Provider, ContentToolResponseFetch)) <- function(
  provider,
  x,
  ...
) {
  x@json
}

#' Describe the schema of a data frame, suitable for sending to an LLM
#'
#' @description
#' `df_schema()` gives a column-by-column description of a data frame. For
#' each column, it gives the name, type, label (if present), and number of
#' missing values. For numeric and date/time columns, it also gives the
#' range. For character and factor columns, it also gives the number of unique
#' values, and if there's only a few (<= 10), their values.
#'
#' The goal is to give the LLM a sense of the structure of the data, so that
#' it can generate useful code, and the output attempts to balance between
#' conciseness and accuracy.
#'
#' @param df A data frame to describe.
#' @param max_cols Maximum number of columns to includes. Defaults to 50 to
#'   avoid accidentally generating very large prompts.
#' @export
#' @examples
#' df_schema(mtcars)
#' df_schema(iris)
df_schema <- function(df, max_cols = 50) {
  if (!is.data.frame(df)) {
    stop_input_type(df, "a data frame")
  }

  df_desc <- sprintf(
    "A data frame with %i rows and %i columns:",
    nrow(df),
    ncol(df)
  )

  if (ncol(df) > max_cols) {
    cli::cli_warn("Truncating to {max_cols} columns.")
    df <- df[seq_len(max_cols)]
    extra <- sprintf("and %i more columns", ncol(df) - max_cols)
  } else {
    extra <- NULL
  }

  cols <- map_chr(df, col_schema)
  col_desc <- paste0("* ", names(cols), ": ", cols, recycle0 = TRUE)

  desc <- paste0(c(df_desc, col_desc, extra), collapse = "\n")
  ellmer_prompt(desc)
}

col_schema <- new_generic("col_schema", "x")

method(col_schema, class_logical) <- function(x) {
  describe_column(
    "logical",
    desc_label(x),
    sprintf("%i TRUEs", sum(x, na.rm = TRUE)),
    sprintf("%i FALSEs", sum(!x, na.rm = TRUE)),
    desc_na(x)
  )
}

method(col_schema, class_numeric) <- function(x) {
  describe_column(
    if (is.integer(x)) "integer" else "numeric",
    desc_label(x),
    desc_range(x),
    desc_na(x)
  )
}

method(col_schema, class_character) <- function(x) {
  describe_column(
    "character",
    desc_label(x),
    desc_na(x),
    desc_unique(x)
  )
}

method(col_schema, class_factor) <- function(x) {
  describe_column(
    if (is.ordered(x)) "ordinal" else "nominal",
    desc_label(x),
    desc_na(x),
    desc_unique(x)
  )
}

method(col_schema, class_Date) <- function(x) {
  describe_column(
    "date",
    desc_label(x),
    desc_range(x),
    desc_na(x)
  )
}

method(col_schema, class_POSIXt) <- function(x) {
  tz <- attr(x, "tzone")

  describe_column(
    "date-time",
    if (!is.null(tz) && tz != "") sprintf("timezone %s", tz),
    desc_label(x),
    desc_range(x),
    desc_na(x)
  )
}

method(col_schema, class_data.frame) <- function(x) {
  describe_column(
    "data frame",
    paste(names(x), collapse = ", "),
    desc_na(x)
  )
}

method(col_schema, class_list) <- function(x) {
  describe_column("list column")
}

method(col_schema, class_any) <- function(x) {
  describe_column(paste0(class(x), collapse = "/"))
}


# Helpers ----------------------------------------------------------------------

describe_column <- function(type, ...) {
  if (missing(...)) {
    ellmer_prompt(type)
  } else {
    props <- c(...)
    ellmer_prompt(paste0(type, " with ", str_flatten(props)))
  }
}

desc_range <- function(x) {
  x <- x[!is.na(x)]
  if (length(x) == 0) {
    return()
  }
  rng <- range(x)
  sprintf(
    "range [%s, %s]",
    format(rng[1], digits = 4),
    format(rng[2], digits = 4)
  )
}

desc_unique <- function(x) {
  if (is.factor(x)) {
    label <- "permitted"
    unique <- levels(x)
  } else {
    label <- "unique"
    unique <- unique(x)
  }
  unique <- unique[!is.na(unique)]

  desc <- sprintf("%i %s values", length(unique), label)
  if (length(unique) > 0 && length(unique) <= 10 && sum(nchar(unique)) < 200) {
    quoted <- encodeString(unique, quote = '"')
    desc <- paste0(desc, " (", str_flatten(quoted, last = NULL), ")")
  }
  desc
}

desc_na <- function(x) {
  sprintf("%i NAs", sum(is.na(x)))
}

desc_label <- function(x) {
  label <- attr(x, "label")
  if (is.null(label)) {
    return()
  }
  sprintf("label (%s)", label)
}


str_flatten <- function(string, collapse = ", ", last = ", and ") {
  n <- length(string)
  if (!is.null(last) && n >= 2) {
    string <- c(
      string[seq2(1, n - 2)],
      paste0(string[[n - 1]], last, string[[n]])
    )
  }

  paste0(string, collapse = collapse)
}

# df_schema -----------------------------------------------------------------

test_that("df_schema aggregates column descriptions", {
  df <- data.frame(x = 1:3, y = letters[1:3])
  expect_snapshot(df_schema(df))
})

test_that("df_schema works on edge case data frames", {
  expect_snapshot({
    df_schema(data.frame())
    df_schema(data.frame(x = 1, y = "a"))
    df_schema(data.frame(x = 1:5))
  })
})

test_that("df_schema checks its inputs", {
  expect_snapshot(error = TRUE, {
    df_schema(1)
  })
})

test_that("warns for wide data frames", {
  df <- vctrs::new_data_frame(set_names(rep(list(1), 26), letters))
  expect_snapshot(df_schema(df, max_cols = 5))
})

# col_schema ----------------------------------------------------------------

test_that("col_schema handles logical vectors", {
  expect_snapshot(col_schema(c(TRUE, FALSE, TRUE, NA)))
})

test_that("col_schema handles numeric vectors", {
  expect_snapshot({
    col_schema(c(1L:5L, NA))
    col_schema(c(1.5, 2.5, 3.5))
    col_schema(c(1.5, NA, NaN, Inf, -Inf, 3.5))
  })
})

test_that("col_schema handles characters and factors", {
  expect_snapshot({
    col_schema(c("a", "b", NA))
    col_schema(letters)
    col_schema(factor(c("a", "b", NA)))
    col_schema(factor(c("a", "b", NA), exclude = NULL))
    col_schema(ordered(c("low", "med", "high")))
  })
})

test_that("col_schema handles date/times", {
  expect_snapshot({
    col_schema(as.Date(c("2024-01-01", "2024-12-31", NA)))
    col_schema(as.POSIXct(c(
      "2024-01-01 10:30:00",
      "2024-12-31 23:59:59"
    )))
    col_schema(as.POSIXct(
      "2024-01-01 10:30:00",
      tz = "America/New_York"
    ))
  })
})

test_that("col_schema handles empty vectors", {
  expect_snapshot({
    col_schema(character(0))
    col_schema(numeric(0))
    col_schema(logical(0))
    col_schema(factor(character(0)))
    col_schema(as.Date(character(0)))
    col_schema(as.POSIXct(character(0)))
  })
})

test_that("col_schema handles data frame columns", {
  df <- data.frame(a = 1:3, b = letters[1:3])
  expect_snapshot(col_schema(df))
})

test_that("col_schema handles list columns", {
  lst <- list(1:3, letters[1:5], NULL)
  expect_snapshot(col_schema(lst))
})

test_that("col_schema handles unknown classes", {
  x <- structure(1:5, class = c("class1", "class2", "class3"))
  expect_snapshot(col_schema(x))
})

test_that("col_schema handles labeled columns", {
  x <- 1:5
  attr(x, "label") <- "My labeled variable"
  expect_snapshot(col_schema(x))
})

# desc_* ------------------------------------------------------------------

test_that("desc_range handles various numeric inputs", {
  expect_snapshot({
    desc_range(numeric())
    desc_range(c(-Inf, 1, 2, Inf))
    desc_range(c(NA, 1, 2, NA))
  })
})

test_that("desc_unique handles character vectors", {
  expect_snapshot({
    desc_unique(letters)
    desc_unique(strrep("x", 10000))
    desc_unique(character())
  })
})

#' @include provider-openai-compatible.R
NULL

#' Chat with a model hosted on perplexity.ai
#'
#' @description
#' Sign up at <https://www.perplexity.ai>.
#'
#' Perplexity AI is a platform for running LLMs that are capable of
#' searching the web in real-time to help them answer questions with
#' information that may not have been available when the model was
#' trained.
#'
#' This function is a Uses OpenAI compatible API via `chat_openai_compatible()` with
#' the defaults tweaked for Perplexity AI.
#'
#' @export
#' @family chatbots
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("PERPLEXITY_API_KEY")`
#' @param model `r param_model("llama-3.1-sonar-small-128k-online")`
#' @param params Common model parameters, usually created by [params()].
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_perplexity()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_perplexity <- function(
  system_prompt = NULL,
  base_url = "https://api.perplexity.ai/",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
) {
  model <- set_default(model, "llama-3.1-sonar-small-128k-online")
  echo <- check_echo(echo)

  credentials <- as_credentials(
    "chat_perplexity",
    function() perplexity_key(),
    credentials = credentials,
    api_key = api_key
  )

  params <- params %||% params()

  provider <- ProviderPerplexity(
    name = "Perplexity",
    base_url = base_url,
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderPerplexity <- new_class(
  "ProviderPerplexity",
  parent = ProviderOpenAICompatible,
)

method(chat_params, ProviderPerplexity) <- function(provider, params) {
  # https://docs.perplexity.ai/api-reference/chat-completions-post
  standardise_params(
    params,
    c(
      max_tokens = "max_tokens",
      temperature = "temperature",
      top_p = "top_p",
      top_k = "top_k",
      presence_penalty = "presence_penalty",
      frequency_penalty = "frequency_penalty"
    )
  )
}

perplexity_key <- function() {
  key_get("PERPLEXITY_API_KEY")
}

#

vcr_clean <- function(url_prefix) {
  vcr_dirs <- file.path(c("inst", "tests/testthat", "vignettes"), "_vcr")
  cassettes <- dir(vcr_dirs, full.names = TRUE)

  first_uri <- map_chr(cassettes, \(path) {
    yaml <- utils::getFromNamespace("read_yaml", "yaml")(path)
    yaml$http_interactions[[1]]$request$uri
  })
  match <- startsWith(first_uri, url_prefix)
  unlink(cassettes[match])
}

vcr_rebuild <- function() {
  withr::local_temp_libpaths()

  cli::cli_rule("Installing package to temporary location")
  # Needed because build_articles() renders in clean session
  utils::getFromNamespace("install", "devtools")(
    quick = TRUE,
    upgrade = FALSE,
    quiet = TRUE
  )
  utils::getFromNamespace("build_articles", "pkgdown")(lazy = FALSE)
  utils::getFromNamespace("build_reference", "pkgdown")(lazy = FALSE)
  utils::getFromNamespace("test", "devtools")()
}

# Standalone file: do not edit by hand
# Source: https://github.com/r-lib/withr/blob/HEAD/R/standalone-defer.R
# Generated by: usethis::use_standalone("r-lib/withr", "defer")
# ----------------------------------------------------------------------
#
# ---
# repo: r-lib/withr
# file: standalone-defer.R
# last-updated: 2024-01-15
# license: https://unlicense.org
# ---
#
# `defer()` is similar to `on.exit()` but with a better default for
# `add` (hardcoded to `TRUE`) and `after` (`FALSE` by default).
# It also supports adding handlers to other frames which is useful
# to implement `local_` functions.
#
#
# ## Changelog
#
# 2024-01-15:
# * Rewritten to be pure base R.
#
# nocov start

defer <- function(expr, envir = parent.frame(), after = FALSE) {
  thunk <- as.call(list(function() expr))
  do.call(
    on.exit,
    list(thunk, add = TRUE, after = after),
    envir = envir
  )
}

# nocov end

#' @include content.R
NULL

#' A chatbot provider
#'
#' A Provider captures the details of one chatbot service/API. This captures
#' how the API works, not the details of the underlying large language model.
#' Different providers might offer the same (open source) model behind a
#' different API.
#'
#' To add support for a new backend, you will need to subclass `Provider`
#' (adding any additional fields that your provider needs) and then implement
#' the various generics that control the behavior of each provider.
#'
#' @export
#' @param name Name of the provider.
#' @param model Name of the model.
#' @param base_url The base URL for the API.
#' @param params A list of standard parameters created by [params()].
#' @param credentials A zero-argument function that returns the credentials to use
#'   for authentication. Can either return a string, representing an API key,
#'   or a named list of headers.
#' @param extra_args Arbitrary extra arguments to be included in the request body.
#' @param extra_headers Arbitrary extra headers to be added to the request.
#' @return An S7 Provider object.
#' @examples
#' Provider(
#'   name = "CoolModels",
#'   model = "my_model",
#'   base_url = "https://cool-models.com"
#' )
Provider <- new_class(
  "Provider",
  properties = list(
    name = prop_string(),
    model = prop_string(),
    base_url = prop_string(),
    params = class_list,
    extra_args = class_list,
    extra_headers = class_character,
    credentials = class_function | NULL
  )
)

test_provider <- function(name = "", model = "", base_url = "", ...) {
  Provider(name = name, model = model, base_url = base_url, ...)
}

# Create a request------------------------------------

base_request <- new_generic("base_request", "provider", function(provider) {
  S7_dispatch()
})

base_request_error <- new_generic(
  "base_request_error",
  "provider",
  function(provider, req) {
    S7_dispatch()
  }
)

chat_request <- new_generic(
  "chat_request",
  "provider",
  function(
    provider,
    stream = TRUE,
    turns = list(),
    tools = list(),
    type = NULL
  ) {
    S7_dispatch()
  }
)

method(chat_request, Provider) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  req <- base_request(provider)
  req <- req_url_path_append(req, chat_path(provider))

  body <- chat_body(
    provider = provider,
    stream = stream,
    turns = turns,
    tools = tools,
    type = type
  )
  body <- modify_list(body, provider@extra_args)
  req <- req_body_json(req, body)
  req <- req_headers(req, !!!provider@extra_headers)

  req
}

chat_body <- new_generic(
  "chat_body",
  "provider",
  function(
    provider,
    stream = TRUE,
    turns = list(),
    tools = list(),
    type = NULL
  ) {
    S7_dispatch()
  }
)

chat_path <- new_generic("chat_path", "provider", function(provider) {
  S7_dispatch()
})

chat_resp_stream <- new_generic(
  "chat_resp_stream",
  "provider",
  function(provider, resp) {
    S7_dispatch()
  }
)
method(chat_resp_stream, Provider) <- function(provider, resp) {
  resp_stream_sse(resp)
}

chat_params <- new_generic(
  "chat_params",
  "provider",
  function(provider, params) {
    S7_dispatch()
  }
)

# Extract data from streaming results ------------------------------------

stream_parse <- new_generic(
  "stream_parse",
  "provider",
  function(provider, event) {
    S7_dispatch()
  }
)
stream_content <- new_generic(
  "stream_content",
  "provider",
  function(provider, event) {
    S7_dispatch()
  }
)

stream_text <- function(provider, event) {
  content <- stream_content(provider, event)
  if (is.null(content)) {
    return(NULL)
  }
  content_text(content)
}

content_text <- function(content) {
  switch(
    class(content)[1],
    "ellmer::ContentThinking" = content@thinking,
    "ellmer::ContentText" = content@text,
    format(content)
  )
}
stream_merge_chunks <- new_generic(
  "stream_merge_chunks",
  "provider",
  function(provider, result, chunk) {
    S7_dispatch()
  }
)

# Extract data from non-streaming results --------------------------------------

value_turn <- new_generic("value_turn", "provider")

# Extract token counts from API response
# Returns a named list produced by token_usage()
value_tokens <- new_generic(
  "value_tokens",
  "provider",
  function(provider, json) {
    S7_dispatch()
  }
)
method(value_tokens, Provider) <- function(provider, json) {
  tokens()
}

# Convert to JSON
as_json <- new_generic(
  "as_json",
  c("provider", "x"),
  function(provider, x, ...) {
    S7_dispatch()
  }
)

method(as_json, list(Provider, class_list)) <- function(provider, x, ...) {
  compact(lapply(x, as_json, provider = provider, ...))
}

method(as_json, list(Provider, ContentJson)) <- function(provider, x, ...) {
  if (!is.null(x@string)) {
    string <- x@string
  } else {
    string <- unclass(jsonlite::toJSON(x@data, auto_unbox = TRUE))
  }
  as_json(provider, ContentText(string), ...)
}

# Batch AI ---------------------------------------------------------------

# Does the provider support batch uploads?
has_batch_support <- new_generic(
  "has_batch_support",
  "provider",
  function(provider) {
    S7_dispatch()
  }
)
method(has_batch_support, Provider) <- function(provider) {
  FALSE
}

# Submit a batch, return an object "batch" object that will be passed to
# batch_poll() and batch_retrieve()
batch_submit <- new_generic(
  "batch_submit",
  "provider",
  function(provider, conversations, type = NULL) {
    S7_dispatch()
  }
)

# Get batch status. Returns an opaque list.
batch_poll <- new_generic(
  "batch_poll",
  "provider",
  function(provider, batch) {
    S7_dispatch()
  }
)

# Given batch status, return a standardised list:
# * working - TRUE/FALSE
# * n_processing = number of requests still processing
# * n_succeeded = number of requests that succeeded
# * n_failed = number of requests that failed
batch_status <- new_generic(
  "batch_status",
  "provider",
  function(provider, batch) {
    S7_dispatch()
  }
)

# Download batched results
batch_retrieve <- new_generic(
  "batch_retrieve",
  "provider",
  function(provider, batch) {
    S7_dispatch()
  }
)

# Process a single result. Returns either a turn or NULL, if the turn
# did not succeed
batch_result_turn <- new_generic(
  "batch_result_turn",
  "provider",
  function(provider, result, has_type = FALSE) {
    S7_dispatch()
  }
)

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

is_snapshot <- function() {
  identical(Sys.getenv("TESTTHAT_IS_SNAPSHOT"), "true")
}

is_replaying <- function() {
  as.logical(Sys.getenv("VCR_IS_REPLAYING", "FALSE"))
}

key_get <- function(name, error_call = caller_env()) {
  val <- Sys.getenv(name)
  if (!identical(val, "")) {
    val
  } else {
    if (is_replaying()) {
      # this value won't get used, but we don't want to error
      ""
    } else if (is_testing()) {
      testthat::skip(sprintf("%s env var is not configured", name))
    } else {
      cli::cli_abort("Can't find env var {.code {name}}.", call = error_call)
    }
  }
}

key_exists <- function(name) {
  !identical(Sys.getenv(name), "")
}

set_default <- function(value, default, arg = caller_arg(value)) {
  if (is.null(value)) {
    if (!is_testing() || is_snapshot()) {
      cli::cli_inform("Using {.field {arg}} = {.val {default}}.")
    }
    default
  } else {
    value
  }
}

last_request_json <- function() {
  print_json(last_request()$body$data)
}
last_response_json <- function() {
  print_json(resp_body_json(last_response()))
}
print_json <- function(x) {
  cat(pretty_json(x))
  cat("\n")
}
pretty_json <- function(x) {
  jsonlite::toJSON(x, pretty = TRUE, auto_unbox = TRUE)
}

parse_json <- function(x) {
  tryCatch(
    jsonlite::fromJSON(x, simplifyVector = FALSE),
    error = function(cnd) NULL
  )
}

read_ndjson <- function(path, fallback = \(line) NULL) {
  lines <- readLines(path, warn = FALSE)
  lapply(lines, function(line) {
    tryCatch(
      jsonlite::fromJSON(line, simplifyVector = FALSE),
      error = function(cnd) fallback(line)
    )
  })
}

prettify <- function(x) {
  tryCatch(
    jsonlite::prettify(x),
    error = function(cnd) x
  )
}

check_echo <- function(echo = NULL) {
  if (identical(echo, "text")) {
    lifecycle::deprecate_soft(
      when = "0.2.0",
      what = I('`echo = "text"`'),
      with = I('`echo = "output"`')
    )
    echo <- "output"
  }

  if (is.null(echo) || identical(echo, c("none", "output", "all"))) {
    option <- getOption("ellmer_echo")

    if (!is.null(option)) {
      option
    } else if (env_is_user_facing(parent.frame(2))) {
      "output"
    } else {
      "none"
    }
  } else if (isTRUE(echo)) {
    "output"
  } else if (isFALSE(echo)) {
    "none"
  } else {
    arg_match(echo, c("none", "output", "all"))
  }
}

dots_named <- function(...) {
  is_named2(list2(...))
}

`paste<-` <- function(x, value) {
  paste0(x, value)
}

`append<-` <- function(x, value) {
  x[[length(x) + 1]] <- value
  x
}

#' Are credentials avaiable?
#'
#' Used for examples/testing.
#'
#' @keywords internal
#' @param provider Provider name.
#' @export
has_credentials <- function(provider) {
  switch(
    provider,
    snowflake = snowflake_credentials_exist(),
    openai = openai_key_exists(),
    claude = anthropic_key_exists(),
    cli::cli_abort("Unknown model {model}.")
  )
}

# In-memory cache for credentials. Analogous to httr2:::cache_mem().
credentials_cache <- function(key) {
  list(
    get = function() env_get(the$credentials_cache, key, default = NULL),
    set = function(creds) env_poke(the$credentials_cache, key, creds),
    clear = function() env_unbind(the$credentials_cache, key)
  )
}

has_connect_viewer_token <- function(...) {
  if (!is_installed("connectcreds")) {
    return(FALSE)
  }
  connectcreds::has_viewer_token(...)
}

modify_list <- function(x, y) {
  if (is.null(x)) {
    return(y)
  }
  if (is.null(y)) {
    return(x)
  }

  utils::modifyList(x, y)
}

is_whitespace <- function(x) {
  grepl("^(\\s|\n)*$", x)
}

paste_c <- function(...) {
  paste(c(...), collapse = "")
}

cli_escape <- function(x) {
  x <- gsub("{", "{{", x, fixed = TRUE)
  gsub("}", "}}", x, fixed = TRUE)
}

api_key_param <- function(key) {
  paste_c(
    "Override the default credentials. ",
    c(
      "You generally should not need this argument; instead set the ",
      c("`", key, "`"),
      " environment variable. "
    ),
    c(
      "The best place to set this is in `.Renviron`,
      which you can easily edit by calling `usethis::edit_r_environ()`."
    ),
    "\n\n",
    c(
      "If you do need additional control, this argument takes a ",
      "zero-argument function that returns either a string (the API key), ",
      "or a named list (added as additional headers to every request)."
    )
  )
}

param_model <- function(default, provider = NULL) {
  paste_c(
    c(
      "The model to use for the chat",
      if (!is.null(default)) c(" (defaults to \"", default, "\")"),
      ".\n"
    ),
    if (!is.null(default)) {
      c(
        "We regularly update the default, so we strongly recommend explicitly ",
        "specifying a model for anything other than casual use.\n"
      )
    },
    if (!is.null(provider)) {
      c("Use `models_", provider, "()` to see all options.\n")
    }
  )
}

unrowname <- function(df) {
  rownames(df) <- NULL
  df
}

color_role <- function(role) {
  switch(
    role,
    user = cli::col_blue(role),
    assistant = cli::col_green(role),
    system = cli::col_br_white(role),
    role
  )
}

counter <- function() {
  count <- 0
  function() {
    count <<- count + 1
    count
  }
}

match_prices <- function(provider, id) {
  prices <- prices[prices$provider == provider, ]
  idx <- match(id, prices$model)
  prices[idx, c("cached_input", "input", "output")]
}

base64_enc <- function(path, raw) {
  check_exclusive(path, raw)
  if (!missing(path)) {
    raw <- readBin(path, "raw", file.info(path)$size)
  }
  gsub("\n", "", openssl::base64_encode(raw), fixed = TRUE)
}

ellmer_output <- function(x) {
  structure(x, class = "ellmer_output")
}
#' @export
print.ellmer_output <- function(x, ...) {
  cat_line(x)
  invisible(x)
}

eval_vignette <- function() {
  name <- tools::file_path_sans_ext(knitr::current_input())

  cassettes <- dir("_vcr", pattern = paste0(name, "*"))
  has_cassette <- length(cassettes) > 0

  has_key <- has_credentials("openai") && has_credentials("claude")

  options(ellmer_echo = "none")

  has_key || has_cassette
}

vcr_example_start <- function(name) {
  options(ellmer_echo = "none")
  vcr::insert_example_cassette(name, package = "ellmer")
}
vcr_example_end <- function() {
  options(ellmer_echo = NULL)
  vcr::eject_cassette()
}

replay <- function(values) {
  i <- 0
  function() {
    i <<- i + 1
    values[[i]]
  }
}

content_types <- function(turns) {
  lapply(turns, function(turn) {
    map_chr(turn@contents, function(x) S7_class(x)@name)
  })
}

request_summary <- function(req) {
  list(
    url = req_get_url(req),
    headers = req_get_headers(req, "reveal"),
    body = req_get_body(req)
  )
}

str_trunc <- function(x, n) {
  ifelse(nchar(x) > n, paste0(substr(x, 1, n - 3), "..."), x)
}

extract_custom_id <- function(json_string) {
  pattern <- '"custom_id"\\s*:\\s*"([^"]*)"'
  match <- regexec(pattern, json_string)

  result <- regmatches(json_string, match)[[1]]
  if (length(result) < 2) {
    return(NA_character_)
  }

  result[2] # Second element is the captured group
}

to_json <- function(x) {
  jsonlite::toJSON(x, auto_unbox = TRUE)
}

test_that("finds key if set", {
  withr::local_envvar(FOO = "abc123")
  expect_true(key_exists("FOO"))
  expect_equal(key_get("FOO"), "abc123")
})


test_that("informative error if no key", {
  withr::local_envvar(FOO = NULL, TESTTHAT = "false")
  expect_false(key_exists("FOO"))
  expect_snapshot(key_get("FOO"), error = TRUE)
})

test_that("detects whitespace", {
  expect_true(is_whitespace("\n\n\n \t"))
  expect_true(is_whitespace(""))

  expect_false(is_whitespace("a"))
  expect_false(is_whitespace("."))
})

test_that('echo="output" replaces echo="text"', {
  expect_snapshot(
    expect_equal(check_echo("text"), "output")
  )
})

#' Upload, downloand, and manage files for Claude
#'
#' @description
#' `r lifecycle::badge("experimental")`
#' Use the beta Files API to upload files to and manage files in Claude.
#' This is currently experimental because the API is in beta and may change.
#' Note that you need `beta-headers = "files-api-2025-04-14"` to use the API.
#'
#' Claude offers 100GB of file storage per organization, with each file
#' having a maximum size of 500MB. For more details see
#' <https://docs.claude.com/en/docs/build-with-claude/files>
#'
#' * `claude_file_upload()` uploads a file and returns an object that
#'   you can use in chat.
#' * `claude_file_list()` lists all uploaded files.
#' * `claude_file_get()` returns an object for an previously uploaded file.
#' * `claude_file_download()` downloads the file with the given ID. Note
#'   that you can only download files created by skills or the code execution
#'   tool.
#' * `claude_file_delete()` deletes the file with the given ID.
#'
#' @inheritParams chat_anthropic
#' @param path Path to a file to upload.
#' @param file_id ID of the file to get information about, download, or delete.
#' @param beta_headers Beta headers to use for the request. Defaults to
#'   `files-api-2025-04-14`.
#' @export
#' @examples
#' \dontrun{
#' file <- claude_file_upload("path/to/file.pdf")
#' chat <- chat_anthropic(beta_headers = "files-api-2025-04-14")
#' chat$chat("Please summarize the document.", file)
#' }

claude_file_upload <- function(
  path,
  base_url = "https://api.anthropic.com/v1/",
  beta_headers = "files-api-2025-04-14",
  credentials = NULL
) {
  check_string(path, allow_empty = FALSE)
  if (!file.exists(path)) {
    cli::cli_abort("{.arg path} must be an existing file.")
  }
  file <- form_file(path, type = guess_mime_type(path))

  req <- request_anthropic_file(base_url, beta_headers, credentials)
  req <- req_url_path_append(req, "/files")
  req <- req_body_multipart(req, file = file)
  resp <- req_perform(req)
  json <- resp_body_json(resp)

  ContentUploaded(uri = json$id, mime_type = json$mime_type)
}

#' @export
#' @rdname claude_file_upload
claude_file_list <- function(
  base_url = "https://api.anthropic.com/v1/",
  credentials = NULL,
  beta_headers = "files-api-2025-04-14"
) {
  req <- request_anthropic_file(base_url, beta_headers, credentials)
  req <- req_url_path_append(req, "/files")
  resp <- req_perform(req)

  data <- resp_body_json(resp)$data
  data.frame(
    id = map_chr(data, "[[", "id"),
    filename = map_chr(data, "[[", "filename"),
    mime_type = map_chr(data, "[[", "mime_type"),
    size = map_dbl(data, "[[", "size_bytes"),
    created_at = map_chr(data, "[[", "created_at")
  )
}

#' @export
#' @rdname claude_file_upload
claude_file_get <- function(
  file_id,
  base_url = "https://api.anthropic.com/v1/",
  credentials = NULL,
  beta_headers = "files-api-2025-04-14"
) {
  req <- request_anthropic_file(base_url, beta_headers, credentials)
  req <- req_url_path_append(req, "files", file_id)
  resp <- req_perform(req)
  json <- resp_body_json(resp)

  ContentUploaded(uri = json$id, mime_type = json$mime_type)
}

#' @export
#' @rdname claude_file_upload
#' @param path Path to download the file to.
claude_file_download <- function(
  file_id,
  path,
  base_url = "https://api.anthropic.com/v1/",
  credentials = NULL,
  beta_headers = "files-api-2025-04-14"
) {
  check_string(path)

  req <- request_anthropic_file(base_url, beta_headers, credentials)
  req <- req_url_path_append(req, "files", file_id, "content")
  req_perform(req, path = path)

  invisible(path)
}

#' @export
#' @rdname claude_file_upload
claude_file_delete <- function(
  file_id,
  base_url = "https://api.anthropic.com/v1/",
  credentials = NULL,
  beta_headers = "files-api-2025-04-14"
) {
  req <- request_anthropic_file(base_url, beta_headers, credentials)
  req <- req_url_path_append(req, "files", file_id)
  req <- req_method(req, "DELETE")
  resp <- req_perform(req)

  invisible()
}

request_anthropic_file <- function(url, beta_headers, credentials) {
  credentials <- as_credentials(
    "chat_anthropic",
    function() anthropic_key(),
    credentials = credentials
  )

  provider <- ProviderAnthropic(
    name = "Anthropic",
    model = "",
    base_url = url,
    params = list(),
    extra_args = list(),
    credentials = credentials,
    beta_headers = beta_headers,
    cache = "none"
  )

  base_request(provider)
}

form_file <- function(path, type = type) {
  curl::form_file(path, type = type)
}

test_that("end-to-end test of all functions", {
  vcr::local_cassette("anthropic-upload-file")
  # Avoid using an absolute path in form_file
  local_mocked_bindings(
    form_file = function(path, type) {
      structure(
        list(path = path, type = type, name = NULL),
        class = "form_file"
      )
    }
  )

  upload <- claude_file_upload(test_path("apples.pdf"))
  defer(claude_file_delete(upload@uri))

  chat <- chat_anthropic_test(beta_headers = "files-api-2025-04-14")
  response <- chat$chat("What's the title of this document?", upload)
  expect_match(response, "Apples are tasty")
  expect_match(
    chat$chat("What apple is not tasty? State only the name"),
    "red delicious",
    ignore.case = TRUE
  )

  # Can't download uploaded files, but at least tests that request succeeds
  path <- withr::local_tempfile()
  expect_error(claude_file_download(upload@uri, path), "not downloadable")

  files <- claude_file_list()
  expect_true("apples.pdf" %in% files$filename)
})

# Given a promise-yielding expression, loop until it resolves or rejects.
# DON'T USE THIS TECHNIQUE IN SHINY, PLUMBER, OR HTTPUV CONTEXTS.
sync <- function(expr) {
  p <- force(expr)

  done <- FALSE
  success <- NULL
  error <- NULL

  promises::then(
    p,
    function(result) {
      success <<- result
      done <<- TRUE
    },
    function(err) {
      error <<- err
      done <<- TRUE
    }
  )

  while (!done) {
    later::run_now(0.25)
  }
  if (!is.null(error)) {
    stop(error)
  } else {
    success
  }
}

#' @include provider.R
#' @include content.R
#' @include turns.R
#' @include tools-def.R
NULL

#' Chat with an OpenAI-compatible model
#'
#' @description
#' This function is for use with OpenAI-compatible APIs, also known as the
#' **chat completions** API. If you want to use OpenAI itself, we recommend
#' [chat_openai()], which uses the newer **responses** API.
#'
#' Many providers offer OpenAI-compatible APIs, including:
#' * [Ollama](https://ollama.com/) for local models
#' * [vLLM](https://docs.vllm.ai/) for self-hosted models
#' * Various cloud providers with OpenAI-compatible endpoints
#'
#' @param base_url The base URL to the endpoint. This parameter is **required**
#'   since there is no default for OpenAI-compatible APIs.
#' @param name The name of the provider; this is shown in [token_usage()] and
#'   is used to compute costs.
#' @param system_prompt A system prompt to set the behavior of the assistant.
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials Credentials to use for authentication. If not provided,
#'   will attempt to use the `OPENAI_API_KEY` environment variable.
#' @param model The model to use for chat. No default; depends on your provider.
#' @param params Common model parameters, usually created by [params()].
#' @param api_args Named list of arbitrary extra arguments appended to the body
#'   of every chat API call. Combined with the body object generated by ellmer
#'   with [modifyList()].
#' @param api_headers Named character vector of arbitrary extra headers appended
#'   to every chat API call.
#' @param preserve_thinking If `TRUE`, reasoning content returned by the model
#'   is included when sending conversation history back to the API. If `FALSE`
#'   (the default), reasoning content is still captured in the turn but dropped
#'   from subsequent requests. Set to `TRUE` if your provider requires or
#'   benefits from seeing prior reasoning in multi-turn conversations.
#' @param echo One of the following options:
#'   * `none`: don't emit any output (default when running in a function).
#'   * `output`: echo text and tool-calling output as it streams in (default
#'     when running at the console).
#'   * `all`: echo all input and output.
#'
#'   Note this only affects the `chat()` method.
#' @family chatbots
#' @export
#' @returns A [Chat] object.
#' @examples
#' \dontrun{
#' # Example with Ollama (requires Ollama running locally)
#' chat <- chat_openai_compatible(
#'   base_url = "http://localhost:11434/v1",
#'   model = "llama2"
#' )
#' chat$chat("What is the difference between a tibble and a data frame?")
#' }
chat_openai_compatible <- function(
  base_url,
  name = "OpenAI-compatible",
  system_prompt = NULL,
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  api_headers = character(),
  preserve_thinking = FALSE,
  echo = c("none", "output", "all")
) {
  if (missing(base_url)) {
    cli::cli_abort(c(
      "{.arg base_url} is required for OpenAI-compatible APIs.",
      "i" = "Use {.fn chat_openai} if you want to use OpenAI's official API."
    ))
  }

  # Fall back to OPENAI_BASE_URL if it exists (for backwards compatibility)
  if (is.null(base_url) || base_url == "") {
    base_url <- Sys.getenv("OPENAI_BASE_URL", "")
    if (base_url == "") {
      cli::cli_abort("{.arg base_url} must be provided.")
    }
  }

  model <- model %||% cli::cli_abort("{.arg model} is required.")
  echo <- check_echo(echo)

  credentials <- as_credentials(
    "chat_openai_compatible",
    function() openai_key(),
    credentials = credentials,
    api_key = api_key,
    token = TRUE
  )

  provider <- ProviderOpenAICompatible(
    name = name,
    base_url = base_url,
    model = model,
    params = params %||% params(),
    extra_args = api_args,
    extra_headers = api_headers,
    credentials = credentials,
    preserve_thinking = preserve_thinking
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

chat_openai_compatible_test <- function(
  system_prompt = "Be terse.",
  ...,
  model = "gpt-4.1-nano",
  params = NULL,
  echo = "none"
) {
  params <- params %||% params()
  params$seed <- params$seed %||% 1014
  params$temperature <- params$temperature %||% 0

  chat <- chat_openai_compatible(
    # Test with real OpenAI API
    name = "OpenAI",
    base_url = "https://api.openai.com/v1",
    system_prompt = system_prompt,
    model = model,
    params = params,
    ...,
    echo = echo
  )
  chat
}

ProviderOpenAICompatible <- new_class(
  "ProviderOpenAICompatible",
  parent = Provider,
  properties = list(
    preserve_thinking = new_property(class_logical, default = FALSE)
  )
)

openai_key_exists <- function() {
  key_exists("OPENAI_API_KEY")
}

openai_key <- function() {
  key_get("OPENAI_API_KEY")
}

# Base request -----------------------------------------------------------------

method(base_request, ProviderOpenAICompatible) <- function(provider) {
  req <- request(provider@base_url)
  req <- ellmer_req_credentials(req, provider@credentials(), "Authorization")
  req <- ellmer_req_robustify(req)
  req <- ellmer_req_user_agent(req)
  req <- base_request_error(provider, req)
  req
}

method(base_request_error, ProviderOpenAICompatible) <- function(
  provider,
  req
) {
  req_error(req, body = function(resp) {
    if (resp_content_type(resp) == "application/json") {
      error <- resp_body_json(resp)$error
      if (is_string(error)) {
        error
      } else if (is.list(error)) {
        error$message
      } else {
        prettify(resp_body_string(resp))
      }
    } else if (resp_content_type(resp) == "text/plain") {
      resp_body_string(resp)
    }
  })
}

# Chat endpoint ----------------------------------------------------------------

method(chat_path, ProviderOpenAICompatible) <- function(provider) {
  "/chat/completions"
}

# https://platform.openai.com/docs/api-reference/chat/create
method(chat_body, ProviderOpenAICompatible) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  messages <- compact(unlist(as_json(provider, turns), recursive = FALSE))
  tools <- as_json(provider, unname(tools))

  if (!is.null(type)) {
    response_format <- list(
      type = "json_schema",
      json_schema = list(
        name = "structured_data",
        schema = as_json(provider, type),
        strict = TRUE
      )
    )
  } else {
    response_format <- NULL
  }

  params <- chat_params(provider, provider@params)

  compact(list2(
    messages = messages,
    model = provider@model,
    !!!params,
    stream = stream,
    stream_options = if (stream) list(include_usage = TRUE),
    tools = tools,
    response_format = response_format
  ))
}


method(chat_params, ProviderOpenAICompatible) <- function(provider, params) {
  standardise_params(
    params,
    c(
      frequency_penalty = "frequency_penalty",
      logprobs = "log_probs",
      max_completion_tokens = "max_tokens",
      presence_penalty = "presence_penalty",
      seed = "seed",
      stop = "stop_sequences",
      temperature = "temperature",
      top_logprobs = "top_k",
      top_p = "top_p"
    )
  )
}

# OpenAI -> ellmer --------------------------------------------------------------

method(stream_parse, ProviderOpenAICompatible) <- function(provider, event) {
  if (is.null(event) || identical(event$data, "[DONE]")) {
    return(NULL)
  }

  jsonlite::parse_json(event$data)
}
method(stream_content, ProviderOpenAICompatible) <- function(provider, event) {
  if (length(event$choices) == 0) {
    return(NULL)
  }
  delta <- event$choices[[1]]$delta

  reasoning <- delta[["reasoning_content"]]
  if (!is.null(reasoning)) {
    return(ContentThinking(reasoning))
  }

  text <- delta[["content"]]
  if (is.null(text)) {
    return(NULL)
  }
  ContentText(text)
}
method(stream_merge_chunks, ProviderOpenAICompatible) <- function(
  provider,
  result,
  chunk
) {
  if (is.null(result)) {
    chunk
  } else {
    merge_dicts(result, chunk)
  }
}

method(value_tokens, ProviderOpenAICompatible) <- function(provider, json) {
  usage <- json$usage
  cached_tokens <- usage$prompt_tokens_details$cached_tokens %||% 0

  tokens(
    input = (usage$prompt_tokens %||% 0) - cached_tokens,
    output = usage$completion_tokens %||% 0,
    cached_input = cached_tokens
  )
}

method(value_turn, ProviderOpenAICompatible) <- function(
  provider,
  result,
  has_type = FALSE
) {
  if (has_name(result$choices[[1]], "delta")) {
    # streaming
    message <- result$choices[[1]]$delta
  } else {
    message <- result$choices[[1]]$message
  }

  thinking <- list()
  reasoning <- message$reasoning_content
  if (is_string(reasoning) && nzchar(reasoning)) {
    thinking <- list(ContentThinking(reasoning))
  }

  if (has_type) {
    if (is_string(message$content)) {
      content <- list(ContentJson(string = message$content[[1]]))
    } else {
      content <- list(ContentJson(data = message$content))
    }
  } else {
    # Some providers (e.g. Databricks) return content: "" instead of
    # content: null for tool-only turns; treat empty strings as null
    if (is_string(message$content) && !nzchar(message$content)) {
      content <- list()
    } else {
      content <- lapply(message$content, as_content)
    }
  }
  if (has_name(message, "tool_calls")) {
    calls <- lapply(message$tool_calls, function(call) {
      name <- call$`function`$name
      # TODO: record parsing error
      args <- tryCatch(
        jsonlite::parse_json(call$`function`$arguments),
        error = function(cnd) list()
      )
      ContentToolRequest(name = name, arguments = args, id = call$id)
    })
    content <- c(content, calls)
  }

  content <- c(thinking, content)

  tokens <- value_tokens(provider, result)
  cost <- get_token_cost(provider, tokens)
  AssistantTurn(content, json = result, tokens = unlist(tokens), cost = cost)
}

# ellmer -> OpenAI --------------------------------------------------------------

method(as_json, list(ProviderOpenAICompatible, Turn)) <- function(
  provider,
  x,
  ...
) {
  if (is_system_turn(x)) {
    list(
      list(role = "system", content = x@contents[[1]]@text)
    )
  } else if (is_user_turn(x)) {
    # Tool results come out of content and go into own element
    x <- turn_contents_expand(x)
    data <- turn_split_tool_results(x)

    if (length(data$contents) > 0) {
      content <- as_json(provider, data$contents, ...)
      user <- list(list(role = "user", content = content))
    } else {
      user <- list()
    }

    tools <- lapply(data$tool_results, function(tool) {
      list(
        role = "tool",
        content = tool_string(tool),
        tool_call_id = tool@request@id
      )
    })

    c(tools, user)
  } else if (is_assistant_turn(x)) {
    # Drop empty ContentText to avoid API errors from providers that reject
    # empty text content blocks (e.g. Databricks returns content: "" for
    # tool-only turns, then rejects it on replay)
    is_empty_text <- map_lgl(x@contents, function(c) {
      S7_inherits(c, ContentText) && !nzchar(c@text)
    })
    contents <- x@contents[!is_empty_text]
    if (length(contents) == 0) {
      return(NULL)
    }
    # Thinking and tool requests come out of content
    is_thinking <- map_lgl(contents, S7_inherits, ContentThinking)
    is_tool <- map_lgl(contents, is_tool_request)
    other <- contents[!is_thinking & !is_tool]
    content <- as_json(provider, other, ...)
    tool_calls <- as_json(provider, contents[is_tool], ...)

    reasoning_content <- NULL
    if (provider@preserve_thinking) {
      thinking_texts <- map_chr(contents[is_thinking], function(c) c@thinking)
      if (length(thinking_texts) > 0) {
        reasoning_content <- paste0(thinking_texts, collapse = "")
      }
    }

    list(
      compact(list(
        role = "assistant",
        reasoning_content = reasoning_content,
        content = content,
        tool_calls = tool_calls
      ))
    )
  } else {
    cli::cli_abort("Unknown role {x@role}", .internal = TRUE)
  }
}

method(as_json, list(ProviderOpenAICompatible, ContentText)) <- function(
  provider,
  x,
  ...
) {
  list(type = "text", text = x@text)
}

method(as_json, list(ProviderOpenAICompatible, ContentImageRemote)) <- function(
  provider,
  x,
  ...
) {
  list(type = "image_url", image_url = list(url = x@url))
}

method(as_json, list(ProviderOpenAICompatible, ContentImageInline)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "image_url",
    image_url = list(
      url = paste0("data:", x@type, ";base64,", x@data)
    )
  )
}

method(as_json, list(ProviderOpenAICompatible, ContentPDF)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "file",
    file = list(
      filename = x@filename,
      file_data = paste0("data:application/pdf;base64,", x@data)
    )
  )
}

method(as_json, list(ProviderOpenAICompatible, ContentToolRequest)) <- function(
  provider,
  x,
  ...
) {
  json_args <- to_json(x@arguments)
  list(
    id = x@id,
    `function` = list(name = x@name, arguments = json_args),
    type = "function"
  )
}

method(as_json, list(ProviderOpenAICompatible, ToolDef)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "function",
    "function" = compact(list(
      name = x@name,
      description = x@description,
      strict = TRUE,
      parameters = as_json(provider, x@arguments, ...)
    ))
  )
}


method(as_json, list(ProviderOpenAICompatible, TypeObject)) <- function(
  provider,
  x,
  ...
) {
  if (x@additional_properties) {
    cli::cli_abort("{.arg .additional_properties} not supported for OpenAI.")
  }

  names <- names2(x@properties)
  properties <- lapply(x@properties, function(x) {
    out <- as_json(provider, x, ...)
    if (!x@required) {
      out$type <- c(out$type, "null")
    }
    out
  })

  names(properties) <- names

  list(
    type = "object",
    description = x@description %||% "",
    properties = properties,
    required = as.list(names),
    additionalProperties = FALSE
  )
}


# Batched requests -------------------------------------------------------------

method(has_batch_support, ProviderOpenAICompatible) <- function(provider) {
  FALSE
}

#' Open a live chat application
#'
#' @description
#'
#' * `live_console()` lets you chat interactively in the console.
#' * `live_browser()` lets you chat interactively in a browser.
#'
#' Note that these functions will mutate the input `chat` object as
#' you chat because your turns will be appended to the history.
#'
#' @param chat A chat object created by [chat_openai()] or friends.
#' @param quiet If `TRUE`, suppresses the initial message that explains how
#'   to use the console.
#' @export
#' @returns (Invisibly) The input `chat`.
#' @examples
#' \dontrun{
#' chat <- chat_anthropic()
#' live_console(chat)
#' live_browser(chat)
#' }
live_console <- function(chat, quiet = FALSE) {
  if (!is_interactive()) {
    cli::cli_abort("The chat console is only available in interactive mode.")
  }

  if (!isTRUE(quiet)) {
    cli::cat_boxx(
      c(
        "Entering chat console.",
        "Use \"\"\" for multi-line input.",
        "Type 'Q' to quit."
      ),
      padding = c(0, 1, 0, 1),
      border_style = "double"
    )
  }

  repeat {
    user_input <- readline(prompt = ">>> ")

    if (user_input == "Q") {
      break
    }

    if (!grepl("\\S", user_input)) {
      next
    }

    if (grepl('^\\s*"""', user_input)) {
      repeat {
        next_input <- readline(prompt = '... ')
        user_input <- paste0(user_input, "\n", next_input)
        if (grepl('"""\\s*$', next_input)) {
          break
        }
      }
      # Strip leading and trailing """, using regex
      user_input <- gsub('^\\s*"""\\s*', '', user_input)
      user_input <- gsub('\\s*"""\\s*$', '', user_input)
    }

    # Process the input using the provided LLM function
    chat$chat(user_input, echo = TRUE)
    cat("\n")
  }

  invisible(chat)
}

#' @export
#' @rdname live_console
live_browser <- function(chat, quiet = FALSE) {
  check_installed("shiny")
  check_installed("shinychat", version = "0.1.1.9001")

  if (!isTRUE(quiet)) {
    cli::cat_boxx(
      c("Entering interactive chat", "Press Ctrl+C to quit."),
      padding = c(0, 1, 0, 1),
      border_style = "double"
    )
  }

  tryCatch(
    shiny::runGadget(shinychat::chat_app(chat, options = list(quiet = TRUE))),
    interrupt = function(cnd) NULL
  )
  invisible(chat)
}

fixture_list_of_tools <- function() {
  list(
    tool_scalar = tool(
      function() 1,
      name = "tool_scalar",
      description = "Tool"
    ),
    my_tool = tool(
      function() 1,
      name = "my_tool",
      description = "Tool"
    ),
    tool_list = tool(
      function() list(a = 1, b = 2),
      name = "tool_list",
      description = "Tool"
    ),
    tool_chr = tool(
      function() letters[1:3],
      name = "tool_chr",
      description = "Tool"
    ),
    tool_abort = tool(
      function() {
        cli::cli_abort(c(
          "Unexpected input",
          "i" = "Please revise and try again."
        ))
      },
      name = "tool_abort",
      description = "Tool"
    )
  )
}

fixture_turn_with_tool_requests <- function(with_tool = TRUE) {
  tools <- fixture_list_of_tools()

  req_success <- ContentToolRequest(
    id = "x1",
    name = "my_tool",
    arguments = list(),
    tool = if (with_tool) tools$my_tool
  )
  req_fail <- ContentToolRequest(
    id = "x2",
    name = "my_tool",
    arguments = list(x = 1),
    tool = if (with_tool) tools$my_tool
  )
  req_list <- ContentToolRequest(
    id = "x3",
    name = "tool_list",
    arguments = list(),
    tool = if (with_tool) tools$tool_list
  )
  req_chr <- ContentToolRequest(
    id = "x4",
    name = "tool_chr",
    arguments = list(),
    tool = if (with_tool) tools$tool_chr
  )
  req_abort <- ContentToolRequest(
    id = "x5",
    name = "tool_abort",
    arguments = list(),
    tool = if (with_tool) tools$tool_abort
  )

  AssistantTurn(
    list(req_success, req_fail, req_list, req_chr, req_abort)
  )
}

#' @include tools-def.R
NULL

#' Record and replay content
#'
#' @description
#' These generic functions can be use to convert [Turn]/[Content] objects
#' into easily serializable representations (i.e. lists and atomic vectors).
#'
#' * `contents_record()` accepts a [Turn] or [Content] and return a simple list.
#' * `contents_replay()` takes the output of `contents_record()` and returns
#'   a [Turn] or [Content] object.
#'
#' @param x A [Turn] or [Content] object to serialize; or a serialized object
#'   to replay.
#' @param tools A named list of tools
#' @keywords internal
#' @export
contents_record <- function(x) {
  class_name <- class(x)[[1]]

  check_is_ellmer_object(x)

  # Assume that we can replay attributes (the underlying data in an S7 object)
  # to the constructor, i.e. do.call(S7_class(obj), attributes(obj))
  # This avoids inspecting read-only and/or dynamic properties
  attr <- attributes(x)
  serializable_props <- intersect(prop_names(x), names(attr))

  # Don't serialize the actual tool definition, as it really belongs to the Chat
  # object() and on replay we will re-apply by matching to existing tools. Note
  # that it's not possible for users to extend this class because ellmer always
  # creates the tool request object, so we can check the class directly.
  if (class_name == "ellmer::ContentToolRequest") {
    serializable_props <- setdiff(serializable_props, "tool")
  }

  prop_values <- lapply(attr[serializable_props], function(value) {
    if (S7_inherits(value)) {
      # Recursive record for S7 objects
      contents_record(value)
    } else if (is_list_of_s7_objects(value)) {
      # Make record of each item in list
      lapply(value, contents_record)
    } else {
      value
    }
  })

  recorded_object(class_name, prop_values)
}

#' @rdname contents_record
#' @param .envir The environment in which to look for class definitions. Used
#'   when the recorded objects include classes that extend [ellmer::Turn] or
#'   [ellmer::Content] but are not from the \pkg{ellmer} package itself.
#' @export
contents_replay <- function(x, tools = list(), .envir = parent.frame()) {
  check_recorded(x)

  class <- recorded_class_info(x, .envir = .envir)

  obj_props <- map(x$props, function(prop_value) {
    if (is_list_of_recorded_objects(prop_value)) {
      # If the prop is a list of recorded objects, replay each one
      map(prop_value, contents_replay, tools = tools, .envir = .envir)
    } else if (is_recorded_object(prop_value)) {
      # If the prop is a recorded object, replay it
      contents_replay(prop_value, tools = tools, .envir = .envir)
    } else {
      prop_value
    }
  })

  env <- if (!is.null(class$pkg)) ns_env(class$pkg) else .envir

  # This is a bit of overkill, but gives nicer tracebacks
  out <- exec(class$name, !!!obj_props, .env = env)

  if (S7_inherits(out, Turn)) {
    out <- match_tools(out, tools)
  }
  out
}

# Helpers ----------------------------------------------------------------------

check_is_ellmer_object <- function(x) {
  if (S7_inherits(x, ellmer::Content) || S7_inherits(x, ellmer::Turn)) {
    return(invisible(x))
  }

  cli::cli_abort(
    c(
      "Cannot record or replay {.obj_type_friendly {x}}.",
      "i" = "Only {.code ellmer::Content} or {.code ellmer::Turn} classes or subclasses are currently supported."
    ),
    call = caller_env()
  )
}

recorded_object <- function(class, props) {
  list(
    version = 1,
    class = class,
    props = props
  )
}

is_recorded_object <- function(x) {
  is.list(x) && all(c("version", "class", "props") %in% names(x))
}

is_list_of_s7_objects <- function(x) {
  is.list(x) && all(map_lgl(x, S7_inherits))
}

is_list_of_recorded_objects <- function(x) {
  is.list(x) && all(map_lgl(x, is_recorded_object))
}

check_recorded <- function(recorded, call = caller_env()) {
  if (!is_recorded_object(recorded)) {
    cli::cli_abort(
      "Expected the recorded object to be a list with at least names 'version', 'class', and 'props'.",
      call = call
    )
  }

  if (!identical(recorded$version, 1)) {
    cli::cli_abort("Unsupported version {.val {recorded$version}}.")
  }

  if (!is_string(recorded$class)) {
    cli::cli_abort(
      "Expected the recorded object to have a single $class name, containing `::` if the class is from a package.",
      call = call
    )
  }
}

recorded_class_info <- function(x, .envir = parent.frame()) {
  class_split <- strsplit(x$class, "::")[[1]]
  if (length(class_split) > 2) {
    cli::cli_abort(
      "Expected the class to be in the form `package::ClassName`, not {.val {x$class}}.",
      call = caller_env()
    )
  }

  if (length(class_split) < 2) {
    pkg <- NULL
    name <- class_split[[1]]
  } else {
    pkg <- class_split[[1]]
    name <- class_split[[2]]
  }

  check_recorded_class(pkg, name, .envir)

  list(pkg = pkg, name = name)
}

check_recorded_class <- function(pkg, name, .envir = parent.frame()) {
  if (is.null(pkg)) {
    cls_fmt <- name
    cls <- get0(name, envir = .envir, inherits = TRUE)
  } else {
    cls_fmt <- sprintf("%s::%s", pkg, name)
    cls <- ns_env(pkg)[[name]]
  }

  if (is.null(cls)) {
    cli::cli_abort(
      "Unable to find the S7 class: {.code {cls_fmt}}.",
      call = caller_env(n = 2)
    )
  }
  if (!S7_inherits(cls)) {
    cli::cli_abort(
      "Expected the object named {.code {cls_fmt}} to be an S7 class, not {.obj_type_friendly {cls}}.",
      call = caller_env(n = 2)
    )
  }

  invisible(cls)
}

test_record_replay <- function(x, tools = list(), .envir = parent.frame()) {
  recorded <- contents_record(x)
  replayed <- contents_replay(recorded, tools = tools, .envir = .envir)
  expect_equal(replayed, x)
}

test_that("can serialize simple objects", {
  expect_equal(
    contents_record(Content()),
    recorded_object("ellmer::Content", set_names(list()))
  )

  expect_equal(
    contents_record(ContentText("hello world")),
    recorded_object("ellmer::ContentText", list(text = "hello world"))
  )
})

test_that("can round trip of Turn record/replay", {
  test_record_replay(UserTurn())

  test_record_replay(UserTurn(
    list(
      ContentText("hello world"),
      ContentText("hello world2")
    )
  ))
})

test_that("can round trip simple content types", {
  test_record_replay(Content())
  test_record_replay(ContentText("hello world"))
  test_record_replay(ContentImageInline("image/png", "abcd123"))
  test_record_replay(ContentImageRemote("https://example.com/image.jpg"))
  test_record_replay(ContentJson(list(a = 1:2, b = "apple")))
  test_record_replay(ContentThinking("A **thought**."))
  test_record_replay(ContentUploaded("https://example.com/image.jpg"))
  test_record_replay(ContentPDF("TYPE", "DATA", "FILENAME"))
})

test_that("can round trip of ContentToolRequest/ContentToolResult", {
  request <- ContentToolRequest(
    "ID",
    "tool_name",
    list(a = 1:2, b = "apple")
  )
  result <- ContentToolResult(
    value = "VALUE",
    extra = list(extra = 1:2, b = "apple")
  )
  test_record_replay(request)
  test_record_replay(result)
})

test_that("can re-match tools if present", {
  turn <- UserTurn(list(ContentToolRequest("123", "mytool")))
  recorded <- contents_record(turn)

  mytool <- tool(function() {}, "mytool")
  replayed <- contents_replay(recorded, tools = list(mytool = mytool))
  expect_equal(replayed@contents[[1]]@tool, mytool)

  # If no match, it still works, but tool is left as NULL
  replayed <- contents_replay(recorded, tools = list())
  expect_equal(replayed@contents[[1]]@tool, NULL)
})

test_that("can re-match tools if present", {
  mytool <- tool(function() {}, "mytool")

  request <- ContentToolRequest("123", "mytool", tool = mytool)
  result <- ContentToolResult("value", request = request)

  turn_request <- UserTurn(list(request))
  turn_result <- UserTurn(list(result))

  test_record_replay(turn_request, tools = list(mytool = mytool))
  test_record_replay(turn_result, tools = list(mytool = mytool))

  # If no tool match, it still works, but tool is left as NULL
  replayed_turn_request <- contents_replay(
    contents_record(turn_request),
    tools = list()
  )
  expect_null(replayed_turn_request@contents[[1]]@tool)

  replayed_turn_result <- contents_replay(
    contents_record(turn_result),
    tools = list()
  )
  expect_null(replayed_turn_result@contents[[1]]@request@tool)
})

test_that("checks recorded value types", {
  bad_names <- list()
  bad_version <- list(version = 2, class = "ellmer::Content", props = list())
  bad_class <- list(version = 1, class = c("a", "b"), props = list())
  expect_snapshot(error = TRUE, {
    contents_replay(bad_names)
    contents_replay(bad_version)
    contents_replay(bad_class)
  })
})

test_that("non-ellmer classes are not recorded/replayed by default", {
  LocalClass <- S7::new_class("LocalClass", package = "foo")
  recorded <- list(version = 1, class = "foo::LocalClass", props = list())

  expect_snapshot(error = TRUE, {
    contents_record(LocalClass())
    contents_replay(recorded)
  })
})

test_that("packaged classes that extend ellmer classes can be replayed", {
  env <- rlang::current_env()

  local_mocked_bindings(
    ns_env = function(x) {
      if (x == "foo") {
        return(env)
      }
      rlang::ns_env(x)
    }
  )

  LocalContentText <- S7::new_class(
    "LocalContentText",
    package = "foo",
    parent = ellmer::ContentText
  )

  test_record_replay(LocalContentText("hello world"))
  test_record_replay(
    UserTurn(list(LocalContentText(text = "hello world")))
  )
})

test_that("local classes that extend ellmer classes can be replayed", {
  LocalContentText <- S7::new_class(
    "LocalContentText",
    package = NULL,
    parent = ellmer::ContentText
  )

  test_record_replay(LocalContentText("hello world"))
  test_record_replay(
    UserTurn(list(LocalContentText(text = "hello world")))
  )
})

test_that("local classes that extend ellmer classes can be replayed", {
  test_content <- S7::new_class(
    "LocalContentText",
    package = NULL,
    parent = ellmer::ContentText
  )

  LocalContentText <- function(...) {
    test_content(...)
  }

  expect_snapshot(
    error = TRUE,
    test_record_replay(test_content("hello world"))
  )
})

test_that("replayed objects must be existing S7 classes", {
  doesnt_exist <- list(version = 1, class = "ellmer::Turn2", props = list())
  not_s7 <- list(version = 1, class = "ellmer::chat_openai", props = list())

  expect_snapshot(error = TRUE, {
    contents_replay(doesnt_exist)
    contents_replay(not_s7)
  })
})

#' Chat with a model hosted on Databricks
#'
#' @description
#' Databricks provides out-of-the-box access to a number of [foundation
#' models](https://docs.databricks.com/en/machine-learning/model-serving/score-foundation-models.html)
#' and can also serve as a gateway for external models hosted by a third party.
#'
#' Built on top of [chat_openai_compatible()].
#'
#' ## Authentication
#'
#' `chat_databricks()` picks up on ambient Databricks credentials for a subset
#' of the [Databricks client unified
#' authentication](https://docs.databricks.com/en/dev-tools/auth/unified-auth.html)
#' model. Specifically, it supports:
#'
#' - Personal access tokens
#' - Service principals via OAuth (OAuth M2M)
#' - User account via OAuth (OAuth U2M)
#' - Authentication via the Databricks CLI
#' - Posit Workbench-managed credentials
#' - Viewer-based credentials on Posit Connect. Requires the \pkg{connectcreds}
#'   package.
#'
#' @family chatbots
#' @param workspace The URL of a Databricks workspace, e.g.
#'   `"https://example.cloud.databricks.com"`. Will use the value of the
#'   environment variable `DATABRICKS_HOST`, if set.
#' @param model `r param_model("databricks-claude-3-7-sonnet")`
#'
#'   Available foundational models include:
#'
#'   - `databricks-claude-3-7-sonnet` (the default)
#'   - `databricks-mixtral-8x7b-instruct`
#'   - `databricks-meta-llama-3-1-70b-instruct`
#'   - `databricks-meta-llama-3-1-405b-instruct`
#' @param token An authentication token for the Databricks workspace, or
#'   `NULL` to use ambient credentials.
#' @param params Common model parameters, usually created by [params()].
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_databricks()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_databricks <- function(
  workspace = databricks_workspace(),
  system_prompt = NULL,
  model = NULL,
  token = NULL,
  params = NULL,
  api_args = list(),
  echo = c("none", "output", "all"),
  api_headers = character()
) {
  check_string(workspace, allow_empty = FALSE)
  check_string(token, allow_empty = FALSE, allow_null = TRUE)
  model <- set_default(model, "databricks-claude-3-7-sonnet")
  echo <- check_echo(echo)
  if (!is.null(token)) {
    credentials <- function() list(Authorization = paste("Bearer", token))
  } else {
    credentials <- default_databricks_credentials(workspace)
  }

  params <- params %||% params()

  provider <- ProviderDatabricks(
    name = "Databricks",
    base_url = workspace,
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderDatabricks <- new_class(
  "ProviderDatabricks",
  parent = ProviderOpenAICompatible
)

method(base_request, ProviderDatabricks) <- function(provider) {
  req <- request(provider@base_url)
  req <- ellmer_req_credentials(req, provider@credentials())
  req <- ellmer_req_robustify(req)
  req <- ellmer_req_user_agent(req, databricks_user_agent())
  req <- base_request_error(provider, req)
  req
}

method(chat_body, ProviderDatabricks) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  body <- chat_body(
    super(provider, ProviderOpenAICompatible),
    stream = stream,
    turns = turns,
    tools = tools,
    type = type
  )

  params <- chat_params(provider, provider@params)
  body <- modify_list(body, params)

  # Databricks doesn't support stream options
  body$stream_options <- NULL

  body
}

method(chat_params, ProviderDatabricks) <- function(provider, params) {
  # https://docs.databricks.com/aws/en/machine-learning/foundation-model-apis/api-reference#chat-request
  standardise_params(
    params,
    c(
      temperature = "temperature",
      top_p = "topP",
      top_k = "topK",
      max_tokens = "maxTokens",
      stop_sequences = "stopSequences"
    )
  )
}

method(chat_path, ProviderDatabricks) <- function(provider) {
  # Note: this API endpoint is undocumented and seems to exist primarily for
  # compatibility with the OpenAI Python SDK. The documented endpoint is
  # `/serving-endpoints/<model>/invocations`.
  "/serving-endpoints/chat/completions"
}

method(base_request_error, ProviderDatabricks) <- function(provider, req) {
  req_error(req, body = function(resp) {
    if (resp_content_type(resp) == "application/json") {
      # Databrick's "OpenAI-compatible" API has a slightly incompatible error
      # response format, which we account for here.
      resp_body_json(resp)$message
    }
  })
}

# See: https://docs.databricks.com/aws/en/machine-learning/foundation-model-apis/api-reference#functionobject
method(as_json, list(ProviderDatabricks, ToolDef)) <- function(
  provider,
  x,
  ...
) {
  # Note: It seems that Databricks doesn't support the "strict" field, despite
  # what their documentation says. It *is* supported for structured outputs,
  # though. I suspect a copy & paste error in their docs.
  compact(list(
    type = "function",
    "function" = compact(list(
      name = x@name,
      description = x@description,
      # Use the same parameter encoding as the OpenAI provider, but only if
      # there actually are parameters.
      parameters = if (length(x@arguments@properties) != 0) {
        as_json(provider, x@arguments, ...)
      }
    ))
  ))
}

# https://docs.databricks.com/aws/en/machine-learning/foundation-model-apis/api-reference#toolcall
method(as_json, list(ProviderDatabricks, ContentToolRequest)) <- function(
  provider,
  x,
  ...
) {
  # Databricks seems to require encoding empty arguments as an empty
  # dictionary, rather than an empty array.
  json_args <- to_json(set_names(x@arguments))
  list(
    id = x@id,
    `function` = list(name = x@name, arguments = json_args),
    type = "function"
  )
}

databricks_workspace <- function(error_call = caller_env()) {
  # Default to the well-known DATABRICKS_HOST environment variable.
  host <- Sys.getenv("DATABRICKS_HOST")
  if (nchar(host) != 0) {
    if (!grepl("^https?://", host)) {
      host <- paste0("https://", host)
    }
    return(host)
  }
  # If we have no environment variable, check for a Databricks config file
  # (likely generated via the Databricks CLI or their VS Code extension).
  cfg_file <- Sys.getenv("DATABRICKS_CONFIG_FILE", "~/.databrickscfg")
  if (file.exists(cfg_file)) {
    profile <- Sys.getenv("DATABRICKS_CONFIG_PROFILE", "DEFAULT")
    host <- databricks_config_field(cfg_file, profile, "host")
    if (!is.null(host)) {
      return(host)
    }
  }
  if (is_testing()) {
    testthat::skip("no Databricks workspace configured")
  }
  cli::cli_abort(
    "No env var {.code DATABRICKS_HOST} set or valid {.file {cfg_file}} found.",
    call = error_call
  )
}

databricks_user_agent <- function() {
  user_agent <- paste0("r-ellmer/", utils::packageVersion("ellmer"))
  if (nchar(Sys.getenv("SPARK_CONNECT_USER_AGENT")) != 0) {
    user_agent <- paste(Sys.getenv("SPARK_CONNECT_USER_AGENT"), user_agent)
  }
  user_agent
}

# Try various ways to get Databricks credentials. This implements a subset of
# the "Databricks client unified authentication" model.
default_databricks_credentials <- function(workspace = databricks_workspace()) {
  host <- gsub("https://|/$", "", workspace)

  # Detect viewer-based credentials from Posit Connect.
  if (has_connect_viewer_token(resource = workspace)) {
    return(function() {
      token <- connectcreds::connect_viewer_token(workspace)
      list(Authorization = paste("Bearer", token$access_token))
    })
  }

  # An explicit PAT takes precedence over everything else.
  token <- Sys.getenv("DATABRICKS_TOKEN")
  if (nchar(token)) {
    return(function() list(Authorization = paste("Bearer", token)))
  }

  # Next up are explicit OAuth2 M2M credentials.
  client_id <- Sys.getenv("DATABRICKS_CLIENT_ID")
  client_secret <- Sys.getenv("DATABRICKS_CLIENT_SECRET")
  if (nchar(client_id) && nchar(client_secret)) {
    # M2M credentials use an OAuth client credentials flow. We cache the token
    # so we don't need to perform this flow before each turn.
    client <- oauth_client(
      client_id,
      paste0("https://", host, "/oidc/v1/token"),
      secret = client_secret,
      auth = "header",
      name = "ellmer-databricks-m2m"
    )
    return(function() {
      token <- oauth_token_cached(
        client,
        oauth_flow_client_credentials,
        # The "all-apis" scope translates to "everything this service principal
        # has access to", not "all Databricks APIs".
        flow_params = list(scope = "all-apis"),
        # Don't use the cached token when testing.
        reauth = is_testing()
      )
      list(Authorization = paste("Bearer", token$access_token))
    })
  }

  # Check for Workbench-provided credentials.
  cfg_file <- Sys.getenv("DATABRICKS_CONFIG_FILE")
  if (grepl("posit-workbench", cfg_file, fixed = TRUE)) {
    wb_token <- workbench_databricks_token(host, cfg_file)
    if (!is.null(wb_token)) {
      return(function() {
        # Ensure we get an up-to-date token.
        token <- workbench_databricks_token(host, cfg_file)
        list(Authorization = paste("Bearer", token))
      })
    }
  }

  # When on desktop, try using the Databricks CLI for auth.
  cli_path <- Sys.getenv("DATABRICKS_CLI_PATH", "databricks")
  if (!is_hosted_session() && nchar(Sys.which(cli_path)) != 0) {
    token <- databricks_cli_token(cli_path, workspace)
    if (!is.null(token)) {
      return(function() {
        # Ensure we get an up-to-date token.
        token <- databricks_cli_token(cli_path, workspace)
        list(Authorization = paste("Bearer", token))
      })
    }
  }

  # Fall back to OAuth U2M, masquerading as the Databricks CLI. Again, this
  # only works on desktop.
  if (is_interactive() && !is_hosted_session()) {
    # U2M credentials use an OAuth authorization code flow. We cache the token
    # so we don't need to perform this flow before each turn.
    client <- oauth_client(
      "databricks-cli",
      paste0("https://", host, "/oidc/v1/token"),
      auth = "body",
      name = "ellmer-databricks-u2m"
    )
    return(function() {
      token <- oauth_token_cached(
        client,
        oauth_flow_auth_code,
        flow_params = list(
          auth_url = paste0("https://", host, "/oidc/v1/authorize"),
          # The "all-apis" scope translates to "everything this user has access
          # to", not "all Databricks APIs".
          scope = "all-apis offline_access",
          # This is the registered redirect URI for the Databricks CLI.
          redirect_uri = "http://localhost:8020"
        )
      )
      list(Authorization = paste("Bearer", token$access_token))
    })
  }

  if (is_testing()) {
    testthat::skip("no Databricks credentials available")
  }

  cli::cli_abort("No Databricks credentials are available.")
}

# Try to determine whether we can redirect the user's browser to a server on
# localhost, which isn't possible if we are running on a hosted platform.
#
# This is based on the strategy pioneered by the {gargle} package and {httr2}.
is_hosted_session <- function() {
  # If RStudio Server or Posit Workbench is running locally (which is possible,
  # though unusual), it's not acting as a hosted environment.
  Sys.getenv("RSTUDIO_PROGRAM_MODE") == "server" &&
    !grepl("localhost", Sys.getenv("RSTUDIO_HTTP_REFERER"), fixed = TRUE)
}

databricks_cli_token <- function(cli_path, host) {
  output <- suppressWarnings(
    system2(
      cli_path,
      c("auth", "token", "--host", host),
      stdout = TRUE,
      stderr = TRUE
    )
  )
  output <- paste(output, collapse = "\n")
  # If we don't get an error message, try to extract the token from the JSON-
  # formatted output.
  if (grepl("access_token", output, fixed = TRUE)) {
    token <- gsub(".*access_token\":\\s?\"([^\"]+).*", "\\1", output)
    return(token)
  }
  NULL
}

# Reads Posit Workbench-managed Databricks credentials from a
# $DATABRICKS_CONFIG_FILE. The generated file will look as follows:
#
# [workbench]
# host = some-host
# token = some-token
workbench_databricks_token <- function(host, cfg_file) {
  databricks_config_field(cfg_file, "workbench", "token")
}

databricks_config_field <- function(cfg_file, profile, field) {
  cfg <- readLines(cfg_file)
  # We don't attempt a full parse of the INI syntax supported by Databricks
  # config files and use the following approach to cheat, instead:
  #
  # 1. Find the indices of the profile header and the field. If none appear in
  #    the file, we're done.
  # 2. Find the index of the next profile header, if any.
  # 3. Find the index -- if any -- of the field between these headers.
  # 4. Extract the field's value.
  profile <- which(grepl(sprintf("[%s]", profile), cfg, fixed = TRUE))
  if (length(profile) != 1) {
    return(NULL)
  }
  entry <- which(grepl(field, cfg, fixed = TRUE))
  entry <- entry[entry > profile]
  if (!any(entry)) {
    return(NULL)
  }
  headers <- which(startsWith(cfg, "["))
  next_profile <- headers[headers > profile]
  if (!any(next_profile)) {
    next_profile <- length(cfg) + 1
  } else {
    next_profile <- next_profile[1]
  }
  entry <- entry[entry < next_profile]
  if (length(entry) != 1) {
    return(NULL)
  }
  gsub(sprintf("%s\\s*=\\s*(.*)", field), "\\1", cfg[entry])
}

test_that("can make simple batch request", {
  chat <- chat_databricks(
    system_prompt = "Be as terse as possible; no punctuation"
  )
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_databricks(
    system_prompt = "Be as terse as possible; no punctuation"
  )
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  # Setting a dummy host ensures we don't skip this test, even if there are no
  # Databricks credentials available.
  withr::local_envvar(DATABRICKS_HOST = "https://example.cloud.databricks.com")
  expect_snapshot(. <- chat_databricks())
})

test_that("supports tool calling", {
  test_tools_simple(chat_databricks)
})

test_that("can continue chat after empty assistant content", {
  # Databricks returns content: "" for tool-only turns (#932). Replaying that
  # empty text back used to cause HTTP 400. Verify that a conversation can

  # continue after an assistant turn with only empty content.
  chat <- chat_databricks(system_prompt = "Be as terse as possible")
  chat$set_turns(list(
    UserTurn(list(ContentText("Hello"))),
    AssistantTurn(list(ContentText("")))
  ))
  result <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(result, "2")
})

test_that("can extract data", {
  test_data_extraction(chat_databricks)
})

test_that("can use images", {
  test_images_inline(chat_databricks)
  # remote images don't appear to work
  # test_images_remote(chat_databricks)
})

# Auth --------------------------------------------------------------------

test_that("Databricks PATs are detected correctly", {
  withr::local_envvar(
    DATABRICKS_HOST = "https://example.cloud.databricks.com",
    DATABRICKS_TOKEN = "token"
  )
  credentials <- default_databricks_credentials()
  expect_equal(credentials(), list(Authorization = "Bearer token"))
})

test_that("Databricks CLI tokens are detected correctly", {
  # Emulate a ~/.databrickscfg file written by other tooling.
  cfg_file <- tempfile("databricks", fileext = ".cfg")
  writeLines(
    c(
      '[DEFAULT]',
      'host      = https://example.cloud.databricks.com',
      'auth_type = databricks-cli',
      ';This profile is autogenerated by the Databricks Extension for VS Code',
      '[ellmer]',
      'host=https://example2.cloud.databricks.com/',
      'auth_type=databricks-cli'
    ),
    cfg_file
  )
  withr::local_envvar(
    DATABRICKS_HOST = NA,
    DATABRICKS_CONFIG_FILE = cfg_file,
    DATABRICKS_CLIENT_ID = NA,
    DATABRICKS_CLIENT_SECRET = NA,
    DATABRICKS_TOKEN = NA
  )
  local_mocked_bindings(
    databricks_cli_token = function(path, host) {
      stopifnot(startsWith(host, "https://"))
      "cli_token"
    }
  )
  # Ensure we can get the Workspace out of the config file.
  workspace <- databricks_workspace()
  expect_equal(workspace, "https://example.cloud.databricks.com")

  # And for a non-default profile, too.
  withr::local_envvar(DATABRICKS_CONFIG_PROFILE = "ellmer")
  expect_equal(databricks_workspace(), "https://example2.cloud.databricks.com/")

  # Ensure we call the CLI for credentials.
  credentials <- default_databricks_credentials(workspace)
  expect_equal(credentials(), list(Authorization = "Bearer cli_token"))
})

test_that("Workbench-managed Databricks credentials are detected correctly", {
  # Emulate a databricks.cfg file written by Workbench.
  db_home <- tempfile("posit-workbench")
  dir.create(db_home)
  writeLines(
    c(
      '[workbench]',
      'host = https://example.cloud.databricks.com',
      'token = token'
    ),
    file.path(db_home, "databricks.cfg")
  )
  withr::local_envvar(
    DATABRICKS_CONFIG_FILE = file.path(db_home, "databricks.cfg"),
    DATABRICKS_CONFIG_PROFILE = "workbench",
    DATABRICKS_HOST = "https://example.cloud.databricks.com",
    DATABRICKS_CLIENT_ID = NA,
    DATABRICKS_CLIENT_SECRET = NA,
    DATABRICKS_TOKEN = NA
  )
  credentials <- default_databricks_credentials()
  expect_equal(credentials(), list(Authorization = "Bearer token"))
})

test_that("M2M authentication requests look correct", {
  withr::local_envvar(
    DATABRICKS_HOST = "https://example.cloud.databricks.com",
    DATABRICKS_CLIENT_ID = "id",
    DATABRICKS_CLIENT_SECRET = "secret",
    DATABRICKS_TOKEN = NA
  )
  local_mocked_responses(function(req) {
    # Snapshot relevant fields of the outgoing request.
    expect_snapshot(str(request_summary(req)))
    response_json(body = list(access_token = "token"))
  })
  credentials <- default_databricks_credentials()
  expect_equal(credentials(), list(Authorization = "Bearer token"))
})

test_that("workspace detection handles URLs with and without an https prefix", {
  withr::with_envvar(
    c(DATABRICKS_HOST = "example.cloud.databricks.com"),
    expect_equal(
      databricks_workspace(),
      "https://example.cloud.databricks.com"
    )
  )
  withr::with_envvar(
    c(DATABRICKS_HOST = "https://example.cloud.databricks.com"),
    expect_equal(
      databricks_workspace(),
      "https://example.cloud.databricks.com"
    )
  )
})

test_that("the user agent respects SPARK_CONNECT_USER_AGENT when set", {
  withr::with_envvar(
    c(SPARK_CONNECT_USER_AGENT = NA),
    expect_match(databricks_user_agent(), "^r-ellmer")
  )
  withr::with_envvar(
    c(SPARK_CONNECT_USER_AGENT = "testing"),
    expect_match(databricks_user_agent(), "^testing r-ellmer")
  )
})

test_that("tokens can be requested from a Connect server", {
  skip_if_not_installed("connectcreds")

  withr::local_envvar(
    DATABRICKS_HOST = "https://example.cloud.databricks.com",
    DATABRICKS_TOKEN = "token"
  )
  connectcreds::local_mocked_connect_responses(token = "token")
  credentials <- default_databricks_credentials()
  expect_equal(credentials(), list(Authorization = "Bearer token"))
})

test_that("chat_databricks() serializes tools correctly", {
  withr::local_envvar(
    DATABRICKS_HOST = "https://example.cloud.databricks.com",
    DATABRICKS_TOKEN = "token"
  )
  chat <- chat_databricks(model = "databricks-claude-3-7-sonnet")
  provider <- chat$get_provider()
  expect_equal(
    as_json(
      provider,
      tool(
        function() format(Sys.Date()),
        name = "current_date",
        description = "Returns the current date in ISO 8601 format."
      )
    ),
    list(
      type = "function",
      "function" = list(
        name = "current_date",
        description = "Returns the current date in ISO 8601 format."
      )
    )
  )
  expect_equal(
    as_json(
      provider,
      tool(
        function(person) {
          if (person == "Joe") "sage green" else "red"
        },
        name = "favourite_colour",
        description = "Returns a person's favourite colour.",
        arguments = list(person = type_string("Name of a person"))
      )
    ),
    list(
      type = "function",
      "function" = list(
        name = "favourite_colour",
        description = "Returns a person's favourite colour.",
        parameters = list(
          type = "object",
          description = "",
          properties = list(
            person = list(
              type = "string",
              description = "Name of a person"
            )
          ),
          required = list("person"),
          additionalProperties = FALSE
        )
      )
    )
  )
})

#' Encode images for chat input
#'
#' These functions are used to prepare image URLs and files for input to the
#' chatbot. The `content_image_url()` function is used to provide a URL to an
#' image, while `content_image_file()` is used to provide the image data itself.
#'
#' @param url The URL of the image to include in the chat input. Can be a
#'   `data:` URL or a regular URL. Valid image types are PNG, JPEG, WebP, and
#'   non-animated GIF.
#' @param detail The [detail
#'   setting](https://platform.openai.com/docs/guides/images/image-input-requirements)
#'   for this image. Can be `"auto"`, `"low"`, or `"high"`.
#' @returns An input object suitable for including in the `...` parameter of
#'   the `chat()`, `stream()`, `chat_async()`, or `stream_async()` methods.
#'
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_openai()
#' chat$chat(
#'   "What do you see in these images?",
#'   content_image_url("https://www.r-project.org/Rlogo.png"),
#'   content_image_file(system.file("httr2.png", package = "ellmer"))
#' )
#'
#' plot(waiting ~ eruptions, data = faithful)
#' chat <- chat_openai()
#' chat$chat(
#'   "Describe this plot in one paragraph, as suitable for inclusion in
#'    alt-text. You should briefly describe the plot type, the axes, and
#'    2-5 major visual patterns.",
#'    content_image_plot()
#' )
#' }
content_image_url <- function(url, detail = c("auto", "low", "high")) {
  detail <- arg_match(detail)

  if (grepl("^data:", url)) {
    parsed <- parse_data_url(url)
    ContentImageInline(parsed$content_type, parsed$base64)
  } else {
    ContentImageRemote(url = url, detail = detail)
  }
}

#' @rdname content_image_url
#' @param path The path to the image file to include in the chat input. Valid
#'   file extensions are `.png`, `.jpeg`, `.jpg`, `.webp`, and (non-animated)
#'   `.gif`.
#' @param content_type The content type of the image (e.g. `image/png`). If
#'   `"auto"`, the content type is inferred from the file extension.
#' @param resize If `"low"`, resize images to fit within 512x512. If `"high"`,
#'   resize to fit within 2000x768 or 768x2000. (See the [OpenAI
#'   docs](https://platform.openai.com/docs/guides/images/image-input-requirements)
#'   for more on why these specific sizes are used.) If `"none"`, do not resize.
#'
#'   You can also pass a custom string to resize the image to a specific size,
#'   e.g. `"200x200"` to resize to 200x200 pixels while preserving aspect ratio.
#'   Append `>` to resize only if the image is larger than the specified size,
#'   and `!` to ignore aspect ratio (e.g. `"300x200>!"`).
#'
#'   All values other than `none` require the `magick` package.
#' @export
content_image_file <- function(path, content_type = "auto", resize = "low") {
  # TODO: Allow vector input?
  check_string(path, allow_empty = FALSE)
  check_string(content_type, allow_empty = FALSE)
  check_string(resize, allow_empty = FALSE)

  if (!file.exists(path) || dir.exists(path)) {
    cli::cli_abort("{path} must be an existing file.")
  }

  if (content_type == "auto") {
    # OpenAI supports .png, .jpeg, .jpg, .webp, .gif
    # https://platform.openai.com/docs/guides/images/image-input-requirements
    ext <- tolower(tools::file_ext(path))
    content_type <- switch(
      ext,
      png = "image/png",
      jpeg = "image/jpeg",
      jpg = "image/jpeg",
      webp = "image/webp",
      gif = "image/gif",
      cli::cli_abort("Unsupported image file extension: {ext}.")
    )
  }

  # Implement resizing logic
  if (resize == "none") {
    base64 <- base64_enc(path = path)
  } else {
    check_installed("magick", "to resize images")

    img <- magick::image_read(path, strip = TRUE)

    if (resize == "low") {
      img <- magick::image_resize(img, "512x512>")
    } else if (resize == "high") {
      # Get current image dimensions
      dims <- magick::image_info(img)
      width <- dims$width
      height <- dims$height

      if (width > height) {
        img <- magick::image_resize(img, "2000x768>")
      } else {
        img <- magick::image_resize(img, "768x2000>")
      }
    } else {
      img <- magick::image_resize(img, resize)
    }
    buf <- magick::image_write(img, format = magick::image_info(img)$format)
    base64 <- base64_enc(raw = buf)
  }

  ContentImageInline(content_type, base64)
}

#' @rdname content_image_url
#' @export
#' @param width,height Width and height in pixels.
content_image_plot <- function(width = 768, height = 768) {
  check_number_whole(width, min = 1)
  check_number_whole(height, min = 1)

  plot <- grDevices::recordPlot()

  if (is.null(plot[[1]])) {
    cli::cli_abort(c(
      "Can't record plot because display list is inhibited.",
      i = "Turn it on with {.code dev.control('enable')}."
    ))
  }

  old <- grDevices::dev.cur()

  path <- tempfile("ellmer-plot-", fileext = ".png")
  defer(unlink(path))

  grDevices::png(path, width = width, height = height)
  grDevices::replayPlot(plot)
  grDevices::dev.off()

  grDevices::dev.set(old)

  content_image_file(path, "image/png", resize = "none")
}


parse_data_url <- function(url, error_call = caller_env()) {
  # https://developer.mozilla.org/en-US/docs/Web/URI/Schemes/data
  parts <- strsplit(sub("^data:", "", url), ";")[[1]]
  if (length(parts) != 2 || !grepl("^base64,", parts[[2]])) {
    cli::cli_abort("{.arg url} is not a valid data url.", call = error_call)
  }

  list(
    content_type = parts[[1]],
    base64 = sub("^base64,", "", parts[[2]])
  )
}

test_that("can create image from url", {
  obj <- content_image_url("https://www.r-project.org/Rlogo.png")
  expect_s3_class(obj, "ellmer::ContentImageRemote")
})

test_that("can create inline image from data url", {
  obj <- content_image_url("data:image/png;base64,abcd")
  expect_s3_class(obj, "ellmer::ContentImageInline")
  expect_equal(obj@type, "image/png")
  expect_equal(obj@data, "abcd")
})

test_that("errors with invalid data urls", {
  expect_snapshot(content_image_url("data:base64,abcd"), error = TRUE)

  expect_error(content_image_url("data:"))
  expect_error(content_image_url("data:;;;"))
  expect_error(content_image_url("data:image/png;abc"))
})

test_that("can create image from path", {
  skip_if_not_installed("magick")

  path <- system.file("httr2.png", package = "ellmer")
  obj <- content_image_file(path)
  expect_s3_class(obj, "ellmer::ContentImageInline")
})

test_that("can create image from plot", {
  withr::local_pdf(NULL)
  dev.control("enable")
  plot(1:10)

  obj <- content_image_plot()
  expect_s3_class(obj, "ellmer::ContentImageInline")
  expect_equal(obj@type, "image/png")
})

test_that("image resizing", {
  skip_if_not_installed("magick")

  img_file <- system.file("httr2.png", package = "ellmer")

  expect_snapshot(error = TRUE, {
    content_image_file("DOESNTEXIST")
    content_image_file(test_path("test-content.R"))
    content_image_file(img_file, resize = TRUE)
    content_image_file(img_file, resize = "blah")
  })

  expect_no_error(content_image_file(img_file))
  expect_no_error(content_image_file(img_file, resize = "low"))
  expect_no_error(content_image_file(img_file, resize = "high"))
  expect_no_error(content_image_file(img_file, resize = "none"))
  expect_no_error(content_image_file(img_file, resize = "100x100"))
  expect_no_error(content_image_file(img_file, resize = "100x100>!"))
})

test_that("useful errors if no display", {
  # file based devices have display list inhibited
  withr::local_pdf(NULL)
  expect_snapshot(content_image_plot(), error = TRUE)
})

# Utility functions for using coro together with R6. This is ordinarily not
# possible because R6 replaces a method's containing environment with its own,
# and coro generators need their containing environments.
#
# We solve this by storing pristine coro generator functions in a special
# environment, and create sacrificial wrapper functions to serve as methods;
# these wrappers do nothing but fetch the real generators and invoke them.
#
# TODO: Check with @hadley if it's safe to do this in a package, given that the
# R6 classes that depend on this will be instantiated at package build time; so
# the coro generator functions will be burned into the package .Rds file.

# So `R CMD check` doesn't get confused about these variables being used from
# methods
utils::globalVariables(c("self", "private", "generator_env", "exits"))

generators <- new_environment()
generators$cur_id <- 1L
new_id <- function() {
  generators$cur_id <- generators$cur_id + 1L
  as.character(generators$cur_id)
}

# Decorator for anonymous functions; the return value is intended to be used as
# an R6 method. Unlike regular R6 methods, the decorated function must have
# `self` as the first argument (which will be automatically passed in by the
# decorator). If necessary we can also provide access to `private` in the same
# way.
generator_method <- function(fn) {
  lambda_expr <- enexpr(fn)
  check_args(fn)
  deferred_method_transform(lambda_expr, coro::generator, parent.frame())
}

# Same as generator_method, but for async logic
async_generator_method <- function(fn) {
  lambda_expr <- enexpr(fn)
  check_args(fn)
  deferred_method_transform(lambda_expr, coro::async_generator, parent.frame())
}

async_method <- function(fn) {
  lambda_expr <- enexpr(fn)
  check_args(fn)
  deferred_method_transform(lambda_expr, coro::async, parent.frame())
}
# Takes a quoted function expression and a transformer function, and returns a
# function that will _lazily_ transform the lambda function using `transformer`
# upon first call. This is necessary because the transformation needs to be done
# not at package build time, but after package load time.
#
# Elsewhere in ellmer, we use rlang::on_load to defer the transformation of
# generators until after package load time. We can't do that for R6 methods
# because nesting R6 class definitions inside of rlang::on_load causes roxygen2
# to get confused.
deferred_method_transform <- function(lambda_expr, transformer, eval_env) {
  transformer <- enexpr(transformer)
  force(eval_env)

  unique_id <- new_id()
  env_bind_lazy(
    generators,
    !!unique_id := inject((!!transformer)(!!lambda_expr)),
    eval.env = eval_env
  )

  inject(
    function(...) {
      # Can't simply use `generators` because the lexical environment of this
      # function is about to get wrecked by R6
      (!!generators)[[!!unique_id]](self, private, ...)
    }
  )
}

check_args <- function(fn) {
  arg_names <- names(formals(fn))
  if (length(arg_names) < 2) {
    cli::cli_abort(
      "Function must have at least two arguments.",
      .internal = TRUE
    )
  } else if (arg_names[[1]] != "self") {
    cli::cli_abort("First argument must be {.arg self}.", .internal = TRUE)
  } else if (arg_names[[2]] != "private") {
    cli::cli_abort("Second argument must be {.arg private}.", .internal = TRUE)
  }
}

#' @include utils-S7.R
#' @include types.R
#' @include ellmer-package.R
NULL

#' Define a tool
#'
#' @description
#' Annotate a function for use in tool calls, by providing a name, description,
#' and type definition for the arguments.
#'
#' Learn more in `vignette("tool-calling")`.
#'
#' # ellmer 0.3.0
#'
#' In ellmer 0.3.0, the definition of the `tool()` function changed quite
#' a bit. To make it easier to update old versions, you can use an LLM with
#' the following system prompt
#'
#' ````
#' Help the user convert an ellmer 0.2.0 and earlier tool definition into a
#' ellmer 0.3.0 tool definition. Here's what changed:
#'
#' * All arguments, apart from the first, should be named, and the argument
#'   names no longer use `.` prefixes. The argument order should be function,
#'   name (as a string), description, then arguments, then anything
#'
#' * Previously `arguments` was passed as `...`, so all type specifications
#'   should now be moved into a named list and passed to the `arguments`
#'   argument. It can be omitted if the function has no arguments.
#'
#' ```R
#' # old
#' tool(
#'   add,
#'   "Add two numbers together"
#'   x = type_number(),
#'   y = type_number()
#' )
#'
#' # new
#' tool(
#'   add,
#'   name = "add",
#'   description = "Add two numbers together",
#'   arguments = list(
#'     x = type_number(),
#'     y = type_number()
#'   )
#' )
#' ```
#'
#' Don't respond; just let the user provide function calls to convert.
#' ````
#'
#' @param fun The function to be invoked when the tool is called. The return
#'   value of the function is sent back to the chatbot.
#'
#'   Expert users can customize the tool result by returning a
#'   [ContentToolResult] object.
#' @param name The name of the function. This can be omitted if `fun` is an
#'   existing function (i.e. not defined inline).
#' @param description A detailed description of what the function does.
#'   Generally, the more information that you can provide here, the better.
#' @param arguments A named list that defines the arguments accepted by the
#'   function. Each element should be created by a [`type_*()`][type_boolean]
#'   function. Use [type_ignore()] if you don't want the LLM to provide that
#'   argument (e.g., because the R function has a suitable default value).
#' @param annotations Additional properties that describe the tool and its
#'   behavior. Usually created by [tool_annotations()], where you can find a
#'   description of the annotation properties recommended by the [Model Context
#'   Protocol](https://modelcontextprotocol.io/introduction).
#' @param convert Should JSON inputs be automatically convert to their
#'   R data type equivalents? Defaults to `TRUE`.
#' @param ... `r lifecycle::badge("deprecated")` Use `arguments` instead.
#' @param .name,.description,.convert,.annotations
#'   `r lifecycle::badge("deprecated")` Please switch to the non-prefixed
#'   equivalents.
#' @return An S7 `ToolDef` object.
#' @examples
#' \dontshow{ellmer:::vcr_example_start("tool")}
#' \dontshow{set.seed(1014)}
#' # First define the metadata that the model uses to figure out when to
#' # call the tool
#' tool_rnorm <- tool(
#'   rnorm,
#'   description = "Draw numbers from a random normal distribution",
#'   arguments = list(
#'     n = type_integer("The number of observations. Must be a positive integer."),
#'     mean = type_number("The mean value of the distribution."),
#'     sd = type_number("The standard deviation of the distribution. Must be a non-negative number.")
#'   )
#' )
#' tool_rnorm(n = 5, mean = 0, sd = 1)
#'
#' chat <- chat_openai()
#' # Then register it
#' chat$register_tool(tool_rnorm)
#'
#' # Then ask a question that needs it.
#' chat$chat("Give me five numbers from a random normal distribution.")
#'
#' # Look at the chat history to see how tool calling works:
#' chat
#' # Assistant sends a tool request which is evaluated locally and
#' # results are sent back in a tool result.
#'
#' \dontshow{ellmer:::vcr_example_end()}
#' @family tool calling helpers
#' @aliases ToolDef
#' @export
tool <- function(
  fun,
  description,
  ...,
  arguments = list(),
  name = NULL,
  convert = TRUE,
  annotations = list(),
  .name = deprecated(),
  .description = deprecated(),
  .convert = deprecated(),
  .annotations = deprecated()
) {
  if (!missing(...)) {
    lifecycle::deprecate_warn("0.3.0", "tool(...)", "tool(arguments)")
    arguments <- list(...)
  }
  if (lifecycle::is_present(.name)) {
    lifecycle::deprecate_warn("0.3.0", "tool(.name)", "tool(name)")
    name <- .name
  }
  if (lifecycle::is_present(.description)) {
    lifecycle::deprecate_warn(
      "0.3.0",
      "tool(.description)",
      "tool(description)"
    )
    description <- .description
  }
  if (lifecycle::is_present(.convert)) {
    lifecycle::deprecate_warn("0.3.0", "tool(.convert)", "tool(convert)")
    convert <- .convert
  }
  if (lifecycle::is_present(.annotations)) {
    lifecycle::deprecate_warn(
      "0.3.0",
      "tool(.annotations)",
      "tool(annotations)"
    )
    annotations <- .annotations
  }

  fun_expr <- substitute(fun)
  check_function(fun)
  check_string(description)
  check_string(name, allow_null = TRUE)
  check_bool(convert)

  if (is.null(name)) {
    if (is.name(fun_expr)) {
      name <- as.character(fun_expr)
    } else {
      name <- unique_tool_name()
    }
  }
  if (!grepl("^[a-zA-Z0-9_-]+$", name)) {
    cli::cli_abort("{.arg name} must contain only letters, numbers, - and _.")
  }

  check_arguments(arguments, formals(fun))

  # Filter out TypeIgnore arguments - they should not be sent to the LLM
  llm_arguments <- arguments[!map_lgl(arguments, S7_inherits, TypeIgnore)]

  ToolDef(
    fun,
    name = name,
    description = description,
    arguments = TypeObject(properties = llm_arguments),
    convert = convert,
    annotations = annotations
  )
}

ToolDef <- new_class(
  "ToolDef",
  parent = class_function,
  properties = list(
    name = prop_string(),
    description = prop_string(),
    arguments = TypeObject,
    convert = prop_bool(TRUE),
    annotations = class_list
  )
)

method(print, ToolDef) <- function(x, ...) {
  fake_call <- call2(x@name, !!!syms(names(x@arguments@properties)))

  cat_line("# <ellmer::ToolDef> ", deparse1(fake_call))
  cat_line("# @name: ", x@name)
  cat_line("# @description: ", x@description)
  cat_line("# @convert: ", x@convert)
  cat_line("#")
  print(S7_data(x))

  invisible(x)
}

check_arguments <- function(arguments, formals, call = caller_env()) {
  if (!is.list(arguments) || !(length(arguments) == 0 || is_named(arguments))) {
    stop_input_type(arguments, "a named list", call = call)
  }

  extra_args <- setdiff(names(arguments), names(formals))
  missing_args <- setdiff(names(formals), names(arguments))
  if (length(extra_args) > 0 || length(missing_args) > 0) {
    cli::cli_abort(
      c(
        "Names of {.arg arguments} must match formals of {.arg fun}",
        "*" = if (length(extra_args) > 0) {
          "Extra type definitions: {.val {extra_args}}"
        },
        "*" = if (length(missing_args) > 0) {
          "Missing type definitions: {.val {missing_args}}"
        }
      ),
      call = call
    )
  }

  for (nm in names(arguments)) {
    arg <- arguments[[nm]]
    if (!S7_inherits(arg, Type)) {
      stop_input_type(
        arg,
        "a <Type>",
        arg = paste0("arguments$", nm),
        call = call
      )
    }
  }

  invisible()
}

check_tool <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (!S7_inherits(x, ToolDef) && !S7_inherits(x, ToolBuiltIn)) {
    stop_input_type(x, "a <ToolDef>", arg = arg, call = call)
  }
}

check_tools <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (!is_list(x)) {
    stop_input_type(x, "a list", arg = arg, call = call)
  }
  for (i in seq_along(x)) {
    check_tool(x[[i]], arg = paste0(arg, "[[", i, "]]"), call = call)
  }
}

#' Tool annotations
#'
#' @description
#' Tool annotations are additional properties that, when passed to the
#' `.annotations` argument of [tool()], provide additional information about the
#' tool and its behavior. This information can be used for display to users, for
#' example in a Shiny app or another user interface.
#'
#' The annotations in `tool_annotations()` are drawn from the [Model Context
#' Protocol](https://modelcontextprotocol.io/introduction) and are considered
#' *hints*. Tool authors should use these annotations to communicate tool
#' properties, but users should note that these annotations are not guaranteed.
#'
#' @examples
#' # See ?tool() for a full example using this function.
#' # We're creating a tool around R's `rnorm()` function to allow the chatbot to
#' # generate random numbers from a normal distribution.
#' tool_rnorm <- tool(
#'   rnorm,
#'   # Describe the tool function to the LLM
#'   .description = "Drawn numbers from a random normal distribution",
#'   # Describe the parameters used by the tool function
#'   n = type_integer("The number of observations. Must be a positive integer."),
#'   mean = type_number("The mean value of the distribution."),
#'   sd = type_number("The standard deviation of the distribution. Must be a non-negative number."),
#'   # Tool annotations optionally provide additional context to the LLM
#'   .annotations = tool_annotations(
#'     title = "Draw Random Normal Numbers",
#'     read_only_hint = TRUE, # the tool does not modify any state
#'     open_world_hint = FALSE # the tool does not interact with the outside world
#'   )
#' )
#'
#' @param title A human-readable title for the tool.
#' @param read_only_hint If `TRUE`, the tool does not modify its environment.
#' @param open_world_hint If `TRUE`, the tool may interact with an "open world"
#'   of external entities. If `FALSE`, the tool's domain of interaction is
#'   closed. For example, the world of a web search tool is open, but the world
#'   of a memory tool is not.
#' @param idempotent_hint If `TRUE`, calling the tool repeatedly with the same
#'   arguments will have no additional effect on its environment. (Only
#'   meaningful when `read_only_hint` is `FALSE`.)
#' @param destructive_hint If `TRUE`, the tool may perform destructive updates
#'   to its environment, otherwise it only performs additive updates. (Only
#'   meaningful when `read_only_hint` is `FALSE`.)
#' @param ... Additional named parameters to include in the tool annotations.
#'
#' @return A list of tool annotations.
#'
#' @family tool calling helpers
#' @export
tool_annotations <- function(
  title = NULL,
  read_only_hint = NULL,
  open_world_hint = NULL,
  idempotent_hint = NULL,
  destructive_hint = NULL,
  ...
) {
  # Snake-cased names and descriptions from the MCP 2025-03-26 Schema
  # https://github.com/modelcontextprotocol/specification/blob/72516795/schema/2025-03-26/schema.json#L2050-L2074
  check_character(title, allow_null = TRUE)
  check_bool(read_only_hint, allow_null = TRUE)
  check_bool(open_world_hint, allow_null = TRUE)
  check_bool(idempotent_hint, allow_null = TRUE)
  check_bool(destructive_hint, allow_null = TRUE)

  compact(list2(
    title = title,
    read_only_hint = read_only_hint,
    open_world_hint = open_world_hint,
    idempotent_hint = idempotent_hint,
    destructive_hint = destructive_hint,
    ...
  ))
}

#' Reject a tool call
#'
#' @description
#' Throws an error to reject a tool call. `tool_reject()` can be used within the
#' tool function to indicate that the tool call should not be processed.
#' `tool_reject()` can also be called in an `Chat$on_tool_request()` callback.
#'  When used in the callback, the tool call is rejected before the tool
#' function is invoked.
#'
#' Here's an example where `utils::askYesNo()` is used to ask the user for
#' permission before accessing their current working directory. This happens
#' directly in the tool function and is appropriate when you write the tool
#' definition and know exactly how it will be called.
#'
#' ```r
#' chat <- chat_openai(model = "gpt-4.1-nano")
#'
#' list_files <- function() {
#'   allow_read <- utils::askYesNo(
#'     "Would you like to allow access to your current directory?"
#'   )
#'   if (isTRUE(allow_read)) {
#'     dir(pattern = "[.](r|R|csv)$")
#'   } else {
#'     tool_reject()
#'   }
#' }
#'
#' chat$register_tool(tool(
#'   list_files,
#'   "List files in the user's current directory"
#' ))
#'
#' chat$chat("What files are available in my current directory?")
#' #> [tool call] list_files()
#' #> Would you like to allow access to your current directory? (Yes/no/cancel) no
#' #> #> Error: Tool call rejected. The user has chosen to disallow the tool #' call.
#' #> It seems I am unable to access the files in your current directory right now.
#' #> If you can tell me what specific files you're looking for or if you can #' provide
#' #> the list, I can assist you further.
#'
#' chat$chat("Try again.")
#' #> [tool call] list_files()
#' #> Would you like to allow access to your current directory? (Yes/no/cancel) yes
#' #> #> app.R
#' #> #> data.csv
#' #> The files available in your current directory are "app.R" and "data.csv".
#' ```
#'
#' You can achieve a similar experience with tools written by others by using a
#' `tool_request` callback. In the next example, imagine the tool is provided by
#' a third-party package. This example implements a simple menu to ask the user
#' for consent before running *any*  tool.
#'
#' ```r
#' packaged_list_files_tool <- tool(
#'   function() dir(pattern = "[.](r|R|csv)$"),
#'   "List files in the user's current directory"
#' )
#'
#' chat <- chat_openai(model = "gpt-4.1-nano")
#' chat$register_tool(packaged_list_files_tool)
#'
#' always_allowed <- c()
#'
#' # ContentToolRequest
#' chat$on_tool_request(function(request) {
#'   if (request@name %in% always_allowed) return()
#'
#'   answer <- utils::menu(
#'     title = sprintf("Allow tool `%s()` to run?", request@name),
#'     choices = c("Always", "Once", "No"),
#'     graphics = FALSE
#'   )
#'
#'   if (answer == 1) {
#'     always_allowed <<- append(always_allowed, request@name)
#'   } else if (answer %in% c(0, 3)) {
#'     tool_reject()
#'   }
#' })
#'
#' # Try choosing different answers to the menu each time
#' chat$chat("What files are available in my current directory?")
#' chat$chat("How about now?")
#' chat$chat("And again now?")
#' ```
#'
#' @param reason A character string describing the reason for rejecting the
#'   tool call.
#' @return Throws an error of class `ellmer_tool_reject` with the provided
#'   reason.
#'
#' @family tool calling helpers
#' @export
tool_reject <- function(
  reason = "The user has chosen to disallow the tool call."
) {
  check_string(reason)

  abort(
    paste("Tool call rejected.", reason),
    class = "ellmer_tool_reject"
  )
}


unique_tool_name <- function() {
  the$cur_tool_id <- (the$cur_tool_id %||% 0) + 1
  sprintf("tool_%03d", the$cur_tool_id)
}

test_that("can call tools directory", {
  f <- tool(function() 1, description = "a simple function")
  expect_equal(f(), 1)
})

test_that("tools have a print method", {
  fun <- function(x = 1, y = 2) {
    x + y
  }
  environment(fun) <- globalenv()

  f <- tool(
    fun,
    name = "my_fun",
    arguments = list(x = type_string(), y = type_number()),
    description = "a simple function"
  )
  expect_snapshot(f)
})

test_that("tool can get name", {
  f <- function() {}
  td <- tool(f, description = "")
  expect_equal(td@name, "f")

  td <- tool(
    function() {},
    description = ""
  )
  expect_match(td@name, "^tool_")
})

test_that("json_schema_parameters generates correct paramters if no arguments", {
  expect_equal(
    as_json(test_provider(), type_object()),
    list(
      type = "object",
      description = "",
      properties = set_names(list()),
      required = list(),
      additionalProperties = FALSE
    )
  )
})

test_that("old arguments are deprecated", {
  expect_snapshot({
    f <- tool(
      function(x) x * 2,
      .name = "double",
      .description = "double the input",
      x = type_number(),
      .convert = FALSE,
      .annotations = tool_annotations(title = "My Tool")
    )
  })

  expect_equal(f@name, "double")
  expect_equal(f@convert, FALSE)
  expect_equal(f@description, "double the input")
  expect_equal(f@arguments, TypeObject(properties = list(x = type_number())))
  expect_equal(f@annotations, tool_annotations(title = "My Tool"))
})

test_that("checks its arguments", {
  expect_snapshot(error = TRUE, {
    tool(1)
    tool(identity, 1)
    tool(identity, "", name = 1)
    tool(identity, "", name = "...")
    tool(identity, "", arguments = 1)
    tool(identity, "", convert = 1)
  })
})

test_that("arguments must match function formals", {
  fun <- function(x, y) {}

  expect_snapshot(error = TRUE, {
    tool(fun, "", arguments = list(z = type_number()))
    tool(fun, "", arguments = list(x = type_number(), y = 1))
  })
})

test_that("type_ignore() filters out arguments from LLM", {
  td <- tool(
    \(x, y = 10) x + y,
    description = "Add two numbers",
    arguments = list(
      x = type_number("First number"),
      y = type_ignore()
    )
  )

  # The tool definition should not include y in its arguments
  expect_equal(names(td@arguments@properties), "x")
  # But the function should still work with the default
  expect_equal(td(x = 5), 15)
})

test_that("type_ignore() works with all arguments ignored", {
  td <- tool(
    \(x = 1, y = 2) x + y,
    description = "Add with defaults",
    arguments = list(
      x = type_ignore(),
      y = type_ignore()
    )
  )

  # No arguments should be sent to LLM
  expect_equal(length(td@arguments@properties), 0)
  # Function should still work
  expect_equal(td(), 3)
})

test_that("can check tool/tools", {
  x <- list(1)
  expect_snapshot(error = TRUE, {
    check_tool(1)
    check_tools(1)
    check_tools(x)
  })
})

# tool_annotations() -------------------------------------------------------

test_that("tool_annotations(): NULL values are stripped", {
  expect_equal(tool_annotations(), set_names(list()))
})

test_that("tool_annotations(): checks its inputs", {
  expect_snapshot(error = TRUE, {
    tool_annotations(title = list("Something unexpected"))
    tool_annotations(read_only_hint = "yes")
    tool_annotations(open_world_hint = "yes")
    tool_annotations(idempotent_hint = "no")
    tool_annotations(destructive_hint = "no")
  })
})

test_that("tool_annotations(): allows additional properties", {
  expect_equal(
    tool_annotations(description = "foo"),
    list(description = "foo")
  )
})

test_that("tool() allows annotations", {
  annotations <- tool_annotations(title = "My Tool", read_only_hint = TRUE)
  tool_def <- tool(function() {}, description = "", annotations = annotations)
  expect_equal(tool_def@annotations, annotations)
})

test_that("ToolBuiltIn has description and annotations", {
  tool <- ToolBuiltIn(
    name = "test_tool",
    description = "A test tool.",
    annotations = tool_annotations(title = "Test Tool", read_only_hint = TRUE),
    json = list(type = "test")
  )

  expect_equal(tool@name, "test_tool")
  expect_equal(tool@description, "A test tool.")
  expect_equal(tool@annotations$title, "Test Tool")
  expect_true(tool@annotations$read_only_hint)
  expect_equal(tool@json, list(type = "test"))
})

test_that("ToolBuiltIn defaults for description and annotations", {
  tool <- ToolBuiltIn(name = "minimal", json = list())

  expect_equal(tool@description, "")
  expect_equal(tool@annotations, list())
})

test_that("built-in tools", {
  get_built_in_tools <- function() {
    exports <- getNamespaceExports("ellmer")
    tool_fns <- exports[grepl("_tool_", exports)]

    tools <- list()
    for (fn_name in tool_fns) {
      fn <- getExportedValue("ellmer", fn_name)
      result <- tryCatch(fn(), error = function(e) NULL)
      if (!is.null(result) && S7_inherits(result, ToolBuiltIn)) {
        tools[[fn_name]] <- result
      }
    }
    tools
  }

  expect_gte(length(get_built_in_tools()), 1)

  built_in_tools <- get_built_in_tools()

  for (fn_name in names(built_in_tools)) {
    test_that(paste0(fn_name, "() sets description and annotations"), {
      tool <- built_in_tools[[fn_name]]
      expect_match(tool@description, ".")
      expect_match(tool@annotations$title, ".")
    })
  }
})

#' @include utils-S7.R
NULL

#' A user, assistant, or system turn
#'
#' @description
#' Every conversation with a chatbot consists of pairs of user and assistant
#' turns, corresponding to an HTTP request and response. These turns are
#' represented by the `Turn` object, which contains a list of [Content]s representing
#' the individual messages within the turn. These might be text, images, tool
#' requests (assistant only), or tool responses (user only).
#'
#' `UserTurn`, `AssistantTurn`, and `SystemTurn` are specialized subclasses
#' of `Turn` for different types of conversation turns. `AssistantTurn` includes
#' additional metadata about the API response.
#'
#' Note that a call to `$chat()` and related functions may result in multiple
#' user-assistant turn cycles. For example, if you have registered tools,
#' ellmer will automatically handle the tool calling loop, which may result in
#' any number of additional cycles. Learn more about tool calling in
#' `vignette("tool-calling")`.
#'
#' @param contents A list of [Content] objects.
#' @param role `r lifecycle::badge("deprecated")`
#'   For system, user and assistant turns, use `SystemTurn()`, `UserTurn()`, and
#'   `AssistantTurn()`, respectively.
#' @export
#' @return An S7 `Turn` object
#' @examples
#' UserTurn(list(ContentText("Hello, world!")))
Turn <- new_class(
  "Turn",
  properties = list(
    contents = prop_list_of(Content),
    text = new_property(
      class = class_character,
      getter = function(self) contents_text(self)
    ),
    role = new_property(
      class = class_character,
      getter = function(self) "unknown"
    )
  ),
  constructor = function(role = NULL, contents = list(), tokens = NULL) {
    if (is.character(contents)) {
      contents <- list(ContentText(paste0(contents, collapse = "\n")))
    }

    if (!is.null(role)) {
      # Quick fallback to allow 0.4.0 release. Remove in future
      # Can then also remove custom constructors for subclasses.
      # We should warn here, but this would cause chattr to fail tests.
      # https://github.com/tidyverse/ellmer/issues/864
      role <- arg_match(role, c("user", "assistant", "system"))

      return(switch(
        role,
        user = UserTurn(contents = contents),
        assistant = AssistantTurn(
          contents = contents,
          tokens = tokens %||% c(NA_real_, NA_real_, NA_real_)
        ),
        system = SystemTurn(contents = contents),
        cli::cli_abort("Unsupported role {.str {role}}.")
      ))
    }

    new_object(S7_object(), contents = contents)
  }
)

#' @rdname Turn
#' @export
UserTurn <- new_class(
  "UserTurn",
  parent = Turn,
  properties = list(
    role = new_property(
      class = class_character,
      getter = function(self) "user"
    )
  ),
  constructor = function(contents = list()) {
    if (is.character(contents)) {
      contents <- list(ContentText(paste0(contents, collapse = "\n")))
    }

    new_object(Turn(), contents = contents)
  }
)

#' @rdname Turn
#' @export
SystemTurn <- new_class(
  "SystemTurn",
  parent = Turn,
  properties = list(
    role = new_property(
      class = class_character,
      getter = function(self) "system"
    )
  ),
  constructor = function(contents = list()) {
    if (is.character(contents)) {
      contents <- list(ContentText(paste0(contents, collapse = "\n")))
    }

    new_object(Turn(), contents = contents)
  }
)

#' @param json The serialized JSON corresponding to the underlying data of
#'   the turns. This is useful if there's information returned by the provider
#'   that ellmer doesn't otherwise expose.
#' @param tokens A numeric vector of length 3 representing the number of
#'   input tokens (uncached), output tokens, and input tokens (cached)
#'   used in this turn.
#' @param cost The cost of the turn in dollars.
#' @param duration The duration of the request in seconds.
#' @export
#' @rdname Turn
#' @return An S7 `AssistantTurn` object
AssistantTurn <- new_class(
  "AssistantTurn",
  parent = Turn,
  properties = list(
    json = class_list,
    tokens = new_property(
      class_numeric,
      default = c(NA_real_, NA_real_, NA_real_),
      validator = function(value) {
        if (length(value) != 3) {
          "must be length three"
        }
      }
    ),
    cost = prop_number_decimal(NA_real_, allow_na = TRUE),
    duration = prop_number_decimal(NA_real_, allow_na = TRUE),
    role = new_property(
      class = class_character,
      getter = function(self) "assistant"
    )
  ),
  constructor = function(
    contents = list(),
    json = list(),
    tokens = c(NA_real_, NA_real_, NA_real_),
    cost = NA_real_,
    duration = NA_real_
  ) {
    if (is.character(contents)) {
      contents <- list(ContentText(paste0(contents, collapse = "\n")))
    }

    new_object(
      Turn(),
      contents = contents,
      json = json,
      tokens = tokens,
      cost = cost,
      duration = duration
    )
  }
)
#' @param reason A character string describing why the turn was interrupted.
#'   Defaults to `"interrupted"`.
#' @export
#' @rdname Turn
#' @return An S7 `AssistantPartialTurn` object
AssistantPartialTurn <- new_class(
  "AssistantPartialTurn",
  parent = AssistantTurn,
  properties = list(
    reason = prop_string("interrupted")
  )
)

method(format, Turn) <- function(x, ...) {
  contents <- map_chr(x@contents, format, ...)
  paste0(contents, "\n", collapse = "")
}
method(contents_text, Turn) <- function(content) {
  paste0(unlist(lapply(content@contents, contents_text)), collapse = "")
}
method(contents_html, Turn) <- function(content) {
  paste0(unlist(lapply(content@contents, contents_html)), collapse = "\n")
}
method(contents_markdown, Turn) <- function(content) {
  paste0(unlist(lapply(content@contents, contents_markdown)), collapse = "\n\n")
}

method(print, Turn) <- function(x, ...) {
  cat(paste_c("<Turn: ", color_role(x@role), ">\n"))
  cat(format(x))
  invisible(x)
}

user_turn <- function(..., .call = caller_env(), .check_empty = TRUE) {
  as_user_turn(
    list2(...),
    call = .call,
    arg = "...",
    check_empty = .check_empty
  )
}

as_user_turn <- function(
  contents,
  check_empty = TRUE,
  call = caller_env(),
  arg = "..."
) {
  if (check_empty && length(contents) == 0) {
    cli::cli_abort("{.arg {arg}} must contain at least one input.", call = call)
  }
  if (is_named(contents)) {
    cli::cli_abort("{.arg {arg}} must be unnamed.", call = call)
  }
  if (S7_inherits(contents, Content)) {
    return(UserTurn(list(contents)))
  }

  contents <- lapply(contents, as_content, error_call = call, error_arg = arg)
  UserTurn(contents)
}

as_user_turns <- function(
  prompts,
  call = caller_env(),
  arg = caller_arg(prompts)
) {
  if (!is.list(prompts) && !is_prompt(prompts)) {
    stop_input_type(prompts, "a list or prompt", call = call, arg = arg)
  }
  turns <- map(seq_along(prompts), function(i) {
    this_arg <- paste0(arg, "[[", i, "]]")
    as_user_turn(prompts[[i]], call = call, arg = this_arg)
  })

  turns
}

is_system_turn <- function(x) {
  S7_inherits(x, SystemTurn)
}
# needed for vitals
is_system_prompt <- is_system_turn

is_user_turn <- function(x) {
  S7_inherits(x, UserTurn)
}

is_assistant_turn <- function(x) {
  S7_inherits(x, AssistantTurn)
}

is_partial_turn <- function(x) {
  S7_inherits(x, AssistantPartialTurn)
}

check_turn <- function(x, call = caller_env(), arg = caller_arg(x)) {
  if (!S7_inherits(x, Turn)) {
    stop_input_type(x, "a <Turn>", call = call, arg = arg)
  }
}

normalize_turns <- function(
  turns = NULL,
  system_prompt = NULL,
  overwrite = FALSE,
  error_call = caller_env()
) {
  check_string(system_prompt, allow_null = TRUE)

  if (!is.null(turns)) {
    if (!is.list(turns) || is_named(turns)) {
      stop_input_type(
        turns,
        "an unnamed list",
        allow_null = TRUE,
        call = error_call
      )
    }
    correct_class <- map_lgl(turns, S7_inherits, Turn)
    if (!all(correct_class)) {
      cli::cli_abort("Every element of {.arg turns} must be a `turn`.")
    }
  } else {
    turns <- list()
  }

  if (!is.null(system_prompt)) {
    system_turn <- SystemTurn(system_prompt)

    # No turns; start with just the system prompt
    if (length(turns) == 0) {
      turns <- list(system_turn)
    } else if (!is_system_turn(turns[[1]])) {
      turns <- c(list(system_turn), turns)
    } else if (overwrite || identical(turns[[1]], system_turn)) {
      # Duplicate system prompt; don't need to do anything
    } else {
      cli::cli_abort(
        "`system_prompt` and `turns[[1]]` can't contain conflicting system prompts.",
        call = error_call
      )
    }
  }

  turns
}

turn_contents_preview <- function(turn) {
  is_text <- map_lgl(turn@contents, S7_inherits, ContentText)
  is_first_text <- is_text & cumsum(is_text) == 1

  contents <- map2_chr(turn@contents, is_first_text, \(x, is_first_text) {
    if (is_first_text) {
      paste0("Text[", str_trunc(x@text, 40), "]")
    } else {
      sub("^ellmer::Content", "", class(x)[[1]])
    }
  })
  paste(contents, collapse = ", ")
}

test_that("Turn() with `role`", {
  suppressWarnings({
    # Promotes system, user, assistant to subclasses of Turn
    expect_s7_class(Turn(role = "system"), SystemTurn)
    expect_s7_class(Turn(role = "user"), UserTurn)
    expect_s7_class(Turn(role = "assistant"), AssistantTurn)
  })

  # No role: unknown
  expect_equal(Turn()@role, "unknown")

  # Invalid roles are errors
  expect_snapshot(error = TRUE, {
    Turn(role = "invalid")
    Turn(role = 42)
  })
})

test_that("system prompt is applied correctly", {
  sys_prompt <- "foo"
  sys_msg <- SystemTurn(sys_prompt)
  user_msg <- UserTurn("bar")

  expect_equal(normalize_turns(), list())
  expect_equal(normalize_turns(list(user_msg)), list(user_msg))
  expect_equal(normalize_turns(list(sys_msg)), list(sys_msg))

  expect_equal(normalize_turns(list(), sys_prompt), list(sys_msg))
  expect_equal(
    normalize_turns(list(user_msg), sys_prompt),
    list(sys_msg, user_msg)
  )
  expect_equal(
    normalize_turns(list(sys_msg, user_msg), sys_prompt),
    list(sys_msg, user_msg)
  )
})

test_that("normalize_turns throws useful errors", {
  sys_prompt <- "foo"
  sys_msg <- SystemTurn("foo")
  user_msg <- UserTurn("bar")

  expect_snapshot(error = TRUE, {
    normalize_turns(1)
    normalize_turns(list(1))
    normalize_turns(list(sys_msg, user_msg), 1)
    normalize_turns(list(sys_msg, user_msg), "foo2")
  })
})


test_that("as_user_turn gives useful errors", {
  expect_snapshot(error = TRUE, {
    as_user_turn(list())
    as_user_turn(list(x = 1))
    as_user_turn(1)
  })
})

test_that("can opt-out of empty check", {
  out <- as_user_turn(list(), check_empty = FALSE)
  expect_equal(out, UserTurn())
})

test_that("as_user_turns gives useful errors", {
  x <- list(list(1))
  expect_snapshot(error = TRUE, {
    as_user_turns(1)
    as_user_turns(x)
  })
})

test_that("can extract text easily", {
  turn <- AssistantTurn(
    list(
      ContentText("ABC"),
      ContentImage(),
      ContentText("DEF")
    )
  )
  expect_equal(turn@text, "ABCDEF")
})

test_that("turns have a reasonable print method", {
  expect_snapshot(UserTurn("hello"))
})

test_that("as_user_turns can create lists of turns from lists of Content objects", {
  content_turns <- as_user_turns(
    list(
      content_image_url("https://www.r-project.org/Rlogo.png"),
      content_image_url("https://www.r-project.org/Rlogo.png")
    )
  )

  expect_length(content_turns, 2)
  expect_s3_class(content_turns[[1]], "ellmer::Turn")
  expect_s3_class(content_turns[[2]], "ellmer::Turn")
})

# turn_contents_preview()

test_that("ContentText shows first (truncated) text", {
  expect_equal(
    turn_contents_preview(UserTurn("This is short")),
    "Text[This is short]"
  )
  expect_equal(
    turn_contents_preview(UserTurn(
      "This is a very long message that should be truncated"
    )),
    "Text[This is a very long message that shou...]"
  )

  expect_equal(
    turn_contents_preview(UserTurn(
      list(ContentText("This is short"), ContentText("Second"))
    )),
    "Text[This is short], Text"
  )
})

test_that("non-text types just show type", {
  expect_equal(
    turn_contents_preview(UserTurn(list(ContentImageInline("")))),
    "ImageInline"
  )

  expect_equal(
    turn_contents_preview(UserTurn(
      list(ContentImageInline(""), ContentImageRemote(""))
    )),
    "ImageInline, ImageRemote"
  )
})

# Standalone file: do not edit by hand
# Source: https://github.com/r-lib/rlang/blob/HEAD/R/standalone-purrr.R
# Generated by: usethis::use_standalone("r-lib/rlang", "purrr")
# ----------------------------------------------------------------------
#
# ---
# repo: r-lib/rlang
# file: standalone-purrr.R
# last-updated: 2023-02-23
# license: https://unlicense.org
# imports: rlang
# ---
#
# This file provides a minimal shim to provide a purrr-like API on top of
# base R functions. They are not drop-in replacements but allow a similar style
# of programming.
#
# ## Changelog
#
# 2023-02-23:
# * Added `list_c()`
#
# 2022-06-07:
# * `transpose()` is now more consistent with purrr when inner names
#   are not congruent (#1346).
#
# 2021-12-15:
# * `transpose()` now supports empty lists.
#
# 2021-05-21:
# * Fixed "object `x` not found" error in `imap()` (@mgirlich)
#
# 2020-04-14:
# * Removed `pluck*()` functions
# * Removed `*_cpl()` functions
# * Used `as_function()` to allow use of `~`
# * Used `.` prefix for helpers
#
# nocov start

map <- function(.x, .f, ...) {
  .f <- as_function(.f, env = global_env())
  lapply(.x, .f, ...)
}
walk <- function(.x, .f, ...) {
  map(.x, .f, ...)
  invisible(.x)
}

map_lgl <- function(.x, .f, ...) {
  .rlang_purrr_map_mold(.x, .f, logical(1), ...)
}
map_int <- function(.x, .f, ...) {
  .rlang_purrr_map_mold(.x, .f, integer(1), ...)
}
map_dbl <- function(.x, .f, ...) {
  .rlang_purrr_map_mold(.x, .f, double(1), ...)
}
map_chr <- function(.x, .f, ...) {
  .rlang_purrr_map_mold(.x, .f, character(1), ...)
}
.rlang_purrr_map_mold <- function(.x, .f, .mold, ...) {
  .f <- as_function(.f, env = global_env())
  out <- vapply(.x, .f, .mold, ..., USE.NAMES = FALSE)
  names(out) <- names(.x)
  out
}

map2 <- function(.x, .y, .f, ...) {
  .f <- as_function(.f, env = global_env())
  out <- mapply(.f, .x, .y, MoreArgs = list(...), SIMPLIFY = FALSE)
  if (length(out) == length(.x)) {
    set_names(out, names(.x))
  } else {
    set_names(out, NULL)
  }
}
map2_lgl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "logical")
}
map2_int <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "integer")
}
map2_dbl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "double")
}
map2_chr <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "character")
}
imap <- function(.x, .f, ...) {
  map2(.x, names(.x) %||% seq_along(.x), .f, ...)
}

pmap <- function(.l, .f, ...) {
  .f <- as.function(.f)
  args <- .rlang_purrr_args_recycle(.l)
  do.call("mapply", c(
    FUN = list(quote(.f)),
    args, MoreArgs = quote(list(...)),
    SIMPLIFY = FALSE, USE.NAMES = FALSE
  ))
}
.rlang_purrr_args_recycle <- function(args) {
  lengths <- map_int(args, length)
  n <- max(lengths)

  stopifnot(all(lengths == 1L | lengths == n))
  to_recycle <- lengths == 1L
  args[to_recycle] <- map(args[to_recycle], function(x) rep.int(x, n))

  args
}

keep <- function(.x, .f, ...) {
  .x[.rlang_purrr_probe(.x, .f, ...)]
}
discard <- function(.x, .p, ...) {
  sel <- .rlang_purrr_probe(.x, .p, ...)
  .x[is.na(sel) | !sel]
}
map_if <- function(.x, .p, .f, ...) {
  matches <- .rlang_purrr_probe(.x, .p)
  .x[matches] <- map(.x[matches], .f, ...)
  .x
}
.rlang_purrr_probe <- function(.x, .p, ...) {
  if (is_logical(.p)) {
    stopifnot(length(.p) == length(.x))
    .p
  } else {
    .p <- as_function(.p, env = global_env())
    map_lgl(.x, .p, ...)
  }
}

compact <- function(.x) {
  Filter(length, .x)
}

transpose <- function(.l) {
  if (!length(.l)) {
    return(.l)
  }

  inner_names <- names(.l[[1]])

  if (is.null(inner_names)) {
    fields <- seq_along(.l[[1]])
  } else {
    fields <- set_names(inner_names)
    .l <- map(.l, function(x) {
      if (is.null(names(x))) {
        set_names(x, inner_names)
      } else {
        x
      }
    })
  }

  # This way missing fields are subsetted as `NULL` instead of causing
  # an error
  .l <- map(.l, as.list)

  map(fields, function(i) {
    map(.l, .subset2, i)
  })
}

every <- function(.x, .p, ...) {
  .p <- as_function(.p, env = global_env())

  for (i in seq_along(.x)) {
    if (!rlang::is_true(.p(.x[[i]], ...))) return(FALSE)
  }
  TRUE
}
some <- function(.x, .p, ...) {
  .p <- as_function(.p, env = global_env())

  for (i in seq_along(.x)) {
    if (rlang::is_true(.p(.x[[i]], ...))) return(TRUE)
  }
  FALSE
}
negate <- function(.p) {
  .p <- as_function(.p, env = global_env())
  function(...) !.p(...)
}

reduce <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(x, y, ...)
  Reduce(f, .x, init = .init)
}
reduce_right <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(y, x, ...)
  Reduce(f, .x, init = .init, right = TRUE)
}
accumulate <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(x, y, ...)
  Reduce(f, .x, init = .init, accumulate = TRUE)
}
accumulate_right <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(y, x, ...)
  Reduce(f, .x, init = .init, right = TRUE, accumulate = TRUE)
}

detect <- function(.x, .f, ..., .right = FALSE, .p = is_true) {
  .p <- as_function(.p, env = global_env())
  .f <- as_function(.f, env = global_env())

  for (i in .rlang_purrr_index(.x, .right)) {
    if (.p(.f(.x[[i]], ...))) {
      return(.x[[i]])
    }
  }
  NULL
}
detect_index <- function(.x, .f, ..., .right = FALSE, .p = is_true) {
  .p <- as_function(.p, env = global_env())
  .f <- as_function(.f, env = global_env())

  for (i in .rlang_purrr_index(.x, .right)) {
    if (.p(.f(.x[[i]], ...))) {
      return(i)
    }
  }
  0L
}
.rlang_purrr_index <- function(x, right = FALSE) {
  idx <- seq_along(x)
  if (right) {
    idx <- rev(idx)
  }
  idx
}

list_c <- function(x) {
  inject(c(!!!x))
}

# nocov end

# Very few providers support anything other than text results; fortunately we
# can fake them them by unrolling the tool result to a forward pointer to
# other user content items.
#
# ContentToolResult(value = ContentImageInline(...))
# ->
# ContentToolResult("See <content-tool> below")
# ContentText("<content-tool>")
# ContentImageInline(...)
# ContentText("</content-tool>")

turn_contents_expand <- function(turn) {
  if (length(turn@contents) == 0) {
    # Early return to avoid unlist(list()) yielding NULL
    return(turn)
  }

  contents <- map(turn@contents, expand_content_if_needed)
  turn@contents <- unlist(contents, recursive = FALSE)
  # Make sure all tool results come first in the expanded turn
  turn_contents <- turn_split_tool_results(turn)
  turn@contents <- c(turn_contents$tool_results, turn_contents$contents)
  turn
}

expand_content_if_needed <- function(content) {
  if (!is_tool_result(content)) {
    return(list(content))
  }

  request <- content@request
  value <- content@value

  if (S7_inherits(value, Content)) {
    expand_tool_value(request, value)
  } else if (is.list(value)) {
    if (all(map_lgl(value, \(x) S7_inherits(x, Content)))) {
      expand_tool_values(request, value)
    } else {
      list(content)
    }
  } else {
    list(content)
  }
}

expand_tool_value <- function(request, value) {
  open <- sprintf('<tool-content call-id="%s">', request@id)
  list(
    ContentToolResult(
      value = sprintf("See %s below.", open),
      request = request
    ),
    ContentText(open),
    value,
    ContentText("</tool-content>")
  )
}

expand_tool_values <- function(request, values) {
  open <- sprintf('<tool-contents call-id="%s">', request@id)
  result <- ContentToolResult(
    value = sprintf('See %s below.', open),
    request = request
  )

  contents <- map(values, function(value) {
    list(ContentText("<tool-content>"), value, ContentText("</tool-content>"))
  })
  contents <- unlist(contents, recursive = FALSE)

  open <- ContentText(open)
  close <- ContentText("</tool-contents>")
  c(list(result, open), contents, list(close))
}

test_that("can expand all contents in a turn", {
  req <- ContentToolRequest(id = "123", name = "my_tool")
  image <- ContentImageInline("image/png", "abc")

  turn <- UserTurn(list(
    ContentText("abc"),
    ContentToolResult(value = image, request = req),
    ContentToolResult(value = "abc", request = req)
  ))
  expect_length(turn_contents_expand(turn)@contents, 1 + 4 + 1)

  turn <- UserTurn(list(
    ContentText("abc"),
    ContentToolResult(value = image, request = req),
    ContentText("ghi"),
    ContentToolResult(value = image, request = req)
  ))
  expect_length(turn_contents_expand(turn)@contents, 1 + 4 + 1 + 4)

  turn <- UserTurn(list(
    ContentText("abc"),
    ContentToolResult(value = list(image, image), request = req),
    ContentText("ghi")
  ))
  expect_length(
    turn_contents_expand(turn)@contents,
    1 + (1 + 1 + 3 + 3 + 1) + 1
  )
})

test_that("expanding generates useful JSON", {
  req <- ContentToolRequest(id = "123", name = "my_tool")
  image <- ContentImageInline("image/png", "abc")
  provider <- ProviderOpenAI("name", "model", "base_url")

  expanded_simple <- expand_tool_value(req, image)
  expect_snapshot(print_json(as_json(provider, UserTurn(expanded_simple))))

  expanded_list <- expand_tool_values(req, list(image, image))
  expect_snapshot(print_json(as_json(provider, UserTurn(expanded_list))))
})

test_that("can expand tool with single value", {
  req <- ContentToolRequest(id = "123", name = "my_tool")
  image <- ContentImageInline("image/png", "abc")

  expanded <- expand_tool_value(req, image)
  expect_s7_class(expanded[[1]], ContentToolResult)
  expect_equal(expanded[[1]]@request, req)
  expect_s7_class(expanded[[2]], ContentText) # <tool-content>
  expect_equal(expanded[[3]], image)
  expect_s7_class(expanded[[4]], ContentText) # </tool-content>
})

test_that("can expand tool with multiple values", {
  req <- ContentToolRequest(id = "123", name = "my_tool")
  image1 <- ContentImageInline("image/png", "abc")
  image2 <- ContentImageInline("image/png", "def")

  expanded <- expand_tool_values(req, list(image1, image2))
  expect_s7_class(expanded[[1]], ContentToolResult)
  expect_equal(expanded[[1]]@request, req)
  expect_s7_class(expanded[[2]], ContentText) # <tool-contents>
  expect_s7_class(expanded[[3]], ContentText) # <tool-content>
  expect_equal(expanded[[4]], image1)
  expect_s7_class(expanded[[5]], ContentText) # </tool-content>
  expect_s7_class(expanded[[6]], ContentText) # <tool-content>
  expect_equal(expanded[[7]], image2)
  expect_s7_class(expanded[[8]], ContentText) # </tool-content>
  expect_s7_class(expanded[[9]], ContentText) # </tool-contents>
})

# Currently performing chat request is not generic as there appears to
# be sufficiently genericity elsewhere to handle the API variations.
# We will recconsider this in the future if necessary.
chat_perform <- function(
  provider,
  mode = c("value", "stream", "async-stream", "async-value"),
  turns,
  tools = NULL,
  type = NULL,
  otel_span = NULL,
  controller = NULL
) {
  mode <- arg_match(mode)
  stream <- mode %in% c("stream", "async-stream")
  tools <- tools %||% list()

  setup_active_promise_otel_span(otel_span)

  req <- chat_request(
    provider = provider,
    turns = turns,
    tools = tools,
    stream = stream,
    type = type
  )

  switch(
    mode,
    "value" = req_perform(req),
    "stream" = chat_perform_stream(
      provider,
      req,
      controller = controller,
      otel_span = otel_span
    ),
    "async-value" = req_perform_promise(req),
    "async-stream" = chat_perform_async_stream(
      provider,
      req,
      controller = controller,
      otel_span = otel_span
    )
  )
}

on_load(
  chat_perform_stream <- coro::generator(function(
    provider,
    req,
    controller = NULL,
    otel_span = NULL
  ) {
    setup_active_promise_otel_span(otel_span)

    resp <- req_perform_connection(req)
    on.exit(close(resp))

    repeat {
      if (!is.null(controller) && controller$cancelled) {
        break
      }
      event <- chat_resp_stream(provider, resp)
      data <- stream_parse(provider, event)
      if (is.null(data)) {
        break
      } else {
        yield(data)
      }
    }
  })
)

on_load(
  chat_perform_async_stream <- coro::async_generator(function(
    provider,
    req,
    controller = NULL,
    otel_span = NULL
  ) {
    setup_active_promise_otel_span(otel_span)

    resp <- req_perform_connection(req, blocking = FALSE)
    on.exit(close(resp))

    repeat {
      if (!is.null(controller) && controller$cancelled) {
        break
      }
      event <- chat_resp_stream(provider, resp)
      if (is.null(event) && !resp_stream_is_complete(resp)) {
        fds <- resp$body$get_fdset()
        await(promises::promise(function(resolve, reject) {
          later::later_fd(
            resolve,
            fds$reads,
            fds$writes,
            fds$exceptions,
            fds$timeout
          )
        }))
        next
      }

      data <- stream_parse(provider, event)
      if (is.null(data)) {
        break
      } else {
        yield(data)
      }
    }
  })
)

# Request helpers --------------------------------------------------------------

ellmer_req_robustify <- function(req, is_transient = NULL, after = NULL) {
  req <- req_timeout(req, getOption("ellmer_timeout_s", 5 * 60))

  req <- req_retry(
    req,
    max_tries = getOption("ellmer_max_tries", 3),
    is_transient = is_transient,
    after = after,
    retry_on_failure = TRUE
  )

  req
}

ellmer_req_user_agent <- function(req, override = "") {
  ua <- if (identical(override, "")) ellmer_user_agent() else override
  req_user_agent(req, ua)
}
ellmer_user_agent <- function() {
  paste0("r-ellmer/", utils::packageVersion("ellmer"))
}
transform_user_agent <- function(x) {
  gsub(ellmer_user_agent(), "<ellmer_user_agent>", x, fixed = TRUE)
}

on_load(
  the$tokens <- tokens_row()
)

tokens <- function(input = 0, output = 0, cached_input = 0) {
  check_number_decimal(input, allow_null = TRUE)
  check_number_decimal(output, allow_null = TRUE)
  check_number_decimal(cached_input, allow_null = TRUE)

  list(
    input = input %||% 0,
    output = output %||% 0,
    cached_input = cached_input %||% 0
  )
}

map_tokens <- function(x, f, ...) {
  out <- t(vapply(x, f, double(3)))
  colnames(out) <- c("input", "output", "cached_input")
  out
}

log_tokens <- function(provider, tokens, cost) {
  i <- vctrs::vec_match(
    data.frame(
      provider = provider@name,
      model = provider@model
    ),
    the$tokens[c("provider", "model")]
  )

  if (is.na(i)) {
    new_row <- tokens_row(
      provider@name,
      provider@model,
      tokens$input,
      tokens$output,
      tokens$cached_input,
      cost
    )
    the$tokens <- rbind(the$tokens, new_row)
  } else {
    the$tokens$input[i] <- the$tokens$input[i] + tokens$input
    the$tokens$output[i] <- the$tokens$output[i] + tokens$output
    the$tokens$cached_input[i] <- the$tokens$cached_input[i] +
      tokens$cached_input
    the$tokens$price[i] <- the$tokens$price[i] + cost
  }

  invisible()
}

log_turn <- function(provider, turn) {
  if (is_partial_turn(turn)) {
    return(invisible())
  }
  log_tokens(provider, exec(tokens, !!!as.list(turn@tokens)), turn@cost)
}

log_turns <- function(provider, turns) {
  for (turn in turns) {
    if (S7_inherits(turn, AssistantTurn)) {
      log_turn(provider, turn)
    }
  }
}


tokens_row <- function(
  provider = character(0),
  model = character(0),
  input = numeric(0),
  output = numeric(0),
  cached_input = numeric(0),
  price = numeric(0)
) {
  data.frame(
    provider = provider,
    model = model,
    input = input,
    output = output,
    cached_input = cached_input,
    price = price
  )
}

local_tokens <- function(frame = parent.frame()) {
  old <- the$tokens
  the$tokens <- tokens_row()

  defer(the$tokens <- old, envir = frame)
}

#' Report on token usage in the current session
#'
#' Call this function to find out the cumulative number of tokens that you
#' have sent and recieved in the current session. The price will be shown
#' if known.
#'
#' @export
#' @return A data frame
#' @examples
#' token_usage()
token_usage <- function() {
  if (nrow(the$tokens) == 0) {
    cli::cli_inform(c(x = "No recorded usage in this session"))
    return(invisible(the$tokens))
  }

  the$tokens
}

# Cost ----------------------------------------------------------------------

has_cost <- function(provider, model) {
  needle <- data.frame(provider = provider@name, model = model)
  vctrs::vec_in(needle, prices[c("provider", "model")])
}

get_token_cost <- function(provider, tokens, variant = "") {
  check_string(variant, .internal = TRUE)

  needle <- data.frame(
    provider = provider@name,
    model = provider@model,
    variant = variant
  )
  idx <- vctrs::vec_match(needle, prices[c("provider", "model", "variant")])

  if (any(is.na(idx))) {
    # Match baseline if we can't match specific variant
    no_match <- is.na(idx)
    needle$variant <- ""

    idx[no_match] <- vctrs::vec_match(
      needle[no_match],
      prices[c("provider", "model", "variant")]
    )
  }

  input_price <- tokens$input * prices$input[idx] / 1e6
  output_price <- tokens$output * prices$output[idx] / 1e6
  cached_input_price <- tokens$cached_input * prices$cached_input[idx] / 1e6

  dollars(input_price + output_price + cached_input_price)
}

dollars <- function(x) {
  structure(x, class = c("ellmer_dollars", "numeric"))
}
#' @export
format.ellmer_dollars <- function(x, ...) {
  paste0(ifelse(is.na(x), "", "$"), format(unclass(round(x, 2)), nsmall = 2))
}
#' @export
print.ellmer_dollars <- function(x, ...) {
  print(format(x), quote = FALSE)
  invisible(x)
}
#' @export
`[.ellmer_dollars` <- function(x, ...) {
  dollars(NextMethod())
}
#' @export
`[[.ellmer_dollars` <- function(x, ...) {
  dollars(NextMethod())
}
#' @export
Summary.ellmer_dollars <- function(x, ...) {
  dollars(NextMethod())
}

test_that("useful message if no tokens", {
  local_tokens()

  expect_snapshot(token_usage())
})

test_that("can retrieve and log tokens", {
  local_tokens()
  provider <- test_provider("testprovider", "test")

  log_tokens(provider, tokens(input = 1), dollars(0))
  expect_equal(the$tokens, tokens_row("testprovider", "test", 1, 0, 0, 0))

  log_tokens(provider, tokens(output = 1), dollars(0))
  expect_equal(the$tokens, tokens_row("testprovider", "test", 1, 1, 0, 0))

  log_tokens(provider, tokens(cached_input = 1), dollars(0))
  expect_equal(the$tokens, tokens_row("testprovider", "test", 1, 1, 1, 0))

  log_tokens(provider, tokens(), dollars(0))
  expect_equal(the$tokens, tokens_row("testprovider", "test", 1, 1, 1, 0))

  expect_snapshot(token_usage())

  log_tokens(provider, tokens(), dollars(NA_real_))
  expect_equal(
    the$tokens,
    tokens_row("testprovider", "test", 1, 1, 1, NA_real_)
  )
})

test_that("can compute price of tokens", {
  provider <- test_provider("OpenAI", "gpt-4o")

  expect_equal(get_token_cost(provider, tokens(input = 1e6)), dollars(2.5))
  expect_equal(get_token_cost(provider, tokens(output = 1e6)), dollars(10))
  expect_equal(
    get_token_cost(provider, tokens(cached_input = 1e6)),
    dollars(1.25)
  )
})

test_that("can compute price of tokens with a variant", {
  provider <- test_provider("OpenAI", "gpt-4o")

  expect_equal(
    get_token_cost(provider, tokens(input = 1e6), variant = "priority"),
    dollars(4.25)
  )

  # fals back to baseline if no match
  expect_equal(
    get_token_cost(provider, tokens(input = 1e6), variant = "tuesday-pm"),
    get_token_cost(provider, tokens(input = 1e6))
  )
})

test_that("informative internal error if variant is missing", {
  provider <- test_provider("OpenAI", "gpt-4o")
  expect_snapshot(
    get_token_cost(provider, tokens(), variant = NULL),
    error = TRUE
  )
})


test_that("price is NA if we don't have the data for it", {
  provider <- test_provider("ClosedAI", "gpt-4o")
  expect_equal(
    get_token_cost(provider, tokens(1, 1, 1)),
    dollars(NA_real_)
  )
})

test_that("token_usage() shows price if available", {
  local_tokens()
  provider <- test_provider("OpenAI", "gpt-4o")

  toks <- tokens(input = 1.5e6, output = 2e5, cached_input = 0)
  cost <- get_token_cost(provider, toks)
  log_tokens(provider, toks, cost)
  expect_snapshot(token_usage())
})

# ellmer_dollars --------------------------------------------------------------

test_that("price is formatted nicely", {
  expect_equal(format(dollars(NA)), "NA")
  expect_equal(format(dollars(0.0001)), "$0.00")
  expect_equal(format(dollars(c(10, 1))), c("$10.00", "$ 1.00"))
})

test_that("dollars looks good, including in data.frames", {
  price <- dollars(1.234567)
  expect_snapshot({
    price
    data.frame(price)
  })
})

test_that("dollars class survives basic transforms", {
  expect_equal(sum(dollars(c(1, 2))), dollars(3))

  d <- dollars(c(1, 2, 3))
  expect_equal(d[1:2], dollars(c(1, 2)))
  expect_equal(d[[1]], dollars(1))
})

# Helpers ---------------------------------------------------------------------

test_that("log_turns ignores non-assistant turns", {
  local_tokens()
  provider <- test_provider("testprovider", "test")

  turn1 <- UserTurn(contents = "text")
  turn2 <- AssistantTurn(
    contents = "Hello",
    tokens = c(8, 3, 2),
    cost = dollars(1)
  )

  log_turns(provider, list(turn1, turn2, NULL))
  expect_equal(the$tokens, tokens_row("testprovider", "test", 8, 3, 2, 1))
})

test_that("log_turns aggregates multiple turns", {
  local_tokens()
  provider <- test_provider("testprovider", "test")

  turn1 <- AssistantTurn(contents = "Hello", tokens = c(8, 3, 2))
  turn2 <- AssistantTurn(contents = "World", tokens = c(1, 1, 1))
  log_turns(provider, list(turn1, turn2))
  expect_equal(
    the$tokens,
    tokens_row("testprovider", "test", 9, 4, 3, NA_real_)
  )
})

otel_tracer_name <- "co.posit.r-package.ellmer"

otel_cache_tracer <- NULL
otel_capture_content_enabled <- NULL
local_chat_otel_span <- NULL
local_tool_otel_span <- NULL
local_agent_otel_span <- NULL

local({
  otel_is_tracing <- FALSE
  otel_tracer <- NULL
  otel_capture_content <- FALSE

  otel_cache_tracer <<- function() {
    if (!requireNamespace("otel", quietly = TRUE)) {
      return()
    }
    otel_tracer <<- otel::get_tracer(otel_tracer_name)
    otel_is_tracing <<- tracer_enabled(otel_tracer)
    otel_capture_content <<- {
      val <- Sys.getenv("OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT")
      tolower(val) %in% c("true", "1")
    }
  }

  otel_capture_content_enabled <<- function() otel_capture_content

  local_chat_otel_span <<- function(
    provider,
    turns = NULL,
    system_prompt = NULL,
    parent = NULL,
    local_envir = parent.frame()
  ) {
    if (!otel_is_tracing) {
      return()
    }
    chat_span <-
      otel::start_span(
        sprintf("chat %s", provider@model),
        options = list(
          parent = parent,
          kind = "client"
        ),
        attributes = list(
          "gen_ai.operation.name" = "chat",
          "gen_ai.provider.name" = tolower(provider@name),
          "gen_ai.request.model" = provider@model
        ),
        tracer = otel_tracer
      )

    defer(otel::end_span(chat_span), envir = local_envir)

    if (otel_capture_content) {
      if (!is.null(system_prompt)) {
        parts <- lapply(system_prompt@contents, as_otel_part)
        chat_span$set_attribute(
          "gen_ai.system_instructions",
          jsonlite::toJSON(parts, auto_unbox = TRUE, null = "null")
        )
      }
      if (length(turns)) {
        # Tool result values are typed `class_any`, so tools can return objects
        # (environments, R6, external pointers) that `jsonlite::toJSON` rejects.
        # Skip emission rather than break the chat — the provider's tool_string
        # path will surface a descriptive error.
        tryCatch(
          {
            msgs <- lapply(turns, as_otel_message)
            chat_span$set_attribute(
              "gen_ai.input.messages",
              jsonlite::toJSON(msgs, auto_unbox = TRUE, null = "null")
            )
          },
          error = function(e) NULL
        )
      }
    }

    chat_span
  }

  # Starts an Open Telemetry span that abides by the semantic conventions for
  # Generative AI tool calls.
  #
  # Must be activated for the calling scope.
  #
  # See: https://opentelemetry.io/docs/specs/semconv/gen-ai/gen-ai-spans/#execute-tool-span
  local_tool_otel_span <<- function(
    request,
    parent = NULL,
    local_envir = parent.frame()
  ) {
    if (!otel_is_tracing) {
      return()
    }
    tool_span <-
      otel::start_span(
        sprintf("execute_tool %s", request@tool@name),
        options = list(parent = parent),
        attributes = compact(list(
          "gen_ai.operation.name" = "execute_tool",
          "gen_ai.tool.name" = request@tool@name,
          "gen_ai.tool.description" = request@tool@description,
          "gen_ai.tool.call.id" = request@id
        )),
        tracer = otel_tracer
      )

    setup_active_promise_otel_span(tool_span, local_envir)

    defer(otel::end_span(tool_span), envir = local_envir)

    tool_span
  }

  # Starts an Open Telemetry span that abides by the semantic conventions for
  # Generative AI "agents".
  #
  # See: https://opentelemetry.io/docs/specs/semconv/gen-ai/gen-ai-spans/#inference
  # local_otel_span_agent
  local_agent_otel_span <<- function(
    provider,
    activate = TRUE,
    local_envir = parent.frame()
  ) {
    if (!otel_is_tracing) {
      return()
    }
    if (activate) {
      abort(c(
        "Activating the agent span is not supported at this time.",
        "*" = "Activating the span here would set it as the active span globally (via otel::local_active_span() until the calling function ends (a long time).",
        "*" = "`coro::setup()` would address this and be appropriate",
        "i" = "Work around: Activate only where necessary or over a single yield in the calling scope."
      ))
    }
    agent_span <-
      otel::start_span(
        "invoke_agent",
        options = list(kind = "client"),
        attributes = list(
          "gen_ai.operation.name" = "invoke_agent",
          "gen_ai.provider.name" = tolower(provider@name),
          "gen_ai.request.model" = provider@model
        ),
        tracer = otel_tracer
      )

    ## Do not activate!
    ## The current usage of `local_agent_otel_span()` is in a multi-step coroutine.
    ## This would require deactivating only after the coroutine is done,
    ## but not between yields, that is too long and unpredictable.
    ## The span should only be activated in the specific steps where it is needed.
    # setup_active_promise_otel_span(agent_span, local_envir)

    defer(otel::end_span(agent_span), envir = local_envir)

    agent_span
  }
})

tracer_enabled <- function(tracer) {
  .subset2(tracer, "is_enabled")()
}

span_recording <- function(span) {
  .subset2(span, "is_recording")()
}

with_otel_record <- function(expr) {
  on.exit(otel_cache_tracer())
  otelsdk::with_otel_record({
    otel_cache_tracer()
    expr
  })
}

record_chat_otel_span_status <- function(span, provider, result) {
  if (is.null(span) || !span_recording(span)) {
    return()
  }
  if (!is.null(result$model)) {
    span$set_attribute("gen_ai.response.model", result$model)
  }
  if (!is.null(result$id)) {
    span$set_attribute("gen_ai.response.id", result$id)
  }
  tokens <- value_tokens(provider, result)
  input <- as.integer(tokens$input + tokens$cached_input)
  output <- as.integer(tokens$output)
  if (input > 0L || output > 0L) {
    span$set_attribute("gen_ai.usage.input_tokens", input)
    span$set_attribute("gen_ai.usage.output_tokens", output)
  }
  # TODO: Consider setting gen_ai.response.finish_reasons.
  span$set_status("ok")
}

# Convert a single Content into a GenAI semconv "part" — a named list emitted
# as one entry of a ChatMessage's `parts` array. New content classes fall
# through to the default method, which emits a schema-valid `generic` part.
as_otel_part <- new_generic("as_otel_part", "content")

method(as_otel_part, Content) <- function(content) {
  list(type = "generic", class = S7_class(content)@name)
}

method(as_otel_part, ContentText) <- function(content) {
  list(type = "text", content = content@text)
}

method(as_otel_part, ContentToolRequest) <- function(content) {
  list(
    type = "tool_call",
    id = content@id,
    name = content@name,
    arguments = content@arguments
  )
}

method(as_otel_part, ContentToolResult) <- function(content) {
  part <- list(type = "tool_call_response")
  if (!is.null(content@request)) {
    part$id <- content@request@id
  }
  part$response <- tool_otel_response(content)
  part
}

tool_otel_response <- function(content) {
  if (tool_errored(content)) {
    return(tool_error_string(content))
  }
  value <- content@value
  if (inherits(value, "json")) {
    # Parse so jsonlite re-emits the structured value rather than encoding the
    # JSON string itself as a quoted string under `auto_unbox = TRUE`.
    return(jsonlite::fromJSON(value, simplifyVector = FALSE))
  }
  value
}

# Produce a GenAI semconv ChatMessage from a Turn. Tool-result UserTurns get
# role "tool" so consumers can filter them out of normal user input — matching
# Python's GenAI instrumentations.
as_otel_message <- function(turn) {
  list(
    role = if (is_tool_result_turn(turn)) "tool" else turn@role,
    parts = lapply(turn@contents, as_otel_part)
  )
}

is_tool_result_turn <- function(turn) {
  S7_inherits(turn, UserTurn) &&
    length(turn@contents) > 0 &&
    all(map_lgl(turn@contents, S7_inherits, ContentToolResult))
}

otel_chat_input <- function(private, user_turn) {
  if (private$has_system_prompt()) {
    sys_turn <- private$.turns[[1]]
    history <- private$.turns[-1]
  } else {
    sys_turn <- NULL
    history <- private$.turns
  }
  list(
    turns = c(history, list(user_turn)),
    system_prompt = sys_turn
  )
}

record_chat_otel_span_output <- function(span, turn) {
  if (is.null(span) || !span_recording(span)) {
    return()
  }
  if (!otel_capture_content_enabled()) {
    return()
  }
  if (!S7_inherits(turn, AssistantTurn)) {
    return()
  }
  msg <- as_otel_message(turn)
  span$set_attribute(
    "gen_ai.output.messages",
    jsonlite::toJSON(list(msg), auto_unbox = TRUE, null = "null")
  )
}

record_tool_otel_span_error <- function(span, error) {
  if (is.null(span) || !span_recording(span)) {
    return()
  }
  span$record_exception(error)
  span$set_status("error")
  span$set_attribute("error.type", class(error)[1L])
}

# Only activate the span if it is non-NULL. If
# otel_promise_domain is TRUE, also ensure that the active span is reactivated upon promise domain restoration.
#' Activate and use handoff promise domain for Open Telemetry span
#'
#' @param otel_span An Open Telemetry span object.
#' @param activation_scope The scope in which to activate the span.
#' @noRd
setup_active_promise_otel_span <- function(
  span,
  activation_scope = parent.frame()
) {
  if (is.null(span) || !span_recording(span)) {
    return()
  }

  promises::local_otel_promise_domain(activation_scope)
  otel::local_active_span(span, activation_scope = activation_scope)

  invisible()
}

#' Helpers for interpolating data into prompts
#'
#' @description
#' These functions are lightweight wrappers around
#' [glue](https://glue.tidyverse.org/) that make it easier to interpolate
#' dynamic data into a static prompt:
#'
#' * `interpolate()` works with a string.
#' * `interpolate_file()` works with a file.
#' * `interpolate_package()` works with a file in the `inst/prompts`
#'   directory of a package.
#'
#' Compared to glue, dynamic values should be wrapped in `{{ }}`, making it
#' easier to include R code and JSON in your prompt.
#'
#' @param prompt A prompt string. You should not generally expose this
#'   to the end user, since glue interpolation makes it easy to run arbitrary
#'   code.
#' @param ... Define additional temporary variables for substitution.
#' @param .envir Environment to evaluate `...` expressions in. Used when
#'   wrapping in another function. See `vignette("wrappers", package = "glue")`
#'   for more details.
#' @return A \{glue\} string.
#' @export
#' @examples
#' joke <- "You're a cool dude who loves to make jokes. Tell me a joke about {{topic}}."
#'
#' # You can supply valuese directly:
#' interpolate(joke, topic = "bananas")
#'
#' # Or allow interpolate to find them in the current environment:
#' topic <- "applies"
#' interpolate(joke)
#'
#'
interpolate <- function(prompt, ..., .envir = parent.frame()) {
  check_string(prompt)

  dots <- list2(...)
  if (length(dots) > 0 && !is_named(dots)) {
    cli::cli_abort("All elements of `...` must be named")
  }

  envir <- list2env(dots, parent = .envir)
  out <- glue::glue(prompt, .open = "{{", .close = "}}", .envir = envir)
  ellmer_prompt(out)
}

#' @param path A path to a prompt file (often a `.md`). In
#'   `interpolate_package()`, this path is relative to `inst/prompts`.
#' @rdname interpolate
#' @export
interpolate_file <- function(path, ..., .envir = parent.frame()) {
  check_string(path)
  string <- read_file(path)
  interpolate(string, ..., .envir = .envir)
}

#' @param package Package name.
#' @rdname interpolate
#' @export
interpolate_package <- function(package, path, ..., .envir = parent.frame()) {
  check_string(package)
  check_string(path)

  prompts_path <- file.path(inst_path(package), "prompts")
  if (!dir.exists(prompts_path)) {
    cli::cli_abort(
      "{.pkg {package}} does not have a {.field prompts/} directory."
    )
  }

  prompt_path <- file.path(prompts_path, path)
  if (!file.exists(prompt_path)) {
    cli::cli_abort(c(
      "{.pkg {package}} does not have {.val {path}} in its {.field prompts/} directory.",
      "i" = 'Run {.run dir(system.file("prompts", package = "{package}"))} to see available prompts.'
    ))
  }

  interpolate_file(prompt_path, ..., .envir = .envir)
}

read_file <- function(path) {
  if (!file.exists(path)) {
    cli::cli_abort(
      "{.arg path} {.path {path}} does not exist.",
      call = caller_env()
    )
  }
  readChar(path, file.size(path))
}

# Prompt class -----------------------------------------------------------------

ellmer_prompt <- function(x) {
  structure(x, class = c("ellmer_prompt", "character"))
}

is_prompt <- function(x) {
  inherits(x, "ellmer_prompt")
}

#' @export
print.ellmer_prompt <- function(
  x,
  ...,
  max_items = 20,
  max_lines = max_items * 10
) {
  check_number_whole(max_items, min = 1)
  check_number_whole(max_lines, min = 1)

  n <- length(x)
  n_extra <- length(x) - max_items
  if (n_extra > 0) {
    x <- x[seq_len(max_items)]
  }

  if (length(x) == 0) {
    cli::cli_inform(c(x = "Zero-length prompt.\n"))
    return(invisible(x))
  }

  bar <- if (cli::is_utf8_output()) "\u2502" else "|"

  id <- format(paste0("[", seq_along(x), "] "), justify = "right")
  indent <- paste0(cli::col_grey(id, bar), " ")
  exdent <- paste0(strrep(" ", nchar(id[[1]])), cli::col_grey(bar), " ")

  x[is.na(x)] <- cli::col_red("NA")
  x <- paste0(indent, x)
  x <- gsub("\n", paste0("\n", exdent), x)

  lines <- strsplit(x, "\n")
  ids <- rep(seq_along(x), lengths(lines))
  first <- c(TRUE, ids[-length(ids)] != ids[-1])
  lines <- unlist(lines)

  if (length(lines) > max_lines) {
    if (first[max_lines + 1]) {
      max_lines <- if (ids[max_lines] > 1) max_lines - 1 else max(max_lines, 1)
    }

    lines <- lines[seq_len(max_lines)]
    lines <- c(lines, paste0(exdent, "..."))
    n_extra <- n - ids[max_lines]
  }

  cat(lines, sep = "\n")
  if (n_extra > 0) {
    cat("... and ", n_extra, " more.\n", sep = "")
  }

  invisible(x)
}

#' @export
`[.ellmer_prompt` <- function(x, i, ...) {
  ellmer_prompt(NextMethod())
}

# Helpers ----------------------------------------------------------------------

inst_path <- function(package) {
  if (is_dev_package(package)) {
    file.path(getNamespaceInfo(package, "path"), "inst")
  } else {
    system.file(package = package)
  }
}

is_dev_package <- function(pkg) {
  !is.null(.getNamespace(pkg)[[".__DEVTOOLS__"]])
}

test_that("checks inputs", {
  expect_snapshot(error = TRUE, {
    interpolate(1)
    interpolate("x", 1)
  })
})

test_that("vectorised interpolation generates a list", {
  expect_equal(
    interpolate("{{x}}", x = 1:2),
    ellmer_prompt(c("1", "2"))
  )
})

test_that("has a nice print method", {
  expect_snapshot(interpolate("Hi!"))
})

test_that("print method truncates many elements", {
  prompt <- ellmer_prompt(c("x\ny", c("a\nb\nc\nd\ne")))
  expect_snapshot({
    print(prompt, max_items = 1)
    print(prompt, max_lines = 2)
    print(prompt, max_lines = 3)
    print(ellmer_prompt("a\nb\nc\nd\ne"), max_lines = 3)
  })
})

test_that("can interpolate from local env or from ...", {
  x <- 1

  expect_equal(interpolate("{{x}}"), ellmer_prompt("1"))
  expect_equal(interpolate("{{x}}", x = 2), ellmer_prompt("2"))
})

test_that("can take a data frame via !!!", {
  df <- data.frame(x = 1, y = 2)
  expect_equal(interpolate("{{x}} + {{y}}", !!!df), ellmer_prompt("1 + 2"))
})

test_that("can interpolate from a file", {
  path <- withr::local_tempfile(lines = "{{x}}")
  expect_equal(interpolate_file(path, x = 1), ellmer_prompt("1"))

  # and errors if file doesn't exist
  expect_snapshot(interpolate_file("does-not-exist.md", x = 1), error = TRUE)
})

# interpolate_package() -------------------------------------------------------

test_that("can interpolate from a package", {
  inst_path <- withr::local_tempdir()
  local_mocked_bindings(inst_path = function(...) inst_path)

  dir.create(file.path(inst_path, "prompts"))
  con <- file(file.path(inst_path, "prompts", "bar.md"), "wb")
  writeLines("{{x}}", con)
  close(con)

  expect_equal(interpolate_package("test", "bar.md", x = 1), ellmer_prompt("1"))
})

test_that("informative errors if can't find prompts", {
  inst_path <- withr::local_tempdir()
  local_mocked_bindings(inst_path = function(...) inst_path)

  expect_snapshot(interpolate_package("x", "does-not-exist.md"), error = TRUE)

  dir.create(file.path(inst_path, "prompts"))
  expect_snapshot(interpolate_package("x", "does-not-exist.md"), error = TRUE)
})

test_that("checks its inputs", {
  expect_snapshot(error = TRUE, {
    interpolate_package(1)
    interpolate_package("x", 1)
  })
})

#' @include provider-openai-compatible.R
#' @include content.R
NULL

#' Chat with a model hosted by vLLM
#'
#' @description
#' [vLLM](https://docs.vllm.ai/en/latest/) is an open source library that
#' provides an efficient and convenient LLMs model server. You can use
#' `chat_vllm()` to connect to endpoints powered by vLLM.
#'
#' Uses OpenAI compatible API via `chat_openai_compatible()`.
#'
#' @inheritParams chat_openai
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("VLLM_API_KEY")`
#' @param model `r param_model(NULL, "vllm")`
#' @param params Common model parameters, usually created by [params()].
#' @inherit chat_openai return
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_vllm("http://my-vllm.com")
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_vllm <- function(
  base_url,
  system_prompt = NULL,
  model,
  params = NULL,
  api_args = list(),
  api_key = NULL,
  credentials = NULL,
  echo = NULL,
  api_headers = character()
) {
  check_string(base_url)

  credentials <- as_credentials(
    "chat_vllm",
    function() vllm_key(),
    credentials = credentials,
    api_key = api_key
  )

  check_string(credentials())
  if (missing(model)) {
    models <- models_vllm(base_url, credentials = credentials)$id
    cli::cli_abort(c(
      "Must specify {.arg model}.",
      i = "Available models: {.str {models}}."
    ))
  }
  echo <- check_echo(echo)

  # https://docs.vllm.ai/en/latest/serving/openai_compatible_server.html
  params <- params %||% params()

  provider <- ProviderVllm(
    name = "VLLM",
    base_url = base_url,
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

chat_vllm_test <- function(..., echo = "none") {
  chat_vllm(
    base_url = "https://llm.nrp-nautilus.io/",
    ...,
    model = "llama3",
    echo = echo
  )
}

ProviderVllm <- new_class(
  "ProviderVllm",
  parent = ProviderOpenAICompatible,
  package = "ellmer",
)

# Just like OpenAI but no strict
method(as_json, list(ProviderVllm, ToolDef)) <- function(provider, x, ...) {
  list(
    type = "function",
    "function" = compact(list(
      name = x@name,
      description = x@description,
      parameters = as_json(provider, x@arguments, ...)
    ))
  )
}

vllm_key <- function() {
  key_get("VLLM_API_KEY")
}

#' @export
#' @rdname chat_vllm
models_vllm <- function(base_url, api_key = NULL, credentials = NULL) {
  credentials <- as_credentials(
    "models_vllm",
    function() vllm_key(),
    credentials = credentials,
    api_key = api_key
  )

  req <- request(base_url)
  req <- req_auth_bearer_token(req, credentials())
  req <- req_url_path_append(req, "/v1/models")
  resp <- req_perform(req)
  json <- resp_body_json(resp)

  data.frame(
    id = map_chr(json$data, "[[", "id")
    # Not accurate?
    # created = .POSIXct(map_dbl(json$data, "[[", "created")),
    # owned_by = map_chr(json$data, "[[", "owned_by")
  )
}

#' Encode PDFs content for chat input
#'
#' @description
#' These functions are used to prepare PDFs as input to the chatbot. The
#' `content_pdf_url()` function is used to provide a URL to an PDF file,
#' while `content_pdf_file()` is used to for local PDF files.
#'
#' Not all providers support PDF input, so check the documentation for the
#' provider you are using.
#'
#' @param path,url Path or URL to a PDF file.
#' @return A `ContentPDF` object
#' @export
content_pdf_file <- function(path) {
  check_string(path, allow_empty = FALSE)
  if (!file.exists(path) || dir.exists(path)) {
    cli::cli_abort("{.arg path} must be an existing file.")
  }

  ContentPDF(
    type = "application/pdf",
    data = base64_enc(path = path),
    filename = basename(path)
  )
}

#' @rdname content_pdf_file
#' @export
content_pdf_url <- function(url) {
  if (grepl("^data:", url)) {
    parsed <- parse_data_url(url)
    ContentPDF(parsed$content_type, parsed$base64, unique_pdf_name())
  } else {
    # TODO: need seperate ContentPDFRemote type so we can use file upload
    # apis where they exist. Might need some kind of mutable state so can
    # record point to uploaded file.
    path <- tempfile(pattern = basename(url), fileext = ".pdf")
    on.exit(unlink(path))

    resp <- httr2::req_perform(httr2::request(url), path = path)
    content_pdf_file(path)
  }
}

unique_pdf_name <- function() {
  the$cur_pdf_id <- (the$cur_pdf_id %||% 0) + 1
  sprintf("file_%03d.pdf", the$cur_pdf_id)
}

#' @include provider.R
#' @include types.R

method(as_json, list(Provider, TypeBasic)) <- function(provider, x, ...) {
  list(type = x@type, description = x@description %||% "")
}

method(as_json, list(Provider, TypeEnum)) <- function(provider, x, ...) {
  list(
    type = "string",
    description = x@description %||% "",
    enum = as.list(x@values)
  )
}

method(as_json, list(Provider, TypeObject)) <- function(provider, x, ...) {
  names <- names2(x@properties)
  required <- map_lgl(x@properties, function(prop) prop@required)

  properties <- as_json(provider, x@properties, ...)
  names(properties) <- names

  list(
    type = "object",
    description = x@description %||% "",
    properties = properties,
    required = as.list(names[required]),
    additionalProperties = x@additional_properties
  )
}

method(as_json, list(Provider, TypeArray)) <- function(provider, x, ...) {
  list(
    type = "array",
    description = x@description %||% "",
    items = as_json(provider, x@items, ...)
  )
}

method(as_json, list(Provider, TypeJsonSchema)) <- function(provider, x, ...) {
  x@json
}

#' @include provider.R
#' @include content.R
#' @include turns.R
#' @include tools-def.R
NULL

#' Chat with an AWS bedrock model
#'
#' @description
#' [AWS Bedrock](https://aws.amazon.com/bedrock/) provides a number of
#' language models, including those from Anthropic's
#' [Claude](https://aws.amazon.com/bedrock/claude/), using the Bedrock
#' [Converse API](https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html).
#'
#' ## Authentication
#'
#' Authentication is handled through \{paws.common\}, so if authentication
#' does not work for you automatically, you'll need to follow the advice
#' at <https://www.paws-r-sdk.com/#credentials>. In particular, if your
#' org uses AWS SSO, you'll need to run `aws sso login` at the terminal.
#'
#' ## Prompt caching
#'
#' Bedrock supports
#' [prompt caching](https://docs.aws.amazon.com/bedrock/latest/userguide/prompt-caching.html)
#' via cache checkpoints. When caching is enabled, ellmer places cache
#' checkpoints on the system prompt and the last turn, so that the
#' conversation history is cached across turns.
#'
#' By default (`cache = "auto"`), caching is enabled for models known to
#' support it (Anthropic Claude and Amazon Nova) and disabled for all other
#' models. You can also set `cache` to `"5m"` or `"1h"` to force a specific
#' TTL, or `"none"` to disable caching entirely. Note that individual models
#' may have minimum input token thresholds before caching takes effect.
#'
#' Note that [token_usage()] does not currently reflect the cost of writing
#' to the cache, which is priced at a premium over regular input tokens.
#' Cache read savings are reported correctly.
#'
#' @param profile AWS profile to use.
#' @param cache How long to cache inputs? The default, `"auto"`, enables
#'   caching with a 5-minute TTL for models known to support it (Anthropic
#'   Claude and Amazon Nova) and disables caching for all other models.
#'   Set to `"5m"` or `"1h"` to force caching on, or `"none"` to disable it.
#'
#'   See details below.
#' @param model `r param_model("anthropic.claude-sonnet-4-5-20250929-v1:0", "models_aws_bedrock")`.
#'
#'   While ellmer provides a default model, there's no guarantee that you'll
#'   have access to it, so you'll need to specify a model that you can.
#'   If you're using [cross-region inference](https://aws.amazon.com/blogs/machine-learning/getting-started-with-cross-region-inference-in-amazon-bedrock/),
#'   you'll need to use the inference profile ID, e.g.
#'   `model="us.anthropic.claude-sonnet-4-5-20250929-v1:0"`.
#' @param params Common model parameters, usually created by [params()].
#' @param api_args Named list of arbitrary extra arguments appended to the body
#'   of every chat API call. Use `params` for common parameters. Model-specific
#'   inference parameters can be provided using the
#'   `additionalModelRequestFields` field, for example to enable thinking effort
#'   in Anthropic Claude models:
#'
#'   ```R
#'   api_args = list(
#'     additionalModelRequestFields = list(
#'       thinking = list(type = "enabled", budget_tokens = 4000)
#'     )
#'   )
#'   ```
#'
#'   See <https://docs.aws.amazon.com/bedrock/latest/userguide/conversation-inference-call.html>
#'   for more details.
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @family chatbots
#' @export
#' @examples
#' \dontrun{
#' # Basic usage
#' chat <- chat_aws_bedrock()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_aws_bedrock <- function(
  system_prompt = NULL,
  base_url = NULL,
  model = NULL,
  profile = NULL,
  cache = c("auto", "5m", "1h", "none"),
  params = NULL,
  api_args = list(),
  api_headers = character(),
  echo = NULL
) {
  check_installed("paws.common", "AWS authentication")
  check_string(base_url, allow_null = TRUE)
  base_url <- base_url %||%
    \(x) sprintf("https://bedrock-runtime.%s.amazonaws.com", x)
  echo <- check_echo(echo)

  params <- params %||% params()

  provider <- provider_aws_bedrock(
    base_url = base_url,
    model = model,
    profile = profile,
    cache_point = cache,
    params = params,
    extra_args = api_args,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}


#' @export
#' @rdname chat_aws_bedrock
models_aws_bedrock <- function(profile = NULL, base_url = NULL) {
  check_string(base_url, allow_null = TRUE)
  base_url <- base_url %||% \(x) sprintf("https://bedrock.%s.amazonaws.com", x)

  provider <- provider_aws_bedrock(
    base_url = base_url,
    model = "",
    profile = profile,
  )
  req <- base_request(provider)
  req <- req_url_path_append(req, "foundation-models")

  resp <- req_perform(req)
  json <- resp_body_json(resp)
  models <- json$modelSummaries

  df <- data.frame(
    id = map_chr(models, "[[", "modelId"),
    name = map_chr(models, "[[", "modelName"),
    provider = map_chr(models, "[[", "providerName")
  )
  df <- cbind(df, match_prices("AWS/Bedrock", df$id))
  df
}

chat_aws_bedrock_test <- function(
  ...,
  model = "us.anthropic.claude-haiku-4-5-20251001-v1:0",
  params = NULL,
  echo = "none"
) {
  params <- params %||% params()
  params$temperature <- params$temperature %||% 0

  chat_aws_bedrock(model = model, params = params, ..., echo = echo)
}

provider_aws_bedrock <- function(
  base_url,
  model = "",
  profile = NULL,
  cache_point = "none",
  params = list(),
  extra_args = list(),
  extra_headers = character()
) {
  cache <- aws_creds_cache(profile)
  credentials <- paws_credentials(profile, cache = cache)

  if (is.function(base_url)) {
    base_url <- base_url(credentials$region)
  }

  model <- set_default(model, "anthropic.claude-sonnet-4-5-20250929-v1:0")

  cache_point <- as_bedrock_cache_point(cache_point, model)

  ProviderAWSBedrock(
    name = "AWS/Bedrock",
    base_url = base_url,
    model = model,
    profile = profile,
    region = credentials$region,
    cache = cache,
    cache_point = cache_point,
    params = params,
    extra_args = extra_args,
    extra_headers = extra_headers
  )
}

ProviderAWSBedrock <- new_class(
  "ProviderAWSBedrock",
  parent = Provider,
  properties = list(
    profile = prop_string(allow_null = TRUE),
    region = prop_string(),
    cache = class_list,
    cache_point = prop_string()
  )
)

method(base_request, ProviderAWSBedrock) <- function(provider) {
  creds <- paws_credentials(provider@profile, provider@cache)

  req <- request(provider@base_url)
  req <- req_auth_aws_v4(
    req,
    aws_access_key_id = creds$access_key_id,
    aws_secret_access_key = creds$secret_access_key,
    aws_session_token = creds$session_token
  )
  req <- ellmer_req_robustify(req)
  req <- ellmer_req_user_agent(req)
  req <- base_request_error(provider, req)
  req
}

method(base_request_error, ProviderAWSBedrock) <- function(provider, req) {
  req_error(req, body = function(resp) {
    body <- resp_body_json(resp)
    body$Message %||% body$message
  })
}

method(chat_params, ProviderAWSBedrock) <- function(provider, params) {
  # https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_InferenceConfiguration.html
  standardise_params(
    params,
    c(
      temperature = "temperature",
      topP = "top_p",
      maxTokens = "max_tokens",
      stopSequences = "stop_sequences"
    )
  )
}

method(chat_request, ProviderAWSBedrock) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  req <- base_request(provider)
  req <- req_url_path_append(
    req,
    "model",
    provider@model,
    if (stream) "converse-stream" else "converse"
  )

  if (length(turns) >= 1 && is_system_turn(turns[[1]])) {
    system <- c(
      list(list(text = turns[[1]]@text)),
      bedrock_cache_point(provider)
    )
  } else {
    system <- NULL
  }

  is_last <- seq_along(turns) == length(turns)
  messages <- compact(map2(turns, is_last, function(turn, is_last) {
    as_json(provider, turn, is_last = is_last)
  }))

  if (!is.null(type)) {
    tool_def <- ToolDef(
      function(...) {},
      name = "structured_tool_call__",
      description = "Extract structured data",
      arguments = type_object(data = type)
    )
    tools[[tool_def@name]] <- tool_def
    tool_choice <- list(tool = list(name = tool_def@name))
    stream <- FALSE
  } else {
    tool_choice <- NULL
  }

  if (length(tools) > 0) {
    tools <- as_json(provider, unname(tools))
    toolConfig <- compact(list(tools = tools, tool_choice = tool_choice))
  } else {
    toolConfig <- NULL
  }

  # Merge params into inferenceConfig, giving precedence to manual api_args
  params <- chat_params(provider, provider@params)

  extra_args <- provider@extra_args
  extra_args$inferenceConfig <- modify_list(params, extra_args$inferenceConfig)

  # https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html
  body <- compact(list2(
    messages = messages,
    system = system,
    toolConfig = toolConfig,
    !!!extra_args
  ))

  req <- req_body_json(req, body)
  req <- req_headers(req, !!!provider@extra_headers)

  req
}

method(chat_resp_stream, ProviderAWSBedrock) <- function(provider, resp) {
  resp_stream_aws(resp)
}

# Bedrock -> ellmer -------------------------------------------------------------

method(stream_parse, ProviderAWSBedrock) <- function(provider, event) {
  if (is.null(event)) {
    return()
  }

  body <- event$body
  body$event_type <- event$headers$`:event-type`
  body$p <- NULL # padding? Looks like: "p": "abcdefghijklmnopqrstuvwxyzABCDEFGHIJ",

  body
}

method(stream_content, ProviderAWSBedrock) <- function(provider, event) {
  if (event$event_type == "contentBlockDelta") {
    text <- event$delta$text
    if (is.null(text)) {
      return(NULL)
    }
    ContentText(text)
  }
}

method(stream_merge_chunks, ProviderAWSBedrock) <- function(
  provider,
  result,
  chunk
) {
  i <- chunk$contentBlockIndex + 1

  if (chunk$event_type == "messageStart") {
    result <- list(role = chunk$role, content = list())
  } else if (chunk$event_type == "contentBlockStart") {
    result$content[[i]] <- list(toolUse = chunk$start$toolUse)
  } else if (chunk$event_type == "contentBlockDelta") {
    if (i > length(result$content)) {
      result$content[[i]] <- list()
    }
    if (has_name(chunk$delta, "text")) {
      paste(result$content[[i]]$text) <- chunk$delta$text
    } else if (has_name(chunk$delta, "toolUse")) {
      paste(result$content[[i]]$toolUse$input) <- chunk$delta$toolUse$input
    } else if (has_name(chunk$delta, "reasoningContent")) {
      if (is.null(result$content[[i]]$reasoningContent)) {
        result$content[[i]]$reasoningContent <- list(reasoningText = list())
      }
      # https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ReasoningContentBlockDelta.html
      delta <- chunk$delta$reasoningContent
      if (has_name(delta, "text")) {
        paste(result$content[[i]]$reasoningContent$reasoningText$text) <-
          delta$text
      } else if (has_name(delta, "signature")) {
        result$content[[i]]$reasoningContent$reasoningText$signature <-
          delta$signature
      }
    } else {
      cli::cli_abort(
        "Unknown chunk type {names(chunk$delta)}",
        .internal = TRUE
      )
    }
  } else if (chunk$event_type == "contentBlockStop") {
    if (has_name(result$content[[i]], "toolUse")) {
      input <- result$content[[i]]$toolUse$input
      if (input == "") {
        result$content[[i]]$toolUse$input <- set_names(list())
      } else {
        result$content[[i]]$toolUse$input <- jsonlite::parse_json(input)
      }
    }
  } else if (chunk$event_type == "messageStop") {
    # match structure of non-streaming
    result <- list(
      output = list(
        message = result
      )
    )
  } else if (chunk$event_type == "metadata") {
    result$usage <- chunk$usage
    result$metrics <- chunk$metrics
  } else {
    cli::cli_inform(c("!" = "Unknown chunk type {.str {event_type}}."))
  }

  result
}

method(value_tokens, ProviderAWSBedrock) <- function(provider, json) {
  usage <- json$usage
  tokens(
    input = usage$inputTokens %||% 0,
    output = usage$outputTokens %||% 0,
    cached_input = usage$cacheReadInputTokens %||% 0
  )
}

method(value_turn, ProviderAWSBedrock) <- function(
  provider,
  result,
  has_type = FALSE
) {
  contents <- lapply(result$output$message$content, function(content) {
    if (has_name(content, "text")) {
      ContentText(content$text)
    } else if (has_name(content, "toolUse")) {
      if (has_type) {
        ContentJson(data = content$toolUse$input$data)
      } else {
        ContentToolRequest(
          name = content$toolUse$name,
          arguments = content$toolUse$input,
          id = content$toolUse$toolUseId
        )
      }
    } else if (has_name(content, "reasoningContent")) {
      ContentThinking(
        content$reasoningContent$reasoningText$text,
        extra = list(
          signature = content$reasoningContent$reasoningText$signature
        )
      )
    } else {
      cli::cli_abort(
        "Unknown content type {.str {names(content)}}.",
        .internal = TRUE
      )
    }
  })

  tokens <- value_tokens(provider, result)
  cost <- get_token_cost(provider, tokens)
  AssistantTurn(contents, json = result, tokens = unlist(tokens), cost = cost)
}

# ellmer -> Bedrock -------------------------------------------------------------

# https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ContentBlock.html
method(as_json, list(ProviderAWSBedrock, Turn)) <- function(
  provider,
  x,
  ...,
  is_last = FALSE
) {
  if (is_system_turn(x)) {
    NULL
  } else if (is_user_turn(x) || is_assistant_turn(x)) {
    x <- turn_contents_expand(x)
    content <- as_json(provider, x@contents, ...)

    if (is_last) {
      content <- c(content, bedrock_cache_point(provider))
    }

    list(role = x@role, content = content)
  } else {
    cli::cli_abort("Unknown role {x@role}", .internal = TRUE)
  }
}

method(as_json, list(ProviderAWSBedrock, ContentText)) <- function(
  provider,
  x,
  ...
) {
  if (is_whitespace(x@text)) {
    list(text = "[empty string]")
  } else {
    list(text = x@text)
  }
}

method(as_json, list(ProviderAWSBedrock, ContentImageRemote)) <- function(
  provider,
  x,
  ...
) {
  cli::cli_abort("Bedrock doesn't support remote images")
}

# https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ImageBlock.html
method(as_json, list(ProviderAWSBedrock, ContentImageInline)) <- function(
  provider,
  x,
  ...
) {
  type <- switch(
    x@type,
    "image/png" = "png",
    "image/gif" = "gif",
    "image/jpeg" = "jpeg",
    "image/webp" = "webp",
    cli::cli_abort("Image type {content@type} is not supported by bedrock")
  )

  list(
    image = list(
      format = type,
      source = list(bytes = x@data)
    )
  )
}

# https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_DocumentBlock.html
method(as_json, list(ProviderAWSBedrock, ContentPDF)) <- function(
  provider,
  x,
  ...
) {
  list(
    document = list(
      #> This field is vulnerable to prompt injections, because the model
      #> might inadvertently interpret it as instructions. Therefore, we
      #> that you specify a neutral name.
      name = bedrock_document_name(),
      format = "pdf",
      source = list(bytes = x@data)
    )
  )
}

# https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ToolUseBlock.html
method(as_json, list(ProviderAWSBedrock, ContentToolRequest)) <- function(
  provider,
  x,
  ...
) {
  list(
    toolUse = list(
      toolUseId = x@id,
      name = x@name,
      input = x@arguments
    )
  )
}

# https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ToolResultBlock.html
method(as_json, list(ProviderAWSBedrock, ContentToolResult)) <- function(
  provider,
  x,
  ...
) {
  list(
    toolResult = list(
      toolUseId = x@request@id,
      content = list(list(text = tool_string(x))),
      status = if (tool_errored(x)) "error" else "success"
    )
  )
}

method(as_json, list(ProviderAWSBedrock, ToolDef)) <- function(
  provider,
  x,
  ...
) {
  list(
    toolSpec = list(
      name = x@name,
      description = x@description,
      inputSchema = list(json = compact(as_json(provider, x@arguments, ...)))
    )
  )
}

method(as_json, list(ProviderAWSBedrock, ContentThinking)) <- function(
  provider,
  x,
  ...
) {
  if (identical(x@thinking, "")) {
    return()
  }

  list(
    reasoningContent = list(
      reasoningText = list(
        text = x@thinking,
        signature = x@extra$signature
      )
    )
  )
}

# Helpers ----------------------------------------------------------------

as_bedrock_cache_point <- function(cache_point, model) {
  cache_point <- arg_match(
    cache_point,
    values = c("auto", "5m", "1h", "none")
  )
  if (cache_point != "auto") {
    return(cache_point)
  }
  supports_caching <-
    grepl("(^|\\.)anthropic\\.", model) || grepl("(^|\\.)amazon\\.nova", model)
  if (supports_caching) "5m" else "none"
}

bedrock_cache_point <- function(provider) {
  if (provider@cache_point == "none") {
    return(list())
  }
  cp <- list(type = "default")
  if (provider@cache_point != "5m") {
    cp$ttl <- provider@cache_point
  }
  list(list(cachePoint = cp))
}

paws_credentials <- function(
  profile,
  cache = aws_creds_cache(profile),
  reauth = FALSE
) {
  creds <- cache$get()
  if (reauth || is.null(creds) || creds$expiration < Sys.time()) {
    cache$clear()
    try_fetch(
      creds <- locate_aws_credentials(profile),
      error = function(cnd) {
        if (is_testing()) {
          testthat::skip("Failed to locate AWS credentials")
        }
        cli::cli_abort("No IAM credentials found.", parent = cnd)
      }
    )
    cache$set(creds)
  }
  creds
}

# Wrapper for paws.common::locate_credentials() so we can mock it in tests.
locate_aws_credentials <- function(profile) {
  paws.common::locate_credentials(profile)
}

aws_creds_cache <- function(profile) {
  credentials_cache(key = hash(c("aws", profile)))
}

bedrock_document_name <- local({
  i <- 1
  function() {
    i <<- i + 1
    paste0("document-", i)
  }
})

test_that("can make simple batch request", {
  chat <- chat_aws_bedrock_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_aws_bedrock_test("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

test_that("can list models", {
  test_models(models_aws_bedrock)
})

test_that("can set api args", {
  chat <- chat_aws_bedrock_test(
    api_args = list(inferenceConfig = list(maxTokens = 1))
  )
  result <- chat$chat("Who are the reindeer?")
  expect_true(nchar(result) < 10)
})

test_that("api args overwrite params", {
  chat <- chat_aws_bedrock_test(
    api_args = list(inferenceConfig = list(maxTokens = 1)),
    params = params(max_tokens = 100)
  )
  result <- chat$chat("Who are the reindeer?")
  expect_true(nchar(result) < 10)
})

test_that("handles errors", {
  chat <- chat_aws_bedrock_test(
    api_args = list(inferenceConfig = list(temperature = "hot"))
  )
  expect_snapshot(error = TRUE, {
    chat$chat("What is 1 + 1?", echo = FALSE)
    chat$chat("What is 1 + 1?", echo = TRUE)
  })
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  expect_snapshot(. <- chat_aws_bedrock())
})

test_that("supports tool calling", {
  chat_fun <- chat_aws_bedrock_test

  test_tools_simple(chat_fun)
})

test_that("can extract data", {
  chat_fun <- chat_aws_bedrock_test

  test_data_extraction(chat_fun)
})

test_that("can use images", {
  chat_fun <- chat_aws_bedrock_test

  test_images_inline(chat_fun)
  test_images_remote_error(chat_fun)
})

test_that("can use pdfs", {
  chat_fun <- chat_aws_bedrock_test

  test_pdf_local(chat_fun)
})

# Prompt caching ----------------------------------------------------------

has_cache_point <- function(content) {
  any(vapply(content, function(b) "cachePoint" %in% names(b), logical(1)))
}

block_types <- function(content) {
  vapply(content, function(b) names(b)[[1]], character(1))
}

test_that("as_bedrock_cache_point() resolves 'auto' for known models", {
  # Anthropic models (direct and cross-region)
  expect_equal(
    as_bedrock_cache_point("auto", "anthropic.claude-3-5-haiku-20241022-v1:0"),
    "5m"
  )
  expect_equal(
    as_bedrock_cache_point(
      "auto",
      "us.anthropic.claude-sonnet-4-5-20250929-v1:0"
    ),
    "5m"
  )

  # Amazon Nova models (direct and cross-region)
  expect_equal(as_bedrock_cache_point("auto", "amazon.nova-pro-v1:0"), "5m")
  expect_equal(as_bedrock_cache_point("auto", "us.amazon.nova-lite-v1:0"), "5m")

  # Unsupported models
  expect_equal(as_bedrock_cache_point("auto", "zai.glm-5"), "none")
  expect_equal(
    as_bedrock_cache_point("auto", "meta.llama3-1-8b-instruct-v1:0"),
    "none"
  )
})

test_that("as_bedrock_cache_point() passes through non-auto values", {
  expect_equal(as_bedrock_cache_point("5m", "zai.glm-5"), "5m")
  expect_equal(as_bedrock_cache_point("1h", "zai.glm-5"), "1h")
  expect_equal(
    as_bedrock_cache_point("none", "anthropic.claude-3-5-haiku-20241022-v1:0"),
    "none"
  )
})

test_that("cache points are inserted in last turn when cache is enabled", {
  provider <- test_aws_bedrock_provider(cache_point = "5m")

  # Non-last turn should not have a cache point
  result <- as_json(provider, UserTurn("Hello"), is_last = FALSE)
  expect_disjoint(block_types(result$content), "cachePoint")

  # Last turn should have a cache point appended
  result <- as_json(provider, UserTurn("Hello"), is_last = TRUE)
  last_block <- result$content[[length(result$content)]]
  expect_equal(last_block, list(cachePoint = list(type = "default")))
})

test_that("cache points are omitted when cache = 'none'", {
  provider <- test_aws_bedrock_provider(cache_point = "none")

  result <- as_json(provider, UserTurn("Hello"), is_last = TRUE)
  expect_disjoint(block_types(result$content), "cachePoint")
})

test_that("cache TTL is included for '1h' but not '5m'", {
  provider_5m <- test_aws_bedrock_provider(cache_point = "5m")
  provider_1h <- test_aws_bedrock_provider(cache_point = "1h")

  # 5m: cachePoint should be list(type = "default") with no ttl
  cp_5m <- bedrock_cache_point(provider_5m)
  expect_equal(cp_5m, list(list(cachePoint = list(type = "default"))))

  # 1h: cachePoint should include ttl = "1h"
  cp_1h <- bedrock_cache_point(provider_1h)
  expect_equal(
    cp_1h,
    list(list(cachePoint = list(type = "default", ttl = "1h")))
  )
})

test_that("cache point is only on the last turn in multi-turn conversations", {
  provider <- test_aws_bedrock_provider(cache_point = "5m")

  # Intermediate turns (is_last = FALSE) should not have cache points
  r1 <- as_json(provider, UserTurn("Hello"), is_last = FALSE)
  expect_false(has_cache_point(r1$content))

  r2 <- as_json(
    provider,
    AssistantTurn(list(ContentText("Hi there!"))),
    is_last = FALSE
  )
  expect_false(has_cache_point(r2$content))

  # Last turn should have a cache point
  r3 <- as_json(provider, UserTurn("How are you?"), is_last = TRUE)
  expect_true(has_cache_point(r3$content))
})

test_that("bedrock_cache_point() is added to the system prompt", {
  provider <- test_aws_bedrock_provider(cache_point = "5m")
  cp <- bedrock_cache_point(provider)

  # Mirrors the system prompt construction in chat_request()
  system <- c(
    list(list(text = "You are a helpful assistant.")),
    cp
  )

  expect_length(system, 2)
  expect_equal(system[[1]], list(text = "You are a helpful assistant."))
  expect_equal(system[[2]], list(cachePoint = list(type = "default")))

  # cache = "none" should not add a cache point
  provider_none <- test_aws_bedrock_provider(cache_point = "none")
  expect_equal(bedrock_cache_point(provider_none), list())
})

# Reasoning / thinking content --------------------------------------------

test_that("can use extended thinking", {
  vcr::local_cassette("aws-bedrock-thinking")

  chat <- chat_aws_bedrock_test(
    params = params(temperature = 1),
    api_args = list(
      additionalModelRequestFields = list(
        thinking = list(type = "enabled", budget_tokens = 4000)
      )
    ),
    model = "us.anthropic.claude-sonnet-4-5-20250929-v1:0"
  )
  resp <- chat$chat("Create a riddle for data scientists", echo = FALSE)

  contents <- chat$last_turn()@contents
  thinking <- Filter(\(x) S7::S7_inherits(x, ContentThinking), contents)
  expect_length(thinking, 1)
  expect_gt(nchar(thinking[[1]]@thinking), 0)
  expect_match(resp, "\\S")
})

# Provider idiosynchronies -----------------------------------------------

test_that("continues to work after whitespace only outputs (#376)", {
  chat <- chat_aws_bedrock_test()
  chat$chat("Respond with only two blank lines")
  expect_equal(
    chat$chat("What's 1+1? Just give me the number"),
    ellmer_output("2")
  )
})

# Auth --------------------------------------------------------------------

test_that("AWS credential caching works as expected", {
  # Mock AWS credentials for different profiles.
  local_mocked_bindings(
    locate_aws_credentials = function(profile) {
      if (!is.null(profile) && profile == "test") {
        list(
          access_key_id = "key1",
          secret_key = "secret1",
          expiration = Sys.time() + 3600
        )
      } else {
        list(
          access_key_id = "key2",
          secret_key = "secret2",
          expiration = Sys.time() + 3600
        )
      }
    }
  )

  creds1 <- paws_credentials(profile = "test", reauth = TRUE)
  creds2 <- paws_credentials(profile = NULL, reauth = TRUE)

  # Verify different credentials were returned.
  expect_false(identical(creds1, creds2))
  expect_equal(creds1$access_key_id, "key1")
  expect_equal(creds2$access_key_id, "key2")

  # Verify cached credentials match original ones.
  expect_identical(creds1, paws_credentials(profile = "test"))
  expect_identical(creds2, paws_credentials(profile = NULL))

  # Simulate a cache entry that has expired.
  creds_modified <- creds1
  creds_modified$expiration <- Sys.time() - 5
  aws_creds_cache(profile = "test")$set(creds_modified)

  # Ensure the new credentials have been updated.
  expect_false(identical(creds_modified, paws_credentials(profile = "test")))
  expect_false(identical(creds1, paws_credentials(profile = "test")))
  expect_false(identical(creds2, paws_credentials(profile = "test")))
})

#' @include provider-openai-compatible.R
#' @include content.R
NULL

# https://learn.microsoft.com/en-us/azure/ai-services/openai/reference#chat-completions

#' Chat with a model hosted on Azure OpenAI
#'
#' @description
#' The [Azure OpenAI server](https://azure.microsoft.com/en-us/products/ai-services/openai-service)
#' hosts a number of open source models as well as proprietary models
#' from OpenAI.
#'
#' Built on top of [chat_openai_compatible()].
#'
#' ## Authentication
#'
#' `chat_azure_openai()` supports API keys and the `credentials` parameter, but
#' it also makes use of:
#'
#' - Azure service principals (when the `AZURE_TENANT_ID`, `AZURE_CLIENT_ID`,
#'   and `AZURE_CLIENT_SECRET` environment variables are set).
#' - Interactive Entra ID authentication, like the Azure CLI.
#' - Viewer-based credentials on Posit Connect. Requires the \pkg{connectcreds}
#'   package.
#'
#' @param endpoint Azure OpenAI endpoint url with protocol and hostname, i.e.
#'  `https://{your-resource-name}.openai.azure.com`. Defaults to using the
#'   value of the `AZURE_OPENAI_ENDPOINT` environment variable.
#' @param model The **deployment id** for the model you want to use.
#' @param deployment_id `r lifecycle::badge("deprecated")` Use `model` instead.
#' @param api_version The API version to use.
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("AZURE_OPENAI_API_KEY")`
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @family chatbots
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_azure_openai(model = "gpt-4o-mini")
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_azure_openai <- function(
  endpoint = azure_endpoint(),
  model,
  params = NULL,
  api_version = NULL,
  system_prompt = NULL,
  api_key = NULL,
  credentials = NULL,
  api_args = list(),
  echo = c("none", "output", "all"),
  api_headers = character(),
  deployment_id = deprecated()
) {
  if (lifecycle::is_present(deployment_id)) {
    lifecycle::deprecate_warn(
      when = "0.4.0",
      what = "chat_azure_openai(deployment_id=)",
      with = "chat_azure_openai(model=)",
    )
    model <- deployment_id
  }
  check_string(endpoint)
  check_string(model)
  params <- params %||% params()
  api_version <- set_default(api_version, "2024-10-21")

  credentials <- as_credentials(
    "chat_azure_openai",
    default_azure_credentials(),
    credentials = credentials,
    api_key = api_key
  )

  echo <- check_echo(echo)

  provider <- ProviderAzureOpenAI(
    name = "Azure/OpenAI",
    base_url = paste0(endpoint, "/openai/deployments/", model),
    model = model,
    params = params,
    api_version = api_version,
    credentials = credentials,
    extra_args = api_args,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

chat_azure_openai_test <- function(
  system_prompt = NULL,
  params = NULL,
  ...,
  echo = "none"
) {
  credentials <- \() key_get("AZURE_OPENAI_API_KEY")
  default_params <- params(seed = 1014, temperature = 0)
  params <- modify_list(default_params, params %||% params())

  chat_azure_openai(
    ...,
    system_prompt = system_prompt,
    credentials = credentials,
    endpoint = "https://ai-hwickhamai260967855527.openai.azure.com",
    model = "gpt-4o-mini",
    params = params,
    echo = echo
  )
}

ProviderAzureOpenAI <- new_class(
  "ProviderAzureOpenAI",
  parent = ProviderOpenAICompatible,
  properties = list(
    api_version = prop_string()
  )
)

# https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/switching-endpoints#api-key
azure_endpoint <- function() {
  key_get("AZURE_OPENAI_ENDPOINT")
}

# https://learn.microsoft.com/en-us/azure/ai-services/openai/reference#chat-completions
method(base_request, ProviderAzureOpenAI) <- function(provider) {
  req <- request(provider@base_url)
  req <- ellmer_req_robustify(req)
  req <- ellmer_req_user_agent(req)
  req <- base_request_error(provider, req)

  req <- req_url_query(req, `api-version` = provider@api_version)
  req <- ellmer_req_credentials(req, provider@credentials(), "api-key")
  req
}

method(base_request_error, ProviderAzureOpenAI) <- function(provider, req) {
  req_error(req, body = function(resp) {
    error <- resp_body_json(resp)$error
    msg <- paste0(error$code, ": ", error$message)
    # Try to be helpful in the (common) case that the user or service
    # principal is missing the necessary role.
    # See: https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/role-based-access-control
    bad_rbac <- identical(
      error$message,
      "Principal does not have access to API/Operation."
    )
    if (bad_rbac) {
      msg <- c(
        "*" = msg,
        "i" = cli::format_inline(
          "Your user or service principal likely needs one of the following
        roles: {.emph Cognitive Services OpenAI User},
        {.emph Cognitive Services OpenAI Contributor}, or
        {.emph Cognitive Services Contributor}.",
          keep_whitespace = FALSE
        )
      )
    }
    msg
  })
}

default_azure_credentials <- function() {
  azure_openai_scope <- "https://cognitiveservices.azure.com/.default"

  # Detect viewer-based credentials from Posit Connect.
  if (has_connect_viewer_token(scope = azure_openai_scope)) {
    return(function() {
      token <- connectcreds::connect_viewer_token(scope = azure_openai_scope)
      list(Authorization = paste("Bearer", token$access_token))
    })
  }

  # Detect Azure service principals.
  tenant_id <- Sys.getenv("AZURE_TENANT_ID")
  client_id <- Sys.getenv("AZURE_CLIENT_ID")
  client_secret <- Sys.getenv("AZURE_CLIENT_SECRET")
  if (nchar(tenant_id) && nchar(client_id) && nchar(client_secret)) {
    # Service principals use an OAuth client credentials flow. We cache the token
    # so we don't need to perform this flow before each turn.
    client <- oauth_client(
      client_id,
      token_url = paste0(
        "https://login.microsoftonline.com/",
        tenant_id,
        "/oauth2/v2.0/token"
      ),
      secret = client_secret,
      auth = "body",
      name = "ellmer-azure-sp"
    )
    return(function() {
      token <- oauth_token_cached(
        client,
        oauth_flow_client_credentials,
        flow_params = list(scope = azure_openai_scope),
        # Don't use the cached token when testing.
        reauth = is_testing()
      )
      list(Authorization = paste("Bearer", token$access_token))
    })
  }

  # If we have an API key, include it in the credentials.
  api_key <- Sys.getenv("AZURE_OPENAI_API_KEY")
  if (nchar(api_key)) {
    return(\() api_key)
  }

  # Masquerade as the Azure CLI.
  client_id <- "04b07795-8ddb-461a-bbee-02f9e1bf7b46"
  if (is_interactive() && !is_hosted_session()) {
    client <- oauth_client(
      client_id,
      token_url = "https://login.microsoftonline.com/common/oauth2/v2.0/token",
      secret = "",
      auth = "body",
      name = paste0("ellmer-", client_id)
    )
    return(function() {
      token <- oauth_token_cached(
        client,
        oauth_flow_auth_code,
        flow_params = list(
          auth_url = "https://login.microsoftonline.com/common/oauth2/v2.0/authorize",
          scope = paste(azure_openai_scope, "offline_access"),
          redirect_uri = "http://localhost:8400",
          auth_params = list(prompt = "select_account")
        )
      )
      list(Authorization = paste("Bearer", token$access_token))
    })
  }

  if (is_testing()) {
    testthat::skip("no Azure credentials available")
  }

  cli::cli_abort("No Azure credentials are available.")
}

#' Chat with a model hosted on PortkeyAI
#'
#' @description
#' [PortkeyAI](https://portkey.ai/docs/product/ai-gateway/universal-api)
#' provides an interface (AI Gateway) to connect through its Universal API to a
#' variety of LLMs providers via a single endpoint.
#'
#' @family chatbots
#' @param model The model name, e.g. `@my-provider/my-model`.
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("PORTKEY_API_KEY")`
#' @param virtual_key `r lifecycle::badge("deprecated")`.
#'   Portkey now recommend supplying the model provider
#'   (formerly known as the `virtual_key`), in the model name, e.g.
#'   `@my-provider/my-model`. See
#'   <https://portkey.ai/docs/support/upgrade-to-model-catalog> for details.
#'
#'   For backward compatibility, the `PORTKEY_VIRTUAL_KEY` env var is still used
#'   if the model doesn't include a provider.
#' @export
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_portkey()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_portkey <- function(
  model,
  system_prompt = NULL,
  base_url = "https://api.portkey.ai/v1",
  api_key = NULL,
  credentials = NULL,
  virtual_key = deprecated(),
  params = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
) {
  check_string(model)
  echo <- check_echo(echo)

  if (lifecycle::is_present(virtual_key)) {
    lifecycle::deprecate_warn(
      when = "0.4.0",
      what = "chat_portkey(virtual_key=)",
      with = "chat_portkey(model=)",
    )
    check_string(virtual_key, allow_null = TRUE)
  } else {
    virtual_key <- NULL
  }

  # For backward compatibility
  if (!grepl("^@", model)) {
    virtual_key <- virtual_key %||% key_get("PORTKEY_VIRTUAL_KEY")
    model <- paste0("@", virtual_key, "/", model)
  }

  credentials <- as_credentials(
    "chat_portkey",
    function() portkey_key(),
    credentials = credentials,
    api_key = api_key
  )

  params <- params %||% params()
  provider <- ProviderPortkeyAI(
    name = "PortkeyAI",
    base_url = base_url,
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

chat_portkey_test <- function(
  ...,
  model = "@open-ai-virtual-7f0dcd/gpt-4.1-nano",
  params = NULL,
  echo = "none"
) {
  params <- params %||% params()
  params$temperature <- params$temperature %||% 0

  chat_portkey(model = model, params = params, ..., echo = echo)
}

ProviderPortkeyAI <- new_class(
  "ProviderPortkeyAI",
  parent = ProviderOpenAICompatible
)

portkey_key <- function() {
  key_get("PORTKEY_API_KEY")
}

method(base_request, ProviderPortkeyAI) <- function(provider) {
  req <- request(provider@base_url)

  req <- ellmer_req_credentials(
    req,
    provider@credentials(),
    "x-portkey-api-key"
  )
  req <- ellmer_req_robustify(req)
  req <- ellmer_req_user_agent(req)
  req <- base_request_error(provider, req)
  req
}


#' @export
#' @rdname chat_portkey
models_portkey <- function(
  base_url = "https://api.portkey.ai/v1",
  api_key = portkey_key()
) {
  provider <- ProviderPortkeyAI(
    name = "PortkeyAI",
    model = "",
    base_url = base_url,
    credentials = function() api_key
  )

  req <- base_request(provider)
  req <- req_url_path_append(req, "/models")
  resp <- req_perform(req)

  json <- resp_body_json(resp)

  id <- map_chr(json$data, "[[", "id")
  slug <- map_chr(json$data, "[[", "slug")

  df <- data.frame(
    id = id,
    slug = slug
  )
  df
}

prop_string <- function(default = NULL, allow_null = FALSE, allow_na = FALSE) {
  force(allow_null)
  force(allow_na)

  new_property(
    class = if (allow_null) NULL | class_character else class_character,
    default = if (is.null(default) && !allow_null) {
      quote(stop("Required"))
    } else {
      default
    },
    validator = function(value) {
      if (allow_null && is.null(value)) {
        return()
      }

      if (length(value) != 1) {
        paste0("must be a single string, not ", obj_type_friendly(value), ".")
      } else if (!allow_na && is.na(value)) {
        "must not be missing."
      }
    }
  )
}

prop_bool <- function(default, allow_null = FALSE, allow_na = FALSE) {
  force(allow_null)
  force(allow_na)

  new_property(
    class = if (allow_null) NULL | class_logical else class_logical,
    default = default,
    validator = function(value) {
      if (allow_null && is.null(value)) {
        return()
      }

      if (length(value) != 1) {
        if (allow_na) {
          paste0(
            "must be a single TRUE or FALSE, not ",
            obj_type_friendly(value),
            "."
          )
        } else {
          paste0(
            "must be a single TRUE, FALSE or NA, not ",
            obj_type_friendly(value),
            "."
          )
        }
      } else if (!allow_na && is.na(value)) {
        paste0("must be a TRUE or FALSE, not NA.")
      }
    }
  )
}

prop_list_of <- function(class, names = c("any", "all", "none")) {
  force(class)
  names <- arg_match(names)

  new_property(
    class = class_list,
    validator = function(value) {
      for (i in seq_along(value)) {
        val <- value[[i]]
        if (!S7_inherits(val, class)) {
          return(paste0(
            "must be a list of <",
            class@name,
            ">s. ",
            "Element ",
            i,
            " is ",
            obj_type_friendly(val),
            "."
          ))
        }
      }
      if (names == "all" && any(names2(value) == "")) {
        "must be a named list."
      } else if (names == "none" && any(names2(value) != "")) {
        "must be an unnamed list."
      }
    }
  )
}

prop_number_whole <- function(
  default = NULL,
  min = NULL,
  max = NULL,
  allow_null = FALSE,
  allow_na = FALSE
) {
  force(allow_null)
  force(allow_na)

  new_property(
    class = if (allow_null) NULL | class_double else class_double,
    default = default,
    validator = function(value) {
      if (allow_null && is.null(value)) {
        return()
      }

      if (length(value) != 1) {
        paste0("must be a whole number, not ", obj_type_friendly(value), ".")
      } else if (!allow_na && is.na(value)) {
        "must not be missing."
      } else if (value != trunc(value)) {
        paste0("must be a whole number, not ", obj_type_friendly(value), ".")
      } else if (!is.null(min) && value < min) {
        paste0("must be at least ", min, ", not ", value, ".")
      } else if (!is.null(max) && value > max) {
        paste0("must be at most ", max, ", not ", value, ".")
      }
    }
  )
}

prop_number_decimal <- function(
  default = NULL,
  min = NULL,
  max = NULL,
  allow_null = FALSE,
  allow_na = FALSE,
  allow_infinite = FALSE
) {
  force(allow_null)
  force(allow_na)

  new_property(
    class = if (allow_null) NULL | class_double else class_double,
    default = default,
    validator = function(value) {
      if (allow_null && is.null(value)) {
        return()
      }

      if (length(value) != 1) {
        paste0("must be a whole number, not ", obj_type_friendly(value), ".")
      } else if (!allow_infinite && is.infinite(value)) {
        paste0("must be finite, not ", value, ".")
      } else if (!is.null(min) && value < min) {
        paste0("must be at least ", min, ", not ", value, ".")
      } else if (!is.null(max) && value > max) {
        paste0("must be at most ", max, ", not ", value, ".")
      } else if (!allow_na && is.na(value)) {
        "must not be missing."
      }
    }
  )
}

# Weakrefs are not serialised, so we can use them to ensure that a value
# is never recorded by saveRDS() and friends.
prop_redacted <- function(name, default = NULL, allow_null = FALSE) {
  force(allow_null)

  # Tie all weakref values to the enclosing environment of `new_property()`
  # This ensures that all the weakrefs will go out of scope and be gc'd if
  # the class is deleted.
  key <- environment()

  new_property(
    name = name,
    default = if (is.null(default) && !allow_null) {
      quote(stop("Required"))
    } else {
      default
    },
    getter = function(self) {
      wref_value(prop(self, name))
    },
    setter = function(self, value) {
      check_string(value, allow_null = allow_null)

      prop(self, name) <- new_weakref(key, value)
      self
    }
  )
}

test_that("prop_whole_number validates inputs", {
  check_prop <- function(...) {
    new_class(
      "class",
      properties = list(prop = prop_number_whole(...)),
      package = NULL
    )
  }
  expect_snapshot(error = TRUE, {
    check_prop()("x")
    check_prop()(c(1, 2))
    check_prop()(1.5)
    check_prop(min = 1)(0)
    check_prop(max = -1)(0)
    check_prop()(NA_real_)
  })
})

test_that("prop_whole_decimal validates inputs", {
  check_prop <- function(...) {
    new_class(
      "class",
      properties = list(prop = prop_number_decimal(...)),
      package = NULL
    )
  }
  expect_snapshot(error = TRUE, {
    check_prop()("x")
    check_prop()(c(1, 2))
    check_prop()(1.5)
    check_prop(min = 1)(0)
    check_prop(max = -1)(0)
    check_prop(max = -1)(Inf)
    check_prop()(NA_real_)
  })
})


test_that("redacted values aren't saved to disk", {
  Test <- new_class("Test", properties = list(prop_redacted("redacted")))

  # Can get and set redacted values
  test <- Test(redacted = "secret")
  expect_equal(test@redacted, "secret")
  test@redacted <- "new secret"
  expect_equal(test@redacted, "new secret")

  # But can't save it to disk
  path <- withr::local_tempfile()
  saveRDS(test, path)
  test <- readRDS(path)
  expect_equal(test@redacted, NULL)
})

test_that("redacted values are instance specific", {
  # Reassure myself about the semantics of the weakrefs: even though the
  # key is shared across all instances of the same class, we are still creating
  # individual weakrefs for each instance.

  Test <- new_class("Test", properties = list(prop_redacted("redacted")))
  test1 <- Test(redacted = "secret1")
  test2 <- Test(redacted = "secret2")
  expect_equal(test1@redacted, "secret1")
  expect_equal(test2@redacted, "secret2")

  test1@redacted <- "secret1a"
  test2@redacted <- "secret2a"
  expect_equal(test1@redacted, "secret1a")
  expect_equal(test2@redacted, "secret2a")
})

#' Submit multiple chats in parallel
#'
#' @description
#' If you have multiple prompts, you can submit them in parallel. This is
#' typically considerably faster than submitting them in sequence, especially
#' with Gemini and OpenAI.
#'
#' If you're using [chat_openai()] or [chat_anthropic()] and you're willing
#' to wait longer, you might want to use [batch_chat()] instead, as it comes
#' with a 50% discount in return for taking up to 24 hours.
#'
#' @param chat A chat object created by a `chat_` function, or a
#'   string passed to [chat()].
#' @param prompts A vector created by [interpolate()] or a list
#'   of character vectors.
#' @param max_active The maximum number of simultaneous requests to send.
#'
#'   For [chat_anthropic()], note that the number of active connections is
#'   limited primarily by the output tokens per minute limit (OTPM) which is
#'   estimated from the `max_tokens` parameter, which defaults to 4096. That
#'   means if your usage tier limits you to 16,000 OTPM, you should either set
#'   `max_active = 4` (16,000 / 4096) to decrease the number of active
#'   connections or use [params()] in `chat_anthropic()` to decrease
#'   `max_tokens`.
#' @param rpm Maximum number of requests per minute.
#' @param on_error What to do when a request fails. One of:
#'   * `"return"` (the default): stop processing new requests, wait for
#'      in flight requests to finish, then return.
#'   * `"continue"`: keep going, performing every request.
#'   * `"stop"`: stop processing and throw an error.
#' @returns
#' For `parallel_chat()`, a list with one element for each prompt. Each element
#' is either a [Chat] object (if successful), a `NULL` (if the request wasn't
#' performed) or an error object (if it failed).
#'
#' For `parallel_chat_text()`, a character vector with one element for each
#' prompt. Requests that weren't succesful get an `NA`.
#'
#' For `parallel_chat_structured()`, a single structured data object with one
#' element for each prompt. Typically, when `type` is an object, this will
#' be a tibble with one row for each prompt, and one column for each
#' property. If the output is a data frame, and some requests error,
#' an `.error` column will be added with the error objects.
#' @export
#' @examples
#' \dontshow{ellmer:::vcr_example_start("parallel_chat")}
#' chat <- chat_openai()
#'
#' # Chat ----------------------------------------------------------------------
#' country <- c("Canada", "New Zealand", "Jamaica", "United States")
#' prompts <- interpolate("What's the capital of {{country}}?")
#' parallel_chat(chat, prompts)
#'
#' # Structured data -----------------------------------------------------------
#' prompts <- list(
#'   "I go by Alex. 42 years on this planet and counting.",
#'   "Pleased to meet you! I'm Jamal, age 27.",
#'   "They call me Li Wei. Nineteen years young.",
#'   "Fatima here. Just celebrated my 35th birthday last week.",
#'   "The name's Robert - 51 years old and proud of it.",
#'   "Kwame here - just hit the big 5-0 this year."
#' )
#' type_person <- type_object(name = type_string(), age = type_number())
#' parallel_chat_structured(chat, prompts, type_person)
#' \dontshow{ellmer:::vcr_example_end()}
parallel_chat <- function(
  chat,
  prompts,
  max_active = 10,
  rpm = 500,
  on_error = c("return", "continue", "stop")
) {
  chat <- as_chat(chat)
  on_error <- arg_match(on_error)

  my_parallel_turns <- function(conversations) {
    parallel_turns(
      provider = chat$get_provider(),
      conversations = conversations,
      tools = chat$get_tools(),
      max_active = max_active,
      rpm = rpm,
      on_error = on_error
    )
  }

  # First build up list of cumulative conversations
  user_turns <- as_user_turns(prompts)
  existing <- chat$get_turns(include_system_prompt = TRUE)
  conversations <- append_turns(list(existing), user_turns)

  # Now get the assistant's response
  assistant_turns <- my_parallel_turns(conversations)

  is_ok <- !map_lgl(assistant_turns, turn_failed)
  repeat {
    if (!any(is_ok)) {
      break
    }
    conversations[is_ok] <- append_turns(
      conversations[is_ok],
      assistant_turns[is_ok]
    )

    tool_turns <- map(assistant_turns[is_ok], function(turn) {
      turn <- match_tools(turn, tools = chat$get_tools())
      tool_results <- coro::collect(invoke_tools(turn))
      tool_results_as_turn(tool_results)
    })
    needs_iter <- !map_lgl(tool_turns, is.null)
    if (!any(needs_iter)) {
      break
    }

    conversations[is_ok][needs_iter] <- append_turns(
      conversations[is_ok][needs_iter],
      tool_turns[needs_iter]
    )

    assistant_turns <- vector("list", length(user_turns))
    assistant_turns[needs_iter] <- my_parallel_turns(conversations[needs_iter])
    is_ok[needs_iter] <- !map_lgl(assistant_turns[needs_iter], turn_failed)
  }

  map(seq_along(conversations), function(i) {
    if (is_ok[[i]]) {
      turns <- conversations[[i]]
      log_turns(chat$get_provider(), turns)
      chat$clone()$set_turns(turns)
    } else {
      assistant_turns[[i]]
    }
  })
}

#' @rdname parallel_chat
#' @export
parallel_chat_text <- function(
  chat,
  prompts,
  max_active = 10,
  rpm = 500,
  on_error = c("return", "continue", "stop")
) {
  chat <- as_chat(chat)
  on_error <- arg_match(on_error)

  chats <- parallel_chat(
    chat,
    prompts,
    max_active = max_active,
    rpm = rpm,
    on_error = on_error
  )

  is_ok <- !map_lgl(chats, turn_failed)
  out <- rep(NA_character_, length(prompts))
  out[is_ok] <- map_chr(chats[is_ok], \(chat) chat$last_turn()@text)
  out
}

#' @param type A type specification for the extracted data. Should be
#'   created with a [`type_()`][type_boolean] function.
#' @param convert If `TRUE`, automatically convert from JSON lists to R
#'   data types using the schema. This typically works best when `type` is
#'   [type_object()] as this will give you a data frame with one column for
#'   each property. If `FALSE`, returns a list.
#' @param include_tokens If `TRUE`, and the result is a data frame, will
#'   add `input_tokens` and `output_tokens` columns giving the total input
#'   and output tokens for each prompt.
#' @param include_cost If `TRUE`, and the result is a data frame, will
#'   add `cost` column giving the cost of each prompt.
#' @export
#' @rdname parallel_chat
parallel_chat_structured <- function(
  chat,
  prompts,
  type,
  convert = TRUE,
  include_tokens = FALSE,
  include_cost = FALSE,
  max_active = 10,
  rpm = 500,
  on_error = c("return", "continue", "stop")
) {
  chat <- as_chat(chat)
  turns <- as_user_turns(prompts)
  check_bool(convert)
  on_error <- arg_match(on_error)

  provider <- chat$get_provider()
  needs_wrapper <- type_needs_wrapper(type, provider)

  # First build up list of cumulative conversations
  user_turns <- as_user_turns(prompts)
  existing <- chat$get_turns(include_system_prompt = TRUE)
  conversations <- append_turns(list(existing), user_turns)

  turns <- parallel_turns(
    provider = provider,
    conversations = conversations,
    tools = chat$get_tools(),
    type = wrap_type_if_needed(type, needs_wrapper),
    max_active = max_active,
    rpm = rpm,
    on_error = on_error
  )
  log_turns(provider, turns)

  multi_convert(
    provider,
    turns,
    type,
    convert = convert,
    include_tokens = include_tokens,
    include_cost = include_cost
  )
}

multi_convert <- function(
  provider,
  turns,
  type,
  convert = TRUE,
  include_tokens = FALSE,
  include_cost = FALSE
) {
  needs_wrapper <- type_needs_wrapper(type, provider)

  rows <- map(turns, \(turn) {
    if (turn_failed(turn)) {
      NULL
    } else {
      safely(
        extract_data(
          turn = turn,
          type = wrap_type_if_needed(type, needs_wrapper),
          convert = FALSE,
          needs_wrapper = needs_wrapper
        )
      )
    }
  })

  is_err <- map_lgl(rows, \(x) !is.null(x$error))
  n_error <- sum(is_err)
  if (n_error > 0) {
    msgs <- map(rows[is_err], \(x) conditionMessage(x$error))
    errors <- paste0(" * ", seq_along(turns)[is_err], ": ", msgs)

    cli::cli_warn(c(
      "Failed to extract data from {n_error}/{length(turns)} turns",
      cli_escape(errors)
    ))
  }
  # convert_from_type() will convert NULL to required type
  row_data <- map(rows, \(x) x$result)

  if (convert) {
    out <- convert_from_type(row_data, type_array(type))
  } else {
    out <- row_data
  }

  if (is.data.frame(out)) {
    is_error <- map_lgl(turns, turn_failed)
    if (any(is_error)) {
      errors <- vector("list", length(turns))
      errors[is_error] <- turns[is_error]
      out$.error <- errors
    }

    if (include_tokens) {
      tokens <- map_tokens(turns, \(turn) {
        if (turn_failed(turn)) c(0L, 0L, 0L) else turn@tokens
      })
      out$input_tokens <- tokens[, 1]
      out$output_tokens <- tokens[, 2]
      out$cached_input_tokens <- tokens[, 3]
    }

    if (include_cost) {
      out$cost <- map_dbl(turns, \(turn) {
        if (turn_failed(turn)) 0 else turn@cost
      })
    }
  }
  out
}

append_turns <- function(old_turns, new_turns) {
  map2(old_turns, new_turns, function(old, new) {
    if (is.null(new)) {
      old
    } else {
      c(old, list(new))
    }
  })
}

turn_failed <- function(turn) {
  is.null(turn) || inherits(turn, "error")
}

parallel_turns <- function(
  provider,
  conversations,
  tools,
  type = NULL,
  max_active = 10,
  rpm = 60,
  on_error = "return"
) {
  reqs <- map(conversations, function(turns) {
    chat_request(
      provider = provider,
      turns = turns,
      type = type,
      tools = tools,
      stream = FALSE
    )
  })
  reqs <- map(reqs, function(req) {
    req_throttle(req, capacity = rpm, fill_time_s = 60)
  })

  # Returns list where elements NULL, an error, or a response
  resps <- req_perform_parallel(
    reqs,
    max_active = max_active,
    on_error = on_error
  )

  is_absent <- map_lgl(resps, is.null)
  if (any(is_absent)) {
    n <- sum(is_absent)
    cli::cli_warn("{n} request{?s} did not complete.")
  }

  is_error <- map_lgl(resps, inherits, "error")
  if (any(is_error)) {
    n <- sum(is_error)
    cli::cli_warn("{n} request{?s} errored.")
  }

  map(resps, function(resp) {
    if (is.null(resp)) {
      NULL
    } else if (inherits(resp, "error")) {
      resp
    } else {
      json <- resp_body_json(resp)
      turn <- value_turn(provider, json, has_type = !is.null(type))
      turn@duration <- resp_timing(resp)[["total"]] %||% NA_real_
      turn
    }
  })
}

# Helpers -----------------------------------------------------------------

safely <- function(code) {
  tryCatch(
    list(result = code, error = NULL),
    error = function(cnd) {
      list(result = NULL, error = cnd)
    }
  )
}

retry_test <- function(code, retries = 1) {
  code <- enquo(code)

  i <- 1
  while (i <= retries) {
    tryCatch(
      {
        return(eval(get_expr(code), get_env(code)))
        break
      },
      expectation_failure = function(cnd) NULL
    )
    cli::cli_inform(c(i = "Retry {i}"))
    i <- i + 1
  }

  eval(get_expr(code), get_env(code))
}

# Params -----------------------------------------------------------------

test_params_stop <- function(chat_fun) {
  chat <- chat_fun(params = params(stop_sequences = "cool"))
  out <- chat$chat("Repeat the following phrase: Dogs are cool")
  expect_no_match(out, "cool")
}

# Tool calls -------------------------------------------------------------

test_tools_simple <- function(chat_fun) {
  chat <- chat_fun(
    system_prompt = "Always use a tool to answer. Reply with 'It is ____.'."
  )
  chat$register_tool(tool(
    function() "2024-01-01",
    name = "current_date",
    description = "Return the current date"
  ))
  chat$register_tool(tool(
    function() "February",
    name = "current_month",
    description = "Return the full name of the current month"
  ))

  result <- chat$chat("What's the current date in Y-M-D format?")
  expect_match(result, "2024-01-01")

  result <- chat$chat("What month is it? Provide the full name")
  expect_match(result, "February")
}

test_tool_image <- function(chat_fun) {
  # has a subtle dependency on imagemagick
  skip_on_cran()

  chat <- chat_fun()
  chat$register_tool(tool(
    \() content_image_file(system.file("smol-animal.jpg", package = "ellmer")),
    name = "draw_animal",
    description = "Draw a cute animal"
  ))
  chat$chat("Draw a picture of a cute animal")
  expect_match(chat$chat("What sort of animal is that?"), "kitten|cat")
}

# Data extraction --------------------------------------------------------

test_data_extraction <- function(chat_fun) {
  article_summary <- type_object(
    "Summary of the article. Preserve existing case.",
    title = type_string("Content title"),
    author = type_string("Name of the author")
  )

  prompt <- "
    # Apples are tasty
    By Hadley Wickham

    Apples are delicious and tasty and I like to eat them.
    Except for red delicious, that is. They are NOT delicious.
  "

  chat <- chat_fun()
  data <- chat$chat_structured(prompt, type = article_summary)
  expect_mapequal(
    data,
    list(title = "Apples are tasty", author = "Hadley Wickham")
  )

  # Check that we can do it again
  data <- chat$chat_structured(prompt, type = article_summary)
  expect_mapequal(
    data,
    list(title = "Apples are tasty", author = "Hadley Wickham")
  )
}

# Built-in tools ---------------------------------------------------------

test_tool_web_fetch <- function(chat_fun, tool) {
  chat <- chat_fun()
  chat$register_tool(tool)

  url <- "https://rvest.tidyverse.org/articles/starwars.html"
  expect_match(
    chat$chat(paste0("What's the first movie listed on ", url, "?")),
    "The Phantom Menace"
  )
  expect_match(chat$chat("Who directed it?"), "George Lucas")
}

test_tool_web_search <- function(chat_fun, tool, hint = NULL) {
  chat <- chat_fun()
  chat$register_tool(tool)

  result <- chat$chat(c(
    "When was ggplot2 1.0.0 released to CRAN?",
    "Answer in YYYY-MM-DD format.",
    hint
  ))
  # for openAI
  result <- gsub("\u2011", "-", result, fixed = TRUE)
  expect_match(result, "2014-05-21")
  expect_match(chat$chat("What month was that?"), "May")
}

# Images -----------------------------------------------------------------

test_images_inline <- function(chat_fun, test_shape = TRUE) {
  # has a subtle dependency on imagemagick
  skip_on_cran()

  chat <- chat_fun()
  response <- chat$chat(
    "What's in this image? (Be sure to mention the outside shape)",
    content_image_file(system.file("httr2.png", package = "ellmer"))
  )
  if (test_shape) {
    expect_match(response, "hex")
  }
  expect_match(response, "baseball")
}

test_images_remote <- function(chat_fun, test_shape = TRUE) {
  chat <- chat_fun()
  response <- chat$chat(
    "What's in this image? (Be sure to mention the outside shape)",
    content_image_url("https://httr2.r-lib.org/logo.png")
  )
  if (test_shape) {
    expect_match(response, "hex")
  }
  expect_match(response, "baseball")
}

test_images_remote_error <- function(chat_fun) {
  chat <- chat_fun()

  image_remote <- content_image_url("https://httr2.r-lib.org/logo.png")
  expect_snapshot(
    . <- chat$chat("What's in this image?", image_remote),
    error = TRUE
  )
  expect_length(chat$get_turns(), 0)
}

# PDF ---------------------------------------------------------------------

test_pdf_local <- function(chat_fun) {
  chat <- chat_fun()
  response <- chat$chat(
    "What's the title of this document?",
    content_pdf_file(test_path("apples.pdf"))
  )
  expect_match(response, "Apples are tasty")
  expect_match(
    chat$chat("What apple is not tasty?"),
    "red delicious",
    ignore.case = TRUE
  )
}

# Models ------------------------------------------------------------------

test_models <- function(models_fun) {
  models <- models_fun()
  expect_gt(nrow(models), 0)
  expect_s3_class(models, "data.frame")
  expect_contains(names(models), "id")
}

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_azure_openai_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_azure_openai_test("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  withr::local_envvar(AZURE_OPENAI_API_KEY = "key")
  expect_snapshot(. <- chat_azure_openai("endpoint", "deployment_id"))
})

test_that("supports standard parameters", {
  chat_fun <- chat_azure_openai_test

  test_params_stop(chat_fun)
})

test_that("supports tool calling", {
  chat_fun <- chat_azure_openai_test

  test_tools_simple(chat_fun)
})

test_that("can extract data", {
  chat_fun <- chat_azure_openai_test

  test_data_extraction(chat_fun)
})

test_that("can use images", {
  skip("Run manually; 24 hour rate limit")
  chat_fun <- chat_azure_openai_test

  test_images_inline(chat_fun)
  test_images_remote(chat_fun)
})

# Authentication --------------------------------------------------------------

test_that("Azure request headers are generated correctly", {
  turn <- UserTurn(
    contents = list(ContentText("What is 1 + 1?"))
  )
  deployment_id <- "gpt-4o-mini"
  base_url <- paste0(
    "https://ai-hwickhamai260967855527.openai.azure.com/openai/deployments/",
    deployment_id
  )

  # API key.
  p <- ProviderAzureOpenAI(
    name = "Azure",
    base_url = base_url,
    model = deployment_id,
    api_version = "2024-06-01",
    credentials = \() "key"
  )
  req <- chat_request(p, FALSE, list(turn))
  expect_snapshot(str(req_get_headers(req, "reveal")))
})

test_that("service principal authentication requests look correct", {
  withr::local_envvar(
    AZURE_TENANT_ID = "aaaa0a0a-bb1b-cc2c-dd3d-eeeeee4e4e4e",
    AZURE_CLIENT_ID = "id",
    AZURE_CLIENT_SECRET = "secret"
  )
  local_mocked_responses(function(req) {
    # Snapshot relevant fields of the outgoing request.
    expect_snapshot(str(request_summary(req)))
    response_json(body = list(access_token = "token"))
  })
  source <- default_azure_credentials()
  expect_equal(source(), list(Authorization = "Bearer token"))
})

test_that("tokens can be requested from a Connect server", {
  skip_if_not_installed("connectcreds")

  connectcreds::local_mocked_connect_responses(token = "token")
  credentials <- default_azure_credentials()
  expect_equal(credentials(), list(Authorization = "Bearer token"))
})

test_that("can make simple batch request", {
  vcr::local_cassette("anthropic-batch")

  chat <- chat_anthropic_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?")
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_anthropic_test(
    "Be as terse as possible; no punctuation"
  )
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

test_that("can list models", {
  vcr::local_cassette("anthropic-list-models")

  test_models(models_anthropic)
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  expect_snapshot(. <- chat_anthropic())
})

test_that("supports standard parameters", {
  vcr::local_cassette("anthropic-standard-params")
  test_params_stop(chat_anthropic_test)
})

test_that("supports tool calling", {
  vcr::local_cassette("anthropic-tool")
  chat_fun <- chat_anthropic_test

  test_tools_simple(chat_fun)
})

test_that("can fetch web pages", {
  vcr::local_cassette("anthropic-web-fetch")
  chat_fun <- \(...) {
    chat_anthropic_test(..., beta_headers = "web-fetch-2025-09-10")
  }
  test_tool_web_fetch(chat_fun, claude_tool_web_fetch())
})

test_that("can search web pages", {
  vcr::local_cassette("anthropic-web-search")
  chat_fun <- \(...) chat_anthropic_test(...)
  test_tool_web_search(chat_fun, claude_tool_web_search())
})

test_that("tools can return images", {
  vcr::local_cassette("anthropic-tool-image")
  chat_fun <- chat_anthropic_test
  test_tool_image(chat_fun)
})

test_that("can extract data", {
  vcr::local_cassette("anthropic-structured-data")
  chat_fun <- chat_anthropic_test

  test_data_extraction(chat_fun)
})

test_that("can use images", {
  vcr::local_cassette("anthropic-images")
  chat_fun <- chat_anthropic_test

  test_images_inline(chat_fun)
  test_images_remote(chat_fun)
})

test_that("can use pdfs", {
  vcr::local_cassette("anthropic-pdfs")
  chat_fun <- chat_anthropic_test

  test_pdf_local(chat_fun)
})

# Custom features --------------------------------------------------------

test_that("can set beta headers", {
  chat <- chat_anthropic_test(beta_headers = c("a", "b"))
  req <- chat_request(chat$get_provider())
  headers <- req_get_headers(req)
  expect_equal(headers$`anthropic-beta`, "a,b")
})

test_that("continues to work after whitespace only outputs (#376)", {
  vcr::local_cassette("anthropic-whitespace")

  chat <- chat_anthropic_test()
  chat$chat("Respond with only two blank lines")
  expect_equal(
    chat$chat("What's 1+1? Just give me the number"),
    ellmer_output("2")
  )
})

test_that("can match prices for some common models", {
  provider <- chat_anthropic_test()$get_provider()

  expect_true(has_cost(provider, "claude-sonnet-4-20250514"))
})

test_that("removes empty final chat messages", {
  chat <- chat_anthropic_test()
  chat$set_turns(
    list(
      UserTurn("Don't say anything"),
      AssistantTurn()
    )
  )

  turns_json <- as_json(chat$get_provider(), chat$get_turns())

  expect_length(turns_json, 1)
  expect_equal(turns_json[[1]]$role, "user")
  expect_equal(
    turns_json[[1]]$content,
    list(list(type = "text", text = "Don't say anything"))
  )
})

test_that("batch chat works", {
  chat <- chat_anthropic_test(system_prompt = "Answer with just the city name")

  prompts <- list(
    "What's the capital of Iowa?",
    "What's the capital of New York?",
    "What's the capital of California?",
    "What's the capital of Texas?"
  )

  out <- batch_chat_text(
    chat,
    prompts,
    path = test_path("batch/state-capitals-anthropic.json")
  )
  expect_equal(out, c("Des Moines", "Albany", "Sacramento", "Austin"))
})

test_that("value_turn() parses server_tool_use input from JSON string", {
  provider <- ProviderAnthropic(
    name = "Anthropic",
    base_url = "https://api.anthropic.com/v1",
    model = "claude-sonnet-4-20250514",
    params = list(),
    extra_args = list(),
    extra_headers = character(),
    credentials = NULL,
    beta_headers = character(),
    cache = ""
  )

  result <- list(
    content = list(
      list(
        type = "server_tool_use",
        id = "srvtoolu_1",
        name = "web_search",
        input = '{"query":"test search"}'
      )
    ),
    stop_reason = "end_turn",
    usage = list(
      input_tokens = 10,
      output_tokens = 5,
      cache_creation_input_tokens = 0,
      cache_read_input_tokens = 0
    )
  )

  turn <- value_turn(provider, result)
  search_content <- turn@contents[[1]]
  expect_s7_class(search_content, ContentToolRequestSearch)
  expect_equal(search_content@query, "test search")
})

test_that("value_turn() parses server_tool_use web_fetch input from JSON string", {
  provider <- ProviderAnthropic(
    name = "Anthropic",
    base_url = "https://api.anthropic.com/v1",
    model = "claude-sonnet-4-20250514",
    params = list(),
    extra_args = list(),
    extra_headers = character(),
    credentials = NULL,
    beta_headers = character(),
    cache = ""
  )

  result <- list(
    content = list(
      list(
        type = "server_tool_use",
        id = "srvtoolu_2",
        name = "web_fetch",
        input = '{"url":"https://example.com"}'
      )
    ),
    stop_reason = "end_turn",
    usage = list(
      input_tokens = 10,
      output_tokens = 5,
      cache_creation_input_tokens = 0,
      cache_read_input_tokens = 0
    )
  )

  turn <- value_turn(provider, result)
  fetch_content <- turn@contents[[1]]
  expect_s7_class(fetch_content, ContentToolRequestFetch)
  expect_equal(fetch_content@url, "https://example.com")
})

test_that("value_turn() prices cache writes at 1.25x while reporting raw tokens", {
  provider <- ProviderAnthropic(
    name = "Anthropic",
    base_url = "https://api.anthropic.com/v1",
    model = "claude-sonnet-4-20250514",
    params = list(),
    extra_args = list(),
    extra_headers = character(),
    credentials = NULL,
    beta_headers = character(),
    cache = ""
  )

  result <- list(
    content = list(list(type = "text", text = "ok")),
    stop_reason = "end_turn",
    usage = list(
      input_tokens = 1000,
      output_tokens = 50,
      cache_creation_input_tokens = 400,
      cache_read_input_tokens = 200
    )
  )

  turn <- value_turn(provider, result)

  # tokens slot reports raw integer counts (no 1.25x weighting on input).
  expect_equal(
    unname(turn@tokens),
    c(1000 + 400, 50, 200)
  )

  # Cost matches the 1.25x cache-write weighting:
  #   (1000 + 400 * 1.25) * $3/1M + 50 * $15/1M + 200 * $0.30/1M
  expected_cost <- ((1000 + 400 * 1.25) * 3 + 50 * 15 + 200 * 0.30) / 1e6
  expect_equal(unclass(turn@cost), expected_cost)
})

test_that("stream_merge_chunks() handles citations_delta", {
  provider <- ProviderAnthropic(
    name = "Anthropic",
    base_url = "https://api.anthropic.com/v1",
    model = "claude-sonnet-4-20250514",
    params = list(),
    extra_args = list(),
    extra_headers = character(),
    credentials = NULL,
    beta_headers = character(),
    cache = ""
  )

  chunks <- list(
    list(
      type = "message_start",
      message = list(
        id = "msg_1",
        type = "message",
        role = "assistant",
        content = list(),
        model = "claude-sonnet-4-20250514",
        stop_reason = NULL,
        stop_sequence = NULL,
        usage = list(input_tokens = 10, output_tokens = 1)
      )
    ),
    list(
      type = "content_block_start",
      index = 0L,
      content_block = list(
        type = "text",
        text = ""
      )
    ),
    list(
      type = "content_block_delta",
      index = 0L,
      delta = list(
        type = "text_delta",
        text = "Hello"
      )
    ),
    list(
      type = "content_block_delta",
      index = 0L,
      delta = list(
        type = "citations_delta",
        citation = list(
          type = "web_search_result_location",
          cited_text = "example text",
          url = "https://example.com"
        )
      )
    ),
    list(type = "content_block_stop", index = 0L),
    list(
      type = "message_delta",
      delta = list(
        stop_reason = "end_turn",
        stop_sequence = NULL
      ),
      usage = list(output_tokens = 5)
    )
  )

  result <- NULL
  expect_no_warning({
    for (chunk in chunks) {
      result <- stream_merge_chunks(provider, result, chunk)
    }
  })

  expect_equal(result$content[[1]]$text, "Hello")
  expect_length(result$content[[1]]$citations, 1)
  expect_equal(result$content[[1]]$citations[[1]]$url, "https://example.com")
})

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_openai_test()
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))

  resp <- chat$chat("Double that", echo = FALSE)
  expect_match(resp, "4")
})

test_that("can make simple streaming request", {
  chat <- chat_openai_test()
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

test_that("can list models", {
  test_models(models_openai)
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  expect_snapshot(. <- chat_openai())
})

# No longer supports stop parameter
# test_that("supports standard parameters", {
#   chat_fun <- chat_openai_test

#   test_params_stop(chat_fun)
# })

test_that("supports tool calling", {
  vcr::local_cassette("openai-v2-tool")
  chat_fun <- chat_openai_test

  test_tools_simple(chat_fun)
})

test_that("tools can return images", {
  vcr::local_cassette("openai-v2-tool-image")
  chat_fun <- chat_openai_test
  test_tool_image(chat_fun)
})

test_that("can extract data", {
  chat_fun <- chat_openai_test

  test_data_extraction(chat_fun)
})

test_that("can search web pages", {
  vcr::local_cassette("openai-v2-web-search")
  chat_fun <- \(...) chat_openai_test(model = "gpt-4.1", ...)
  test_tool_web_search(
    chat_fun,
    openai_tool_web_search(),
    hint = "The CRAN archive page has this info."
  )
})

test_that("can use images", {
  vcr::local_cassette("openai-v2-image")
  # Needs mini to get shape correct
  chat_fun <- \(...) chat_openai_test(model = "gpt-4.1-mini", ...)

  test_images_inline(chat_fun)
  test_images_remote(chat_fun)
})

test_that("can use pdfs", {
  vcr::local_cassette("openai-v2-pdf")
  chat_fun <- chat_openai_test

  test_pdf_local(chat_fun)
})

test_that("can match prices for some common models", {
  provider <- chat_openai_test()$get_provider()

  expect_true(has_cost(provider, "gpt-4.1"))
  expect_true(has_cost(provider, "gpt-4.1-2025-04-14"))
})

# Custom tests -----------------------------------------------------------------

test_that("can retrieve log_probs (#115)", {
  chat <- chat_openai_test(params = params(log_probs = TRUE))
  chat$chat("Hi")
  expect_length(chat$last_turn()@json$output[[1]]$content[[1]]$logprobs, 2)
})

test_that("structured data work with and without wrapper", {
  chat <- chat_openai_test()
  out <- chat$chat_structured(
    "Extract the number: apple, green, eleven",
    type = type_number()
  )
  expect_equal(out, 11)

  out <- chat$chat_structured(
    "Extract the number: apple, green, eleven",
    type = type_object(number = type_number())
  )
  expect_equal(out, list(number = 11))
})

test_that("service tier affects pricing", {
  vcr::local_cassette("openai-v2-service-tier")
  chat <- chat_openai_test(service_tier = "priority")
  chat$chat("Tell me a joke")

  last_turn <- chat$last_turn()
  tokens <- as.list(last_turn@tokens)
  priority_cost <- get_token_cost(chat$get_provider(), tokens, "priority")
  expect_equal(last_turn@cost, priority_cost)

  # Confirm we have pricing for the priority tier
  default_cost <- get_token_cost(chat$get_provider(), tokens)
  expect_gt(last_turn@cost, default_cost)
})


test_that("batch retrieve succeeds even if JSON is mangled", {
  local_mocked_bindings(
    openai_download_file = function(provider, id, path) {
      writeLines('{"custom_id": "123", ', path)
    }
  )
  provider <- chat_openai_test()$get_provider()
  out <- batch_retrieve(provider, list(output_file_id = "123"))
  expect_equal(out, list(list(status_code = 500)))
  expect_equal(batch_result_turn(provider, out[[1]]), NULL)
})

test_that("can extract dummy response from malformed JSON", {
  expect_equal(
    openai_json_fallback('{"custom_id": "123", '),
    list(custom_id = "123", response = list(status_code = 500))
  )
})

test_that("value_turn handles NULL service_tier gracefully", {
  provider <- chat_openai_test()$get_provider()

  result <- list(
    output = list(
      list(type = "message", content = list(list(text = "Hello")))
    ),
    usage = NULL,
    service_tier = NULL
  )

  expect_no_error(value_turn(provider, result))
})

test_that("value_turn() handles web_search_call action types", {
  provider <- chat_openai_test()$get_provider()

  make_result <- function(action) {
    list(
      id = "chatcmpl-1",
      object = "chat.completion",
      model = "gpt-4.1-mini",
      choices = list(list(
        index = 0L,
        message = list(
          role = "assistant",
          content = NULL
        ),
        finish_reason = "stop"
      )),
      output = list(
        list(
          id = "ws_1",
          type = "web_search_call",
          status = "completed",
          action = action
        )
      ),
      usage = list(
        prompt_tokens = 10,
        completion_tokens = 5,
        total_tokens = 15
      ),
      service_tier = "default"
    )
  }

  # search action with query
  turn <- value_turn(
    provider,
    make_result(list(type = "search", query = "test query"))
  )
  expect_equal(turn@contents[[1]]@query, "test query")

  # open_page action with url
  turn <- value_turn(
    provider,
    make_result(list(type = "open_page", url = "https://example.com"))
  )
  expect_equal(turn@contents[[1]]@query, "https://example.com")

  # find_in_page action with queries
  turn <- value_turn(
    provider,
    make_result(list(type = "find_in_page", queries = list("find this")))
  )
  expect_equal(turn@contents[[1]]@query, "find this")

  # search action without query
  turn <- value_turn(provider, make_result(list(type = "search")))
  expect_equal(turn@contents[[1]]@query, "web search")

  # find_in_page action with empty queries falls back
  turn <- value_turn(
    provider,
    make_result(list(type = "find_in_page", queries = list()))
  )
  expect_equal(turn@contents[[1]]@query, "web search")
})

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_openai_compatible_test()
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_openai_compatible_test()
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

test_that("can list models", {
  test_models(models_openai)
})


# Common provider interface -----------------------------------------------

test_that("supports standard parameters", {
  chat_fun <- chat_openai_compatible_test

  test_params_stop(chat_fun)
})

test_that("supports tool calling", {
  vcr::local_cassette("openai-tool")
  chat_fun <- chat_openai_compatible_test

  test_tools_simple(chat_fun)
})

test_that("can extract data", {
  chat_fun <- chat_openai_compatible_test

  test_data_extraction(chat_fun)
})

test_that("can use images", {
  vcr::local_cassette("openai-image")
  # Needs mini to get shape correct
  chat_fun <- \(...) chat_openai_compatible_test(model = "gpt-4.1-mini", ...)

  test_images_inline(chat_fun)
  test_images_remote(chat_fun)
})

test_that("can use pdfs", {
  vcr::local_cassette("openai-pdf")
  chat_fun <- chat_openai_compatible_test

  test_pdf_local(chat_fun)
})

test_that("can match prices for some common models", {
  provider <- chat_openai_compatible_test()$get_provider()

  expect_true(has_cost(provider, "gpt-4.1"))
  expect_true(has_cost(provider, "gpt-4.1-2025-04-14"))
})

# Custom tests -----------------------------------------------------------------

test_that("can retrieve log_probs (#115)", {
  chat <- chat_openai_compatible_test(params = params(log_probs = TRUE))
  pieces <- coro::collect(chat$stream("Hi"))

  logprobs <- chat$last_turn()@json$choices[[1]]$logprobs$content
  expect_equal(
    length(logprobs),
    length(pieces) - 2 # leading "" + trailing \n
  )
})

test_that("structured data work with and without wrapper", {
  chat <- chat_openai_compatible_test()
  out <- chat$chat_structured(
    "Extract the number: apple, green, eleven",
    type = type_number()
  )
  expect_equal(out, 11)

  out <- chat$chat_structured(
    "Extract the number: apple, green, eleven",
    type = type_object(number = type_number())
  )
  expect_equal(out, list(number = 11))
})

# Custom -----------------------------------------------------------------

test_that("value_turn() treats empty content string as null", {
  stub <- ProviderOpenAICompatible(name = "", base_url = "", model = "")

  result <- list(
    choices = list(list(
      message = list(
        role = "assistant",
        content = "",
        tool_calls = list(list(
          id = "call_1",
          `function` = list(name = "fn", arguments = "{}")
        ))
      )
    ))
  )

  turn <- value_turn(stub, result)
  # Empty content string should not produce ContentText("")
  expect_false(
    any(map_lgl(turn@contents, function(c) S7_inherits(c, ContentText)))
  )
  # Tool request should still be preserved
  expect_equal(length(turn@contents), 1)
  expect_true(S7_inherits(turn@contents[[1]], ContentToolRequest))
})

test_that("empty ContentText is dropped during serialization", {
  stub <- ProviderOpenAICompatible(name = "", base_url = "", model = "")

  # Assistant turn with only an empty ContentText should be dropped entirely
  turn <- AssistantTurn(list(ContentText("")))
  expect_null(as_json(stub, turn))

  # Multiple empty ContentText values are all dropped

  turn <- AssistantTurn(list(ContentText(""), ContentText("")))
  expect_null(as_json(stub, turn))

  # Empty ContentText is stripped but other content is preserved
  turn <- AssistantTurn(list(
    ContentText(""),
    ContentText("Hello")
  ))
  result <- as_json(stub, turn)
  expect_equal(
    result,
    list(list(
      role = "assistant",
      content = list(list(type = "text", text = "Hello"))
    ))
  )
})

test_that("empty ContentText is stripped but tool requests are preserved", {
  stub <- ProviderOpenAICompatible(name = "", base_url = "", model = "")

  turn <- AssistantTurn(list(
    ContentText(""),
    ContentToolRequest(name = "fn", arguments = list(), id = "call_1")
  ))
  result <- as_json(stub, turn)
  expect_equal(result[[1]]$role, "assistant")
  expect_equal(length(result[[1]]$tool_calls), 1)
  expect_null(result[[1]]$content)
})

test_that("stream_content extracts reasoning_content", {
  stub <- ProviderOpenAICompatible(name = "", base_url = "", model = "")

  event <- list(choices = list(list(delta = list(reasoning_content = "think"))))
  result <- stream_content(stub, event)
  expect_s3_class(result, "ellmer::ContentThinking")
  expect_equal(result@thinking, "think")

  event <- list(choices = list(list(delta = list(content = "hello"))))
  result <- stream_content(stub, event)

  expect_s3_class(result, "ellmer::ContentText")
  expect_equal(result@text, "hello")
})

test_that("value_turn extracts reasoning_content", {
  stub <- ProviderOpenAICompatible(name = "", base_url = "", model = "")

  result <- list(
    choices = list(list(
      message = list(
        role = "assistant",
        reasoning_content = "Let me think...",
        content = "The answer is 42."
      )
    ))
  )

  turn <- value_turn(stub, result)
  expect_equal(length(turn@contents), 2)
  expect_s3_class(turn@contents[[1]], "ellmer::ContentThinking")
  expect_equal(turn@contents[[1]]@thinking, "Let me think...")
  expect_s3_class(turn@contents[[2]], "ellmer::ContentText")
  expect_equal(turn@contents[[2]]@text, "The answer is 42.")
})

test_that("as_json drops reasoning_content by default", {
  stub <- ProviderOpenAICompatible(name = "", base_url = "", model = "")

  turn <- AssistantTurn(list(
    ContentThinking("Let me think..."),
    ContentText("The answer is 42.")
  ))
  result <- as_json(stub, turn)
  expect_null(result[[1]]$reasoning_content)
  expect_equal(
    result[[1]]$content,
    list(list(type = "text", text = "The answer is 42."))
  )
})

test_that("as_json preserves reasoning_content when preserve_thinking = TRUE", {
  stub <- ProviderOpenAICompatible(
    name = "",
    base_url = "",
    model = "",
    preserve_thinking = TRUE
  )

  turn <- AssistantTurn(list(
    ContentThinking("Let me think..."),
    ContentText("The answer is 42.")
  ))
  result <- as_json(stub, turn)
  expect_equal(result[[1]]$reasoning_content, "Let me think...")
  expect_equal(
    result[[1]]$content,
    list(list(type = "text", text = "The answer is 42."))
  )
})

test_that("as_json specialised for OpenAI", {
  stub <- ProviderOpenAI(name = "", base_url = "", model = "")

  expect_snapshot(
    as_json(stub, type_object(.additional_properties = TRUE)),
    error = TRUE
  )

  obj <- type_object(x = type_number(required = FALSE))
  expect_equal(
    as_json(stub, obj),
    list(
      type = "object",
      description = "",
      properties = list(x = list(type = c("number", "null"), description = "")),
      required = list("x"),
      additionalProperties = FALSE
    )
  )
})

test_that("can make simple request", {
  chat <- chat_huggingface_test()
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_huggingface_test()
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  expect_snapshot(. <- chat_huggingface())
})

test_that("supports standard parameters", {
  chat_fun <- chat_huggingface_test
  test_params_stop(chat_fun)
})

test_that("supports tool calling", {
  chat_fun <- chat_huggingface_test
  test_tools_simple(chat_fun)
})

test_that("can extract data", {
  chat_fun <- chat_huggingface_test
  test_data_extraction(chat_fun)
})

test_that("can use images", {
  chat_fun <- function(...) {
    chat_huggingface_test(model = "Qwen/Qwen2.5-VL-7B-Instruct")
  }

  # Thinks hexagon is a diamond
  test_images_inline(chat_fun, test_shape = FALSE)
  test_images_remote(chat_fun, test_shape = FALSE)
})

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_portkey_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_portkey_test("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

# Common provider interface -----------------------------------------------

test_that("supports tool calling", {
  chat_fun <- chat_portkey_test
  test_tools_simple(chat_fun)
})

test_that("can extract data", {
  chat_fun <- chat_portkey_test
  test_data_extraction(chat_fun)
})

test_that("can use images", {
  # Needs mini to get shape correct
  chat_fun <- \(...) chat_portkey_test(model = "gpt-4.1-mini", ...)

  test_images_inline(chat_fun)
  test_images_remote(chat_fun)
})

# Provider specifics ------------------------------------------------------

test_that("virtual_key is deprecated", {
  expect_snapshot(chat <- chat_portkey(model = "def", virtual_key = "abc"))
  expect_equal(chat$get_provider()@model, "@abc/def")
})

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_cloudflare_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_cloudflare_test("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  expect_snapshot(. <- chat_cloudflare())
})

# Not supported
# test_that("supports standard parameters", {
#   chat_fun <- chat_cloudflare_test

#   test_params_stop(chat_fun)
# })

# Doesn't appear to work
# test_that("supports tool calling", {
#   chat_fun <- function(...) {
#     chat_cloudflare_test(
#       ...,
#       model = "@hf/nousresearch/hermes-2-pro-mistral-7b"
#     )
#   }

#   test_tools_simple(chat_fun)
# })

test_that("can extract data", {
  chat_fun <- chat_cloudflare_test

  test_data_extraction(chat_fun)
})

# Can't find model that works
# test_that("can use images", {
#   chat_fun <- function(...)
#     chat_cloudflare_test(model = "@cf/llava-hf/llava-1.5-7b-hf", ...)

#   test_images_inline(chat_fun)
#   # test_images_remote(chat_fun)
# })

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_google_gemini_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?")
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_google_gemini_test("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

test_that("can handle errors", {
  chat <- chat_google_gemini_test(model = "doesnt-exist")
  expect_snapshot(chat$chat("Hi"), error = TRUE)
})

test_that("can list models", {
  test_models(models_google_gemini)
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  expect_snapshot(. <- chat_google_gemini())
})

test_that("supports standard parameters", {
  chat_fun <- chat_google_gemini_test

  test_params_stop(chat_fun)
})

test_that("supports tool calling", {
  vcr::local_cassette("google-tool")
  chat_fun <- chat_google_gemini_test

  test_tools_simple(chat_fun)
})

test_that("can extract data", {
  chat_fun <- chat_google_gemini_test

  test_data_extraction(chat_fun)
})

test_that("can fetch web pages", {
  vcr::local_cassette("google-web-fetch")
  chat_fun <- chat_google_gemini_test
  test_tool_web_fetch(chat_fun, google_tool_web_fetch())
})

test_that("can search web pages", {
  vcr::local_cassette("google-web-search")
  chat_fun <- chat_google_gemini_test
  test_tool_web_search(chat_fun, google_tool_web_search())
})

test_that("can use images", {
  vcr::local_cassette("google-image")
  chat_fun <- chat_google_gemini_test

  test_images_inline(chat_fun)
  test_images_remote_error(chat_fun)
})

test_that("can use pdfs", {
  chat_fun <- chat_google_gemini_test

  test_pdf_local(chat_fun)
})

test_that("can match prices for some common models", {
  provider <- chat_google_gemini_test()$get_provider()

  expect_true(has_cost(provider, "gemini-2.5-flash"))
  expect_false(has_cost(provider, "gemini-1.0-pro-latest"))
})

# custom behaviour -------------------------------------------------------------

test_that("vertex generates expected base_url", {
  chat <- chat_google_vertex("{location}", "{project}")

  service_endpoint <- "https://{location}-aiplatform.googleapis.com/v1"
  model <- "/projects/{project}/locations/{location}/publishers/google/"
  expect_equal(chat$get_provider()@base_url, paste0(service_endpoint, model))
})

test_that("can merge text output", {
  # output from "tell me a joke" with text changed
  messages <- c(
    '{"candidates": [{"content": {"parts": [{"text": "a"}],"role": "model"}}],"usageMetadata": {"promptTokenCount": 5,"totalTokenCount": 5},"modelVersion": "gemini-1.5-flash"}',
    '{"candidates": [{"content": {"parts": [{"text": "b"}],"role": "model"}}],"usageMetadata": {"promptTokenCount": 5,"totalTokenCount": 5},"modelVersion": "gemini-1.5-flash"}',
    '{"candidates": [{"content": {"parts": [{"text": "c"}],"role": "model"},"finishReason": "STOP"}],"usageMetadata": {"promptTokenCount": 5,"candidatesTokenCount": 17,"totalTokenCount": 22},"modelVersion": "gemini-1.5-flash"}'
  )
  chunks <- lapply(messages, jsonlite::parse_json)

  out <- merge_gemini_chunks(chunks[[1]], chunks[[2]])
  out <- merge_gemini_chunks(out, chunks[[3]])

  expect_equal(out$candidates[[1]]$content$parts[[1]]$text, "abc")
  expect_equal(
    out$usageMetadata,
    list(
      promptTokenCount = 5,
      candidatesTokenCount = 17,
      totalTokenCount = 22
    )
  )
  expect_equal(out$candidates[[1]]$finishReason, "STOP")
})

test_that("can handle citations", {
  # based on "Write me a 5-paragraph essay on the history of the tidyverse."
  messages <- c(
    '{"candidates": [{"content": {"parts": [{"text": "a"}]}, "role": "model"}]}',
    '{"candidates": [{
      "content": {"parts": [{"text": "a"}]},
      "role": "model",
      "citationMetadata": {
        "citationSources": [
          {
            "startIndex": 1,
            "endIndex": 2,
            "uri": "https://example.com",
            "license": ""
          }
        ]
      }
    }]}'
  )
  chunks <- lapply(messages, jsonlite::parse_json)

  out <- merge_gemini_chunks(chunks[[1]], chunks[[2]])
  source <- out$candidates[[1]]$citationMetadata$citationSources[[1]]
  expect_equal(source$startIndex, 1)
  expect_equal(source$endIndex, 2)
  expect_equal(source$uri, "https://example.com")
  expect_equal(source$license, "")
})

test_that("can generate images", {
  vcr::local_cassette("google-image-gen")

  chat <- chat_google_gemini_test(model = "gemini-2.5-flash-image")
  chat$chat("Draw a cat")

  turn <- chat$get_turns()[[2]]
  expect_s7_class(turn@contents[[1]], ContentImageInline)
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  expect_snapshot(. <- chat_groq())
})

test_that("supports tool calling", {
  chat_fun <- chat_groq
  test_tools_simple(chat_fun)
})

test_that("supports structured data", {
  # current default model does not support structured data
  chat_fun <- function(model = "openai/gpt-oss-20b", ...) {
    chat_groq(model = model, ...)
  }
  test_data_extraction(chat_fun)
})

#' @include utils-S7.R
NULL

#' Type definitions for function calling and structured data extraction.
#'
#' These S7 classes are provided for use by package devlopers who are
#' extending ellmer. In every day use, use [type_boolean()] and friends.
#'
#' @name Type
#' @inheritParams type_boolean
#' @return S7 objects inheriting from `Type`
#' @examples
#' TypeBasic(type = "boolean")
#' TypeArray(items = TypeBasic(type = "boolean"))
NULL

Type <- new_class(
  "Type",
  properties = list(
    description = prop_string(allow_null = TRUE),
    required = prop_bool(TRUE)
  )
)

#' @export
#' @rdname Type
#' @param type Basic type name. Must be one of `boolean`, `integer`,
#'   `number`, or `string`.
TypeBasic <- new_class(
  "TypeBasic",
  Type,
  properties = list(
    type = prop_string()
  )
)

#' @export
#' @rdname Type
TypeEnum <- new_class(
  "TypeEnum",
  Type,
  properties = list(
    values = class_character
  )
)

#' @export
#' @rdname Type
TypeArray <- new_class(
  "TypeArray",
  Type,
  properties = list(
    items = Type
  )
)

#' @export
#' @param json A JSON schema object as a list.
#' @name Type
TypeJsonSchema <- new_class(
  "TypeJsonSchema",
  Type,
  properties = list(
    json = class_list
  )
)

#' @export
#' @rdname Type
TypeIgnore <- new_class("TypeIgnore", Type)

#' @export
#' @rdname Type
#' @param properties Named list of properties stored inside the object.
#'   Each element should be an S7 `Type` object.`
TypeObject <- new_class(
  "TypeObject",
  Type,
  properties = list(
    properties = prop_list_of(Type, names = "all"),
    additional_properties = prop_bool(FALSE)
  )
)

#' Type specifications
#'
#' @description
#' These functions specify object types in a way that chatbots understand and
#' are used for tool calling and structured data extraction. Their names are
#' based on the [JSON schema](https://json-schema.org), which is what the APIs
#' expect behind the scenes. The translation from R concepts to these types is
#' fairly straightforward.
#'
#' * `type_boolean()`, `type_integer()`, `type_number()`, and `type_string()`
#'   each represent scalars. These are equivalent to length-1 logical,
#'   integer, double, and character vectors (respectively).
#'
#' * `type_enum()` is equivalent to a length-1 factor; it is a string that can
#'   only take the specified values.
#'
#' * `type_array()` is equivalent to a vector in R. You can use it to represent
#'   an atomic vector: e.g. `type_array(type_boolean())` is equivalent
#'   to a logical vector and `type_array(type_string())` is equivalent
#'   to a character vector). You can also use it to represent a list of more
#'   complicated types where every element is the same type (R has no base
#'   equivalent to this), e.g. `type_array(type_array(type_string()))`
#'   represents a list of character vectors.
#'
#' * `type_object()` is equivalent to a named list in R, but where every element
#'   must have the specified type. For example,
#'   `type_object(a = type_string(), b = type_array(type_integer()))` is
#'   equivalent to a list with an element called `a` that is a string and
#'   an element called `b` that is an integer vector.
#'
#'
#' * `type_ignore()` is used in tool calling to indicate that an argument should
#'   not be provided by the LLM. This is useful when the R function has a
#'   default value for the argument and you don't want the LLM to supply it.
#'
#' * `type_from_schema()` allows you to specify the full schema that you want to
#'   get back from the LLM as a JSON schema. This is useful if you have a
#'   pre-defined schema that you want to use directly without manually creating
#'   the type using the `type_*()` functions. You can point to a file with the
#'   `path` argument or provide a JSON string with `text`. The schema must be a
#'   valid JSON schema object.
#'
#' @param description,.description The purpose of the component. This is
#'   used by the LLM to determine what values to pass to the tool or what
#'   values to extract in the structured data, so the more detail that you can
#'   provide here, the better.
#' @param required,.required Is the component or argument required?
#'
#'   In type descriptions for structured data, if `required = FALSE` and the
#'   component does not exist in the data, the LLM may hallucinate a value. Only
#'   applies when the element is nested inside of a `type_object()`.
#'
#'   In tool definitions, `required = TRUE` signals that the LLM should always
#'   provide a value. Arguments with `required = FALSE` should have a default
#'   value in the tool function's definition. If the LLM does not provide a
#'   value, the default value will be used.
#' @export
#' @examples
#' # An integer vector
#' type_array(type_integer())
#'
#' # The closest equivalent to a data frame is an array of objects
#' type_array(type_object(
#'    x = type_boolean(),
#'    y = type_string(),
#'    z = type_number()
#' ))
#'
#' # There's no specific type for dates, but you use a string with the
#' # requested format in the description (it's not guaranteed that you'll
#' # get this format back, but you should most of the time)
#' type_string("The creation date, in YYYY-MM-DD format.")
#' type_string("The update date, in dd/mm/yyyy format.")
type_boolean <- function(description = NULL, required = TRUE) {
  TypeBasic(type = "boolean", description = description, required = required)
}
#' @export
#' @rdname type_boolean
type_integer <- function(description = NULL, required = TRUE) {
  TypeBasic(type = "integer", description = description, required = required)
}
#' @export
#' @rdname type_boolean
type_number <- function(description = NULL, required = TRUE) {
  TypeBasic(type = "number", description = description, required = required)
}
#' @export
#' @rdname type_boolean
type_string <- function(description = NULL, required = TRUE) {
  TypeBasic(type = "string", description = description, required = required)
}

#' @param values Character vector of permitted values.
#' @export
#' @rdname type_boolean
type_enum <- function(values, description = NULL, required = TRUE) {
  TypeEnum(values = values, description = description, required = required)
}

#' @param items The type of the array items. Can be created by any of the
#'   `type_` function.
#' @export
#' @rdname type_boolean
type_array <- function(items, description = NULL, required = TRUE) {
  TypeArray(items = items, description = description, required = required)
}

#' @param ... <[`dynamic-dots`][rlang::dyn-dots]> Name-type pairs defining
#'   the components that the object must possess.
#' @param .additional_properties Can the object have arbitrary additional
#'   properties that are not explicitly listed? Only supported by Claude.
#' @export
#' @rdname type_boolean
type_object <- function(
  .description = NULL,
  ...,
  .required = TRUE,
  .additional_properties = FALSE
) {
  TypeObject(
    properties = list2(...),
    description = .description,
    required = .required,
    additional_properties = .additional_properties
  )
}

#' @param text A JSON string.
#' @param path A file path to a JSON file.
#' @export
#' @rdname type_boolean
type_from_schema <- function(text, path) {
  check_exclusive(text, path)
  if (!missing(text)) {
    json <- jsonlite::fromJSON(text, simplifyVector = FALSE)
  } else {
    json <- jsonlite::read_json(path)
  }
  TypeJsonSchema(json = json)
}

#' @export
#' @rdname type_boolean
type_ignore <- function() {
  TypeIgnore()
}

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_ollama_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_ollama_test("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

test_that("can list models", {
  skip_if_no_ollama()
  test_models(models_ollama)
})

test_that("includes list of models in error message if `model` is missing", {
  skip_if_no_ollama()

  local_mocked_bindings(
    models_ollama = function(...) list(id = "llama3")
  )

  expect_snapshot(chat_ollama(), error = TRUE)
})

test_that("checks that requested model is installed", {
  skip_if_no_ollama()
  local_mocked_bindings(
    models_ollama = function(...) list(id = "llama3")
  )
  expect_snapshot(
    chat_ollama(model = "not-a-real-model"),
    error = TRUE
  )
})

# Common provider interface -----------------------------------------------

test_that("supports tool calling", {
  chat_fun <- chat_ollama_test
  test_tools_simple(chat_fun)

  # Work, but don't match quite the right format because they include
  # additional (blank) ContentText
})

# Currently no other tests because I can't find a model that returns reliable
# results and is reasonably performant.

# Custom -----------------------------------------------------------------

test_that("as_json specialised for Ollama", {
  stub <- ProviderOllama(name = "", base_url = "", model = "")

  expect_snapshot(
    as_json(stub, type_object(.additional_properties = TRUE)),
    error = TRUE
  )

  obj <- type_object(
    x = type_number(required = FALSE),
    y = type_string(required = TRUE)
  )
  expect_equal(
    as_json(stub, obj),
    list(
      type = "object",
      description = "",
      properties = list(
        x = list(type = c("number"), description = ""),
        y = list(type = c("string"), description = "")
      ),
      required = list("y"),
      additionalProperties = FALSE
    )
  )
})

test_that("structured data is round-tripped", {
  chat <- chat_openai_test()
  data <- chat$chat_structured(
    "Generate the name and age of a random person.",
    type = type_object(
      name = type_string(),
      age = type_number()
    )
  )
  expect_match(chat$chat("What is the name of the person?"), data$name)
})

# Object from ContentJSON -----------------------------------------------------

test_that("useful error if no ContentJson", {
  turn <- AssistantTurn(list(ContentText("Hello")))
  expect_snapshot(extract_data(turn), error = TRUE)
})

test_that("can extract data from ContentJson", {
  turn <- AssistantTurn(list(ContentJson(list(x = 1))))
  type <- type_object(x = type_integer())
  expect_equal(extract_data(turn, type), list(x = 1))
})

test_that("can opt-out of conversion data from ContentJson", {
  turn <- AssistantTurn(list(ContentJson(list(x = list(1, 2)))))
  type <- type_object(x = type_array(type_integer()))
  expect_equal(extract_data(turn, type, convert = TRUE), list(x = c(1L, 2L)))
  expect_equal(extract_data(turn, type, convert = FALSE), list(x = list(1, 2)))
})

test_that("can extract data when wrapper is used", {
  turn <- AssistantTurn(list(ContentJson(list(wrapper = list(x = 1)))))
  type <- wrap_type_if_needed(type_object(x = type_integer()), TRUE)
  expect_equal(extract_data(turn, type, needs_wrapper = TRUE), list(x = 1))
})

test_that("warns if multiple ContentJson (and uses first)", {
  turn <- AssistantTurn(
    list(
      ContentJson(list(x = 1)),
      ContentJson(list(x = 2)),
      ContentJson(list(x = 3))
    )
  )
  type <- type_object(x = type_integer())
  expect_snapshot(result <- extract_data(turn, type))
  expect_equal(result, list(x = 1))
})

# Type coercion ---------------------------------------------------------------

test_that("required base types (scalars) become NA", {
  expect_equal(convert_from_type(NULL, type_boolean()), NA)
  expect_equal(convert_from_type(NULL, type_integer()), NA_integer_)
  expect_equal(convert_from_type(NULL, type_number()), NA_real_)
  expect_equal(convert_from_type(NULL, type_string()), NA_character_)
})

test_that("optional base types (scalars) stay as NULL", {
  expect_equal(convert_from_type(NULL, type_boolean(required = FALSE)), NULL)
  expect_equal(convert_from_type(NULL, type_integer(required = FALSE)), NULL)
  expect_equal(convert_from_type(NULL, type_number(required = FALSE)), NULL)
  expect_equal(convert_from_type(NULL, type_string(required = FALSE)), NULL)
})

test_that("can convert arrays of basic types to simple vectors", {
  expect_equal(
    convert_from_type(list(FALSE, TRUE), type_array(type_boolean())),
    c(FALSE, TRUE)
  )
  expect_identical(
    convert_from_type(list(1L, 2L), type_array(type_integer())),
    c(1L, 2L)
  )
  expect_identical(
    convert_from_type(list(1, 2), type_array(type_integer())),
    c(1L, 2L)
  )
  expect_identical(
    convert_from_type(list(1L, 2L), type_array(type_number())),
    c(1, 2)
  )
  expect_equal(
    convert_from_type(list(1.2, 2.5), type_array(type_number())),
    c(1.2, 2.5)
  )
  expect_equal(
    convert_from_type(list("x", "y"), type_array(type_string())),
    c("x", "y")
  )
})

test_that("NULL turns into empty vector", {
  expect_equal(convert_from_type(NULL, type_array(type_boolean())), logical())
  expect_equal(convert_from_type(NULL, type_array(type_integer())), integer())
  expect_equal(convert_from_type(NULL, type_array(type_number())), double())
  expect_equal(convert_from_type(NULL, type_array(type_string())), character())
})

test_that("values of wrong type are silently converted to NA", {
  expect_equal(
    convert_from_type(list(1.2, "x"), type_array(type_number())),
    c(1.2, NA)
  )
})

test_that("values of incorrect length silently truncated", {
  expect_equal(
    convert_from_type(list(c(1, 2), c()), type_array(type_number())),
    c(1, NA)
  )
})

test_that("handles empty and NULL vectors of basic types", {
  type <- type_array(type_boolean(required = FALSE))
  expect_equal(convert_from_type(list(FALSE, TRUE), type), c(FALSE, TRUE))
  expect_equal(convert_from_type(list(NULL, TRUE), type), c(NA, TRUE))

  type <- type_array(type_integer(required = FALSE))
  expect_identical(convert_from_type(list(), type), integer())
  expect_identical(convert_from_type(list(NULL), type), NA_integer_)

  type <- type_array(type_number(required = FALSE))
  expect_identical(convert_from_type(list(), type), double())
  expect_identical(convert_from_type(list(NULL), type), NA_real_)

  type <- type_array(type_string(required = FALSE))
  expect_equal(convert_from_type(list(), type), character())
  expect_equal(convert_from_type(list(NULL), type), NA_character_)
})

test_that("scalar enums are converted to strings", {
  type <- type_enum(c("A", "B", "C"))
  expect_equal(convert_from_type("A", type), "A")
  expect_equal(convert_from_type(NULL, type), NA_character_)
})


test_that("completely missing optional components become NULL", {
  type <- type_integer(required = FALSE)
  expect_equal(convert_from_type(NULL, type), NULL)

  type <- type_array(type_integer(), required = FALSE)
  expect_equal(convert_from_type(NULL, type), NULL)

  type <- type_object(
    x = type_integer(),
    y = type_integer(required = FALSE)
  )
  expect_equal(
    convert_from_type(list(x = 1), type),
    list(x = 1, y = NULL)
  )
})

test_that("objects take order from type", {
  x <- list(y = 1, x = 2)
  type1 <- type_object(x = type_integer(), y = type_integer())
  expect_equal(convert_from_type(x, type1), list(x = 2, y = 1))

  type2 <- type_object(x = type_integer(), .additional_properties = TRUE)
  expect_equal(convert_from_type(x, type2), list(x = 2, y = 1))
})

test_that("additional properties are ignored, unless specified by type", {
  x <- list(y = 1, x = 2, z = 3)
  type <- type_object(x = type_integer())
  expect_equal(convert_from_type(x, type), list(x = 2))

  type <- type_object(x = type_integer(), .additional_properties = TRUE)
  expect_equal(convert_from_type(x, type), list(x = 2, y = 1, z = 3))
})

test_that("can handle missing optional values in objects (#384)", {
  data <- list(
    list(fruit = "Apples", year = NULL),
    list(fruit = "Oranges", year = NULL)
  )
  type <- type_array(
    type_object(
      fruit = type_string(),
      year = type_integer(required = FALSE)
    )
  )
  expect_equal(
    convert_from_type(data, type),
    tibble::tibble(
      fruit = c("Apples", "Oranges"),
      year = c(NA_integer_, NA_integer_)
    )
  )
})

test_that("can covert array of arrays to lists of vectors", {
  expect_equal(
    convert_from_type(
      list(list(1, 2), list(3, 4)),
      type_array(type_array(type_integer()))
    ),
    list(c(1L, 2L), c(3L, 4L))
  )
})

test_that("arrays of enums are converted to factors", {
  type <- type_array(type_enum(c("x", "y", "z")))
  expect_equal(
    convert_from_type(list("x", "y"), type),
    factor(c("x", "y"), levels = c("x", "y", "z"))
  )

  expect_equal(
    convert_from_type(NULL, type),
    factor(character(), levels = c("x", "y", "z"))
  )
})

test_that("can convert arrays of objects to data frames", {
  x <- list(list(x = 1, y = "x"), list(x = 3, y = "y"))
  type <- type_array(type_object(x = type_integer(), y = type_string()))
  expect_equal(
    convert_from_type(x, type),
    tibble::tibble(x = c(1L, 3L), y = c("x", "y"))
  )

  # unless they have additional properties
  type <- type_array(type_object(
    x = type_integer(),
    .additional_properties = TRUE
  ))
  expect_equal(convert_from_type(x, type), x)

  # in which case the order should still be preserved
  type2 <- type_array(type_object(
    y = type_integer(),
    .additional_properties = TRUE
  ))
  expect_equal(
    convert_from_type(x, type2),
    list(list(y = "x", x = 1), list(y = "y", x = 3))
  )
})

test_that("array of object with nested objects becomes packed data frame", {
  type <- type_array(
    type_object(
      x = type_object(a = type_integer()),
      y = type_object(a = type_integer())
    )
  )

  data <- list(
    list(x = list(a = 1), y = list(a = 3)),
    list(x = list(a = 5), y = list(a = 7))
  )

  expect_equal(
    convert_from_type(data, type),
    tibble::tibble(
      x = tibble::tibble(a = c(1, 5)),
      y = tibble::tibble(a = c(3, 7))
    )
  )
})

test_that("can handle mix of present and absent rows", {
  type <- type_array(type_object(
    x = type_integer(),
    y = type_string()
  ))
  data <- list(
    list(x = 1, y = "x"),
    NULL,
    list(x = 3, y = "y")
  )
  expect_equal(
    convert_from_type(data, type),
    tibble::tibble(x = c(1L, NA_integer_, 3L), y = c("x", NA_character_, "y"))
  )
})

test_that("can recursively convert objects contents", {
  expect_equal(
    convert_from_type(
      list(x = 1, y = list(1, 2, 3)),
      type_object(x = type_integer(), y = type_array(type_integer()))
    ),
    list(x = 1, y = c(1, 2, 3))
  )
})

test_that("can convert basic types to json schema", {
  expect_equal(
    as_json(test_provider(), type_boolean("desc")),
    list(type = "boolean", description = "desc")
  )

  expect_equal(
    as_json(test_provider(), type_enum(letters[1:3], "desc")),
    list(type = "string", description = "desc", enum = as.list(letters[1:3]))
  )

  expect_equal(
    as_json(test_provider(), type_array(type_boolean("b"), "a")),
    list(
      type = "array",
      description = "a",
      items = list(type = "boolean", description = "b")
    )
  )
})

test_that("can convert an object to json schema", {
  obj <- type_object(
    "a",
    integer = type_integer(),
    number = type_number(),
    string = type_string()
  )

  expect_equal(
    as_json(test_provider(), obj),
    list(
      type = "object",
      description = "a",
      properties = list(
        integer = list(type = "integer", description = ""),
        number = list(type = "number", description = ""),
        string = list(type = "string", description = "")
      ),
      required = list("integer", "number", "string"),
      additionalProperties = FALSE
    )
  )
})

test_that("invalid inputs give useful errors", {
  chat <- chat_openai_test()

  expect_snapshot(error = TRUE, {
    chat$chat(question = "Are unicorns real?")
    chat$chat(TRUE)
  })
})

test_that("can create content from a vector", {
  expect_equal(
    as_content(c("a", "b")),
    ContentText("a\n\nb")
  )
})

test_that("turn contents can be converted to text, markdown and HTML", {
  turn <- UserTurn(
    contents = list(
      ContentText("User input."),
      ContentImageInline("image/png", "abcd123"),
      ContentImageRemote("https://example.com/image.jpg", detail = ""),
      ContentJson(list(a = 1:2, b = "apple"))
    )
  )

  expect_snapshot(cat(contents_text(turn)))
  expect_snapshot(cat(contents_markdown(turn)))

  turns <- list(
    turn,
    AssistantTurn(list(ContentText("Here's your answer.")))
  )
  chat <- Chat$new(test_provider())
  chat$set_turns(turns)
  expect_snapshot(cat(contents_markdown(chat)))

  skip_if_not_installed("commonmark")
  expect_snapshot(cat(contents_html(turn)))
})


# Content types ----------------------------------------------------------------

test_that("thinking has useful representations", {
  ct <- ContentThinking("A **thought**.")
  expect_equal(contents_text(ct), NULL)
  expect_equal(format(ct), "<thinking>\nA **thought**.\n</thinking>\n")
  expect_equal(
    contents_markdown(ct),
    "<thinking>\nA **thought**.\n</thinking>\n"
  )
  expect_snapshot(cat(contents_html(ct)))
})

test_that("ContentToolRequest shows converted arguments", {
  my_tool <- tool(
    function(x, y, z) {},
    name = "my_tool",
    description = "A tool",
    arguments = list(
      x = type_array(type_number()),
      y = type_array(type_string()),
      z = type_enum(c("a", "b", "c"))
    )
  )
  content <- ContentToolRequest(
    "id",
    name = "my_tool",
    arguments = list(x = c(1, 2), y = c("a", "b"), z = "a"),
    tool = my_tool
  )
  expect_snapshot(cat(format(content)))

  # and very long arguments are truncated
  content <- ContentToolRequest(
    "id",
    name = "my_tool",
    arguments = list(x = rep(123, 1000), y = c("a", "b"), z = "a"),
    tool = my_tool
  )
  expect_snapshot(cat(format(content)))
})

test_that("ContentToolResult@error requires a string or an error condition", {
  expect_snapshot(error = TRUE, {
    ContentToolResult("id", error = TRUE)
    ContentToolResult("id", error = c("one", "two"))
  })
})

#' @include provider.R
#' @include content.R
#' @include turns.R
#' @include tools-def.R
NULL

#' Chat with an Anthropic Claude model
#'
#' @description
#' [Anthropic](https://www.anthropic.com) provides a number of chat based models
#' under the [Claude](https://claude.com/product/overview) moniker. Note that a
#' Claude Pro membership does not give you the ability to call models via the
#' API; instead, you will need to sign up (and pay for) a
#' [developer account](https://platform.claude.com/).
#'
#' # Caching
#'
#' Caching with Claude is a bit more complicated than other providers but we
#' believe that on average it will save you both money and time, so we have
#' enabled it by default. With other providers, like OpenAI and Google,
#' you only pay for cache reads, which cost 10% of the normal price. With
#' Claude, you also pay for cache writes, which cost 125% of the normal price
#' for 5 minute caching and 200% of the normal price for 1 hour caching.
#'
#' How does this affect the total cost of a conversation? Imagine the first
#' turn sends 1000 input tokens and receives 200 output tokens. The second
#' turn must first send both the input and output from the previous turn
#' (1200 tokens). It then sends a further 1000 tokens and receives 200 tokens
#' back.
#'
#' To compare the prices of these two approaches we can ignore the cost of
#' output tokens, because they are the same for both. How much will the input
#' tokens cost? If we don't use caching, we send 1000 tokens in the first turn
#' and 2200 (1000 + 200 + 1000) tokens in the second turn for a total of 3200
#' tokens. If we use caching, we'll send (the equivalent of) 1000 * 1.25 = 1250
#' tokens in the first turn. In the second turn, 1000 of the input tokens will
#' be cached so the total cost is 1000 * 0.1 + (200 + 1000) * 1.25 = 1600
#' tokens. That makes a total of 2850 tokens, i.e. 11% fewer tokens,
#' decreasing the overall cost.
#'
#' Obviously, the details will vary from conversation to conversation, but
#' if you have a large system prompt that you re-use many times you should
#' expect to see larger savings. You can see exactly how many input and
#' cache input tokens each turn uses, along with the total cost,
#' with `chat$get_tokens()`. If you don't see savings for your use case, you can
#' suppress caching with `cache = "none"`.
#'
#' I know this is already quite complicated, but there's one final wrinkle:
#' Claude will only cache longer prompts, with caching requiring at least
#' 1024-4096 tokens, depending on the model. So don't be surprised it if you
#' don't see any differences with caching if you have a short prompt.
#'
#' See all the details at
#' <https://docs.claude.com/en/docs/build-with-claude/prompt-caching>.
#'
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @param model `r param_model("claude-sonnet-4-5-20250929", "anthropic")`
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("ANTHROPIC_API_KEY")`
#' @param base_url The base URL to the endpoint; the default is Claude's
#'   public API.
#' @param cache How long to cache inputs? Defaults to "5m" (five minutes).
#'   Set to "none" to disable caching or "1h" to cache for one hour.
#'
#'   See details below.
#' @param beta_headers Optionally, a character vector of beta headers to opt-in
#'   claude features that are still in beta.
#' @param api_headers Named character vector of arbitrary extra headers appended
#'   to every chat API call.
#' @family chatbots
#' @export
#' @examples
#' \dontshow{ellmer:::vcr_example_start("chat_anthropic")}
#' chat <- chat_anthropic()
#' chat$chat("Tell me three jokes about statisticians")
#' \dontshow{ellmer:::vcr_example_end()}
chat_anthropic <- function(
  system_prompt = NULL,
  params = NULL,
  model = NULL,
  cache = c("5m", "1h", "none"),
  api_args = list(),
  base_url = "https://api.anthropic.com/v1",
  beta_headers = character(),
  api_key = NULL,
  credentials = NULL,
  api_headers = character(),
  echo = NULL
) {
  echo <- check_echo(echo)

  model <- set_default(model, "claude-sonnet-4-5-20250929")
  cache <- arg_match(cache)

  credentials <- as_credentials(
    "chat_anthropic",
    function() anthropic_key(),
    credentials = credentials,
    api_key = api_key
  )

  provider <- ProviderAnthropic(
    name = "Anthropic",
    model = model,
    params = params %||% params(),
    extra_args = api_args,
    extra_headers = api_headers,
    base_url = base_url,
    beta_headers = beta_headers,
    credentials = credentials,
    cache = cache
  )

  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

#' @rdname chat_anthropic
#' @export
chat_claude <- chat_anthropic

chat_anthropic_test <- function(
  ...,
  model = "claude-sonnet-4-5-20250929",
  params = NULL,
  echo = "none"
) {
  params <- params %||% params()
  params$temperature <- params$temperature %||% 0

  chat_anthropic(model = model, params = params, ..., echo = echo)
}

ProviderAnthropic <- new_class(
  "ProviderAnthropic",
  parent = Provider,
  properties = list(
    beta_headers = class_character,
    cache = prop_string()
  )
)

anthropic_key <- function() {
  key_get("ANTHROPIC_API_KEY")
}
anthropic_key_exists <- function() {
  key_exists("ANTHROPIC_API_KEY")
}

method(base_request, ProviderAnthropic) <- function(provider) {
  req <- request(provider@base_url)
  # <https://docs.anthropic.com/en/api/versioning>
  req <- req_headers(req, `anthropic-version` = "2023-06-01")
  # <https://docs.anthropic.com/en/api/getting-started#authentication>
  req <- ellmer_req_credentials(req, provider@credentials(), "x-api-key")

  # <https://docs.anthropic.com/en/api/rate-limits>
  # <https://docs.anthropic.com/en/api/errors#http-errors>
  req <- ellmer_req_robustify(req, is_transient = function(resp) {
    resp_status(resp) %in% c(429, 503, 529)
  })

  if (length(provider@beta_headers) > 0) {
    req <- req_headers(req, `anthropic-beta` = provider@beta_headers)
  }

  # <https://docs.anthropic.com/en/api/errors>
  req <- req_error(req, body = function(resp) {
    if (resp_content_type(resp) == "application/json") {
      json <- resp_body_json(resp)
      paste0(json$error$message, " [", json$error$type, "]")
    }
  })

  req
}


# https://docs.anthropic.com/en/api/messages
method(chat_path, ProviderAnthropic) <- function(provider) {
  "messages"
}
method(chat_body, ProviderAnthropic) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  if (length(turns) >= 1 && is_system_turn(turns[[1]])) {
    system <- list(list(type = "text", text = turns[[1]]@text))
    # Always cache system prompt
    system[[1]]$cache_control <- cache_control(provider)
  } else {
    system <- NULL
  }

  is_last <- seq_along(turns) == length(turns)
  messages <- compact(map2(turns, is_last, function(turn, is_last) {
    as_json(provider, turn, is_last = is_last)
  }))

  if (!is.null(type)) {
    tool_def <- ToolDef(
      function(...) {},
      name = "_structured_tool_call",
      description = "Extract structured data",
      arguments = type_object(data = type)
    )
    tools[[tool_def@name]] <- tool_def
    tool_choice <- list(type = "tool", name = tool_def@name)
    stream <- FALSE
  } else {
    tool_choice <- NULL
  }
  tools <- as_json(provider, unname(tools))

  params <- chat_params(provider, provider@params)
  if (has_name(params, "budget_tokens")) {
    thinking <- list(
      type = "enabled",
      budget_tokens = params$budget_tokens
    )
    params$budget_tokens <- NULL
  } else {
    thinking <- NULL
  }

  compact(list2(
    model = provider@model,
    system = system,
    messages = messages,
    stream = stream,
    tools = tools,
    tool_choice = tool_choice,
    thinking = thinking,
    !!!params
  ))
}

method(chat_params, ProviderAnthropic) <- function(provider, params) {
  params <- standardise_params(
    params,
    c(
      temperature = "temperature",
      top_p = "top_p",
      top_k = "top_k",
      max_tokens = "max_tokens",
      stop_sequences = "stop_sequences",
      budget_tokens = "reasoning_tokens"
    )
  )

  # Unlike other providers, Claude requires that this be set
  params$max_tokens <- params$max_tokens %||% 4096

  params$stop_sequences <- as.list(params$stop_sequences)

  params
}

# Claude -> ellmer --------------------------------------------------------------

method(stream_parse, ProviderAnthropic) <- function(provider, event) {
  if (is.null(event)) {
    cli::cli_abort("Connection closed unexpectedly")
  }

  data <- jsonlite::parse_json(event$data)
  if (identical(data$type, "message_stop")) {
    return(NULL)
  }

  data
}
method(stream_content, ProviderAnthropic) <- function(provider, event) {
  if (event$type == "content_block_delta") {
    if (identical(event$delta$type, "thinking_delta")) {
      return(ContentThinking(event$delta$thinking))
    }
    text <- event$delta$text
    if (is.null(text)) {
      return(NULL)
    }
    ContentText(text)
  }
}
method(stream_merge_chunks, ProviderAnthropic) <- function(
  provider,
  result,
  chunk
) {
  if (chunk$type == "ping") {
    # nothing to do
  } else if (chunk$type == "message_start") {
    result <- chunk$message
  } else if (chunk$type == "content_block_start") {
    result$content[[chunk$index + 1L]] <- chunk$content_block
  } else if (chunk$type == "content_block_delta") {
    # https://docs.anthropic.com/en/api/messages-streaming#delta-types
    i <- chunk$index + 1L

    if (chunk$delta$type == "text_delta") {
      paste(result$content[[i]]$text) <- chunk$delta$text
    } else if (chunk$delta$type == "input_json_delta") {
      if (chunk$delta$partial_json != "") {
        # See issue #228 about partial_json sometimes being ""
        paste(result$content[[i]]$input) <- chunk$delta$partial_json
      }
    } else if (chunk$delta$type == "thinking_delta") {
      paste(result$content[[i]]$thinking) <- chunk$delta$thinking
    } else if (chunk$delta$type == "signature_delta") {
      paste(result$content[[i]]$signature) <- chunk$delta$signature
    } else if (chunk$delta$type == "citations_delta") {
      # https://docs.claude.com/en/docs/build-with-claude/citations#streaming-support
      result$content[[i]]$citations <- c(
        result$content[[i]]$citations,
        list(chunk$delta$citation)
      )
    } else {
      cli::cli_inform(c("!" = "Unknown delta type {.str {chunk$delta$type}}."))
    }
  } else if (chunk$type == "content_block_stop") {
    # nothing to do
  } else if (chunk$type == "message_delta") {
    result$stop_reason <- chunk$delta$stop_reason
    result$stop_sequence <- chunk$delta$stop_sequence
    result$usage$output_tokens <- chunk$usage$output_tokens
  } else if (chunk$type == "error") {
    if (chunk$error$type == "overloaded_error") {
      # https://docs.anthropic.com/en/api/messages-streaming#error-events
      # TODO: track number of retries
      wait <- backoff_default(1)
      Sys.sleep(wait)
    } else {
      cli::cli_abort("{chunk$error$message}")
    }
  } else {
    cli::cli_inform(c("!" = "Unknown chunk type {.str {chunk$type}}."))
  }
  result
}

method(value_tokens, ProviderAnthropic) <- function(provider, json) {
  usage <- json$usage
  tokens(
    input = (usage$input_tokens %||% 0) +
      (usage$cache_creation_input_tokens %||% 0),
    output = usage$output_tokens %||% 0,
    cached_input = usage$cache_read_input_tokens %||% 0
  )
}

method(value_turn, ProviderAnthropic) <- function(
  provider,
  result,
  has_type = FALSE
) {
  contents <- lapply(result$content, function(content) {
    if (content$type == "text") {
      ContentText(content$text)
    } else if (content$type == "tool_use") {
      if (has_type) {
        ContentJson(data = content$input$data)
      } else {
        if (is_string(content$input)) {
          content$input <- jsonlite::parse_json(content$input)
        }
        ContentToolRequest(content$id, content$name, content$input)
      }
    } else if (content$type == "server_tool_use") {
      if (is_string(content$input)) {
        content$input <- jsonlite::parse_json(content$input)
      }
      if (content$name == "web_search") {
        # https://docs.claude.com/en/docs/agents-and-tools/tool-use/web-search-tool#response
        ContentToolRequestSearch(
          query = content$input$query,
          json = content
        )
      } else if (content$name == "web_fetch") {
        # https://docs.claude.com/en/docs/agents-and-tools/tool-use/web-fetch-tool#response
        ContentToolRequestFetch(
          url = content$input$url,
          json = content
        )
      } else {
        cli::cli_abort("Unknown server tool {.str {content$name}}.")
      }
    } else if (content$type == "web_search_tool_result") {
      urls <- map_chr(content$content, \(x) x$url)
      ContentToolResponseSearch(
        urls = urls,
        json = content
      )
    } else if (content$type == "web_fetch_tool_result") {
      ContentToolResponseFetch(url = content$url %||% "failed", json = content)
    } else if (content$type == "thinking") {
      ContentThinking(
        content$thinking,
        extra = list(signature = content$signature)
      )
    } else {
      cli::cli_abort(
        "Unknown content type {.str {content$type}}.",
        .internal = TRUE
      )
    }
  })

  tokens <- value_tokens(provider, result)
  cache_write <- result$usage$cache_creation_input_tokens %||% 0
  # Anthropic charges 1.25x the input rate for cache writes; tokens$input
  # already counts them at 1.0x, so add the 0.25x surcharge for pricing.
  cost_tokens <- tokens
  cost_tokens$input <- cost_tokens$input + cache_write * 0.25
  cost <- get_token_cost(provider, cost_tokens)
  AssistantTurn(contents, json = result, tokens = unlist(tokens), cost = cost)
}

# ellmer -> Claude --------------------------------------------------------------

method(as_json, list(ProviderAnthropic, Turn)) <- function(
  provider,
  x,
  ...,
  is_last = FALSE
) {
  if (is_system_turn(x)) {
    # claude passes system prompt as separate arg
    NULL
  } else if (is_user_turn(x) || is_assistant_turn(x)) {
    if (is_assistant_turn(x) && identical(x@contents, list())) {
      # Drop empty assistant turns to avoid an API error
      # (all messages must have non-empty content)
      return(NULL)
    }
    x <- turn_contents_expand(x)
    content <- as_json(provider, x@contents, ...)

    # Add caching to the last content block in the last turn
    # https://docs.claude.com/en/docs/build-with-claude/prompt-caching#how-automatic-prefix-checking-works
    if (is_last) {
      content[[length(content)]]$cache_control <- cache_control(provider)
    }
    list(role = x@role, content = content)
  } else {
    cli::cli_abort("Unknown role {x@role}", .internal = TRUE)
  }
}

method(as_json, list(ProviderAnthropic, ContentText)) <- function(
  provider,
  x,
  ...
) {
  if (is_whitespace(x@text)) {
    list(type = "text", text = "[empty string]")
  } else {
    list(type = "text", text = x@text)
  }
}

method(as_json, list(ProviderAnthropic, ContentPDF)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "document",
    source = list(
      type = "base64",
      media_type = x@type,
      data = x@data
    )
  )
}

method(as_json, list(ProviderAnthropic, ContentUploaded)) <- function(
  provider,
  x
) {
  # https://docs.claude.com/en/docs/build-with-claude/files#using-a-file-in-messages
  block_type <- switch(
    x@mime_type,
    "application/pdf" = "document",
    "text/plain" = "document",
    "image/jpeg" = "image",
    "image/png" = "image",
    "image/gif" = "image",
    "image/webp" = "image",
    "text/csv" = "container_upload",
    "application/json" = "container_upload",
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet" = "container_upload",
    "application/vnd.ms-excel" = "container_upload",
    "text/xml" = "container_upload",
    "application/xml" = "container_upload"
  )

  list(
    type = block_type,
    source = list(
      type = "file",
      file_id = x@uri
    )
  )
}

method(as_json, list(ProviderAnthropic, ContentImageRemote)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "image",
    source = list(
      type = "url",
      url = x@url
    )
  )
}

method(as_json, list(ProviderAnthropic, ContentImageInline)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "image",
    source = list(
      type = "base64",
      media_type = x@type,
      data = x@data
    )
  )
}

# https://docs.anthropic.com/en/docs/build-with-claude/tool-use#handling-tool-use-and-tool-result-content-blocks
method(as_json, list(ProviderAnthropic, ContentToolRequest)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "tool_use",
    id = x@id,
    name = x@name,
    input = x@arguments
  )
}

# https://docs.anthropic.com/en/docs/build-with-claude/tool-use#handling-tool-use-and-tool-result-content-blocks
method(as_json, list(ProviderAnthropic, ContentToolResult)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "tool_result",
    tool_use_id = x@request@id,
    content = tool_string(x),
    is_error = tool_errored(x)
  )
}

method(as_json, list(ProviderAnthropic, ToolDef)) <- function(
  provider,
  x,
  ...
) {
  list(
    name = x@name,
    description = x@description,
    input_schema = compact(as_json(provider, x@arguments, ...))
  )
}

method(as_json, list(ProviderAnthropic, ContentThinking)) <- function(
  provider,
  x,
  ...
) {
  if (identical(x@thinking, "")) {
    return()
  }

  list(
    type = "thinking",
    thinking = x@thinking,
    signature = x@extra$signature
  )
}

# Batch chat -------------------------------------------------------------------

method(has_batch_support, ProviderAnthropic) <- function(provider) {
  TRUE
}

# https://docs.anthropic.com/en/api/creating-message-batches
method(batch_submit, ProviderAnthropic) <- function(
  provider,
  conversations,
  type = NULL
) {
  req <- base_request(provider)
  req <- req_url_path_append(req, "/messages/batches")

  requests <- map(seq_along(conversations), function(i) {
    params <- chat_body(
      provider,
      stream = FALSE,
      turns = conversations[[i]],
      type = type
    )
    list(
      custom_id = paste0("chat-", i),
      params = params
    )
  })
  req <- req_body_json(req, list(requests = requests))

  resp <- req_perform(req)
  resp_body_json(resp)
}

# https://docs.anthropic.com/en/api/retrieving-message-batches
method(batch_poll, ProviderAnthropic) <- function(provider, batch) {
  req <- base_request(provider)
  req <- req_url_path_append(req, "/messages/batches", batch$id)
  resp <- req_perform(req)

  resp_body_json(resp)
}

method(batch_status, ProviderAnthropic) <- function(provider, batch) {
  counts <- batch$request_counts
  list(
    working = batch$processing_status != "ended",
    n_processing = batch$request_counts$processing,
    n_succeeded = batch$request_counts$succeeded,
    n_failed = counts$errored + counts$canceled + counts$expired
  )
}

# https://docs.anthropic.com/en/api/retrieving-message-batch-results
method(batch_retrieve, ProviderAnthropic) <- function(provider, batch) {
  req <- base_request(provider)
  req <- req_url(req, batch$results_url)
  req <- req_progress(req, "down")

  path <- local_tempfile()
  req <- req_perform(req, path = path)

  lines <- readLines(path, warn = FALSE)
  json <- lapply(lines, jsonlite::fromJSON, simplifyVector = FALSE)

  ids <- as.numeric(gsub("chat-", "", map_chr(json, "[[", "custom_id")))
  results <- lapply(json, "[[", "result")
  results[order(ids)]
}

method(batch_result_turn, ProviderAnthropic) <- function(
  provider,
  result,
  has_type = FALSE
) {
  if (result$type == "succeeded") {
    value_turn(provider, result$message, has_type = has_type)
  } else {
    NULL
  }
}

# Models -----------------------------------------------------------------------

#' @export
#' @rdname chat_anthropic
models_claude <- function(
  base_url = "https://api.anthropic.com/v1",
  api_key = NULL,
  credentials = NULL
) {
  credentials <- as_credentials(
    "models_anthropic",
    function() anthropic_key(),
    credentials = credentials,
    api_key = api_key
  )

  provider <- ProviderAnthropic(
    name = "Anthropic",
    model = "",
    base_url = base_url,
    credentials = credentials,
    cache = "none"
  )

  req <- base_request(provider)
  req <- req_url_path_append(req, "/models")
  resp <- req_perform(req)

  json <- resp_body_json(resp)

  id <- map_chr(json$data, "[[", "id")
  display_name <- map_chr(json$data, "[[", "display_name")
  created_at <- as.POSIXct(map_chr(json$data, "[[", "created_at"))

  df <- data.frame(
    id = id,
    name = display_name,
    created_at = created_at
  )
  df <- cbind(df, match_prices("Anthropic", df$id))
  df[order(-xtfrm(df$created_at)), ]
}

#' @export
#' @rdname chat_anthropic
models_anthropic <- models_claude

# Helpers ----------------------------------------------------------------

# From httr2
backoff_default <- function(i) {
  round(min(stats::runif(1, min = 1, max = 2^i), 60), 1)
}

cache_control <- function(provider) {
  if (provider@cache == "none") {
    NULL
  } else {
    list(
      type = "ephemeral",
      ttl = provider@cache
    )
  }
}

#' Submit multiple chats in one batch
#'
#' @description
#' `batch_chat()` and `batch_chat_structured()` currently only work with
#' [chat_openai()] and [chat_anthropic()]. They use the
#' [OpenAI](https://platform.openai.com/docs/guides/batch) and
#' [Anthropic](https://docs.claude.com/en/docs/build-with-claude/batch-processing)
#' batch APIs which allow you to submit multiple requests simultaneously.
#' The results can take up to 24 hours to complete, but in return you pay 50%
#' less than usual (but note that ellmer doesn't include this discount in
#' its pricing metadata). If you want to get results back more quickly, or
#' you're working with a different provider, you may want to use
#' [parallel_chat()] instead.
#'
#' Since batched requests can take a long time to complete, `batch_chat()`
#' requires a file path that is used to store information about the batch so
#' you never lose any work. You can either set `wait = FALSE` or simply
#' interrupt the waiting process, then later, either call `batch_chat()` to
#' resume where you left off or call `batch_chat_completed()` to see if the
#' results are ready to retrieve. `batch_chat()` will store the chat responses
#' in this file, so you can either keep it around to cache the results,
#' or delete it to free up disk space.
#'
#' This API is marked as experimental since I don't yet know how to handle
#' errors in the most helpful way. Fortunately they don't seem to be common,
#' but if you have ideas, please let me know!
#'
#' @inheritParams parallel_chat
#' @param path Path to file (with `.json` extension) to store state.
#'
#'   The file records a hash of the provider, the prompts, and the existing
#'   chat turns. If you attempt to reuse the same file with any of these being
#'   different, you'll get an error.
#' @param wait If `TRUE`, will wait for batch to complete. If `FALSE`,
#'   it will return `NULL` if the batch is not complete, and you can retrieve
#'   the results later by re-running `batch_chat()` when
#'   `batch_chat_completed()` is `TRUE`.
#' @param ignore_hash If `TRUE`, will only warn rather than error when the hash
#'   doesn't match. You can use this if ellmer has changed the hash structure
#'   and you're confident that you're reusing the same inputs.
#' @returns
#' For `batch_chat()`, a list of [Chat] objects, one for each prompt.
#' For `batch_chat_test()`, a character vector of text responses.
#' For `batch_chat_structured()`, a single structured data object with one
#' element for each prompt. Typically, when `type` is an object, this will
#' will be a data frame with one row for each prompt, and one column for each
#' property.
#'
#' For any of the aboves, will return `NULL` if `wait = FALSE` and the job
#' is not complete.
#' @examplesIf has_credentials("openai")
#' chat <- chat_openai(model = "gpt-4.1-nano")
#'
#' # Chat ----------------------------------------------------------------------
#'
#' prompts <- interpolate("What do people from {{state.name}} bring to a potluck dinner?")
#' \dontrun{
#' chats <- batch_chat(chat, prompts, path = "potluck.json")
#' chats
#' }
#'
#' # Structured data -----------------------------------------------------------
#' prompts <- list(
#'   "I go by Alex. 42 years on this planet and counting.",
#'   "Pleased to meet you! I'm Jamal, age 27.",
#'   "They call me Li Wei. Nineteen years young.",
#'   "Fatima here. Just celebrated my 35th birthday last week.",
#'   "The name's Robert - 51 years old and proud of it.",
#'   "Kwame here - just hit the big 5-0 this year."
#' )
#' type_person <- type_object(name = type_string(), age = type_number())
#' \dontrun{
#' data <- batch_chat_structured(
#'   chat = chat,
#'   prompts = prompts,
#'   path = "people-data.json",
#'   type = type_person
#' )
#' data
#' }
#' @export
batch_chat <- function(chat, prompts, path, wait = TRUE, ignore_hash = FALSE) {
  chat <- as_chat(chat)

  job <- BatchJob$new(
    chat = chat,
    prompts = prompts,
    path = path,
    wait = wait,
    ignore_hash = ignore_hash
  )
  job$step_until_done()

  assistant_turns <- job$result_turns()
  map2(job$user_turns, assistant_turns, function(user, assistant) {
    if (!is.null(assistant)) {
      # Logged on retrieval
      chat$clone()$add_turn(user, assistant, log_tokens = FALSE)
    } else {
      NULL
    }
  })
}

#' @export
#' @rdname batch_chat
batch_chat_text <- function(
  chat,
  prompts,
  path,
  wait = TRUE,
  ignore_hash = FALSE
) {
  chat <- as_chat(chat)
  chats <- batch_chat(
    chat,
    prompts,
    path,
    wait = wait,
    ignore_hash = ignore_hash
  )
  map_chr(chats, \(chat) {
    if (is.null(chat)) NA_character_ else chat$last_turn()@text
  })
}

#' @export
#' @rdname batch_chat
#' @inheritParams parallel_chat_structured
batch_chat_structured <- function(
  chat,
  prompts,
  path,
  type,
  wait = TRUE,
  ignore_hash = FALSE,
  convert = TRUE,
  include_tokens = FALSE,
  include_cost = FALSE
) {
  chat <- as_chat(chat)
  provider <- chat$get_provider()
  needs_wrapper <- type_needs_wrapper(type, provider)

  job <- BatchJob$new(
    chat = chat,
    prompts = prompts,
    type = wrap_type_if_needed(type, needs_wrapper),
    path = path,
    wait = wait,
    ignore_hash = ignore_hash
  )
  job$step_until_done()
  turns <- job$result_turns()

  multi_convert(
    provider,
    turns,
    type,
    convert = convert,
    include_tokens = include_tokens,
    include_cost = include_cost
  )
}

#' @export
#' @rdname batch_chat
batch_chat_completed <- function(chat, prompts, path) {
  job <- BatchJob$new(
    chat = chat,
    prompts = prompts,
    path = path
  )
  switch(
    job$stage,
    "submitting" = FALSE,
    "waiting" = !job$poll()$working,
    "retrieving" = TRUE,
    "done" = TRUE,
    cli::cli_abort("Unexpected stage: {job$stage}", .internal = TRUE)
  )
}

BatchJob <- R6::R6Class(
  "BatchJob",
  public = list(
    chat = NULL,
    user_turns = NULL,
    path = NULL,
    should_wait = TRUE,
    ignore_hash = FALSE,
    type = NULL,

    # Internal state
    provider = NULL,
    started_at = NULL,
    stage = NULL,
    batch = NULL,
    results = NULL,

    initialize = function(
      chat,
      prompts,
      path,
      type = NULL,
      wait = TRUE,
      ignore_hash = FALSE,
      call = caller_env(2)
    ) {
      self$provider <- chat$get_provider()
      check_has_batch_support(self$provider, call = call)

      user_turns <- as_user_turns(prompts, call = call)
      check_string(path, allow_empty = FALSE, call = call)
      check_bool(wait, call = call)
      check_bool(ignore_hash, call = call)

      self$chat <- chat
      self$user_turns <- user_turns
      self$type <- type
      self$path <- path
      self$should_wait <- wait
      self$ignore_hash <- ignore_hash

      if (file.exists(path)) {
        state <- jsonlite::read_json(path)
        self$stage <- state$stage
        self$batch <- state$batch
        self$results <- state$results
        self$started_at <- .POSIXct(state$started_at)

        self$check_hash(state$hash, call = call)
      } else {
        self$stage <- "submitting"
        self$batch <- NULL
        self$started_at <- Sys.time()
      }
    },

    save_state = function() {
      jsonlite::write_json(
        list(
          version = 1,
          stage = self$stage,
          batch = self$batch,
          results = self$results,
          started_at = as.integer(self$started_at),
          hash = self$compute_hash()
        ),
        self$path,
        auto_unbox = TRUE,
        pretty = TRUE
      )
    },

    step = function() {
      if (self$stage == "submitting") {
        self$submit()
      } else if (self$stage == "waiting") {
        self$wait()
      } else if (self$stage == "retrieving") {
        self$retrieve()
      } else {
        cli::cli_abort("Unknown stage: {self$stage}", .internal = TRUE)
      }
    },

    step_until_done = function() {
      while (self$stage != "done") {
        if (!self$step()) {
          return(invisible())
        }
      }
      invisible(self)
    },

    submit = function() {
      existing <- self$chat$get_turns(include_system_prompt = TRUE)
      conversations <- append_turns(list(existing), self$user_turns)

      self$batch <- batch_submit(self$provider, conversations, type = self$type)
      self$stage <- "waiting"
      self$save_state()
      TRUE
    },

    wait = function() {
      # always poll once, even when wait = FALSE
      status <- self$poll()

      if (self$should_wait) {
        cli::cli_progress_bar(
          format = paste(
            "{cli::pb_spin} Processing...",
            "[{self$elapsed()}]",
            "{status$n_processing} pending |",
            "{cli::col_green({status$n_succeeded})} done |",
            "{cli::col_red({status$n_failed})} failed"
          )
        )
        while (status$working) {
          Sys.sleep(0.5)
          cli::cli_progress_update()
          status <- self$poll()
        }
        cli::cli_progress_done()
      }

      if (!status$working) {
        self$stage <- "retrieving"
        self$save_state()
        TRUE
      } else {
        FALSE
      }
    },
    poll = function() {
      self$batch <- batch_poll(self$provider, self$batch)
      self$save_state()

      batch_status(self$provider, self$batch)
    },
    elapsed = function() {
      pretty_sec(as.integer(Sys.time()) - as.integer(self$started_at))
    },

    retrieve = function() {
      self$results <- batch_retrieve(self$provider, self$batch)
      log_turns(self$provider, self$result_turns())

      self$stage <- "done"
      self$save_state()
      TRUE
    },

    result_turns = function() {
      if (length(self$results) != length(self$user_turns)) {
        cli::cli_abort(c(
          "Provider returned unexpected number of responses.",
          x = "Expected {length(self$user_turns)}, got {length(self$results)}."
        ))
      }
      map2(self$results, self$user_turns, function(result, user_turn) {
        batch_result_turn(self$provider, result, has_type = !is.null(self$type))
      })
    },

    compute_hash = function() {
      # TODO: replace with JSON serialization when available
      list(
        provider = hash(provider_hash(self$provider)),
        prompts = hash(lapply(self$user_turns, format)),
        user_turns = hash(lapply(self$chat$get_turns(TRUE), format))
      )
    },

    check_hash = function(old_hash, call = caller_env()) {
      new_hash <- self$compute_hash()
      same <- map2_lgl(old_hash, new_hash, `==`)

      if (all(same)) {
        return(invisible())
      }
      differences <- names(new_hash)[!same]

      if (self$ignore_hash) {
        cli::cli_warn(
          c("!" = "{differences} {?does/do}n't match stored value{?s}."),
          call = call
        )
      } else {
        cli::cli_abort(
          c(
            "{differences} {?does/do}n't match stored value{?s}.",
            i = "Do you need to pick a different {.arg path}?",
            i = "Or set {.code ignore_hash = TRUE} to ignore this check?"
          ),
          call = call
        )
      }
    }
  )
)

provider_hash <- function(x) {
  list(
    name = x@name,
    model = x@model,
    base_url = x@base_url
  )
}

check_has_batch_support <- function(provider, call = caller_env()) {
  if (has_batch_support(provider)) {
    return(invisible())
  }

  cli::cli_abort(
    "Batch requests are not currently supported by this provider.",
    call = call
  )
}

test_that("can get chats/data from completed request", {
  chat <- chat_openai_test()

  prompts <- list(
    "What's the capital of Iowa?",
    "What's the capital of New York?",
    "What's the capital of California?",
    "What's the capital of Texas?"
  )
  chats <- batch_chat(
    chat,
    prompts,
    path = test_path("batch/state-capitals.json")
  )
  expect_length(chats, 4)

  out <- batch_chat_text(
    chat,
    prompts,
    path = test_path("batch/state-capitals.json")
  )
  expect_equal(
    gsub("\\.$", "", out),
    c("Des Moines", "Albany", "Sacramento", "Austin")
  )

  type_state <- type_object(name = type_string("State name"))
  data <- batch_chat_structured(
    chat,
    prompts,
    path = test_path("batch/state-name.json"),
    type = type_state
  )
  expect_equal(nrow(data), 4)
})

test_that("tokens only logged on first retrieval", {
  local_tokens()

  prompts <- list(
    "What's the capital of Iowa?",
    "What's the capital of New York?",
    "What's the capital of California?",
    "What's the capital of Texas?"
  )
  suppressWarnings(batch_chat(
    chat_openai_test(),
    prompts,
    path = test_path("batch/state-capitals.json"),
    ignore_hash = TRUE
  ))
  expect_equal(nrow(the$tokens), 0)
})

test_that("errors if chat/provider/prompts don't match previous run", {
  chat <- chat_anthropic_test(system_prompt = "Be cool")
  prompts <- list("What's the capital of Iowa?")
  path <- test_path("batch/state-capitals.json")
  expect_snapshot(batch_chat(chat, prompts, path), error = TRUE)

  expect_snapshot(
    batch_chat_structured(chat, prompts, path, type = type_string()),
    error = TRUE
  )
})

test_that("can override hash check", {
  chat <- chat_openai_test(system_prompt = "Be cool")
  prompts <- list("a", "b", "c", "d")
  path <- test_path("batch/state-capitals.json")
  expect_snapshot(. <- batch_chat(chat, prompts, path, ignore_hash = TRUE))
})

test_that("steps through in logical order, writing to disk at end step", {
  chat <- chat_openai_test()
  prompts <- list("What's your name")
  local_mocked_bindings(
    batch_submit = function(...) list(id = "123"),
    batch_poll = function(...) list(id = "123", results = TRUE),
    batch_status = function(...) list(working = FALSE),
    batch_retrieve = function(...) list(x = 1),
    batch_result_turn = function(...) list()
  )

  path <- withr::local_tempfile()
  read_stage <- function() jsonlite::read_json(path)$stage

  job <- BatchJob$new(chat, prompts, path)
  completed <- \() batch_chat_completed(chat, prompts, path)

  expect_equal(job$stage, "submitting")
  expect_false(completed())

  job$step()
  expect_equal(job$stage, "waiting")
  expect_equal(read_stage(), "waiting")
  expect_equal(job$batch, list(id = "123"))
  expect_true(completed())

  job$step()
  expect_equal(job$stage, "retrieving")
  expect_equal(read_stage(), "retrieving")
  expect_equal(job$batch, list(id = "123", results = TRUE))
  expect_true(completed())

  job$step()
  expect_equal(job$stage, "done")
  expect_equal(read_stage(), "done")
  expect_equal(job$results, list(x = 1))
  expect_true(completed())
})

test_that("can run all steps at once", {
  local_mocked_bindings(
    batch_submit = function(...) list(id = "123"),
    batch_poll = function(...) list(id = "123", results = TRUE),
    batch_status = function(...) list(working = FALSE),
    batch_retrieve = function(...) list(x = 1),
    batch_result_turn = function(...) list()
  )

  path <- withr::local_tempfile()
  job <- BatchJob$new(
    chat = chat_openai_test(),
    prompts = list("What's your name"),
    path = path
  )
  job$step_until_done()
  expect_equal(job$stage, "done")
  expect_equal(job$results, list(x = 1))
})

test_that("errors if wait = FALSE and not complete", {
  local_mocked_bindings(
    batch_submit = function(...) list(id = "123"),
    batch_poll = function(...) list(id = "123", results = TRUE),
    batch_status = function(...) list(working = TRUE)
  )

  path <- withr::local_tempfile()
  job <- BatchJob$new(
    chat = chat_openai_test(),
    prompts = list("What's your name"),
    path = path,
    wait = FALSE
  )
  expect_equal(job$step_until_done(), NULL)
})

test_that("informative error for bad inputs", {
  chat_openai <- chat_openai_test()
  chat_ollama <- chat_openai_test()
  chat_ollama$.__enclos_env__$private$provider <- ProviderOllama(
    "ollama",
    "model",
    "base_url"
  )

  expect_snapshot(error = TRUE, {
    batch_chat(1)
    batch_chat(chat_ollama)
    batch_chat(chat_openai, "a")
    batch_chat(chat_openai, list("a"), path = 1)
    batch_chat(chat_openai, list("a"), path = "x", wait = 1)
  })
})

#' Chat with a local Ollama model
#'
#' @description
#' To use `chat_ollama()` first download and install
#' [Ollama](https://ollama.com). Then install some models either from the
#' command line (e.g. with `ollama pull llama3.1`) or within R using
#' {[ollamar](https://hauselin.github.io/ollama-r/)} (e.g.
#' `ollamar::pull("llama3.1")`).
#'
#' Built on top of [chat_openai_compatible()].
#'
#' ## Known limitations
#'
#' * Tool calling is not supported with streaming (i.e. when `echo` is
#'   `"text"` or `"all"`)
#' * Models can only use 2048 input tokens, and there's no way
#'   to get them to use more, except by creating a custom model with a
#'   different default.
#' * Tool calling generally seems quite weak, at least with the models I have
#'   tried it with.
#'
#' @inheritParams chat_openai
#' @param model `r param_model(NULL, "ollama")`
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials Ollama doesn't require credentials for local usage and in most
#'   cases you do not need to provide `credentials`.
#'
#'   However, if you're accessing an Ollama instance hosted behind a reverse
#'   proxy or secured endpoint that enforces bearer‐token authentication, you
#'   can set the `OLLAMA_API_KEY` environment variable or provide a callback
#'   function to `credentials`.
#' @param params Common model parameters, usually created by [params()].
#' @inherit chat_openai return
#' @family chatbots
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_ollama(model = "llama3.2")
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_ollama <- function(
  system_prompt = NULL,
  base_url = Sys.getenv("OLLAMA_BASE_URL", "http://localhost:11434"),
  model,
  params = NULL,
  api_args = list(),
  echo = NULL,
  api_key = NULL,
  credentials = NULL,
  api_headers = character()
) {
  # ollama doesn't require an API key for local usage, but one might be needed
  # if ollama is served behind a proxy (see #501)
  credentials <- ollama_credentials(credentials, api_key)

  if (!has_ollama(base_url, credentials)) {
    cli::cli_abort("Can't find locally running ollama.")
  }

  models <- models_ollama(base_url, credentials)$id

  if (missing(model)) {
    cli::cli_abort(c(
      "Must specify {.arg model}.",
      i = "Locally installed models: {.str {models}}."
    ))
  } else if (!model %in% models) {
    cli::cli_abort(
      c(
        "Model {.val {model}} is not installed locally.",
        i = "Run {.code ollama pull {model}} in your terminal or {.run ollamar::pull(\"{model}\")} in R to install the model.",
        i = "See locally installed models with {.run ellmer::models_ollama()}."
      )
    )
  }

  echo <- check_echo(echo)

  provider <- ProviderOllama(
    name = "Ollama",
    base_url = file.path(base_url, "v1"), ## the v1 portion of the path is added for openAI compatible API
    model = model,
    params = params %||% params(),
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )

  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderOllama <- new_class(
  "ProviderOllama",
  parent = ProviderOpenAICompatible,
  properties = list(
    model = prop_string()
  )
)

ollama_credentials <- function(credentials = NULL, api_key = NULL) {
  as_credentials(
    "chat_ollama",
    function() Sys.getenv("OLLAMA_API_KEY", ""),
    credentials = credentials,
    api_key = api_key
  )
}

method(chat_params, ProviderOllama) <- function(provider, params) {
  # https://github.com/ollama/ollama/blob/main/docs/api.md#generate-a-chat-completion
  standardise_params(
    params,
    c(
      frequency_penalty = "frequency_penalty",
      presence_penalty = "presence_penalty",
      seed = "seed",
      stop = "stop_sequences",
      temperature = "temperature",
      top_p = "top_p",
      top_k = "top_k",
      max_tokens = "max_tokens"
    )
  )
}

chat_ollama_test <- function(..., model = "qwen3:4b", echo = "none") {
  # model: Note that tests require a model with tool capabilities

  skip_if_no_ollama()
  testthat::skip_if_not(
    model %in% models_ollama()$id,
    sprintf("Ollama: model '%s' is not installed", model)
  )

  chat_ollama(..., model = model, echo = echo)
}

skip_if_no_ollama <- function() {
  if (!has_ollama()) {
    testthat::skip("ollama not found")
  }
}

#' @export
#' @rdname chat_ollama
models_ollama <- function(
  base_url = "http://localhost:11434",
  credentials = NULL
) {
  credentials <- as_credentials(
    "models_ollama",
    function() Sys.getenv("OLLAMA_API_KEY", ""),
    credentials = credentials
  )

  req <- request(base_url)
  req <- ellmer_req_credentials(req, credentials(), "Authorization")
  req <- req_url_path_append(req, "api/tags")
  resp <- req_perform(req)
  json <- resp_body_json(resp)

  names <- map_chr(json$models, "[[", "name")
  names <- gsub(":latest$", "", names)

  modified_at <- as.POSIXct(map_chr(json$models, "[[", "modified_at"))
  size <- map_dbl(json$models, "[[", "size")

  df <- data.frame(
    id = names,
    created_at = modified_at,
    size = size,
    capabilities = ollama_model_capabilities(base_url, names, credentials)
  )
  df[order(-xtfrm(df$created_at)), ]
}

the$ollama_cache <- new_environment()

ollama_model_details <- function(
  base_url,
  model,
  credentials = ollama_credentials()
) {
  # https://github.com/ollama/ollama/blob/main/docs/api.md#show-model-information
  if (env_has(the$ollama_cache, model)) {
    return(the$ollama_cache[[model]])
  }

  req <- request(base_url)
  req <- ellmer_req_credentials(req, credentials(), "Authorization")
  req <- req_url_path_append(req, "api/show")
  req <- req_body_json(req, list(model = model, verbose = FALSE))

  resp <- req_perform(req)

  details <- resp_body_json(resp)

  # Cache model information (very unlikely to change during a session)
  the$ollama_cache[[model]] <- details
  details
}

ollama_model_capabilities <- function(
  base_url,
  models,
  credentials = ollama_credentials()
) {
  res <- map(models, function(m) {
    tryCatch(
      ollama_model_details(base_url, m, credentials),
      error = function(e) NULL
    )
  })
  map_chr(res, \(x) paste(x$capabilities, collapse = ","))
}


has_ollama <- function(
  base_url = "http://localhost:11434",
  credentials = ollama_credentials()
) {
  check_credentials(credentials)

  tryCatch(
    {
      req <- request(base_url)
      req <- ellmer_req_credentials(req, credentials(), "Authorization")
      req <- req_url_path_append(req, "api/tags")
      req_perform(req)
      TRUE
    },
    httr2_error = function(cnd) FALSE
  )
}

method(as_json, list(ProviderOllama, TypeObject)) <- function(
  provider,
  x,
  ...
) {
  if (x@additional_properties) {
    cli::cli_abort("{.arg .additional_properties} not supported for Ollama.")
  }

  # Unlike OpenAI, Ollama uses the `required` field to list required tool args
  required <- map_lgl(x@properties, function(prop) prop@required)

  compact(list(
    type = "object",
    description = x@description %||% "",
    properties = as_json(provider, x@properties, ...),
    required = as.list(names2(x@properties)[required]),
    additionalProperties = FALSE
  ))
}

#' Chat with any provider
#'
#' This is a generic interface to all the other `chat_` functions that allow
#' to you pick the provider and the model with a simple string.
#'
#' @inheritParams chat_openai
#' @param name Provider (and optionally model) name in the form
#'   `"provider/model"` or `"provider"` (which will use the default model
#'   for that provider).
#' @param ... Arguments passed to the provider function.
#' @rdname chat-any
#' @export
chat <- function(
  name,
  ...,
  system_prompt = NULL,
  params = NULL,
  echo = c("none", "output", "all")
) {
  check_string(name, allow_empty = FALSE)
  pieces <- strsplit(name, "/", fixed = TRUE)[[1]]

  if (length(pieces) == 1) {
    provider <- pieces[[1]]
    model <- NULL
  } else {
    provider <- pieces[[1]]
    model <- paste(pieces[-1], collapse = "/")
  }

  provider_fn_name <- paste0("chat_", provider)
  chat_fun <- env_get(asNamespace("ellmer"), provider_fn_name, default = NULL)
  if (is.null(chat_fun)) {
    cli::cli_abort("Can't find provider {.code ellmer::{provider_fn_name}()}.")
  }

  required_params <- c("model", "system_prompt", "params")
  if (any(!required_params %in% fn_fmls_names(chat_fun))) {
    cli::cli_abort(
      "{.fn ellmer::chat} does not support {.fn ellmer::{provider_fn_name}}, please call it directly.",
    )
  }

  chat_fun(
    model = model,
    ...,
    system_prompt = system_prompt,
    params = params,
    echo = check_echo(echo)
  )
}

#' @include provider-openai-compatible.R
#' @include content.R
NULL

#' Chat with a local LM Studio model
#'
#' @description
#' To use `chat_lmstudio()` first download and install
#' [LM Studio](https://lmstudio.ai). Then load a model using the LM Studio
#' GUI and start the local server. To learn more about running LM Studio
#' locally, see <https://lmstudio.ai/docs/developer/core/server>/.
#'
#' Built on top of [chat_openai_compatible()].
#'
#' @examples
#' \dontrun{
#' # https://lmstudio.ai/models/zai-org/glm-4.7-flash
#' chat <- chat_lmstudio(model = "zai-org/glm-4.7-flash")
#' chat$chat("Tell me three jokes about statisticians")
#' }
#'
#' @inheritParams chat_openai
#' @param model `r param_model(NULL, "lmstudio")`
#' @param credentials LM Studio doesn't require credentials for local usage
#'   and in most cases you do not need to provide `credentials`.
#'
#'   However, if you're accessing an LM Studio instance hosted behind a
#'   reverse proxy or secured endpoint that enforces bearer-token
#'   authentication, you can set the `LMSTUDIO_API_KEY` environment variable
#'   or provide a callback function to `credentials`.
#' @param params Common model parameters, usually created by [params()].
#'
#' @inherit chat_openai return
#'
#' @family chatbots
#' @export
chat_lmstudio <- function(
  system_prompt = NULL,
  base_url = Sys.getenv("LMSTUDIO_BASE_URL", "http://localhost:1234"),
  model,
  params = NULL,
  api_args = list(),
  echo = NULL,
  credentials = NULL,
  api_headers = character()
) {
  credentials <- lmstudio_credentials(credentials)

  if (!has_lmstudio(base_url, credentials)) {
    cli::cli_abort("Can't find locally running LM Studio.")
  }

  models <- models_lmstudio(base_url, credentials)$id

  if (missing(model)) {
    cli::cli_abort(c(
      "Must specify {.arg model}.",
      i = "Locally available models: {.str {models}}."
    ))
  } else if (!model %in% models) {
    cli::cli_abort(
      c(
        "Model {.val {model}} is not available in LM Studio.",
        i = "Download the model using the LM Studio GUI.",
        i = "See locally available models with {.run ellmer::models_lmstudio()}."
      )
    )
  }

  echo <- check_echo(echo)

  provider <- ProviderLMStudio(
    name = "LM Studio",
    base_url = file.path(base_url, "v1"),
    model = model,
    params = params %||% params(),
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )

  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderLMStudio <- new_class(
  "ProviderLMStudio",
  parent = ProviderOpenAICompatible,
  properties = list(
    model = prop_string()
  )
)

lmstudio_credentials <- function(credentials = NULL) {
  as_credentials(
    "chat_lmstudio",
    function() Sys.getenv("LMSTUDIO_API_KEY", ""),
    credentials = credentials
  )
}

method(chat_params, ProviderLMStudio) <- function(provider, params) {
  # https://lmstudio.ai/docs/developer/openai-compat/chat-completions#supported-payload-parameters
  standardise_params(
    params,
    c(
      frequency_penalty = "frequency_penalty",
      max_tokens = "max_tokens",
      presence_penalty = "presence_penalty",
      seed = "seed",
      stop = "stop_sequences",
      temperature = "temperature",
      top_k = "top_k",
      top_p = "top_p"
    )
  )
}

chat_lmstudio_test <- function(..., model = NULL, echo = "none") {
  skip_if_no_lmstudio()
  if (is.null(model)) {
    models <- models_lmstudio()$id
    if (length(models) == 0) {
      testthat::skip("No models loaded in LM Studio")
    }
    models <- union(
      # Prefer gemma4 or glm-4.7-flash if they're available
      intersect(c("google/gemma-4-26b-a4b", "zai-org/glm-4.7-flash"), models),
      models
    )
    model <- models[[1]]
  }
  chat_lmstudio(..., model = model, echo = echo)
}

skip_if_no_lmstudio <- function() {
  if (!has_lmstudio()) {
    testthat::skip("LM Studio not found")
  }
}

#' @export
#' @rdname chat_lmstudio
models_lmstudio <- function(
  base_url = "http://localhost:1234",
  credentials = NULL
) {
  credentials <- as_credentials(
    "models_lmstudio",
    function() Sys.getenv("LMSTUDIO_API_KEY", ""),
    credentials = credentials
  )

  req <- request(base_url)
  req <- ellmer_req_credentials(req, credentials(), "Authorization")
  req <- req_url_path_append(req, "/v1/models")
  resp <- req_perform(req)
  json <- resp_body_json(resp)

  data.frame(
    id = map_chr(json$data, "[[", "id")
  )
}

has_lmstudio <- function(
  base_url = "http://localhost:1234",
  credentials = lmstudio_credentials()
) {
  check_credentials(credentials)

  tryCatch(
    {
      req <- request(base_url)
      req <- ellmer_req_credentials(req, credentials(), "Authorization")
      req <- req_url_path_append(req, "/v1/models")
      req_perform(req)
      TRUE
    },
    httr2_error = function(cnd) FALSE
  )
}

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_lmstudio_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_lmstudio_test("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

test_that("can list models", {
  skip_if_no_lmstudio()
  test_models(models_lmstudio)
})

test_that("includes list of models in error message if `model` is missing", {
  skip_if_no_lmstudio()

  local_mocked_bindings(
    models_lmstudio = function(...) {
      list(id = c("llama3", "google/gemma-4-26b-a4b"))
    }
  )

  expect_snapshot(chat_lmstudio(), error = TRUE)
})

test_that("checks that requested model is loaded", {
  skip_if_no_lmstudio()
  local_mocked_bindings(
    models_lmstudio = function(...) list(id = "llama3")
  )
  expect_snapshot(
    chat_lmstudio(model = "not-a-real-model"),
    error = TRUE
  )
})

# Common provider interface -----------------------------------------------

test_that("supports tool calling", {
  chat_fun <- chat_lmstudio_test
  test_tools_simple(chat_fun)
})

#' Chat with a model hosted on Hugging Face Serverless Inference API
#'
#' @description
#' [Hugging Face](https://huggingface.co/) hosts a variety of open-source
#' and proprietary AI models available via their Inference API.
#' To use the Hugging Face API, you must have an Access Token, which you can obtain
#' from your [Hugging Face account](https://huggingface.co/settings/tokens)
#' (ensure that at least "Make calls to Inference Providers" and
#' "Make calls to your Inference Endpoints" is checked).
#'
#' Built on top of [chat_openai_compatible()].
#'
#' ## Known limitations
#'
#' * Some models do not support the chat interface or parts of it, for example
#'   `google/gemma-2-2b-it` does not support a system prompt. You will need to
#'   carefully choose the model.
#'
#' @family chatbots
#' @param model `r param_model("meta-llama/Llama-3.1-8B-Instruct")`
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("HUGGINGFACE_API_KEY")`
#' @export
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_huggingface()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_huggingface <- function(
  system_prompt = NULL,
  params = NULL,
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
) {
  model <- set_default(model, "meta-llama/Llama-3.1-8B-Instruct")
  echo <- check_echo(echo)
  params <- params %||% params()

  credentials <- as_credentials(
    "chat_huggingface",
    function() hf_key(),
    credentials = credentials,
    api_key = api_key
  )

  # https://huggingface.co/docs/inference-providers/en/index?python-clients=requests#http--curl
  base_url <- "https://router.huggingface.co/v1/"

  provider <- ProviderHuggingFace(
    name = "HuggingFace",
    base_url = base_url,
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderHuggingFace <- new_class(
  "ProviderHuggingFace",
  parent = ProviderOpenAICompatible
)

chat_huggingface_test <- function(
  system_prompt = "Be terse",
  ...,
  model = NULL,
  echo = "none"
) {
  model <- model %||% "meta-llama/Llama-3.1-8B-Instruct"
  chat_huggingface(model = model, ..., echo = echo)
}

hf_key <- function() {
  key_get("HUGGINGFACE_API_KEY")
}

#' Standard model parameters
#'
#' @description
#' This helper function makes it easier to create a list of parameters used
#' across many models. The parameter names are automatically standardised and
#' included in the correctly place in the API call.
#'
#' Note that parameters that are not supported by a given provider will generate
#' a warning, not an error. This allows you to use the same set of parameters
#' across multiple providers.
#'
#' @param temperature Temperature of the sampling distribution.
#' @param top_p The cumulative probability for token selection.
#' @param top_k The number of highest probability vocabulary tokens to keep.
#' @param frequency_penalty Frequency penalty for generated tokens.
#' @param presence_penalty Presence penalty for generated tokens.
#' @param seed Seed for random number generator.
#' @param max_tokens Maximum number of tokens to generate.
#' @param log_probs Include the log probabilities in the output?
#' @param stop_sequences A character vector of tokens to stop generation on.
#' @param reasoning_effort,reasoning_tokens How much effort to spend thinking?
#'   `ressoning_effort` is a string, like "low", "medium", "high".
#'   `reasoning_tokens` is an integer, giving a maximum token budget.
#'   Each provider only takes one of these two parameters.
#' @param ... Additional named parameters to send to the provider.
#' @export
params <- function(
  temperature = NULL,
  top_p = NULL,
  top_k = NULL,
  frequency_penalty = NULL,
  presence_penalty = NULL,
  seed = NULL,
  max_tokens = NULL,
  log_probs = NULL,
  stop_sequences = NULL,
  reasoning_effort = NULL,
  reasoning_tokens = NULL,
  ...
) {
  check_number_decimal(temperature, allow_null = TRUE, min = 0)
  check_number_decimal(top_p, allow_null = TRUE, min = 0)
  check_number_whole(top_k, allow_null = TRUE, min = 0)
  check_number_decimal(frequency_penalty, allow_null = TRUE)
  check_number_decimal(presence_penalty, allow_null = TRUE)
  check_number_whole(seed, allow_null = TRUE)
  check_number_whole(max_tokens, allow_null = TRUE, min = 1)
  check_bool(log_probs, allow_null = TRUE)
  check_character(stop_sequences, allow_null = TRUE)
  check_string(reasoning_effort, allow_null = TRUE)
  check_number_whole(reasoning_tokens, min = 0, allow_null = TRUE)

  compact(list2(
    temperature = temperature,
    top_p = top_p,
    top_k = top_k,
    frequency_penalty = frequency_penalty,
    presence_penalty = presence_penalty,
    seed = seed,
    max_tokens = max_tokens,
    log_probs = log_probs,
    stop_sequences = stop_sequences,
    reasoning_effort = reasoning_effort,
    reasoning_tokens = reasoning_tokens,
    extra_args = list2(...)
  ))
}

standardise_params <- function(params, provider_params) {
  standard <- params[names(params) != "extra_args"]

  unknown <- setdiff(names(standard), provider_params)
  if (length(unknown) > 0) {
    cli::cli_warn("Ignoring unsupported parameters: {.str {unknown}}")
    standard <- standard[names(standard) %in% provider_params]
  }

  names(standard) <- names(provider_params)[match(
    names(standard),
    provider_params
  )]

  c(standard, params$extra_args)
}

## To implement a `chat_params()` method for a new provider, use this template:
# method(chat_params, ProviderNew) <- function(provider, params) {
#   # <link to api docs>
#   standardise_params(
#     params,
#     c(
#       paramFromAPI = "ellmer_params_name",
#       maxTokens = "max_tokens",
#       # ... all supported parameters that overlap with `params()` ...
#     )
#   )
# }

# params -------------------------------------------------------------------

test_that("NULL values are stripped", {
  expect_equal(params(), set_names(list()))
})

test_that("checks its inputs", {
  expect_snapshot(error = TRUE, {
    params(temperature = "x")
    params(top_p = "x")
    params(top_k = "x")
    params(frequency_penalty = "x")
    params(presence_penalty = "x")
    params(seed = "x")
    params(max_tokens = "x")
    params(log_probs = 1)
    params(stop_sequences = 1)
  })
})

# standardise_params -------------------------------------------------------

test_that("standardise_params warns about unknown args", {
  test_params <- list(temperature = 0.7, top_p = 0.9)
  provider_params <- c("temperature" = "temperature")
  expect_snapshot(. <- standardise_params(test_params, provider_params))
})

test_that("standardise_params renames supported parameters", {
  test_params <- list(
    top_p = 0.9,
    temperature = 0.7,
    max_tokens = 100
  )

  provider_params <- c(
    temp = "temperature",
    topP = "top_p",
    maxTokens = "max_tokens"
  )
  expect_equal(
    standardise_params(test_params, provider_params),
    list(topP = 0.9, temp = 0.7, maxTokens = 100)
  )
})

test_that("standardise_params handles empty parameters correctly", {
  test_params <- list(extra_args = list())
  provider_params <- c("temperature" = "temperature", "top_p" = "top_p")
  expect_equal(standardise_params(test_params, provider_params), list())
})

test_that("standardise_params leavees extra_args as is", {
  test_params <- list(top_k = 2, extra_args = list(a = 1, b = 2))

  provider_params <- c(top_k = "top_k")
  expect_equal(
    standardise_params(test_params, provider_params),
    list(top_k = 2, a = 1, b = 2)
  )
})

#' @include utils-coro.R
NULL

#' The Chat object
#'
#' @description
#' A `Chat` is a sequence of user and assistant [Turn]s sent
#' to a specific [Provider]. A `Chat` is a mutable R6 object that takes care of
#' managing the state associated with the chat; i.e. it records the messages
#' that you send to the server, and the messages that you receive back.
#' If you register a tool (i.e. an R function that the assistant can call on
#' your behalf), it also takes care of the tool loop.
#'
#' You should generally not create this object yourself,
#' but instead call [chat_openai()] or friends instead.
#'
#' @return A Chat object
#' @examples
#' \dontshow{ellmer:::vcr_example_start("Chat")}
#' chat <- chat_openai()
#' chat$chat("Tell me a funny joke")
#' \dontshow{ellmer:::vcr_example_end()}
Chat <- R6::R6Class(
  "Chat",
  public = list(
    #' @param provider A provider object.
    #' @param system_prompt System prompt to start the conversation with.
    #' @param echo One of the following options:
    #'   * `none`: don't emit any output (default when running in a function).
    #'   * `output`: echo text and tool-calling output as it streams in (default
    #'     when running at the console).
    #'   * `all`: echo all input and output.
    #'
    #'  Note this only affects the `chat()` method. You can override the default
    #'  by setting the `ellmer_echo` option.
    initialize = function(provider, system_prompt = NULL, echo = "none") {
      private$provider <- provider
      private$echo <- echo
      private$callback_on_tool_request <- CallbackManager$new(args = "request")
      private$callback_on_tool_result <- CallbackManager$new(args = "result")
      self$set_system_prompt(system_prompt)
    },

    #' @description Retrieve the turns that have been sent and received so far
    #'   (optionally starting with the system prompt, if any).
    #' @param include_system_prompt Whether to include the system prompt in the
    #'   turns (if any exists).
    get_turns = function(include_system_prompt = FALSE) {
      if (length(private$.turns) == 0) {
        return(private$.turns)
      }

      if (!include_system_prompt && is_system_turn(private$.turns[[1]])) {
        private$.turns[-1]
      } else {
        private$.turns
      }
    },

    #' @description Replace existing turns with a new list.
    #' @param value A list of [Turn]s.
    set_turns = function(value) {
      private$.turns <- normalize_turns(
        value,
        self$get_system_prompt(),
        overwrite = TRUE
      )
      invisible(self)
    },

    #' @description Add a pair of turns to the chat.
    #' @param user The user [Turn].
    #' @param assistant The system [Turn].
    #' @param log_tokens Should tokens used in the turn be logged to the
    #'   session counter?
    add_turn = function(user, assistant, log_tokens = TRUE) {
      check_turn(user)
      check_turn(assistant)

      if (log_tokens) {
        log_turn(private$provider, assistant)
      }

      private$.turns[[length(private$.turns) + 1]] <- user
      private$.turns[[length(private$.turns) + 1]] <- assistant
      invisible(self)
    },

    #' @description If set, the system prompt, it not, `NULL`.
    get_system_prompt = function() {
      if (private$has_system_prompt()) {
        private$.turns[[1]]@text
      } else {
        NULL
      }
    },

    #' @description Retrieve the model name
    get_model = function() {
      private$provider@model
    },

    #' @description Update the system prompt
    #' @param value A character vector giving the new system prompt
    set_system_prompt = function(value) {
      check_character(value, allow_null = TRUE)
      if (length(value) > 1) {
        value <- paste(value, collapse = "\n\n")
      }

      # Remove prompt, if present
      if (private$has_system_prompt()) {
        private$.turns <- private$.turns[-1]
      }
      # Add prompt, if new
      if (is.character(value)) {
        system_turn <- SystemTurn(value)
        private$.turns <- c(list(system_turn), private$.turns)
      }
      invisible(self)
    },

    #' @description A data frame with token usage and cost data. There are four
    #'   columns: `input`, `output`, `cached_input`, and `cost`. There is one
    #'   row for each assistant turn, because token counts and costs are only
    #'   available when the API returns the assistant's response.
    #' @param include_system_prompt `r lifecycle::badge("deprecated")`
    get_tokens = function(include_system_prompt = deprecated()) {
      if (lifecycle::is_present(include_system_prompt)) {
        lifecycle::deprecate_warn(
          "0.4.0",
          "get_tokens(include_system_prompt)",
          "get_tokens()"
        )
      }

      turns <- self$get_turns()
      assistant_turns <- keep(turns, is_assistant_turn)
      complete_turns <- discard(assistant_turns, is_partial_turn)
      tokens <- map_tokens(complete_turns, \(turn) turn@tokens)
      tokens <- tibble::as_tibble(tokens)
      tokens$cost <- dollars(map_dbl(complete_turns, \(turn) turn@cost))

      user_turns <- keep(turns, is_user_turn)
      tokens$input_preview <- map_chr(user_turns, turn_contents_preview)
      tokens
    },

    #' @description The cost of this chat
    #' @param include The default, `"all"`, gives the total cumulative cost
    #'   of this chat. Alternatively, use `"last"` to get the cost of just the
    #'   most recent turn. Incomplete turns (from cancelled or interrupted
    #'   streams) are excluded because they lack token data.
    get_cost = function(include = c("all", "last")) {
      include <- arg_match(include)

      turns <- self$get_turns()
      assistant_turns <- keep(turns, is_assistant_turn)
      complete_turns <- discard(assistant_turns, is_partial_turn)

      if (length(complete_turns) == 0) {
        return(dollars(0))
      }

      if (include == "last") {
        cost <- complete_turns[[length(complete_turns)]]@cost
      } else {
        cost <- sum(map_dbl(complete_turns, \(turn) turn@cost))
      }

      dollars(cost)
    },

    #' @description The last turn returned by the assistant.
    #' @param role Optionally, specify a role to find the last turn with
    #'   for the role.
    #' @return Either a `Turn` or `NULL`, if no turns with the specified
    #'   role have occurred.
    last_turn = function(role = c("assistant", "user", "system")) {
      role <- arg_match(role)

      n <- length(private$.turns)
      switch(
        role,
        system = if (private$has_system_prompt()) private$.turns[[1]],
        assistant = if (n > 1) private$.turns[[n]],
        user = if (n > 1) private$.turns[[n - 1]]
      )
    },

    #' @description Submit input to the chatbot, and return the response as a
    #'   simple string (probably Markdown).
    #' @param ... The input to send to the chatbot. Can be strings or images
    #'   (see [content_image_file()] and [content_image_url()].
    #' @param echo Whether to emit the response to stdout as it is received. If
    #'   `NULL`, then the value of `echo` set when the chat object was created
    #'   will be used.
    chat = function(..., echo = NULL) {
      finish_tools <- private$complete_dangling_tool_requests()

      turn <- user_turn(!!!finish_tools, ...)
      echo <- check_echo(echo %||% private$echo)

      # Returns a single turn (the final response from the assistant), even if
      # multiple rounds of back and forth happened.
      coro::collect(private$chat_impl(
        turn,
        stream = echo != "none",
        echo = echo,
        controller = stream_controller()
      ))

      text <- ellmer_output(self$last_turn()@text)
      if (echo == "none") text else invisible(text)
    },

    #' @description Extract structured data
    #' @param ... The input to send to the chatbot. This is typically the text
    #'   you want to extract data from, but it can be omitted if the data is
    #'   obvious from the existing conversation.
    #' @param type A type specification for the extracted data. Should be
    #'   created with a [`type_()`][type_boolean] function.
    #' @param echo Whether to emit the response to stdout as it is received.
    #'   Set to "text" to stream JSON data as it's generated (not supported by
    #'   all providers).
    #' @param convert Automatically convert from JSON lists to R data types
    #'   using the schema. For example, this will turn arrays of objects into
    #'   data frames and arrays of strings into a character vector.
    chat_structured = function(..., type, echo = "none", convert = TRUE) {
      finish_tools <- private$complete_dangling_tool_requests()

      turn <- user_turn(!!!finish_tools, ..., .check_empty = FALSE)
      echo <- check_echo(echo %||% private$echo)
      check_bool(convert)

      needs_wrapper <- type_needs_wrapper(type, private$provider)
      type <- wrap_type_if_needed(type, needs_wrapper)

      coro::collect(private$submit_turns(
        turn,
        type = type,
        stream = echo != "none",
        echo = echo,
        controller = stream_controller()
      ))

      turn <- self$last_turn()
      extract_data(turn, type, convert = convert, needs_wrapper = needs_wrapper)
    },

    #' @description Extract structured data, asynchronously. Returns a promise
    #'   that resolves to an object matching the type specification.
    #' @param ... The input to send to the chatbot. Will typically include
    #'   the phrase "extract structured data".
    #' @param type A type specification for the extracted data. Should be
    #'   created with a [`type_()`][type_boolean] function.
    #' @param echo Whether to emit the response to stdout as it is received.
    #'   Set to "text" to stream JSON data as it's generated (not supported by
    #'   all providers).
    #' @param convert Automatically convert from JSON lists to R data types
    #'   using the schema. For example, this will turn arrays of objects into
    #'   data frames and arrays of strings into a character vector.
    chat_structured_async = function(..., type, echo = "none", convert = TRUE) {
      finish_tools <- private$complete_dangling_tool_requests()

      turn <- user_turn(!!!finish_tools, ..., .check_empty = FALSE)
      echo <- check_echo(echo %||% private$echo)
      check_bool(convert)

      needs_wrapper <- type_needs_wrapper(type, private$provider)
      type <- wrap_type_if_needed(type, needs_wrapper)

      done <- coro::async_collect(private$submit_turns_async(
        turn,
        type = type,
        stream = echo != "none",
        echo = echo,
        controller = stream_controller()
      ))

      promises::then(done, function(dummy) {
        turn <- self$last_turn()
        extract_data(
          turn,
          type,
          convert = convert,
          needs_wrapper = needs_wrapper
        )
      })
    },

    #' @description Submit input to the chatbot, and receive a promise that
    #'   resolves with the response all at once. Returns a promise that resolves
    #'   to a string (probably Markdown).
    #' @param ... The input to send to the chatbot. Can be strings or images.
    #' @param tool_mode Whether tools should be invoked one-at-a-time
    #'   (`"sequential"`) or concurrently (`"concurrent"`). Sequential mode is
    #'   best for interactive applications, especially when a tool may involve
    #'   an interactive user interface. Concurrent mode is the default and is
    #'   best suited for automated scripts or non-interactive applications.
    chat_async = function(..., tool_mode = c("concurrent", "sequential")) {
      finish_tools <- private$complete_dangling_tool_requests()

      turn <- user_turn(!!!finish_tools, ...)
      tool_mode <- arg_match(tool_mode)

      # Returns a single turn (the final response from the assistant), even if
      # multiple rounds of back and forth happened.
      done <- coro::async_collect(
        private$chat_impl_async(
          turn,
          stream = FALSE,
          echo = "none",
          tool_mode = tool_mode,
          controller = stream_controller()
        )
      )
      promises::then(done, function(dummy) {
        self$last_turn()@text
      })
    },

    #' @description Submit input to the chatbot, returning streaming results.
    #'   Returns A [coro
    #'   generator](https://coro.r-lib.org/articles/generator.html#iterating)
    #'   that yields strings. While iterating, the generator will block while
    #'   waiting for more content from the chatbot.
    #' @param ... The input to send to the chatbot. Can be strings or images.
    #' @param stream Whether the stream should yield only `"text"` or ellmer's
    #'   rich content types. When `stream = "content"`, `stream()` yields
    #'   [Content] objects.
    #' @param controller An optional [stream_controller()] used to cancel the
    #'   stream from outside the iteration loop.
    stream = function(..., stream = c("text", "content"), controller = NULL) {
      controller <- as_controller(controller)
      finish_tools <- private$complete_dangling_tool_requests()

      turn <- user_turn(!!!finish_tools, ...)
      stream <- arg_match(stream)
      private$chat_impl(
        turn,
        stream = TRUE,
        echo = "none",
        yield_as_content = stream == "content",
        controller = controller
      )
    },

    #' @description Submit input to the chatbot, returning asynchronously
    #'   streaming results. Returns a [coro async
    #'   generator](https://coro.r-lib.org/reference/async_generator.html) that
    #'   yields string promises.
    #' @param ... The input to send to the chatbot. Can be strings or images.
    #' @param tool_mode Whether tools should be invoked one-at-a-time
    #'   (`"sequential"`) or concurrently (`"concurrent"`). Sequential mode is
    #'   best for interactive applications, especially when a tool may involve
    #'   an interactive user interface. Concurrent mode is the default and is
    #'   best suited for automated scripts or non-interactive applications.
    #' @param stream Whether the stream should yield only `"text"` or ellmer's
    #'   rich content types. When `stream = "content"`, `stream()` yields
    #'   [Content] objects.
    #' @param controller An optional [stream_controller()] used to cancel the
    #'   stream from outside the iteration loop.
    stream_async = function(
      ...,
      tool_mode = c("concurrent", "sequential"),
      stream = c("text", "content"),
      controller = NULL
    ) {
      controller <- as_controller(controller)
      finish_tools <- private$complete_dangling_tool_requests()

      turn <- user_turn(!!!finish_tools, ...)
      tool_mode <- arg_match(tool_mode)
      stream <- arg_match(stream)
      private$chat_impl_async(
        turn,
        stream = TRUE,
        echo = "none",
        tool_mode = tool_mode,
        yield_as_content = stream == "content",
        controller = controller
      )
    },

    #' @description Register a tool (an R function) that the chatbot can use.
    #'   Learn more in `vignette("tool-calling")`.
    #' @param tool A tool definition created by [tool()].
    register_tool = function(tool) {
      check_tool(tool)
      if (has_name(private$tools, tool@name)) {
        cli::cli_inform("Replacing existing {tool@name} tool.")
      }

      private$tools[[tool@name]] <- tool
      invisible(self)
    },

    #' @description Register a list of tools.
    #'   Learn more in `vignette("tool-calling")`.
    #' @param tools A list of tool definitions created by [tool()].
    register_tools = function(tools) {
      check_tools(tools)

      for (tool in tools) {
        self$register_tool(tool)
      }
      invisible(self)
    },

    #' @description Get the underlying provider object. For expert use only.
    get_provider = function() {
      private$provider
    },

    #' @description Retrieve the list of registered tools.
    get_tools = function() {
      private$tools
    },

    #' @description Sets the available tools. For expert use only; most users
    #'   should use `register_tool()`.
    #'
    #' @param tools A list of tool definitions created with [ellmer::tool()].
    set_tools = function(tools) {
      check_tools(tools)

      private$tools <- list()
      for (tool_def in tools) {
        self$register_tool(tool_def)
      }
      invisible(self)
    },

    #' @description Register a callback for a tool request event.
    #'
    #' @param callback A function to be called when a tool request event occurs,
    #'   which must have `request` as its only argument.
    #'
    #' @return A function that can be called to remove the callback.
    on_tool_request = function(callback) {
      private$callback_on_tool_request$add(callback)
    },

    #' @description Register a callback for a tool result event.
    #'
    #' @param callback A function to be called when a tool result event occurs,
    #'   which must have `result` as its only argument.
    #'
    #' @return A function that can be called to remove the callback.
    on_tool_result = function(callback) {
      private$callback_on_tool_result$add(callback)
    }
  ),
  private = list(
    provider = NULL,

    .turns = list(),
    echo = NULL,
    tools = list(),
    callback_on_tool_request = NULL,
    callback_on_tool_result = NULL,

    # If stream = TRUE, yields completion deltas. If stream = FALSE, yields
    # complete assistant turns.
    chat_impl = generator_method(function(
      self,
      private,
      user_turn,
      stream,
      echo,
      yield_as_content = FALSE,
      controller = NULL
    ) {
      tool_errors <- list()
      defer(warn_tool_errors(tool_errors))

      agent_span <- local_agent_otel_span(private$provider, activate = FALSE)

      while (!is.null(user_turn)) {
        assistant_chunks <- private$submit_turns(
          user_turn,
          stream = stream,
          echo = echo,
          yield_as_content = yield_as_content,
          controller = controller,
          otel_span = agent_span
        )
        for (chunk in assistant_chunks) {
          yield(chunk)
        }

        assistant_turn <- self$last_turn()
        user_turn <- NULL

        # Don't invoke tools if the stream was cancelled
        if (controller$cancelled) {
          break
        }

        if (turn_has_tool_request(assistant_turn)) {
          tool_calls <- invoke_tools(
            assistant_turn,
            echo = echo,
            on_tool_request = private$callback_on_tool_request$invoke,
            on_tool_result = private$callback_on_tool_result$invoke,
            yield_request = yield_as_content,
            otel_span = agent_span
          )

          tool_results <- list()

          for (tool_step in tool_calls) {
            if (yield_as_content) {
              yield(tool_step)
            }
            if (is_tool_result(tool_step)) {
              tool_results <- c(tool_results, list(tool_step))
            }
          }

          user_turn <- tool_results_as_turn(tool_results)
        }

        if (echo == "all") {
          cat(format(user_turn))
        } else if (echo == "none") {
          tool_errors <- c(tool_errors, turn_get_tool_errors(user_turn))
        }
      }
    }),

    # If stream = TRUE, yields completion deltas. If stream = FALSE, yields
    # complete assistant turns.
    chat_impl_async = async_generator_method(function(
      self,
      private,
      user_turn,
      stream,
      echo,
      tool_mode = "concurrent",
      yield_as_content = FALSE,
      controller = NULL
    ) {
      tool_errors <- list()
      defer(warn_tool_errors(tool_errors))

      agent_span <- local_agent_otel_span(private$provider, activate = FALSE)

      while (!is.null(user_turn)) {
        assistant_chunks <- private$submit_turns_async(
          user_turn,
          stream = stream,
          echo = echo,
          yield_as_content = yield_as_content,
          controller = controller,
          otel_span = agent_span
        )
        for (chunk in await_each(assistant_chunks)) {
          yield(chunk)
        }

        assistant_turn <- self$last_turn()
        user_turn <- NULL

        # Don't invoke tools if the stream was cancelled
        if (controller$cancelled) {
          break
        }

        if (turn_has_tool_request(assistant_turn)) {
          tool_calls <- invoke_tools_async(
            assistant_turn,
            echo = echo,
            on_tool_request = private$callback_on_tool_request$invoke_async,
            on_tool_result = private$callback_on_tool_result$invoke_async,
            yield_request = yield_as_content,
            otel_span = agent_span
          )
          if (tool_mode == "sequential") {
            tool_results <- list()
            for (tool_step in await_each(tool_calls)) {
              if (yield_as_content) {
                yield(tool_step)
              }
              if (is_tool_result(tool_step)) {
                tool_results <- c(tool_results, list(tool_step))
              }
            }
          } else {
            tool_results <- coro::collect(tool_calls)
            if (yield_as_content) {
              # Filter out and yield tool requests before awaiting tool results
              is_request <- map_lgl(tool_results, is_tool_request)
              for (tool_step in tool_results[is_request]) {
                yield(tool_step)
              }
              tool_results <- tool_results[!is_request]
            }
            tool_results <- await(promises::promise_all(.list = tool_results))
            if (yield_as_content) {
              for (tool_result in tool_results) {
                yield(tool_result)
              }
            }
          }

          user_turn <- tool_results_as_turn(tool_results)
        }

        if (echo == "all") {
          cat(format(user_turn))
        } else if (echo == "none") {
          tool_errors <- c(tool_errors, turn_get_tool_errors(user_turn))
        }
      }
    }),

    # If stream = TRUE, yields completion deltas. If stream = FALSE, yields
    # complete assistant turns.
    submit_turns = generator_method(function(
      self,
      private,
      user_turn,
      stream,
      echo,
      type = NULL,
      yield_as_content = FALSE,
      controller = NULL,
      otel_span = NULL
    ) {
      if (echo == "all") {
        cat_line(format(user_turn), prefix = "> ")
      }

      otel_input <- otel_chat_input(private, user_turn)
      chat_span <- local_chat_otel_span(
        private$provider,
        turns = otel_input$turns,
        system_prompt = otel_input$system_prompt,
        parent = otel_span
      )

      response <- chat_perform(
        provider = private$provider,
        mode = if (stream) "stream" else "value",
        turns = c(private$.turns, list(user_turn)),
        tools = if (is.null(type)) private$tools,
        type = type,
        controller = controller,
        otel_span = chat_span
      )

      emit <- emitter(echo)
      any_text <- FALSE
      turn <- NULL
      acc <- TurnAccumulator$new(self, private, controller)

      if (stream) {
        acc$begin_turn(user_turn)
        on.exit(acc$finalize_turn(), add = TRUE)

        result <- NULL
        for (chunk in response) {
          content <- stream_content(private$provider, chunk)
          if (!is.null(content)) {
            text <- content_text(content)
            emit(text)
            yield(if (yield_as_content) content else text)
            acc$update_turn(content)
            any_text <- TRUE
          }

          result <- stream_merge_chunks(private$provider, result, chunk)
        }

        record_chat_otel_span_status(chat_span, private$provider, result)
        turn <- acc$complete_turn(result, type = type)
        record_chat_otel_span_output(chat_span, turn)
      } else {
        result <- resp_body_json(response)
        duration <- resp_timing(response)[["total"]] %||% NA_real_
        record_chat_otel_span_status(chat_span, private$provider, result)
        turn <- acc$add_turn(user_turn, result, duration, type = type)
        record_chat_otel_span_output(chat_span, turn)

        text <- turn@text
        if (!is.null(text)) {
          emit(text)
          if (yield_as_content) {
            yield(ContentText(text))
          } else {
            yield(text)
          }
          any_text <- TRUE
        }
      }

      if (!is.null(turn) && !is_partial_turn(turn)) {
        # Ensure turns always end in a newline
        if (any_text) {
          emit("\n")
          if (yield_as_content) {
            yield(ContentText("\n"))
          } else {
            yield("\n")
          }
        }

        if (echo == "all") {
          echo_non_text_contents(turn)
        }
        # When `echo="output"`, tool calls are emitted in `invoke_tools()`
      }

      coro::exhausted()
    }),

    # If stream = TRUE, yields completion deltas. If stream = FALSE, yields
    # complete assistant turns.
    submit_turns_async = async_generator_method(function(
      self,
      private,
      user_turn,
      stream,
      echo,
      type = NULL,
      yield_as_content = FALSE,
      controller = NULL,
      otel_span = NULL
    ) {
      otel_input <- otel_chat_input(private, user_turn)
      chat_span <- local_chat_otel_span(
        private$provider,
        turns = otel_input$turns,
        system_prompt = otel_input$system_prompt,
        parent = otel_span
      )

      response <- chat_perform(
        provider = private$provider,
        mode = if (stream) "async-stream" else "async-value",
        turns = c(private$.turns, list(user_turn)),
        tools = if (is.null(type)) private$tools,
        type = type,
        controller = controller,
        otel_span = chat_span
      )

      emit <- emitter(echo)
      any_text <- FALSE
      turn <- NULL
      acc <- TurnAccumulator$new(self, private, controller)

      if (stream) {
        acc$begin_turn(user_turn)
        on.exit(acc$finalize_turn(), add = TRUE)

        result <- NULL
        for (chunk in await_each(response)) {
          content <- stream_content(private$provider, chunk)
          if (!is.null(content)) {
            text <- content_text(content)
            emit(text)
            yield(if (yield_as_content) content else text)
            acc$update_turn(content)
            any_text <- TRUE
          }

          result <- stream_merge_chunks(private$provider, result, chunk)
        }

        record_chat_otel_span_status(chat_span, private$provider, result)
        turn <- acc$complete_turn(result, type = type)
        record_chat_otel_span_output(chat_span, turn)
      } else {
        response <- await(response)
        result <- resp_body_json(response)
        duration <- resp_timing(response)[["total"]] %||% NA_real_
        record_chat_otel_span_status(chat_span, private$provider, result)
        turn <- acc$add_turn(user_turn, result, duration, type = type)
        record_chat_otel_span_output(chat_span, turn)

        text <- turn@text
        if (!is.null(text)) {
          emit(text)
          if (yield_as_content) {
            yield(ContentText(text))
          } else {
            yield(text)
          }
          any_text <- TRUE
        }
      }

      if (!is.null(turn) && !is_partial_turn(turn)) {
        # Ensure turns always end in a newline
        if (any_text) {
          emit("\n")
          if (yield_as_content) {
            yield(ContentText("\n"))
          } else {
            yield("\n")
          }
        }

        if (echo == "all") {
          echo_non_text_contents(turn)
        }
        # When `echo="output"`, tool calls are echoed via `invoke_tools_async()`
      }
      coro::exhausted()
    }),

    has_system_prompt = function() {
      length(private$.turns) > 0 && is_system_turn(private$.turns[[1]])
    },

    complete_dangling_tool_requests = function() {
      if (length(private$.turns) == 0) {
        return(NULL)
      }

      last_turn <- private$.turns[[length(private$.turns)]]
      if (last_turn@role != "assistant") {
        return(NULL)
      }

      tool_requests <- keep(last_turn@contents, is_tool_request)
      if (length(tool_requests) == 0) {
        return(NULL)
      }

      lapply(tool_requests, function(req) {
        ContentToolResult(
          error = "Chat ended before the tool could be invoked.",
          request = req
        )
      })
    }
  )
)

#' @export
print.Chat <- function(x, ...) {
  provider <- x$get_provider()
  turns <- x$get_turns(include_system_prompt = TRUE)

  assistant_turns <- keep(turns, \(x) x@role == "assistant")
  complete_turns <- discard(assistant_turns, is_partial_turn)
  total_tokens <- colSums(map_tokens(complete_turns, \(x) x@tokens))
  total_cost <- sum(map_dbl(complete_turns, \(x) x@cost))

  cat(paste_c(
    "<Chat",
    c(" ", provider@name, "/", provider@model),
    c(" turns=", length(turns)),
    turn_cost(total_tokens, total_cost, prefix = " "),
    ">\n"
  ))

  for (i in seq_along(turns)) {
    turn <- turns[[i]]
    if (is_partial_turn(turn)) {
      label <- paste0(" [", turn@reason, "]")
    } else if (turn@role == "assistant") {
      label <- turn_cost(turn@tokens, turn@cost, prefix = " [", suffix = "]")
    } else {
      label <- ""
    }

    cli::cat_rule(cli::format_inline("{color_role(turn@role)}{label}"))
    cat(format(turns[[i]]))
  }

  invisible(x)
}

turn_cost <- function(tokens, cost, prefix, suffix = "") {
  out <- paste0(prefix, "input=")

  if (!is.na(tokens[[3]]) && tokens[[3]] > 0) {
    out <- paste0(out, tokens[[1]], "+", tokens[[3]])
  } else {
    out <- paste0(out, tokens[[1]])
  }
  out <- paste0(out, " output=", tokens[[2]])

  if (!is.na(cost)) {
    out <- paste0(out, " cost=", format(dollars(cost)))
  }
  out <- paste0(out, suffix)
  out
}

TurnAccumulator <- R6::R6Class(
  "TurnAccumulator",
  public = list(
    chat = NULL,
    chat_private = NULL,
    provider = NULL,
    controller = NULL,
    turn_idx = NULL,
    start_time = NULL,

    initialize = function(chat, chat_private, controller) {
      self$chat <- chat
      self$chat_private <- chat_private
      self$provider <- chat$get_provider()
      self$controller <- controller
    },

    begin_turn = function(user_turn) {
      self$chat$add_turn(user_turn, AssistantPartialTurn(), log_tokens = FALSE)
      self$turn_idx <- length(self$chat_private$.turns)
      self$start_time <- proc.time()[["elapsed"]]
      invisible(self)
    },

    update_turn = function(content) {
      idx <- self$turn_idx
      turn <- self$chat_private$.turns[[idx]]
      turn@contents <- c(turn@contents, list(content))
      self$chat_private$.turns[[idx]] <- turn
      invisible(self)
    },

    complete_turn = function(result, type = NULL) {
      if (self$controller$cancelled) {
        return(invisible(self))
      }
      duration <- proc.time()[["elapsed"]] - self$start_time
      turn <- self$value_turn(result, type, duration = duration)
      self$chat_private$.turns[[self$turn_idx]] <- turn
      # log_turn() is called manually here because the streaming path
      # replaces a partial turn in-place rather than using Chat$add_turn(),
      # which handles logging automatically for the non-streaming path.
      log_turn(self$provider, turn)
      turn
    },

    finalize_turn = function() {
      idx <- self$turn_idx
      if (is.null(idx)) {
        return(invisible())
      }
      turn <- self$chat_private$.turns[[idx]]
      if (!is_partial_turn(turn)) {
        return(invisible())
      }
      turn@contents <- merge_content_text(turn@contents)
      turn@reason <- self$controller$reason %||% "interrupted"
      turn@duration <- proc.time()[["elapsed"]] - self$start_time
      self$chat_private$.turns[[idx]] <- turn
      log_turn(self$provider, turn)
    },

    add_turn = function(user_turn, result, duration = NA_real_, type = NULL) {
      turn <- self$value_turn(result, type, duration = duration)
      self$chat$add_turn(user_turn, turn)
      turn
    },

    value_turn = function(result, type, duration = NA_real_) {
      turn <- value_turn(
        self$provider,
        result,
        has_type = !is.null(type)
      )
      turn@duration <- duration
      match_tools(turn, self$chat$get_tools())
    }
  )
)

echo_non_text_contents <- function(turn) {
  is_text <- map_lgl(turn@contents, S7_inherits, ContentText)
  formatted <- map_chr(turn@contents[!is_text], format)
  cat_line(formatted, prefix = "< ")
}

merge_content_text <- function(contents) {
  reduce(contents, .init = list(), function(acc, item) {
    n <- length(acc)
    if (n > 0 && every(list(acc[[n]], item), S7_inherits, ContentText)) {
      acc[[n]] <- ContentText(paste0(acc[[n]]@text, item@text))
    } else {
      acc <- c(acc, list(item))
    }
    acc
  })
}
method(contents_markdown, new_S3_class("Chat")) <- function(
  content,
  heading_level = 2
) {
  turns <- content$get_turns()
  if (length(turns) == 0) {
    return("")
  }

  hh <- strrep("#", heading_level)

  res <- vector("character", length(turns))
  for (i in seq_along(res)) {
    role <- turns[[i]]@role
    substr(role, 0, 1) <- toupper(substr(role, 0, 1))
    res[i] <- glue::glue("{hh} {role}\n\n{contents_markdown(turns[[i]])}")
  }

  paste(res, collapse = "\n\n")
}

# Registry ---------------------------------------------------------------------

test_that("can only set/register tools", {
  chat <- chat_openai_test()
  tool_def <- tool(function() 1, "tool")

  expect_snapshot(error = TRUE, {
    chat$register_tools(list(tool_def, 1))
    chat$set_tools(list(tool_def, 1))
  })
})

test_that("overwriting a tool yields a message", {
  my_tool <- tool(function() 1, "desc", name = "my_tool")

  chat <- chat_openai_test()
  chat$register_tool(my_tool)
  expect_snapshot(chat$register_tool(my_tool))
})

test_that("chat can get and register a list of tools", {
  chat <- chat_openai_test()
  chat2 <- chat_openai_test()

  tools <- list(
    "sys_time" = tool(
      function() strftime(Sys.time(), "%F %T"),
      name = "sys_time",
      description = "Get the current system time"
    ),
    "r_version" = tool(
      function() R.version.string,
      name = "r_version",
      description = "Get the R version of the current session"
    )
  )

  chat$register_tools(tools)
  chat2$set_tools(tools)

  expect_equal(chat$get_tools(), tools)
  expect_equal(chat2$get_tools(), chat$get_tools())

  # action = "replace" overwrites existing tools
  tool_r_major <- tool(
    function() R.version$major,
    name = "r_version_major",
    description = "Get the major version of R"
  )
  new_tools <- list("r_version_major" = tool_r_major)
  chat$set_tools(new_tools)
  expect_equal(chat$get_tools(), new_tools)
})

# Execution --------------------------------------------------------------------

test_that("can handle parallel tools", {
  vcr::local_cassette("chat-tools-parallel")

  chat <- chat_openai_test("Be terse")
  chat$register_tool(tool(
    replay(c(2, 5)),
    name = "dice",
    description = "Rolls a six-sided die"
  ))
  result <- chat$chat("Roll two dice and compute the total.")
  expect_match(result, "7")
  expect_equal(
    content_types(chat$get_turns()),
    list(
      "ContentText",
      c("ContentToolRequest", "ContentToolRequest"),
      c("ContentToolResult", "ContentToolResult"),
      "ContentText"
    )
  )
})

test_that("can handle sequential tools", {
  vcr::local_cassette("chat-tools-sequential")

  chat <- chat_openai_test("Be terse")
  chat$register_tool(tool(
    function() 1,
    name = "dice",
    description = "Rolls a dice"
  ))
  chat$register_tool(tool(
    function(roll) "Pants",
    name = "clothes",
    description = "Pick clothes to wear based on a dice roll",
    arguments = list(roll = type_number())
  ))

  result <- chat$chat(
    "Which clothes should I wear today? Roll a dice to decide."
  )
  expect_match(result, "pants", ignore.case = TRUE)
  expect_equal(
    content_types(chat$get_turns()),
    list(
      "ContentText",
      "ContentToolRequest",
      "ContentToolResult",
      "ContentToolRequest",
      "ContentToolResult",
      "ContentText"
    )
  )
})

test_that("chat warns on tool failures", {
  vcr::local_cassette("chat-tools-failure")
  chat <- chat_openai_test()

  chat$register_tool(tool(
    function(user) stop("User denied tool request"),
    name = "user_favorite_color",
    description = "Find out a user's favorite color",
    arguments = list(user = type_string("User's name"))
  ))

  expect_snapshot(
    . <- chat$chat("What are Joe, Hadley, Simon, and Tom's favorite colors?"),
    transform = function(value) gsub(" \\(\\w+_[a-z0-9A-Z]+\\)", " (ID)", value)
  )
})

test_that("tool calls can be rejected via `tool_request` callbacks", {
  vcr::local_cassette("chat-tools-reject-callback")
  chat <- chat_openai_test(model = "gpt-4.1-mini")

  chat$register_tool(tool(
    function(user) "red",
    name = "user_favorite_color",
    description = "Find out a user's favorite color",
    arguments = list(user = type_string("User's name"))
  ))

  chat$on_tool_request(function(request) {
    if (request@arguments$user == "Joe") {
      tool_reject("Joe denied the request.")
    }
  })

  expect_warning(
    result <- chat$chat(
      "What are Joe and Hadley's favorite colors?",
      "Write 'Joe ____ Hadley ____'. Use 'unknown' if you don't know."
    ),
    class = "ellmer_tool_failure"
  )
  expect_equal(result, ellmer_output("Joe unknown Hadley red"))
})

test_that("tool calls can be rejected via the tool function", {
  vcr::local_cassette("chat-tools-reject-tool-function")
  chat <- chat_openai_test(model = "gpt-4.1")

  chat$register_tool(tool(
    function(user) if (user == "Joe") tool_reject() else "red",
    name = "user_favorite_color",
    description = "Find out a user's favorite color",
    arguments = list(user = type_string("User's name"))
  ))

  expect_warning(
    result <- chat$chat(
      "What are Joe and Hadley's favorite colors?",
      "Write 'Joe ____ Hadley ____'. Use 'unknown' if you don't know."
    ),
    class = "ellmer_tool_failure"
  )
  expect_equal(result, ellmer_output("Joe unknown Hadley red"))
})

# Async ------------------------------------------------------------------------

test_that("can use async tools", {
  chat <- chat_openai_test("Be very terse, not even punctuation.")
  chat$register_tool(tool(
    coro::async(function() "2024-01-01"),
    name = "current_date",
    description = "Return the current date"
  ))

  result <- sync(chat$chat_async("What's the current date in Y-M-D format?"))
  expect_match(result, "2024-01-01")

  # Can't use async tools in sync context
  expect_error(chat$chat("Great. Do it again."), class = "tool_async_error")
})

test_that("chat callbacks for tool requests/results", {
  vcr::local_cassette("chat-tools-callbacks")
  chat <- chat_openai_test()

  test_tool <- tool(
    function(user) c("red", "blue")[nchar(user) %% 2 + 1],
    name = "user_favorite_color",
    description = "Find out a user's favorite color",
    arguments = list(user = type_string("User's name"))
  )

  chat$register_tool(test_tool)

  last_request <- NULL
  cb_count_request <- 0
  cb_count_result <- 0

  chat$on_tool_request(function(request) {
    cb_count_request <<- cb_count_request + 1
    cli::cli_inform(
      "[{cb_count_request}] Tool request: {request@arguments$user}"
    )

    expect_s7_class(request, ContentToolRequest)
    expect_equal(request@tool, test_tool)
    last_request <<- request
  })
  chat$on_tool_result(function(result) {
    cb_count_result <<- cb_count_result + 1
    cli::cli_inform("[{cb_count_result}] Tool result: {result@value}")

    expect_s7_class(result, ContentToolResult)
    expect_equal(result@request, last_request)
  })

  expect_snapshot(
    . <- chat$chat("What are Joe and Hadley's favorite colors?")
  )
  expect_equal(cb_count_request, 2L)
  expect_equal(cb_count_result, 2L)

  expect_snapshot(error = TRUE, {
    chat$on_tool_request(function(data) NULL)
    chat$on_tool_result(function(data) NULL)
  })
})

test_that("$chat_async() can run tools concurrently", {
  res <- list()

  chat <- chat_openai_test()
  chat$register_tool(tool(
    coro::async(function(i) {
      res[[i]] <<- list(start = Sys.time())
      coro::await(coro::async_sleep(0.5))
      res[[i]]$end <<- Sys.time()
      i
    }),
    name = "test_async_tool",
    description = "Tests async tool usage",
    arguments = list(i = type_string("ID of the tool call"))
  ))

  sync(chat$chat_async(
    "Run `test_async_tool` twice with inputs '1' and '2'.",
    tool_mode = "concurrent"
  ))

  # The calls overlap, and both start before the first ends
  expect_true(res[[1]]$start < res[[1]]$end)
  expect_true(res[[2]]$start < res[[1]]$end)
  expect_true(res[[2]]$start < res[[2]]$end)
})

test_that("$chat_async() can run tools sequentially", {
  res <- list()

  chat <- chat_openai_test()
  chat$register_tool(tool(
    coro::async(function(i) {
      res[[i]] <<- list(start = Sys.time())
      coro::await(coro::async_sleep(0.5))
      res[[i]]$end <<- Sys.time()
      i
    }),
    name = "test_async_tool",
    description = "Tests async tool usage",
    arguments = list(i = type_string("ID of the tool call"))
  ))

  sync(chat$chat_async(
    "Run `test_async_tool` twice with inputs '1' and '2'.",
    tool_mode = "sequential"
  ))

  # The calls don't overlap, the first ends before the second starts
  expect_true(res[[1]]$start < res[[1]]$end)
  expect_true(res[[1]]$end < res[[2]]$start)
  expect_true(res[[2]]$start < res[[2]]$end)
})

test_that("$stream(stream='content') yields tool request/result contents", {
  chat <- chat_openai_test()
  tool_current_date <- tool(
    function() "2024-01-01",
    description = "Return the current date"
  )
  chat$register_tool(tool_current_date)

  res <- coro::collect(
    chat$stream(
      "What's the current date in Y-M-D format?",
      stream = "content"
    )
  )

  # 1. Tool request
  # 2. Tool result (paired with request)
  # 3. ...rest of assistant message
  expect_s7_class(res[[1]], ContentToolRequest)
  expect_equal(res[[1]]@tool, tool_current_date)
  expect_s7_class(res[[2]], ContentToolResult)
  expect_equal(res[[2]]@value, "2024-01-01")
  expect_equal(res[[2]]@request, res[[1]])

  for (delta in res[-(1:2)]) {
    expect_s7_class(delta, ContentText)
  }
})

test_that("$stream_async(stream='content', tool_mode='concurrent') yields tool request/result contents concurrently", {
  chat <- chat_openai_test()
  tool_current_date <- tool(
    coro::async(function() {
      coro::await(coro::async_sleep(0.1))
      "2024-01-01"
    }),
    name = "current_date",
    description = "Return the current date"
  )
  chat$register_tool(tool_current_date)

  res <- sync(
    coro::async_collect(
      chat$stream_async(
        "Confirm the current data by calling `current_date` twice.",
        "Write YES if the dates match or NO if not.",
        stream = "content",
        tool_mode = "concurrent"
      )
    )
  )

  # 1. Tool request 1
  expect_s7_class(res[[1]], ContentToolRequest)
  expect_equal(res[[1]]@tool, tool_current_date)
  # 2. Tool request 2
  expect_s7_class(res[[2]], ContentToolRequest)
  expect_equal(res[[2]]@tool, tool_current_date)
  # 3. Tool result 1
  expect_s7_class(res[[3]], ContentToolResult)
  expect_equal(res[[3]]@value, "2024-01-01")
  expect_equal(res[[3]]@request, res[[1]])
  # 4. Tool result 2
  expect_s7_class(res[[4]], ContentToolResult)
  expect_equal(res[[4]]@value, "2024-01-01")
  expect_equal(res[[4]]@request, res[[2]])

  # 5. ...rest of assistant message
  for (delta in res[-(1:4)]) {
    expect_s7_class(delta, ContentText)
  }
})

test_that("$stream_async(stream='content', tool_mode='sequential') yields tool request/result contents sequentially", {
  chat <- chat_openai_test()
  tool_current_date <- tool(
    coro::async(function() {
      coro::await(coro::async_sleep(0.1))
      "2024-01-01"
    }),
    name = "current_date",
    description = "Return the current date"
  )
  chat$register_tool(tool_current_date)

  res <- sync(
    coro::async_collect(
      chat$stream_async(
        "Confirm the current data by calling `current_date` twice.",
        "Write YES if the dates match or NO if not.",
        stream = "content",
        tool_mode = "sequential"
      )
    )
  )

  # 1. Tool request 1
  expect_s7_class(res[[1]], ContentToolRequest)
  expect_equal(res[[1]]@tool, tool_current_date)
  # 2. Tool result 1
  expect_s7_class(res[[2]], ContentToolResult)
  expect_equal(res[[2]]@value, "2024-01-01")
  expect_equal(res[[2]]@request, res[[1]])
  # 3. Tool request 2
  expect_s7_class(res[[3]], ContentToolRequest)
  expect_equal(res[[3]]@tool, tool_current_date)
  # 4. Tool result 2
  expect_s7_class(res[[4]], ContentToolResult)
  expect_equal(res[[4]]@value, "2024-01-01")
  expect_equal(res[[4]]@request, res[[3]])

  # 5. ...rest of assistant message
  for (delta in res[-(1:4)]) {
    expect_s7_class(delta, ContentText)
  }
})

# Invocation ------------------------------------------------------------------

test_that("invoke_tool returns a ContentToolResult", {
  tool_f <- tool(function() 1, name = "my_tool", description = "A tool")

  res <- invoke_tool(
    ContentToolRequest(
      id = "x",
      name = "my_tool",
      arguments = list(),
      tool = tool_f
    )
  )
  expect_s3_class(res, "ellmer::ContentToolResult")
  expect_equal(res@error, NULL)
  expect_false(tool_errored(res))
  expect_equal(res@value, 1)
  expect_s3_class(res@request, "ellmer::ContentToolRequest")
  expect_equal(res@request@id, "x")
  expect_equal(res@request@tool, tool_f)
  expect_equal(res@request@arguments, list())

  res <- invoke_tool(
    ContentToolRequest(
      id = "x",
      name = "my_tool",
      arguments = list(x = 1),
      tool = tool_f
    )
  )
  expect_s3_class(res, "ellmer::ContentToolResult")
  expect_s3_class(res@error, "condition")
  expect_true(tool_errored(res))
  expect_match(tool_error_string(res), "unused argument", ignore.case = TRUE)
  expect_equal(res@value, NULL)
  expect_equal(res@extra, list())
  expect_s3_class(res@request, "ellmer::ContentToolRequest")
  expect_equal(res@request@id, "x")
  expect_equal(res@request@tool, tool_f)
  expect_equal(res@request@arguments, list(x = 1))

  res <- invoke_tool(
    ContentToolRequest(
      id = "x",
      arguments = list(x = 1),
      name = "my_tool"
    )
  )
  expect_s3_class(res, "ellmer::ContentToolResult")
  expect_equal(res@error, "Unknown tool")
  expect_equal(tool_error_string(res), "Unknown tool")
  expect_true(tool_errored(res))
  expect_equal(res@value, NULL)
  expect_equal(res@extra, list())
  expect_s3_class(res@request, "ellmer::ContentToolRequest")
  expect_equal(res@request@id, "x")
  expect_equal(res@request@tool, NULL)
  expect_equal(res@request@arguments, list(x = 1))

  tool_ctr <- tool(
    function() ContentToolResult(value = 1, extra = list(a = 1)),
    name = "my_tool",
    description = "A tool that returns ContentToolResult"
  )
  res <- invoke_tool(
    ContentToolRequest(
      id = "x",
      name = "my_tool",
      arguments = list(),
      tool = tool_ctr
    )
  )
  expect_s3_class(res, "ellmer::ContentToolResult")
  expect_equal(res@error, NULL)
  expect_false(tool_errored(res))
  expect_equal(res@value, 1)
  expect_equal(res@extra, list(a = 1))
  expect_s3_class(res@request, "ellmer::ContentToolRequest")
  expect_equal(res@request@id, "x")
  expect_equal(res@request@tool, tool_ctr)
  expect_equal(res@request@arguments, list())
})

test_that("invoke_tool_async returns a ContentToolResult", {
  tool_f <- tool(function() 1, name = "my_tool", description = "A tool")

  res <- sync(invoke_tool_async(
    ContentToolRequest(
      id = "x",
      name = "my_tool",
      arguments = list(),
      tool = tool_f
    )
  ))
  expect_s3_class(res, "ellmer::ContentToolResult")
  expect_equal(res@error, NULL)
  expect_false(tool_errored(res))
  expect_equal(res@value, 1)
  expect_s3_class(res@request, "ellmer::ContentToolRequest")
  expect_equal(res@request@id, "x")
  expect_equal(res@request@tool, tool_f)
  expect_equal(res@request@arguments, list())

  res <- sync(invoke_tool_async(
    ContentToolRequest(
      id = "x",
      name = "my_tool",
      arguments = list(x = 1),
      tool = tool_f
    )
  ))
  expect_s3_class(res, "ellmer::ContentToolResult")
  expect_s3_class(res@error, "condition")
  expect_true(tool_errored(res))
  expect_match(tool_error_string(res), "unused argument", ignore.case = TRUE)
  expect_equal(res@value, NULL)
  expect_equal(res@extra, list())
  expect_s3_class(res@request, "ellmer::ContentToolRequest")
  expect_equal(res@request@id, "x")
  expect_equal(res@request@tool, tool_f)
  expect_equal(res@request@arguments, list(x = 1))

  res <- sync(invoke_tool_async(
    ContentToolRequest(
      id = "x",
      arguments = list(x = 1),
      name = "my_tool"
    )
  ))
  expect_s3_class(res, "ellmer::ContentToolResult")
  expect_equal(res@error, "Unknown tool")
  expect_equal(tool_error_string(res), "Unknown tool")
  expect_true(tool_errored(res))
  expect_equal(res@value, NULL)
  expect_equal(res@extra, list())
  expect_s3_class(res@request, "ellmer::ContentToolRequest")
  expect_equal(res@request@id, "x")
  expect_equal(res@request@tool, NULL)
  expect_equal(res@request@arguments, list(x = 1))

  tool_ctr <- tool(
    function() ContentToolResult(value = 1, extra = list(a = 1)),
    name = "my_tool",
    description = "A tool that returns ContentToolResult"
  )
  res <- sync(invoke_tool_async(
    ContentToolRequest(
      id = "x",
      name = "my_tool",
      arguments = list(),
      tool = tool_ctr
    )
  ))
  expect_s3_class(res, "ellmer::ContentToolResult")
  expect_equal(res@error, NULL)
  expect_false(tool_errored(res))
  expect_equal(res@value, 1)
  expect_equal(res@extra, list(a = 1))
  expect_s3_class(res@request, "ellmer::ContentToolRequest")
  expect_equal(res@request@id, "x")
  expect_equal(res@request@tool, tool_ctr)
  expect_equal(res@request@arguments, list())
})

test_that("invoke_tools() echoes tool requests and results", {
  turn <- fixture_turn_with_tool_requests()

  expect_silent(coro::collect(invoke_tools(turn)))
  expect_snapshot(. <- coro::collect(invoke_tools(turn, echo = "output")))
})

test_that("invoke_tools() yields tool requests and results", {
  turn <- fixture_turn_with_tool_requests()

  steps <- coro::collect(invoke_tools(turn, yield_request = TRUE))
  for (i in seq_along(steps)) {
    # Threaded requests and results
    if (i %% 2 == 1) {
      expect_s3_class(steps[[i]], "ellmer::ContentToolRequest")
    } else {
      expect_s3_class(steps[[i]], "ellmer::ContentToolResult")
      # results are paired with previous request
      expect_equal(steps[[i]]@request, steps[[i - 1]])
    }
  }
})

test_that("invoke_tools_async() echoes tool requests and results", {
  turn <- fixture_turn_with_tool_requests()

  expect_silent(sync({
    # Concurrent tool calls
    gen_async_promise_all(invoke_tools_async(turn))
    # Sequential tool calls
    coro::async_collect(invoke_tools_async(turn))
  }))
  expect_snapshot({
    # Concurrent tool calls
    . <- sync(gen_async_promise_all(invoke_tools_async(turn, echo = "output")))
    # Sequential tool calls
    . <- sync(coro::async_collect(invoke_tools_async(turn, echo = "output")))
  })
})

test_that("invoke_tools_async() yields tool requests and promises results", {
  turn <- fixture_turn_with_tool_requests()

  steps <- coro::collect(invoke_tools_async(turn, yield_request = TRUE))

  for (i in seq_along(steps)) {
    # Threaded requests and promise for results
    if (i %% 2 == 1) {
      expect_s3_class(steps[[i]], "ellmer::ContentToolRequest")
    } else {
      expect_s3_class(steps[[i]], "promise")
      result <- sync(steps[[i]])
      expect_s3_class(result, "ellmer::ContentToolResult")
      # results are paired with previous request
      expect_equal(result@request, steps[[i - 1]])
    }
  }
})

test_that("invoke_tools() converts to R data structures", {
  out <- NULL
  tool <- tool(
    function(x, y) out <<- list(x = x, y = y),
    description = "A tool",
    arguments = list(
      x = type_array(type_number()),
      y = type_array(type_string())
    )
  )

  req <- ContentToolRequest(
    id = "x",
    name = "my_tool",
    arguments = list(x = list(1, 2), y = list()),
    tool = tool
  )

  args <- tool_request_args(req)
  expect_equal(args$x, c(1, 2))
  expect_equal(args$y, character())

  res <- invoke_tool(req)
  expect_equal(out$x, c(1, 2))
  expect_equal(out$y, character())
})

test_that("invoke_tools_async() converts to R data structures", {
  out <- NULL
  tool <- tool(
    function(x, y) out <<- list(x = x, y = y),
    description = "A tool",
    arguments = list(
      x = type_array(type_number()),
      y = type_array(type_string())
    )
  )

  req <-
    ContentToolRequest(
      id = "x",
      name = "my_tool",
      arguments = list(x = list(1, 2), y = list()),
      tool = tool
    )

  args <- tool_request_args(req)
  expect_equal(args$x, c(1, 2))
  expect_equal(args$y, character())

  res <- sync(invoke_tool_async(req))
  expect_equal(out$x, c(1, 2))
  expect_equal(out$y, character())
})

test_that("invoke_tools() can invoke tools with args with default values", {
  out <- NULL
  tool <- tool(
    function(x, y, z = "z") out <<- list(x = x, y = y, z = z),
    description = "A tool",
    arguments = list(
      x = type_array(type_number()),
      y = type_array(type_string()),
      z = type_array(type_string(), required = FALSE)
    )
  )

  req <- ContentToolRequest(
    id = "x",
    name = "my_tool",
    arguments = list(x = list(1, 2), y = NULL, z = NULL),
    tool = tool
  )

  args <- tool_request_args(req)
  expect_equal(args$x, c(1, 2))
  expect_equal(args$y, character()) # Required arg
  expect_equal(args$z, NULL) # Optional arg

  res <- invoke_tool(req)
  expect_equal(out$x, c(1, 2))
  expect_equal(out$y, character())
  expect_equal(out$z, "z")
})

test_that("invoke_tools_async() can invoke tools with args with default values", {
  out <- NULL
  tool <- tool(
    function(x, y, z = "z") out <<- list(x = x, y = y, z = z),
    description = "A tool",
    arguments = list(
      x = type_array(type_number()),
      y = type_array(type_string()),
      z = type_array(type_string(), required = FALSE)
    )
  )

  req <- ContentToolRequest(
    id = "x",
    name = "my_tool",
    arguments = list(x = list(1, 2), y = NULL, z = NULL),
    tool = tool
  )

  args <- tool_request_args(req)
  expect_equal(args$x, c(1, 2))
  expect_equal(args$y, character()) # Required arg
  expect_equal(args$z, NULL) # Optional arg

  res <- sync(invoke_tool_async(req))
  expect_equal(out$x, c(1, 2))
  expect_equal(out$y, character())
  expect_equal(out$z, "z")
})

test_that("tool error warnings", {
  errors <- list(
    ContentToolResult(
      error = "The JSON was invalid: {[1, 2, 3]}",
      request = ContentToolRequest(
        id = "call1",
        name = "returns_json"
      )
    ),
    ContentToolResult(
      error = rlang::catch_cnd(stop("went boom!")),
      request = ContentToolRequest(
        id = "call2",
        name = "throws"
      )
    )
  )

  expect_snapshot(
    warn_tool_errors(errors)
  )
})

test_that("match_tools() matches tools in a turn to a list of tools", {
  turn_single <- AssistantTurn(
    list(ContentToolRequest("y1", "unknown", list()))
  )
  expect_null(turn_single@contents[[1]]@tool)
  expect_s7_class(match_tools(turn_single, list()), Turn)
  # unmatched requests have NULL tool
  expect_null(match_tools(turn_single, list())@contents[[1]]@tool)

  tools <- fixture_list_of_tools()
  turn <- fixture_turn_with_tool_requests(with_tool = FALSE)

  turn_matched <- match_tools(turn, tools)
  expect_equal(turn_matched, fixture_turn_with_tool_requests(with_tool = TRUE))
})

# Dangling tool requests ------------------------------------------------------
test_that("dangling tool requests are inserted into user message", {
  skip_if_not_installed("testthat", "3.3.0")
  local_mocked_r6_class(
    Chat,
    private = list(
      chat_impl = generator_method(function(self, private, user_turn, ...) {
        self$add_turn(
          user_turn,
          AssistantTurn("Assistant response", tokens = c(0, 0, 0))
        )
      })
    )
  )
  provider <- ProviderOpenAI("name", "model", "base_url")
  chat <- Chat$new(provider)

  # Simulate a broken chat history with dangling tool request
  tool_request <- ContentToolRequest(id = "1", name = "get_date")
  chat$set_turns(list(
    UserTurn("What year is it today?"),
    AssistantTurn(list(tool_request))
  ))
  chat$chat("try again")

  turns <- chat$get_turns()
  expect_s3_class(turns[[3]], "ellmer::UserTurn")
  # And UserTurn must contain a result that matches the request
  expect_equal(
    turns[[3]],
    UserTurn(list(
      ContentToolResult(
        error = "Chat ended before the tool could be invoked.",
        request = tool_request
      ),
      ContentText("try again")
    ))
  )
})

test_that("can resume chat after dangling tool requests", {
  vcr::local_cassette("chat-tools-dangling")

  chat <- chat_openai_test("Be terse")
  chat$register_tool(tool(
    \() "2001-02-09",
    name = "get_date",
    description = "Get todays date"
  ))

  # Simulate a broken chat history with dangling tool request
  tool_request <- ContentToolRequest(id = "1", name = "get_date")
  chat$set_turns(list(
    UserTurn("What year is it today?"),
    AssistantTurn(list(tool_request))
  ))

  # resume chat and get answer, not error
  expect_match(chat$chat("try again"), "2001")
})

# See test-chat-tools.R for tests of tool calling

test_that("can get and set the system prompt", {
  chat <- chat_openai_test()
  chat$set_turns(list(
    UserTurn("Hi"),
    AssistantTurn("Hello")
  ))

  # NULL -> NULL
  chat$set_system_prompt(NULL)
  expect_equal(chat$get_system_prompt(), NULL)

  # NULL -> string
  chat$set_system_prompt("x")
  expect_equal(chat$get_system_prompt(), "x")

  # string -> string
  chat$set_system_prompt("y")
  expect_equal(chat$get_system_prompt(), "y")

  # string -> NULL
  chat$set_system_prompt(NULL)
  expect_equal(chat$get_system_prompt(), NULL)
})

test_that("system prompt can be a vector", {
  chat <- chat_openai_test(c("This is", "the system prompt"))
  expect_equal(chat$get_system_prompt(), "This is\n\nthe system prompt")
})

test_that("system prompt must be a character vector", {
  expect_snapshot(error = TRUE, {
    chat_openai_test(1)
  })
})

test_that("can retrieve system prompt with last_turn()", {
  chat1 <- chat_openai_test(system_prompt = NULL)
  expect_equal(chat1$last_turn("system"), NULL)

  chat2 <- chat_openai_test(system_prompt = "You are from New Zealand")
  expect_equal(
    chat2$last_turn("system"),
    SystemTurn("You are from New Zealand")
  )
})

test_that("can get and set turns", {
  chat <- chat_openai_test()
  expect_equal(chat$get_turns(), list())

  turns <- list(UserTurn(), AssistantTurn())
  chat$set_turns(turns)
  expect_equal(chat$get_turns(), list(UserTurn(), AssistantTurn()))
})

test_that("can get model", {
  chat <- chat_openai_test(model = "abc")
  expect_equal(chat$get_model(), "abc")
})

test_that("setting turns usually preserves, but can set system prompt", {
  chat <- chat_openai_test(system_prompt = "You're a funny guy")
  chat$set_turns(list())
  expect_equal(chat$get_system_prompt(), "You're a funny guy")

  chat$set_turns(list(SystemTurn(list(ContentText("You're a cool guy")))))
  expect_equal(chat$get_system_prompt(), "You're a cool guy")
})


test_that("can perform a simple batch chat", {
  chat <- chat_openai_test()

  result <- chat$chat("What's 1 + 1. Just give me the answer, no punctuation")
  expect_equal(result, ellmer_output("2"))
  expect_equal(chat$last_turn()@contents[[1]]@text, "2")
})

test_that("can chat with a single prompt", {
  chat <- chat_openai_test()
  expect_no_error(chat$chat(interpolate("What's 1 + 1?")))
})

test_that("can't chat with multiple prompts", {
  chat <- chat_openai_test()
  prompt <- interpolate("{{x}}", x = 1:2)
  expect_snapshot(error = TRUE, {
    chat$chat(prompt)
  })
})

test_that("can perform a simple async batch chat", {
  chat <- chat_openai_test()

  result <- chat$chat_async(
    "What's 1 + 1. Just give me the answer, no punctuation"
  )
  expect_s3_class(result, "promise")

  result <- sync(result)
  expect_equal(result, "2")
  expect_equal(chat$last_turn()@contents[[1]]@text, "2")
})

test_that("can perform a simple streaming chat", {
  chat <- chat_openai_test()

  chunks <- coro::collect(chat$stream(
    "
    What are the canonical colors of the ROYGBIV rainbow?
    Put each colour on its own line. Don't use punctuation.
  "
  ))
  expect_gt(length(chunks), 2)

  rainbow_re <- "^red *\norange *\nyellow *\ngreen *\nblue *\nindigo *\nviolet *\n?$"
  expect_match(paste(chunks, collapse = ""), rainbow_re, ignore.case = TRUE)
  expect_match(
    chat$last_turn()@contents[[1]]@text,
    rainbow_re,
    ignore.case = TRUE
  )
})

test_that("can perform a simple async batch chat", {
  chat <- chat_openai_test()

  chunks <- coro::async_collect(chat$stream_async(
    "
    What are the canonical colors of the ROYGBIV rainbow?
    Put each colour on its own line. Don't use punctuation.
  "
  ))
  expect_s3_class(chunks, "promise")

  chunks <- sync(chunks)
  expect_gt(length(chunks), 2)
  rainbow_re <- "^red *\norange *\nyellow *\ngreen *\nblue *\nindigo *\nviolet *\n?$"
  expect_match(paste(chunks, collapse = ""), rainbow_re, ignore.case = TRUE)
  expect_match(
    chat$last_turn()@contents[[1]]@text,
    rainbow_re,
    ignore.case = TRUE
  )
})

test_that("can extract structured data", {
  person <- type_object(name = type_string(), age = type_integer())

  chat <- chat_openai_test()
  data <- chat$chat_structured("John, age 15, won first prize", type = person)
  expect_equal(data, list(name = "John", age = 15))
})

test_that("can extract structured data (async)", {
  person <- type_object(name = type_string(), age = type_integer())

  chat <- chat_openai_test()
  data <- sync(chat$chat_structured_async(
    "John, age 15, won first prize",
    type = person
  ))
  expect_equal(data, list(name = "John", age = 15))
})

test_that("chat_structured() doesn't require a prompt", {
  chat <- chat_openai_test()
  chat$chat("What's the biggest city in the world? What country is it in?")

  out <- chat$chat_structured(
    type = type_object(
      city = type_string(),
      county = type_string()
    )
  )
  expect_equal(out, list(city = "Tokyo", county = "Japan"))
})

test_that("has a basic print method", {
  chat <- chat_openai_test()
  chat$set_turns(list(
    UserTurn("What's 1 + 1?\nWhat's 1 + 2?"),
    AssistantTurn("2\n\n3", tokens = c(10, 5, 5))
  ))
  expect_snapshot(chat)
})

test_that("print method shows interrupted for partial turns", {
  chat <- chat_openai_test(model = "gpt-4o", system_prompt = NULL)
  chat$set_turns(list(
    UserTurn("Input 1"),
    AssistantTurn("Output 1", tokens = c(15000, 500, 0), cost = 0.2),
    UserTurn("Input 2"),
    AssistantPartialTurn("Partial output...")
  ))
  expect_snapshot(chat)
})

test_that("print method shows custom reason for partial turns", {
  chat <- chat_openai_test(model = "gpt-4o", system_prompt = NULL)
  chat$set_turns(list(
    UserTurn("Input 1"),
    AssistantPartialTurn("Partial output...", reason = "cancelled")
  ))
  expect_snapshot(chat)
})

test_that("print method shows cumulative tokens & cost", {
  chat <- chat_openai_test(model = "gpt-4o", system_prompt = NULL)
  chat$set_turns(list(
    UserTurn("Input 1"),
    AssistantTurn("Output 1", tokens = c(15000, 500, 0), cost = 0.2),
    UserTurn("Input 2"),
    AssistantTurn("Output 1", tokens = c(30000, 1000, 0), cost = 0.1)
  ))
  expect_snapshot(chat)
})

test_that("can compute costs", {
  chat <- chat_openai_test(model = "gpt-4o", system_prompt = NULL)
  chat$set_turns(list(
    UserTurn("Input 1"),
    AssistantTurn("Output 1", tokens = c(15000, 500, 0), cost = 0.2),
    UserTurn("Input 2"),
    AssistantTurn("Output 1", tokens = c(30000, 1000, 0), cost = 0.1)
  ))

  expect_equal(chat$get_cost(), dollars(0.3))
  expect_equal(chat$get_cost("last"), dollars(0.1))

  details <- chat$get_tokens()
  expect_equal(details$cost, dollars(c(0.2, 0.1)))
  expect_equal(details$input, c(15000, 30000))
  expect_equal(details$output, c(500, 1000))
  expect_equal(details$cached_input, c(0, 0))

  expect_snapshot(details)
})

test_that("can optionally echo", {
  chat <- chat_openai_test("Repeat the input back to me exactly", echo = TRUE)
  expect_output(chat$chat("Echo this."), "Echo this.")
  expect_output(chat$chat("Echo this.", echo = FALSE), NA)

  chat <- chat_openai_test("Repeat the input back to me exactly")
  expect_output(chat$chat("Echo this."), NA)
  expect_output(chat$chat("Echo this.", echo = TRUE), "Echo this.")
})

test_that("can retrieve last_turn for user and assistant", {
  chat <- chat_openai_test()
  expect_equal(chat$last_turn("user"), NULL)
  expect_equal(chat$last_turn("assistant"), NULL)

  chat$chat("Hi")
  expect_equal(chat$last_turn("user")@role, "user")
  expect_equal(chat$last_turn("assistant")@role, "assistant")
})

test_that("api_headers parameter works correctly", {
  chat <- chat_openai_test(api_headers = c("X-Test" = "value"))
  expect_equal(chat$get_provider()@extra_headers, c("X-Test" = "value"))

  req <- chat_request(chat$get_provider())
  expect_equal(req_get_headers(req), list("X-Test" = "value"))
})

test_that("assistant turns track duration", {
  vcr::local_cassette("chat-duration")

  chat <- chat_openai_test()
  chat$chat("What's 1 + 1?")

  assistant_turn <- chat$last_turn()

  # These assistant durations are usually not NA, but are during replay (#479)
  expect_true(is.na(assistant_turn@duration) || assistant_turn@duration > 0)
})

# stream_controller() ----------------------------------------------------------

test_that("stream_controller() creates correct object", {
  ctrl <- stream_controller()
  expect_s3_class(ctrl, "ellmer_stream_controller")
  expect_false(ctrl$cancelled)
  expect_null(ctrl$reason)
  expect_true(is.function(ctrl$cancel))
  expect_true(is.function(ctrl$reset))
})

test_that("stream_controller()$cancel() sets cancelled to TRUE", {
  ctrl <- stream_controller()
  ctrl$cancel()
  expect_true(ctrl$cancelled)
  expect_equal(ctrl$reason, "cancelled")
})

test_that("stream_controller()$cancel() accepts a custom reason", {
  ctrl <- stream_controller()
  ctrl$cancel(reason = "timeout")
  expect_true(ctrl$cancelled)
  expect_equal(ctrl$reason, "timeout")
})

test_that("stream_controller()$reset() clears cancelled state and reason", {
  ctrl <- stream_controller()
  ctrl$cancel(reason = "timeout")
  expect_true(ctrl$cancelled)
  expect_equal(ctrl$reason, "timeout")
  ctrl$reset()
  expect_false(ctrl$cancelled)
  expect_null(ctrl$reason)
})

test_that("as_controller() resets a pre-cancelled controller", {
  ctrl <- stream_controller()
  ctrl$cancel()
  result <- expect_silent(as_controller(ctrl, reset = TRUE))
  expect_false(ctrl$cancelled)
  expect_identical(result, ctrl)
})

test_that("stream() rejects non-controller object", {
  chat <- chat_openai_test()
  expect_snapshot(error = TRUE, {
    coro::collect(chat$stream("hi", controller = TRUE))
  })
})

test_that("stream_async() rejects non-controller object", {
  chat <- chat_openai_test()
  expect_snapshot(error = TRUE, {
    sync(coro::async_collect(chat$stream_async("hi", controller = list())))
  })
})

test_that("as_controller() accepts a valid stream_controller() or NULL", {
  ctrl <- stream_controller()
  expect_identical(as_controller(ctrl), ctrl)

  default <- as_controller(NULL)
  expect_false(default$cancelled)
  expect_null(default$reason)
})

test_that("stream_controller() rejects invalid cancelled values", {
  ctrl <- stream_controller()
  expect_error(ctrl$cancelled <- "banana")
  expect_error(ctrl$cancelled <- NA)
  expect_error(ctrl$cancelled <- c(TRUE, FALSE))
})

test_that("stream_controller() rejects invalid reason values", {
  ctrl <- stream_controller()
  expect_error(ctrl$reason <- 123)
  expect_error(ctrl$reason <- NA_character_)
  expect_error(ctrl$reason <- c("a", "b"))
})

test_that("stream_controller() environment is locked", {
  ctrl <- stream_controller()
  expect_error(ctrl$typo <- TRUE)
})

test_that("finalize_turn() merges adjacent ContentText", {
  chat <- chat_openai_test()
  acc <- TurnAccumulator$new(
    chat,
    chat$.__enclos_env__$private,
    stream_controller()
  )

  user_turn <- Turn("user", list(ContentText("hi")))
  acc$begin_turn(user_turn)
  acc$update_turn(ContentText("Hello "))
  acc$update_turn(ContentText("world"))
  acc$finalize_turn()

  turn <- chat$last_turn()
  expect_s7_class(turn, AssistantPartialTurn)
  expect_length(turn@contents, 1)
  expect_equal(turn@text, "Hello world")
  expect_equal(turn@reason, "interrupted")
  # No token data
  expect_true(all(is.na(turn@tokens)))
  expect_true(is.na(turn@cost))
})

test_that("finalize_turn() uses controller reason", {
  chat <- chat_openai_test()
  ctrl <- stream_controller()
  ctrl$cancel(reason = "timeout")
  acc <- TurnAccumulator$new(chat, chat$.__enclos_env__$private, ctrl)

  user_turn <- Turn("user", list(ContentText("hi")))
  acc$begin_turn(user_turn)
  acc$update_turn(ContentText("partial"))
  acc$finalize_turn()

  turn <- chat$last_turn()
  expect_s7_class(turn, AssistantPartialTurn)
  expect_equal(turn@reason, "timeout")
})

test_that("finalize_turn() is a no-op for complete turns", {
  chat <- chat_openai_test()
  acc <- TurnAccumulator$new(
    chat,
    chat$.__enclos_env__$private,
    stream_controller()
  )

  user_turn <- Turn("user", list(ContentText("hi")))
  chat$add_turn(
    user_turn,
    AssistantTurn(contents = list(ContentText("done"))),
    log_tokens = FALSE
  )
  # Manually set turn_idx so finalize_turn has something to check
  acc$.__enclos_env__$private$turn_idx <- 2L

  acc$finalize_turn()
  turn <- chat$last_turn()

  expect_s7_class(turn, AssistantTurn)
  expect_false(S7_inherits(turn, AssistantPartialTurn))
})

test_that("update_turn() appends content incrementally", {
  chat <- chat_openai_test()
  acc <- TurnAccumulator$new(
    chat,
    chat$.__enclos_env__$private,
    stream_controller()
  )

  user_turn <- Turn("user", list(ContentText("hi")))
  acc$begin_turn(user_turn)
  acc$update_turn(ContentText("a"))
  acc$update_turn(ContentText("b"))

  turn <- chat$last_turn()
  expect_length(turn@contents, 2)
  expect_equal(turn@contents[[1]]@text, "a")
  expect_equal(turn@contents[[2]]@text, "b")
})

test_that("merge_content_text() merges adjacent text, preserves non-text", {
  contents <- list(
    ContentText("a"),
    ContentText("b"),
    ContentThinking("thought"),
    ContentText("c")
  )
  merged <- merge_content_text(contents)

  expect_length(merged, 3)
  expect_s7_class(merged[[1]], ContentText)
  expect_equal(merged[[1]]@text, "ab")
  expect_s7_class(merged[[2]], ContentThinking)
  expect_s7_class(merged[[3]], ContentText)
  expect_equal(merged[[3]]@text, "c")
})

test_that("can chat in parallel", {
  vcr::local_cassette("parallel-basic")

  chat <- chat_openai_test()
  chats <- parallel_chat(chat, list("What's 1 + 1?", "What's 2 + 2?"))

  expect_type(chats, "list")
  expect_length(chats, 2)

  expect_s3_class(chats[[1]], "Chat")
  expect_s3_class(chats[[2]], "Chat")

  expect_equal(chats[[1]]$last_turn()@contents[[1]]@text, "2")
  expect_equal(chats[[2]]$last_turn()@contents[[1]]@text, "4")
})

test_that("can just get text parallel ", {
  vcr::local_cassette("parallel-basic")

  chat <- chat_openai_test()
  out <- parallel_chat_text(chat, list("What's 1 + 1?", "What's 2 + 2?"))
  expect_equal(out, c("2", "4"))
})

test_that("can call tools in parallel", {
  vcr::local_cassette("parallel-tool")

  prompts <- rep(list("Roll the dice, please! Reply with 'You rolled ____'"), 2)

  chat <- chat_openai_test()
  chat$register_tool(tool(
    counter(),
    name = "roll",
    description = "Rolls a six-sided die."
  ))
  chats <- parallel_chat(chat, prompts)

  turns_1 <- chats[[1]]$get_turns()
  expect_s3_class(turns_1[[2]]@contents[[1]], "ellmer::ContentToolRequest")
  expect_s3_class(turns_1[[3]]@contents[[1]], "ellmer::ContentToolResult")
  expect_equal(contents_text(turns_1[[4]]), "You rolled 1")

  turns_1 <- chats[[2]]$get_turns()
  expect_equal(contents_text(turns_1[[4]]), "You rolled 2")
})

test_that("can have uneven number of turns", {
  vcr::local_cassette("parallel-tool-uneven")

  prompts <- list(
    "Roll the dice, please! Reply with 'You rolled ____'",
    "reply with the word 'boop'",
    "Roll the dice, please! Reply with 'You rolled ____'",
    "reply with the word 'beep'"
  )

  chat <- chat_openai_test()
  chat$register_tool(tool(
    counter(),
    name = "roll",
    description = "Rolls a six-sided die."
  ))
  chats <- parallel_chat(chat, prompts)

  lengths <- map_int(chats, \(chat) length(chat$get_turns()))
  expect_equal(lengths, c(4, 2, 4, 2))

  text <- map_chr(chats, \(chat) chat$last_turn()@text)
  expect_equal(text, c("You rolled 1", "boop", "You rolled 2", "beep"))
})

# structured data --------------------------------------------------------------

test_that("can extract data in parallel", {
  vcr::local_cassette("parallel-data")

  person <- type_object(name = type_string(), age = type_integer())

  chat <- chat_openai_test()
  data <- parallel_chat_structured(
    chat,
    list("John, age 15", "Jane, age 16"),
    type = person
  )
  expect_equal(data, tibble::tibble(name = c("John", "Jane"), age = c(15, 16)))
})

test_that("can get tokens", {
  vcr::local_cassette("parallel-data")
  person <- type_object(name = type_string(), age = type_integer())
  chat <- chat_openai_test()

  data <- parallel_chat_structured(
    chat,
    list("John, age 15", "Jane, age 16"),
    type = person,
    include_tokens = TRUE
  )
  # These are pretty weak, but it's hard to know how to do better.
  expect_contains(names(data), c("input_tokens", "output_tokens"))
  expect_equal(data$input_tokens > 0, c(TRUE, TRUE))
  expect_equal(data$output_tokens > 0, c(TRUE, TRUE))
})

test_that("can get cost", {
  vcr::local_cassette("parallel-data")
  person <- type_object(name = type_string(), age = type_integer())
  chat <- chat_openai_test()

  data <- parallel_chat_structured(
    chat,
    list("John, age 15", "Jane, age 16"),
    type = person,
    include_cost = TRUE
  )
  expect_contains(names(data), "cost")
  expect_equal(data$cost > 0, c(TRUE, TRUE))
})

test_that("can get tokens & cost", {
  vcr::local_cassette("parallel-data")
  person <- type_object(name = type_string(), age = type_integer())
  chat <- chat_openai_test()

  data <- parallel_chat_structured(
    chat,
    list("John, age 15", "Jane, age 16"),
    type = person,
    include_cost = TRUE,
    include_tokens = TRUE
  )
  expect_contains(names(data), c("input_tokens", "output_tokens", "cost"))
})

test_that("parallel_chat logs tokens", {
  local_tokens()
  vcr::local_cassette("parallel-logging")

  chat <- chat_openai_test()
  prompts <- list("Say 1", "Say 2", "Say 3")
  results <- parallel_chat(chat, prompts)

  input_tokens <- map_dbl(results, \(x) x$get_tokens()$input)
  expect_equal(the$tokens$input, sum(input_tokens))
})

test_that("parallel_chat_structured logs tokens", {
  local_tokens()
  vcr::local_cassette("parallel-structured-logging")

  chat <- chat_openai_test()
  type <- type_object(msg = type_string("A greeting message"))
  prompts <- list("Say hi", "Say bye")
  results <- parallel_chat_structured(
    chat,
    prompts,
    type = type,
    include_tokens = TRUE,
    include_cost = TRUE
  )

  expect_equal(the$tokens$input, sum(results$input_tokens))
  expect_equal(the$tokens$output, sum(results$output_tokens))
})

# error handling ---------------------------------------------------------------

test_that("handles errors and NULLs in parallel functions", {
  chat <- chat_openai(
    credentials = \() "test-key",
    base_url = "http://localhost:1234",
    model = "mock"
  )
  prompts <- list("prompt1", "prompt2", "prompt3")
  responses <- list(
    AssistantTurn("Success", tokens = c(10, 20, 0)),
    simpleError("Request failed"),
    NULL
  )
  local_mocked_bindings(parallel_turns = function(...) responses)

  chats <- parallel_chat(chat, prompts)
  expect_length(chats, 3)
  expect_s3_class(chats[[1]], "Chat")
  expect_s3_class(chats[[2]], "error")
  expect_null(chats[[3]])

  expect_equal(parallel_chat_text(chat, prompts), c("Success", NA, NA))

  responses <- list(
    AssistantTurn(list(ContentJson(list(x = 1))), tokens = c(10, 20, 0)),
    simpleError("Request failed"),
    NULL
  )
  type <- type_object(x = type_number())
  out <- parallel_chat_structured(chat, prompts, type)
  expect_s3_class(out, "data.frame")
  expect_equal(nrow(out), 3)
  expect_named(out, c("x", ".error"))
  expect_equal(out$x, c(1, NA, NA))
})

test_that("errors in conversion become warnings", {
  chat <- chat_openai_test()
  provider <- chat$get_provider()
  type <- type_object(x = type_integer())

  turns <- list(
    AssistantTurn(list(ContentJson(data = list(x = 1)))),
    # no json
    AssistantTurn(list(ContentText("Hello"))),
    # invalid json
    AssistantTurn(list(ContentJson(string = "{")))
  )

  expect_snapshot(out <- multi_convert(provider, turns, type = type))
  expect_equal(out, tibble::tibble(x = c(1, NA, NA)))
})

test_that("assistant turns track duration in parallel", {
  vcr::local_cassette("parallel-duration")

  chat <- chat_openai_test()
  chats <- parallel_chat(chat, list("What's 1 + 1?", "What's 2 + 2?"))

  assistant_duration_1 <- chats[[1]]$last_turn()@duration
  assistant_duration_2 <- chats[[2]]$last_turn()@duration

  # These assistant durations are usually not NA, but are during replay (#479)
  expect_true(is.na(assistant_duration_1) || assistant_duration_1 > 0)
  expect_true(is.na(assistant_duration_2) || assistant_duration_2 > 0)
})

test_that("useful errors", {
  expect_snapshot(error = TRUE, {
    chat()
    chat("")
    chat("susan")
    chat("susan/jones")
  })
})


test_that("can set model or use default", {
  chat1 <- chat("openai")
  expect_equal(chat1$get_provider()@name, "OpenAI")
  expect_equal(chat1$get_provider()@model, chat_openai()$get_provider()@model)

  chat2 <- chat("openai/gpt-4.1-mini")
  expect_equal(chat2$get_provider()@name, "OpenAI")
  expect_equal(chat2$get_provider()@model, "gpt-4.1-mini")

  # We split at the first slash, the remainder is passed to `model`
  chat3 <- chat("openai/provider/model")
  expect_equal(chat3$get_provider()@name, "OpenAI")
  expect_equal(chat3$get_provider()@model, "provider/model")
})

test_that("works for chat functions that don't include `params`", {
  local_mocked_bindings(
    has_ollama = function(...) TRUE,
    models_ollama = function(...) {
      list(id = "qwen3:4b")
    }
  )
  expect_s3_class(chat("ollama/qwen3:4b"), "Chat")

  chat <- chat("ollama/qwen3:4b")
  expect_equal(chat$get_provider()@name, "Ollama")
  expect_equal(chat$get_provider()@model, "qwen3:4b")
})

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_vllm_test("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_vllm_test("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
})

test_that("can list models", {
  test_models(\(...) models_vllm("https://llm.nrp-nautilus.io/"))
})

# Getting started --------------------------------------------------------

test_that("can make simple request", {
  chat <- chat_snowflake("Be as terse as possible; no punctuation")
  resp <- chat$chat("What is 1 + 1?", echo = FALSE)
  expect_match(resp, "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

test_that("can make simple streaming request", {
  chat <- chat_snowflake("Be as terse as possible; no punctuation")
  resp <- coro::collect(chat$stream("What is 1 + 1?"))
  expect_match(paste0(unlist(resp), collapse = ""), "2")
  expect_equal(unname(chat$last_turn()@tokens[1:2] > 0), c(TRUE, TRUE))
})

# Common provider interface -----------------------------------------------

test_that("defaults are reported", {
  # Setting a dummy account ensures we don't skip this test, even if there are
  # no Snowflake credentials available.
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_TOKEN = "token"
  )
  expect_snapshot(. <- chat_snowflake())
})

test_that("supports tool calling", {
  test_tools_simple(chat_snowflake)
})

test_that("can extract data", {
  test_data_extraction(chat_snowflake)
})

test_that("can use images", {
  # Snowflake models don't support images.
  #
  # test_images_inline(chat_snowflake)
  # test_images_remote(chat_snowflake)
})

# Auth --------------------------------------------------------------------

test_snowflake_key <- "-----BEGIN PRIVATE KEY-----
MIIEvgIBADANBgkqhkiG9w0BAQEFAASCBKgwggSkAgEAAoIBAQCbxG4OC5HU9QlK
dmtbQCa7r+uoKyDSisxqJQchfkDy64v6V6WsovI8evUGPQpkAbqsmXY3DR3T/Mco
P2oHyzGsfd2t7v6NLHNtGbMiEJYjVJvOw52Yn1m4WH5bEtl5JP/8W2qyTdr6qym+
m47X8hAqb+ToQjnolRq9xlme9n6vhwGi8mlco5POLCcEDhcYiqcPxI/WRqHDcdi8
/nU1eRhGTSe77NUnC0QQOojjRZ3P59NuA7zpgFdMLdE0I5qrfL3e6SQlnmTizdng
qiZUAI5p1ISZffn1FLf9GZlnD/usG0Dbp2MDGbYMhx8a5ii0RGyEINYrATdxkKaW
/AKGeKbdAgMBAAECggEAH+a022+HKGQeyP9DsWaMCDhZPRHIIRaIEt0Ofs+KobWX
72dv6NFeZwCPmf16WUz5XEv5qACpsTa92wJRxtLYk4kbk3m07FjEMv3mb/2Roh67
4jax2gYYq+aDykcr/uGTA639RhMn29qeLAlT0eojYW2VJfQaRAX1ehRbWnEFNRFR
Pyy1pBOCReDG0yyw1OtDv85H09UdiRWyNC7HkxdaMZns//GOQ3MwuZL61st7Aezg
Xz+mGw7v+SEIL0zk92GSIOHA0TXiUAhIWGxIyqSeNqw+Cl0+4r6ZuT+z2lILPR9C
UPVMtXUzUBhBPhtvPpq2RoRqcHzXWsdUcfteKyN/iQKBgQC5JvzZOwOnBcqaTUpn
ykrYwyiAOk0h3uOs4Mrs7A40xWmQ35VOb1gWVnTgvC91SBBfP/jGf02ZdLk5NG1/
oe13aKvQ6mh/jTImPLEPxsMm+469+nklitHwF8b6R3zrSPHoqdF4XUOHOcbK5V8W
MgUIIXDGtLCqxTns41VbIM9/5wKBgQDXXvgG5238F1LtHFG0FNZRilRt3d+cO1CU
HctSPGRXVe8ZGEJZ4F/TV6pWEOrdsuk5bp/IoDGKE2b9FI6K9BKy3Xc8qx5Um9zF
q5ca671UmZkcqu8jh99JSn9sKM7PP9QZInhP1eca7J9r2lhROHk0hsyTWtzuVcWO
JttBO0lamwKBgQCJWFGCNxO7h0FGewUxvs8MwqA9loH3GScc69e8LlNPdA2eKSzR
dSkL0PB8cTxnLKDwdzzsyixfJEXuGGUNo6nKxTuHCwufarcDxEu4H0JOnZbCeJX7
cmHPT2QL7pHM21yPscEwH0bjfcloYwPJLCutX1kQHaNb2lfg0LZVlh42iwKBgQCW
3yp0+66qiFRJUitSMb6pRHQ8us8ojMy31d9W7oOEQujJ9ZqVh37ZeHIU9KjzQZ/r
4bkBPGc3yLu+0qXAZZarwkUDNQR8VOtldfzWmQn6t9bwpDX99/LNTujQhg3KVXZp
XSJXGwtYayaK0VxJGXye9UdeeqqGM4O/Py0dF0EdvQKBgDo82ImF2mKzJUEBK33r
uGtR8Fxbg4cNRAc0W6xME86IVTnLnqLp1yeTZZGCFek6hDqERLCbQhQk8t1Szm0V
OdYSh6YfkxhsBGp6hHefOTWuoto4zHZ98uuu0GD8NkzGmnZApZ7It1MiH+SZPG9w
AK4HbizZMWlkvg87OphvnQhC
-----END PRIVATE KEY-----
"

test_that("Snowflake keypair token caching works as expected", {
  skip_if_not_installed("jose")

  # Random RSA key for testing.
  testkey <- openssl::read_key(test_snowflake_key)

  token1 <- snowflake_keypair_token("test1", "user", testkey)
  token2 <- snowflake_keypair_token("test2", "user", testkey)

  # Verify different tokens were returned
  expect_false(identical(token1, token2))

  # Verify cached tokens match original ones
  expect_identical(token1, snowflake_keypair_token("test1", "user", testkey))
  expect_identical(token2, snowflake_keypair_token("test2", "user", testkey))

  # Simulate a cache entry that has expired
  cache <- snowflake_keypair_cache("test1", testkey)
  creds_modified <- cache$get()
  creds_modified$expiry <- Sys.time() - 5
  cache$set(creds_modified)

  # Ensure the new token has been updated
  expect_false(
    identical(
      creds_modified,
      snowflake_keypair_token("test1", "user", testkey)
    )
  )
  expect_false(
    identical(token1, snowflake_keypair_token("test1", "user", testkey))
  )
  expect_false(
    identical(token2, snowflake_keypair_token("test1", "user", testkey))
  )
})

test_that("Snowflake OAuth tokens are detected correctly", {
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_TOKEN = "token"
  )
  credentials <- default_snowflake_credentials()
  expect_identical(
    credentials(),
    list(
      Authorization = "Bearer token",
      `X-Snowflake-Authorization-Token-Type` = "OAUTH"
    )
  )
})

test_that("Snowflake key-pair credentials are detected correctly", {
  skip_if_not_installed("jose")

  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_USER = "user",
    SNOWFLAKE_PRIVATE_KEY = test_snowflake_key
  )
  # Warm the cache so we can compare more easily.
  token <- snowflake_keypair_token(
    "testorg-test_account",
    "user",
    openssl::read_key(test_snowflake_key)
  )
  credentials <- default_snowflake_credentials()
  expect_identical(
    credentials(),
    list(
      Authorization = paste("Bearer", token),
      `X-Snowflake-Authorization-Token-Type` = "KEYPAIR_JWT"
    )
  )
})

test_that("snowflake account identifier is extracted from privatelink account and used in JWT claim", {
  skip_if_not_installed("jose")

  testkey <- openssl::read_key(test_snowflake_key)
  token <- snowflake_keypair_token(
    account = "mycompany.us-east-1.privatelink",
    user = "testuser",
    key = testkey
  )
  pubkey <- as.list(testkey)$pubkey
  decoded <- jose::jwt_decode_sig(token, pubkey = pubkey)
  expect_equal(decoded$sub, "MYCOMPANY.TESTUSER")
})

test_that("tokens can be requested from a Connect server", {
  skip_if_not_installed("connectcreds")

  withr::local_envvar(SNOWFLAKE_ACCOUNT = "testorg-test_account")
  connectcreds::local_mocked_connect_responses(token = "token")
  credentials <- default_snowflake_credentials()
  expect_identical(
    credentials(),
    list(
      Authorization = "Bearer token",
      `X-Snowflake-Authorization-Token-Type` = "OAUTH"
    )
  )
})

test_that("we can merge Snowflake's chunk format", {
  # Setting a dummy account ensures we don't skip this test, even if there are
  # no Snowflake credentials available.
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_TOKEN = "token"
  )
  chat <- chat_snowflake()
  provider <- chat$get_provider()
  chunk1 <- list(
    id = "id",
    model = "claude-3-5-sonnet",
    choices = list(
      list(
        delta = list(
          type = "text",
          content = "I",
          content_list = list(list(type = "text", text = "I")),
          text = "I"
        )
      )
    ),
    usage = structure(list(), names = character(0))
  )
  chunk2 <- list(
    id = "id",
    model = "claude-3-5-sonnet",
    choices = list(
      list(
        delta = list(
          type = "text",
          content = " aim to be direct and honest: I don't actually",
          content_list = list(
            list(
              type = "text",
              text = " aim to be direct and honest: I don't actually"
            )
          ),
          text = " aim to be direct and honest: I don't actually"
        )
      )
    ),
    usage = structure(list(), names = character(0))
  )
  expect_equal(
    stream_merge_chunks(
      provider,
      stream_merge_chunks(provider, NULL, chunk1),
      chunk2
    ),
    list(
      id = "id",
      model = "claude-3-5-sonnet",
      choices = list(
        list(
          message = list(
            content_list = list(
              list(
                type = "text",
                text = "I aim to be direct and honest: I don't actually"
              )
            )
          )
        )
      ),
      usage = structure(list(), names = character(0))
    )
  )
})

test_that("we can merge Snowflake's tool_use chunk format", {
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_TOKEN = "token"
  )
  chat <- chat_snowflake()
  provider <- chat$get_provider()

  # Simulate a response that starts with a tool call (no preceding text).
  # This is the scenario that triggers the bug in #938.
  chunk1 <- list(
    choices = list(
      list(
        delta = list(
          type = "tool_use",
          tool_use_id = "toolu_123",
          name = "my_tool",
          content_list = list(
            list(tool_use_id = "toolu_123", name = "my_tool")
          ),
          text = ""
        )
      )
    ),
    usage = structure(list(), names = character(0))
  )
  chunk2 <- list(
    choices = list(
      list(
        delta = list(
          type = "tool_use",
          input = "{\"x\":",
          content_list = list(list(input = "{\"x\":")),
          text = ""
        )
      )
    ),
    usage = structure(list(), names = character(0))
  )
  chunk3 <- list(
    choices = list(
      list(
        delta = list(
          type = "tool_use",
          input = " 5}",
          content_list = list(list(input = " 5}")),
          text = ""
        )
      )
    ),
    usage = structure(list(), names = character(0))
  )
  chunk4 <- list(
    choices = list(
      list(
        delta = list(
          type = "text",
          content_list = list(list(type = "text", text = "")),
          text = ""
        )
      )
    ),
    usage = list(prompt_tokens = 10, completion_tokens = 5)
  )

  result <- stream_merge_chunks(provider, NULL, chunk1)
  result <- stream_merge_chunks(provider, result, chunk2)
  result <- stream_merge_chunks(provider, result, chunk3)
  result <- stream_merge_chunks(provider, result, chunk4)

  expect_equal(
    result$choices[[1]]$message$content_list,
    list(
      list(
        type = "tool_use",
        tool_use_id = "toolu_123",
        name = "my_tool",
        input = "{\"x\": 5}"
      ),
      list(type = "text", text = "")
    )
  )

  # value_turn filters empty text and produces only the tool request
  turn <- value_turn(provider, result)
  expect_length(turn@contents, 1)
  expect_s3_class(turn@contents[[1]], "ellmer::ContentToolRequest")
  expect_equal(turn@contents[[1]]@id, "toolu_123")
  expect_equal(turn@contents[[1]]@name, "my_tool")
  expect_equal(turn@contents[[1]]@arguments, list(x = 5))
})

test_that("stream_content extracts text from Snowflake's delta format", {
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_TOKEN = "token"
  )
  chat <- chat_snowflake()
  provider <- chat$get_provider()

  # Text chunk with content yields ContentText
  text_event <- list(
    choices = list(
      list(
        delta = list(type = "text", content = "hello", text = "hello")
      )
    )
  )
  result <- stream_content(provider, text_event)
  expect_s3_class(result, "ellmer::ContentText")
  expect_equal(result@text, "hello")

  # Text chunk using delta$text (no delta$content)
  text_event2 <- list(
    choices = list(
      list(
        delta = list(type = "text", text = "world")
      )
    )
  )
  result2 <- stream_content(provider, text_event2)
  expect_s3_class(result2, "ellmer::ContentText")
  expect_equal(result2@text, "world")

  # Tool_use chunk returns NULL (not text)
  tool_event <- list(
    choices = list(
      list(
        delta = list(type = "tool_use", tool_use_id = "id", name = "fn")
      )
    )
  )
  expect_null(stream_content(provider, tool_event))

  # Empty text chunk returns NULL
  empty_event <- list(
    choices = list(
      list(
        delta = list(type = "text", text = "")
      )
    )
  )
  expect_null(stream_content(provider, empty_event))
})

test_that("as_json handles tool-only assistant turns", {
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_TOKEN = "token"
  )
  chat <- chat_snowflake()
  provider <- chat$get_provider()

  # An assistant turn with only a tool request (no preceding text)
  turn <- Turn(
    "assistant",
    contents = list(
      ContentToolRequest("toolu_123", "my_tool", list(x = 5))
    )
  )
  result <- as_json(provider, turn)
  expect_equal(result$role, "assistant")
  expect_equal(result$content_list[[1]]$type, "tool_use")
  expect_equal(result$content_list[[1]]$tool_use$tool_use_id, "toolu_123")
})

test_that("chat_snowflake() supports parameters", {
  # Setting a dummy account ensures we don't skip this test, even if there are
  # no Snowflake credentials available.
  withr::local_envvar(
    SNOWFLAKE_ACCOUNT = "testorg-test_account",
    SNOWFLAKE_TOKEN = "token"
  )
  chat <- chat_snowflake(
    model = "claude-3-5-sonnet",
    params = params(
      temperature = 0.7,
      max_tokens = 100,
      top_p = 0.9,
      # This is a Snowflake-specific parameter.
      guardrails = list(
        enabled = TRUE
      )
    )
  )
  provider <- chat$get_provider()
  expect_equal(
    chat_body(provider),
    list(
      model = "claude-3-5-sonnet",
      temperature = 0.7,
      top_p = 0.9,
      max_tokens = 100,
      guardrails = list(enabled = TRUE),
      stream = TRUE
    )
  )
  # Warn/ignore unsupported parameters.
  provider@params <- params(seed = 1L)
  expect_warning(
    out <- chat_body(provider),
    regexp = "unsupported parameters"
  )
  expect_equal(out, list(model = "claude-3-5-sonnet", stream = TRUE))
})

test_that("tracing works as expected for synchronous chats", {
  skip_if_not_installed("otelsdk")
  withr::local_envvar(OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT = NA)

  # Capture spans from a typical synchronous chat with tool calls.
  spans <- with_otel_record({
    test_tools_simple(chat_openai_test)
  })[["traces"]]

  # Check we have two top-level "invoke_agent" spans (one for each chat()
  # invocation) that start their respective traces.
  agent_spans <- Filter(function(x) x$name == "invoke_agent", spans)
  expect_length(agent_spans, 2L)
  expect_equal(agent_spans[[1L]]$parent, agent_spans[[2L]]$parent)
  expect_equal(agent_spans[[1L]]$kind, "client")
  agent_span_ids <- sapply(agent_spans, function(x) x$span_id)

  # We should have (at least) two "execute_tool" spans
  # (one for each tool invocation) that are children of the agent spans.
  # Note 2025/11: Some models may call tools early and cache the results, so we can not check for existance of multiple parent ids (newer openai models)
  # Note 2025/11: Some models call more tools than necessary
  # Ex: anthropic calls the date tool twice. So we check for at least 2.
  tool_spans <- Filter(function(x) startsWith(x$name, "execute_tool"), spans)
  expect_gte(length(tool_spans), 2L)
  expect_true(all(vapply(
    tool_spans,
    function(x) x$parent %in% agent_span_ids,
    logical(1)
  )))
  expect_equal(tool_spans[[1L]]$kind, "internal")

  # And "chat" spans that correspond to model calls before and after each
  # tool call -- these are also children of the agent spans and siblings of one
  # another.
  # Note 2025/09: Not all models are calling tools the same. There must be AT LEAST 2 chat spans. Ex: anthropic has 4, openai has 3. :shrug:
  chat_spans <- Filter(function(x) startsWith(x$name, "chat"), spans)
  expect_gte(length(chat_spans), 2L)
  expect_true(all(vapply(
    chat_spans,
    function(x) x$parent %in% agent_span_ids,
    logical(1)
  )))
  expect_equal(chat_spans[[1L]]$kind, "client")

  # Ensure we record the result (and therefore set the status) of chat spans.
  expect_true(all(vapply(chat_spans, function(x) x$status == "ok", logical(1))))

  # And that token usage attributes are recorded with positive integer counts.
  expect_true(all(vapply(
    chat_spans,
    function(x) {
      input <- x$attributes[["gen_ai.usage.input_tokens"]]
      output <- x$attributes[["gen_ai.usage.output_tokens"]]
      !is.null(input) &&
        !is.null(output) &&
        input > 0 &&
        output > 0 &&
        input == round(input) &&
        output == round(output)
    },
    logical(1)
  )))

  # We should also get some underlying HTTP spans that correspond to the model
  # calls.
  chat_span_ids <- sapply(chat_spans, function(x) x$span_id)
  http_spans <- Filter(function(x) startsWith(x$name, "POST"), spans)
  expect_true(all(vapply(
    http_spans,
    function(x) x$parent %in% chat_span_ids,
    logical(1)
  )))

  # Content-capture attributes are off by default.
  expect_null(chat_spans[[1]]$attributes[["gen_ai.system_instructions"]])
  expect_null(chat_spans[[1]]$attributes[["gen_ai.input.messages"]])
  expect_null(chat_spans[[1]]$attributes[["gen_ai.output.messages"]])
})

test_that("tracing works as expected for synchronous streams", {
  skip_if_not_installed("otelsdk")

  # Capture spans when a stream is suspended.
  spans <- with_otel_record({
    chat <- chat_openai_test(system_prompt = "Be terse.")
    stream <- chat$stream("What's your favourite Apache Spark feature?")
    local(otel::start_local_active_span("simultaneous"))
    coro::collect(stream)
  })[["traces"]]

  # Check we have one top-level "invoke_agent" span.
  expect_length(spans[names(spans) == "invoke_agent"], 1L)

  # And one child span for the stream (confirming that it ends when collected).
  chat_spans <- Filter(function(x) startsWith(x$name, "chat"), spans)
  expect_length(chat_spans, 1L)
  expect_equal(chat_spans[[1L]]$parent, spans[["invoke_agent"]]$span_id)

  # Token usage attributes are recorded on the streamed chat span.
  expect_true(all(vapply(
    chat_spans,
    function(x) {
      input <- x$attributes[["gen_ai.usage.input_tokens"]]
      output <- x$attributes[["gen_ai.usage.output_tokens"]]
      !is.null(input) &&
        !is.null(output) &&
        input > 0 &&
        output > 0 &&
        input == round(input) &&
        output == round(output)
    },
    logical(1)
  )))

  # Verify that the span started when the stream was suspended is not part of
  # the agent trace.
  expect_equal(spans[["simultaneous"]]$parent, spans[["invoke_agent"]]$parent)

  # But that HTTP spans are.
  chat_span_ids <- sapply(chat_spans, function(x) x$span_id)
  http_spans <- Filter(function(x) startsWith(x$name, "POST"), spans)
  expect_true(all(vapply(
    http_spans,
    function(x) x$parent %in% chat_span_ids,
    logical(1)
  )))
})

test_that("tracing works as expected for asynchronous chats", {
  skip_if_not_installed("otelsdk")

  # chat <- chat_openai_test(
  chat <- chat_anthropic_test(
    system_prompt = "Always use a tool to answer. Reply with 'It is ____.'."
  )
  chat$register_tool(tool(
    coro::async(function() "2024-01-01"),
    "Return the current date",
    name = "current_date"
  ))

  # Capture spans for an async chat with async tool calls interleaved with
  # other synchronous and asynchronous spans.
  spans <- with_otel_record({
    p1 <- chat$chat_async("What's the current date in Y-M-D format?") |>
      promises::then(function(result) {
        chat$chat_async("What date will it be 47 days from now?")
      })
    p2 <- promises::promise(function(resolve, reject) {
      span <- otel::start_span("concurrent")
      otel::local_active_span(span) # just to try to mess with things
      later::later(
        function() {
          otel::local_active_span(span) # just to try to mess with things
          resolve(NULL)
          span$end()
        },
        0.1
      )
    })

    local({
      # Typical external usage
      local(otel::start_local_active_span("simultaneous"))
    })

    sync(p2)
    sync(p1)
  })[["traces"]]

  ## Debug span output; name, id, parent
  # ignore <- Map(
  #   spans,
  #   format(names(spans), justify = "right"),
  #   f = function(span, name) {
  #     message(name, " - ", span$span_id, " - ", span$parent)
  #   }
  # )

  # Check we have two top-level extra spans
  expect_length(spans[names(spans) == "concurrent"], 1L)
  expect_length(spans[names(spans) == "simultaneous"], 1L)
  expect_equal(spans[["concurrent"]]$parent, spans[["simultaneous"]]$parent)

  # Check we have two top-level "invoke_agent" spans (one for each chat()
  # invocation) that start their respective traces.
  agent_spans <- Filter(function(x) x$name == "invoke_agent", spans)
  expect_length(agent_spans, 2L)
  expect_equal(agent_spans[[1]]$parent, spans[["concurrent"]]$parent)
  expect_equal(agent_spans[[2]]$parent, spans[["concurrent"]]$parent)

  agent_span_ids <- sapply(agent_spans, function(x) x$span_id)

  # We should have two "execute_tool" spans (one for each tool invocation)
  # that are children of the agent spans.
  tool_spans <- Filter(function(x) startsWith(x$name, "execute_tool"), spans)
  expect_gte(length(tool_spans), 1L)
  expect_true(all(vapply(
    tool_spans,
    function(x) x$parent %in% agent_span_ids,
    logical(1)
  )))

  # And four "chat" spans that correspond to model calls before and after each
  # tool call -- these are also children of the agent spans and siblings of one
  # another.
  chat_spans <- Filter(function(x) startsWith(x$name, "chat"), spans)
  expect_gte(length(chat_spans), 2 * length(tool_spans))
  expect_true(all(vapply(
    chat_spans,
    function(x) x$parent %in% agent_span_ids,
    logical(1)
  )))

  # Ensure we record the result (and therefore set the status) of chat spans.
  expect_true(all(vapply(chat_spans, function(x) x$status == "ok", logical(1))))

  # Token usage attributes are recorded with positive integer counts. This
  # exercises the Anthropic path that previously folded a 1.25x cache-write
  # pricing weight into the input count.
  expect_true(all(vapply(
    chat_spans,
    function(x) {
      input <- x$attributes[["gen_ai.usage.input_tokens"]]
      output <- x$attributes[["gen_ai.usage.output_tokens"]]
      !is.null(input) &&
        !is.null(output) &&
        input > 0 &&
        output > 0 &&
        input == round(input) &&
        output == round(output)
    },
    logical(1)
  )))

  # We should also get some underlying HTTP spans that correspond to the model
  # calls.
  chat_span_ids <- sapply(chat_spans, function(x) x$span_id)
  http_spans <- Filter(function(x) startsWith(x$name, "POST"), spans)
  expect_true(all(vapply(
    http_spans,
    function(x) x$parent %in% chat_span_ids,
    logical(1)
  )))
})

test_that("tracing works as expected for asynchronous streams", {
  skip_if_not_installed("otelsdk")

  # Capture spans when an async stream is used in concert with other
  # synchronous and asynchronous spans.
  spans <- with_otel_record({
    chat <- chat_openai_test(system_prompt = "Be terse.")
    stream <- chat$stream_async("What's your favourite Apache Spark feature?")
    p <- promises::promise(function(resolve, reject) {
      span <- local(otel::start_span("concurrent"))
      later::later(
        function() {
          on.exit(span$end())
          otel::with_active_span(span, {
            resolve(NULL)
          })
        },
        0.1
      )
    })
    local(otel::start_local_active_span("simultaneous"))
    sync(p)
    sync(coro::async_collect(stream))
  })[["traces"]]

  # Check we have one top-level "invoke_agent" span.
  expect_length(spans[names(spans) == "invoke_agent"], 1L)

  # And one child span for the stream (confirming that it ends when collected).
  chat_spans <- Filter(function(x) startsWith(x$name, "chat"), spans)
  expect_length(chat_spans, 1L)
  expect_equal(chat_spans[[1L]]$parent, spans[["invoke_agent"]]$span_id)

  # Verify that the spans started when the stream was suspended are not part of
  # the agent trace.
  expect_equal(spans[["concurrent"]]$parent, spans[["invoke_agent"]]$parent)
  expect_equal(spans[["simultaneous"]]$parent, spans[["invoke_agent"]]$parent)

  # But that HTTP spans are.
  chat_span_ids <- sapply(chat_spans, function(x) x$span_id)
  http_spans <- Filter(function(x) startsWith(x$name, "POST"), spans)
  expect_true(all(vapply(
    http_spans,
    function(x) x$parent %in% chat_span_ids,
    logical(1)
  )))
})

test_that("captures content when OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT is set", {
  skip_if_not_installed("otelsdk")
  withr::local_envvar(
    OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT = "true"
  )

  spans <- with_otel_record({
    chat <- chat_openai_test(system_prompt = "Be terse.")
    chat$chat("Say hi.")
  })[["traces"]]

  chat_span <- Filter(function(x) startsWith(x$name, "chat"), spans)[[1]]

  # System instructions: JSON parts array, first part is text.
  sys <- jsonlite::fromJSON(
    chat_span$attributes[["gen_ai.system_instructions"]],
    simplifyVector = FALSE
  )
  expect_equal(sys[[1]]$type, "text")

  # Input messages: last is the user turn with a text part.
  msgs <- jsonlite::fromJSON(
    chat_span$attributes[["gen_ai.input.messages"]],
    simplifyVector = FALSE
  )
  last <- msgs[[length(msgs)]]
  expect_equal(last$role, "user")

  # Output: assistant message with at least one part.
  out <- jsonlite::fromJSON(
    chat_span$attributes[["gen_ai.output.messages"]],
    simplifyVector = FALSE
  )
  expect_equal(out[[1]]$role, "assistant")
  expect_gt(length(out[[1]]$parts), 0)
})

test_that("tool_call_response part omits id when request is unknown", {
  part <- as_otel_part(ContentToolResult(value = "ok"))
  expect_false("id" %in% names(part))
  expect_equal(part$type, "tool_call_response")
  expect_equal(part$response, "ok")

  request <- ContentToolRequest(id = "abc", name = "f", arguments = list())
  part <- as_otel_part(ContentToolResult(value = "ok", request = request))
  expect_equal(part$id, "abc")
})

test_that("tool_call_response part decodes json-class values", {
  value <- structure('{"foo":1,"bar":[2,3]}', class = "json")
  part <- as_otel_part(ContentToolResult(value = value))
  expect_equal(part$response, list(foo = 1, bar = list(2, 3)))

  encoded <- jsonlite::toJSON(part, auto_unbox = TRUE, null = "null")
  expect_equal(
    jsonlite::fromJSON(encoded, simplifyVector = FALSE)$response,
    list(foo = 1, bar = list(2, 3))
  )
})

#' @include provider.R
#' @include provider-openai-compatible.R
#' @include content.R
#' @include turns.R
#' @include tools-def.R
NULL

#' Chat with an OpenAI model
#'
#' @description
#' This is the main interface to [OpenAI](https://openai.com/)'s models,
#' using the **responses API**. You can use this to access OpenAI's latest
#' models and features like image generation and web search. If you need to use
#' an OpenAI-compatible API from another provider, or the **chat completions**
#' API with OpenAI,use [chat_openai_compatible()] instead.
#'
#' Note that a ChatGPT Plus membership does not grant access to the API.
#' You will need to sign up for a developer account (and pay for it) at the
#' [developer platform](https://platform.openai.com).
#'
#' @param system_prompt A system prompt to set the behavior of the assistant.
#' @param base_url The base URL to the API endpoint.
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("OPENAI_API_KEY")`
#' @param model `r param_model("gpt-4.1", "openai")`
#' @param params Common model parameters, usually created by [params()].
#' @param api_args Named list of arbitrary extra arguments appended to the body
#'   of every chat API call. Combined with the body object generated by ellmer
#'   with [modifyList()].
#' @param api_headers Named character vector of arbitrary extra headers appended
#'   to every chat API call.
#' @param service_tier Request a specific service tier. There are four options:
#'   * `"auto"` (default): uses the service tier configured in Project settings.
#'   * `"default"`: standard pricing and performance.
#'   * `"flex"`: slower and cheaper.
#'   * `"priority"`: faster and more expensive.
#' @param echo One of the following options:
#'   * `none`: don't emit any output (default when running in a function).
#'   * `output`: echo text and tool-calling output as it streams in (default
#'     when running at the console).
#'   * `all`: echo all input and output.
#'
#'   Note this only affects the `chat()` method.
#' @family chatbots
#' @export
#' @returns A [Chat] object.
#' @examples
#' \dontshow{ellmer:::vcr_example_start("chat_openai")}
#' chat <- chat_openai()
#' chat$chat("
#'   What is the difference between a tibble and a data frame?
#'   Answer with a bulleted list
#' ")
#'
#' chat$chat("Tell me three funny jokes about statisticians")
#' \dontshow{ellmer:::vcr_example_end()}
chat_openai <- function(
  system_prompt = NULL,
  base_url = "https://api.openai.com/v1",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  api_headers = character(),
  service_tier = c("auto", "default", "flex", "priority"),
  echo = c("none", "output", "all")
) {
  model <- set_default(model, "gpt-4.1")
  echo <- check_echo(echo)
  service_tier <- arg_match(service_tier)

  credentials <- as_credentials(
    "chat_openai",
    \() openai_key(),
    credentials = credentials,
    api_key = api_key
  )

  provider <- ProviderOpenAI(
    name = "OpenAI",
    base_url = base_url,
    model = model,
    params = params %||% params(),
    extra_args = api_args,
    extra_headers = api_headers,
    credentials = credentials,
    service_tier = service_tier
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

#' @rdname chat_openai
#' @export
models_openai <- function(
  base_url = "https://api.openai.com/v1",
  api_key = NULL,
  credentials = NULL
) {
  credentials <- as_credentials(
    "models_openai",
    \() openai_key(),
    credentials = credentials,
    api_key = api_key
  )

  provider <- ProviderOpenAICompatible(
    name = "OpenAI",
    model = "",
    base_url = base_url,
    credentials = credentials
  )

  req <- base_request(provider)
  req <- req_url_path_append(req, "/models")
  resp <- req_perform(req)

  json <- resp_body_json(resp)

  id <- map_chr(json$data, "[[", "id")
  created <- as.Date(.POSIXct(map_int(json$data, "[[", "created")))
  owned_by <- map_chr(json$data, "[[", "owned_by")

  df <- data.frame(
    id = id,
    created_at = created,
    owned_by = owned_by
  )
  df <- cbind(df, match_prices(provider@name, df$id))
  df[order(-xtfrm(df$created_at)), ]
}

chat_openai_test <- function(
  system_prompt = "Be terse.",
  ...,
  model = "gpt-4.1-nano",
  params = NULL,
  echo = "none"
) {
  params <- params %||% params()
  params$temperature <- params$temperature %||% 0

  chat_openai(
    system_prompt = system_prompt,
    model = model,
    params = params,
    ...,
    echo = echo
  )
}

ProviderOpenAI <- new_class(
  "ProviderOpenAI",
  parent = ProviderOpenAICompatible,
  properties = list(
    service_tier = class_character
  )
)


# Chat endpoint ----------------------------------------------------------------

method(chat_path, ProviderOpenAI) <- function(provider) {
  "/responses"
}

# https://platform.openai.com/docs/api-reference/responses
method(chat_body, ProviderOpenAI) <- function(
  provider,
  stream = TRUE,
  turns = list(),
  tools = list(),
  type = NULL
) {
  input <- compact(unlist(as_json(provider, turns), recursive = FALSE))
  tools <- as_json(provider, unname(tools))

  if (!is.null(type)) {
    # https://platform.openai.com/docs/api-reference/responses/create#responses-create-text
    text <- list(
      format = list(
        type = "json_schema",
        name = "structured_data",
        schema = as_json(provider, type),
        strict = TRUE
      )
    )
  } else {
    text <- NULL
  }

  # https://platform.openai.com/docs/api-reference/responses/create#responses-create-include
  params <- chat_params(provider, provider@params)

  if (has_name(params, "reasoning_effort")) {
    reasoning <- list(
      effort = params$reasoning_effort,
      summary = "auto"
    )
    params$reasoning_effort <- NULL
  } else {
    reasoning <- NULL
  }

  include <- c(
    if (isTRUE(params$log_probs)) "message.output_text.logprobs",
    if (is_openai_reasoning(provider@model)) "reasoning.encrypted_content"
  )
  params$log_probs <- NULL

  compact(list2(
    input = input,
    include = as.list(include),
    model = provider@model,
    !!!params,
    stream = stream,
    tools = tools,
    text = text,
    reasoning = reasoning,
    store = FALSE,
    service_tier = provider@service_tier
  ))
}


method(chat_params, ProviderOpenAI) <- function(provider, params) {
  standardise_params(
    params,
    c(
      temperature = "temperature",
      top_p = "top_p",
      frequency_penalty = "frequency_penalty",
      max_tokens = "max_output_tokens",
      log_probs = "log_probs",
      top_logprobs = "top_k",
      reasoning_effort = "reasoning_effort"
    )
  )
}

# OpenAI -> ellmer --------------------------------------------------------------

method(stream_content, ProviderOpenAI) <- function(provider, event) {
  if (event$type == "response.output_text.delta") {
    # https://platform.openai.com/docs/api-reference/responses-streaming/response/output_text/delta
    if (is.null(event$delta)) {
      return(NULL)
    }
    ContentText(event$delta)
  } else if (event$type == "response.reasoning_summary_text.delta") {
    # https://platform.openai.com/docs/api-reference/responses-streaming/response/reasoning_summary_text/delta
    ContentThinking(event$delta)
  } else if (event$type == "response.reasoning_summary_text.done") {
    # https://platform.openai.com/docs/api-reference/responses-streaming/response/reasoning_summary_text/done
    NULL
  }
}
method(stream_merge_chunks, ProviderOpenAI) <- function(
  provider,
  result,
  chunk
) {
  if (chunk$type == "response.completed") {
    # https://platform.openai.com/docs/api-reference/responses-streaming/response/completed
    chunk$response
  } else if (chunk$type == "response.failed") {
    # https://platform.openai.com/docs/api-reference/responses-streaming/response/failed
    error <- chunk$response$error
    cli::cli_abort(c("Request failed ({error$code})", "{error$message}"))
  } else if (chunk$type == "error") {
    # https://platform.openai.com/docs/api-reference/responses-streaming/error
    error <- chunk$error
    cli::cli_abort(c("Request errored ({error$type})", "{error$message}"))
  }
}

method(value_tokens, ProviderOpenAI) <- function(provider, json) {
  usage <- json$usage
  cached_tokens <- usage$input_tokens_details$cached_tokens %||% 0

  tokens(
    input = (usage$input_tokens %||% 0) - cached_tokens,
    output = usage$output_tokens %||% 0,
    cached_input = cached_tokens
  )
}

method(value_turn, ProviderOpenAI) <- function(
  provider,
  result,
  has_type = FALSE
) {
  contents <- lapply(result$output, function(output) {
    if (output$type == "message") {
      if (has_type) {
        ContentJson(jsonlite::parse_json(output$content[[1]]$text))
      } else {
        ContentText(output$content[[1]]$text)
      }
    } else if (output$type == "function_call") {
      arguments <- jsonlite::parse_json(output$arguments)
      ContentToolRequest(output$id, output$name, arguments)
    } else if (output$type == "reasoning") {
      thinking <- paste0(map_chr(output$summary, "[[", "text"), collapse = "")
      ContentThinking(thinking = thinking, extra = output)
    } else if (output$type == "image_generation_call") {
      mime_type <- switch(
        output$output_format,
        png = "image/png",
        jpeg = "image/jpeg",
        webp = "image/webp",
        "unknown"
      )
      ContentImageInline(mime_type, output$result)
    } else if (output$type == "web_search_call") {
      # https://platform.openai.com/docs/guides/tools-web-search#output-and-citations
      first_query <- if (length(output$action$queries)) {
        output$action$queries[[1]]
      }
      query <- output$action$query %||%
        first_query %||%
        output$action$url %||%
        "web search"
      ContentToolRequestSearch(query = query, json = output)
    } else {
      browser()
      cli::cli_abort(
        "Unknown content type {.str {output$type}}.",
        .internal = TRUE
      )
    }
  })

  tokens <- value_tokens(provider, result)
  variant <- result$service_tier %||% "default"
  cost <- get_token_cost(provider, tokens, variant = variant)
  AssistantTurn(
    contents = contents,
    json = result,
    tokens = unlist(tokens),
    cost = cost
  )
}

# ellmer -> OpenAI --------------------------------------------------------------

method(as_json, list(ProviderOpenAI, Turn)) <- function(
  provider,
  x,
  ...
) {
  x <- turn_contents_expand(x)

  # While the user turn can contain multiple contents, the assistant turn
  # can't. Fortunately, we can send multiple user turns with out issue.
  as_json(provider, x@contents, ..., role = x@role)
}

method(as_json, list(ProviderOpenAI, ContentText)) <- function(
  provider,
  x,
  ...,
  role
) {
  type <- if (role %in% c("user", "system")) "input_text" else "output_text"
  list(
    role = role,
    content = list(list(type = type, text = x@text))
  )
}

method(as_json, list(ProviderOpenAI, ContentThinking)) <- function(
  provider,
  x,
  ...
) {
  x@extra
}

method(as_json, list(ProviderOpenAI, ContentImageRemote)) <- function(
  provider,
  x,
  ...
) {
  list(
    role = "user",
    content = list(
      list(type = "input_image", image_url = x@url)
    )
  )
}

method(as_json, list(ProviderOpenAI, ContentImageInline)) <- function(
  provider,
  x,
  ...
) {
  list(
    role = "user",
    content = list(
      list(
        type = "input_image",
        image_url = paste0("data:", x@type, ";base64,", x@data)
      )
    )
  )
}

method(as_json, list(ProviderOpenAI, ContentPDF)) <- function(
  provider,
  x,
  ...
) {
  # https://platform.openai.com/docs/guides/pdf-files?api-mode=responses
  list(
    role = "user",
    content = list(list(
      type = "input_file",
      filename = x@filename,
      file_data = paste0("data:application/pdf;base64,", x@data)
    ))
  )
}

method(as_json, list(ProviderOpenAI, ContentToolRequest)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "function_call",
    call_id = x@id,
    name = x@name,
    arguments = to_json(x@arguments)
  )
}

method(as_json, list(ProviderOpenAI, ContentToolResult)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "function_call_output",
    call_id = x@request@id,
    output = tool_string(x)
  )
}

method(as_json, list(ProviderOpenAI, ToolDef)) <- function(
  provider,
  x,
  ...
) {
  list(
    type = "function",
    name = x@name,
    description = x@description,
    strict = TRUE,
    parameters = as_json(provider, x@arguments, ...)
  )
}

# Batched requests -------------------------------------------------------------

method(has_batch_support, ProviderOpenAI) <- function(provider) {
  TRUE
}

# https://platform.openai.com/docs/api-reference/batch
method(batch_submit, ProviderOpenAI) <- function(
  provider,
  conversations,
  type = NULL
) {
  path <- local_tempfile()

  # First put the requests in a file
  # https://platform.openai.com/docs/api-reference/batch/request-input
  requests <- map(seq_along(conversations), function(i) {
    body <- chat_body(
      provider,
      stream = FALSE,
      turns = conversations[[i]],
      type = type
    )

    list(
      custom_id = paste0("chat-", i),
      method = "POST",
      url = "/v1/responses",
      body = body
    )
  })
  json <- map_chr(requests, to_json)
  writeLines(json, path)
  # Then upload it
  uploaded <- openai_upload(provider, path)

  # Now we can submit the
  req <- base_request(provider)
  req <- req_url_path_append(req, "/batches")
  req <- req_body_json(
    req,
    list(
      input_file_id = uploaded$id,
      endpoint = "/v1/responses",
      completion_window = "24h"
    )
  )

  resp <- req_perform(req)
  resp_body_json(resp)
}

# https://platform.openai.com/docs/api-reference/files/create
openai_upload <- function(provider, path, purpose = "batch") {
  req <- base_request(provider)
  req <- req_url_path_append(req, "/files")
  req <- req_body_multipart(
    req,
    purpose = purpose,
    file = curl::form_file(path)
  )
  req <- req_progress(req, "up")

  resp <- req_perform(req)
  resp_body_json(resp)
}

# https://platform.openai.com/docs/api-reference/batch/retrieve
method(batch_poll, ProviderOpenAI) <- function(provider, batch) {
  req <- base_request(provider)
  req <- req_url_path_append(req, "/batches/", batch$id)

  resp <- req_perform(req)
  resp_body_json(resp)
}
method(batch_status, ProviderOpenAI) <- function(provider, batch) {
  list(
    working = batch$status != "completed",
    n_processing = batch$request_counts$total - batch$request_counts$completed,
    n_succeeded = batch$request_counts$completed,
    n_failed = batch$request_counts$failed
  )
}


# https://platform.openai.com/docs/api-reference/batch/retrieve
method(batch_retrieve, ProviderOpenAI) <- function(provider, batch) {
  # output file
  path_output <- withr::local_tempfile()
  openai_download_file(provider, batch$output_file_id, path_output)
  json <- read_ndjson(path_output, fallback = openai_json_fallback)

  # error file
  if (length(batch$error_file_id) == 1) {
    path_error <- withr::local_tempfile()
    openai_download_file(provider, batch$error_file_id, path_error)

    json <- c(json, read_ndjson(path_error, fallback = openai_json_fallback))
  }

  ids <- as.numeric(gsub("chat-", "", map_chr(json, "[[", "custom_id")))
  results <- lapply(json, "[[", "response")
  results[order(ids)]
}

openai_download_file <- function(provider, id, path) {
  req <- base_request(provider)
  req <- req_url_path_append(req, "/files/", id, "/content")
  req <- req_progress(req, "down")
  req_perform(req, path = path)

  invisible(path)
}
openai_json_fallback <- function(line) {
  list(
    custom_id = extract_custom_id(line),
    response = list(status_code = 500)
  )
}
extract_custom_id <- function(json_string) {
  pattern <- '"custom_id"\\s*:\\s*"([^"]*)"'
  match <- regexec(pattern, json_string)

  result <- regmatches(json_string, match)[[1]]
  if (length(result) != 2) {
    NA_character_
  } else {
    result[[2]] # Second element is the captured group
  }
}

method(batch_result_turn, ProviderOpenAI) <- function(
  provider,
  result,
  has_type = FALSE
) {
  if (result$status_code == 200) {
    value_turn(provider, result$body, has_type = has_type)
  } else {
    NULL
  }
}

# Helpers ------------------------------------------------------------------

is_openai_reasoning <- function(model) {
  # https://platform.openai.com/docs/models/compare
  startsWith(model, "o") || startsWith(model, "gpt-5")
}

# tests use `chat_openai()` to verify environments flow through correctly

test_that("api_key is deprecated", {
  expect_snapshot(chat <- chat_openai_compatible_test(api_key = "abc"))
  expect_equal(chat$get_provider()@credentials(), "Bearer abc")
})

test_that("errors if both credentials and api_key are provided", {
  expect_snapshot(
    chat_openai_compatible_test(credentials = "abc", api_key = "def"),
    error = TRUE
  )
})

test_that("verifies all properties of credentials", {
  expect_snapshot(error = TRUE, {
    chat_openai_test(credentials = 1)
    chat_openai_test(credentials = \(a, b) a + b)
    chat_openai_test(credentials = \() 1)
  })
})

as_credentials <- function(
  fun_name,
  default,
  credentials = NULL,
  api_key = NULL,
  token = FALSE,
  env = caller_env(),
  user_env = caller_env(2)
) {
  if (is.null(credentials) && is.null(api_key)) {
    default
  } else if (!is.null(credentials) && is.null(api_key)) {
    check_credentials(credentials, error_call = env)
    credentials
  } else if (is.null(credentials) && !is.null(api_key)) {
    check_string(api_key, allow_null = TRUE, call = env)

    lifecycle::deprecate_warn(
      "0.4.0",
      paste0(fun_name, "(api_key)"),
      paste0(fun_name, "(credentials)"),
      env = env,
      user_env = user_env
    )
    if (token) {
      function() paste0("Bearer ", api_key)
    } else {
      function() api_key
    }
  } else {
    cli::cli_abort(
      "Must supply one of {.arg api_key} or {.arg credentials}.",
      call = env
    )
  }
}

check_credentials <- function(credentials, error_call = caller_env()) {
  check_function(credentials, allow_null = TRUE, call = error_call)
  if (length(formals(credentials)) != 0) {
    cli::cli_abort(
      "{.arg credentials} must not have arguments.",
      call = error_call
    )
  }

  creds <- credentials()

  if (is.function(creds)) {
    # Carve out for special internal cases, e.g. OAuth for gemini
    check_function2(creds, arg = "credentials()", args = "req")
  } else if (is_string(creds)) {
    # api key
  } else if (is_named(creds) && is.list(creds)) {
    # headers
  } else {
    stop_input_type(
      creds,
      c("a string", "a named list"),
      call = error_call,
      arg = "credentials()"
    )
  }

  invisible()
}

ellmer_req_credentials <- function(req, credentials, key_name = NULL) {
  if (is.function(credentials)) {
    return(credentials(req))
  }

  if (is_string(credentials)) {
    if (is.null(key_name)) {
      cli::cli_abort(
        "A `credentials()` function that returns a string is not supported by this provider.",
        call = NULL
      )
    }

    if (key_name == "Authorization") {
      credentials <- paste0("Bearer ", credentials)
    }

    credentials <- set_names(list(credentials), key_name)
  }
  req_headers_redacted(req, !!!credentials)
}

#' @include provider-openai-compatible.R
NULL

#' Chat with a model hosted on Groq
#'
#' @description
#' Sign up at <https://groq.com>.
#'
#' Built on top of [chat_openai_compatible()].
#'
#' @export
#' @family chatbots
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("GROQ_API_KEY")`
#' @param model `r param_model("llama-3.1-8b-instant")`
#' @param params Common model parameters, usually created by [params()].
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_groq()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_groq <- function(
  system_prompt = NULL,
  base_url = "https://api.groq.com/openai/v1",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
) {
  model <- set_default(model, "llama-3.1-8b-instant")
  echo <- check_echo(echo)

  credentials <- as_credentials(
    "chat_groq",
    function() groq_key(),
    credentials = credentials,
    api_key = api_key
  )

  # https://console.groq.com/docs/api-reference#chat-create (same as OpenAI)
  params <- params %||% params()

  provider <- ProviderGroq(
    name = "Groq",
    base_url = base_url,
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

ProviderGroq <- new_class("ProviderGroq", parent = ProviderOpenAICompatible)


method(as_json, list(ProviderGroq, ToolDef)) <- function(provider, x, ...) {
  list(
    type = "function",
    "function" = compact(list(
      name = x@name,
      description = x@description,
      parameters = as_json(provider, x@arguments, ...)
    ))
  )
}

groq_key <- function() {
  key_get("GROQ_API_KEY")
}

#' Upload a file to gemini
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' This function uploads a file then waits for Gemini to finish processing it
#' so that you can immediately use it in a prompt. It's experimental because
#' it's currently Gemini specific, and we expect other providers to evolve
#' similar feature in the future.
#'
#' Uploaded files are automatically deleted after 2 days. Each file must be
#' less than 2 GB and you can upload a total of 20 GB. ellmer doesn't currently
#' provide a way to delete files early; please
#' [file an issue](https://github.com/tidyverse/ellmer/issues) if this would
#' be useful for you.
#'
#' @inheritParams chat_google_gemini
#' @param path Path to a file to upload.
#' @param mime_type Optionally, specify the mime type of the file.
#'   If not specified, will be guesses from the file extension.
#' @returns A `<ContentUploaded>` object that can be passed to `$chat()`.
#' @export
#' @examples
#' \dontrun{
#' file <- google_upload("path/to/file.pdf")
#'
#' chat <- chat_google_gemini()
#' chat$chat(file, "Give me a three paragraph summary of this PDF")
#' }
google_upload <- function(
  path,
  base_url = "https://generativelanguage.googleapis.com/",
  api_key = NULL,
  credentials = NULL,
  mime_type = NULL
) {
  credentials <- as_credentials(
    "google_upload",
    default_google_credentials(variant = "gemini"),
    credentials = credentials,
    api_key = api_key
  )

  mime_type <- mime_type %||% guess_mime_type(path)

  upload_url <- google_upload_init(
    path = path,
    base_url = base_url,
    credentials = credentials,
    mime_type = mime_type
  )

  status <- google_upload_send(
    upload_url = upload_url,
    path = path,
    credentials = credentials
  )
  google_upload_wait(status, credentials)

  ContentUploaded(uri = status$uri, mime_type = status$mimeType)
}

# https://ai.google.dev/api/files#method:-media.upload
google_upload_init <- function(path, base_url, credentials, mime_type) {
  file_size <- file.size(path)
  display_name <- basename(path)

  req <- request(base_url)
  req <- ellmer_req_credentials(req, credentials(), "x-goog-api-key")
  req <- req_url_path_append(req, "upload/v1beta/files")
  req <- req_headers(
    req,
    "X-Goog-Upload-Protocol" = "resumable",
    "X-Goog-Upload-Command" = "start",
    "X-Goog-Upload-Header-Content-Length" = toString(file_size),
    "X-Goog-Upload-Header-Content-Type" = mime_type,
  )
  req <- req_body_json(req, list(file = list(display_name = display_name)))

  resp <- req_perform(req)
  resp_header(resp, "x-goog-upload-url")
}

google_upload_send <- function(upload_url, path, credentials) {
  file_size <- file.size(path)

  req <- request(upload_url)
  req <- ellmer_req_credentials(req, credentials(), "x-goog-api-key")
  req <- req_headers(
    req,
    "Content-Length" = toString(file_size),
    "X-Goog-Upload-Offset" = "0",
    "X-Goog-Upload-Command" = "upload, finalize"
  )
  req <- req_body_file(req, path)
  req <- req_progress(req, "up")

  resp <- req_perform(req)
  resp_body_json(resp)$file
}

google_upload_status <- function(uri, credentials) {
  req <- request(uri)
  req <- ellmer_req_credentials(req, credentials(), "x-goog-api-key")

  resp <- req_perform(req)
  resp_body_json(resp)
}

google_upload_wait <- function(status, credentials) {
  cli::cli_progress_bar(
    format = "{cli::pb_spin} Processing [{cli::pb_elapsed}] "
  )

  while (status$state == "PROCESSING") {
    cli::cli_progress_update()
    status <- google_upload_status(status$uri, credentials)
    Sys.sleep(0.5)
  }
  if (status$state == "FAILED") {
    cli::cli_abort("Upload failed: {status$error$message}")
  }

  invisible()
}

# Helpers ----------------------------------------------------------------------

guess_mime_type <- function(file_path, call = caller_env()) {
  ext <- tolower(tools::file_ext(file_path))

  if (has_name(mime_types, ext)) {
    mime_types[[ext]]
  } else {
    cli::cli_abort(
      c(
        "x" = "Couldn't determine mime type for {.arg path} because it has an unknown file extension, {ext}.",
        "i" = "Please supply the {.arg mime_type} manually."
      )
    )
  }
}

mime_types <- list(
  # Images
  jpg = "image/jpeg",
  jpeg = "image/jpeg",
  png = "image/png",
  gif = "image/gif",
  bmp = "image/bmp",
  svg = "image/svg+xml",
  webp = "image/webp",
  tiff = "image/tiff",
  ico = "image/x-icon",

  # Documents
  pdf = "application/pdf",
  doc = "application/msword",
  docx = "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
  xls = "application/vnd.ms-excel",
  xlsx = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
  ppt = "application/vnd.ms-powerpoint",
  pptx = "application/vnd.openxmlformats-officedocument.presentationml.presentation",
  txt = "text/plain",
  rtf = "application/rtf",

  # Audio
  mp3 = "audio/mpeg",
  wav = "audio/wav",
  ogg = "audio/ogg",
  m4a = "audio/mp4",
  flac = "audio/flac",
  aac = "audio/aac",

  # Video
  mp4 = "video/mp4",
  avi = "video/x-msvideo",
  mkv = "video/x-matroska",
  mov = "video/quicktime",
  wmv = "video/x-ms-wmv",
  webm = "video/webm",

  # Web
  html = "text/html",
  htm = "text/html",
  css = "text/css",
  js = "application/javascript",
  json = "application/json",
  xml = "application/xml",

  # Data
  csv = "text/csv",
  tsv = "text/tab-separated-values",
  sql = "application/sql"
)

parse_ms <- function(ms) {
  stopifnot(is.numeric(ms))

  data.frame(
    days = floor(ms / 86400000),
    hours = floor((ms / 3600000) %% 24),
    minutes = floor((ms / 60000) %% 60),
    seconds = round((ms / 1000) %% 60, 1)
  )
}

first_positive <- function(x) which(x > 0)[1]

trim <- function(x) gsub("^\\s+|\\s+$", "", x)

pretty_ms <- function(ms, compact = FALSE) {
  stopifnot(is.numeric(ms))

  parsed <- t(parse_ms(ms))

  if (compact) {
    units <- c("d", "h", "m", "s")
    parsed2 <- parsed
    parsed2[] <- paste0(parsed, units)
    idx <- cbind(
      apply(parsed, 2, first_positive),
      seq_len(length(ms))
    )
    tmp <- paste0("~", parsed2[idx])

    # handle NAs
    tmp[is.na(parsed2[idx])] <- NA_character_
    tmp
  } else {
    ## Exact for small ones
    exact <- paste0(ceiling(ms), "ms")
    exact[is.na(ms)] <- NA_character_

    ## Approximate for others, in seconds
    merge_pieces <- function(pieces) {
      ## handle NAs
      if (all(is.na(pieces))) {
        return(NA_character_)
      }

      ## handle non-NAs
      paste0(
        if (pieces[1]) paste0(pieces[1], "d "),
        if (pieces[2]) paste0(pieces[2], "h "),
        if (pieces[3]) paste0(pieces[3], "m "),
        if (pieces[4]) paste0(pieces[4], "s ")
      )
    }
    approx <- trim(apply(parsed, 2, merge_pieces))

    ifelse(ms < 1000, exact, approx)
  }
}

pretty_sec <- function(sec, compact = FALSE) {
  pretty_ms(sec * 1000, compact = compact)
}

#' Google web search (grounding) tool
#'
#' @description
#' Enables Gemini models to search the web for up-to-date information and ground
#' responses with citations to sources. The model automatically decides when
#' (and how) to search the web based on your prompt. Search results are
#' incorporated into the response with grounding metadata including source
#' URLs and titles.
#'
#' Learn more in <https://ai.google.dev/gemini-api/docs/google-search>.
#'
#' @family built-in tools
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_google_gemini()
#' chat$register_tool(google_tool_web_search())
#' chat$chat("What was in the news today?")
#' chat$chat("What's the biggest news in the economy?")
#' }
google_tool_web_search <- function() {
  ToolBuiltIn(
    name = "web_search",
    description = "Search the web for up-to-date information.",
    annotations = tool_annotations(
      title = "Web search",
      read_only_hint = TRUE,
      open_world_hint = TRUE
    ),
    json = list(google_search = set_names(list()))
  )
}

#' Google URL fetch tool
#'
#' @description
#' When this tool is enabled, you can include URLs directly in your prompts and
#' Gemini will fetch and analyze the content.
#'
#' Learn more in <https://ai.google.dev/gemini-api/docs/url-context>.
#'
#' @family built-in tools
#' @export
#' @examples
#' \dontrun{
#' chat <- chat_google_gemini()
#' chat$register_tool(google_tool_web_fetch())
#' chat$chat("What are the latest package releases on https://tidyverse.org/blog?")
#' }
google_tool_web_fetch <- function() {
  ToolBuiltIn(
    name = "web_fetch",
    description = "Fetch and analyze content from a web URL.",
    annotations = tool_annotations(
      title = "Web fetch",
      read_only_hint = TRUE,
      open_world_hint = TRUE
    ),
    json = list(url_context = set_names(list()))
  )
}

# fmt: skip file

#' A function for foo-ing three numbers.
#'
#' @param x The first param
#' @param y The second param
#' @param z Take a guess
#' @returns The result of x %foo% y %foo% z.
has_roxygen_comments <- function(x, y, z = pi - 3.14) {
  super_secret_code()
}


  #' A function for foo-ing three numbers.
  #'
  #' @param x The first param
  #' @param y The second param
  #' @param z Take a guess
  #' @returns The result of x %foo% y %foo% z.
  indented_comments <- function(x, y, z = pi - 3.14) {
    super_secret_code()
  }

no_roxygen_comments <- function(i, j, k = pi - 3.14) {
  super_secret_code()
}

no_srcfile <- eval(quote(
  #' A function for foo-ing three numbers.
  function(a, b, c = pi - 3.14) {
    super_secret_code()
  }
))

extract_data <- function(turn, type, convert = TRUE, needs_wrapper = FALSE) {
  is_json <- map_lgl(turn@contents, S7_inherits, ContentJson)
  n <- sum(is_json)
  if (n == 0) {
    cli::cli_abort("Data extraction failed: no JSON responses found.")
  }

  if (n > 1) {
    cli::cli_warn("Found {n} JSON response{?s}, using the first.")
  }

  json <- turn@contents[[which(is_json)[1]]]
  out <- json@parsed # might error

  if (needs_wrapper) {
    out <- out$wrapper
    type <- type@properties[[1]]
  }
  if (convert) {
    out <- convert_from_type(out, type)
  }
  out
}

# OpenAI only works with objects, so we wrap and unwrap as needed
type_needs_wrapper <- function(type, provider) {
  S7_inherits(provider, ProviderOpenAICompatible) &&
    !(S7_inherits(type, TypeObject) || S7_inherits(type, TypeJsonSchema))
}
wrap_type_if_needed <- function(type, needs_wrapper = FALSE) {
  if (needs_wrapper) {
    type_object(wrapper = type)
  } else {
    type
  }
}

convert_from_type <- function(x, type) {
  if (is.null(x) && !type@required) {
    x
  } else if (S7_inherits(type, TypeArray)) {
    if (S7_inherits(type@items, TypeBasic)) {
      list_to_atomic(x, type@items@type)
    } else if (S7_inherits(type@items, TypeArray)) {
      lapply(x, function(y) convert_from_type(y, type@items))
    } else if (S7_inherits(type@items, TypeEnum)) {
      factor(as.character(x), levels = type@items@values)
    } else if (S7_inherits(type@items, TypeObject)) {
      if (type@items@additional_properties) {
        # don't convert to data frame, but put known properties first
        lapply(x, \(y) y[union(names(type@items@properties), names(y))])
      } else {
        cols <- lapply(names(type@items@properties), function(name) {
          vals <- lapply(x, function(y) y[[name]])
          convert_from_type(vals, type_array(type@items@properties[[name]]))
        })
        names(cols) <- names(type@items@properties)
        tibble::as_tibble(vctrs::new_data_frame(cols))
      }
    } else {
      x
    }
  } else if (S7_inherits(type, TypeObject)) {
    out <- lapply(names(type@properties), function(name) {
      convert_from_type(x[[name]], type@properties[[name]])
    })
    out <- set_names(out, names(type@properties))

    if (type@additional_properties) {
      extra_props <- setdiff(names(x), names(type@properties))
      out <- c(out, as.list(x[extra_props]))
    }
    out
  } else if (S7_inherits(type, TypeBasic)) {
    if (is.null(x)) {
      switch(
        type@type,
        boolean = NA,
        integer = NA_integer_,
        number = NA_real_,
        string = NA_character_,
        cli::cli_abort("Unknown type {type@type}", .internal = TRUE)
      )
    } else {
      x
    }
  } else if (S7_inherits(type, TypeEnum)) {
    if (is.null(x)) {
      NA_character_
    } else {
      as.character(x)
    }
  } else {
    x
  }
}

list_to_atomic <- function(x, type) {
  r_type <- switch(
    type,
    boolean = "logical",
    integer = "integer",
    number = "double",
    string = "character",
    cli::cli_abort("Unknown type {type}", .internal = TRUE)
  )

  if (length(x) == 0) {
    return(vector(r_type, 0))
  }

  types <- map_chr(x, typeof)
  if (r_type == "integer") {
    doubles <- which(types == "double")
    x[doubles] <- lapply(x[doubles], as.integer)
    types[doubles] <- "integer"
  } else if (r_type == "double") {
    integers <- which(types == "integer")
    x[integers] <- lapply(x[integers], as.double)
    types[integers] <- "double"
  }

  # silently replace incorrect types (incl NULLs) to NA
  wrong_type <- types != r_type
  x[wrong_type] <- list(`mode<-`(NA, r_type))

  # silently truncate incorrect lengths to 1
  x[lengths(x) != 1] <- lapply(x[lengths(x) != 1], \(x) x[1])

  unlist(x, use.names = FALSE, recursive = FALSE)
}

#' @include tools-def.R
#' @include utils-S7.R
NULL

#' Format contents into a textual representation
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' These generic functions can be use to convert [Turn] contents or [Content]
#' objects into textual representations.
#'
#' * `contents_text()` is the most minimal and only includes [ContentText]
#'   objects in the output.
#' * `contents_markdown()` returns the text content (which it assumes to be
#'   markdown and does not convert it) plus markdown representations of images
#'   and other content types.
#' * `contents_html()` returns the text content, converted from markdown to
#'   HTML with [commonmark::markdown_html()], plus HTML representations of
#'   images and other content types.
#'
#' These content types will continue to grow and change as ellmer evolves to
#' support more providers and as providers add more content types.
#' @examples
#' turns <- list(
#'   UserTurn(list(
#'     ContentText("What's this image?"),
#'     content_image_url("https://placehold.co/200x200")
#'   )),
#'   AssistantTurn("It's a placeholder image.")
#' )
#'
#' lapply(turns, contents_text)
#' lapply(turns, contents_markdown)
#' if (rlang::is_installed("commonmark")) {
#'   contents_html(turns[[1]])
#' }
#'
#' @param content The [Turn] or [Content] object to be converted into text.
#'   `contents_markdown()` also accepts [Chat] instances to turn the entire
#'   conversation history into markdown text.
#' @param ... Additional arguments passed to methods.
#'
#' @return A string of text, markdown or HTML.
#' @export
contents_text <- new_generic("contents_text", "content")

#' @rdname contents_text
#' @export
contents_html <- new_generic("contents_html", "content")

#' @rdname contents_text
#' @export
contents_markdown <- new_generic("contents_markdown", "content")


#' Content types received from and sent to a chatbot
#'
#' @description
#' Use these functions if you're writing a package that extends ellmer and need
#' to customise methods for various types of content. For normal use, see
#' [content_image_url()] and friends.
#'
#' ellmer abstracts away differences in the way that different [Provider]s
#' represent various types of content, allowing you to more easily write
#' code that works with any chatbot. This set of classes represents types of
#' content that can be either sent to and received from a provider:
#'
#' * `ContentText`: simple text (often in markdown format). This is the only
#'   type of content that can be streamed live as it's received.
#' * `ContentImageRemote` and `ContentImageInline`: images, either as a pointer
#'   to a remote URL or included inline in the object. See
#'   [content_image_file()] and friends for convenient ways to construct these
#'   objects.
#' * `ContentToolRequest`: a request to perform a tool call (sent by the
#'    assistant).
#' * `ContentToolResult`: the result of calling the tool (sent by the user).
#'   This object is automatically created from the value returned by calling the
#'   [tool()] function. Alternatively, expert users can return a
#'   `ContentToolResult` from a [tool()] function to include additional data or
#'   to customize the display of the result.
#'
#' @export
#' @return S7 objects that all inherit from `Content`
#' @examples
#' Content()
#' ContentText("Tell me a joke")
#' ContentImageRemote("https://www.r-project.org/Rlogo.png")
#' ContentToolRequest(id = "abc", name = "mean", arguments = list(x = 1:5))
Content <- new_class("Content")

method(contents_text, Content) <- function(content) {
  NULL
}

method(contents_markdown, Content) <- function(content) {
  # Fall back to text representation in markdown
  contents_text(content)
}

method(contents_html, Content) <- function(content) {
  NULL
}

#' @rdname Content
#' @export
#' @param text A single string.
ContentText <- new_class(
  "ContentText",
  parent = Content,
  properties = list(text = prop_string()),
)
method(format, ContentText) <- function(x, ...) {
  x@text
}

method(contents_text, ContentText) <- function(content) {
  content@text
}

method(contents_html, ContentText) <- function(content) {
  check_installed("commonmark")
  commonmark::markdown_html(content@text)
}

method(contents_markdown, ContentText) <- function(content) {
  content@text
}

# Images -----------------------------------------------------------------

#' @rdname Content
#' @export
ContentImage <- new_class(
  "ContentImage",
  parent = Content
)

#' @rdname Content
#' @export
#' @param url URL to a remote image.
#' @param detail Not currently used.
ContentImageRemote <- new_class(
  "ContentImageRemote",
  parent = ContentImage,
  properties = list(
    url = prop_string(),
    detail = prop_string(default = "")
  )
)
method(format, ContentImageRemote) <- function(x, ...) {
  cli::format_inline("[{.strong remote image}]: {.url {x@url}}")
}
method(contents_html, ContentImageRemote) <- function(content) {
  sprintf('<img src="%s">', content@url)
}
method(contents_markdown, ContentImageRemote) <- function(content) {
  sprintf('![](%s)', content@url)
}

#' @rdname Content
#' @export
#' @param type MIME type of the image.
#' @param data Base64 encoded image data.
ContentImageInline <- new_class(
  "ContentImageInline",
  parent = ContentImage,
  properties = list(
    type = prop_string(),
    data = prop_string(allow_null = TRUE)
  )
)
method(format, ContentImageInline) <- function(x, ...) {
  show_image(x)
  cli::format_inline("[{.strong inline image}]")
}
method(contents_html, ContentImageInline) <- function(content) {
  sprintf('<img src="data:%s;base64,%s">', content@type, content@data)
}
method(contents_markdown, ContentImageInline) <- function(content) {
  sprintf('![](data:%s;base64,%s)', content@type, content@data)
}

show_image <- function(x) {
  if (x@type != "image/png" || !is_installed("png")) {
    return(invisible())
  }

  data <- jsonlite::base64_dec(x@data)
  png <- png::readPNG(data)
  grid::grid.newpage()
  grid::grid.raster(png)
}

# Tools ------------------------------------------------------------------

#' @rdname Content
#' @export
#' @param id Tool call id (used to associate a request and a result).
#'   Automatically managed by \pkg{ellmer}.
#' @param name Function name
#' @param arguments Named list of arguments to call the function with.
#' @param tool ellmer automatically matches a tool request to the tools defined
#'   for the chatbot. If `NULL`, the request did not match a defined tool.
#' @param extra Provider-specific metadata associated with the tool request.
#'   Automatically managed by \pkg{ellmer}.
ContentToolRequest <- new_class(
  "ContentToolRequest",
  parent = Content,
  properties = list(
    id = prop_string(),
    name = prop_string(),
    arguments = class_list,
    tool = NULL | ToolDef,
    extra = class_list
  )
)
method(format, ContentToolRequest) <- function(
  x,
  ...,
  show = c("all", "call", "call_short")
) {
  show <- arg_match(show)

  arguments <- tool_request_args(x)
  if (is_tool_result(arguments)) {
    # Failed to convert the arguments so just use unconverted
    arguments <- x@arguments
  }
  call <- call2(x@name, !!!arguments)
  call_str <- deparse(call)
  if (show == "call") {
    return(call_str)
  }

  if (length(call_str) > 1) {
    call_str <- paste0(call_str[1], "...)")
  }

  if (show == "call_short") {
    return(call_str)
  }

  cli::format_inline("[{.strong tool request} ({x@id})]: {call_str}")
}

#' @rdname Content
#' @export
#' @param value The results of calling the tool function, if it succeeded.
#' @param error The error message, as a string, or the error condition thrown
#'   as a result of a failure when calling the tool function. Must be `NULL`
#'   when the tool call is successful.
#' @param extra Optional additional data associated with the tool result that
#'   isn't included in the `value` that's shown to the LLM. Useful for including
#'   additional data for displaying the tool result in a client, like a Shiny
#'   app, without including the data in the response to the LLM.
#' @param request The [ContentToolRequest] associated with the tool result,
#'   automatically added by \pkg{ellmer} when evaluating the tool call.
ContentToolResult <- new_class(
  "ContentToolResult",
  parent = Content,
  properties = list(
    value = class_any,
    error = new_property(
      class = NULL | class_character | new_S3_class("condition"),
      default = NULL,
      validator = function(value) {
        ok <- is.null(value) || is_string(value) || inherits(value, "condition")
        if (ok) {
          return()
        }

        paste0(
          "must be a single string or a condition object, not ",
          obj_type_friendly(value),
          "."
        )
      }
    ),
    extra = class_list,
    request = NULL | ContentToolRequest
  )
)
method(format, ContentToolResult) <- function(
  x,
  ...,
  show = c("all", "header")
) {
  show <- arg_match(show)

  header <- cli::format_inline("[{.strong tool result}  ({x@request@id})]:")

  if (show == "header") {
    return(header)
  }

  if (tool_errored(x)) {
    value <- paste0(cli::col_red("Error: "), tool_error_string(x))
  } else {
    value <- tool_string_preview(x)
  }

  if (!is_string(value) || !grepl("\n", value)) {
    paste0(header, " ", value)
  } else {
    paste(c(header, value), collapse = "\n")
  }
}

tool_errored <- function(x) !is.null(x@error)
tool_error_string <- function(x) {
  if (inherits(x@error, "condition")) conditionMessage(x@error) else x@error
}
tool_string <- function(x) {
  if (tool_errored(x)) {
    paste0("Tool calling failed with error ", tool_error_string(x))
  } else if (inherits(x@value, "AsIs")) {
    x@value
  } else if (inherits(x@value, "json")) {
    x@value
  } else if (is.character(x@value)) {
    paste(x@value, collapse = "\n")
  } else {
    withCallingHandlers(
      to_json(x@value),
      error = function(err) {
        cli::cli_abort(
          c(
            "Could not convert tool result from {.obj_type_friendly {x@value}} to JSON.",
            "i" = "If you are the tool author, update the tool to convert the result to a string or JSON."
          ),
          parent = err
        )
      }
    )
  }
}

tool_string_preview <- function(x) {
  tryCatch(
    tool_string(x),
    error = function(err) {
      cli::col_silver("<tool result preview not available>")
    }
  )
}

ContentJson <- new_class(
  "ContentJson",
  parent = Content,
  properties = list(
    # Some providers guarantee JSON returning it as part of the parsed results.
    # Others return as a string, which may or may not be valid JSON. We don't
    # want to force conversion on construction() since that's likely to be
    # inconveniently early
    data = class_any,
    string = prop_string(allow_null = TRUE),

    # Virtual property that always returns parsed JSON or errors
    parsed = new_property(getter = function(self) {
      if (is.null(self@string)) {
        self@data
      } else {
        jsonlite::parse_json(self@string)
      }
    })
  )
)

method(format, ContentJson) <- function(x, ...) {
  paste0(
    cli::format_inline("[{.strong data}] "),
    pretty_json(x@parsed)
  )
}
method(contents_html, ContentJson) <- function(content) {
  sprintf('<pre><code>%s</code></pre>\n', pretty_json(content@parsed))
}
method(contents_markdown, ContentJson) <- function(content) {
  sprintf('```json\n%s\n```\n', pretty_json(content@parsed))
}

ContentUploaded <- new_class(
  "ContentUploaded",
  parent = Content,
  properties = list(
    uri = prop_string(),
    mime_type = prop_string(default = "")
  )
)
method(format, ContentUploaded) <- function(x, ...) {
  cli::format_inline("[{.strong uploaded file}]: [{x@mime_type}]")
}
method(contents_html, ContentUploaded) <- function(content) {
  sprintf("[uploaded %s file]", content@mime_type)
}
method(contents_markdown, ContentUploaded) <- function(content) {
  sprintf("`[uploaded %s file]`", content@mime_type)
}

# Thinking ---------------------------------------------------------------------

#' @rdname Content
#' @param thinking The text of the thinking output.
#' @param extra Additional data.
#' @export
ContentThinking <- new_class(
  "ContentThinking",
  parent = Content,
  properties = list(
    thinking = prop_string(),
    extra = class_list
  )
)

method(format, ContentThinking) <- function(x, ...) {
  paste0("<thinking>\n", x@thinking, "\n</thinking>\n")
}

method(contents_html, ContentThinking) <- function(content) {
  check_installed("commonmark")
  paste0(
    "<details><summary>Thinking</summary>\n",
    commonmark::markdown_html(content@thinking),
    "</details>\n"
  )
}

method(contents_markdown, ContentThinking) <- function(content) {
  format(content)
}

# Helpers ----------------------------------------------------------------------

as_content <- function(x, error_call = caller_env(), error_arg = "...") {
  if (is.null(x)) {
    list()
  } else if (is_prompt(x)) {
    if (length(x) == 1) {
      ContentText(x[[1]])
    } else {
      cli::cli_abort(
        "{.arg {error_arg}} can only accept a single prompt.",
        call = error_call
      )
    }
  } else if (is.character(x)) {
    ContentText(paste0(x, collapse = "\n\n"))
  } else if (S7_inherits(x, Content)) {
    x
  } else {
    stop_input_type(
      x,
      what = "made up strings or <content> objects",
      arg = error_arg,
      error_call = error_call
    )
  }
}

#' @rdname Content
#' @param filename File name, used to identify the PDF.
#' @export
ContentPDF <- new_class(
  "ContentPDF",
  parent = Content,
  properties = list(
    type = prop_string(),
    data = prop_string(),
    filename = prop_string()
  )
)

method(format, ContentPDF) <- function(x, ...) {
  "<PDF document>"
}

#' @include provider-openai-compatible.R
NULL

#' Chat with one of the many models hosted on OpenRouter
#'
#' @description
#' Sign up at <https://openrouter.ai>.
#'
#' Support for features depends on the underlying model that you use; see
#' <https://openrouter.ai/models> for details.
#'
#' @export
#' @family chatbots
#' @param api_key `r lifecycle::badge("deprecated")` Use `credentials` instead.
#' @param credentials `r api_key_param("OPENROUTER_API_KEY")`
#' @param model `r param_model("gpt-4o")`
#' @param params Common model parameters, usually created by [params()].
#' @inheritParams chat_openai
#' @inherit chat_openai return
#' @examples
#' \dontrun{
#' chat <- chat_openrouter()
#' chat$chat("Tell me three jokes about statisticians")
#' }
chat_openrouter <- function(
  system_prompt = NULL,
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  echo = c("none", "output", "all"),
  api_headers = character()
) {
  model <- set_default(model, "gpt-4o")
  echo <- check_echo(echo)

  credentials <- as_credentials(
    "chat_openrouter",
    function() openrouter_key(),
    credentials = credentials,
    api_key = api_key
  )

  params <- params %||% params()

  provider <- ProviderOpenRouter(
    name = "OpenRouter",
    base_url = "https://openrouter.ai/api/v1",
    model = model,
    params = params,
    extra_args = api_args,
    credentials = credentials,
    extra_headers = api_headers,
    preserve_thinking = TRUE
  )
  Chat$new(provider = provider, system_prompt = system_prompt, echo = echo)
}

chat_openrouter_test <- function(..., echo = "none") {
  chat_openrouter(..., model = "openai/gpt-4o-mini-2024-07-18", echo = echo)
}

ProviderOpenRouter <- new_class(
  "ProviderOpenRouter",
  parent = ProviderOpenAICompatible,
)

method(chat_params, ProviderOpenRouter) <- function(provider, params) {
  # https://openrouter.ai/docs/api-reference/parameters
  standardise_params(
    params,
    c(
      temperature = "temperature",
      top_p = "top_p",
      top_k = "top_k",
      frequency_penalty = "frequency_penalty",
      presence_penalty = "presence_penalty",
      seed = "seed",
      max_tokens = "max_tokens",
      logprobs = "log_probs",
      stop = "stop_sequences"
    )
  )
}

openrouter_key <- function() {
  key_get("OPENROUTER_API_KEY")
}

method(base_request, ProviderOpenRouter) <- function(provider) {
  req <- base_request(super(provider, ProviderOpenAICompatible))
  # https://openrouter.ai/docs/api-keys
  req <- req_headers(
    req,
    `HTTP-Referer` = "https://ellmer.tidyverse.org",
    `X-Title` = "ellmer"
  )

  req
}

method(value_turn, ProviderOpenRouter) <- function(
  provider,
  result,
  has_type = FALSE
) {
  # https://openrouter.ai/docs/errors
  check_openrouter_error(result$error)

  value_turn(
    super(provider, ProviderOpenAICompatible),
    result = result,
    has_type = has_type
  )
}

method(stream_parse, ProviderOpenRouter) <- function(provider, event) {
  if (is.null(event) || identical(event$data, "[DONE]")) {
    return(NULL)
  }

  result <- jsonlite::parse_json(event$data)
  check_openrouter_error(result$error)
  result
}

check_openrouter_error <- function(error, call = caller_env()) {
  if (is.null(error)) {
    return()
  }
  message <- error$message
  if (is.null(error$metadata$raw$data)) {
    details <- NULL
  } else {
    details <- prettify(error$metadata$raw$data)
    # don't line wrap
    details <- gsub(" ", "\u00a0", details, fixed = TRUE)
  }

  abort(
    c("message", i = if (!is.null(details)) details),
    call = call
  )
}

method(chat_resp_stream, ProviderOpenRouter) <- function(provider, resp) {
  repeat {
    event <- resp_stream_sse(resp)
    if (is.null(event)) {
      break
    }

    # https://openrouter.ai/docs/responses#sse-streaming-comments
    if (!identical(event$data, character())) {
      break
    }
    Sys.sleep(0.1)
  }

  event
}

method(as_json, list(ProviderOpenRouter, ContentText)) <- function(
  provider,
  x,
  ...
) {
  if (identical(x@text, "")) {
    # Tool call requests can include a Content with empty text,
    # but it doesn't like it if you send this back
    NULL
  } else {
    list(type = "text", text = x@text)
  }
}

test_that("ContentJson converted to ContentText", {
  test_provider <- ProviderOpenAICompatible("test", "model", "base_url")
  expect_equal(
    as_json(test_provider, ContentJson(list(x = 1))),
    list(type = "text", text = "{\"x\":1}")
  )
})

You have a single purpose: take the documentation for an R function and turn it into a tool call definition. A tool call definition consists of a call to the `tool()` function with the following arguments:

* The first argument, which should be unnamed, is the function.
* The second argument, `name`, is the name of the function as a string.
* The third argument, `description`, is a brief description of the function, in the same language as the provided documentation.
* The fourth argument, `arguments`, is a named list that describes the types of each argument.
  It should have one element for each argument to the function. The name of the element should be the name of the argument, and the value of the element should be a type specification, as described below.

## Type specification

There are four basic types that represent scalars: `type_string()`, `type_number()`, `type_integer()` and `type_boolean()`. These can be combined with `type_array()` to represent vectors, e.g. `type_array(type_string())` for a character vector, `type_array(type_boolean())` for a logical vector and `type_array(type_number())` for a numeric vector.

The first argument to each `type_` function is the `description`. It should include a 1-2 sentence description of the argument, in the same language as the provided documentation.

Any arguments that don't use one of these basic class should be given type `NULL`, indicating that it can't be easily used by the LLM.

Types default to `required = TRUE`, but if the function argument has a default value, you can set `required = FALSE` to indicate that the argument is optional. Be sure to describe the default value in the `description` field.

## Example

Here's an example for a simple function:

<user>
<name>stats::median</name>
<documentation>
median                  package:stats                  R Documentation

Median Value

Description:

     Compute the sample median.

Usage:

     median(x, na.rm = FALSE, ...)
     ## Default S3 method:
     median(x, na.rm = FALSE, ...)

Arguments:

       x: an object for which a method has been defined, or a numeric
          vector containing the values whose median is to be computed.

   na.rm: a logical value indicating whether ‘NA’ values should be
          stripped before the computation proceeds.

     ...: potentially further arguments for methods; not used in the
          default method.

Details:

     This is a generic function for which methods can be written.
     However, the default method makes use of ‘is.na’, ‘sort’ and
     ‘mean’ from package ‘base’ all of which are generic, and so the
     default method will work for most classes (e.g., ‘"Date"’) for
     which a median is a reasonable concept.

Value:

     The default method returns a length-one object of the same type as
     ‘x’, except when ‘x’ is logical or integer of even length, when
     the result will be double.

     If there are no values or if ‘na.rm = FALSE’ and there are ‘NA’
     values the result is ‘NA’ of the same type as ‘x’ (or more
     generally the result of ‘x[NA_integer_]’).

References:

     Becker, R. A., Chambers, J. M. and Wilks, A. R. (1988) _The New S
     Language_.  Wadsworth & Brooks/Cole.

See Also:

     ‘quantile’ for general quantiles.

Examples:

     median(1:4)                # = 2.5 [even number]
     median(c(1:3, 100, 1000))  # = 3 [odd, robust]
</documentation>
</user>

<assistant>
tool(
  stats::median,
  name = "median",
  description = "Compute the median value",
  arguments = list(
     x = type_array("Input vector", items = type_number()),
     na.rm = type_boolean(
       "Should missing values be removed? Defaults to FALSE",
       required = FALSE
     ),
     ... = NULL
  ),
)
</assistant>

---
title: "Tool/function calling"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tool/function calling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = ellmer:::eval_vignette()
)
vcr::setup_knitr()
```

## Introduction

One of the most interesting aspects of modern chat models is their ability to make use of external tools that are defined by the caller.

When making a chat request to the chat model, the caller advertises one or more tools (defined by their function name, description, and a list of expected arguments), and the chat model can choose to respond with one or more "tool calls". These tool calls are requests *from the chat model to the caller* to execute the function with the given arguments; the caller is expected to execute the functions and "return" the results by submitting another chat request with the conversation so far, plus the results. The chat model can then use those results in formulating its response, or, it may decide to make additional tool calls.

*Note that the chat model does not directly execute any external tools!* It only makes requests for the caller to execute them. It's easy to think that tool calling might work like this:

![Diagram showing the wrong mental model of tool calls: a user initiates a request that flows to the assistant, which then runs the code, and returns the result back to the user."](tool-calling-wrong.svg)

But in fact it works like this:

![Diagram showing the correct mental model for tool calls: a user sends a request that needs a tool call, the assistant requests that the user runs that tool, returns the result to the assistant, which uses it to generate the final answer.](tool-calling-right.svg)

The value that the chat model brings is not in helping with execution, but with knowing when it makes sense to call a tool, what values to pass as arguments, and how to use the results in formulating its response.

```{r setup}
library(ellmer)
```


### Motivating example

Let's take a look at an example where we really need an external tool. Chat models generally do not know the current time, which makes questions like these impossible.

```{r}
#| label: no-tool
#| cassette: true

chat <- chat_openai(model = "gpt-4o")
chat$chat("How long ago did Neil Armstrong touch down on the moon?")
```

Since the model doesn't know what day it is, the result is incorrect.

### Defining a tool function

The first thing we'll do is define an R function that returns the current time.

```{r}
#' Gets the current time in the given time zone.
#'
#' @param tz The time zone to get the current time in.
#' @return The current time in the given time zone.
get_current_time <- function(tz = "UTC") {
  format(Sys.time(), tz = tz, usetz = TRUE)
}
```

Note that we've gone through the trouble of creating [roxygen2 comments](https://roxygen2.r-lib.org/). This isn't necessary, but as we'll see shortly, can make it a bit easier to generate a tool defintion.

```{r}
#| include: false

# Fake it for the vignette so we always get the same results
get_current_time <- function(tz = "UTC") {
  format(
    as.POSIXct("2025-06-25 11:53:23", tz = "America/Chicago"),
    tz = tz,
    usetz = TRUE
  )
}
```

To turn a function into a tool, we provide some additional metadata that the model will use:

```{r}
get_current_time <- tool(
  get_current_time,
  name = "get_current_time",
  description = "Returns the current time.",
  arguments = list(
    tz = type_string(
      "Time zone to display the current time in. Defaults to `\"UTC\"`.",
      required = FALSE
    )
  )
)
```

This is a fair amount of code to write, even for such a simple function. Fortunately, you don't have to write this by hand! I generated the above `tool()` call by calling `create_tool_def(get_current_time)`, which uses an LLM to generate the `tool()` call for you. `create_tool_def()` is not perfect, so you must review the generated code before using it, but it is a big time-saver.

Note that a tool is just a special type of function so we can still call it:

```{r}
get_current_time()
```

### Registering and using tools

Now we need to give our chat object access to our tool. We do this with `$register_tool()`:

```{r}
chat$register_tool(get_current_time)
```

That's all we need to do! Let's retry our query:

```{r}
#| label: with-tool
#| cassette: true

chat$chat("How long ago did Neil Armstrong touch down on the moon?")
```

That's correct! Without any further guidance, the chat model decided to call our tool function and successfully used its result in formulating its response.

If we print the chat we can see where the model decided to use the tool:

```{r}
chat
```

(Full disclosure: I originally tried this example with the default model of `gpt-4o-mini` and it got the tool calling right but the date math wrong, hence the explicit `model="gpt-4o"`.)

This tool example was extremely simple, but you can imagine doing much more interesting things from tool functions: calling APIs, reading from or writing to a database, kicking off a complex simulation, or even calling a complementary GenAI model (like an image generator). Or if you are using ellmer in a Shiny app, you could use tools to set reactive values, setting off a chain of reactive updates.

### Tool inputs and outputs

Remember that tool arguments come from the LLM, and tool results are returned to the LLM. This implies that you should keep both as simple as possible.

Inputs to a tool call, must be defined by `type_boolean()`, `type_integer()`, `type_number()`, `type_string()`, `type_enum()`, `type_array()`, or `type_object()`. We recommend keeping them as simple as possible, focusing on basic scalar types as much as you can.

The output of the tool call will be interpreted by the LLM, just as if you had typed that information into the data. That means you'll generally want to produce text or other atomic vectors. For more complex data, ellmer will automatically serialize the result to JSON, which LLMs generally seem to be good at understanding.  If you must have more direct control of the structure of the JSON that's returned, you can return a JSON-serializable value wrapped in `I()`, which ellmer will leave alone until the entire request is JSON-serialized.

To show off these ideas, here's a slightly more complicated example simulating a weather API that returns data for multiple cities at once. The `get_weather()` function returns a data frame that ellmer will automatically convert into JSON in row-major format, which our experiments suggest is good for LLMs.

```{r}
get_weather <- tool(
  function(cities) {
    raining <- c(London = "heavy", Houston = "none", Chicago = "overcast")
    temperature <- c(London = "cool", Houston = "hot", Chicago = "warm")
    wind <- c(London = "strong", Houston = "weak", Chicago = "strong")

    data.frame(
      city = cities,
      raining = unname(raining[cities]),
      temperature = unname(temperature[cities]),
      wind = unname(wind[cities])
    )
  },
  name = "get_weather",
  description = "
    Report on weather conditions in multiple cities. For efficiency, request
    all weather updates using a single tool call
  ",
  arguments = list(
    cities = type_array(type_string(), "City names")
  )
)
```

Now we register and use it:

```{r}
#| label: inputs-outputs
#| cassette: true

chat <- chat_openai()
chat$register_tool(get_weather)
chat$chat("Give me a weather update for London and Chicago")
```

We can print the chat to confirm that the model only performed a single tool call:

```{r}
chat
```

### Image and PDF tool output

ellmer allow tools to return image or PDF content that can be returned with the tool result, if the LLM or API supports vision capabilities.

Simply return a `content_image_file()`, `content_pdf_file()`, or similar content type from the tool function.
For example, here's a simple tool to screenshot a website:

```{r}
#| eval: false

screenshot_website <- tool(
  function(url) {
    tmpf <- withr::local_tempfile(fileext = ".png")
    webshot2::webshot(url, file = tmpf)
    content_image_file(tmpf)
  },
  name = "screenshot_website",
  description = "Take a screenshot of a website.",
  arguments = list(
    url = type_string("The URL of the website")
  )
)
```

You could use this tool to allow the LLM to "see" websites, like [the tidyverse website](https://tidyverse.org):

```{r}
#| eval: false

chat <- chat_openai()
#> Using model = "gpt-4.1".
chat$register_tool(screenshot_website)
chat$chat("Describe the design aesthetic of https://tidyverse.org")
#> https://tidyverse.org screenshot completed
#> The design aesthetic of the Tidyverse website (https://tidyverse.org) is
#> clean, modern, and minimalistic, with several distinct features:
#>
#> - **Color Palette**: The overall site uses a lot of white space with navy
#>   and dark backgrounds for some elements, accentuated by the colorful
#>   hexagonal logos for various R packages.
#> - **Typography**: Simple, sans-serif fonts contribute to readability and
#>   a contemporary look.
#> - **Hexagonal Icons**: Prominent display of tidyverse package logos in
#>   hexagonal shapes, emphasizing the modular, package-oriented
#>   nature of the Tidyverse.
#> - **Layout**: A balanced, spacious two-column layout. The left side
#>   features graphic elements; the right side provides concise, text-based
#>   information.
#>
#> Overall, the design communicates clarity, ease of use, and a focus on
#> modern data science tools.
```

---
title: "Getting started with ellmer"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with ellmer}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = ellmer:::eval_vignette()
)
vcr::setup_knitr()
```

```{r setup}
library(ellmer)
```

ellmer makes it easy to access the wealth of large language models (LLMs) from R. But what can you do with those models once you have access to them? This vignette will give you the basic vocabulary you need to use an LLM effectively and will show you some examples to ignite your creativity.

In this vignette we'll mostly ignore how LLMs work, using them as convenient black boxes. If you want to get a sense of how they actually work, we recommend watching Jeremy Howard's posit::conf(2023) keynote: [A hacker's guide to open source LLMs](https://www.youtube.com/watch?v=sYliwvml9Es).

## Vocabulary

We'll start by laying out the key vocab that you'll need to understand LLMs. Unfortunately the vocab is all a little entangled: to understand one term you'll often have to know a little about some of the others. So we'll start with some simple definitions of the most important terms then iteratively go a little deeper.

It all starts with a **prompt**, which is the text (typically a question or a request) that you send to the LLM. This starts a **conversation**, a sequence of turns that alternate between user prompts and model responses. Inside the model, both the prompt and response are represented by a sequence of **tokens**, which represent either individual words or subcomponents of a word. The tokens are used to compute the cost of using a model and to measure the size of the **context**, the combination of the current prompt and any previous prompts and responses used to generate the next response.

It's useful to make the distinction between providers and models. A **provider** is a web API that gives access to one or more **models**. The distinction is a bit subtle because providers are often synonymous with a model, like OpenAI and GPT, Anthropic and Claude, and Google and Gemini. But other providers, like Ollama, can host many different models, typically open source models like LLaMa and Mistral. Still other providers support both open and closed models, typically by partnering with a company that provides a popular closed model. For example, Azure OpenAI offers both open source models and OpenAI's GPT, while AWS Bedrock offers both open source models and Anthropic's Claude.

### What is a token?

An LLM is a _model_, and like all models needs some way to represent its inputs numerically. For LLMs, that means we need some way to convert words to numbers. This is the goal of the **tokenizer**. For example, using the GPT 4o tokenizer, the string "When was R created?" is converted to 5 tokens: 5958 ("When"), 673 (" was"), 460 (" R"), 5371 (" created"), 30 ("?"). As you can see, many simple strings can be represented by a single token. But more complex strings require multiple tokens. For example, the string "counterrevolutionary" requires 4 tokens: 32128 ("counter"), 264 ("re"), 9477 ("volution"), 815 ("ary"). (You can see how various strings are tokenized at <http://tiktokenizer.vercel.app/>).

It's important to have a rough sense of how text is converted to tokens because tokens are used to determine the cost of a model and how much context can be used to predict the next response. On average an English word needs ~1.5 tokens so a page might require 375-400 tokens and a complete book might require 75,000 to 150,000 tokens. Other languages will typically require more tokens, because (in brief) LLMs are trained on data from the internet, which is primarily in English.

LLMs are priced per million tokens. State of the art models (like GPT-4.1 or Claude 3.5 sonnet) cost $2-3 per million input tokens, and $10-15 per million output tokens. Cheaper models can cost much less, e.g. GPT-4.1 nano costs $0.10 per million input tokens and $0.40 per million output tokens. Even $10 of API credit will give you a lot of room for experimentation, particularly with cheaper models, and prices are likely to decline as model performance improves.

Tokens also used to measure the context window, which is how much text the LLM can use to generate the next response. As we'll discuss shortly, the context length includes the full state of your conversation so far (both your prompts and the model's responses), which means that cost grow rapidly with the number of conversational turns.

In ellmer, you can see how many tokens a conversations has used by printing it, and you can see total usage for a session with `token_usage()`.

```{r}
#| cassette: token-usage
chat <- chat_openai(model = "gpt-4.1")
. <- chat$chat("Who created R?", echo = FALSE)
chat

token_usage()
```

If you want to learn more about tokens and tokenizers, I'd recommend watching the first 20-30 minutes of [Let's build the GPT Tokenizer](https://www.youtube.com/watch?v=zduSFxRajkE) by Andrej Karpathy. You certainly don't need to learn how to build your own tokenizer, but the intro will give you a bunch of useful background knowledge that will help improve your undersstanding of how LLM's work.

### What is a conversation?

A conversation with an LLM takes place through a series of HTTP requests and responses: you send your question to the LLM as an HTTP request, and it sends back its reply as an HTTP response. In other words, a conversation consists of a sequence of a paired turns: a sent prompt and a returned response.

It's important to note that a request includes not only the current user prompt, but every previous user prompt and model response. This means that:

* The cost of a conversation grows quadratically with the number of turns: if you want to save money, keep your conversations short.

* Each response is affected by all previous prompts and responses. This can make a conversation get stuck in a local optimum, so it's generally better to iterate by starting a new conversation with a better prompt rather than having a long back-and-forth.

* ellmer has full control over the conversational history. Because it's ellmer's responsibility to send the previous turns of the conversation, it's possible to start a conversation with one model and finish it with another.

### What is a prompt?

The user prompt is the question that you send to the model. There are two other important prompts that underlie the user prompt:

* The **platform prompt**, which is unchangeable, set by the model provider, and affects every conversation. You can see what these look like from Anthropic, who [publishes their core system prompts](https://platform.claude.com/docs/en/release-notes/system-prompts).

* The **system prompt** (aka developer prompt), which is set when you create a new conversation, and affects every response. It's used to provide additional instructions to the model, shaping its responses to your needs. For example, you might use the system prompt to ask the model to always respond in Spanish or to write dependency-free base R code. You can also use the system prompt to provide the model with information it wouldn't otherwise know, like the details of your database schema, or your preferred ggplot2 theme and color palette.

OpenAI calls this the [chain of command](https://cdn.openai.com/spec/model-spec-2024-05-08.html#follow-the-chain-of-command): if there are conflicts or inconsistencies in the prompts, the platform prompt overrides the system prompt, which in turn overrides the user prompt.

When you use a chat app like ChatGPT or Claude.ai you can only iterate on the user prompt. But when you're programming with LLMs, you'll primarily iterate on the system prompt. For example, if you're developing an app that helps a user write tidyverse code, you'd work with the system prompt to ensure that user gets the style of code they want.

Writing a good prompt, which is called __prompt design__, is key to effective use of LLMs. It is discussed in more detail in `vignette("prompt-design")`.

## Example uses

Now that you've got the basic vocab under your belt, I'm going to fire a bunch of interesting potential use cases at you. While there are special purpose tools that might solve these cases faster and/or cheaper, an LLM allows you to rapidly prototype a solution. This can be extremely valuable even if you end up using those more specialised tools in your final product.

In general, we recommend avoiding LLMs where accuracy is critical. That said, there are still many cases for their use. For example, even though they always require some manual fiddling, you might save a bunch of time ever with an 80% correct solution. In fact, even a not-so-good solution can still be useful because it makes it easier to get started: it's easier to react to something rather than to have to start from scratch with a blank page.

### Chatbots

A great place to start with ellmer and LLMs is to build a chatbot with a custom prompt. Chatbots are a familiar interface to LLMs and are easy to create in R with [shinychat](https://github.com/jcheng5/shinychat). And there's a surprising amount of value to creating a custom chatbot that has a prompt stuffed with useful knowledge. For example:

* Help people use your new package. To do so, you need a custom prompt because LLMs were trained on data prior to your package's existence. You can create a surprisingly useful tool just by preloading the prompt with your README and vignettes. This is how the [ellmer assistant](https://github.com/jcheng5/ellmer-assistant) works.

* Build language specific prompts for R and/or Python. [Shiny Assistant](https://shiny.posit.co/blog/posts/shiny-assistant/) helps you build shiny apps (either in R or Python) by combining a [prompt](https://github.com/posit-dev/shiny-assistant/blob/main/shinyapp/app_prompt.md) that gives general advice on building apps with a prompt for [R](https://github.com/posit-dev/shiny-assistant/blob/main/shinyapp/app_prompt_r.md) or [python](https://github.com/posit-dev/shiny-assistant/blob/main/shinyapp/app_prompt_python.md). The Python prompt is very detailed because there's much less information about Shiny for Python in the existing LLM knowledgebases.

* Help people find the answers to their questions. Even if you've written a bunch of documentation for something, you might find that you still get questions because folks can't easily find exactly what they're looking for. You can reduce the need to answer these questions by creating a chatbot with a prompt that contains your documentation. For example, if you're a teacher, you could create a chatbot that includes your syllabus in the prompt. This eliminates a common class of question where the data necessary to answer the question is available, but hard to find.

Another direction is to give the chatbot additional context about your current environment. For example, [aidea](https://github.com/cpsievert/aidea) allows the user to interactively explore a dataset with the help of the LLM. It adds summary statistics about the dataset to the [prompt](https://github.com/cpsievert/aidea/blob/main/inst/app/prompt.md) so that the LLM knows something about your data. Along these lines, imagine writing a chatbot to help with data import that has a prompt which include all the files in the current directory along with their first few lines.

### Structured data extraction

LLMs are often very good at extracting structured data from unstructured text. This can give you traction to analyse data that was previously unaccessible. For example:

* Customer tickets and GitHub issues: you can use LLMs for quick and dirty sentiment analysis by extracting any specifically mentioned products and summarising the discussion as a few bullet points.

* Geocoding: LLMs do a surprisingly good job at geocoding, especially extracting addresses or finding the latitute/longitude of cities. There are specialised tools that do this better, but using an LLM makes it easy to get started.

* Recipes: I've extracted structured data from baking and cocktail recipes. Once you have the data in a structured form you can use your R skills to better understand how recipes vary within a cookbook or to look for recipes that use the ingredients currently in your kitchen. You could even use shiny assistant to help make those techniques available to anyone, not just R users.

Structured data extraction also works well with images. It's not the fastest or cheapest way to extract data but it makes it really easy to prototype ideas. For example, maybe you have a bunch of scanned documents that you want to index. You can convert PDFs to images (e.g. using {imagemagick}) then use structured data extraction to pull out key details.

Learn more about structured data extraction in `vignette("structured-data")`.

### Programming

LLMs can also be useful to solve general programming problems. For example:

* Write a detailed prompt that explains how to update code to use a new version of a package. You could combine this with the rstudioapi package to allow the user to select code, transform it, and replace it in the existing text. A comprehensive example of this sort of app is [chores](https://simonpcouch.github.io/chores/), which includes prompts for automatically generating roxygen documentation blocks, updating testthat code to the 3rd edition, and converting `stop()` and `abort()` to use `cli::cli_abort()`.

* You could automatically look up the documentation for an R function, and include it in the prompt to make it easier to figure out how to use a specific function.

* You can use LLMs to explain code, or even ask them to [generate a diagram](https://bsky.app/profile/daviddiviny.com/post/3lb6kjaen4c2u).

* You can ask an LLM to analyse your code for potential code smells or security issues. You can do this a function at a time, or explore the entire source code of your package or script in the prompt.

* You could use [gh](https://gh.r-lib.org) to find unlabelled issues, extract the text, and ask the LLM to figure out what labels might be most appropriate. Or maybe an LLM might be able to help people create better reprexes, or simplify reprexes that are too complicated?

* I find it useful to have an LLM document a function for me, even knowing that it's likely to be mostly incorrect. Having something to react to make it much easier for me to get started.

* If you're working with code or data from another programming language, you can ask an LLM to convert it to R code for you. Even if it's not perfect, it's still typically much faster than doing everything yourself.

## Miscellaneous

To finish up here are a few other ideas that seem cool but didn't seem to fit the above categories:

* Automatically generate alt text for plots, using `content_image_plot()`.

* Analyse the text of your statistical report to look for flaws in your statistical reasoning (e.g. misinterpreting p-values or assuming causation where only correlation exists).

* Use your existing company style guide to generate a [brand.yaml](https://posit-dev.github.io/brand-yml/articles/llm-brand-yml-prompt/) specification to automatically style your reports, apps, dashboards and plots to match your corporate style guide.

---
title: "Streaming and async APIs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Streaming and async APIs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


### Streaming results

The `chat()` method does not return any results until the entire response is received. (It can _print_ the streaming results to the console but it _returns_ the result only when the response is complete.)

If you want to process the response as it arrives, you can use the `stream()` method. This is useful when you want to send the response, in realtime, somewhere other than the R console (e.g., to a file, an HTTP response, or a Shiny chat window), or when you want to manipulate the response before displaying it without giving up the immediacy of streaming.

With the `stream()` method, which returns a [coro](https://coro.r-lib.org/) [generator](https://coro.r-lib.org/articles/generator.html), you can process the response by looping over it as it arrives.

```{r eval=FALSE}
stream <- chat$stream("What are some common uses of R?")
coro::loop(for (chunk in stream) {
  cat(toupper(chunk))
})
#>  R IS COMMONLY USED FOR:
#>
#>  1. **STATISTICAL ANALYSIS**: PERFORMING COMPLEX STATISTICAL TESTS AND ANALYSES.
#>  2. **DATA VISUALIZATION**: CREATING GRAPHS, CHARTS, AND PLOTS USING PACKAGES LIKE  GGPLOT2.
#>  3. **DATA MANIPULATION**: CLEANING AND TRANSFORMING DATA WITH PACKAGES LIKE DPLYR AND TIDYR.
#>  4. **MACHINE LEARNING**: BUILDING PREDICTIVE MODELS WITH LIBRARIES LIKE CARET AND #>  RANDOMFOREST.
#>  5. **BIOINFORMATICS**: ANALYZING BIOLOGICAL DATA AND GENOMIC STUDIES.
#>  6. **ECONOMETRICS**: PERFORMING ECONOMIC DATA ANALYSIS AND MODELING.
#>  7. **REPORTING**: GENERATING DYNAMIC REPORTS AND DASHBOARDS WITH R MARKDOWN.
#>  8. **TIME SERIES ANALYSIS**: ANALYZING TEMPORAL DATA AND FORECASTING.
#>
#>  THESE USES MAKE R A POWERFUL TOOL FOR DATA SCIENTISTS, STATISTICIANS, AND RESEARCHERS.
```

## Async usage

ellmer also supports async usage. This is useful when you want to run multiple, concurrent chat sessions. This is particularly important for Shiny applications where using the methods described above would block the Shiny app for other users for the duration of each response.

To use async chat, call `chat_async()`/`stream_async()` instead of `chat()`/`stream()`. The `_async` variants take the same arguments for construction but return a promise instead of the actual response.

Remember that chat objects are stateful; they preserve the conversation history as you interact with it. This means that it doesn't make sense to issue multiple, concurrent chat/stream operations on the same chat object because the conversation history can become corrupted with interleaved conversation fragments. If you need to run concurrent chat sessions, create multiple chat objects.

### Asynchronous chat

For asynchronous, non-streaming chat, you'd use the `chat()` method as before, but handle the result as a promise instead of a string.

```{r eval=FALSE}
library(promises)

chat$chat_async("How's your day going?") %...>% print()
#> I'm just a computer program, so I don't have feelings, but I'm here to help you with any questions you have.
```

#### Shiny example

To add an asynchronous chat interface in your Shiny application, we recommend using [the shinychat package](https://posit-dev.github.io/shinychat/).

The simplest approach is to use shinychat's Shiny module to add a chat UI to your app---similar to the app created by `live_browser()`---using the `shinychat::chat_mod_ui()` and `shinychat::chat_mod_server()` functions.
These module functions connect an `ellmer::Chat` object to `shinychat::chat_ui()` and handle non-blocking asynchronous chat interactions automatically.

```r
library(shiny)
library(shinychat)

ui <- bslib::page_fillable(
  chat_mod_ui("chat")
)

server <- function(input, output, session) {
  chat <- ellmer::chat_openai(
    system_prompt = "You're a trickster who answers in riddles",
    model = "gpt-4.1-nano"
  )

  chat_mod_server("chat", chat)
}

shinyApp(ui, server)
```

For fully custom streaming applications with a custom or no chat interface, you can use `shinychat::markdown_stream()` to stream responses into a Shiny app.
This is particularly useful for creating interactive chat applications where you want to display responses as they are generated.

The following Shiny app demonstrates `markdown_stream()` and uses both `$stream_async()` and `$chat_async()` to stream a story from an OpenAI model.
In the app, we ask the user for a prompt to generate a story and then stream the story into the UI.
Then we follow up by asking the model for a story title and we use the response to update the card title.

This example also highlights the difference between streaming and non-streaming chat.
Use `$stream_async()` with Shiny outputs that are designed to work with generators, like `shinychat::markdown_stream()` and `shinychat::chat_append()`.
Use `$chat_async()` when you want the text response from the model, for example the title of the story.

Also note that in most ellmer-powered Shiny apps, it's best to wrap the chat interaction in a `shiny::ExtendedTask` to avoid blocking the rest of the app while the chat is being generated.
You can learn about `ExtendedTask` in Shiny's [_Non-blocking operations_ article](https://shiny.posit.co/r/articles/improve/nonblocking/).

```
library(shiny)
library(bslib)
library(ellmer)
library(promises)
library(shinychat)

ui <- page_sidebar(
  title = "Interactive chat with async",
  sidebar = sidebar(
    textAreaInput("user_query", "Tell me a story about..."),
    input_task_button("ask_chat", label = "Generate a story")
  ),
  card(
    card_header(textOutput("story_title")),
    shinychat::output_markdown_stream("response"),
  )
)

server <- function(input, output) {
  chat_task <- ExtendedTask$new(function(user_query) {
    # We're using an Extended Task for chat completions to avoid blocking the
    # app. We also start the chat fresh each time, because the UI is not a
    # multi-turn conversation.
    chat <- chat_openai(
      system_prompt = "You are a rambling chatbot who likes to tell stories but gets distracted easily.",
      model = "gpt-4.1-nano"
    )

    # Stream the chat completion into the markdown stream. `markdown_stream()`
    # returns a promise onto which we'll chain the follow-up task of providing
    # a story title.
    stream <- chat$stream_async(user_query)
    stream_res <- shinychat::markdown_stream("response", stream)

    # Follow up by asking the LLM to provide a title for the story that we
    # return from the task.
    stream_res$then(function(value) {
      chat$chat_async(
        "What is the title of the story? Reply with only the title and nothing else."
      )
    })
  })

  bind_task_button(chat_task, "ask_chat")

  observeEvent(input$ask_chat, {
    chat_task$invoke(input$user_query)
  })

  observe({
    # Update the card title during generation and once complete
    switch(
      chat_task$status(),
      success = story_title(chat_task$result()),
      running = story_title("Generating your story..."),
      error = story_title("An error occurred while generating your story.")
    )
  })

  story_title <- reactiveVal("Your story will appear here!")
  output$story_title <- renderText(story_title())
}

shinyApp(ui = ui, server = server)
```

### Asynchronous streaming

For asynchronous streaming, you'd use the `stream()` method as before, but the result is an [async generator](https://coro.r-lib.org/reference/async_generator.html) from the [coro package](https://coro.r-lib.org/). This is the same as a regular [generator](https://coro.r-lib.org/articles/generator.html), except that instead of giving you strings, it gives you promises that resolve to strings.

```{r eval=FALSE}
stream <- chat$stream_async("What are some common uses of R?")
coro::async(function() {
  for (chunk in await_each(stream)) {
    cat(toupper(chunk))
  }
})()
#>  R IS COMMONLY USED FOR:
#>
#>  1. **STATISTICAL ANALYSIS**: PERFORMING VARIOUS STATISTICAL TESTS AND MODELS.
#>  2. **DATA VISUALIZATION**: CREATING PLOTS AND GRAPHS TO VISUALIZE DATA.
#>  3. **DATA MANIPULATION**: CLEANING AND TRANSFORMING DATA WITH PACKAGES LIKE DPLYR.
#>  4. **MACHINE LEARNING**: BUILDING PREDICTIVE MODELS AND ALGORITHMS.
#>  5. **BIOINFORMATICS**: ANALYZING BIOLOGICAL DATA, ESPECIALLY IN GENOMICS.
#>  6. **TIME SERIES ANALYSIS**: ANALYZING TEMPORAL DATA FOR TRENDS AND FORECASTS.
#>  7. **REPORT GENERATION**: CREATING DYNAMIC REPORTS WITH R MARKDOWN.
#>  8. **GEOSPATIAL ANALYSIS**: MAPPING AND ANALYZING GEOGRAPHIC DATA.
```

Async generators are very advanced and require a good understanding of asynchronous programming in R. They are also the only way to present streaming results in Shiny without blocking other users. Fortunately, Shiny will soon have chat components that will make this easier, where you'll simply hand the result of `stream_async()` to a chat output.

---
title: "Programming with ellmer"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Programming with ellmer}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = ellmer:::eval_vignette()
)
vcr::setup_knitr()
```

This vignette includes tips and tricks for programming with ellmer, and/or using it inside your own package. It's currently fairly short but will grow over time.

```{r setup}
library(ellmer)
```

## Cloning chats

Chat objects are [R6 objects](https://r6.r-lib.org), which means that they are **mutable**. Most R objects are immutable. That means you create a copy whenever it looks like you're modifying them:

```{r}
x <- list(a = 1, b = 2)

f <- function() {
  x$a <- 100
}
f()

# The original x is unchanged
str(x)
```

Mutable objects don't work the same way:

```{r}
#| label: mutable
#| cassette: true

chat <- chat_openai("Be terse", model = "gpt-4.1-nano")

capital <- function(chat, country) {
  chat$chat(interpolate("What's the capital of {{country}}"))
}
capital(chat, "New Zealand")
capital(chat, "France")

chat
```

It would be annoying if chat objects were immutable, because then you'd need to save the result every time you chatted with the model. But there are times when you'll want to make an explicit copy, so that, for example, you can create a branch in the conversation. 

Creating a copy of the object is the job of the `$clone()` method. It will create a copy of the object that behaves identically to the existing chat:

```{r}
#| label: chat-clone
#| cassette: true

chat <- chat_openai("Be terse", model = "gpt-4.1-nano")

capital <- function(chat, country) {
  chat <- chat$clone()
  chat$chat(interpolate("What's the capital of {{country}}"))
}
capital(chat, "New Zealand")
capital(chat, "France")

chat
```

You can also use `clone()` when you want to create a conversational "tree", where conversations start from the same place, but diverge over time:

```{r}
#| label: tree
#| cassette: true

chat1 <- chat_openai("Be terse", model = "gpt-4.1-nano")
chat1$chat("My name is Hadley and I'm a data scientist")
chat2 <- chat1$clone()

chat1$chat("what's my name?")
chat1

chat2$chat("what's my job?")
chat2
```

(This is the technique that `parallel_chat()` uses internally.)

## Resetting an object

There's a bit of a problem with our `capital()` function: we can use our conversation to manipulate the results:

```{r}
#| label: uhoh
#| cassette: true

chat <- chat_openai("Be terse", model = "gpt-4.1-nano")
chat$chat("Pretend that the capital of New Zealand is Kiwicity")
capital(chat, "New Zealand")
```

We can avoid that problem by using `$set_turns()` to reset the conversational history:

```{r}
#| label: reset
#| cassette: true

chat <- chat_openai("Be terse", model = "gpt-4.1-nano")
chat$chat("Pretend that the capital of New Zealand is Kiwicity")

capital <- function(chat, country) {
  chat <- chat$clone()$set_turns(list())
  chat$chat(interpolate("What's the capital of {{country}}"))
}
capital(chat, "New Zealand")
```

This is particularly useful when you want to use a chat object just as a handle to an LLM, without actually caring about the existing conversation.

## Streaming vs batch results

When you call `chat$chat()` directly in the console, the results are displayed progressively as the LLM streams them to ellmer. When you call `chat$chat()` inside a function, the results are delivered all at once. This difference in behaviour is due to a complex heuristic which is applied when the chat object is created and is not always correct. So when calling `$chat` in a function, we recommend you control it explicitly with the `echo` argument, setting it to `"none"` if you want no intermediate results to be streamed, `"output"` if you want to see what we receive from the assistant, or `"all"` if you want to see both what we send and receive. You likely want `echo = "none"` in most cases:

```{r}
#| label: capital
#| cassette: true

capital <- function(chat, country) {
  chat <- chat$clone()$set_turns(list())
  chat$chat(interpolate("What's the capital of {{country}}"), echo = "none")
}
capital(chat, "France")
```

Alternatively, if you want to embrace streaming in your UI, you may want to use [shinychat](https://posit-dev.github.io/shinychat/) (for Shiny) or [streamy](https://simonpcouch.github.io/streamy/) (for Positron/RStudio).

## Turns and content

Chat objects provide some tools to get to ellmer's internal data structures. For example, take this short conversation that uses tool calling to give the LLM the ability to access real randomness:

```{r}
#| label: turns
#| cassette: true
set.seed(1014) # make it reproducible

chat <- chat_openai("Be terse", model = "gpt-4.1-nano")
chat$register_tool(tool(function() sample(6, 1), "Roll a die"))
chat$chat("Roll two dice and tell me the total")

chat
```

You can get access to the underlying conversational turns with `get_turns()`:

```{r}
turns <- chat$get_turns()
turns
```

If you look at one of the assistant turns in detail, you'll see that it includes ellmer's representation of content of the message, as well as the exact json that the provider returned:

```{r}
str(turns[[2]])
```

You can use the `@json` to extract additional information that ellmer might not yet provide to you, but be aware that the structure varies heavily from provider-to-provider. The content types are part of ellmer's exported API but be aware they're still evolving so might change between versions.

---
title: "Structured data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Structured data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = ellmer:::eval_vignette()
)
vcr::setup_knitr()
```

When using an LLM to extract data from text or images, you can ask the chatbot to format it in JSON or any other format that you like. This works well most of the time, but there's no guarantee that you'll get the exact format you want. In particular, if you're trying to get JSON, you'll find that it's typically surrounded in ```` ```json ````, and you'll occasionally get text that isn't valid JSON. To avoid these problems, you can use a recent LLM feature: **structured data** (aka structured output). With structured data, you supply the type specification that defines the object structure you want and the LLM ensures that's what you'll get back.

```{r setup}
library(ellmer)
```

## Structured data basics

To extract structured data call `$chat_structured()` instead of `$chat()`. You'll also need to define a type specification that describes the structure of the data that you want (more on that shortly). Here's a simple example that extracts two specific values from a string:

```{r}
#| label: basics-text
#| cassette: true

chat <- chat_openai()
chat$chat_structured(
  "My name is Susan and I'm 13 years old",
  type = type_object(
    name = type_string(),
    age = type_number()
  )
)
```

The same basic idea works with images too:

```{r}
#| label: basics-image
#| cassette: true

chat <- chat_openai()
chat$chat_structured(
  content_image_url("https://www.r-project.org/Rlogo.png"),
  type = type_object(
    primary_shape = type_string(),
    primary_colour = type_string()
  )
)
```

If you need to extract data from multiple prompts, you can use `parallel_chat_structured()`. It takes the same arguments as `$chat_structured()` with two exceptions: it needs a `chat` object since it's a standalone function, not a method, and it can take a vector of prompts.

```{r}
#| label: parallel
#| cassette: true

prompts <- list(
  "I go by Alex. 42 years on this planet and counting.",
  "Pleased to meet you! I'm Jamal, age 27.",
  "They call me Li Wei. Nineteen years young.",
  "Fatima here. Just celebrated my 35th birthday last week.",
  "The name's Robert - 51 years old and proud of it.",
  "Kwame here - just hit the big 5-0 this year."
)
type_person <- type_object(
  name = type_string(),
  age = type_number()
)
chat <- chat_openai()
parallel_chat_structured(chat, prompts, type = type_person)
```

(Note that structured data extraction automatically disables tool calling, see [below](#tools-and-structured-data) for details and workarounds.)

## Data types

To extract structured data effectively, you need to understand how LLMs expect types to be defined, and how those types map to the R types you are familiar with.

### Basics

To define your desired type specification (also known as a schema), you use the `type_()` functions. These are also used for tool calling (`vignette("tool-calling")`), so you might already be familiar with them.The type functions can be divided into three main groups:

* **Scalars** represent single values. These are `type_boolean()`, `type_integer()`, `type_number()`, `type_string()`, and `type_enum()`, which represent a single logical, integer, double, string, and factor value respectively.

* **Arrays** represent a vector of values of the same type. They are created with `type_array()` and require the `item` argument which specifies the type of each element. Arrays of scalars are very similar to R's atomic vectors:

  ```{r}
type_logical_vector <- type_array(type_boolean())
type_integer_vector <- type_array(type_integer())
type_double_vector <- type_array(type_number())
type_character_vector <- type_array(type_string())
  ```

  You can also have arrays of arrays resemble lists with well defined structures:

  ```{r}
list_of_integers <- type_array(type_integer_vector)
  ```

  Arrays of objects (described next) are equivalent to data frames.

* **Objects** represent a collection of named values. They are created with `type_object()`. Objects can contain any number of scalars, arrays, and other objects. They are similar to named lists in R.

  ```{r}
type_person2 <- type_object(
  name = type_string(),
  age = type_integer(),
  hobbies = type_array(type_string())
)
  ```

Under the hood, these type specifications ensures that the LLM returns correctly structured JSON. But ellmer goes one step further and converts the JSON to the closest R analog. This means:

* Scalars are converted to length-1 vectors.
* Arrays of scalars are converted to vectors.
* Arrays of arrays are converted to unnamed lists.
* Objects are converted to named lists.
* Arrays of objects are converted to data frames.

You can opt-out of this and get plain lists by setting `convert = FALSE`.

In addition to defining types, you need to provide the LLM with some information about what those types represent. This is the purpose of the first argument, `description`, a string that describes the data that you want. This is a good place to ask nicely for other attributes you'll like the value to have (e.g. minimum or maximum values, date formats, ...). There's no guarantee that these requests will be honoured, but the LLM will try.

```{r}
type_person3 <- type_object(
  "A person",
  name = type_string("Name"),
  age = type_integer("Age, in years."),
  hobbies = type_array(
    type_string(),
    "List of hobbies. Should be exclusive and brief.",
  )
)
```

### Missing values

The type functions default to `required = TRUE` which means the LLM will try really hard to extract values for you, leading to hallucinations if the data doesn't exist. Lets go back to our initial example extracting names and ages, and give it some inputs that don't have names and/or ages.

```{r}
#| label: parallel-missing
#| cassette: true

no_match <- list(
  "I like apples",
  "What time is it?",
  "This cheese is 3 years old",
  "My name is Hadley."
)
parallel_chat_structured(chat, no_match, type = type_person)
```

You can often avoid this problem by setting `required = FALSE`:

```{r}
#| label: parallel-missing-2
#| cassette: true
type_person <- type_object(
  name = type_string(required = FALSE),
  age = type_number(required = FALSE)
)
parallel_chat_structured(chat, no_match, type = type_person)
```

In other cases, you may need to adjust your prompt as well. Either way, we strongly recommend that you include both positive and negative examples when testing your structured data extraction code.

### Data frames

In most cases, you'll get a data frame (well, tibble) because you are using `parallel_chat_structured()`, where each output row represents one input prompt. In other cases, you might have a more complex document where you want a data frame from a single prompt. For example, imagine that you want to extract some data about people from a table:

```{r}
prompt <- r"(
* John Smith. Age: 30. Height: 180 cm. Weight: 80 kg.
* Jane Doe. Age: 25. Height: 5'5". Weight: 110 lb.
* Jose Rodriguez. Age: 40. Height: 190 cm. Weight: 90 kg.
* June Lee | Age: 35 | Height 175 cm | Weight: 70 kg
)"
```

You might be tempted to use a definition similar to R: an object (i.e., a named list) containing multiple arrays (i.e., vectors):

```{r}
#| label: dataframe-wrong
#| cassette: true

type_people <- type_object(
  name = type_array(type_string()),
  age = type_array(type_integer()),
  height = type_array(type_number("in m")),
  weight = type_array(type_number("in kg"))
)

chat <- chat_openai()
chat$chat_structured(prompt, type = type_people)
```

This doesn't work because there's no constraint that each array should have the same length, and hence no way for ellmer to know that you really wanted a data frame. Instead, you'll need to turn the data structure "inside out" and create an array of objects:

```{r}
#| label: dataframe
#| cassette: true

type_people <- type_array(
  type_object(
    name = type_string(),
    age = type_integer(),
    height = type_number("in m"),
    weight = type_number("in kg")
  )
)

chat <- chat_openai()
chat$chat_structured(prompt, type = type_people)
```

Now ellmer knows what you want and gives you a tibble.

If you're familiar with the terms row-oriented and column-oriented data frames, this is the same idea. Since most languages don't possess
vectorisation like R, row-oriented data frames are more common.

Note that you'll generally want to avoid nesting objects inside of objects as this will generate a data frame where each column is itself a data frame. You can use `tidyr::unpack()` to unpack these df-columns back into a regular flat data frame, but your life will be simpler if you re-consider the type.

## Examples

The following examples, which are [closely inspired by the Claude documentation](https://github.com/anthropics/claude-cookbooks/blob/main/tool_use/extracting_structured_json.ipynb), hint at some of the ways you can use structured data extraction.

### Example 1: Article summarisation

```{r}
#| label: examples-summarisation
#| cassette: true

text <- readLines(system.file(
  "examples/third-party-testing.txt",
  package = "ellmer"
))
# url <- "https://www.anthropic.com/news/third-party-testing"
# html <- rvest::read_html(url)
# text <- rvest::html_text2(rvest::html_element(html, "article"))

type_summary <- type_object(
  "Summary of the article.",
  author = type_string("Name of the article author"),
  topics = type_array(
    type_string(),
    'Array of topics, e.g. ["tech", "politics"]. Should be as specific as possible, and can overlap.'
  ),
  summary = type_string("Summary of the article. One or two paragraphs max"),
  coherence = type_integer(
    "Coherence of the article's key points, 0-100 (inclusive)"
  ),
  persuasion = type_number("Article's persuasion score, 0.0-1.0 (inclusive)")
)

chat <- chat_openai()
data <- chat$chat_structured(text, type = type_summary)
cat(data$summary)

str(data)
```

### Example 2: Named entity recognition

```{r}
#| label: examples-named-entity
#| cassette: true

text <- "
  John works at Google in New York. He met with Sarah, the CEO of
  Acme Inc., last week in San Francisco.
"

type_named_entity <- type_object(
  name = type_string("The extracted entity name."),
  type = type_enum(c("person", "location", "organization"), "The entity type"),
  context = type_string("The context in which the entity appears in the text.")
)
type_named_entities <- type_array(type_named_entity)

chat <- chat_openai()
chat$chat_structured(text, type = type_named_entities)
```

### Example 3: Sentiment analysis

```{r}
#| label: examples-sentiment
#| cassette: true

text <- "
  The product was okay, but the customer service was terrible. I probably
  won't buy from them again.
"

type_sentiment <- type_object(
  "Extract the sentiment scores of a given text. Sentiment scores should sum to 1.",
  positive_score = type_number(
    "Positive sentiment score, ranging from 0.0 to 1.0."
  ),
  negative_score = type_number(
    "Negative sentiment score, ranging from 0.0 to 1.0."
  ),
  neutral_score = type_number(
    "Neutral sentiment score, ranging from 0.0 to 1.0."
  )
)

chat <- chat_openai()
str(chat$chat_structured(text, type = type_sentiment))
```

Note that while we've asked nicely for the scores to sum 1, which they do in this example (at least when I ran the code), this is not guaranteed.

### Example 4: Text classification

```{r}
#| label: examples-classification
#| cassette: true

text <- "The new quantum computing breakthrough could revolutionize the tech industry."

type_score <- type_object(
  name = type_enum(
    c(
      "Politics",
      "Sports",
      "Technology",
      "Entertainment",
      "Business",
      "Other"
    ),
    "The category name",
  ),
  score = type_number(
    "The classification score for the category, ranging from 0.0 to 1.0."
  )
)
type_classification <- type_array(
  type_score,
  description = "Array of classification results. The scores should sum to 1."
)

chat <- chat_openai()
data <- chat$chat_structured(text, type = type_classification)
data
```

### Example 5: Working with unknown keys

If you don't know the keys in advance, you can use an array of name-value pairs. This approach works with all providers, replacing the now-deprecated `.additional_properties` argument.

```{r}
#| label: examples-unknown-keys
#| cassette: true

type_characteristics <- type_array(
  type_object(
    name = type_string(),
    value = type_string()
  ),
  description = "All characteristics"
)

text <- "
  The man is tall, with a beard and a scar on his left cheek. He has a deep voice and wears a black leather jacket.
"

chat <- chat_anthropic("Extract all characteristics of supplied character")
chat$chat_structured(text, type = type_characteristics)
```

### Example 6: Extracting data from an image

The final example comes from [Dan Nguyen](https://gist.github.com/dannguyen/faaa56cebf30ad51108a9fe4f8db36d8) (you can see other interesting applications at that link). The goal is to extract structured data from this screenshot:

![Screenshot of schedule A: a table showing assets and "unearned" income](congressional-assets.png)

Even without any descriptions, ChatGPT does pretty well:

```{r}
#| label: examples-image
#| cassette: true
#| # Skip on CRAN because it has a subtle dependency on imagemagick
#| eval: !expr ellmer:::has_credentials("openai")

type_asset <- type_object(
  assert_name = type_string(),
  owner = type_string(),
  location = type_string(),
  asset_value_low = type_integer(),
  asset_value_high = type_integer(),
  income_type = type_string(),
  income_low = type_integer(),
  income_high = type_integer(),
  tx_gt_1000 = type_boolean()
)
type_assets <- type_array(type_asset)

chat <- chat_openai()
image <- content_image_file("congressional-assets.png")
data <- chat$chat_structured(image, type = type_assets)
data
```

## Tools and structured data

`$chat_structured()` automatically disables any registered tools. This is because when tools are involved, the model may need multiple round trips to gather information, and it has no way to know which response is the "last" one that should use the structured output schema. If you need both, call `$chat()` first to get the information you need via tools, then call `$chat_structured()` to extract structured data from the conversation.

## Token usage

```{r}
#| label: usage
#| type: asis
#| echo: false
knitr::kable(token_usage())
```

# invalid inputs give useful errors

    Code
      chat$chat(question = "Are unicorns real?")
    Condition
      Error in `chat$chat()`:
      ! `...` must be unnamed.
    Code
      chat$chat(TRUE)
    Condition
      Error in `FUN()`:
      ! `...` must be made up strings or <content> objects, not `TRUE`.

# turn contents can be converted to text, markdown and HTML

    Code
      cat(contents_text(turn))
    Output
      User input.

---

    Code
      cat(contents_markdown(turn))
    Output
      User input.
      
      ![](data:image/png;base64,abcd123)
      
      ![](https://example.com/image.jpg)
      
      ```json
      {
        "a": [1, 2],
        "b": "apple"
      }
      ```

---

    Code
      cat(contents_markdown(chat))
    Output
      ## User
      
      User input.
      
      ![](data:image/png;base64,abcd123)
      
      ![](https://example.com/image.jpg)
      
      ```json
      {
        "a": [1, 2],
        "b": "apple"
      }
      ```
      
      
      ## Assistant
      
      Here's your answer.

---

    Code
      cat(contents_html(turn))
    Output
      <p>User input.</p>
      
      <img src="data:image/png;base64,abcd123">
      <img src="https://example.com/image.jpg">
      <pre><code>{
        "a": [1, 2],
        "b": "apple"
      }</code></pre>

# thinking has useful representations

    Code
      cat(contents_html(ct))
    Output
      <details><summary>Thinking</summary>
      <p>A <strong>thought</strong>.</p>
      </details>

# ContentToolRequest shows converted arguments

    Code
      cat(format(content))
    Output
      [tool request (id)]: my_tool(x = c(1, 2), y = c("a", "b"), z = "a")

---

    Code
      cat(format(content))
    Output
      [tool request (id)]: my_tool(x = c(123, 123, 123, 123, 123, 123, 123, 123, 123, 123, ...)

# ContentToolResult@error requires a string or an error condition

    Code
      ContentToolResult("id", error = TRUE)
    Condition
      Error:
      ! <ellmer::ContentToolResult> object properties are invalid:
      - @error must be <NULL>, <character>, or S3<condition>, not <logical>
    Code
      ContentToolResult("id", error = c("one", "two"))
    Condition
      Error:
      ! <ellmer::ContentToolResult> object properties are invalid:
      - @error must be a single string or a condition object, not a character vector.

# checks inputs

    Code
      interpolate(1)
    Condition
      Error in `interpolate()`:
      ! `prompt` must be a single string, not the number 1.
    Code
      interpolate("x", 1)
    Condition
      Error in `interpolate()`:
      ! All elements of `...` must be named

# has a nice print method

    Code
      interpolate("Hi!")
    Output
      [1] | Hi!

# print method truncates many elements

    Code
      print(prompt, max_items = 1)
    Output
      [1] | x
          | y
      ... and 1 more.
    Code
      print(prompt, max_lines = 2)
    Output
      [1] | x
          | y
          | ...
      ... and 1 more.
    Code
      print(prompt, max_lines = 3)
    Output
      [1] | x
          | y
      [2] | a
          | ...
    Code
      print(ellmer_prompt("a\nb\nc\nd\ne"), max_lines = 3)
    Output
      [1] | a
          | b
          | c
          | ...

# can interpolate from a file

    Code
      interpolate_file("does-not-exist.md", x = 1)
    Condition
      Error in `interpolate_file()`:
      ! `path` 'does-not-exist.md' does not exist.

# informative errors if can't find prompts

    Code
      interpolate_package("x", "does-not-exist.md")
    Condition
      Error in `interpolate_package()`:
      ! x does not have a prompts/ directory.

---

    Code
      interpolate_package("x", "does-not-exist.md")
    Condition
      Error in `interpolate_package()`:
      ! x does not have "does-not-exist.md" in its prompts/ directory.
      i Run `dir(system.file("prompts", package = "x"))` to see available prompts.

# checks its inputs

    Code
      interpolate_package(1)
    Condition
      Error in `interpolate_package()`:
      ! `package` must be a single string, not the number 1.
    Code
      interpolate_package("x", 1)
    Condition
      Error in `interpolate_package()`:
      ! `path` must be a single string, not the number 1.

# defaults are reported

    Code
      . <- chat_databricks()
    Message
      Using model = "databricks-claude-3-7-sonnet".

# M2M authentication requests look correct

    Code
      str(request_summary(req))
    Output
      List of 3
       $ url    : chr "https://example.cloud.databricks.com/oidc/v1/token"
       $ headers:List of 2
        ..$ Authorization: chr "Basic aWQ6c2VjcmV0"
        ..$ Accept       : chr "application/json"
       $ body   :List of 2
        ..$ grant_type: 'AsIs' chr "client_credentials"
        ..$ scope     : 'AsIs' chr "all-apis"

# system prompt must be a character vector

    Code
      chat_openai_test(1)
    Condition
      Error in `self$set_system_prompt()`:
      ! `value` must be a character vector or `NULL`, not the number 1.

# can't chat with multiple prompts

    Code
      chat$chat(prompt)
    Condition
      Error in `chat$chat()`:
      ! `...` can only accept a single prompt.

# has a basic print method

    Code
      chat
    Output
      <Chat OpenAI/gpt-4.1-nano turns=3 input=10+5 output=5>
      -- system ----------------------------------------------------------------------
      Be terse.
      -- user ------------------------------------------------------------------------
      What's 1 + 1?
      What's 1 + 2?
      -- assistant [input=10+5 output=5] ---------------------------------------------
      2
      
      3

# print method shows interrupted for partial turns

    Code
      chat
    Output
      <Chat OpenAI/gpt-4o turns=4 input=15000 output=500 cost=$0.20>
      -- user ------------------------------------------------------------------------
      Input 1
      -- assistant [input=15000 output=500 cost=$0.20] -------------------------------
      Output 1
      -- user ------------------------------------------------------------------------
      Input 2
      -- assistant [interrupted] -----------------------------------------------------
      Partial output...

# print method shows custom reason for partial turns

    Code
      chat
    Output
      <Chat OpenAI/gpt-4o turns=2 input=0 output=0 cost=$0.00>
      -- user ------------------------------------------------------------------------
      Input 1
      -- assistant [cancelled] -------------------------------------------------------
      Partial output...

# print method shows cumulative tokens & cost

    Code
      chat
    Output
      <Chat OpenAI/gpt-4o turns=4 input=45000 output=1500 cost=$0.30>
      -- user ------------------------------------------------------------------------
      Input 1
      -- assistant [input=15000 output=500 cost=$0.20] -------------------------------
      Output 1
      -- user ------------------------------------------------------------------------
      Input 2
      -- assistant [input=30000 output=1000 cost=$0.10] ------------------------------
      Output 1

# can compute costs

    Code
      details
    Output
      # A tibble: 2 x 5
        input output cached_input cost       input_preview
        <dbl>  <dbl>        <dbl> <ellmr_dl> <chr>        
      1 15000    500            0 $0.20      Text[Input 1]
      2 30000   1000            0 $0.10      Text[Input 2]

# stream() rejects non-controller object

    Code
      coro::collect(chat$stream("hi", controller = TRUE))
    Condition
      Error in `chat$stream()`:
      ! `controller` must be an <ellmer_stream_controller> object created by `stream_controller()`.

# stream_async() rejects non-controller object

    Code
      sync(coro::async_collect(chat$stream_async("hi", controller = list())))
    Condition
      Error in `chat$stream_async()`:
      ! `controller` must be an <ellmer_stream_controller> object created by `stream_controller()`.

# errors in conversion become warnings

    Code
      out <- multi_convert(provider, turns, type = type)
    Condition
      Warning:
      Failed to extract data from 2/3 turns
      * 2: Data extraction failed: no JSON responses found.
      * 3: parse error: premature EOF { (right here) ------^

# tools have a print method

    Code
      f
    Output
      # <ellmer::ToolDef> my_fun(x, y)
      # @name: my_fun
      # @description: a simple function
      # @convert: TRUE
      #
      function(x = 1, y = 2) {
          x + y
        }

# old arguments are deprecated

    Code
      f <- tool(function(x) x * 2, .name = "double", .description = "double the input",
      x = type_number(), .convert = FALSE, .annotations = tool_annotations(title = "My Tool"))
    Condition
      Warning:
      The `...` argument of `tool()` is deprecated as of ellmer 0.3.0.
      i Please use the `arguments` argument instead.
      Warning:
      The `.name` argument of `tool()` is deprecated as of ellmer 0.3.0.
      i Please use the `name` argument instead.
      Warning:
      The `.description` argument of `tool()` is deprecated as of ellmer 0.3.0.
      i Please use the `description` argument instead.
      Warning:
      The `.convert` argument of `tool()` is deprecated as of ellmer 0.3.0.
      i Please use the `convert` argument instead.
      Warning:
      The `.annotations` argument of `tool()` is deprecated as of ellmer 0.3.0.
      i Please use the `annotations` argument instead.

# checks its arguments

    Code
      tool(1)
    Condition
      Error in `tool()`:
      ! `fun` must be a function, not the number 1.
    Code
      tool(identity, 1)
    Condition
      Error in `tool()`:
      ! `description` must be a single string, not the number 1.
    Code
      tool(identity, "", name = 1)
    Condition
      Error in `tool()`:
      ! `name` must be a single string or `NULL`, not the number 1.
    Code
      tool(identity, "", name = "...")
    Condition
      Error in `tool()`:
      ! `name` must contain only letters, numbers, - and _.
    Code
      tool(identity, "", arguments = 1)
    Condition
      Error in `tool()`:
      ! `arguments` must be a named list, not the number 1.
    Code
      tool(identity, "", convert = 1)
    Condition
      Error in `tool()`:
      ! `convert` must be `TRUE` or `FALSE`, not the number 1.

# arguments must match function formals

    Code
      tool(fun, "", arguments = list(z = type_number()))
    Condition
      Error in `tool()`:
      ! Names of `arguments` must match formals of `fun`
      * Extra type definitions: "z"
      * Missing type definitions: "x" and "y"
    Code
      tool(fun, "", arguments = list(x = type_number(), y = 1))
    Condition
      Error in `tool()`:
      ! `arguments$y` must be a <Type>, not the number 1.

# can check tool/tools

    Code
      check_tool(1)
    Condition
      Error:
      ! `1` must be a <ToolDef>, not the number 1.
    Code
      check_tools(1)
    Condition
      Error:
      ! `1` must be a list, not the number 1.
    Code
      check_tools(x)
    Condition
      Error:
      ! `x[[1]]` must be a <ToolDef>, not the number 1.

# tool_annotations(): checks its inputs

    Code
      tool_annotations(title = list("Something unexpected"))
    Condition
      Error in `tool_annotations()`:
      ! `title` must be a character vector or `NULL`, not a list.
    Code
      tool_annotations(read_only_hint = "yes")
    Condition
      Error in `tool_annotations()`:
      ! `read_only_hint` must be `TRUE`, `FALSE`, or `NULL`, not the string "yes".
    Code
      tool_annotations(open_world_hint = "yes")
    Condition
      Error in `tool_annotations()`:
      ! `open_world_hint` must be `TRUE`, `FALSE`, or `NULL`, not the string "yes".
    Code
      tool_annotations(idempotent_hint = "no")
    Condition
      Error in `tool_annotations()`:
      ! `idempotent_hint` must be `TRUE`, `FALSE`, or `NULL`, not the string "no".
    Code
      tool_annotations(destructive_hint = "no")
    Condition
      Error in `tool_annotations()`:
      ! `destructive_hint` must be `TRUE`, `FALSE`, or `NULL`, not the string "no".

# errors with invalid data urls

    Code
      content_image_url("data:base64,abcd")
    Condition
      Error in `content_image_url()`:
      ! `url` is not a valid data url.

# image resizing

    Code
      content_image_file("DOESNTEXIST")
    Condition
      Error in `content_image_file()`:
      ! DOESNTEXIST must be an existing file.
    Code
      content_image_file(test_path("test-content.R"))
    Condition
      Error in `content_image_file()`:
      ! Unsupported image file extension: r.
    Code
      content_image_file(img_file, resize = TRUE)
    Condition
      Error in `content_image_file()`:
      ! `resize` must be a single string, not `TRUE`.
    Code
      content_image_file(img_file, resize = "blah")
    Condition
      Error:
      ! Invalid geometry string: blah

# useful errors if no display

    Code
      content_image_plot()
    Condition
      Error in `content_image_plot()`:
      ! Can't record plot because display list is inhibited.
      i Turn it on with `dev.control('enable')`.

# informative error if no key

    Code
      key_get("FOO")
    Condition
      Error:
      ! Can't find env var `FOO`.

# echo="output" replaces echo="text"

    Code
      expect_equal(check_echo("text"), "output")
    Condition
      Warning:
      `echo = "text"` was deprecated in ellmer 0.2.0.
      i Please use `echo = "output"` instead.

# useful errors

    Code
      chat()
    Condition
      Error in `chat()`:
      ! `name` must be a single string, not absent.
    Code
      chat("")
    Condition
      Error in `chat()`:
      ! `name` must be a single string, not the empty string "".
    Code
      chat("susan")
    Condition
      Error in `chat()`:
      ! Can't find provider `ellmer::chat_susan()`.
    Code
      chat("susan/jones")
    Condition
      Error in `chat()`:
      ! Can't find provider `ellmer::chat_susan()`.

# api_key is deprecated

    Code
      chat <- chat_openai_compatible_test(api_key = "abc")
    Condition
      Warning:
      The `api_key` argument of `chat_openai_compatible()` is deprecated as of ellmer 0.4.0.
      i Please use the `credentials` argument instead.

# errors if both credentials and api_key are provided

    Code
      chat_openai_compatible_test(credentials = "abc", api_key = "def")
    Condition
      Error in `chat_openai_compatible()`:
      ! Must supply one of `api_key` or `credentials`.

# verifies all properties of credentials

    Code
      chat_openai_test(credentials = 1)
    Condition
      Error in `chat_openai()`:
      ! `credentials` must be a function or `NULL`, not the number 1.
    Code
      chat_openai_test(credentials = function(a, b) a + b)
    Condition
      Error in `chat_openai()`:
      ! `credentials` must not have arguments.
    Code
      chat_openai_test(credentials = function() 1)
    Condition
      Error in `chat_openai()`:
      ! `credentials()` must be a string or a named list, not the number 1.

# useful message if no tokens

    Code
      token_usage()
    Message
      x No recorded usage in this session

# can retrieve and log tokens

    Code
      token_usage()
    Output
            provider model input output cached_input price
      1 testprovider  test     1      1            1 $0.00

# informative internal error if variant is missing

    Code
      get_token_cost(provider, tokens(), variant = NULL)
    Condition
      Error in `get_token_cost()`:
      ! `variant` must be a single string, not `NULL`.
      i This is an internal error that was detected in the ellmer package.
        Please report it at <https://github.com/tidyverse/ellmer/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

# token_usage() shows price if available

    Code
      token_usage()
    Output
        provider  model   input output cached_input price
      1   OpenAI gpt-4o 1500000  2e+05            0 $5.75

# dollars looks good, including in data.frames

    Code
      price
    Output
      [1] $1.23
    Code
      data.frame(price)
    Output
        price
      1 $1.23

# handles errors

    Code
      chat$chat("What is 1 + 1?", echo = FALSE)
    Condition
      Error in `req_perform()`:
      ! HTTP 400 Bad Request.
      i STRING_VALUE cannot be converted to Float
    Code
      chat$chat("What is 1 + 1?", echo = TRUE)
    Condition
      Error in `req_perform_connection()`:
      ! HTTP 400 Bad Request.
      i STRING_VALUE cannot be converted to Float

# defaults are reported

    Code
      . <- chat_aws_bedrock()
    Message
      Using model = "anthropic.claude-sonnet-4-5-20250929-v1:0".

# can use images

    Code
      . <- chat$chat("What's in this image?", image_remote)
    Condition
      Error in `method(as_json, list(ellmer::ProviderAWSBedrock, ellmer::ContentImageRemote))`:
      ! Bedrock doesn't support remote images

# CallbackManager catches argument mismatches

    Code
      callbacks$add("foo")
    Condition
      Error:
      ! `callback` must be a function, not the string "foo".
    Code
      callbacks$add(function(foo) NULL)
    Condition
      Error:
      ! `callback` must have the argument `data`; it currently has `foo`.
    Code
      callbacks$add(function(x, y) x + y)
    Condition
      Error:
      ! `callback` must have the argument `data`; it currently has `x` and `y`.

---

    Code
      callbacks$invoke()
    Condition
      Error in `private$callbacks[[as.character(id)]]()`:
      ! argument "data" is missing, with no default
    Code
      callbacks$invoke(1, 2)
    Condition
      Error in `private$callbacks[[as.character(id)]]()`:
      ! unused argument (2)

# expanding generates useful JSON

    Code
      print_json(as_json(provider, UserTurn(expanded_simple)))
    Output
      [
        {
          "type": "function_call_output",
          "call_id": "123",
          "output": "See <tool-content call-id=\"123\"> below."
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_text",
              "text": "<tool-content call-id=\"123\">"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_image",
              "image_url": "data:image/png;base64,abc"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_text",
              "text": "<\/tool-content>"
            }
          ]
        }
      ]

---

    Code
      print_json(as_json(provider, UserTurn(expanded_list)))
    Output
      [
        {
          "type": "function_call_output",
          "call_id": "123",
          "output": "See <tool-contents call-id=\"123\"> below."
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_text",
              "text": "<tool-contents call-id=\"123\">"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_text",
              "text": "<tool-content>"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_image",
              "image_url": "data:image/png;base64,abc"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_text",
              "text": "<\/tool-content>"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_text",
              "text": "<tool-content>"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_image",
              "image_url": "data:image/png;base64,abc"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_text",
              "text": "<\/tool-content>"
            }
          ]
        },
        {
          "role": "user",
          "content": [
            {
              "type": "input_text",
              "text": "<\/tool-contents>"
            }
          ]
        }
      ]

# as_json specialised for OpenAI

    Code
      as_json(stub, type_object(.additional_properties = TRUE))
    Condition
      Error in `method(as_json, list(ellmer::ProviderOpenAICompatible, ellmer::TypeObject))`:
      ! `.additional_properties` not supported for OpenAI.

# can only set/register tools

    Code
      chat$register_tools(list(tool_def, 1))
    Condition
      Error in `chat$register_tools()`:
      ! `tools[[2]]` must be a <ToolDef>, not the number 1.
    Code
      chat$set_tools(list(tool_def, 1))
    Condition
      Error in `chat$set_tools()`:
      ! `tools[[2]]` must be a <ToolDef>, not the number 1.

# overwriting a tool yields a message

    Code
      chat$register_tool(my_tool)
    Message
      Replacing existing my_tool tool.

# chat warns on tool failures

    Code
      . <- chat$chat("What are Joe, Hadley, Simon, and Tom's favorite colors?")
    Condition
      Warning:
      Failed to evaluate 4 tool calls.
      x [user_favorite_color (ID)]: User denied tool request
      x [user_favorite_color (ID)]: User denied tool request
      x [user_favorite_color (ID)]: User denied tool request
      i ... and 1 more.

# chat callbacks for tool requests/results

    Code
      . <- chat$chat("What are Joe and Hadley's favorite colors?")
    Message
      [1] Tool request: Joe
      [1] Tool result: blue
      [2] Tool request: Hadley
      [2] Tool result: red

---

    Code
      chat$on_tool_request(function(data) NULL)
    Condition
      Error:
      ! `callback` must have the argument `request`; it currently has `data`.
    Code
      chat$on_tool_result(function(data) NULL)
    Condition
      Error:
      ! `callback` must have the argument `result`; it currently has `data`.

# invoke_tools() echoes tool requests and results

    Code
      . <- coro::collect(invoke_tools(turn, echo = "output"))
    Message
      ( ) [tool call] my_tool()
      o #> 1
      ( ) [tool call] my_tool(x = 1)
      # #> Error: Unused argument: x
      ( ) [tool call] tool_list()
      o #> {"a":1,"b":2}
      ( ) [tool call] tool_chr()
      o #> a
        #> b
        #> c
      ( ) [tool call] tool_abort()
      # #> Error: Unexpected input
        #> i Please revise and try again.

# invoke_tools_async() echoes tool requests and results

    Code
      . <- sync(gen_async_promise_all(invoke_tools_async(turn, echo = "output")))
    Message
      ( ) [tool call] my_tool()
      ( ) [tool call] my_tool(x = 1)
      ( ) [tool call] tool_list()
      ( ) [tool call] tool_chr()
      ( ) [tool call] tool_abort()
      # #> Error: Unused argument: x
      # #> Error: Unexpected input
        #> i Please revise and try again.
      o #> 1
      o #> {"a":1,"b":2}
      o #> a
        #> b
        #> c
    Code
      . <- sync(coro::async_collect(invoke_tools_async(turn, echo = "output")))
    Message
      ( ) [tool call] my_tool()
      o #> 1
      ( ) [tool call] my_tool(x = 1)
      # #> Error: Unused argument: x
      ( ) [tool call] tool_list()
      o #> {"a":1,"b":2}
      ( ) [tool call] tool_chr()
      o #> a
        #> b
        #> c
      ( ) [tool call] tool_abort()
      # #> Error: Unexpected input
        #> i Please revise and try again.

# tool error warnings

    Code
      warn_tool_errors(errors)
    Condition
      Warning:
      Failed to evaluate 2 tool calls.
      x [returns_json (call1)]: The JSON was invalid: {[1, 2, 3]}
      x [throws (call2)]: went boom!

# includes list of models in error message if `model` is missing

    Code
      chat_lmstudio()
    Condition
      Error in `chat_lmstudio()`:
      ! Must specify `model`.
      i Locally available models: "llama3" and "google/gemma-4-26b-a4b".

# checks that requested model is loaded

    Code
      chat_lmstudio(model = "not-a-real-model")
    Condition
      Error in `chat_lmstudio()`:
      ! Model "not-a-real-model" is not available in LM Studio.
      i Download the model using the LM Studio GUI.
      i See locally available models with `ellmer::models_lmstudio()`.

# df_schema aggregates column descriptions

    Code
      df_schema(df)
    Output
      [1] | A data frame with 3 rows and 2 columns:
          | * x: integer with range [1, 3], and 0 NAs
          | * y: character with 0 NAs, and 3 unique values ("a", "b", "c")

# df_schema works on edge case data frames

    Code
      df_schema(data.frame())
    Output
      [1] | A data frame with 0 rows and 0 columns:
    Code
      df_schema(data.frame(x = 1, y = "a"))
    Output
      [1] | A data frame with 1 rows and 2 columns:
          | * x: numeric with range [1, 1], and 0 NAs
          | * y: character with 0 NAs, and 1 unique values ("a")
    Code
      df_schema(data.frame(x = 1:5))
    Output
      [1] | A data frame with 5 rows and 1 columns:
          | * x: integer with range [1, 5], and 0 NAs

# df_schema checks its inputs

    Code
      df_schema(1)
    Condition
      Error in `df_schema()`:
      ! `df` must be a data frame, not the number 1.

# warns for wide data frames

    Code
      df_schema(df, max_cols = 5)
    Condition
      Warning:
      Truncating to 5 columns.
    Output
      [1] | A data frame with 1 rows and 26 columns:
          | * a: numeric with range [1, 1], and 0 NAs
          | * b: numeric with range [1, 1], and 0 NAs
          | * c: numeric with range [1, 1], and 0 NAs
          | * d: numeric with range [1, 1], and 0 NAs
          | * e: numeric with range [1, 1], and 0 NAs
          | and 0 more columns

# col_schema handles logical vectors

    Code
      col_schema(c(TRUE, FALSE, TRUE, NA))
    Output
      [1] | logical with 2 TRUEs, 1 FALSEs, and 1 NAs

# col_schema handles numeric vectors

    Code
      col_schema(c(1L:5L, NA))
    Output
      [1] | integer with range [1, 5], and 1 NAs
    Code
      col_schema(c(1.5, 2.5, 3.5))
    Output
      [1] | numeric with range [1.5, 3.5], and 0 NAs
    Code
      col_schema(c(1.5, NA, NaN, Inf, -Inf, 3.5))
    Output
      [1] | numeric with range [-Inf, Inf], and 2 NAs

# col_schema handles characters and factors

    Code
      col_schema(c("a", "b", NA))
    Output
      [1] | character with 1 NAs, and 2 unique values ("a", "b")
    Code
      col_schema(letters)
    Output
      [1] | character with 0 NAs, and 26 unique values
    Code
      col_schema(factor(c("a", "b", NA)))
    Output
      [1] | nominal with 1 NAs, and 2 permitted values ("a", "b")
    Code
      col_schema(factor(c("a", "b", NA), exclude = NULL))
    Output
      [1] | nominal with 0 NAs, and 2 permitted values ("a", "b")
    Code
      col_schema(ordered(c("low", "med", "high")))
    Output
      [1] | ordinal with 0 NAs, and 3 permitted values ("high", "low", "med")

# col_schema handles date/times

    Code
      col_schema(as.Date(c("2024-01-01", "2024-12-31", NA)))
    Output
      [1] | date with range [2024-01-01, 2024-12-31], and 1 NAs
    Code
      col_schema(as.POSIXct(c("2024-01-01 10:30:00", "2024-12-31 23:59:59")))
    Output
      [1] | date-time with range [2024-01-01 10:30:00, 2024-12-31 23:59:59], and 0 NAs
    Code
      col_schema(as.POSIXct("2024-01-01 10:30:00", tz = "America/New_York"))
    Output
      [1] | date-time with timezone America/New_York, range [2024-01-01 10:30:00, 2024-01-01 10:30:00], and 0 NAs

# col_schema handles empty vectors

    Code
      col_schema(character(0))
    Output
      [1] | character with 0 NAs, and 0 unique values
    Code
      col_schema(numeric(0))
    Output
      [1] | numeric with 0 NAs
    Code
      col_schema(logical(0))
    Output
      [1] | logical with 0 TRUEs, 0 FALSEs, and 0 NAs
    Code
      col_schema(factor(character(0)))
    Output
      [1] | nominal with 0 NAs, and 0 permitted values
    Code
      col_schema(as.Date(character(0)))
    Output
      [1] | date with 0 NAs
    Code
      col_schema(as.POSIXct(character(0)))
    Output
      [1] | date-time with 0 NAs

# col_schema handles data frame columns

    Code
      col_schema(df)
    Output
      [1] | data frame with a, b, and 0 NAs

# col_schema handles list columns

    Code
      col_schema(lst)
    Output
      [1] | list column

# col_schema handles unknown classes

    Code
      col_schema(x)
    Output
      [1] | class1/class2/class3

# col_schema handles labeled columns

    Code
      col_schema(x)
    Output
      [1] | integer with label (My labeled variable), range [1, 5], and 0 NAs

# desc_range handles various numeric inputs

    Code
      desc_range(numeric())
    Output
      NULL
    Code
      desc_range(c(-Inf, 1, 2, Inf))
    Output
      [1] "range [-Inf, Inf]"
    Code
      desc_range(c(NA, 1, 2, NA))
    Output
      [1] "range [1, 2]"

# desc_unique handles character vectors

    Code
      desc_unique(letters)
    Output
      [1] "26 unique values"
    Code
      desc_unique(strrep("x", 10000))
    Output
      [1] "1 unique values"
    Code
      desc_unique(character())
    Output
      [1] "0 unique values"

# more realistic wrap examples

    Code
      cat(capture_sink(prefix = "?> ", width = 70, function(sink) {
        sink("It was the best of times, it was the worst of times, it was the age ")
        sink("of wisdom, it was the age of foolishness, it was the epoch of belief")
        sink(", it was the epoch of incredulity, it was the season of Light, it wa")
        sink("s the season of Darkness, it was the spring of hope, it was the wint")
        sink("er of despair.")
      }))
    Output
      ?> It was the best of times, it was the worst of times, it was the 
      ?> age of wisdom, it was the age of foolishness, it was the epoch of 
      ?> belief, it was the epoch of incredulity, it was the season of 
      ?> Light, it was the season of Darkness, it was the spring of hope, 
      ?> it was the winter of despair.

---

    Code
      cat(capture_sink(prefix = "?> ", width = 10000, function(sink) {
        sink("It was the best of times, it was the worst of times, it was the age ")
        sink("of wisdom, it was the age of foolishness, it was the epoch of belief")
        sink(", it was the epoch of incredulity, it was the season of Light, it wa")
        sink("s the season of Darkness, it was the spring of hope, it was the wint")
        sink("er of despair.")
      }))
    Output
      ?> It was the best of times, it was the worst of times, it was the age of wisdom, it was the age of foolishness, it was the epoch of belief, it was the epoch of incredulity, it was the season of Light, it was the season of Darkness, it was the spring of hope, it was the winter of despair.

---

    Code
      cat(capture_sink(prefix = "", width = 70, function(sink) {
        sink("It was the best of times, it was the worst of times, it was the age ")
        sink("of wisdom, it was the age of foolishness, it was the epoch of belief")
        sink(", it was the epoch of incredulity, it was the season of Light, it wa")
        sink("s the season of Darkness, it was the spring of hope, it was the wint")
        sink("er of despair.")
      }))
    Output
      It was the best of times, it was the worst of times, it was the age 
      of wisdom, it was the age of foolishness, it was the epoch of belief,
      it was the epoch of incredulity, it was the season of Light, it was 
      the season of Darkness, it was the spring of hope, it was the winter 
      of despair.

---

    Code
      cat(capture_sink(prefix = "", width = 10000, function(sink) {
        sink("It was the best of times, it was the worst of times, it was the age ")
        sink("of wisdom, it was the age of foolishness, it was the epoch of belief")
        sink(", it was the epoch of incredulity, it was the season of Light, it wa")
        sink("s the season of Darkness, it was the spring of hope, it was the wint")
        sink("er of despair.")
      }))
    Output
      It was the best of times, it was the worst of times, it was the age of wisdom, it was the age of foolishness, it was the epoch of belief, it was the epoch of incredulity, it was the season of Light, it was the season of Darkness, it was the spring of hope, it was the winter of despair.

---

    Code
      cat(capture_sink(prefix = "Dickens> ", function(sink) {
        sink("It was the best of times,\nit was the worst of times,\n")
        sink("it was the age ")
        sink("of wisdom, it was the age of ")
        sink("foolishness,\nit was the epoch of belief")
        sink(", it was the epoch of incredulity, ")
        sink(paste(rep_len(" ", 200), collapse = ""))
        sink("it was the season of Light, it wa")
        sink("s the season of Darkness,\n\n\nit was the spring of hope,")
        sink(" it was the wint")
        sink("er of despair.")
      }))
    Output
      Dickens> It was the best of times,
      Dickens> it was the worst of times,
      Dickens> it was the age of wisdom, it was the age of 
      Dickens> foolishness,
      Dickens> it was the epoch of belief, it was the epoch of 
      Dickens> incredulity,
      Dickens> it was the season of Light, it was the season of 
      Dickens> Darkness,
      Dickens> 
      Dickens> 
      Dickens> it was the spring of hope, it was the winter of 
      Dickens> despair.

# handles errors

    Code
      chat$chat("What is 1 + 1?", echo = FALSE)
    Condition
      Error in `req_perform()`:
      ! HTTP 400 Bad Request.
      i Invalid input: expected number, received string
    Code
      chat$chat("What is 1 + 1?", echo = TRUE)
    Condition
      Error in `req_perform_connection()`:
      ! HTTP 400 Bad Request.
      i Invalid input: expected number, received string

# virtual_key is deprecated

    Code
      chat <- chat_portkey(model = "def", virtual_key = "abc")
    Condition
      Warning:
      The `virtual_key` argument of `chat_portkey()` is deprecated as of ellmer 0.4.0.
      i Please use the `model` argument instead.

# can handle errors

    Code
      chat$chat("Hi")
    Condition
      Error in `req_perform()`:
      ! HTTP 400 Bad Request.
      i Invalid model: doesnt-exist

# defaults are reported

    Code
      . <- chat_mistral()
    Message
      Using model = "mistral-large-latest".

# can handle errors

    Code
      chat$chat("Hi")
    Condition
      Error in `req_perform()`:
      ! HTTP 404 Not Found.
      i models/doesnt-exist is not found for API version v1beta, or is not supported for generateContent. Call ListModels to see the list of available models and their supported methods.

# defaults are reported

    Code
      . <- chat_google_gemini()
    Message
      Using model = "gemini-2.5-flash".

# can use images

    Code
      . <- chat$chat("What's in this image?", image_remote)
    Condition
      Error in `method(as_json, list(ellmer::ProviderGoogleGemini, ellmer::ContentImageRemote))`:
      ! Gemini doesn't support remote images

# Turn() with `role`

    Code
      Turn(role = "invalid")
    Condition
      Error in `Turn()`:
      ! `role` must be one of "user", "assistant", or "system", not "invalid".
    Code
      Turn(role = 42)
    Condition
      Error in `Turn()`:
      ! `role` must be a character vector, not the number 42.

# normalize_turns throws useful errors

    Code
      normalize_turns(1)
    Condition
      Error:
      ! `turns` must be an unnamed list or `NULL`, not the number 1.
    Code
      normalize_turns(list(1))
    Condition
      Error in `normalize_turns()`:
      ! Every element of `turns` must be a `turn`.
    Code
      normalize_turns(list(sys_msg, user_msg), 1)
    Condition
      Error in `normalize_turns()`:
      ! `system_prompt` must be a single string or `NULL`, not the number 1.
    Code
      normalize_turns(list(sys_msg, user_msg), "foo2")
    Condition
      Error:
      ! `system_prompt` and `turns[[1]]` can't contain conflicting system prompts.

# as_user_turn gives useful errors

    Code
      as_user_turn(list())
    Condition
      Error:
      ! `...` must contain at least one input.
    Code
      as_user_turn(list(x = 1))
    Condition
      Error:
      ! `...` must be unnamed.
    Code
      as_user_turn(1)
    Condition
      Error in `FUN()`:
      ! `...` must be made up strings or <content> objects, not the number 1.

# as_user_turns gives useful errors

    Code
      as_user_turns(1)
    Condition
      Error:
      ! `1` must be a list or prompt, not the number 1.
    Code
      as_user_turns(x)
    Condition
      Error in `FUN()`:
      ! `x[[1]]` must be made up strings or <content> objects, not the number 1.

# turns have a reasonable print method

    Code
      UserTurn("hello")
    Output
      <Turn: user>
      hello

# includes list of models in error message if `model` is missing

    Code
      chat_ollama()
    Condition
      Error in `chat_ollama()`:
      ! Must specify `model`.
      i Locally installed models: "llama3".

# checks that requested model is installed

    Code
      chat_ollama(model = "not-a-real-model")
    Condition
      Error in `chat_ollama()`:
      ! Model "not-a-real-model" is not installed locally.
      i Run `ollama pull not-a-real-model` in your terminal or `ollamar::pull("not-a-real-model")` in R to install the model.
      i See locally installed models with `ellmer::models_ollama()`.

# as_json specialised for Ollama

    Code
      as_json(stub, type_object(.additional_properties = TRUE))
    Condition
      Error in `method(as_json, list(ellmer::ProviderOllama, ellmer::TypeObject))`:
      ! `.additional_properties` not supported for Ollama.

# checks its inputs

    Code
      params(temperature = "x")
    Condition
      Error in `params()`:
      ! `temperature` must be a number or `NULL`, not the string "x".
    Code
      params(top_p = "x")
    Condition
      Error in `params()`:
      ! `top_p` must be a number or `NULL`, not the string "x".
    Code
      params(top_k = "x")
    Condition
      Error in `params()`:
      ! `top_k` must be a whole number or `NULL`, not the string "x".
    Code
      params(frequency_penalty = "x")
    Condition
      Error in `params()`:
      ! `frequency_penalty` must be a number or `NULL`, not the string "x".
    Code
      params(presence_penalty = "x")
    Condition
      Error in `params()`:
      ! `presence_penalty` must be a number or `NULL`, not the string "x".
    Code
      params(seed = "x")
    Condition
      Error in `params()`:
      ! `seed` must be a whole number or `NULL`, not the string "x".
    Code
      params(max_tokens = "x")
    Condition
      Error in `params()`:
      ! `max_tokens` must be a whole number or `NULL`, not the string "x".
    Code
      params(log_probs = 1)
    Condition
      Error in `params()`:
      ! `log_probs` must be `TRUE`, `FALSE`, or `NULL`, not the number 1.
    Code
      params(stop_sequences = 1)
    Condition
      Error in `params()`:
      ! `stop_sequences` must be a character vector or `NULL`, not the number 1.

# standardise_params warns about unknown args

    Code
      . <- standardise_params(test_params, provider_params)
    Condition
      Warning:
      Ignoring unsupported parameters: "top_p"

# errors if chat/provider/prompts don't match previous run

    Code
      batch_chat(chat, prompts, path)
    Condition
      Error in `batch_chat()`:
      ! provider, prompts, and user_turns don't match stored values.
      i Do you need to pick a different `path`?
      i Or set `ignore_hash = TRUE` to ignore this check?

---

    Code
      batch_chat_structured(chat, prompts, path, type = type_string())
    Condition
      Error in `batch_chat_structured()`:
      ! provider, prompts, and user_turns don't match stored values.
      i Do you need to pick a different `path`?
      i Or set `ignore_hash = TRUE` to ignore this check?

# can override hash check

    Code
      . <- batch_chat(chat, prompts, path, ignore_hash = TRUE)
    Condition
      Warning in `batch_chat()`:
      ! prompts and user_turns don't match stored values.

# informative error for bad inputs

    Code
      batch_chat(1)
    Condition
      Error in `batch_chat()`:
      ! `chat` must be a string or a <Chat> object, not the number 1.
    Code
      batch_chat(chat_ollama)
    Condition
      Error in `batch_chat()`:
      ! Batch requests are not currently supported by this provider.
    Code
      batch_chat(chat_openai, "a")
    Condition
      Error in `batch_chat()`:
      ! `prompts` must be a list or prompt, not the string "a".
    Code
      batch_chat(chat_openai, list("a"), path = 1)
    Condition
      Error in `batch_chat()`:
      ! `path` must be a single string, not the number 1.
    Code
      batch_chat(chat_openai, list("a"), path = "x", wait = 1)
    Condition
      Error in `batch_chat()`:
      ! `wait` must be `TRUE` or `FALSE`, not the number 1.

# useful error if no ContentJson

    Code
      extract_data(turn)
    Condition
      Error in `extract_data()`:
      ! Data extraction failed: no JSON responses found.

# warns if multiple ContentJson (and uses first)

    Code
      result <- extract_data(turn, type)
    Condition
      Warning:
      Found 3 JSON responses, using the first.

# defaults are reported

    Code
      . <- chat_azure_openai("endpoint", "deployment_id")
    Message
      Using api_version = "2024-10-21".

# Azure request headers are generated correctly

    Code
      str(req_get_headers(req, "reveal"))
    Output
      List of 1
       $ api-key: chr "key"

# service principal authentication requests look correct

    Code
      str(request_summary(req))
    Output
      List of 3
       $ url    : chr "https://login.microsoftonline.com/aaaa0a0a-bb1b-cc2c-dd3d-eeeeee4e4e4e/oauth2/v2.0/token"
       $ headers:List of 1
        ..$ Accept: chr "application/json"
       $ body   :List of 4
        ..$ grant_type   : 'AsIs' chr "client_credentials"
        ..$ scope        : 'AsIs' chr "https%3A%2F%2Fcognitiveservices.azure.com%2F.default"
        ..$ client_id    : 'AsIs' chr "id"
        ..$ client_secret: 'AsIs' chr "secret"

# checks recorded value types

    Code
      contents_replay(bad_names)
    Condition
      Error in `contents_replay()`:
      ! Expected the recorded object to be a list with at least names 'version', 'class', and 'props'.
    Code
      contents_replay(bad_version)
    Condition
      Error in `check_recorded()`:
      ! Unsupported version 2.
    Code
      contents_replay(bad_class)
    Condition
      Error in `contents_replay()`:
      ! Expected the recorded object to have a single $class name, containing `::` if the class is from a package.

# non-ellmer classes are not recorded/replayed by default

    Code
      contents_record(LocalClass())
    Condition
      Error in `contents_record()`:
      ! Cannot record or replay a <foo::LocalClass> object.
      i Only `ellmer::Content` or `ellmer::Turn` classes or subclasses are currently supported.
    Code
      contents_replay(recorded)
    Condition
      Error in `loadNamespace()`:
      ! there is no package called 'foo'

# local classes that extend ellmer classes can be replayed

    Code
      test_record_replay(test_content("hello world"))
    Condition
      Error in `contents_replay()`:
      ! Expected the object named `LocalContentText` to be an S7 class, not a function.

# replayed objects must be existing S7 classes

    Code
      contents_replay(doesnt_exist)
    Condition
      Error in `contents_replay()`:
      ! Unable to find the S7 class: `ellmer::Turn2`.
    Code
      contents_replay(not_s7)
    Condition
      Error in `contents_replay()`:
      ! Expected the object named `ellmer::chat_openai` to be an S7 class, not a function.

# roxygen2 comment extraction works

    Code
      extract_comments_and_signature(has_roxygen_comments)
    Output
      [1] "#' A function for foo-ing three numbers.\n#'\n#' @param x The first param\n#' @param y The second param\n#' @param z Take a guess\n#' @returns The result of x %foo% y %foo% z.\nfunction (x, y, z = pi - 3.14)  ..."

---

    Code
      extract_comments_and_signature(aliased_function)
    Output
      [1] "#' A function for foo-ing three numbers.\n#'\n#' @param x The first param\n#' @param y The second param\n#' @param z Take a guess\n#' @returns The result of x %foo% y %foo% z.\nfunction (x, y, z = pi - 3.14)  ..."

---

    Code
      extract_comments_and_signature(indented_comments)
    Output
      [1] "  #' A function for foo-ing three numbers.\n  #'\n  #' @param x The first param\n  #' @param y The second param\n  #' @param z Take a guess\n  #' @returns The result of x %foo% y %foo% z.\nfunction (x, y, z = pi - 3.14)  ..."

---

    Code
      extract_comments_and_signature(no_srcfile)
    Output
      [1] "  #' A function for foo-ing three numbers.\nfunction (a, b, c = pi - 3.14)  ..."

# basic signature extraction works

    Code
      extract_comments_and_signature(no_roxygen_comments)
    Output
      [1] "function (i, j, k = pi - 3.14)  ..."

# checks its inputs

    Code
      create_tool_def(print, chat = 1)
    Condition
      Error in `create_tool_def()`:
      ! `chat` must be a <Chat> object or `NULL`, not the number 1.

# prop_whole_number validates inputs

    Code
      check_prop()("x")
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be <double>, not <character>
    Code
      check_prop()(c(1, 2))
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be a whole number, not a double vector.
    Code
      check_prop()(1.5)
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be a whole number, not the number 1.5.
    Code
      check_prop(min = 1)(0)
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be at least 1, not 0.
    Code
      check_prop(max = -1)(0)
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be at most -1, not 0.
    Code
      check_prop()(NA_real_)
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must not be missing.

# prop_whole_decimal validates inputs

    Code
      check_prop()("x")
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be <double>, not <character>
    Code
      check_prop()(c(1, 2))
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be a whole number, not a double vector.
    Code
      check_prop()(1.5)
    Output
      <class>
       @ prop: num 1.5
    Code
      check_prop(min = 1)(0)
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be at least 1, not 0.
    Code
      check_prop(max = -1)(0)
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be at most -1, not 0.
    Code
      check_prop(max = -1)(Inf)
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must be finite, not Inf.
    Code
      check_prop()(NA_real_)
    Condition
      Error:
      ! <class> object properties are invalid:
      - @prop must not be missing.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/turns.R
\name{Turn}
\alias{Turn}
\alias{UserTurn}
\alias{SystemTurn}
\alias{AssistantTurn}
\alias{AssistantPartialTurn}
\title{A user, assistant, or system turn}
\usage{
Turn(role = NULL, contents = list(), tokens = NULL)

UserTurn(contents = list())

SystemTurn(contents = list())

AssistantTurn(
  contents = list(),
  json = list(),
  tokens = c(NA_real_, NA_real_, NA_real_),
  cost = NA_real_,
  duration = NA_real_
)

AssistantPartialTurn(
  contents = list(),
  json = list(),
  tokens = c(NA_real_, NA_real_, NA_real_),
  cost = NA_real_,
  duration = NA_real_,
  reason = "interrupted"
)
}
\arguments{
\item{role}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
For system, user and assistant turns, use \code{SystemTurn()}, \code{UserTurn()}, and
\code{AssistantTurn()}, respectively.}

\item{contents}{A list of \link{Content} objects.}

\item{tokens}{A numeric vector of length 3 representing the number of
input tokens (uncached), output tokens, and input tokens (cached)
used in this turn.}

\item{json}{The serialized JSON corresponding to the underlying data of
the turns. This is useful if there's information returned by the provider
that ellmer doesn't otherwise expose.}

\item{cost}{The cost of the turn in dollars.}

\item{duration}{The duration of the request in seconds.}

\item{reason}{A character string describing why the turn was interrupted.
Defaults to \code{"interrupted"}.}
}
\value{
An S7 \code{Turn} object

An S7 \code{AssistantTurn} object

An S7 \code{AssistantPartialTurn} object
}
\description{
Every conversation with a chatbot consists of pairs of user and assistant
turns, corresponding to an HTTP request and response. These turns are
represented by the \code{Turn} object, which contains a list of \link{Content}s representing
the individual messages within the turn. These might be text, images, tool
requests (assistant only), or tool responses (user only).

\code{UserTurn}, \code{AssistantTurn}, and \code{SystemTurn} are specialized subclasses
of \code{Turn} for different types of conversation turns. \code{AssistantTurn} includes
additional metadata about the API response.

Note that a call to \verb{$chat()} and related functions may result in multiple
user-assistant turn cycles. For example, if you have registered tools,
ellmer will automatically handle the tool calling loop, which may result in
any number of additional cycles. Learn more about tool calling in
\code{vignette("tool-calling")}.
}
\examples{
UserTurn(list(ContentText("Hello, world!")))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-any.R
\name{chat}
\alias{chat}
\title{Chat with any provider}
\usage{
chat(
  name,
  ...,
  system_prompt = NULL,
  params = NULL,
  echo = c("none", "output", "all")
)
}
\arguments{
\item{name}{Provider (and optionally model) name in the form
\code{"provider/model"} or \code{"provider"} (which will use the default model
for that provider).}

\item{...}{Arguments passed to the provider function.}

\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}
}
\description{
This is a generic interface to all the other \code{chat_} functions that allow
to you pick the provider and the model with a simple string.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{has_credentials}
\alias{has_credentials}
\title{Are credentials avaiable?}
\usage{
has_credentials(provider)
}
\arguments{
\item{provider}{Provider name.}
}
\description{
Used for examples/testing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-ollama.R
\name{chat_ollama}
\alias{chat_ollama}
\alias{models_ollama}
\title{Chat with a local Ollama model}
\usage{
chat_ollama(
  system_prompt = NULL,
  base_url = Sys.getenv("OLLAMA_BASE_URL", "http://localhost:11434"),
  model,
  params = NULL,
  api_args = list(),
  echo = NULL,
  api_key = NULL,
  credentials = NULL,
  api_headers = character()
)

models_ollama(base_url = "http://localhost:11434", credentials = NULL)
}
\arguments{
\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{base_url}{The base URL to the API endpoint.}

\item{model}{The model to use for the chat.
Use \code{models_ollama()} to see all options.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{Ollama doesn't require credentials for local usage and in most
cases you do not need to provide \code{credentials}.

However, if you're accessing an Ollama instance hosted behind a reverse
proxy or secured endpoint that enforces bearer‐token authentication, you
can set the \code{OLLAMA_API_KEY} environment variable or provide a callback
function to \code{credentials}.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}
}
\value{
A \link{Chat} object.
}
\description{
To use \code{chat_ollama()} first download and install
\href{https://ollama.com}{Ollama}. Then install some models either from the
command line (e.g. with \verb{ollama pull llama3.1}) or within R using
{\href{https://hauselin.github.io/ollama-r/}{ollamar}} (e.g.
\code{ollamar::pull("llama3.1")}).

Built on top of \code{\link[=chat_openai_compatible]{chat_openai_compatible()}}.
\subsection{Known limitations}{
\itemize{
\item Tool calling is not supported with streaming (i.e. when \code{echo} is
\code{"text"} or \code{"all"})
\item Models can only use 2048 input tokens, and there's no way
to get them to use more, except by creating a custom model with a
different default.
\item Tool calling generally seems quite weak, at least with the models I have
tried it with.
}
}
}
\examples{
\dontrun{
chat <- chat_ollama(model = "llama3.2")
chat$chat("Tell me three jokes about statisticians")
}
}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-openai-compatible.R
\name{chat_openai_compatible}
\alias{chat_openai_compatible}
\title{Chat with an OpenAI-compatible model}
\usage{
chat_openai_compatible(
  base_url,
  name = "OpenAI-compatible",
  system_prompt = NULL,
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  api_headers = character(),
  preserve_thinking = FALSE,
  echo = c("none", "output", "all")
)
}
\arguments{
\item{base_url}{The base URL to the endpoint. This parameter is \strong{required}
since there is no default for OpenAI-compatible APIs.}

\item{name}{The name of the provider; this is shown in \code{\link[=token_usage]{token_usage()}} and
is used to compute costs.}

\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{Credentials to use for authentication. If not provided,
will attempt to use the \code{OPENAI_API_KEY} environment variable.}

\item{model}{The model to use for chat. No default; depends on your provider.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}

\item{preserve_thinking}{If \code{TRUE}, reasoning content returned by the model
is included when sending conversation history back to the API. If \code{FALSE}
(the default), reasoning content is still captured in the turn but dropped
from subsequent requests. Set to \code{TRUE} if your provider requires or
benefits from seeing prior reasoning in multi-turn conversations.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}
}
\value{
A \link{Chat} object.
}
\description{
This function is for use with OpenAI-compatible APIs, also known as the
\strong{chat completions} API. If you want to use OpenAI itself, we recommend
\code{\link[=chat_openai]{chat_openai()}}, which uses the newer \strong{responses} API.

Many providers offer OpenAI-compatible APIs, including:
\itemize{
\item \href{https://ollama.com/}{Ollama} for local models
\item \href{https://docs.vllm.ai/}{vLLM} for self-hosted models
\item Various cloud providers with OpenAI-compatible endpoints
}
}
\examples{
\dontrun{
# Example with Ollama (requires Ollama running locally)
chat <- chat_openai_compatible(
  base_url = "http://localhost:11434/v1",
  model = "llama2"
)
chat$chat("What is the difference between a tibble and a data frame?")
}
}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-openai-tools.R
\name{openai_tool_web_search}
\alias{openai_tool_web_search}
\title{OpenAI web search tool}
\usage{
openai_tool_web_search(
  allowed_domains = NULL,
  user_location = NULL,
  external_web_access = TRUE
)
}
\arguments{
\item{allowed_domains}{Character vector. Restrict searches to specific domains
(e.g., \code{c("nytimes.com", "bbc.com")}). Maximum 20 domains. URLs will be
automatically cleaned (http/https prefixes removed).}

\item{user_location}{List with optional elements: \code{country} (2-letter ISO code),
\code{city}, \code{region}, and \code{timezone} (IANA timezone) to localize search results.}

\item{external_web_access}{Logical. Whether to allow live internet access
(\code{TRUE}, default) or use only cached/indexed results (\code{FALSE}).}
}
\description{
Enables OpenAI models to search the web for up-to-date information. The search
behavior varies by model: non-reasoning models perform simple searches, while
reasoning models can perform agentic, iterative searches.

Learn more at \url{https://platform.openai.com/docs/guides/tools-web-search}
}
\examples{
\dontrun{
chat <- chat_openai()
chat$register_tool(openai_tool_web_search())
chat$chat("Very briefly summarise the top 3 news stories of the day")
chat$chat("Of those stories, which one do you think was the most interesting?")
}
}
\seealso{
Other built-in tools:
\code{\link[=claude_tool_web_fetch]{claude_tool_web_fetch()}},
\code{\link[=claude_tool_web_search]{claude_tool_web_search()}},
\code{\link[=google_tool_web_fetch]{google_tool_web_fetch()}},
\code{\link[=google_tool_web_search]{google_tool_web_search()}}
}
\concept{built-in tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel-chat.R
\name{parallel_chat}
\alias{parallel_chat}
\alias{parallel_chat_text}
\alias{parallel_chat_structured}
\title{Submit multiple chats in parallel}
\usage{
parallel_chat(
  chat,
  prompts,
  max_active = 10,
  rpm = 500,
  on_error = c("return", "continue", "stop")
)

parallel_chat_text(
  chat,
  prompts,
  max_active = 10,
  rpm = 500,
  on_error = c("return", "continue", "stop")
)

parallel_chat_structured(
  chat,
  prompts,
  type,
  convert = TRUE,
  include_tokens = FALSE,
  include_cost = FALSE,
  max_active = 10,
  rpm = 500,
  on_error = c("return", "continue", "stop")
)
}
\arguments{
\item{chat}{A chat object created by a \code{chat_} function, or a
string passed to \code{\link[=chat]{chat()}}.}

\item{prompts}{A vector created by \code{\link[=interpolate]{interpolate()}} or a list
of character vectors.}

\item{max_active}{The maximum number of simultaneous requests to send.

For \code{\link[=chat_anthropic]{chat_anthropic()}}, note that the number of active connections is
limited primarily by the output tokens per minute limit (OTPM) which is
estimated from the \code{max_tokens} parameter, which defaults to 4096. That
means if your usage tier limits you to 16,000 OTPM, you should either set
\code{max_active = 4} (16,000 / 4096) to decrease the number of active
connections or use \code{\link[=params]{params()}} in \code{chat_anthropic()} to decrease
\code{max_tokens}.}

\item{rpm}{Maximum number of requests per minute.}

\item{on_error}{What to do when a request fails. One of:
\itemize{
\item \code{"return"} (the default): stop processing new requests, wait for
in flight requests to finish, then return.
\item \code{"continue"}: keep going, performing every request.
\item \code{"stop"}: stop processing and throw an error.
}}

\item{type}{A type specification for the extracted data. Should be
created with a \code{\link[=type_boolean]{type_()}} function.}

\item{convert}{If \code{TRUE}, automatically convert from JSON lists to R
data types using the schema. This typically works best when \code{type} is
\code{\link[=type_object]{type_object()}} as this will give you a data frame with one column for
each property. If \code{FALSE}, returns a list.}

\item{include_tokens}{If \code{TRUE}, and the result is a data frame, will
add \code{input_tokens} and \code{output_tokens} columns giving the total input
and output tokens for each prompt.}

\item{include_cost}{If \code{TRUE}, and the result is a data frame, will
add \code{cost} column giving the cost of each prompt.}
}
\value{
For \code{parallel_chat()}, a list with one element for each prompt. Each element
is either a \link{Chat} object (if successful), a \code{NULL} (if the request wasn't
performed) or an error object (if it failed).

For \code{parallel_chat_text()}, a character vector with one element for each
prompt. Requests that weren't succesful get an \code{NA}.

For \code{parallel_chat_structured()}, a single structured data object with one
element for each prompt. Typically, when \code{type} is an object, this will
be a tibble with one row for each prompt, and one column for each
property. If the output is a data frame, and some requests error,
an \code{.error} column will be added with the error objects.
}
\description{
If you have multiple prompts, you can submit them in parallel. This is
typically considerably faster than submitting them in sequence, especially
with Gemini and OpenAI.

If you're using \code{\link[=chat_openai]{chat_openai()}} or \code{\link[=chat_anthropic]{chat_anthropic()}} and you're willing
to wait longer, you might want to use \code{\link[=batch_chat]{batch_chat()}} instead, as it comes
with a 50\% discount in return for taking up to 24 hours.
}
\examples{
\dontshow{ellmer:::vcr_example_start("parallel_chat")}
chat <- chat_openai()

# Chat ----------------------------------------------------------------------
country <- c("Canada", "New Zealand", "Jamaica", "United States")
prompts <- interpolate("What's the capital of {{country}}?")
parallel_chat(chat, prompts)

# Structured data -----------------------------------------------------------
prompts <- list(
  "I go by Alex. 42 years on this planet and counting.",
  "Pleased to meet you! I'm Jamal, age 27.",
  "They call me Li Wei. Nineteen years young.",
  "Fatima here. Just celebrated my 35th birthday last week.",
  "The name's Robert - 51 years old and proud of it.",
  "Kwame here - just hit the big 5-0 this year."
)
type_person <- type_object(name = type_string(), age = type_number())
parallel_chat_structured(chat, prompts, type_person)
\dontshow{ellmer:::vcr_example_end()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-claude-tools.R
\name{claude_tool_web_search}
\alias{claude_tool_web_search}
\title{Claude web search tool}
\usage{
claude_tool_web_search(
  max_uses = NULL,
  allowed_domains = NULL,
  blocked_domains = NULL,
  user_location = NULL
)
}
\arguments{
\item{max_uses}{Integer. Maximum number of searches allowed per request.}

\item{allowed_domains}{Character vector. Restrict searches to specific domains
(e.g., \code{c("nytimes.com", "bbc.com")}). Cannot be used with \code{blocked_domains}.}

\item{blocked_domains}{Character vector. Exclude specific domains from searches.
Cannot be used with \code{allowed_domains}.}

\item{user_location}{List with optional elements: \code{country} (2-letter code),
\code{city}, \code{region}, and \code{timezone} (IANA timezone) to localize search results.}
}
\description{
Enables Claude to search the web for up-to-date information. Your organization
administrator must enable web search in the Anthropic Console before using
this tool, as it costs extra ($10 per 1,000 tokens at time of writing).

Learn more in \url{https://docs.claude.com/en/docs/agents-and-tools/tool-use/web-search-tool}.
}
\examples{
\dontrun{
chat <- chat_claude()
chat$register_tool(claude_tool_web_search())
chat$chat("What was in the news today?")
chat$chat("What's the biggest news in the economy?")
}
}
\seealso{
Other built-in tools:
\code{\link[=claude_tool_web_fetch]{claude_tool_web_fetch()}},
\code{\link[=google_tool_web_fetch]{google_tool_web_fetch()}},
\code{\link[=google_tool_web_search]{google_tool_web_search()}},
\code{\link[=openai_tool_web_search]{openai_tool_web_search()}}
}
\concept{built-in tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider.R
\name{Provider}
\alias{Provider}
\title{A chatbot provider}
\usage{
Provider(
  name = stop("Required"),
  model = stop("Required"),
  base_url = stop("Required"),
  params = list(),
  extra_args = list(),
  extra_headers = character(0),
  credentials = function() NULL
)
}
\arguments{
\item{name}{Name of the provider.}

\item{model}{Name of the model.}

\item{base_url}{The base URL for the API.}

\item{params}{A list of standard parameters created by \code{\link[=params]{params()}}.}

\item{extra_args}{Arbitrary extra arguments to be included in the request body.}

\item{extra_headers}{Arbitrary extra headers to be added to the request.}

\item{credentials}{A zero-argument function that returns the credentials to use
for authentication. Can either return a string, representing an API key,
or a named list of headers.}
}
\value{
An S7 Provider object.
}
\description{
A Provider captures the details of one chatbot service/API. This captures
how the API works, not the details of the underlying large language model.
Different providers might offer the same (open source) model behind a
different API.
}
\details{
To add support for a new backend, you will need to subclass \code{Provider}
(adding any additional fields that your provider needs) and then implement
the various generics that control the behavior of each provider.
}
\examples{
Provider(
  name = "CoolModels",
  model = "my_model",
  base_url = "https://cool-models.com"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools-def.R
\name{tool}
\alias{tool}
\alias{ToolDef}
\title{Define a tool}
\usage{
tool(
  fun,
  description,
  ...,
  arguments = list(),
  name = NULL,
  convert = TRUE,
  annotations = list(),
  .name = deprecated(),
  .description = deprecated(),
  .convert = deprecated(),
  .annotations = deprecated()
)
}
\arguments{
\item{fun}{The function to be invoked when the tool is called. The return
value of the function is sent back to the chatbot.

Expert users can customize the tool result by returning a
\link{ContentToolResult} object.}

\item{description}{A detailed description of what the function does.
Generally, the more information that you can provide here, the better.}

\item{...}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{arguments} instead.}

\item{arguments}{A named list that defines the arguments accepted by the
function. Each element should be created by a \code{\link[=type_boolean]{type_*()}}
function. Use \code{\link[=type_ignore]{type_ignore()}} if you don't want the LLM to provide that
argument (e.g., because the R function has a suitable default value).}

\item{name}{The name of the function. This can be omitted if \code{fun} is an
existing function (i.e. not defined inline).}

\item{convert}{Should JSON inputs be automatically convert to their
R data type equivalents? Defaults to \code{TRUE}.}

\item{annotations}{Additional properties that describe the tool and its
behavior. Usually created by \code{\link[=tool_annotations]{tool_annotations()}}, where you can find a
description of the annotation properties recommended by the \href{https://modelcontextprotocol.io/introduction}{Model Context Protocol}.}

\item{.name, .description, .convert, .annotations}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Please switch to the non-prefixed
equivalents.}
}
\value{
An S7 \code{ToolDef} object.
}
\description{
Annotate a function for use in tool calls, by providing a name, description,
and type definition for the arguments.

Learn more in \code{vignette("tool-calling")}.
}
\section{ellmer 0.3.0}{
In ellmer 0.3.0, the definition of the \code{tool()} function changed quite
a bit. To make it easier to update old versions, you can use an LLM with
the following system prompt

\if{html}{\out{<div class="sourceCode">}}\preformatted{Help the user convert an ellmer 0.2.0 and earlier tool definition into a
ellmer 0.3.0 tool definition. Here's what changed:

* All arguments, apart from the first, should be named, and the argument
  names no longer use `.` prefixes. The argument order should be function,
  name (as a string), description, then arguments, then anything

* Previously `arguments` was passed as `...`, so all type specifications
  should now be moved into a named list and passed to the `arguments`
  argument. It can be omitted if the function has no arguments.

```R
# old
tool(
  add,
  "Add two numbers together"
  x = type_number(),
  y = type_number()
)

# new
tool(
  add,
  name = "add",
  description = "Add two numbers together",
  arguments = list(
    x = type_number(),
    y = type_number()
  )
)
```

Don't respond; just let the user provide function calls to convert.
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{ellmer:::vcr_example_start("tool")}
\dontshow{set.seed(1014)}
# First define the metadata that the model uses to figure out when to
# call the tool
tool_rnorm <- tool(
  rnorm,
  description = "Draw numbers from a random normal distribution",
  arguments = list(
    n = type_integer("The number of observations. Must be a positive integer."),
    mean = type_number("The mean value of the distribution."),
    sd = type_number("The standard deviation of the distribution. Must be a non-negative number.")
  )
)
tool_rnorm(n = 5, mean = 0, sd = 1)

chat <- chat_openai()
# Then register it
chat$register_tool(tool_rnorm)

# Then ask a question that needs it.
chat$chat("Give me five numbers from a random normal distribution.")

# Look at the chat history to see how tool calling works:
chat
# Assistant sends a tool request which is evaluated locally and
# results are sent back in a tool result.

\dontshow{ellmer:::vcr_example_end()}
}
\seealso{
Other tool calling helpers:
\code{\link[=tool_annotations]{tool_annotations()}},
\code{\link[=tool_reject]{tool_reject()}}
}
\concept{tool calling helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-claude-files.R
\name{claude_file_upload}
\alias{claude_file_upload}
\alias{claude_file_list}
\alias{claude_file_get}
\alias{claude_file_download}
\alias{claude_file_delete}
\title{Upload, downloand, and manage files for Claude}
\usage{
claude_file_upload(
  path,
  base_url = "https://api.anthropic.com/v1/",
  beta_headers = "files-api-2025-04-14",
  credentials = NULL
)

claude_file_list(
  base_url = "https://api.anthropic.com/v1/",
  credentials = NULL,
  beta_headers = "files-api-2025-04-14"
)

claude_file_get(
  file_id,
  base_url = "https://api.anthropic.com/v1/",
  credentials = NULL,
  beta_headers = "files-api-2025-04-14"
)

claude_file_download(
  file_id,
  path,
  base_url = "https://api.anthropic.com/v1/",
  credentials = NULL,
  beta_headers = "files-api-2025-04-14"
)

claude_file_delete(
  file_id,
  base_url = "https://api.anthropic.com/v1/",
  credentials = NULL,
  beta_headers = "files-api-2025-04-14"
)
}
\arguments{
\item{path}{Path to download the file to.}

\item{base_url}{The base URL to the endpoint; the default is Claude's
public API.}

\item{beta_headers}{Beta headers to use for the request. Defaults to
\code{files-api-2025-04-14}.}

\item{credentials}{Override the default credentials. You generally should not need this argument; instead set the \code{ANTHROPIC_API_KEY} environment variable. The best place to set this is in \code{.Renviron},
which you can easily edit by calling \code{usethis::edit_r_environ()}.

If you do need additional control, this argument takes a zero-argument function that returns either a string (the API key), or a named list (added as additional headers to every request).}

\item{file_id}{ID of the file to get information about, download, or delete.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Use the beta Files API to upload files to and manage files in Claude.
This is currently experimental because the API is in beta and may change.
Note that you need \code{beta-headers = "files-api-2025-04-14"} to use the API.

Claude offers 100GB of file storage per organization, with each file
having a maximum size of 500MB. For more details see
\url{https://docs.claude.com/en/docs/build-with-claude/files}
\itemize{
\item \code{claude_file_upload()} uploads a file and returns an object that
you can use in chat.
\item \code{claude_file_list()} lists all uploaded files.
\item \code{claude_file_get()} returns an object for an previously uploaded file.
\item \code{claude_file_download()} downloads the file with the given ID. Note
that you can only download files created by skills or the code execution
tool.
\item \code{claude_file_delete()} deletes the file with the given ID.
}
}
\examples{
\dontrun{
file <- claude_file_upload("path/to/file.pdf")
chat <- chat_anthropic(beta_headers = "files-api-2025-04-14")
chat$chat("Please summarize the document.", file)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-chat.R
\name{batch_chat}
\alias{batch_chat}
\alias{batch_chat_text}
\alias{batch_chat_structured}
\alias{batch_chat_completed}
\title{Submit multiple chats in one batch}
\usage{
batch_chat(chat, prompts, path, wait = TRUE, ignore_hash = FALSE)

batch_chat_text(chat, prompts, path, wait = TRUE, ignore_hash = FALSE)

batch_chat_structured(
  chat,
  prompts,
  path,
  type,
  wait = TRUE,
  ignore_hash = FALSE,
  convert = TRUE,
  include_tokens = FALSE,
  include_cost = FALSE
)

batch_chat_completed(chat, prompts, path)
}
\arguments{
\item{chat}{A chat object created by a \code{chat_} function, or a
string passed to \code{\link[=chat]{chat()}}.}

\item{prompts}{A vector created by \code{\link[=interpolate]{interpolate()}} or a list
of character vectors.}

\item{path}{Path to file (with \code{.json} extension) to store state.

The file records a hash of the provider, the prompts, and the existing
chat turns. If you attempt to reuse the same file with any of these being
different, you'll get an error.}

\item{wait}{If \code{TRUE}, will wait for batch to complete. If \code{FALSE},
it will return \code{NULL} if the batch is not complete, and you can retrieve
the results later by re-running \code{batch_chat()} when
\code{batch_chat_completed()} is \code{TRUE}.}

\item{ignore_hash}{If \code{TRUE}, will only warn rather than error when the hash
doesn't match. You can use this if ellmer has changed the hash structure
and you're confident that you're reusing the same inputs.}

\item{type}{A type specification for the extracted data. Should be
created with a \code{\link[=type_boolean]{type_()}} function.}

\item{convert}{If \code{TRUE}, automatically convert from JSON lists to R
data types using the schema. This typically works best when \code{type} is
\code{\link[=type_object]{type_object()}} as this will give you a data frame with one column for
each property. If \code{FALSE}, returns a list.}

\item{include_tokens}{If \code{TRUE}, and the result is a data frame, will
add \code{input_tokens} and \code{output_tokens} columns giving the total input
and output tokens for each prompt.}

\item{include_cost}{If \code{TRUE}, and the result is a data frame, will
add \code{cost} column giving the cost of each prompt.}
}
\value{
For \code{batch_chat()}, a list of \link{Chat} objects, one for each prompt.
For \code{batch_chat_test()}, a character vector of text responses.
For \code{batch_chat_structured()}, a single structured data object with one
element for each prompt. Typically, when \code{type} is an object, this will
will be a data frame with one row for each prompt, and one column for each
property.

For any of the aboves, will return \code{NULL} if \code{wait = FALSE} and the job
is not complete.
}
\description{
\code{batch_chat()} and \code{batch_chat_structured()} currently only work with
\code{\link[=chat_openai]{chat_openai()}} and \code{\link[=chat_anthropic]{chat_anthropic()}}. They use the
\href{https://platform.openai.com/docs/guides/batch}{OpenAI} and
\href{https://docs.claude.com/en/docs/build-with-claude/batch-processing}{Anthropic}
batch APIs which allow you to submit multiple requests simultaneously.
The results can take up to 24 hours to complete, but in return you pay 50\%
less than usual (but note that ellmer doesn't include this discount in
its pricing metadata). If you want to get results back more quickly, or
you're working with a different provider, you may want to use
\code{\link[=parallel_chat]{parallel_chat()}} instead.

Since batched requests can take a long time to complete, \code{batch_chat()}
requires a file path that is used to store information about the batch so
you never lose any work. You can either set \code{wait = FALSE} or simply
interrupt the waiting process, then later, either call \code{batch_chat()} to
resume where you left off or call \code{batch_chat_completed()} to see if the
results are ready to retrieve. \code{batch_chat()} will store the chat responses
in this file, so you can either keep it around to cache the results,
or delete it to free up disk space.

This API is marked as experimental since I don't yet know how to handle
errors in the most helpful way. Fortunately they don't seem to be common,
but if you have ideas, please let me know!
}
\examples{
\dontshow{if (has_credentials("openai")) withAutoprint(\{ # examplesIf}
chat <- chat_openai(model = "gpt-4.1-nano")

# Chat ----------------------------------------------------------------------

prompts <- interpolate("What do people from {{state.name}} bring to a potluck dinner?")
\dontrun{
chats <- batch_chat(chat, prompts, path = "potluck.json")
chats
}

# Structured data -----------------------------------------------------------
prompts <- list(
  "I go by Alex. 42 years on this planet and counting.",
  "Pleased to meet you! I'm Jamal, age 27.",
  "They call me Li Wei. Nineteen years young.",
  "Fatima here. Just celebrated my 35th birthday last week.",
  "The name's Robert - 51 years old and proud of it.",
  "Kwame here - just hit the big 5-0 this year."
)
type_person <- type_object(name = type_string(), age = type_number())
\dontrun{
data <- batch_chat_structured(
  chat = chat,
  prompts = prompts,
  path = "people-data.json",
  type = type_person
)
data
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-google-tools.R
\name{google_tool_web_search}
\alias{google_tool_web_search}
\title{Google web search (grounding) tool}
\usage{
google_tool_web_search()
}
\description{
Enables Gemini models to search the web for up-to-date information and ground
responses with citations to sources. The model automatically decides when
(and how) to search the web based on your prompt. Search results are
incorporated into the response with grounding metadata including source
URLs and titles.

Learn more in \url{https://ai.google.dev/gemini-api/docs/google-search}.
}
\examples{
\dontrun{
chat <- chat_google_gemini()
chat$register_tool(google_tool_web_search())
chat$chat("What was in the news today?")
chat$chat("What's the biggest news in the economy?")
}
}
\seealso{
Other built-in tools:
\code{\link[=claude_tool_web_fetch]{claude_tool_web_fetch()}},
\code{\link[=claude_tool_web_search]{claude_tool_web_search()}},
\code{\link[=google_tool_web_fetch]{google_tool_web_fetch()}},
\code{\link[=openai_tool_web_search]{openai_tool_web_search()}}
}
\concept{built-in tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{Content}
\alias{Content}
\alias{ContentText}
\alias{ContentImage}
\alias{ContentImageRemote}
\alias{ContentImageInline}
\alias{ContentToolRequest}
\alias{ContentToolResult}
\alias{ContentThinking}
\alias{ContentPDF}
\title{Content types received from and sent to a chatbot}
\usage{
Content()

ContentText(text = stop("Required"))

ContentImage()

ContentImageRemote(url = stop("Required"), detail = "")

ContentImageInline(type = stop("Required"), data = NULL)

ContentToolRequest(
  id = stop("Required"),
  name = stop("Required"),
  arguments = list(),
  tool = NULL,
  extra = list()
)

ContentToolResult(value = NULL, error = NULL, extra = list(), request = NULL)

ContentThinking(thinking = stop("Required"), extra = list())

ContentPDF(
  type = stop("Required"),
  data = stop("Required"),
  filename = stop("Required")
)
}
\arguments{
\item{text}{A single string.}

\item{url}{URL to a remote image.}

\item{detail}{Not currently used.}

\item{type}{MIME type of the image.}

\item{data}{Base64 encoded image data.}

\item{id}{Tool call id (used to associate a request and a result).
Automatically managed by \pkg{ellmer}.}

\item{name}{Function name}

\item{arguments}{Named list of arguments to call the function with.}

\item{tool}{ellmer automatically matches a tool request to the tools defined
for the chatbot. If \code{NULL}, the request did not match a defined tool.}

\item{extra}{Additional data.}

\item{value}{The results of calling the tool function, if it succeeded.}

\item{error}{The error message, as a string, or the error condition thrown
as a result of a failure when calling the tool function. Must be \code{NULL}
when the tool call is successful.}

\item{request}{The \link{ContentToolRequest} associated with the tool result,
automatically added by \pkg{ellmer} when evaluating the tool call.}

\item{thinking}{The text of the thinking output.}

\item{filename}{File name, used to identify the PDF.}
}
\value{
S7 objects that all inherit from \code{Content}
}
\description{
Use these functions if you're writing a package that extends ellmer and need
to customise methods for various types of content. For normal use, see
\code{\link[=content_image_url]{content_image_url()}} and friends.

ellmer abstracts away differences in the way that different \link{Provider}s
represent various types of content, allowing you to more easily write
code that works with any chatbot. This set of classes represents types of
content that can be either sent to and received from a provider:
\itemize{
\item \code{ContentText}: simple text (often in markdown format). This is the only
type of content that can be streamed live as it's received.
\item \code{ContentImageRemote} and \code{ContentImageInline}: images, either as a pointer
to a remote URL or included inline in the object. See
\code{\link[=content_image_file]{content_image_file()}} and friends for convenient ways to construct these
objects.
\item \code{ContentToolRequest}: a request to perform a tool call (sent by the
assistant).
\item \code{ContentToolResult}: the result of calling the tool (sent by the user).
This object is automatically created from the value returned by calling the
\code{\link[=tool]{tool()}} function. Alternatively, expert users can return a
\code{ContentToolResult} from a \code{\link[=tool]{tool()}} function to include additional data or
to customize the display of the result.
}
}
\examples{
Content()
ContentText("Tell me a joke")
ContentImageRemote("https://www.r-project.org/Rlogo.png")
ContentToolRequest(id = "abc", name = "mean", arguments = list(x = 1:5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content-replay.R
\name{contents_record}
\alias{contents_record}
\alias{contents_replay}
\title{Record and replay content}
\usage{
contents_record(x)

contents_replay(x, tools = list(), .envir = parent.frame())
}
\arguments{
\item{x}{A \link{Turn} or \link{Content} object to serialize; or a serialized object
to replay.}

\item{tools}{A named list of tools}

\item{.envir}{The environment in which to look for class definitions. Used
when the recorded objects include classes that extend \link{Turn} or
\link{Content} but are not from the \pkg{ellmer} package itself.}
}
\description{
These generic functions can be use to convert \link{Turn}/\link{Content} objects
into easily serializable representations (i.e. lists and atomic vectors).
\itemize{
\item \code{contents_record()} accepts a \link{Turn} or \link{Content} and return a simple list.
\item \code{contents_replay()} takes the output of \code{contents_record()} and returns
a \link{Turn} or \link{Content} object.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-google-upload.R
\name{google_upload}
\alias{google_upload}
\title{Upload a file to gemini}
\usage{
google_upload(
  path,
  base_url = "https://generativelanguage.googleapis.com/",
  api_key = NULL,
  credentials = NULL,
  mime_type = NULL
)
}
\arguments{
\item{path}{Path to a file to upload.}

\item{base_url}{The base URL to the API endpoint.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{A function that returns a list of authentication headers
or \code{NULL}, the default, to use ambient credentials. See above for details.}

\item{mime_type}{Optionally, specify the mime type of the file.
If not specified, will be guesses from the file extension.}
}
\value{
A \verb{<ContentUploaded>} object that can be passed to \verb{$chat()}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This function uploads a file then waits for Gemini to finish processing it
so that you can immediately use it in a prompt. It's experimental because
it's currently Gemini specific, and we expect other providers to evolve
similar feature in the future.

Uploaded files are automatically deleted after 2 days. Each file must be
less than 2 GB and you can upload a total of 20 GB. ellmer doesn't currently
provide a way to delete files early; please
\href{https://github.com/tidyverse/ellmer/issues}{file an issue} if this would
be useful for you.
}
\examples{
\dontrun{
file <- google_upload("path/to/file.pdf")

chat <- chat_google_gemini()
chat$chat(file, "Give me a three paragraph summary of this PDF")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-azure.R
\name{chat_azure_openai}
\alias{chat_azure_openai}
\title{Chat with a model hosted on Azure OpenAI}
\usage{
chat_azure_openai(
  endpoint = azure_endpoint(),
  model,
  params = NULL,
  api_version = NULL,
  system_prompt = NULL,
  api_key = NULL,
  credentials = NULL,
  api_args = list(),
  echo = c("none", "output", "all"),
  api_headers = character(),
  deployment_id = deprecated()
)
}
\arguments{
\item{endpoint}{Azure OpenAI endpoint url with protocol and hostname, i.e.
\verb{https://\{your-resource-name\}.openai.azure.com}. Defaults to using the
value of the \code{AZURE_OPENAI_ENDPOINT} environment variable.}

\item{model}{The \strong{deployment id} for the model you want to use.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_version}{The API version to use.}

\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{Override the default credentials. You generally should not need this argument; instead set the \code{AZURE_OPENAI_API_KEY} environment variable. The best place to set this is in \code{.Renviron},
which you can easily edit by calling \code{usethis::edit_r_environ()}.

If you do need additional control, this argument takes a zero-argument function that returns either a string (the API key), or a named list (added as additional headers to every request).}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}

\item{deployment_id}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{model} instead.}
}
\value{
A \link{Chat} object.
}
\description{
The \href{https://azure.microsoft.com/en-us/products/ai-services/openai-service}{Azure OpenAI server}
hosts a number of open source models as well as proprietary models
from OpenAI.

Built on top of \code{\link[=chat_openai_compatible]{chat_openai_compatible()}}.
\subsection{Authentication}{

\code{chat_azure_openai()} supports API keys and the \code{credentials} parameter, but
it also makes use of:
\itemize{
\item Azure service principals (when the \code{AZURE_TENANT_ID}, \code{AZURE_CLIENT_ID},
and \code{AZURE_CLIENT_SECRET} environment variables are set).
\item Interactive Entra ID authentication, like the Azure CLI.
\item Viewer-based credentials on Posit Connect. Requires the \pkg{connectcreds}
package.
}
}
}
\examples{
\dontrun{
chat <- chat_azure_openai(model = "gpt-4o-mini")
chat$chat("Tell me three jokes about statisticians")
}
}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-github.R
\name{chat_github}
\alias{chat_github}
\alias{models_github}
\title{Chat with a model hosted on the GitHub model marketplace}
\usage{
chat_github(
  system_prompt = NULL,
  base_url = "https://models.github.ai/inference/",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  echo = NULL,
  api_headers = character()
)

models_github(
  base_url = "https://models.github.ai/",
  api_key = NULL,
  credentials = NULL
)
}
\arguments{
\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{base_url}{The base URL to the API endpoint.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{Override the default credentials. You generally should not need this argument; instead set the \code{GITHUB_PAT} environment variable. The best place to set this is in \code{.Renviron},
which you can easily edit by calling \code{usethis::edit_r_environ()}.

If you do need additional control, this argument takes a zero-argument function that returns either a string (the API key), or a named list (added as additional headers to every request).}

\item{model}{The model to use for the chat (defaults to "gpt-4o").
We regularly update the default, so we strongly recommend explicitly specifying a model for anything other than casual use.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}
}
\value{
A \link{Chat} object.
}
\description{
GitHub Models hosts a number of open source and OpenAI models. To access the
GitHub model marketplace, you will need to apply for and be accepted into the
beta access program. See \url{https://github.com/marketplace/models} for details.

This function is a lightweight wrapper around \code{\link[=chat_openai_compatible]{chat_openai_compatible()}} with
the defaults tweaked for the GitHub Models marketplace.

GitHub also supports the Azure AI Inference SDK, which you can use by setting
\code{base_url} to \code{"https://models.inference.ai.azure.com/"}. This endpoint was
used in \pkg{ellmer} v0.3.0 and earlier.
}
\examples{
\dontrun{
chat <- chat_github()
chat$chat("Tell me three jokes about statisticians")
}
}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools-def.R
\name{tool_annotations}
\alias{tool_annotations}
\title{Tool annotations}
\usage{
tool_annotations(
  title = NULL,
  read_only_hint = NULL,
  open_world_hint = NULL,
  idempotent_hint = NULL,
  destructive_hint = NULL,
  ...
)
}
\arguments{
\item{title}{A human-readable title for the tool.}

\item{read_only_hint}{If \code{TRUE}, the tool does not modify its environment.}

\item{open_world_hint}{If \code{TRUE}, the tool may interact with an "open world"
of external entities. If \code{FALSE}, the tool's domain of interaction is
closed. For example, the world of a web search tool is open, but the world
of a memory tool is not.}

\item{idempotent_hint}{If \code{TRUE}, calling the tool repeatedly with the same
arguments will have no additional effect on its environment. (Only
meaningful when \code{read_only_hint} is \code{FALSE}.)}

\item{destructive_hint}{If \code{TRUE}, the tool may perform destructive updates
to its environment, otherwise it only performs additive updates. (Only
meaningful when \code{read_only_hint} is \code{FALSE}.)}

\item{...}{Additional named parameters to include in the tool annotations.}
}
\value{
A list of tool annotations.
}
\description{
Tool annotations are additional properties that, when passed to the
\code{.annotations} argument of \code{\link[=tool]{tool()}}, provide additional information about the
tool and its behavior. This information can be used for display to users, for
example in a Shiny app or another user interface.

The annotations in \code{tool_annotations()} are drawn from the \href{https://modelcontextprotocol.io/introduction}{Model Context Protocol} and are considered
\emph{hints}. Tool authors should use these annotations to communicate tool
properties, but users should note that these annotations are not guaranteed.
}
\examples{
# See ?tool() for a full example using this function.
# We're creating a tool around R's `rnorm()` function to allow the chatbot to
# generate random numbers from a normal distribution.
tool_rnorm <- tool(
  rnorm,
  # Describe the tool function to the LLM
  .description = "Drawn numbers from a random normal distribution",
  # Describe the parameters used by the tool function
  n = type_integer("The number of observations. Must be a positive integer."),
  mean = type_number("The mean value of the distribution."),
  sd = type_number("The standard deviation of the distribution. Must be a non-negative number."),
  # Tool annotations optionally provide additional context to the LLM
  .annotations = tool_annotations(
    title = "Draw Random Normal Numbers",
    read_only_hint = TRUE, # the tool does not modify any state
    open_world_hint = FALSE # the tool does not interact with the outside world
  )
)

}
\seealso{
Other tool calling helpers:
\code{\link[=tool]{tool()}},
\code{\link[=tool_reject]{tool_reject()}}
}
\concept{tool calling helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools-def-auto.R
\name{create_tool_def}
\alias{create_tool_def}
\title{Create metadata for a tool}
\usage{
create_tool_def(topic, chat = NULL, echo = interactive(), verbose = FALSE)
}
\arguments{
\item{topic}{A symbol or string literal naming the function to create
metadata for. Can also be an expression of the form \code{pkg::fun}.}

\item{chat}{A \code{Chat} object used to generate the output. If \code{NULL}
(the default) uses \code{\link[=chat_openai]{chat_openai()}}.}

\item{echo}{Emit the registration code to the console. Defaults to \code{TRUE} in
interactive sessions.}

\item{verbose}{If \code{TRUE}, print the input we send to the LLM, which may be
useful for debugging unexpectedly poor results.}
}
\value{
A \code{register_tool} call that you can copy and paste into your code.
Returned invisibly if \code{echo} is \code{TRUE}.
}
\description{
In order to use a function as a tool in a chat, you need to craft the right
call to \code{\link[=tool]{tool()}}. This function helps you do that for documented functions by
extracting the function's R documentation and using an LLM to generate the
\code{tool()} call. It's meant to be used interactively while writing your
code, not as part of your final code.

If the function has package documentation, that will be used. Otherwise, if
the source code of the function can be automatically detected, then the
comments immediately preceding the function are used (especially helpful if
those are roxygen2 comments). If neither are available, then just the function
signature is used.

Note that this function is inherently imperfect. It can't handle all possible
R functions, because not all parameters are suitable for use in a tool call
(for example, because they're not serializable to simple JSON objects). The
documentation might not specify the expected shape of arguments to the level
of detail that would allow an exact JSON schema to be generated. Please be
sure to review the generated code before using it!
}
\examples{
\dontrun{
  # These are all equivalent
  create_tool_def(rnorm)
  create_tool_def(stats::rnorm)
  create_tool_def("rnorm")
  create_tool_def("rnorm", chat = chat_azure_openai())
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-claude-tools.R
\name{claude_tool_web_fetch}
\alias{claude_tool_web_fetch}
\title{Claude web fetch tool}
\usage{
claude_tool_web_fetch(
  max_uses = NULL,
  allowed_domains = NULL,
  blocked_domains = NULL,
  citations = FALSE,
  max_content_tokens = NULL
)
}
\arguments{
\item{max_uses}{Integer. Maximum number of fetches allowed per request.}

\item{allowed_domains}{Character vector. Restrict fetches to specific domains.
Cannot be used with \code{blocked_domains}.}

\item{blocked_domains}{Character vector. Exclude specific domains from fetches.
Cannot be used with \code{allowed_domains}.}

\item{citations}{Logical. Whether to include citations in the response. Default is \code{TRUE}.}

\item{max_content_tokens}{Integer. Maximum number of tokens to fetch from each URL.}
}
\description{
Enables Claude to fetch and analyze content from web URLs. Claude can only
fetch URLs that appear in the conversation context (user messages or
previous tool results). For security reasons, Claude cannot dynamically
construct URLs to fetch.

Requires the \code{web-fetch-2025-09-10} beta header.
Learn more in \url{https://docs.claude.com/en/docs/agents-and-tools/tool-use/web-fetch-tool}.
}
\examples{
\dontrun{
chat <- chat_claude(beta_headers = "web-fetch-2025-09-10")
chat$register_tool(claude_tool_web_fetch())
chat$chat("What are the latest package releases on https://tidyverse.org/blog")
}
}
\seealso{
Other built-in tools:
\code{\link[=claude_tool_web_search]{claude_tool_web_search()}},
\code{\link[=google_tool_web_fetch]{google_tool_web_fetch()}},
\code{\link[=google_tool_web_search]{google_tool_web_search()}},
\code{\link[=openai_tool_web_search]{openai_tool_web_search()}}
}
\concept{built-in tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-claude.R
\name{chat_anthropic}
\alias{chat_anthropic}
\alias{chat_claude}
\alias{models_claude}
\alias{models_anthropic}
\title{Chat with an Anthropic Claude model}
\usage{
chat_anthropic(
  system_prompt = NULL,
  params = NULL,
  model = NULL,
  cache = c("5m", "1h", "none"),
  api_args = list(),
  base_url = "https://api.anthropic.com/v1",
  beta_headers = character(),
  api_key = NULL,
  credentials = NULL,
  api_headers = character(),
  echo = NULL
)

chat_claude(
  system_prompt = NULL,
  params = NULL,
  model = NULL,
  cache = c("5m", "1h", "none"),
  api_args = list(),
  base_url = "https://api.anthropic.com/v1",
  beta_headers = character(),
  api_key = NULL,
  credentials = NULL,
  api_headers = character(),
  echo = NULL
)

models_claude(
  base_url = "https://api.anthropic.com/v1",
  api_key = NULL,
  credentials = NULL
)

models_anthropic(
  base_url = "https://api.anthropic.com/v1",
  api_key = NULL,
  credentials = NULL
)
}
\arguments{
\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{model}{The model to use for the chat (defaults to "claude-sonnet-4-5-20250929").
We regularly update the default, so we strongly recommend explicitly specifying a model for anything other than casual use.
Use \code{models_anthropic()} to see all options.}

\item{cache}{How long to cache inputs? Defaults to "5m" (five minutes).
Set to "none" to disable caching or "1h" to cache for one hour.

See details below.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{base_url}{The base URL to the endpoint; the default is Claude's
public API.}

\item{beta_headers}{Optionally, a character vector of beta headers to opt-in
claude features that are still in beta.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{Override the default credentials. You generally should not need this argument; instead set the \code{ANTHROPIC_API_KEY} environment variable. The best place to set this is in \code{.Renviron},
which you can easily edit by calling \code{usethis::edit_r_environ()}.

If you do need additional control, this argument takes a zero-argument function that returns either a string (the API key), or a named list (added as additional headers to every request).}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}
}
\value{
A \link{Chat} object.
}
\description{
\href{https://www.anthropic.com}{Anthropic} provides a number of chat based models
under the \href{https://claude.com/product/overview}{Claude} moniker. Note that a
Claude Pro membership does not give you the ability to call models via the
API; instead, you will need to sign up (and pay for) a
\href{https://platform.claude.com/}{developer account}.
}
\section{Caching}{
Caching with Claude is a bit more complicated than other providers but we
believe that on average it will save you both money and time, so we have
enabled it by default. With other providers, like OpenAI and Google,
you only pay for cache reads, which cost 10\% of the normal price. With
Claude, you also pay for cache writes, which cost 125\% of the normal price
for 5 minute caching and 200\% of the normal price for 1 hour caching.

How does this affect the total cost of a conversation? Imagine the first
turn sends 1000 input tokens and receives 200 output tokens. The second
turn must first send both the input and output from the previous turn
(1200 tokens). It then sends a further 1000 tokens and receives 200 tokens
back.

To compare the prices of these two approaches we can ignore the cost of
output tokens, because they are the same for both. How much will the input
tokens cost? If we don't use caching, we send 1000 tokens in the first turn
and 2200 (1000 + 200 + 1000) tokens in the second turn for a total of 3200
tokens. If we use caching, we'll send (the equivalent of) 1000 * 1.25 = 1250
tokens in the first turn. In the second turn, 1000 of the input tokens will
be cached so the total cost is 1000 * 0.1 + (200 + 1000) * 1.25 = 1600
tokens. That makes a total of 2850 tokens, i.e. 11\% fewer tokens,
decreasing the overall cost.

Obviously, the details will vary from conversation to conversation, but
if you have a large system prompt that you re-use many times you should
expect to see larger savings. You can see exactly how many input and
cache input tokens each turn uses, along with the total cost,
with \code{chat$get_tokens()}. If you don't see savings for your use case, you can
suppress caching with \code{cache = "none"}.

I know this is already quite complicated, but there's one final wrinkle:
Claude will only cache longer prompts, with caching requiring at least
1024-4096 tokens, depending on the model. So don't be surprised it if you
don't see any differences with caching if you have a short prompt.

See all the details at
\url{https://docs.claude.com/en/docs/build-with-claude/prompt-caching}.
}

\examples{
\dontshow{ellmer:::vcr_example_start("chat_anthropic")}
chat <- chat_anthropic()
chat$chat("Tell me three jokes about statisticians")
\dontshow{ellmer:::vcr_example_end()}
}
\seealso{
Other chatbots:
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content-pdf.R
\name{content_pdf_file}
\alias{content_pdf_file}
\alias{content_pdf_url}
\title{Encode PDFs content for chat input}
\usage{
content_pdf_file(path)

content_pdf_url(url)
}
\arguments{
\item{path, url}{Path or URL to a PDF file.}
}
\value{
A \code{ContentPDF} object
}
\description{
These functions are used to prepare PDFs as input to the chatbot. The
\code{content_pdf_url()} function is used to provide a URL to an PDF file,
while \code{content_pdf_file()} is used to for local PDF files.

Not all providers support PDF input, so check the documentation for the
provider you are using.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content-image.R
\name{content_image_url}
\alias{content_image_url}
\alias{content_image_file}
\alias{content_image_plot}
\title{Encode images for chat input}
\usage{
content_image_url(url, detail = c("auto", "low", "high"))

content_image_file(path, content_type = "auto", resize = "low")

content_image_plot(width = 768, height = 768)
}
\arguments{
\item{url}{The URL of the image to include in the chat input. Can be a
\verb{data:} URL or a regular URL. Valid image types are PNG, JPEG, WebP, and
non-animated GIF.}

\item{detail}{The \href{https://platform.openai.com/docs/guides/images/image-input-requirements}{detail setting}
for this image. Can be \code{"auto"}, \code{"low"}, or \code{"high"}.}

\item{path}{The path to the image file to include in the chat input. Valid
file extensions are \code{.png}, \code{.jpeg}, \code{.jpg}, \code{.webp}, and (non-animated)
\code{.gif}.}

\item{content_type}{The content type of the image (e.g. \code{image/png}). If
\code{"auto"}, the content type is inferred from the file extension.}

\item{resize}{If \code{"low"}, resize images to fit within 512x512. If \code{"high"},
resize to fit within 2000x768 or 768x2000. (See the \href{https://platform.openai.com/docs/guides/images/image-input-requirements}{OpenAI docs}
for more on why these specific sizes are used.) If \code{"none"}, do not resize.

You can also pass a custom string to resize the image to a specific size,
e.g. \code{"200x200"} to resize to 200x200 pixels while preserving aspect ratio.
Append \code{>} to resize only if the image is larger than the specified size,
and \code{!} to ignore aspect ratio (e.g. \code{"300x200>!"}).

All values other than \code{none} require the \code{magick} package.}

\item{width, height}{Width and height in pixels.}
}
\value{
An input object suitable for including in the \code{...} parameter of
the \code{chat()}, \code{stream()}, \code{chat_async()}, or \code{stream_async()} methods.
}
\description{
These functions are used to prepare image URLs and files for input to the
chatbot. The \code{content_image_url()} function is used to provide a URL to an
image, while \code{content_image_file()} is used to provide the image data itself.
}
\examples{
\dontrun{
chat <- chat_openai()
chat$chat(
  "What do you see in these images?",
  content_image_url("https://www.r-project.org/Rlogo.png"),
  content_image_file(system.file("httr2.png", package = "ellmer"))
)

plot(waiting ~ eruptions, data = faithful)
chat <- chat_openai()
chat$chat(
  "Describe this plot in one paragraph, as suitable for inclusion in
   alt-text. You should briefly describe the plot type, the axes, and
   2-5 major visual patterns.",
   content_image_plot()
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-google.R
\name{chat_google_gemini}
\alias{chat_google_gemini}
\alias{chat_google_vertex}
\alias{models_google_gemini}
\alias{models_google_vertex}
\title{Chat with a Google Gemini or Vertex AI model}
\usage{
chat_google_gemini(
  system_prompt = NULL,
  base_url = "https://generativelanguage.googleapis.com/v1beta/",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  api_headers = character(),
  echo = NULL
)

chat_google_vertex(
  location,
  project_id,
  system_prompt = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  api_headers = character(),
  echo = NULL
)

models_google_gemini(
  base_url = "https://generativelanguage.googleapis.com/v1beta/",
  api_key = NULL,
  credentials = NULL
)

models_google_vertex(location, project_id, credentials = NULL)
}
\arguments{
\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{base_url}{The base URL to the API endpoint.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{A function that returns a list of authentication headers
or \code{NULL}, the default, to use ambient credentials. See above for details.}

\item{model}{The model to use for the chat (defaults to "gemini-2.5-flash").
We regularly update the default, so we strongly recommend explicitly specifying a model for anything other than casual use.
Use \code{models_google_gemini()} to see all options.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}

\item{location}{Location, e.g. \code{us-east1}, \code{me-central1}, \code{africa-south1} or
\code{global}.}

\item{project_id}{Project ID.}
}
\value{
A \link{Chat} object.
}
\description{
Google's AI offering is broken up into two parts: Gemini and Vertex AI.
Most enterprises are likely to use Vertex AI, and individuals are likely
to use Gemini.

Use \code{\link[=google_upload]{google_upload()}} to upload files (PDFs, images, video, audio, etc.)
\subsection{Authentication}{

These functions try a number of authentication strategies, in this order:
\itemize{
\item An API key set in the \code{GOOGLE_API_KEY} env var, or,
for \code{chat_google_gemini()} only, \code{GEMINI_API_KEY}.
\item Google's default application credentials, if the \pkg{gargle} package
is installed.
\item Viewer-based credentials on Posit Connect, if the \pkg{connectcreds}
package.
\item \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}. An browser-based OAuth flow, if
you're in an interactive session. This currently uses an unverified
OAuth app (so you will get a scary warning); we plan to verify in the
near future.
}
}
}
\examples{
\dontrun{
chat <- chat_google_gemini()
chat$chat("Tell me three jokes about statisticians")
}
}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/live.R
\name{live_console}
\alias{live_console}
\alias{live_browser}
\title{Open a live chat application}
\usage{
live_console(chat, quiet = FALSE)

live_browser(chat, quiet = FALSE)
}
\arguments{
\item{chat}{A chat object created by \code{\link[=chat_openai]{chat_openai()}} or friends.}

\item{quiet}{If \code{TRUE}, suppresses the initial message that explains how
to use the console.}
}
\value{
(Invisibly) The input \code{chat}.
}
\description{
\itemize{
\item \code{live_console()} lets you chat interactively in the console.
\item \code{live_browser()} lets you chat interactively in a browser.
}

Note that these functions will mutate the input \code{chat} object as
you chat because your turns will be appended to the history.
}
\examples{
\dontrun{
chat <- chat_anthropic()
live_console(chat)
live_browser(chat)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{Chat}
\alias{Chat}
\title{The Chat object}
\value{
A Chat object
}
\description{
A \code{Chat} is a sequence of user and assistant \link{Turn}s sent
to a specific \link{Provider}. A \code{Chat} is a mutable R6 object that takes care of
managing the state associated with the chat; i.e. it records the messages
that you send to the server, and the messages that you receive back.
If you register a tool (i.e. an R function that the assistant can call on
your behalf), it also takes care of the tool loop.

You should generally not create this object yourself,
but instead call \code{\link[=chat_openai]{chat_openai()}} or friends instead.
}
\examples{
\dontshow{ellmer:::vcr_example_start("Chat")}
chat <- chat_openai()
chat$chat("Tell me a funny joke")
\dontshow{ellmer:::vcr_example_end()}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Chat-initialize}{\code{Chat$new()}}
    \item \href{#method-Chat-get_turns}{\code{Chat$get_turns()}}
    \item \href{#method-Chat-set_turns}{\code{Chat$set_turns()}}
    \item \href{#method-Chat-add_turn}{\code{Chat$add_turn()}}
    \item \href{#method-Chat-get_system_prompt}{\code{Chat$get_system_prompt()}}
    \item \href{#method-Chat-get_model}{\code{Chat$get_model()}}
    \item \href{#method-Chat-set_system_prompt}{\code{Chat$set_system_prompt()}}
    \item \href{#method-Chat-get_tokens}{\code{Chat$get_tokens()}}
    \item \href{#method-Chat-get_cost}{\code{Chat$get_cost()}}
    \item \href{#method-Chat-last_turn}{\code{Chat$last_turn()}}
    \item \href{#method-Chat-chat}{\code{Chat$chat()}}
    \item \href{#method-Chat-chat_structured}{\code{Chat$chat_structured()}}
    \item \href{#method-Chat-chat_structured_async}{\code{Chat$chat_structured_async()}}
    \item \href{#method-Chat-chat_async}{\code{Chat$chat_async()}}
    \item \href{#method-Chat-stream}{\code{Chat$stream()}}
    \item \href{#method-Chat-stream_async}{\code{Chat$stream_async()}}
    \item \href{#method-Chat-register_tool}{\code{Chat$register_tool()}}
    \item \href{#method-Chat-register_tools}{\code{Chat$register_tools()}}
    \item \href{#method-Chat-get_provider}{\code{Chat$get_provider()}}
    \item \href{#method-Chat-get_tools}{\code{Chat$get_tools()}}
    \item \href{#method-Chat-set_tools}{\code{Chat$set_tools()}}
    \item \href{#method-Chat-on_tool_request}{\code{Chat$on_tool_request()}}
    \item \href{#method-Chat-on_tool_result}{\code{Chat$on_tool_result()}}
    \item \href{#method-Chat-clone}{\code{Chat$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-initialize}{}}}
\subsection{\code{Chat$new()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$new(provider, system_prompt = NULL, echo = "none")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{provider}}{A provider object.}
      \item{\code{system_prompt}}{System prompt to start the conversation with.}
      \item{\code{echo}}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method. You can override the default
by setting the \code{ellmer_echo} option.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-get_turns"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-get_turns}{}}}
\subsection{\code{Chat$get_turns()}}{
  Retrieve the turns that have been sent and received so far
(optionally starting with the system prompt, if any).
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$get_turns(include_system_prompt = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{include_system_prompt}}{Whether to include the system prompt in the
turns (if any exists).}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-set_turns"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-set_turns}{}}}
\subsection{\code{Chat$set_turns()}}{
  Replace existing turns with a new list.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$set_turns(value)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{value}}{A list of \link{Turn}s.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-add_turn"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-add_turn}{}}}
\subsection{\code{Chat$add_turn()}}{
  Add a pair of turns to the chat.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$add_turn(user, assistant, log_tokens = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{user}}{The user \link{Turn}.}
      \item{\code{assistant}}{The system \link{Turn}.}
      \item{\code{log_tokens}}{Should tokens used in the turn be logged to the
session counter?}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-get_system_prompt"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-get_system_prompt}{}}}
\subsection{\code{Chat$get_system_prompt()}}{
  If set, the system prompt, it not, \code{NULL}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$get_system_prompt()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-get_model"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-get_model}{}}}
\subsection{\code{Chat$get_model()}}{
  Retrieve the model name
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$get_model()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-set_system_prompt"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-set_system_prompt}{}}}
\subsection{\code{Chat$set_system_prompt()}}{
  Update the system prompt
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$set_system_prompt(value)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{value}}{A character vector giving the new system prompt}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-get_tokens"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-get_tokens}{}}}
\subsection{\code{Chat$get_tokens()}}{
  A data frame with token usage and cost data. There are four
columns: \code{input}, \code{output}, \code{cached_input}, and \code{cost}. There is one
row for each assistant turn, because token counts and costs are only
available when the API returns the assistant's response.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$get_tokens(include_system_prompt = deprecated())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{include_system_prompt}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-get_cost"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-get_cost}{}}}
\subsection{\code{Chat$get_cost()}}{
  The cost of this chat
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$get_cost(include = c("all", "last"))}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{include}}{The default, \code{"all"}, gives the total cumulative cost
of this chat. Alternatively, use \code{"last"} to get the cost of just the
most recent turn. Incomplete turns (from cancelled or interrupted
streams) are excluded because they lack token data.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-last_turn"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-last_turn}{}}}
\subsection{\code{Chat$last_turn()}}{
  The last turn returned by the assistant.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$last_turn(role = c("assistant", "user", "system"))}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{role}}{Optionally, specify a role to find the last turn with
for the role.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Either a \code{Turn} or \code{NULL}, if no turns with the specified
role have occurred.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-chat"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-chat}{}}}
\subsection{\code{Chat$chat()}}{
  Submit input to the chatbot, and return the response as a
simple string (probably Markdown).
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$chat(..., echo = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{The input to send to the chatbot. Can be strings or images
(see \code{\link[=content_image_file]{content_image_file()}} and \code{\link[=content_image_url]{content_image_url()}}.}
      \item{\code{echo}}{Whether to emit the response to stdout as it is received. If
\code{NULL}, then the value of \code{echo} set when the chat object was created
will be used.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-chat_structured"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-chat_structured}{}}}
\subsection{\code{Chat$chat_structured()}}{
  Extract structured data
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$chat_structured(..., type, echo = "none", convert = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{The input to send to the chatbot. This is typically the text
you want to extract data from, but it can be omitted if the data is
obvious from the existing conversation.}
      \item{\code{type}}{A type specification for the extracted data. Should be
created with a \code{\link[=type_boolean]{type_()}} function.}
      \item{\code{echo}}{Whether to emit the response to stdout as it is received.
Set to "text" to stream JSON data as it's generated (not supported by
all providers).}
      \item{\code{convert}}{Automatically convert from JSON lists to R data types
using the schema. For example, this will turn arrays of objects into
data frames and arrays of strings into a character vector.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-chat_structured_async"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-chat_structured_async}{}}}
\subsection{\code{Chat$chat_structured_async()}}{
  Extract structured data, asynchronously. Returns a promise
that resolves to an object matching the type specification.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$chat_structured_async(..., type, echo = "none", convert = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{The input to send to the chatbot. Will typically include
the phrase "extract structured data".}
      \item{\code{type}}{A type specification for the extracted data. Should be
created with a \code{\link[=type_boolean]{type_()}} function.}
      \item{\code{echo}}{Whether to emit the response to stdout as it is received.
Set to "text" to stream JSON data as it's generated (not supported by
all providers).}
      \item{\code{convert}}{Automatically convert from JSON lists to R data types
using the schema. For example, this will turn arrays of objects into
data frames and arrays of strings into a character vector.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-chat_async"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-chat_async}{}}}
\subsection{\code{Chat$chat_async()}}{
  Submit input to the chatbot, and receive a promise that
resolves with the response all at once. Returns a promise that resolves
to a string (probably Markdown).
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$chat_async(..., tool_mode = c("concurrent", "sequential"))}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{The input to send to the chatbot. Can be strings or images.}
      \item{\code{tool_mode}}{Whether tools should be invoked one-at-a-time
(\code{"sequential"}) or concurrently (\code{"concurrent"}). Sequential mode is
best for interactive applications, especially when a tool may involve
an interactive user interface. Concurrent mode is the default and is
best suited for automated scripts or non-interactive applications.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-stream"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-stream}{}}}
\subsection{\code{Chat$stream()}}{
  Submit input to the chatbot, returning streaming results.
Returns A \href{https://coro.r-lib.org/articles/generator.html#iterating}{coro generator}
that yields strings. While iterating, the generator will block while
waiting for more content from the chatbot.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$stream(..., stream = c("text", "content"), controller = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{The input to send to the chatbot. Can be strings or images.}
      \item{\code{stream}}{Whether the stream should yield only \code{"text"} or ellmer's
rich content types. When \code{stream = "content"}, \code{stream()} yields
\link{Content} objects.}
      \item{\code{controller}}{An optional \code{\link[=stream_controller]{stream_controller()}} used to cancel the
stream from outside the iteration loop.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-stream_async"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-stream_async}{}}}
\subsection{\code{Chat$stream_async()}}{
  Submit input to the chatbot, returning asynchronously
streaming results. Returns a \href{https://coro.r-lib.org/reference/async_generator.html}{coro async generator} that
yields string promises.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$stream_async(
  ...,
  tool_mode = c("concurrent", "sequential"),
  stream = c("text", "content"),
  controller = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{The input to send to the chatbot. Can be strings or images.}
      \item{\code{tool_mode}}{Whether tools should be invoked one-at-a-time
(\code{"sequential"}) or concurrently (\code{"concurrent"}). Sequential mode is
best for interactive applications, especially when a tool may involve
an interactive user interface. Concurrent mode is the default and is
best suited for automated scripts or non-interactive applications.}
      \item{\code{stream}}{Whether the stream should yield only \code{"text"} or ellmer's
rich content types. When \code{stream = "content"}, \code{stream()} yields
\link{Content} objects.}
      \item{\code{controller}}{An optional \code{\link[=stream_controller]{stream_controller()}} used to cancel the
stream from outside the iteration loop.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-register_tool"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-register_tool}{}}}
\subsection{\code{Chat$register_tool()}}{
  Register a tool (an R function) that the chatbot can use.
Learn more in \code{vignette("tool-calling")}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$register_tool(tool)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{tool}}{A tool definition created by \code{\link[=tool]{tool()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-register_tools"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-register_tools}{}}}
\subsection{\code{Chat$register_tools()}}{
  Register a list of tools.
Learn more in \code{vignette("tool-calling")}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$register_tools(tools)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{tools}}{A list of tool definitions created by \code{\link[=tool]{tool()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-get_provider"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-get_provider}{}}}
\subsection{\code{Chat$get_provider()}}{
  Get the underlying provider object. For expert use only.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$get_provider()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-get_tools"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-get_tools}{}}}
\subsection{\code{Chat$get_tools()}}{
  Retrieve the list of registered tools.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$get_tools()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-set_tools"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-set_tools}{}}}
\subsection{\code{Chat$set_tools()}}{
  Sets the available tools. For expert use only; most users
should use \code{register_tool()}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$set_tools(tools)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{tools}}{A list of tool definitions created with \code{\link[=tool]{tool()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-on_tool_request"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-on_tool_request}{}}}
\subsection{\code{Chat$on_tool_request()}}{
  Register a callback for a tool request event.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$on_tool_request(callback)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{callback}}{A function to be called when a tool request event occurs,
which must have \code{request} as its only argument.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    A function that can be called to remove the callback.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-on_tool_result"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-on_tool_result}{}}}
\subsection{\code{Chat$on_tool_result()}}{
  Register a callback for a tool result event.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$on_tool_result(callback)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{callback}}{A function to be called when a tool result event occurs,
which must have \code{result} as its only argument.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    A function that can be called to remove the callback.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chat-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Chat-clone}{}}}
\subsection{\code{Chat$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Chat$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-openai.R
\name{chat_openai}
\alias{chat_openai}
\alias{models_openai}
\title{Chat with an OpenAI model}
\usage{
chat_openai(
  system_prompt = NULL,
  base_url = "https://api.openai.com/v1",
  api_key = NULL,
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  api_headers = character(),
  service_tier = c("auto", "default", "flex", "priority"),
  echo = c("none", "output", "all")
)

models_openai(
  base_url = "https://api.openai.com/v1",
  api_key = NULL,
  credentials = NULL
)
}
\arguments{
\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{base_url}{The base URL to the API endpoint.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{Override the default credentials. You generally should not need this argument; instead set the \code{OPENAI_API_KEY} environment variable. The best place to set this is in \code{.Renviron},
which you can easily edit by calling \code{usethis::edit_r_environ()}.

If you do need additional control, this argument takes a zero-argument function that returns either a string (the API key), or a named list (added as additional headers to every request).}

\item{model}{The model to use for the chat (defaults to "gpt-4.1").
We regularly update the default, so we strongly recommend explicitly specifying a model for anything other than casual use.
Use \code{models_openai()} to see all options.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}

\item{service_tier}{Request a specific service tier. There are four options:
\itemize{
\item \code{"auto"} (default): uses the service tier configured in Project settings.
\item \code{"default"}: standard pricing and performance.
\item \code{"flex"}: slower and cheaper.
\item \code{"priority"}: faster and more expensive.
}}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}
}
\value{
A \link{Chat} object.
}
\description{
This is the main interface to \href{https://openai.com/}{OpenAI}'s models,
using the \strong{responses API}. You can use this to access OpenAI's latest
models and features like image generation and web search. If you need to use
an OpenAI-compatible API from another provider, or the \strong{chat completions}
API with OpenAI,use \code{\link[=chat_openai_compatible]{chat_openai_compatible()}} instead.

Note that a ChatGPT Plus membership does not grant access to the API.
You will need to sign up for a developer account (and pay for it) at the
\href{https://platform.openai.com}{developer platform}.
}
\examples{
\dontshow{ellmer:::vcr_example_start("chat_openai")}
chat <- chat_openai()
chat$chat("
  What is the difference between a tibble and a data frame?
  Answer with a bulleted list
")

chat$chat("Tell me three funny jokes about statisticians")
\dontshow{ellmer:::vcr_example_end()}
}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ellmer-package.R
\docType{package}
\name{ellmer-package}
\alias{ellmer}
\alias{ellmer-package}
\title{ellmer: Chat with Large Language Models}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Chat with large language models from a range of providers including 'Claude' \url{https://claude.ai}, 'OpenAI' \url{https://chatgpt.com}, and more. Supports streaming, asynchronous calls, tool calling, and structured data extraction.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://ellmer.tidyverse.org}
  \item \url{https://github.com/tidyverse/ellmer}
  \item Report bugs at \url{https://github.com/tidyverse/ellmer/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@posit.co} (\href{https://orcid.org/0000-0003-4757-117X}{ORCID})

Authors:
\itemize{
  \item Hadley Wickham \email{hadley@posit.co} (\href{https://orcid.org/0000-0003-4757-117X}{ORCID})
  \item Joe Cheng
  \item Aaron Jacobs
  \item Garrick Aden-Buie \email{garrick@posit.co} (\href{https://orcid.org/0000-0002-7111-0077}{ORCID})
  \item Barret Schloerke \email{barret@posit.co} (\href{https://orcid.org/0000-0001-9986-114X}{ORCID})
}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/params.R
\name{params}
\alias{params}
\title{Standard model parameters}
\usage{
params(
  temperature = NULL,
  top_p = NULL,
  top_k = NULL,
  frequency_penalty = NULL,
  presence_penalty = NULL,
  seed = NULL,
  max_tokens = NULL,
  log_probs = NULL,
  stop_sequences = NULL,
  reasoning_effort = NULL,
  reasoning_tokens = NULL,
  ...
)
}
\arguments{
\item{temperature}{Temperature of the sampling distribution.}

\item{top_p}{The cumulative probability for token selection.}

\item{top_k}{The number of highest probability vocabulary tokens to keep.}

\item{frequency_penalty}{Frequency penalty for generated tokens.}

\item{presence_penalty}{Presence penalty for generated tokens.}

\item{seed}{Seed for random number generator.}

\item{max_tokens}{Maximum number of tokens to generate.}

\item{log_probs}{Include the log probabilities in the output?}

\item{stop_sequences}{A character vector of tokens to stop generation on.}

\item{reasoning_effort, reasoning_tokens}{How much effort to spend thinking?
\code{ressoning_effort} is a string, like "low", "medium", "high".
\code{reasoning_tokens} is an integer, giving a maximum token budget.
Each provider only takes one of these two parameters.}

\item{...}{Additional named parameters to send to the provider.}
}
\description{
This helper function makes it easier to create a list of parameters used
across many models. The parameter names are automatically standardised and
included in the correctly place in the API call.

Note that parameters that are not supported by a given provider will generate
a warning, not an error. This allows you to use the same set of parameters
across multiple providers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-snowflake.R
\name{chat_snowflake}
\alias{chat_snowflake}
\title{Chat with a model hosted on Snowflake}
\usage{
chat_snowflake(
  system_prompt = NULL,
  account = snowflake_account(),
  credentials = NULL,
  model = NULL,
  params = NULL,
  api_args = list(),
  echo = c("none", "output", "all"),
  api_headers = character()
)
}
\arguments{
\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{account}{A Snowflake \href{https://docs.snowflake.com/en/user-guide/admin-account-identifier}{account identifier},
e.g. \code{"testorg-test_account"}. Defaults to the value of the
\code{SNOWFLAKE_ACCOUNT} environment variable.}

\item{credentials}{A list of authentication headers to pass into
\code{\link[httr2:req_headers]{httr2::req_headers()}}, a function that returns them when called, or
\code{NULL}, the default, to use ambient credentials.}

\item{model}{The model to use for the chat (defaults to "claude-3-7-sonnet").
We regularly update the default, so we strongly recommend explicitly specifying a model for anything other than casual use.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}
}
\value{
A \link{Chat} object.
}
\description{
The Snowflake provider allows you to interact with LLM models available
through the \href{https://docs.snowflake.com/en/user-guide/snowflake-cortex/cortex-llm-rest-api}{Cortex LLM REST API}.
\subsection{Authentication}{

\code{chat_snowflake()} picks up the following ambient Snowflake credentials:
\itemize{
\item A static OAuth token defined via the \code{SNOWFLAKE_TOKEN} environment
variable.
\item Key-pair authentication credentials defined via the \code{SNOWFLAKE_USER} and
\code{SNOWFLAKE_PRIVATE_KEY} (which can be a PEM-encoded private key or a path
to one) environment variables.
\item Posit Workbench-managed Snowflake credentials for the corresponding
\code{account}.
\item Viewer-based credentials on Posit Connect. Requires the \pkg{connectcreds}
package.
}
}

\subsection{Known limitations}{

Note that Snowflake-hosted models do not support images.
}
}
\examples{
\dontshow{if (has_credentials("snowflake")) withAutoprint(\{ # examplesIf}
chat <- chat_snowflake()
chat$chat("Tell me a joke in the form of a SQL query.")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-aws.R
\name{chat_aws_bedrock}
\alias{chat_aws_bedrock}
\alias{models_aws_bedrock}
\title{Chat with an AWS bedrock model}
\usage{
chat_aws_bedrock(
  system_prompt = NULL,
  base_url = NULL,
  model = NULL,
  profile = NULL,
  cache = c("auto", "5m", "1h", "none"),
  params = NULL,
  api_args = list(),
  api_headers = character(),
  echo = NULL
)

models_aws_bedrock(profile = NULL, base_url = NULL)
}
\arguments{
\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{base_url}{The base URL to the API endpoint.}

\item{model}{The model to use for the chat (defaults to "anthropic.claude-sonnet-4-5-20250929-v1:0").
We regularly update the default, so we strongly recommend explicitly specifying a model for anything other than casual use.
Use \code{models_models_aws_bedrock()} to see all options.
.

While ellmer provides a default model, there's no guarantee that you'll
have access to it, so you'll need to specify a model that you can.
If you're using \href{https://aws.amazon.com/blogs/machine-learning/getting-started-with-cross-region-inference-in-amazon-bedrock/}{cross-region inference},
you'll need to use the inference profile ID, e.g.
\code{model="us.anthropic.claude-sonnet-4-5-20250929-v1:0"}.}

\item{profile}{AWS profile to use.}

\item{cache}{How long to cache inputs? The default, \code{"auto"}, enables
caching with a 5-minute TTL for models known to support it (Anthropic
Claude and Amazon Nova) and disables caching for all other models.
Set to \code{"5m"} or \code{"1h"} to force caching on, or \code{"none"} to disable it.

See details below.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Use \code{params} for common parameters. Model-specific
inference parameters can be provided using the
\code{additionalModelRequestFields} field, for example to enable thinking effort
in Anthropic Claude models:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{api_args = list(
  additionalModelRequestFields = list(
    thinking = list(type = "enabled", budget_tokens = 4000)
  )
)
}\if{html}{\out{</div>}}

See \url{https://docs.aws.amazon.com/bedrock/latest/userguide/conversation-inference-call.html}
for more details.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}
}
\value{
A \link{Chat} object.
}
\description{
\href{https://aws.amazon.com/bedrock/}{AWS Bedrock} provides a number of
language models, including those from Anthropic's
\href{https://aws.amazon.com/bedrock/claude/}{Claude}, using the Bedrock
\href{https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html}{Converse API}.
\subsection{Authentication}{

Authentication is handled through \{paws.common\}, so if authentication
does not work for you automatically, you'll need to follow the advice
at \url{https://www.paws-r-sdk.com/#credentials}. In particular, if your
org uses AWS SSO, you'll need to run \verb{aws sso login} at the terminal.
}

\subsection{Prompt caching}{

Bedrock supports
\href{https://docs.aws.amazon.com/bedrock/latest/userguide/prompt-caching.html}{prompt caching}
via cache checkpoints. When caching is enabled, ellmer places cache
checkpoints on the system prompt and the last turn, so that the
conversation history is cached across turns.

By default (\code{cache = "auto"}), caching is enabled for models known to
support it (Anthropic Claude and Amazon Nova) and disabled for all other
models. You can also set \code{cache} to \code{"5m"} or \code{"1h"} to force a specific
TTL, or \code{"none"} to disable caching entirely. Note that individual models
may have minimum input token thresholds before caching takes effect.

Note that \code{\link[=token_usage]{token_usage()}} does not currently reflect the cost of writing
to the cache, which is priced at a premium over regular input tokens.
Cache read savings are reported correctly.
}
}
\examples{
\dontrun{
# Basic usage
chat <- chat_aws_bedrock()
chat$chat("Tell me three jokes about statisticians")
}
}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-lmstudio.R
\name{chat_lmstudio}
\alias{chat_lmstudio}
\alias{models_lmstudio}
\title{Chat with a local LM Studio model}
\usage{
chat_lmstudio(
  system_prompt = NULL,
  base_url = Sys.getenv("LMSTUDIO_BASE_URL", "http://localhost:1234"),
  model,
  params = NULL,
  api_args = list(),
  echo = NULL,
  credentials = NULL,
  api_headers = character()
)

models_lmstudio(base_url = "http://localhost:1234", credentials = NULL)
}
\arguments{
\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{base_url}{The base URL to the API endpoint.}

\item{model}{The model to use for the chat.
Use \code{models_lmstudio()} to see all options.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}

\item{credentials}{LM Studio doesn't require credentials for local usage
and in most cases you do not need to provide \code{credentials}.

However, if you're accessing an LM Studio instance hosted behind a
reverse proxy or secured endpoint that enforces bearer-token
authentication, you can set the \code{LMSTUDIO_API_KEY} environment variable
or provide a callback function to \code{credentials}.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}
}
\value{
A \link{Chat} object.
}
\description{
To use \code{chat_lmstudio()} first download and install
\href{https://lmstudio.ai}{LM Studio}. Then load a model using the LM Studio
GUI and start the local server. To learn more about running LM Studio
locally, see \url{https://lmstudio.ai/docs/developer/core/server}/.

Built on top of \code{\link[=chat_openai_compatible]{chat_openai_compatible()}}.
}
\examples{
\dontrun{
# https://lmstudio.ai/models/zai-org/glm-4.7-flash
chat <- chat_lmstudio(model = "zai-org/glm-4.7-flash")
chat$chat("Tell me three jokes about statisticians")
}

}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_databricks]{chat_databricks()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-vllm.R
\name{chat_vllm}
\alias{chat_vllm}
\alias{models_vllm}
\title{Chat with a model hosted by vLLM}
\usage{
chat_vllm(
  base_url,
  system_prompt = NULL,
  model,
  params = NULL,
  api_args = list(),
  api_key = NULL,
  credentials = NULL,
  echo = NULL,
  api_headers = character()
)

models_vllm(base_url, api_key = NULL, credentials = NULL)
}
\arguments{
\item{base_url}{The base URL to the API endpoint.}

\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{model}{The model to use for the chat.
Use \code{models_vllm()} to see all options.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{api_key}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{credentials} instead.}

\item{credentials}{Override the default credentials. You generally should not need this argument; instead set the \code{VLLM_API_KEY} environment variable. The best place to set this is in \code{.Renviron},
which you can easily edit by calling \code{usethis::edit_r_environ()}.

If you do need additional control, this argument takes a zero-argument function that returns either a string (the API key), or a named list (added as additional headers to every request).}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}
}
\value{
A \link{Chat} object.
}
\description{
\href{https://docs.vllm.ai/en/latest/}{vLLM} is an open source library that
provides an efficient and convenient LLMs model server. You can use
\code{chat_vllm()} to connect to endpoints powered by vLLM.

Uses OpenAI compatible API via \code{chat_openai_compatible()}.
}
\examples{
\dontrun{
chat <- chat_vllm("http://my-vllm.com")
chat$chat("Tell me three jokes about statisticians")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpolate.R
\name{interpolate}
\alias{interpolate}
\alias{interpolate_file}
\alias{interpolate_package}
\title{Helpers for interpolating data into prompts}
\usage{
interpolate(prompt, ..., .envir = parent.frame())

interpolate_file(path, ..., .envir = parent.frame())

interpolate_package(package, path, ..., .envir = parent.frame())
}
\arguments{
\item{prompt}{A prompt string. You should not generally expose this
to the end user, since glue interpolation makes it easy to run arbitrary
code.}

\item{...}{Define additional temporary variables for substitution.}

\item{.envir}{Environment to evaluate \code{...} expressions in. Used when
wrapping in another function. See \code{vignette("wrappers", package = "glue")}
for more details.}

\item{path}{A path to a prompt file (often a \code{.md}). In
\code{interpolate_package()}, this path is relative to \code{inst/prompts}.}

\item{package}{Package name.}
}
\value{
A \{glue\} string.
}
\description{
These functions are lightweight wrappers around
\href{https://glue.tidyverse.org/}{glue} that make it easier to interpolate
dynamic data into a static prompt:
\itemize{
\item \code{interpolate()} works with a string.
\item \code{interpolate_file()} works with a file.
\item \code{interpolate_package()} works with a file in the \code{inst/prompts}
directory of a package.
}

Compared to glue, dynamic values should be wrapped in \code{{{ }}}, making it
easier to include R code and JSON in your prompt.
}
\examples{
joke <- "You're a cool dude who loves to make jokes. Tell me a joke about {{topic}}."

# You can supply valuese directly:
interpolate(joke, topic = "bananas")

# Or allow interpolate to find them in the current environment:
topic <- "applies"
interpolate(joke)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/types.R
\name{Type}
\alias{Type}
\alias{TypeBasic}
\alias{TypeEnum}
\alias{TypeArray}
\alias{TypeJsonSchema}
\alias{TypeIgnore}
\alias{TypeObject}
\title{Type definitions for function calling and structured data extraction.}
\usage{
TypeBasic(description = NULL, required = TRUE, type = stop("Required"))

TypeEnum(description = NULL, required = TRUE, values = character(0))

TypeArray(description = NULL, required = TRUE, items = Type())

TypeJsonSchema(description = NULL, required = TRUE, json = list())

TypeIgnore(description = NULL, required = TRUE)

TypeObject(
  description = NULL,
  required = TRUE,
  properties = list(),
  additional_properties = FALSE
)
}
\arguments{
\item{description}{The purpose of the component. This is
used by the LLM to determine what values to pass to the tool or what
values to extract in the structured data, so the more detail that you can
provide here, the better.}

\item{required}{Is the component or argument required?

In type descriptions for structured data, if \code{required = FALSE} and the
component does not exist in the data, the LLM may hallucinate a value. Only
applies when the element is nested inside of a \code{type_object()}.

In tool definitions, \code{required = TRUE} signals that the LLM should always
provide a value. Arguments with \code{required = FALSE} should have a default
value in the tool function's definition. If the LLM does not provide a
value, the default value will be used.}

\item{type}{Basic type name. Must be one of \code{boolean}, \code{integer},
\code{number}, or \code{string}.}

\item{values}{Character vector of permitted values.}

\item{items}{The type of the array items. Can be created by any of the
\code{type_} function.}

\item{json}{A JSON schema object as a list.}

\item{properties}{Named list of properties stored inside the object.
Each element should be an S7 \code{Type} object.`}

\item{additional_properties}{Can the object have arbitrary additional
properties that are not explicitly listed? Only supported by Claude.}
}
\value{
S7 objects inheriting from \code{Type}
}
\description{
These S7 classes are provided for use by package devlopers who are
extending ellmer. In every day use, use \code{\link[=type_boolean]{type_boolean()}} and friends.
}
\examples{
TypeBasic(type = "boolean")
TypeArray(items = TypeBasic(type = "boolean"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tokens.R
\name{token_usage}
\alias{token_usage}
\title{Report on token usage in the current session}
\usage{
token_usage()
}
\value{
A data frame
}
\description{
Call this function to find out the cumulative number of tokens that you
have sent and recieved in the current session. The price will be shown
if known.
}
\examples{
token_usage()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools-def.R
\name{tool_reject}
\alias{tool_reject}
\title{Reject a tool call}
\usage{
tool_reject(reason = "The user has chosen to disallow the tool call.")
}
\arguments{
\item{reason}{A character string describing the reason for rejecting the
tool call.}
}
\value{
Throws an error of class \code{ellmer_tool_reject} with the provided
reason.
}
\description{
Throws an error to reject a tool call. \code{tool_reject()} can be used within the
tool function to indicate that the tool call should not be processed.
\code{tool_reject()} can also be called in an \code{Chat$on_tool_request()} callback.
When used in the callback, the tool call is rejected before the tool
function is invoked.

Here's an example where \code{utils::askYesNo()} is used to ask the user for
permission before accessing their current working directory. This happens
directly in the tool function and is appropriate when you write the tool
definition and know exactly how it will be called.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{chat <- chat_openai(model = "gpt-4.1-nano")

list_files <- function() \{
  allow_read <- utils::askYesNo(
    "Would you like to allow access to your current directory?"
  )
  if (isTRUE(allow_read)) \{
    dir(pattern = "[.](r|R|csv)$")
  \} else \{
    tool_reject()
  \}
\}

chat$register_tool(tool(
  list_files,
  "List files in the user's current directory"
))

chat$chat("What files are available in my current directory?")
#> [tool call] list_files()
#> Would you like to allow access to your current directory? (Yes/no/cancel) no
#> #> Error: Tool call rejected. The user has chosen to disallow the tool #' call.
#> It seems I am unable to access the files in your current directory right now.
#> If you can tell me what specific files you're looking for or if you can #' provide
#> the list, I can assist you further.

chat$chat("Try again.")
#> [tool call] list_files()
#> Would you like to allow access to your current directory? (Yes/no/cancel) yes
#> #> app.R
#> #> data.csv
#> The files available in your current directory are "app.R" and "data.csv".
}\if{html}{\out{</div>}}

You can achieve a similar experience with tools written by others by using a
\code{tool_request} callback. In the next example, imagine the tool is provided by
a third-party package. This example implements a simple menu to ask the user
for consent before running \emph{any}  tool.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{packaged_list_files_tool <- tool(
  function() dir(pattern = "[.](r|R|csv)$"),
  "List files in the user's current directory"
)

chat <- chat_openai(model = "gpt-4.1-nano")
chat$register_tool(packaged_list_files_tool)

always_allowed <- c()

# ContentToolRequest
chat$on_tool_request(function(request) \{
  if (request@name \%in\% always_allowed) return()

  answer <- utils::menu(
    title = sprintf("Allow tool `\%s()` to run?", request@name),
    choices = c("Always", "Once", "No"),
    graphics = FALSE
  )

  if (answer == 1) \{
    always_allowed <<- append(always_allowed, request@name)
  \} else if (answer \%in\% c(0, 3)) \{
    tool_reject()
  \}
\})

# Try choosing different answers to the menu each time
chat$chat("What files are available in my current directory?")
chat$chat("How about now?")
chat$chat("And again now?")
}\if{html}{\out{</div>}}
}
\seealso{
Other tool calling helpers:
\code{\link[=tool]{tool()}},
\code{\link[=tool_annotations]{tool_annotations()}}
}
\concept{tool calling helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-google-tools.R
\name{google_tool_web_fetch}
\alias{google_tool_web_fetch}
\title{Google URL fetch tool}
\usage{
google_tool_web_fetch()
}
\description{
When this tool is enabled, you can include URLs directly in your prompts and
Gemini will fetch and analyze the content.

Learn more in \url{https://ai.google.dev/gemini-api/docs/url-context}.
}
\examples{
\dontrun{
chat <- chat_google_gemini()
chat$register_tool(google_tool_web_fetch())
chat$chat("What are the latest package releases on https://tidyverse.org/blog?")
}
}
\seealso{
Other built-in tools:
\code{\link[=claude_tool_web_fetch]{claude_tool_web_fetch()}},
\code{\link[=claude_tool_web_search]{claude_tool_web_search()}},
\code{\link[=google_tool_web_search]{google_tool_web_search()}},
\code{\link[=openai_tool_web_search]{openai_tool_web_search()}}
}
\concept{built-in tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.R
\name{contents_text}
\alias{contents_text}
\alias{contents_html}
\alias{contents_markdown}
\title{Format contents into a textual representation}
\usage{
contents_text(content, ...)

contents_html(content, ...)

contents_markdown(content, ...)
}
\arguments{
\item{content}{The \link{Turn} or \link{Content} object to be converted into text.
\code{contents_markdown()} also accepts \link{Chat} instances to turn the entire
conversation history into markdown text.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A string of text, markdown or HTML.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

These generic functions can be use to convert \link{Turn} contents or \link{Content}
objects into textual representations.
\itemize{
\item \code{contents_text()} is the most minimal and only includes \link{ContentText}
objects in the output.
\item \code{contents_markdown()} returns the text content (which it assumes to be
markdown and does not convert it) plus markdown representations of images
and other content types.
\item \code{contents_html()} returns the text content, converted from markdown to
HTML with \code{\link[commonmark:markdown_html]{commonmark::markdown_html()}}, plus HTML representations of
images and other content types.
}

These content types will continue to grow and change as ellmer evolves to
support more providers and as providers add more content types.
}
\examples{
turns <- list(
  UserTurn(list(
    ContentText("What's this image?"),
    content_image_url("https://placehold.co/200x200")
  )),
  AssistantTurn("It's a placeholder image.")
)

lapply(turns, contents_text)
lapply(turns, contents_markdown)
if (rlang::is_installed("commonmark")) {
  contents_html(turns[[1]])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{df_schema}
\alias{df_schema}
\title{Describe the schema of a data frame, suitable for sending to an LLM}
\usage{
df_schema(df, max_cols = 50)
}
\arguments{
\item{df}{A data frame to describe.}

\item{max_cols}{Maximum number of columns to includes. Defaults to 50 to
avoid accidentally generating very large prompts.}
}
\description{
\code{df_schema()} gives a column-by-column description of a data frame. For
each column, it gives the name, type, label (if present), and number of
missing values. For numeric and date/time columns, it also gives the
range. For character and factor columns, it also gives the number of unique
values, and if there's only a few (<= 10), their values.

The goal is to give the LLM a sense of the structure of the data, so that
it can generate useful code, and the output attempts to balance between
conciseness and accuracy.
}
\examples{
df_schema(mtcars)
df_schema(iris)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provider-databricks.R
\name{chat_databricks}
\alias{chat_databricks}
\title{Chat with a model hosted on Databricks}
\usage{
chat_databricks(
  workspace = databricks_workspace(),
  system_prompt = NULL,
  model = NULL,
  token = NULL,
  params = NULL,
  api_args = list(),
  echo = c("none", "output", "all"),
  api_headers = character()
)
}
\arguments{
\item{workspace}{The URL of a Databricks workspace, e.g.
\code{"https://example.cloud.databricks.com"}. Will use the value of the
environment variable \code{DATABRICKS_HOST}, if set.}

\item{system_prompt}{A system prompt to set the behavior of the assistant.}

\item{model}{The model to use for the chat (defaults to "databricks-claude-3-7-sonnet").
We regularly update the default, so we strongly recommend explicitly specifying a model for anything other than casual use.

Available foundational models include:
\itemize{
\item \code{databricks-claude-3-7-sonnet} (the default)
\item \verb{databricks-mixtral-8x7b-instruct}
\item \verb{databricks-meta-llama-3-1-70b-instruct}
\item \verb{databricks-meta-llama-3-1-405b-instruct}
}}

\item{token}{An authentication token for the Databricks workspace, or
\code{NULL} to use ambient credentials.}

\item{params}{Common model parameters, usually created by \code{\link[=params]{params()}}.}

\item{api_args}{Named list of arbitrary extra arguments appended to the body
of every chat API call. Combined with the body object generated by ellmer
with \code{\link[=modifyList]{modifyList()}}.}

\item{echo}{One of the following options:
\itemize{
\item \code{none}: don't emit any output (default when running in a function).
\item \code{output}: echo text and tool-calling output as it streams in (default
when running at the console).
\item \code{all}: echo all input and output.
}

Note this only affects the \code{chat()} method.}

\item{api_headers}{Named character vector of arbitrary extra headers appended
to every chat API call.}
}
\value{
A \link{Chat} object.
}
\description{
Databricks provides out-of-the-box access to a number of \href{https://docs.databricks.com/en/machine-learning/model-serving/score-foundation-models.html}{foundation models}
and can also serve as a gateway for external models hosted by a third party.

Built on top of \code{\link[=chat_openai_compatible]{chat_openai_compatible()}}.
\subsection{Authentication}{

\code{chat_databricks()} picks up on ambient Databricks credentials for a subset
of the \href{https://docs.databricks.com/en/dev-tools/auth/unified-auth.html}{Databricks client unified authentication}
model. Specifically, it supports:
\itemize{
\item Personal access tokens
\item Service principals via OAuth (OAuth M2M)
\item User account via OAuth (OAuth U2M)
\item Authentication via the Databricks CLI
\item Posit Workbench-managed credentials
\item Viewer-based credentials on Posit Connect. Requires the \pkg{connectcreds}
package.
}
}
}
\examples{
\dontrun{
chat <- chat_databricks()
chat$chat("Tell me three jokes about statisticians")
}
}
\seealso{
Other chatbots:
\code{\link[=chat_anthropic]{chat_anthropic()}},
\code{\link[=chat_aws_bedrock]{chat_aws_bedrock()}},
\code{\link[=chat_azure_openai]{chat_azure_openai()}},
\code{\link[=chat_cloudflare]{chat_cloudflare()}},
\code{\link[=chat_deepseek]{chat_deepseek()}},
\code{\link[=chat_github]{chat_github()}},
\code{\link[=chat_google_gemini]{chat_google_gemini()}},
\code{\link[=chat_groq]{chat_groq()}},
\code{\link[=chat_huggingface]{chat_huggingface()}},
\code{\link[=chat_lmstudio]{chat_lmstudio()}},
\code{\link[=chat_mistral]{chat_mistral()}},
\code{\link[=chat_ollama]{chat_ollama()}},
\code{\link[=chat_openai]{chat_openai()}},
\code{\link[=chat_openai_compatible]{chat_openai_compatible()}},
\code{\link[=chat_openrouter]{chat_openrouter()}},
\code{\link[=chat_perplexity]{chat_perplexity()}},
\code{\link[=chat_portkey]{chat_portkey()}}
}
\concept{chatbots}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/types.R
\name{type_boolean}
\alias{type_boolean}
\alias{type_integer}
\alias{type_number}
\alias{type_string}
\alias{type_enum}
\alias{type_array}
\alias{type_object}
\alias{type_from_schema}
\alias{type_ignore}
\title{Type specifications}
\usage{
type_boolean(description = NULL, required = TRUE)

type_integer(description = NULL, required = TRUE)

type_number(description = NULL, required = TRUE)

type_string(description = NULL, required = TRUE)

type_enum(values, description = NULL, required = TRUE)

type_array(items, description = NULL, required = TRUE)

type_object(
  .description = NULL,
  ...,
  .required = TRUE,
  .additional_properties = FALSE
)

type_from_schema(text, path)

type_ignore()
}
\arguments{
\item{description, .description}{The purpose of the component. This is
used by the LLM to determine what values to pass to the tool or what
values to extract in the structured data, so the more detail that you can
provide here, the better.}

\item{required, .required}{Is the component or argument required?

In type descriptions for structured data, if \code{required = FALSE} and the
component does not exist in the data, the LLM may hallucinate a value. Only
applies when the element is nested inside of a \code{type_object()}.

In tool definitions, \code{required = TRUE} signals that the LLM should always
provide a value. Arguments with \code{required = FALSE} should have a default
value in the tool function's definition. If the LLM does not provide a
value, the default value will be used.}

\item{values}{Character vector of permitted values.}

\item{items}{The type of the array items. Can be created by any of the
\code{type_} function.}

\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}> Name-type pairs defining
the components that the object must possess.}

\item{.additional_properties}{Can the object have arbitrary additional
properties that are not explicitly listed? Only supported by Claude.}

\item{text}{A JSON string.}

\item{path}{A file path to a JSON file.}
}
\description{
These functions specify object types in a way that chatbots understand and
are used for tool calling and structured data extraction. Their names are
based on the \href{https://json-schema.org}{JSON schema}, which is what the APIs
expect behind the scenes. The translation from R concepts to these types is
fairly straightforward.
\itemize{
\item \code{type_boolean()}, \code{type_integer()}, \code{type_number()}, and \code{type_string()}
each represent scalars. These are equivalent to length-1 logical,
integer, double, and character vectors (respectively).
\item \code{type_enum()} is equivalent to a length-1 factor; it is a string that can
only take the specified values.
\item \code{type_array()} is equivalent to a vector in R. You can use it to represent
an atomic vector: e.g. \code{type_array(type_boolean())} is equivalent
to a logical vector and \code{type_array(type_string())} is equivalent
to a character vector). You can also use it to represent a list of more
complicated types where every element is the same type (R has no base
equivalent to this), e.g. \code{type_array(type_array(type_string()))}
represents a list of character vectors.
\item \code{type_object()} is equivalent to a named list in R, but where every element
must have the specified type. For example,
\code{type_object(a = type_string(), b = type_array(type_integer()))} is
equivalent to a list with an element called \code{a} that is a string and
an element called \code{b} that is an integer vector.
\item \code{type_ignore()} is used in tool calling to indicate that an argument should
not be provided by the LLM. This is useful when the R function has a
default value for the argument and you don't want the LLM to supply it.
\item \code{type_from_schema()} allows you to specify the full schema that you want to
get back from the LLM as a JSON schema. This is useful if you have a
pre-defined schema that you want to use directly without manually creating
the type using the \verb{type_*()} functions. You can point to a file with the
\code{path} argument or provide a JSON string with \code{text}. The schema must be a
valid JSON schema object.
}
}
\examples{
# An integer vector
type_array(type_integer())

# The closest equivalent to a data frame is an array of objects
type_array(type_object(
   x = type_boolean(),
   y = type_string(),
   z = type_number()
))

# There's no specific type for dates, but you use a string with the
# requested format in the description (it's not guaranteed that you'll
# get this format back, but you should most of the time)
type_string("The creation date, in YYYY-MM-DD format.")
type_string("The update date, in dd/mm/yyyy format.")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream-controller.R
\name{stream_controller}
\alias{stream_controller}
\title{Create a stream controller}
\usage{
stream_controller()
}
\value{
An \code{ellmer_stream_controller} object with the following
elements:
\itemize{
\item \verb{$cancel(reason = "cancelled")}: Cancel the stream. The \code{reason}
string is stored on the controller and used as the
\link[=Turn]{AssistantPartialTurn}'s \code{reason} property.
\item \verb{$reset()}: Clear the cancelled state and reason.
\item \verb{$cancelled}: A logical flag indicating whether the controller
has been cancelled.
\item \verb{$reason}: The cancellation reason string, or \code{NULL} if not
cancelled.
}
}
\description{
Creates a controller that can cancel an in-progress stream. Pass it to
\link{Chat}'s \verb{$stream()} or \verb{$stream_async()} via the \code{controller} argument,
then call \verb{$cancel()} from anywhere (e.g. a Shiny observer) to stop the
stream after the next chunk arrives.

The same controller can be reused across multiple streams. Call
\verb{$reset()} to clear the cancelled state, or pass it directly to a new
\verb{$stream()} call — it will be reset automatically.
}
\section{Async cancellation in Shiny}{


In a Shiny app, use an \link[shiny:ExtendedTask]{ExtendedTask} for
non-blocking chat and a \code{stream_controller()} to wire up a cancel
button:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{controller <- stream_controller()

chat_task <- ExtendedTask$new(function(user_query, controller = NULL) \{
  chat <- chat_openai(model = "gpt-4.1-nano")
  stream <- chat$stream_async(user_query, controller = controller)
  shinychat::markdown_stream("response", stream)
\})

observeEvent(input$ask, \{
  controller <<- stream_controller()
  chat_task$invoke(input$query, controller = controller)
\})

observeEvent(input$cancel, \{
  controller$cancel()
\})
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
chat <- chat_openai(model = "gpt-5.4-nano")

ctrl <- stream_controller()
stream <- chat$stream("Write a short story.", controller = ctrl)

i <- 0
coro::loop(for (chunk in stream) {
  i <- i + 1
  if (i > 10) ctrl$cancel()
})

chat
\dontshow{\}) # examplesIf}
}
